Package: propertee
Version: 1.0.6
Title: Standardization-Based Effect Estimation with Optional Prior
        Covariance Adjustment
Description: The Prognostic Regression Offsets with Propagation of
    ERrors (for Treatment Effect Estimation) package facilitates
    direct adjustment for experiments and observational studies that
    is compatible with a range of study designs and covariance
    adjustment strategies. It uses explicit specification of clusters,
    blocks and treatment allocations to furnish probability of
    assignment-based weights targeting any of several average
    treatment effect parameters, and for standard error calculations
    reflecting these design parameters. For covariance adjustment of
    its Hajek and (one-way) fixed effects estimates, it enables
    offsetting the outcome against predictions from a dedicated
    covariance model, with standard error calculations propagating
    error as appropriate from the covariance model.
Authors@R: c(person("Josh", "Errickson", role = "aut",
                    email = "jerrick@umich.edu"),
             person("Josh", "Wasserman", role = c("cre", "aut"),
                    email = "jwass@umich.edu"),
             person("Mark", "Fredrickson", role = "ctb",
                    email = "mark.m.fredrickson@gmail.com"),
             person("Adam", "Sales", role = "ctb",
                    email = "asales@wpi.edu"),
             person("Xinhe", "Wang", role = "ctb",
                    email = "xinhew@umich.edu"),
             person("Ben", "Hansen", role = "aut",
                    email = "ben.hansen@umich.edu"))
License: MIT + file LICENSE
License_is_FOSS: yes
License_restricts_use: no
Encoding: UTF-8
LazyData: true
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), multcomp, MASS
Config/testthat/edition: 3
Imports: stats, methods, sandwich
Enhances: robustbase
Depends: R (>= 4.1.0)
VignetteBuilder: knitr
URL: https://github.com/benbhansen-stats/propertee
BugReports: https://github.com/benbhansen-stats/propertee/issues
Collate: 'StudySpecification.R' 'StudySpecificationAccessors.R'
        'SandwichLayer.R' 'block_center_residuals.R'
        'SandwichLayerVariance.R' 'StudySpecificationConverters.R'
        'StudySpecificationStructure.R' 'StudySpecificationUtilities.R'
        'WeightedStudySpecification.R' 'areg.center.R' 'confint_lm.R'
        'teeMod.R' 'as.lmitt.R' 'as_data_frame.R' 'assigned.R'
        'bread.R' 'c_weightedStudySpecification.R' 'cov_adj.R' 'data.R'
        'dichotomy.R' 'expand.model.frame_tee.R' 'get_cov_adj.R'
        'get_data_from_model.R' 'get_spec.R' 'glmrobMethods.R'
        'lmitt.R' 'lmrob_methods.R' 'merge_preserve_order.R'
        'propertee_package.R' 'specification_table.R'
        'summary.StudySpecification.R' 'summary.teeMod.R' 'update_by.R'
        'validWeights.R' 'weights_internal.R' 'weights_exported.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-19 17:29:27 UTC; jwass
Author: Josh Errickson [aut],
  Josh Wasserman [cre, aut],
  Mark Fredrickson [ctb],
  Adam Sales [ctb],
  Xinhe Wang [ctb],
  Ben Hansen [aut]
Maintainer: Josh Wasserman <jwass@umich.edu>
Repository: CRAN
Date/Publication: 2026-07-20 12:20:09 UTC

# Generated by roxygen2: do not edit by hand

S3method(bread,glmrob)
S3method(bread,lmrob)
S3method(bread,teeMod)
S3method(confint,teeMod)
S3method(estfun,glmrob)
S3method(estfun,lmrob)
S3method(estfun,teeMod)
S3method(getCall,teeMod)
S3method(lmitt,formula)
S3method(lmitt,lm)
S3method(print,summary.StudySpecification)
S3method(print,summary.teeMod)
S3method(summary,StudySpecification)
S3method(summary,teeMod)
S3method(update,teeMod)
S3method(vcov,teeMod)
export("blocks<-")
export("clusters<-")
export("forcings<-")
export("treatment<-")
export("unitids<-")
export("units_of_assignment<-")
export(a.)
export(adopters)
export(as.SandwichLayer)
export(as.lmitt)
export(as.teeMod)
export(as_obs_spec)
export(as_rct_spec)
export(as_rd_spec)
export(assigned)
export(atc)
export(ate)
export(ato)
export(att)
export(block)
export(blocks)
export(cluster)
export(clusters)
export(cov_adj)
export(default)
export(etc)
export(ett)
export(forcing)
export(forcings)
export(get_structure)
export(has_binary_treatment)
export(has_blocks)
export(identical_StudySpecifications)
export(identify_small_blocks)
export(lmitt)
export(obs_spec)
export(obs_specification)
export(obsstudy_spec)
export(obsstudy_specification)
export(olw)
export(owt)
export(pwt)
export(rct_spec)
export(rct_specification)
export(rd_spec)
export(rd_specification)
export(specification_data_concordance)
export(specification_table)
export(stable)
export(treatment)
export(unit_of_assignment)
export(unitid)
export(unitids)
export(units_of_assignment)
export(uoa)
export(var_names)
export(var_table)
export(vcov_tee)
export(z.)
exportMethods("*")
exportMethods("+")
exportMethods("-")
exportMethods("/")
exportMethods("[")
exportMethods("blocks<-")
exportMethods("clusters<-")
exportMethods("forcings<-")
exportMethods("treatment<-")
exportMethods("unitids<-")
exportMethods("units_of_assignment<-")
exportMethods(blocks)
exportMethods(c)
exportMethods(clusters)
exportMethods(forcings)
exportMethods(show)
exportMethods(subset)
exportMethods(treatment)
exportMethods(unitids)
exportMethods(units_of_assignment)
exportMethods(weights)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,validObject)
importFrom(sandwich,bread)
importFrom(sandwich,estfun)
importFrom(stats,aggregate)
importFrom(stats,as.formula)
importFrom(stats,complete.cases)
importFrom(stats,confint)
importFrom(stats,cov)
importFrom(stats,family)
importFrom(stats,formula)
importFrom(stats,getCall)
importFrom(stats,lm)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,na.action)
importFrom(stats,na.pass)
importFrom(stats,predict)
importFrom(stats,printCoefmat)
importFrom(stats,pt)
importFrom(stats,reformulate)
importFrom(stats,residuals)
importFrom(stats,summary.lm)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(stats,weighted.mean)
importFrom(stats,weights)
importFrom(utils,capture.output)
importFrom(utils,getFromNamespace)
importFrom(utils,getS3method)
importFrom(utils,methods)

## ----installPropertee, eval=FALSE,include=FALSE-------------------------------
# ### this is for Adam, to remember how to do this each time
# ### LMK if it causes problems, I'll remove
# devtools::install_github('benbhansen-stats/propertee', auth_token=GITHUB_PAT)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(propertee)

## ----loadData-----------------------------------------------------------------
data(lsoSynth)

## ----rddPlots,fig.width=7-----------------------------------------------------

figDat <- aggregate(lsoSynth[,c('nextGPA','lhsgrade_pct')],by=list(R=lsoSynth$R),
                    FUN=mean,na.rm=TRUE)
figDat$n <- as.vector(table(lsoSynth$R))

with(figDat,
     plot(R,nextGPA,cex=log(n+1)/3,main='Average Subsequent GPA\n as a function of 1st-Year GPA,\n Centered at AP Cutoff'))
abline(v=0,lty=2)

## ----defineUnitID-------------------------------------------------------------
lsoSynth$id <- rownames(lsoSynth)

## ----rdDesign-----------------------------------------------------------------
lsoSynth$Z <- lsoSynth$R<0

lsoSynthW <- subset(lsoSynth,abs(R)<=0.5)

#lsoSynthW$id <- 1:nrow(lsoSynthW)
spec <- rd_spec(Z ~ forcing(R) + unitid(id), data=lsoSynth, subset=abs(lsoSynth$R) <= 0.5)

## ----olsGmod1, eval=FALSE,include=TRUE----------------------------------------
# ### this doesn't work:
# g1 <- lm(nextGPA ~ R + Z, data = lsoSynth, weights = ett(spec)

## ----olsGmod2-----------------------------------------------------------------
##this works, but it's annoying to enter in the subset expression a 2nd time:
g1 <- lm(nextGPA ~ R + Z, data = lsoSynth, subset = abs(R) <= 0.5)


## ----robGmod------------------------------------------------------------------
g2 <-
if(requireNamespace("robustbase", quietly = TRUE)){
  robustbase::lmrob(nextGPA~poly(R,5)+Z,data=lsoSynthW)
} else g1

## ----yhat---------------------------------------------------------------------

yhat1 <- predict(g1,data.frame(R=forcings(spec)[[1]],Z=FALSE))
yhat2 <- predict(g2,data.frame(R=forcings(spec)[[1]],Z=FALSE))

plot(yhat1,yhat2)

## ----estEff1------------------------------------------------------------------
### method 1:

mean(lsoSynthW$nextGPA[lsoSynthW$Z]-yhat1[lsoSynthW$Z])-
  mean(lsoSynthW$nextGPA[!lsoSynthW$Z]-yhat1[!lsoSynthW$Z])

#### method 2:
coef(lm(nextGPA~Z, offset=yhat1,data=lsoSynthW))['ZTRUE']



## ----estEff2------------------------------------------------------------------
### method 1:

mean(lsoSynthW$nextGPA[lsoSynthW$Z]-yhat2[lsoSynthW$Z])-
  mean(lsoSynthW$nextGPA[!lsoSynthW$Z]-yhat2[!lsoSynthW$Z])

#### method 2:
coef(lm(nextGPA~Z, offset=yhat2,data=lsoSynthW))['ZTRUE']

data("STARplus")

##' A prognostic model fitted to experimental + non-experimental controls
y0hat_read <- lm(read_yr1 ~ gender*dob +dobNA + race,
                 data = STARplus,
                 subset = cond_at_entry!="small")

STARspec <- rct_spec(cond_at_entry ~ unit_of_assignment(stdntid) +
                         block(grade_at_entry, school_at_entry),
                     subset=!is.na(grade_at_entry),# excludes non-experimentals
                     data = STARplus)
ett_wts    <- ett(STARspec, data = STARplus,
                  dichotomy= cond_at_entry =="small" ~.)

ett_read <- lm(read_yr1 ~ assigned(dichotomy= cond_at_entry =="small" ~.),
               offset = cov_adj(y0hat_read),
               data = STARplus,
               weights = ett_wts)
coef(ett_read)
ett_read |> as.lmitt() # brings in control-group means of outcome, predictions

ate_read <- lmitt(read_yr1 ~ 1, STARspec, STARplus,
                  dichotomy= cond_at_entry =="small" ~.,
                  offset = cov_adj(y0hat_read),
                  weights = "ate")
show(ate_read)
vcov(ate_read, type = "HC0", cov_adj_rcorrect = "HC0") |> unname()

ate_read_loc <-
    lmitt(read_yr1 ~ race, STARspec, STARplus,
          dichotomy= cond_at_entry =="small" ~.,
          offset = cov_adj(y0hat_read, newdata = STARplus),
          weights = "ate")
show(ate_read_loc)

##' @title Table of elements from a \code{StudySpecification}
##'
##' @description Produces a table (1-dimensional, or 2-dimensional if \code{y}
##'   is specified) of the elements of the \code{StudySpecification}.
##'
##' @param specification A \code{StudySpecification} object
##' @param x One of "treatment", "unit of assignment", (synonym "uoa"), "block".
##'   Abbreviations are accepted. "unit of assignment" can be replaced by
##'   "unitid" or "cluster" if the \code{StudySpecification} was created with
##'   that element.
##' @param y Optionally, another string similar to \code{x}. A 1-dimensional
##'   table is produced if \code{y} is left at its default, \code{NULL}.
##' @param sort Ignored if \code{y} is not \code{NULL}. If \code{FALSE}
##'   (default), one-way table is sorted according to "names" of levels. If set
##'   to \code{TRUE}, one-way table is sorted according to values.
##' @param decreasing If \code{sort} is \code{TRUE}, choose whether to sort
##'   descending (\code{TRUE}, default) or ascending (\code{FALSE}).
##' @param use_var_names If \code{TRUE}, name dimensions of table returned by
##'   variable names. If \code{FALSE} (default), name by their function (e.g.
##'   "treatment" or "blocks"). Passing the \code{dnn} argument in \code{...}
##'   (an argument of [table()]) overrides whatever is requested here.
##' @param ... additional arguments [table()]
##' @return A table of the requested variables.
##' @export
##' @rdname specification_table
##' @order 2
##' @examples
##' data(simdata)
##' spec <- obs_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid),
##'                   data = simdata)
##' specification_table(spec, "treatment")
##' specification_table(spec, "treatment", "block", sort = TRUE, use_var_names = TRUE)
stable <- function(specification,
                   x,
                   y = NULL,
                   sort = FALSE,
                   decreasing = TRUE,
                   use_var_names = FALSE,
                   ...) {

  # Internal function to match partial names and standardize across uoa/unit of
  # assignment
  .fix_xy <- function(z, which) {
    if (!is.null(z)) {
      z <- tolower(z)
      if (grepl(paste0("^", z), "treatments")) {
        z <- "treatment"
      } else if (grepl(paste0("^", z), "blocks")) {
        z <- "blocks"
      } else if (grepl(paste0("^", z), "uoa") ||
                 grepl(paste0("^", z), "unit of assignment") ||
                 grepl(paste0("^", z), "units of assignment")) {
        z <- "units_of_assignment"
      } else if (grepl(paste0("^", z), "clusters")) {
        z <- "clusters"
      } else {
        stop(paste("Invalid input", which))
      }
    }
    return(z)
  }

  x <- .fix_xy(x, "x")
  y <- .fix_xy(y, "y")

  # Internal function to obtain proper data from the StudySpecification structure, and if
  # its multi-dimensional (e.g. 2 variables uniquely identify block) collapse
  # appropriately.
  .get_xydat <- function(z, specification) {
    if (!is.null(z)) {

      # Since .fix_xy ensures that `x` and `y` contain valid function names,
      # this executes them.
      zdat <- match.fun(z)(specification)

      if (ncol(zdat) == 0) {
        # This is hit if user request an element not in the specification (e.g.
        # requests block, but specification is created without blocks
        warning(paste(z, "not found in specification"))
        return(NULL)
      }

      # If we have multiple variables in the specification definition, collapse them
      if (ncol(zdat) > 1) {
        zdat <- apply(zdat, 1, paste, collapse = ",")
      } else {
        # If we have a single variable, drop second dimension
        zdat <- zdat[, 1]
      }
      return(zdat)
    } else {
      return(NULL)
    }
  }

  xdat <- .get_xydat(x, specification)
  ydat <- .get_xydat(y, specification)

  # If user passes in `dnn`, we need to not pass our own when creating table
  # below.
  dots_have_dnn <- "dnn" %in% names(list(...))

  # If x or y were asking for units of assignment, `x` or `y` are now
  # "units_of_assignment". Replace "_" with " ". Does nothing in other cases.
  x <- gsub("_", " ", x)
  y <- gsub("_", " ", y)


  if (is.null(ydat) | is.null(xdat)) {
    if (is.null(xdat)) {
      # We're here if the user requsted both an x and a y, but the x is invalid
      # (e.g. specification was created without block, but user requested blocks). In
      # this case, it should be equivalent to a user only requesting x, so move
      # y into x. (Note that warning to use arises earlier in `.get_xydat`.)
      x <- y
      xdat <- ydat
    }
    # 1-dimensional table

    if (dots_have_dnn) {
      outtable <- table(xdat, ...)
    } else {
      if (use_var_names) {
        x <- paste(var_names(specification, substr(x, 1, 1)), collapse = ", ")
      }
      outtable <- table(xdat, dnn = list(x), ...)
    }

    if (sort) {
      outtable <- sort(outtable, decreasing = decreasing)
    }
  } else {
    # 2-dimensional table

    if (dots_have_dnn) {
      outtable <- table(ydat, xdat, ...)
    } else {
      if (use_var_names) {
        x <- paste(var_names(specification, substr(x, 1, 1)))
        y <- paste(var_names(specification, substr(y, 1, 1)))
      }
      outtable <- table(ydat, xdat, dnn = list(y, x), ...)
    }

  }

  return(outtable)


}

##' @export
##' @rdname specification_table
##' @order 1
specification_table <- stable

test_that("1-way tables", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  tt <- specification_table(spec, "treatment")

  expect_s3_class(tt, "table")
  expect_length(tt, 2)
  expect_identical(as.character(unique(spec@structure$z)), names(tt))

  expect_identical(tt, specification_table(spec, "t"))
  expect_identical(tt, specification_table(spec, "Treat"))

  bt <- specification_table(spec, "block")

  expect_s3_class(bt, "table")
  expect_length(bt, 3)
  expect_identical(as.character(unique(spec@structure$bid)), names(bt))

  expect_identical(bt, specification_table(spec, "bl"))
  expect_identical(bt, specification_table(spec, "Blo"))

  ut <- specification_table(spec, "units of assignment")

  expect_s3_class(ut, "table")
  expect_length(dim(ut), 1)
  expect_length(ut, 10)

  expect_identical(ut, specification_table(spec, "unit of as"))
  expect_identical(ut, specification_table(spec, "un"))
  expect_identical(ut, specification_table(spec, "UnitS"))
  expect_identical(ut, specification_table(spec, "UO"))

  # Missing element
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  expect_warning(bt <- specification_table(spec, "block"), "blocks not found")
  expect_s3_class(ut, "table")
  expect_equal(dim(bt), 0)

  expect_error(stable(spec, "unit"), "StudySpecification specified with")


# cluster instead of uoa
  ct <- specification_table(spec, "clusters")

  expect_s3_class(ct, "table")
  expect_length(dim(ct), 1)
  expect_length(ct, 10)

  expect_identical(ct, specification_table(spec, "cl"))
  expect_identical(ct, specification_table(spec, "cL"))


  # sortings
  simdata$bid <- rep(1:3, times = c(10, 30, 10))
  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  bt <- specification_table(spec, "block")

  expect_true(!all(bt == sort(bt, decreasing = TRUE)))
  expect_true(all(names(bt) == sort(names(bt))))

  bt2 <- specification_table(spec, "block", sort = TRUE)
  expect_true(!all(bt == bt2))
  expect_true(all(bt2 == sort(bt2, decreasing = TRUE)))
  expect_true(!all(names(bt2) == sort(names(bt2))))

  bt3 <- specification_table(spec, "block", sort = TRUE, decreasing = FALSE)
  expect_true(!all(bt == bt3))
  expect_true(all(bt3 == sort(bt3, decreasing = FALSE)))
  expect_true(!all(names(bt3) == sort(names(bt3))))


  # use_var_names

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  expect_equal(names(dimnames(specification_table(spec, "t")))[1], "treatment")
  expect_equal(names(dimnames(specification_table(spec, "b")))[1], "blocks")
  expect_equal(names(dimnames(specification_table(spec, "u")))[1], "units of assignment")

  expect_equal(names(dimnames(specification_table(spec, "t", use_var_names = TRUE)))[1],
               "z")
  expect_equal(names(dimnames(specification_table(spec, "b", use_var_names = TRUE)))[1],
               "bid")
  expect_equal(names(dimnames(specification_table(spec, "u", use_var_names = TRUE)))[1],
               "uoa1, uoa2")

  # other arguments

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  expect_length(specification_table(spec, "t", useNA = "always"), 3)

  expect_equal(names(dimnames(specification_table(spec, "t", dnn = list("abc")))),
               "abc")

})

test_that("2 way tables", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)


  tbl <- stable(spec, "trea", "bl")
  expect_s3_class(tbl, "table")
  expect_true(all.equal(dim(tbl), c(3, 2)))

  tbl2 <- stable(spec, "bl", "treat")

  expect_identical(tbl, t(tbl2))

  # Missing entry
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)

  expect_warning(tbl <- stable(spec, "trea", "bl"), "blocks not found")
  expect_identical(tbl, stable(spec, "tr"))

  # var names
  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  tbl <- stable(spec, "trea", "bl")
  expect_equal(names(dimnames(tbl)), c("blocks", "treatment"))
  tbl <- stable(spec, "trea", "bl", use_var_names = TRUE)
  expect_equal(names(dimnames(tbl)), c("bid", "z"))

})

test_that("Passing dnn", {
  data(simdata)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  tbl <- stable(spec, "trea", "bl", dnn = c("a", "b"))
  expect_identical(names(attr(tbl, "dimnames")),
                   c("a", "b"))



})

##' Whenever a function in a model
##' (\code{ate()}/\code{ett()}/\code{cov_adj()}/\code{assigned()}) is called
##' without an explicit \code{data=} argument, this will attempt to extract the
##' data from the model itself.
##'
##' The \code{form} specifies what columns of the data are needed. For current
##' use cases (\code{ate()}/\code{ett()} and \code{assigned()}), this will be
##' only the unit of assignment variables, so e.g. \code{form = ~ uoavar}, to
##' enable merging of UOA level variables to the model data. However, this can
##' easily be expanded if other variables are needed.
##' @title (Internal) Locate data in call stack
##' @param which_fn Identify calling function, "weights" or "assigned", helps
##'   separate logic for the two functions.
##' @param form Formula on which to apply \code{model.frame()}. See details
##' @param by translation of unit of assignment/unitid/cluster ID names, passed
##'   down from weights.
##' @return \code{data.frame}
##' @keywords internal
.get_data_from_model <- function(which_fn,
                                 form = NULL,
                                 by = NULL) {

  if (!which_fn %in% c("weights", "cov_adj", "assigned")) {
    stop(paste("Internal error: which_fn is invalid,", which_fn))
  }

  # Formula should be from a StudySpecification; it is used inside `model.frame` below
  if (!(is.call(form) | inherits(form, "formula") | is.name(form))) {
    stop("internal error: form must be a formula or name")
  }

  # Evaluate as needed.
  if (is.name(form)) {
    # dynGet searches iteratively backwards in the
    # callstack. Per the documentation, it is experimental
    # and should be used with caution. We should only be hitting
    # this if the user passes a predefined formula, e.g.:
    # f <- y ~ x; rct_spec(f, ...)
    form <- dynGet(form)
  } else if (is.call(form)) {
    form <- as.formula(form)
  }
  # By this point we should have hit an earlier error
  if (!inherits(form, "formula")) {
    stop("internal error: unable to convert form to formula class")
  }

  if (!is.null(by)) {
    .check_by(by)
    # if we're passed by, need to update the formula.

    # First, convert passed-as-string new variable names to `name` objects
    # to avoid quotation (y ~ "x")` in formula
    by <- lapply(by, as.name)

    # Next use substitute to do the replacing.
    # Cannot just call `substitute(form, by)` since we want to pass
    # the actual formula object, not just the name "form"
    form <- as.formula(do.call("substitute", list(form, by)))
  }

  data <- NULL

  # Obtain the names of all functions in the callstack
  fns_called <- as.character(lapply(sys.calls(), `[[`, 1))

  # Identify all frames which have `model.frame.default` called (or
  # `lmitt`)
  mf_pos <- which(fns_called %in% c("model.frame.default", "lmitt.formula"))

  # identify whether we're looking inside weights or assigned
  if (which_fn == "weights") {
    # Find all frames with `weights` argument

    weights_args <- lapply(sys.calls(), `[[`, "weights")

    # Within each `weights` argument, we're looking for the string "ate(" or
    # "ett(", ensuring that the previous character (if there is one: ?) is not a
    # letter, thus ensuring we don't have some oddity like `translate(x)`
    # triggering.
    pos <- vapply(weights_args, function(aa) {
      is.call(aa) && (aa[[1]] == quote(ate) || aa[[1]] == quote(ett) ||
                        any(vapply(aa, function(x) {
                         x == quote(ate()) || x == quote(ett())
                        }, logical(1))))
    },   logical(1))

    fn_pos <- which(pos)
  } else if (which_fn == "cov_adj") {
    offset_args <- lapply(sys.calls(), `[[`, "offset")

    pos <- vapply(offset_args, function(aa) {
      is.call(aa) && (aa[[1]] == quote(cov_adj) ||
                        any(vapply(aa, function(x) {
                         x == quote(cov_adj)
                        }, logical(1))))
    },  logical(1))

    fn_pos <- which(pos)
  } else if (which_fn == "assigned") {
    # Find all frames with `formula`
    adopter_args <- lapply(sys.frames(), function(x)
      tryCatch(get("formula", envir = x, inherits = FALSE),
               error = function(e) NULL))

    # Ensure only terms or formula are found #124
    adopter_args <- lapply(adopter_args, function(x) {
      if (!(is(x, "formula") | is(x, "terms"))) {
        return(NULL)
      } else {
        return(x)
      }
    })

    # Search these formula for "assigned(", ensuring that the previous character
    # is not a letter, in case there are functions like customassigned().
    pos <- vapply(adopter_args, function(x) {
      any(c("assigned", "z.", "a.", "adopters") %in%
            all.names(as.formula(x)))
    },
    logical(1))

    fn_pos <- which(pos)
  }


  if (length(mf_pos) == 0) {
    # If no model.frames were identified
    warning(paste0("No call to `model.frame` with StudySpecification weights in the ",
                   "call stack found."))
  } else {
    # We've identified at least one model.frame.default
    if (length(mf_pos) > 1) {
      # If we have more than one, pick the one that has weights/assigned that is
      # lowest in the call-stack. (Pick lowest to avoid any issues where there
      # is pre-processing done at an in-bteween step.)
      mf_pos <- max(mf_pos[mf_pos %in% fn_pos])
    }

    # Regenerate the model.frame with the appropriate data
    environment(form) <-
      environment(get("formula", sys.frame(mf_pos)))

    # Try to get the data from the appropriate frame
    passed_data <- tryCatch(get("data", sys.frame(mf_pos)))

    ### #193 - There's no harm adding this every time, as if it's not needed,
    ### it'll be dropped in the model.frame call below
    passed_data$..uoa.. <- rownames(passed_data)

    if (is.null(passed_data)) {
      # If we can't get the data directly, try getting it via name
      passed_data <- tryCatch(eval(sys.call(mf_pos)$data,
                                   sys.frame(mf_pos)),
                              error = function(e) NULL)
    }
    try(data <- model.frame(form,
                            data = passed_data,
                            na.action = na.pass),
        silent = TRUE)
  }

  # search for a teeMod object in the call stack
  if (is.null(data) || !is.data.frame(data)) {
    for (i in seq_len(sys.nframe())) {
      # recursion stops at the rlang_trace_top_env option if it's been set (i.e.
      # in test_that), otherwise where the call stack ends
      if (identical(parent.frame(i), getOption("rlang_trace_top_env")) |
          is.null(sys.call(-i))) break

      objs <- sapply(ls(parent.frame(i)), get, envir = parent.frame(i))
      lmitts <- vapply(objs, inherits, logical(1), "teeMod")
      if (any(lmitts)) {
        found_lmitt <- objs[[which(lmitts)]]
        try(data <- get("data", envir = environment(formula(found_lmitt))),
            silent = TRUE)
        break
      }
    }
  }

  # Ensure data is actually data at this point - if not, try fallback method
  if (is.null(data) || !is.data.frame(data)) {
    warning(paste("Unable to detect data by normal means,",
                  "trying fallback method to obtain data"))
    data <- .fallback_data_search()
  }

  # Finally if even fallback fails, exit
  if (is.null(data) || !is.data.frame(data)) {
    stop("Could not determine appropriate data")
  }

  return(data)
}


##' We try to be intelligent about finding the appropriate data. If this fails,
##' we may have need for a brute force method that just loops through frames and
##' looks for a `data` object.
##' @title (Internal) Fallback brute force method to locate \code{data} in the
##'   call stack.
##' @return If found, the data.
##' @keywords internal
.fallback_data_search <- function() {
  for (i in seq(1, sys.nframe())) {
    data <- mget("data", envir = sys.frame(i), ifnotfound = list(NULL))$data
    if (!is.null(data) && is.data.frame(data)) {
      break()
    }
  }
  if (is.null(data) || !is.data.frame(data)) {
    stop("Could not determine appropriate data")
  }
  return(data)
}

##' @title Convert object to \code{data.frame} or produce meaningful error
##' @param x An object
##' @return \code{x} as a \code{data.frame}
##' @keywords internal
.as_data_frame <- function(x) {
  tryCatch( {
    result <- as.data.frame(x)
    return(result)
  },
  error = function(e) {
    msg <- paste("Failed to convert to data frame:",
                 conditionMessage(e))
    stop(msg, call. = FALSE)
  },
  warning = function(w) {
    msg <- paste("Warning occurred while converting to data frame:",
                 conditionMessage(w))
    warning(msg, call. = FALSE)
    return(as.data.frame(x))
  }
  )
}

##' Helper function used to update the variable names of a
##' \code{StudySpecification} when user passes a \code{by=} argument to align
##' variable names between data sets.
##' @title (Internal) Use \code{by} to update \code{StudySpecification} with new
##'   variable names
##' @param specification A \code{StudySpecification}
##' @param data \code{Data set}
##' @param by named vector or list connecting names of unit of
##'   assignment/unitid/cluster variables in \code{specification} to unit of
##'   assignment/unitid/cluster variables in \code{data}. Names represent
##'   variables in the StudySpecification; values represent variables in the
##'   data.
##' @return A \code{StudySpecification} with updated variable names
##' @keywords internal
.update_by <- function(specification, data, by) {
  .check_by(by)

  # Ensure all names and replacements are valid
  missingnames <- !(names(by) %in% colnames(specification@structure))
  if (any(missingnames)) {
    warning(paste("by labels not found in StudySpecification. unknown elements:",
                  paste(names(by)[missingnames], collapse = ", ")))
  }
  missingdata <-  !(by %in% colnames(data))
  if (any(missingdata)) {
    warning(paste("by replacement values not found in data. unknown elements:",
                  paste(by[missingnames], collapse = ", ")))
  }

  # if we have any names or replacements missing in the specification or data,
  # there's a warning, and then don't try to replace that element
  by <- by[!missingnames & !missingdata]

  newnames <- vapply(colnames(specification@structure), function(x) {
    pos <- names(by) == x
    if (any(pos)) {
      return(by[[which(pos)]])
    }
    return(x)
  }, "character")

  colnames(specification@structure) <- newnames
  validObject(specification)
  return(specification)
}

##' Thie ensures that the \code{by=} argument is of the proper type, is named,
##' and consists of only unique entries.
##'
##' @title (Internal) A few checks to ensure \code{by=} is valid
##' @param by named vector or list connecting names of unit of
##'   assignment/unitid/cluster variables in \code{specification} to unit of
##'   assignment/unitid/cluster variables in \code{data}. Names represent
##'   variables in the StudySpecification; values represent variables in the
##'   data.
##' @return \code{NULL} if no errors are found
##' @keywords internal
.check_by <- function(by) {
  if (!(is.vector(by) || is.list(by)) ||
        is.null(names(by)) ||
        any(names(by) == "")) {
    stop("'by' must be named vector or named list")
  }
  if (any(duplicated(names(by))) || any(duplicated(by))) {
    stop("all entries in 'by' must be unique")
  }
  invisible(NULL)
}

#' @include validWeights.R
NULL


##' Called from \code{ate()} or \code{ett()}.
##' @title (Internal) Worker function for weight calculation
##' @param specification a \code{StudySpecification} object created by one of
##'   \code{rct_spec()}, \code{rd_spec()}, or \code{obs_spec()}.
##' @param target One of "ate" or "ett"; \code{ate()} and \code{ett()} chooses
##'   these automatically.
##' @param weightAlias An alias for the weight target, currently one of "ate",
##'   "ett", "att". Chosen by \code{ate()} and \code{ett()} automatically.
##' @param dichotomy optional; a formula defining the dichotomy of the treatment
##'   variable if it isn't already \code{0}/\code{1}. See details of help for
##'   \code{ate()} or \code{ett()} e.g. for details.
##' @param by optional; named vector or list connecting names of unit of
##'   assignment/ variables in \code{specification} to unit of
##'   assignment/cluster variables in \code{data}. Names represent variables in
##'   the StudySpecification; values represent variables in the data. Only
##'   needed if variable names differ.
##' @param data optionally the data for the analysis to be performed on. May be
##'   excluded if these functions are included as the \code{weights} argument of
##'   a model.
##' @return a \code{WeightedStudySpecification} object
##' @keywords internal
.weights_calc <- function(specification,
                          target,
                          weightAlias,
                          dichotomy,
                          by,
                          data) {
  if (!(.isValidWeightTarget(target))) {
    stop("Invalid weight target")
  }
  
  if (is.null(specification)) {
    specification <- .get_spec()
  }
  
  # get `dichotomy` argument and validate against any up the call stack
  if (is.null(dichotomy)) {
    possible_dichotomies <- .find_dichotomies()
    dichotomy <- .validate_dichotomy(possible_dichotomies)
  } else {
    dichotomy <- .validate_dichotomy(dichotomy)
  }
  
  if (is.null(data)) {
    # Only thing we need from the data is unit of assignment info to enable
    # later merge
    form <- as.formula(paste("~", paste(var_names(specification, "u"),
                                        collapse = "+")))
    
    data <- .get_data_from_model("weights", form, by)
  } else {
    # #174 handle tibbles
    data <- .as_data_frame(data)
  }
  
  if (!is.null(by)) {
    # .update_by handles checking input
    specification <- .update_by(specification, data, by)
  }
  
  # get the units of assignment from the StudySpecification to calculate weights
  txt <- .bin_txt(specification, dichotomy = dichotomy)
  
  #### generate weights
  
  if (!has_blocks(specification)) {
    # If no block is specified, then e_z is the proportion of units of
    # assignment who receive the treatment.
    e_z <- mean(txt, na.rm = TRUE)
  } else {
    # If a block is specified, then e_z varies by block and is the proportion
    # of units of assignments within the block that receive the treatment.
    blks <- specification@structure[, c(var_names(specification, "u"),
                                        var_names(specification, "b"))]
    
    # Generate e_z and merge back into blks
    e_z_ <- aggregate(txt ~ .,
                      data = blks[,var_names(specification, "b"), drop=FALSE],
                      FUN=function(x) sum(x)/length(x))
    blks <- .merge_preserve_order(blks, e_z_, all.x = TRUE)
    # Rename for clarity
    names(blks)[ncol(blks)] <- "e_z"

    e_z <- blks[["e_z"]]
    
  }
  
  weights  <-
    switch(target,
           ate = txt / e_z + (1 - txt) / (1 - e_z),
           ett = txt + (1 - txt) * e_z / (1 - e_z),
           etc= (1-txt) + txt * (1-e_z)/e_z,
           ato= txt*(1-e_z) + (1-txt)*e_z
    )
  
  to_reset_to_0 <- e_z == 1 | e_z == 0
  if (any(to_reset_to_0, na.rm = TRUE)) {
    weights[to_reset_to_0] <- 0
  }
  
  return(.join_spec_weights(weights,
                            specification,
                            target = target,
                            weightAlias = weightAlias,
                            data = data,
                            dichotomy = dichotomy))
}

##' Helper function called during creation of the weights via \code{ate()} or
##' \code{ett()}
##' @title (Internal) Expand unit of assignment level weights to the level of
##'   the data
##' @param weights a vector of weights sorted according to the
##'   \code{StudySpecification}
##' @param specification a \code{StudySpecification}
##' @param target One of "ate" or "ett"
##' @param weightAlias Any currently supported alias
##' @param data New data
##' @param dichotomy formula used to specify a dichotomy of a non-binary
##'   treatment variable. The output \code{WeightedStudySpecification} object
##'   will store this as its \code{dichotomy} slot, unless it is NULL, in which
##'   case it will be translated to an empty \code{formula}.
##' @return a \code{WeightedStudySpecification}
##' @keywords internal
.join_spec_weights <- function(weights,
                               specification,
                               target,
                               weightAlias,
                               data,
                               dichotomy) {
  uoanames <- var_names(specification, "u")
  
  # Merge uoa data with weights at uoa level
  uoadata <- specification@structure[, var_names(specification, "u"), drop = FALSE]
  uoadata$specification_weights <- weights
  
  # Merge with data to expand weights to unit of analysis level
  if (uoanames[1L] == "..uoa.." & length(uoanames) == 1L) {
    data[["..uoa.."]] <- rownames(data)
  }
  weights <- .merge_preserve_order(data, uoadata,
                                   by = uoanames,
                                   all.x = TRUE)$specification_weights
  
  # Replace NA weights with 0 so they don't contribute to the model, but aren't
  # droppde
  weights[is.na(weights)] <- 0
  
  return(new("WeightedStudySpecification",
             weights,
             StudySpecification = specification,
             target = target,
             weightAlias = weightAlias,
             dichotomy = as.formula(dichotomy)))
}

##' @title Summarizing \code{teeMod} objects
##'
##' @description [summary()] method for class \code{teeMod}
##'
##' @details If a \code{teeMod} object is fit with a \code{SandwichLayer}
##'   offset, then the usual \code{stats::summary.lm()} output is enhanced by
##'   the use of covariance-adjusted sandwich standard errors, with t-test
##'   values recalculated to reflect the new standard errors.
##'
##' @param object \code{teeMod} object
##' @param vcov.type A string indicating the desired variance estimator. See
##'   [vcov_tee()] for details on accepted types.
##' @param ... Additional arguments to [vcov_tee()], such as the desired finite
##'   sample heteroskedasticity-robust standard error adjustment.
##' @return object of class \code{summary.teeMod}
##' @export
##' @method summary teeMod
##' @rdname teeMod_summary
summary.teeMod <- function(object,
                                   vcov.type = "HC0",
                                   ...) {
  mc <- match.call()
  orig.coefficients <- object$coefficients
  object$coefficients <- replace(orig.coefficients, is.na(orig.coefficients), 0)
  out <- summary(as(object, "lm"))

  if (object$rank > 0) {
    dots <- list(...)
    toprint <- !grepl("^offset:", names(orig.coefficients))
    out$coefficients <- rbind(
      out$coef,
      matrix(NA, nrow = sum(toprint) - nrow(out$coef), ncol = 4)
    )
    rownames(out$coefficients) <- names(orig.coefficients)[toprint]

    out$coefficients[, 1L] <- orig.coefficients[toprint]
    covmat <- vcov_tee(object, type = vcov.type, ...)
    # get cluster ID's and, if there's covariance adjustment, undo the ordering
    cluster_ids <- .make_uoa_ids(object, substr(vcov.type, 1, 2), dots$cluster)
    if (inherits(object$model$`(offset)`, "SandwichLayer")) {
      ord <- .order_samples(object)
      cluster_ids <- cluster_ids[seq_along(c(ord$Q_not_C, ord$Q_in_C))][
        match(seq_along(c(ord$Q_not_C, ord$Q_in_C)),
              as.numeric(c(names(ord$Q_not_C), names(ord$Q_in_C))))
      ]
    }
    if (is.null(dof.type <- eval(mc$dof.type))) {
      dof.type <- if (vcov.type %in% paste0(c("MB", "HC", "CR"), 2)) {
        "IK"
      } else "stata"
    }
    dof <- vapply(seq_len(object$rank),
                  .get_dof,
                  numeric(1L),
                  x = object,
                  vcov.type = vcov.type,
                  cluster_ids = cluster_ids,
                  dof.type = dof.type)
    out$coefficients[!is.na(orig.coefficients), 2L] <- sqrt(diag(covmat))[
      names(orig.coefficients)[toprint & !is.na(orig.coefficients)]]
    out$coefficients[, 3L] <- out$coefficients[, 1L] / out$coefficients[, 2L]
    out$coefficients[, 4L] <- 2*stats::pt(
      abs(out$coefficients[, 3L]),
      c(dof,
        # just compute quick stata dof for control condition means
        rep(.get_dof(object, 2, "CR0", "stata", cluster_ids),
            sum(toprint) - nrow(out$coef))),
      lower.tail = FALSE
    )
    out$vcov.type <- attr(covmat, "type")
    out$vcov.cov_adj_bias_correction <- attr(covmat, "cov_adj_correction")
  }
  class(out) <- "summary.teeMod"
  out$teeMod <- object
  return(out)
}


##' @param x \code{summary.teeMod} object
##' @param digits the number of significant digits to use when printing.
##' @param signif.stars logical. If ‘TRUE’, ‘significance stars’ are printed for
##'   each coefficient.
##' @importFrom stats pt printCoefmat
##' @export
##' @rdname teeMod_summary
print.summary.teeMod <- function(x,
                                         digits =
                                           max(3L, getOption("digits") - 3L),
                                         signif.stars =
                                           getOption("show.signif.stars"),
                                         ...) {

  cat("\nCall:\n", paste(deparse1(x$teeMod@lmitt_call), sep = "\n",
                         collapse = "\n"), "\n", sep = "")

  df <- x$df

  if (x$teeMod@lmitt_fitted) {
    coefmatname <- "Treatment Effects"
  } else {
    coefmatname <- "Coefficients"
  }

  if (length(x$aliased) == 0L) {
    cat("\nNo Coefficients\n")
    return(invisible(x))
  }
  else {
    if (nsingular <- df[3L] - df[1L])
      cat("\n", coefmatname, ": (", nsingular,
          " not defined because of singularities)\n",
          sep = "")
    else cat("\n", coefmatname, ":\n")
    coefs <- x$coefficients
    if (any(aliased <- x$aliased)) {
      cn <- names(aliased)
      coefs <- matrix(NA, length(aliased), 4,
                      dimnames = list(cn, colnames(coefs)))
      coefs[!aliased, ] <- x$coefficients
    }
    if (x$teeMod@lmitt_fitted) {
      toprint <- (
        grepl(paste0("^\\`?", var_names(x$teeMod@StudySpecification, "t"), "\\."), rownames(coefs)) |
          grepl(paste0("^", formula(x$teeMod)[[2L]], ":"), rownames(coefs))
      )
    } else {
      toprint <- !grepl("^offset:", rownames(coefs))
    }
    stats::printCoefmat(coefs[toprint, , drop = FALSE],
                        digits = digits,
                        signif.stars = signif.stars,
                        na.print = "NA", ...)
  }

  if (sum(toprint) > 0 & any(!is.na(coefs[toprint, 1]))) {
    # Only print if we estimate at least one treatment effect
    cat(paste0("Std. Error calculated via type \"", x$vcov.type, "\"\n\n"))
    if (!is.null(bc <- x$vcov.cov_adj_bias_correction)) {
      cat(paste0("Residuals from covariance adjustment model adjusted by an \"",
                 bc, "\" bias correction\n\n"))
    }
  } else {
    cat("\n")
  }
  invisible(x)
}

#' @include StudySpecification.R
NULL

# (Internal) Extract SandwichLayer from model. Can exist in two places,
# depending on whether model was called as `lm(y ~ z, offset = cov_adj(spec))`
# or `lm(y ~ z + offset(cov_adj(spec))`. Returns `NULL` if it can't find it.
.get_cov_adj <- function(x) {

  # Look for an `offset = ` argument
  offset_from_arg <- x$model$"(offset)"

  if (inherits(offset_from_arg, "PreSandwichLayer") ||
      inherits(offset_from_arg, "SandwichLayer")) {
    return(offset_from_arg)
  }

  # Look for a `+ offset(` term in the formula
  offset_col <- x$model[grepl("offset\\(", names(x$model))]

  which_are_SL <- vapply(offset_col,
                         function(x) {
                           inherits(x, "PreSandwichLayer") ||
                             inherits(x, "SandwichLayer")
                          },
                         logical(1))

  if (sum(which_are_SL) == 1) {
    # Most of the time there will only be one offset, so return it
    return(offset_col[, which_are_SL])
  } else if (sum(which_are_SL) > 1) {
    if (length(unique(as.list(offset_col))) == 1) {
      # All cov_adj are the same, return any of them
      return(offset_col[, 1])
    }
    # At least one cov_Adj is different, error
    stop("Multiple cov_adj() calls found in offset")
  }

  return(NULL)
}

test_that("Finding cov_adj in model", {

  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  ca <- lm(y ~ x, data = simdata)

  cap <- cov_adj(ca, newdata = simdata, specification = spec)

  mod1 <- lm(y ~ z, data = simdata, weights = ate(spec), offset = cov_adj(ca))
  expect_identical(.get_cov_adj(mod1), cap)

  mod2 <- lm(y ~ z  + offset(cov_adj(ca)), data = simdata, weights = ate(spec))
  expect_identical(.get_cov_adj(mod2), cap)


  # Weird variable named offset
  names(simdata)[8] <- "offset(abc)"
  mod3 <- lm(y ~ z + `offset(abc)`, data = simdata)
  expect_null(.get_cov_adj(mod3))

  # An offset not from us
  mod4 <- lm(y ~ z + offset(force), data = simdata)
  expect_null(.get_cov_adj(mod4))
  mod5 <- lm(y ~ z, data = simdata, offset = force)
  expect_null(.get_cov_adj(mod5))

  # proper offset, but weird name cov_adj variable
  names(simdata)[8] <- "cov_adj"
  mod6 <- lm(y ~ z + offset(cov_adj), data = simdata)
  expect_null(.get_cov_adj(mod6))

  # Double offset
  names(simdata)[8] <- "x"
  mod7 <- lm(y ~ z + offset(cov_adj(ca)) + offset(x), data = simdata,
             weights = ate(spec))
  expect_identical(.get_cov_adj(mod7), cap)

  # Double offset, both with the identical cov_adj
  mod8 <- lm(y ~ z + offset(cov_adj(ca)) + offset(cap), data = simdata,
             weights = ate(spec))
  expect_identical(.get_cov_adj(mod8), cap)

  # Double offset, but with different cov_adj
  ca2 <- lm(y ~ force, data = simdata)
  mod9 <- lm(y ~ z + offset(cov_adj(ca)) + offset(cov_adj(ca2)), data = simdata,
             weights = ate(spec))
  expect_error(.get_cov_adj(mod9), "Multiple cov_adj")
})

##' A variation of expand.model.frame which works for \code{teeMod}
##' objects
##'
##' When building a \code{teeMod} object inside \code{lmitt()}, we do a lot of
##' manipulation of the variables involved in the model such that by the time
##' the \code{teeMod} is produced, neither the outcome nor predictors actually
##' fit in the model exist in the \code{data} passed into the call.
##'
##' (E.g. to be specific, if a user calls \code{myda <- lmitt(y ~ 1, data =
##' mydata)}, then \code{model.frame(myda)} would contain column names not found
##' in \code{mydata}.)
##'
##' This is a clone of \code{stats::expand.model.frame()} which has one addition
##' - after extracting the \code{model$call$data} from \code{model}, it adds
##' columns from \code{model.frame(model)} to the object. This ensures that the
##' additional variables created during \code{lmitt()} can be found.
##'
##' Trivial modifications from \code{stats::expand.model.frame()} include
##' ensuring \code{model} is a \code{teeMod} object, and using the \code{::}
##' syntax as appropriate.
##' @title Add new variables to a model frame from a \code{teeMod} object
##' @param model A \code{teeMod} object
##' @param extras one-sided formula or vector of character strings describing
##'   new variables to be added
##' @param envir an environment to evaluate things in
##' @param na.expand logical; see \code{stats::expand.model.frame} for details
##' @return A \code{data.frame}
##' @keywords internal
.expand.model.frame_teeMod <- function (model,
                                    extras,
                                    envir = environment(formula(model)),
                                    na.expand = FALSE) {
  # R4.2.3 or earlier
  if (as.numeric(version$major) < 4 |
        (as.numeric(version$major) == 4 & as.numeric(version$minor) < 3)) {
    stopifnot(is(model, "teeMod")) # JE addition
    f <- stats::formula(model) # JE modification
    data <- eval(model$call$data, envir)
    data <- cbind(data, stats::model.frame(model)) # JE addition
    ff <- foo ~ bar + baz
    gg <- if (is.call(extras))
            extras
    else str2lang(paste("~", paste(extras, collapse = "+")))
    ff[[2L]] <- f[[2L]]
    ff[[3L]][[2L]] <- f[[3L]]
    ff[[3L]][[3L]] <- gg[[2L]]
    if (!na.expand) {
      naa <- model$call$na.action
      subset <- model$call$subset
      rval <- eval(call("model.frame", ff, data = data, subset = subset,
                        na.action = naa), envir)
    }
    else {
      subset <- model$call$subset
      rval <- eval(call("model.frame", ff, data = data, subset = subset,
                        na.action = I), envir)
      oldmf <- stats::model.frame(model) # JE modification
      keep <- match(rownames(oldmf), rownames(rval))
      rval <- rval[keep, ]
      class(rval) <- "data.frame"
    }
    return(rval)
  } else {
    stopifnot(is(model, "teeMod")) # JE addition
    f <- stats::formula(model) # JE modification
    cl <- getCall(model)
    data <- cl$data
    data <- cbind(data, stats::model.frame(model)) # JE addition
    f <- formula(model)
    ff <- foo ~ bar + baz
    gg <- if (is.call(extras))
        extras
    else str2lang(paste("~", paste(extras, collapse = "+")))
    ff[[2L]] <- f[[2L]]
    ff[[3L]][[2L]] <- f[[3L]]
    ff[[3L]][[3L]] <- gg[[2L]]
    environment(ff) <- envir
    if (!na.expand) {
        rval <- eval(call("model.frame", ff, data = data,
            subset = cl$subset, na.action = cl$na.action), envir)
    }
    else {
        rval <- eval(call("model.frame", ff, data = data,
            subset = cl$subset, na.action = I), envir)
        oldmf <- stats::model.frame(model) # JE modification
        keep <- match(rownames(oldmf), rownames(rval))
        rval <- rval[keep, ]
        class(rval) <- "data.frame"
    }
    return(rval)
}

}

test_that("expand.m.f.da basics", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec)
  expect_true(all.equal(stats::expand.model.frame(mod, ~ o),
                        propertee:::.expand.model.frame_teeMod(mod, ~ o)))
  expect_true(all.equal(stats::expand.model.frame(mod, ~ o, na.expand = TRUE),
                        propertee:::.expand.model.frame_teeMod(mod, ~ o,
                                                         na.expand = TRUE)))

  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE)
  expect_true(all.equal(stats::expand.model.frame(mod, ~ o),
                        propertee:::.expand.model.frame_teeMod(mod, ~ o)))
  expect_true(all.equal(stats::expand.model.frame(mod, ~ o, na.expand = TRUE),
                        propertee:::.expand.model.frame_teeMod(mod, ~ o,
                                                         na.expand = TRUE)))


})

##' @title (Internal) Find \code{dichotomy} formulas in the call stack
##' @details \code{.find_dichotomies()} searches for \code{lmitt.formula()}
##'   calls and their \code{weights} arguments for any \code{dichotomy}
##'   arguments.
##' @return A list where elements are either formulas or NULL, depending on
##'   whether a \code{dichotomy} argument was found in \code{lmitt.formula()} or
##'   its \code{weights} argument
##' @keywords internal
.find_dichotomies <- function() {
  found_lmitt <- grepl("^lmitt\\.formula$",
                       lapply(sys.calls(), "[[", 1),
                       perl = TRUE)
  # if this is called within `.weights_calc()` don't look for dichotomies in the
  # weights arguments (no need to compare the dichotomy to itself)
  weights.calls <- if (any(grepl("weights_calc", lapply(sys.calls(), "[[", 1),
                                 fixed = TRUE))) FALSE else {
    lapply(sys.calls(), "[[", "weights")
  }

  dichotomies <- list(lmitt = NULL, weights = NULL)
  if (any(found_lmitt)) {
    lmitt.call <- get("lmitt.call", sys.frame(which(found_lmitt)[1L]))
    dichotomies$lmitt <- lmitt.call$dichotomy
  }
  if (length(valid_pos <- which(!vapply(weights.calls, is.null, logical(1L)))) > 0) {
    dichotomies$weights <- mapply(
      function(cc, pos) {
        if (is.call(cc)) cc$dichotomy else if (
          inherits(wd <- eval("weights", pos), "WeightedStudySpecification")) wd@dichotomy else NULL
        },
      weights.calls[valid_pos],
      valid_pos,
      SIMPLIFY = FALSE)[[1L]]
  }

  return(dichotomies)
}

#' @title (Internal) Validate a dichotomy against other dichotomies found in the
#'   call stack
#' @param possible_dichotomy a list or a formula, the former coming from
#'   \code{.find_dichotomies()}
#' @return formula for a dichotomy or NULL
#' @keywords internal
.validate_dichotomy <- function(possible_dichotomy) {
  if (inherits(possible_dichotomy, "list")) {
    # possible_dichotomy is a list if it's passed the output of .find_dichotomies().
    # in that case, there's no need to compare it to other dichotomy arguments up
    # the stack, so we return a dichotomy if it exists in the list and NULL otherwise
    dichotomy <- possible_dichotomy$lmitt
    if (inherits(dichotomy, "name")) {
      # if the dichotomy passed to weights or lmitt was a stored object, find and
      # evaluate it
      for (s in seq_len(sys.nframe())) {
        eval_d <- tryCatch(eval.parent(dichotomy, s), error = function(e) NULL)
        if (!is.null(eval_d)) break
      }
      dichotomy <- eval_d
    }
    if (is.null(dichotomy)) {
      dichotomy <- possible_dichotomy$weights
    }
    if (inherits(dichotomy, "name")) {
      for (s in seq_len(sys.nframe())) {
        eval_d <- tryCatch(eval.parent(dichotomy, s), error = function(e) NULL)
        if (!is.null(eval_d)) break
      }
      dichotomy <- eval_d
    }
    if (!is.null(dichotomy)) dichotomy <- as.formula(dichotomy)
  } else if (inherits(possible_dichotomy, "call")) {
    possible_dichotomy <- as.formula(possible_dichotomy)
  } else if (inherits(possible_dichotomy, "name")) {
    # if the dichotomy passed to weights or lmitt was a stored object, find and
    # evaluate it
    for (s in seq_len(nf <- sys.nframe())) {
      eval_pd <- tryCatch(
        eval.parent(possible_dichotomy, s),
        error = function(e) if (s == nf) {
          stop(paste("Could not find", deparse1(possible_dichotomy), "in call stack"),
               call. = FALSE)
        } else return(NULL)
      )
      if (!is.null(eval_pd)) break
    }
    possible_dichotomy <- eval_pd
    if (!is.null(possible_dichotomy)) possible_dichotomy <- as.formula(possible_dichotomy)
  }

  if (inherits(possible_dichotomy, "formula")) {
    dichotomy <- possible_dichotomy
    # find any other dichotomies up the call stack to compare dichotomy against
    other_dichotomies <- .find_dichotomies()
    if (!is.null(other_dichotomies$lmitt)) {
      if (inherits(other_dichotomies$lmitt, "name")) {
        for (s in seq_len(sys.nframe())) {
          ok <- tryCatch(eval.parent(other_dichotomies$lmitt, s),
                         error = function(e) NULL)
          if (!is.null(ok)) break
        }
        if (is.null(ok)) stop(paste("Could not find", deparse1(other_dichotomies$lmitt), "in call stack"))
        other_dichotomies$lmitt <- ok
      }
      if (!identical(deparse1(other_dichotomies$lmitt),
                     do.call("deparse1", list(eval(dichotomy))))) {
        warning(paste("`dichotomy` passed to", paste0("`", sys.call(-1L)[[1L]], "()`"),
                      "is not the same as the `dichotomy` passed to `lmitt()`"))
      }
    }
    if (!is.null(other_dichotomies$weights)) {
      if (inherits(other_dichotomies$weights, "name")) {
        for (s in seq_len(sys.nframe())) {
          ok <- tryCatch(eval.parent(other_dichotomies$weights, s),
                         error = function(e) NULL)
          if (!is.null(ok)) break
        }
        if (is.null(ok)) stop(paste("Could not find", deparse1(other_dichotomies$weights), "in call stack"))
        other_dichotomies$weights <- ok
      }
      if (!identical(deparse1(other_dichotomies$weights),
                     do.call("deparse1", list(eval(dichotomy))))) {
        warning(paste("`dichotomy` passed to", paste0("`", sys.call(-1L)[[1L]], "()`"),
                      "is not the same as the `dichotomy` passed to weights"))
      }
    }
  } else if (!exists("dichotomy")) {
    stop("No formulas provided to .validate_dichotomy")
  }

  return(invisible(dichotomy))
}

#' @include StudySpecification.R
NULL
# The above ensures that `StudySpecification` is defined prior to `WeightedStudySpecification`

#' (Internal) Modeling weights with an accompanying StudySpecification
#'
#' @slot .Data numeric vector of modeling weights
#' @slot StudySpecification a StudySpecification
#' @slot target character string, e.g. "ate"
#' @slot weightAlias alias for target appearing in an originating call
#' @slot dichotomy formula describing a treatment/comparison dichotomy
#' @details
#'  `@target` is used for calculation purpose; defining what weight to calculate
#'   `@weightAlias` is only to store the alias used in creation of the 
#'     weights in case we want to report it later.
#' @keywords internal
setClass("WeightedStudySpecification",
         contains = "numeric",
         slots = c(StudySpecification = "StudySpecification",
                   target = "character",
                   weightAlias = "character",
                   dichotomy = "formula"))

setValidity("WeightedStudySpecification", function(object) {
  if (!.isValidWeightTarget(object@target)) {
    return(paste0("@target must be one of [", .listValidWeightTargets(),
                  "]. unknown @target: ", object@target))
  }
  if (!.isValidWeightAlias(object@weightAlias)) {
    return(paste0("@weightAlias must be one of [", .listValidWeightAliases(),
                 "]. unknown @weightAlias: ", object@weightAlias))
  }
  return(TRUE)
})

##' @title Show a \code{WeightedStudySpecification}
##'
##' @description Prints out the weights from a \code{WeightedStudySpecification}
##'
##' @param object a \code{WeightedStudySpecification} object
##' @return an invisible copy of \code{object}
##' @export
setMethod("show", "WeightedStudySpecification", function(object) {
  print(object@.Data)
  invisible(object)
})


##' @title \code{WeightedStudySpecification} Operations
##'
##' @description Algebraic operators on \code{WeightedStudySpecification}
##'   objects and numeric vectors. \code{WeightedStudySpecification}s do not
##'   support addition or subtraction.
##'
##' @details These are primarily used to either combine weights via
##'   multiplication, or to invert weights. Addition and subtraction are not
##'   supported and will produce errors.
##'
##' @param e1,e2 \code{WeightedStudySpecification} or \code{numeric} objects
##' @rdname WeightedStudySpecificationOps
##' @return a \code{WeightedStudySpecification} object
##' @export
setMethod("+", signature(e1 = "WeightedStudySpecification", e2 = "numeric"),
          function(e1, e2) addsubtracterror()
          )

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("+", signature(e1 = "numeric", e2 = "WeightedStudySpecification"),
          function(e1, e2) addsubtracterror()
          )

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("-", signature(e1 = "WeightedStudySpecification", e2 = "numeric"),
          function(e1, e2) addsubtracterror()
          )

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("-", signature(e1 = "numeric", e2 = "WeightedStudySpecification"),
          function(e1, e2) addsubtracterror()
          )

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("*", signature(e1 = "WeightedStudySpecification", e2 = "numeric"),
          function(e1, e2) {
            e1@.Data <- e1@.Data * e2
            validObject(e1)
            return(e1)
          })

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("*", signature(e1 = "numeric", e2 = "WeightedStudySpecification"),
          function(e1, e2) {
            e2@.Data <- e1 * e2@.Data
            validObject(e2)
            return(e2)
          })

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("/", signature(e1 = "WeightedStudySpecification", e2 = "numeric"),
          function(e1, e2) {
            e1@.Data <- e1@.Data/e2
            validObject(e1)
            return(e1)
          })

##' @rdname WeightedStudySpecificationOps
##' @export
setMethod("/", signature(e1 = "numeric", e2 = "WeightedStudySpecification"),
          function(e1, e2) {
            e2@.Data <- e1/e2@.Data
            validObject(e2)
            return(e2)
          })

addsubtracterror <- function() {
  stop("Cannot perform addition or subtraction on WeightedStudySpecifications")
}

setGeneric("weights")

##' @title Extract Weights from \code{WeightedStudySpecification}
##'
##' @description A \code{WeightedStudySpecification} object contains a numeric
##'   vector with a few additional slots, this extracts only the numeric vector.
##'
##' @param object a \code{WeightedStudySpecification} object
##' @param ... Ignored
##' @return A numeric \code{vector} of the weights
##' @export
setMethod("weights", "WeightedStudySpecification", function(object, ...) {
  return(as.numeric(object))
})

setGeneric("subset")

##' @title \code{WeightedStudySpecification} subsetting
##'
##' @description Provides functionality to subset the weights of a
##'   \code{WeightedStudySpecification} object.
##'
##' @param subset Logical vector identifying values to keep or drop
##' @return A \code{WeightedStudySpecification} object which is a subsetted
##'   version of \code{x}.
##' @export
##' @rdname WeightedStudySpecification.subset
setMethod("subset", "WeightedStudySpecification", function(x, subset) {
  x@.Data <- subset(x@.Data, subset = subset)
  return(x)
})

setGeneric("[")

##' @param x \code{WeightedStudySpecification} object
##' @param i indices specifying elements to extract or replace. See
##'   \code{help("[")} for further details.
##' @export
##' @importFrom methods callNextMethod
##' @rdname WeightedStudySpecification.subset
setMethod("[", "WeightedStudySpecification",
          function(x, i) {
            dat <- methods::callNextMethod()
            x@.Data <- dat
            return(x)

          })

#' @importFrom stats summary.lm residuals
bread.lm <- function(x, ...) {
  if (!is.null(x$na.action))
    class(x$na.action) <- "exclude"
  sx <- stats::summary.lm(x)
  n <- length(residuals(x))
  return(sx$cov.unscaled * n)
}


#' @importFrom stats residuals weights
bread.glm <- function(x, ...) {
  if (!is.null(x$na.action))
    class(x$na.action) <- "exclude"
  sx <- summary(x)
  wres <- as.vector(stats::residuals(x, "working")) * stats::weights(x, "working")
  dispersion <- if (substr(x$family$family, 1L, 17L) %in%
                      c("poisson", "binomial", "Negative Binomial"))
                  1
  else sum(wres^2)/sum(weights(x, "working"))
  return(sx$cov.unscaled * length(sx$deviance.resid) * dispersion)
}

#' @importFrom stats summary.lm residuals
bread.mlm <- function(x, ...) {
  if (!is.null(x$na.action))
    class(x$na.action) <- "exclude"
  K <- x$rank
  piv <- x$qr$pivot[seq_len(K)]
  cf <- x$coef[piv,,drop=FALSE]
  rval <- summary.lm(x)
  n <- nrow(residuals(x))

  rval <- kronecker(
    structure(diag(ncol(cf)), dimnames = rep.int(list(colnames(cf)),  2L)),
    structure(rval$cov.unscaled, dimnames = rep.int(list(rownames(cf)), 2L)) * n,
    make.dimnames = TRUE)

  return(rval)
}

#' @include WeightedStudySpecification.R
NULL

##' @title Concatenate weights
##'
##' @description Given several variations of weights generated from a single
##'   \code{StudySpecification}, combine into a single weight.
##'
##' @details Concatenating \code{WeightedStudySpecification} objects with [c()]
##'   requires both individual \code{WeightedStudySpecification} objects to come
##'   from the same \code{StudySpecification} and have the same target (e.g all
##'   created with [ate()] or all created with [ett()], no mixing-and-matching).
##'   All arguments to [c()] must be \code{WeightedStudySpecification}.
##'
##'   \code{WeightedStudySpecification} objects may be concatenated together
##'   even without having the same \code{@dichotomy} slot. This procedure only
##'   prompts a warning for differing dichotomies if the argument
##'   \code{warn_dichotomy_not_equal} is set to \code{TRUE}.
##' @param x, .. a \code{WeightedStudySpecification} object, typically created
##'   from [ate()] or [ett()]
##' @param ... any number of additional \code{WeightedStudySpecification}
##'   objects with equivalent \code{StudySpecification} to \code{x} and
##'   eachother
##' @param warn_dichotomy_not_equal if \code{FALSE} (default),
##'   \code{WeightedStudySpecification}s are considered equivalent even if their
##'   \code{dichotomy} differs. If \code{TRUE}, a warning is produced.
##' @return A numeric \code{vector} with the weights concatenated in the input
##'   order.
##' @export
##' @importFrom methods slot
##' @importFrom stats formula
##' @examples
##' data(simdata)
##' spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
##' w1 <- ate(spec, data = simdata[1:30,])
##' w2 <- ate(spec, data = simdata[31:40,])
##' w3 <- ate(spec, data = simdata[41:50,])
##' c_w <- c(w1, w2, w3)
##' c(length(w1), length(w2), length(w3), length(c_w))
##'
##' spec <- rct_spec(dose ~ unit_of_assignment(uoa1, uoa2), data = simdata)
##' w1 <- ate(spec, data = simdata[1:10, ], dichotomy = dose >= 300 ~ .)
##' w2 <- ate(spec, data = simdata[11:30, ], dichotomy = dose >= 200 ~ .)
##' w3 <- ate(spec, data = simdata[31:50, ], dichotomy = dose >= 100 ~ .)
##' c_w <- c(w1, w2, w3)
setMethod("c", signature(x = "WeightedStudySpecification"),
          function(x, ..., warn_dichotomy_not_equal = FALSE) {
  dots <- list(...)
  # x must be a WeightedStudySpecification to get here; ensure all other elements
  # are as well
  if (any_numeric <- any(vapply(dots, function(k) !inherits(k, "WeightedStudySpecification"), TRUE))) {
    message(paste("Concatenating a WeightedStudySpecification with a non-WeightedStudySpecification vector.",
                  "If non-WeightedStudySpecification vector has been formed from previous",
                  "concatenation of WeightedStudySpecification objects, equality of `StudySpecification`",
                  "slots cannot be confirmed"))
  }

  if (!any_numeric) {
    # Ensure all WeightedStudySpecifications have the same target
    targets <- c(x@target, lapply(dots, methods::slot, "target"))
    if (length(unique(targets)) > 1) {
      stop(paste("WeightedStudySpecifications can only be concatenated with",
                 "the same target"))
    }

    specifications <- c(x@StudySpecification, lapply(dots, methods::slot, "StudySpecification"))
    # temporarily remove calls from the StudySpecifications to avoid weird discrepancies here
    spectmp <- lapply(specifications, function(x) {
      x@call <- call("c") # placeholder for unique check
      x
    })

    if (length(unique(spectmp)) == 1) {
      if (warn_dichotomy_not_equal) {
        dichotomies <- Reduce(c, lapply(dots, methods::slot, "dichotomy"), init = x@dichotomy)
        if (length(unique(dichotomies)) > 1) {
          warning("Concatenating WeightedStudySpecifications with different `dichotomy` slots")
        }
      }
    } else {
      stop("Cannot combine WeightedStudySpecifications from differing StudySpecifications")
    }
  }

  return(Reduce(c, lapply(dots, methods::slot, ".Data"), init = x@.Data))
})

##' @title Convert \code{StudySpecification} between types
##'
##' @description Convert a \code{StudySpecification} between a observational
##'   study, a randomized control trial, and a regression discontinuity (created
##'   from \code{obs_spec}, \code{rct_spec} and \code{rd_spec} respectively).
##'
##' @param StudySpecification a \code{StudySpecification} to convert
##' @param data converting to an RD requires adding a \code{forcing} variable,
##'   which requires access to the original data.
##' @param ... Ignored.
##' @param loseforcing converting from RD to another \code{StudySpecification}
##'   type will error to avoid losing the forcing variable. Setting
##'   \code{loseforcing = TRUE} allows the conversion to automatically drop the
##'   forcing variable. Default \code{FALSE}.
##' @param forcing converting to an RD requires adding a \code{forcing}
##'   variable. This should be entered as a formula which would be passed to
##'   \code{update}, e.g. \code{forcing = . ~ . + forcing(forcevar)}.
##' @return \code{StudySpecification} of the updated type
##' @export
##' @importFrom stats as.formula update
##' @rdname specificationconversion
##' @examples
##' spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
##' spec
##' as_obs_spec(spec)
##' as_rd_spec(spec, simdata, forcing = ~ . + forcing(force))
##' spec2 <- rd_spec(o ~ uoa(uoa1, uoa2) + forcing(force), data = simdata)
##' spec2
##' # as_rct_spec(spec2) # this will produce an error
##' as_rct_spec(spec2, loseforcing = TRUE)
as_rct_spec <- function(StudySpecification, ..., loseforcing = FALSE) {
  if (StudySpecification@type == "RD") {
    if (!loseforcing) {
      stop(paste("Any `forcing` variables will be dropped. Pass option",
                 "`loseforcing = TRUE` to proceed."))
    }
    StudySpecification <- .remove_forcing(StudySpecification)
  }
  StudySpecification@type <- "RCT"
  StudySpecification@call[[1]] <- as.name("rct_spec")
  validObject(StudySpecification)
  return(StudySpecification)
}

##' @export
##' @rdname specificationconversion
as_obs_spec <- function(StudySpecification, ..., loseforcing = FALSE) {
  if (StudySpecification@type == "RD") {
    if (!loseforcing) {
      stop(paste("Any `forcing` variables will be dropped. Pass option",
                 " `force = TRUE` to proceed."))
    }
    StudySpecification <- .remove_forcing(StudySpecification)
  }
  StudySpecification@type <- "Obs"
  StudySpecification@call[[1]] <- as.name("obs_spec")
  validObject(StudySpecification)
  return(StudySpecification)
}

##' @export
##' @rdname specificationconversion
as_rd_spec <- function(StudySpecification, data, ..., forcing) {
  if (!inherits(forcing, "formula")) {
    stop('`forcing` must be entered as a formula such as "~ . + forcing(a, b)"')
  }

  origcall <- StudySpecification@call

  StudySpecification <- rd_spec(formula =
                        stats::update(stats::as.formula(StudySpecification@call[[2]]),
                                      forcing),
                      data = data,
                      subset = eval(StudySpecification@call[["subset"]]))

  StudySpecification@call <- origcall
  StudySpecification@call[[1]] <- as.name("rd_spec")
  StudySpecification@call[[2]] <- as.call(stats::update(stats::as.formula(origcall[[2]]),
                                            forcing))

  validObject(StudySpecification)
  return(StudySpecification)
}

##' In preparation for converting an RD \code{StudySpecification} to another
##' \code{StudySpecification}, this will strip the forcing variable entirely. It
##' is removed from the data (both \code{@structure} and \code{@column_index}),
##' as well as from the formula stored in \code{@call}.
##'
##' Note that the output \code{StudySpecification} will fail a validity check
##' (with \code{validObject()}) due to an RD \code{StudySpecification} requiring
##' a forcing variable, so change the \code{@type} immediately.
##'
##' @title (Internal) Removes the forcing column entirely from a
##'   \code{StudySpecification}
##' @param spec A \code{StudySpecification}
##' @return The \code{StudySpecification} without any forcing variable
##' @keywords internal
.remove_forcing <- function(spec) {
  spec@structure <- spec@structure[, spec@column_index != "f"]
  spec@column_index <- spec@column_index[spec@column_index != "f"]

  # Remove the "forcing" element from the call formula
  spec@call[[2]][[3]] <- spec@call[[2]][[3]][!grepl("forcing",
                                                  spec@call[[2]][[3]])]
  return(spec)
}

# Where to pull the data from
extdataURLs  <-
    list(CCD="https://nces.ed.gov/ccd/data/zip/sc132a_txt.zip",
         MME="https://www.michigan.gov/cepi/-/media/Project/Websites/cepi/MiSchoolData/historical/Historical_Assessments/2011-2014MME.zip")

# pads numerical values with leading zeros for fixed character length
zero_lpad <- function(vec, char_len) {
  vapply(
    vec,
    function(x) paste(c(rep(0, char_len - nchar(x)), x), collapse = ""),
    character(1L)
  )
}

CCD_ID_COLS <- c("STID", "LEANM", "SEASCH", "SCHNAM", "CONUM", "CONAME")
CCD_CAT_COLS <- c("TITLEISTAT", "MAGNET", "CHARTR", "TYPE")
# school grade from kindergarten to G12
CCD_ENROLLMENT_COLS <- c("KG", paste0("G", zero_lpad(seq(1, 12), 2)))
RACEETH_COLS <- c("AM", "ASIAN", "HISP", "BLACK", "WHITE", "PACIFIC", "TR")
RACEGENDER_PREFIXES <- c("AM", "AS", "HI", "BL", "WH", "HP", "TR")
# Total Free and Reduced-Price Lunch Eligible Student
FRL_COL <- "TOTFRL"

# recode variable into 0 or 1
recode_binary_variable <- function(col) {
  if (inherits(col, "character")) {
    nonmissing_levels <- unique(col[!(col %in% c("N", "M"))])
    if (length(setdiff(nonmissing_levels, c("1", "2"))) == 0) {
      col[col == "1"] <- "0"
      col[col == "2"] <- "1"
    }
  } else if (inherits(col, c("numeric", "integer"))) {
    nonmissing_levels <- unique(col[!(col %in% c(-2, -1))])
    if (length(setdiff(nonmissing_levels, c(1, 2))) == 0) {
      col[col == 1] <- 0
      col[col == 2] <- 1
    }
  } else {
    warning("Encountered an unexpected variable type")
  }
  
  return(col)
}

# sum of column values / number of columns 
# if number of columns or MEMBER equal 0, return 0 
# if any column has -1 (missing data), output -1
# if any column has -2 or MEMBER column contains -2 or -1, output -2 
make_demo_perc <- function(df, cols, total_col) {
  out <- rowSums(df[, cols, drop=FALSE]) / df[[total_col]]
  out[df[[total_col]] == 0 | df$MEMBER == 0] <- 0
  out[apply(df[, cols, drop=FALSE], 1, function(x) any(x == -1))] <- -1
  out[apply(df[, cols, drop=FALSE], 1,
            function(x) any(x == -2)) | df$MEMBER %in% c(-2, -1)] <- -2
  
  return(out)
}

clean_ccd <- function(ccd) {
  rownames(ccd) <- NULL
  # keep Michigan schools only 
  ccd <- ccd[ccd$LSTATE == "MI",]
  
  # calculate percentage of gender
  gender_percs <- mapply(
    make_demo_perc,
    lapply(c("ALM", "ALF"), function(pf) paste0(RACEGENDER_PREFIXES, pf)),
    rep("MEMBER", 2),
    MoreArgs = list(df = ccd),
    SIMPLIFY = TRUE
  )
  colnames(gender_percs) <- paste(c("MALE", "FEMALE"), "PERC", sep = "_")
  
  # calculate percentage of G11 gender
  gender_g11_percs <- mapply(
    make_demo_perc,
    lapply(paste0(11, c("M", "F")),
           function(pf) paste0(RACEGENDER_PREFIXES, pf)),
    rep("G11", 2),
    MoreArgs = list(df = ccd),
    SIMPLIFY = TRUE
  )
  colnames(gender_g11_percs) <- paste(c("MALE", "FEMALE"), "G11_PERC", sep = "_")
  
  # calculate percentage of ethinicity 
  raceeth_percs <- mapply(
    make_demo_perc,
    RACEETH_COLS,
    rep("MEMBER", length(RACEETH_COLS)),
    MoreArgs = list(df = ccd),
    SIMPLIFY = TRUE
  )
  colnames(raceeth_percs) <- paste(RACEETH_COLS, "PERC", sep = "_")
  
  # calculate percentage of G11 ethinicity
  raceeth_g11_percs <- mapply(
    make_demo_perc,
    lapply(RACEGENDER_PREFIXES,
           function(pf) paste0(pf, 11, c("M", "F"))),
    rep("G11", length(RACEGENDER_PREFIXES)),
    MoreArgs = list(df = ccd),
    SIMPLIFY = TRUE
  )
  colnames(raceeth_g11_percs) <- paste(RACEETH_COLS, "G11_PERC", sep = "_")
  
  frl_perc <- make_demo_perc(ccd, FRL_COL, "MEMBER")
  
  cleaned_cat_cols <- sapply(ccd[, CCD_CAT_COLS], recode_binary_variable)
  
  # create a new column called merge_id by concatenating STID and SEASCH
  # extract columns listed in CCD_ID_COLS as a data frame 
  # calculate total enrollment for each row by summing up the values in CCD_ENROLLMENT_COLS
  # adds the rest of the columns we processed earlier 
  cbind(merge_id = paste0(ccd$STID, ccd$SEASCH),
        ccd[, CCD_ID_COLS, drop=FALSE],
        sapply(ccd[, CCD_ENROLLMENT_COLS, drop=FALSE], as.numeric),
        TOTAL_ENROLLMENT = as.numeric(apply(
          ccd[, CCD_ENROLLMENT_COLS], 1,
          function(r) sum(r[!(r %in% c(-2, -1))]))),
        cleaned_cat_cols,
        gender_percs, raceeth_percs, TOTFRL_PERC = frl_perc,
        gender_g11_percs, raceeth_g11_percs)
}

clean_scores <- function(all_scores, ccd) {
  all_scores <- all_scores[
    !is.na(all_scores$Building.Code) &
      !(all_scores$Building.Name %in% c("STATEWIDE", "ISDWIDE", "DISTRICTWIDE")),]

  stopifnot(sum(substr(colnames(all_scores),1,24)=="Average.Scale.Score.2011")==2)
  colnames(all_scores)[substr(colnames(all_scores),1,24
                              )=="Average.Scale.Score.2011"]  <-
      paste0("Average.Scale.Score.201", 2:1)
  
  out_cols <- Reduce(
    cbind,
    lapply(
      seq(2011, 2014),
      function(df, yr) {
        colnms <- paste0(c("Total.Tested.", "Average.Scale.Score."), yr)
        out_cols <- df[, colnms, drop=FALSE]
        out_cols[, 3] <- as.numeric(out_cols[, 1] == "<10")
        out_cols[, 4] <- as.numeric(out_cols[, 1] == "NULL")
        out_cols[, 5] <- as.numeric(out_cols[, 2] == "NULL" &
                                      !(out_cols[, 1] %in% c("<10", "NULL")))
        out_cols[out_cols[, 1] %in% c("<10", "NULL"), 1] <- NA_character_
        out_cols[out_cols[, 2] == "NULL", 2] <- NA_character_
        colnames(out_cols) <- c(colnms,
                                paste0(c("Masked.Average.Scale.Score.",
                                         "Missing.Students.Tested.",
                                         "Missing.Average.Scale.Score."), yr))
        sapply(out_cols, as.numeric)
      },
      df = all_scores
    )
  )
  all_scores[, grepl("(.Tested|Average.Scale)", colnames(all_scores))] <- NULL
  all_scores <- cbind(all_scores, out_cols)
  
  nchar_ccd_distid <- max(nchar(ccd$STID[!grepl("^[A-Z]", ccd$STID)]))
  nchar_ccd_schid <- max(nchar(ccd$SEASCH[!grepl("^[A-Z]", ccd$SEASCH)]))
  all_scores$District.Code <- zero_lpad(all_scores$District.Code, nchar_ccd_distid)
  all_scores$Building.Code <- zero_lpad(all_scores$Building.Code, nchar_ccd_schid)
  all_scores$merge_id <- paste0(all_scores$District.Code, all_scores$Building.Code)
  
  all_scores
}

##' Retrieve external data files from NCES and MI's CEPI
stopifnot(exists("extdataURLs"),
          require("readxl"), require("httr")
          )

.tf <- tempfile(fileext=".zip")
download.file(extdataURLs$CCD, .tf)
ccd <- read.delim(unz(.tf, "sc132a.txt"))
unlink(.tf)
cleaned_ccd <- clean_ccd(ccd)

.tf  <- tempfile(fileext=".zip")
headers = c(
  `user-agent` = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.61 Safari/537.36'
)
httr::GET(url=extdataURLs$MME,
                 httr::add_headers(.headers=headers)
          )$content |>
writeBin(.tf)

.td  <- tempdir()
mme_xls  <-
    unzip(.tf,
          files="Spring2011-2014MMEFourYearDemographicDataFile-Sortable.xls",
          exdir=.td)

mme  <-  read_excel(mme_xls, .name_repair="universal")
rm(.tf, mme_xls)

all_schools <- clean_scores(mme, cleaned_ccd) |>
               merge(cleaned_ccd, by = "merge_id", all = TRUE)

analysis1data <- all_schools[
  !is.na(all_schools$DemographicGroup) &
    all_schools$DemographicGroup == "All Students" &
    all_schools$TYPE == "1" &
    all_schools$Subject == "M",]
rownames(analysis1data) <- NULL
analysis1data$schoolid <- analysis1data$merge_id
analysis1data$merge_id <- NULL
analysis2data <- all_schools[
  !is.na(all_schools$DemographicGroup) &
    all_schools$DemographicGroup %in% c("White", "Black or African American",
                                        "Hispanic of any race", "Asian",
                                        "American Indian or Alaska Native",
                                        "Two or more races") &
    all_schools$TYPE == "1" &
    all_schools$Subject == "M",]
analysis2data$schoolid <- analysis2data$merge_id
analysis2data$merge_id <- NULL
analysis2data$DemographicGroup[!(analysis2data$DemographicGroup %in%
                                   c("White", "Black or African American"))] <-
  "Other Race/Ethnicity"
rownames(analysis2data) <- NULL
analysis2data$uniqueid <- seq_len(nrow(analysis2data))

##' @title Generate matrix of estimating equations for \code{lmrob()} fit
##' @details This is part of a workaround for an issue in the robustbase code
##'   affecting sandwich covariance estimation. The issue in question is issue
##'   #6471, robustbase project on R-Forge. This function contributes to
##'   providing sandwich estimates of covariance-adjusted standard errors for
##'   robust linear covariance adjustment models.
##' @param x An \code{lmrob} object produced using an MM/SM estimator chain
##' @param ... Additional arguments to be passed to \code{estfun}
##' @return A \eqn{n\times }(p+1) matrix where the first column corresponds to
##'   the scale estimate and the remaining \eqn{p} colums correspond to the
##'   coefficients
##' @author Ben B. Hansen
##' @rdname lmrob_methods
##' @exportS3Method
estfun.lmrob <- function(x, ...) {
  ctrl <- x$control
  if (!is(ctrl, "lmrobCtrl") & !inherits(ctrl, "list")) {
    stop("Model object must have a `control` element of type `lmrobCtrl` (robustbase 0.99-0 and newer) or `list`")
  }
  if (!(ctrl$method %in% c("SM", "MM"))) {
    stop("estfun.lmrob() supports only SM or MM estimates")
  }
  if (is.null(ctrl$psi)) {
    stop("parameter psi is not defined")
  }

  xmat <- stats::model.matrix(x)
  xmat <- stats::naresid(x$na.action, xmat)
  psi <- chi <- ctrl$psi
  stopifnot(is.numeric(c.chi <- ctrl$tuning.chi),
            is.numeric(c.psi <- ctrl$tuning.psi))
  r0 <- x$init$resid
  r <- x$residuals
  scale <- x$scale
  n <- length(r)
  stopifnot(n == length(r0), is.matrix(xmat), n == nrow(xmat))
  p <- ncol(xmat)
  r0.s <- r0 / scale
  w0 <- robustbase::Mchi(r0.s, cc = c.chi, psi = chi)
  Usigma <- scale(w0, center=TRUE, scale=FALSE)
  colnames(Usigma) <- "sigma"
  r.s <- r / scale
  w <- robustbase::Mpsi(r.s, cc = c.psi, psi = psi)
  Ubeta <- w * xmat
  rval <- cbind(Usigma, Ubeta)
  attr(rval, "assign") <- NULL
  attr(rval, "contrasts") <- NULL

  return(rval)
}

##' @title Extract bread matrix from an \code{lmrob()} fit
##' @details This is part of a workaround for an issue in the robustbase code
##'   affecting sandwich covariance estimation. The issue in question is issue
##'   #6471, robustbase project on R-Forge. This function contributes to
##'   providing sandwich estimates of covariance-adjusted standard errors for
##'   robust linear covariance adjustment models.
##'
##' @param x An \code{lmrob} object produced using an MM/SM estimator chain
##' @param ... Additional arguments to be passed to \code{bread}
##' @return A \eqn{p\times }(p+1) matrix where the first column corresponds to
##'   the scale estimate and the remaining \eqn{p} colums correspond to the
##'   coefficients
##' @author Ben B. Hansen
##' @rdname lmrob_methods
##' @exportS3Method
bread.lmrob <- function(x, ...) {
  ctrl <- x$control
  if (!is(ctrl, "lmrobCtrl") & !inherits(ctrl, "list")) {
    stop("Model object must have a `control` element of type `lmrobCtrl` (robustbase 0.99-0 and newer) or `list`")
  }
  if (!(ctrl$method %in% c("SM", "MM"))) {
    stop("estfun.lmrob() supports only SM or MM estimates")
  }
  if (is.null(ctrl$psi)) {
    stop("parameter psi is not defined")
  }

  psi <- chi <- ctrl$psi
  stopifnot(is.numeric(c.chi <- ctrl$tuning.chi),
            is.numeric(c.psi <- ctrl$tuning.psi))
  r0 <- x$init$resid
  r <- x$residuals
  scale <- x$scale
  xmat <- stats::model.matrix(x)
  bb <- 1 / 2
  n <- length(r)
  stopifnot(n == length(r0), is.matrix(xmat), n == nrow(xmat))
  p <- ncol(xmat)
  r.s <- r / scale
  r0.s <- r0 / scale
  w <- robustbase::Mpsi(r.s, cc = c.psi, psi = psi, deriv = 1)
  w0 <- robustbase::Mchi(r0.s, cc = c.chi, psi = chi, deriv = 1)
  x.wx <- crossprod(xmat, xmat * w)
  A <- tryCatch(solve(x.wx) * scale, error = function(e) {
    tryCatch({
      out <- solve(x.wx, tol = 0) * scale
      warning("X'WX is almost singular", call. = FALSE)
      out
    }, error = function(e) {
      stop("X'WX is singular", call. = FALSE)
    })
  })

  # At this point A has no sample size scaling, as in robustbase:::.vcov.avar1
  # The lack of scaling there precisely compensates for the lack of scaling of
  # the crossproduct
  a <- A %*% (crossprod(xmat, w * r.s) / mean(w0 * r0.s))
  colnames(a) <- "sigma"

  # Now we restore sample size scaling to A
  A <- n * A
  rval <- cbind(a, A)

  return(rval)
}

##' @title Linear Model for Intention To Treat
##'
##' @description Generates a linear model object to estimate a treatment effect,
##'   with proper estimation of variances accounting for the study
##'   specification.
##'
##' @details The first argument to [lmitt()] should be a formula specifying the
##'   outcome on the left hand side. The right hand side of the formula can be
##'   any of the following:
##'
##' \itemize{
##'   \item \code{1}: Estimates a main treatment effect.
##'   \item a subgroup variable: Estimates a treatment effect within each level
##'        of your subgrouping variable.
##'   \item a continuous moderator: Estimates a main treatment effect as well as
##'        a treatment by moderator interaction. The moderator is not
##'        automatically centered.
##' }
##'
##' Alternatively, \code{obj} can be a pre-created \code{lm} object. No
##' modification is made to the formula of the object. See the help for
##' \code{as.lmitt()} for details of this conversion.
##'
##' The [lmitt()] function's \code{subset=} argument governs
##' the subsetting of \code{data} prior to model fitting, just as with [lm()].
##' Functions such as [rct_spec()] that create \code{StudySpecification}s
##' also take an optional \code{subset=} argument, but its role differs
##' from that of the \code{subset=} argument of [lm()] or [lmitt()]. The \code{subset=}
##' argument when creating a \code{StudySpecification} restricts the data used
##' to generate the \code{StudySpecification}, but has no direct impact on the
##' future \code{lm()} or \code{lmitt()} calls using that
##' \code{StudySpecification}. (It can have an indirect impact by excluding
##' particular units from receiving a treatment assignment or weight. When
##' treatment assignments or weights are reconstructed from the \code{StudySpecification},
##' these units will receive NAs, and will be excluded from the \code{lm()}
##' or \code{lmitt()} fit under typical \code{na.action} settings.)
##'
##' To avoid variable name collision, the treatment variable defined in the
##' \code{specification} will have a "\code{.}" appended to it. For example, if
##' you request a main treatment effect (with a formula of \code{~ 1}) with a
##' treatment variable named "txt", you can obtain its estimate from the
##' returned \code{teeMod} object via \code{$coefficients["txt."]}.
##'
##' [lmitt()] will produce a message if the \code{StudySpecification}
##' designates treatment assignment by block but the blocking
##' structure appears not to be reflected in the \code{weights}, nor
##' in a block fixed effect adjustment (via \code{absorb=TRUE}). While
##' not an error, this is at odds with intended uses of
##' \code{propertee}, so \code{lmitt()} flags it as a potential
##' oversight on the part of the analyst. To disable this message, run
##' \code{options("propertee_message_on_unused_blocks" = FALSE)}.
##'
##' [lmitt()] returns objects of class \sQuote{\code{teeMod}}, for
##' Treatment Effect Estimate Model, extending the lm class to add a
##' summary of the response distribution under control (the
##' coefficients of a controls-only regression of the response on an
##' intercept and any moderator variable).  \code{teeMod} objects also
##' record the underlying \code{StudySpecification} and information
##' about any externally fitted models \code{mod} that may have been
##' used for covariance adjustment by passing
##' \code{offset=cov_adj(mod)}. In the latter case, responses are
##' offsetted by predictions from \code{mod} prior to treatment effect
##' estimation, but estimates of the response variable distribution
##' under control are calculated without reference to \code{mod}.
##'
##' The response distribution under control is also characterized when
##' treatment effects are estimated with block fixed effects, i.e. for
##' [lmitt()] with a \code{formula} first argument with option
##' \code{absorb=TRUE}.  Here as otherwise, the supplementary
##' coefficients describe a regression of the response on an intercept
##' and moderator variables, to which only control observations
##' contribute; but in this case the weights are modified for this
##' supplementary regression. The treatment effect estimates adjusted
##' for block fixed effects can be seen to coincide with estimates
##' calculated without block effect but with weights multiplied by an
##' additional factor specific to the combination of block and
##' treatment condition. For block \eqn{s} containing units with weights
##' \eqn{w_i} and binary treatment assignments \eqn{z_i}, define
##' \eqn{\hat{\pi}_s} by \eqn{\hat{\pi}_s\sum_sw_i=\sum_sz_iw_i}. If
##' \eqn{\hat{\pi}_s} is 0 or 1, the block doesn't contribute to
##' effect estimation and the additional weighting factor is 0; if
##' \eqn{0 < \hat{\pi}_s < 1}, the additional weighting factor is
##' \eqn{1 - \hat{\pi}_s} for treatment group members and
##' \eqn{\hat{\pi}_s} for controls. When estimating a main effect only
##' or a main effect with continuous moderator, supplementary
##' coefficients under option \code{absorb=TRUE} reflect regressions
##' with additional weighting factor equal to 0 or \eqn{\hat{\pi}_s},
##' respectively, for treatment or control group members of block
##' \eqn{s}. With a categorical moderator and \code{absorb=TRUE},
##' this additional weighting factor determining supplementary coefficients
##' is calculated separately for each level \eqn{\ell} of the moderator
##' variable, with the sums defining \eqn{\hat{\pi}_{s\ell}} restricted
##' not only to block \eqn{s} but also to observations with moderator
##' equal to \eqn{\ell}. 
##' 
##' 
##' @param obj A \code{formula} or a \code{lm} object. See Details.
##' @param specification The \code{StudySpecification} to be used.
##'   Alternatively, a formula creating a specification. (Of the type of that
##'   would be passed as the first argument to [rd_spec()], [rct_spec()], or
##'   [obs_spec()], with the difference that [cluster()], [uoa()] 
#'    and [unit_of_assignment()] terms can be omitted when each row of 
#'    \code{data} represents a distinct unit of assignment.) If the formula includes a [forcing()] element, an RD
##'   specification is created. Otherwise an observational specification is
##'   created. An RCT specification must be created manually using [rct_spec()].
##' @param data A \code{data.frame} such as would be passed into [lm()].
##' @param absorb If \code{TRUE}, fixed effects are included for blocks
##'   identified in the \code{StudySpecification}. Excluded in \code{FALSE}.
##'   Default is \code{FALSE}. The estimates of these fixed effects are
##'   suppressed from the returned object.
##' @param offset Offset of the kind which would be passed into [lm()]. Ideally,
##'   this should be the output of [cov_adj()].
##' @param weights Which weights should be generated? Options are \code{"ate"}
##'   or \code{"ett"}. Alternatively, the output of a manually run \code{ate()}
##'   or \code{ett()} can be used.
##' @param ... Additional arguments passed to [lm()] and other functions.
##'   An example of the latter is \code{dichotomy=}, a formula passed to
##'   [assigned()] and, as appropriate, [ate()], [att()], [atc()] or
##'   [ato()]. It is used to dichotomize a non-binary treatment
##'   variable in \code{specification}. See the Details section of the
##'   \code{ate()} help page for examples.
##' @return \code{teeMod} object (see Details)
##' @export
##' @importFrom stats lm predict weights weighted.mean reformulate residuals
##' @rdname lmitt
##' @examples
##' data(simdata)
##' spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
##' mod1 <- lmitt(y ~ 1, data = simdata, specification = spec, weights = "ate")
##' mod2 <- lmitt(y ~ as.factor(o), data = simdata, specification = spec, weights = "ate")
##' ### observational study with treatment z assigned row-wise within blocks:
##' mod3 <- lmitt(y ~ 1, data=simdata, specification=z ~ block(bid), weights="att")
##' ### regression discontinuity study with units of assignment
##' ### given by combinations of uoa1, uoa2:
##' mod4 <- lmitt(y ~ 1, data = simdata,
##'               specification = z ~ uoa(uoa1, uoa2) + forcing(force))
lmitt <- function(obj,
                  specification,
                  data,
                  ...) {
  UseMethod("lmitt")
}

##' @export
##' @rdname lmitt
lmitt.formula <- function(obj,
                          specification,
                          data,
                          absorb = FALSE,
                          offset = NULL,
                          weights = NULL,
                          ...) {
  lmitt.call <- match.call()
  m <- match(c("obj", "data", "subset", "weights", "na.action",
               "method", "model", "x", "y", "qr", "singular.ok",
               "contrasts", "offset"), names(lmitt.call), 0L)
  lm.call <- lmitt.call[c(1L, m)]
  names(lm.call)[2] <- "formula"
  lm.call[[1]] <- quote(stats::lm)
  #lmitt.call contains specification, absorb. lm.call contains only things that get
  #passed to lm, model.matrix

  data <- .as_data_frame(data)

  # Save the subset and remove it from `lm.call`, to be re-added at the enc
  savedsubset <- lm.call$subset
  lm.call$subset <- NULL

  ### Allow users to pass in "ate" and "ett" rather than functions if they have
  ### no special modifications/additional arguments
  dichotomy <- eval.parent(lmitt.call$dichotomy)
  wt <- lmitt.call$weights
  if (is(wt, "character")) {
    if (.isValidWeightAlias(wt_call <- tolower(wt))) {
      lm.call$weights <- lmitt.call$weights <- call(wt_call, dichotomy = dichotomy)
    } else {
      warning(paste0("Character other than [",
                    .listValidWeightAliases(),
                    "] passed to `weights=` argument.\nIf you are trying ",
                    "to pass a character to the internal `lm` you can ",
                    "disregard this warning.\nIf you are attempting to ",
                    "use `propertee`\'s weight generation, only [",
                    .listValidWeightAliases(),
                    "] are accepted."))
    }
  } else if (is.call(wt) & is.null(dichotomy)) {
    for (s in seq_len(sys.nframe())) {
      dichotomy <- tryCatch(eval.parent(wt$dichotomy, s), error = function(e) NULL)
      if (!is.null(dichotomy)) break()
    }
  }

  # First, make sure we have a valid `specification=` - if given a formula, make a new
  # `StudySpecification`, otherwise ensure `specification=` is `StudySpecification` class.
  if (!missing(specification)) {
    if (inherits(specification, "formula")) {
      # If there's a `forcing()`, user wants RDD. If not, force Obs. To do RCT,
      # must create StudySpecification manually.
      if (!is.null(attr(terms(specification, specials = "forcing"),
                        "specials")$forcing)) {
        spec_call <- "rd_spec"
        type <- "RD"
      } else {
        spec_call <- "obs_spec"
        type <- "Obs"
      }

      # Build new call. All calls must include obj and data
      new_spec_call <- paste0(spec_call, "(",
                              "formula = ", deparse1(specification),
                              ", data = ", deparse1(lmitt.call$data), ")")
      new_internal_call <- paste0(".new_StudySpecification(",
                                  "form = ", deparse1(specification),
                                  ", data = ", deparse1(lmitt.call$data),
                                  ", type = ", deparse1(type),
                                  ", call = str2lang(\"", new_spec_call, "\")",
                                  ", called_from_lmitt = TRUE",
                                  ")")
      # str2lang converts character into call
      specification <- eval(str2lang(new_internal_call))
    } else if (is(specification, "WeightedStudySpecification")) {
      specification <- specification@StudySpecification
    } else if (!is(specification, "StudySpecification")) {
      stop(paste("`specification=` must be an object created by `*_spec`",
                 "function, or a formula specifying such a specification"))
    }

    # #126 block on factor treatments
    if (is.factor(treatment(specification)[, 1]) & is.null(dichotomy)) {
      if (is.ordered(treatment(specification)[, 1])) {
        fact_or_ord <- "Ordered"
      } else {
        fact_or_ord <- "Factor"
      }
      stop(paste(fact_or_ord, "treatment variables are not supported, use",
                 "`dichotomy=` to define a binary treatment."))
    }
    
    # #205 block on continuous treatments
    if (!has_binary_treatment(specification) & is.null(dichotomy)) {
      stop("Specify a dichotomy when estimating effects of a continuous treatment variable")
    }
  }


  # Extract formula bits
  rhs <- trimws(strsplit(deparse1(obj[[3]]), "+", fixed = TRUE)[[1]])
  rhs <- rhs[rhs != "+"]
  if (length(rhs) > 1) {
    # At max two terms ("sbgrp" and "1")
    stop(paste("Too many terms on right hand side. To estimate subgroup",
               "effects and main effect simultaneously, fit two",
               "separate models."))
  }
  if (rhs == "0") {
    # We don't support y ~ 0
    stop("'0' is not a valid entry on right hand side")
  }
  .check_for_assigned_and_aliases <- function(fn) {
    if (grepl(paste0(fn, "\\("), rhs)) {
      stop(paste("Do not specify `assigned()` or any of its aliases in",
                 "the right hand side of `lmitt()`.\nTo estimate only",
                 "a treatment effect, pass `~ 1` as the right hand side."))
    }
  }
  .check_for_assigned_and_aliases("assigned")
  .check_for_assigned_and_aliases("adopters")
  .check_for_assigned_and_aliases("a\\.")
  .check_for_assigned_and_aliases("z\\.")

  # `rhs` is now either "1" or subgrouping variable

  # Get weights and offset if passed
  lm.call$weights <- eval.parent(lm.call$weights)
  lm.call$offset <- eval.parent(lm.call$offset)


  # Ensure same specification is used in weights and offset, if they're there
  wtspec <- try(lm.call$weights@StudySpecification, silent = TRUE)
  offspec <- try(lm.call$offset@StudySpecification, silent = TRUE)

  if (missing(specification) & is(wtspec, "StudySpecification")) {
    stop(paste("You've passed a `StudySpecification` into the weight function (`ate()`",
                "or `ett()`) but not the `lmitt()` call. Please pass the",
                "`StudySpecification` into the `specification=` argument of `lmitt()`. It is",
                "not needed in `ate()` or `ett()` when passed as the",
                "`specification=` argument to `lmitt()`."))
  }

  if (missing(specification) & is(offspec, "StudySpecification")) {
    stop(paste("You've passed a `StudySpecification` into the offset function",
                " (`cov_adj()`) but not the `lmitt()` call. Please pass the",
                "`StudySpecification` into the `specification=` argument of `lmitt()`. It is",
                "not needed in `cov_adj()` when passed as the `specification=`",
                "argument to `lmitt()`."))
  }

  if (is(wtspec, "StudySpecification")) {
    if (!identical_StudySpecifications(specification, wtspec)) {
      stop(paste("Multiple differing `StudySpecification` found (`specification` argument to",
                 " `lmitt` and `specification` object inside the weights differ)."))
    }
  }

  if (is(offspec, "StudySpecification")) {
    if (!identical_StudySpecifications(specification, offspec)) {
      stop(paste("Multiple differing `StudySpecification` found (`specification` argument to",
                 " `lmitt` and `specification` object inside the offset differ)."))
    }
  }

  if (!is(lm.call$weights, "WeightedStudySpecification") & !absorb) {
    if ("b" %in% specification@column_index) {
      if (options()$propertee_message_on_unused_blocks) {
        message(paste("The StudySpecification object contains block-level information,",
                      "but it is not used in this model. Block information",
                      "is used when weights are defined via `ate()` or `ett()`",
                      "or if the `absorb=TRUE` argument is passed."))
      }
    }
  }

  blocks <- blocks(specification,
                  eval(lm.call$data, parent.frame()),
                  all.x = TRUE,
                  implicit = TRUE)[,1]

  # Identify whether RHS is intercept, continuous moderator, or subgroup
  if (rhs != "1") {
    new.form <- reformulate(rhs, intercept = FALSE)
    modfinder.call <- lm.call
    modfinder.call[[2]] <- str2lang(deparse1(new.form))
    modfinder.call[[1]] <- quote(stats::model.matrix)
    names(modfinder.call)[2] <- "object"
    numcol <- ncol(eval(modfinder.call, parent.frame()))
    if (numcol > 1) {
      rhstype <- "categorical"
    } else {
      rhstype <- "continuous"
    }
  } else {
    rhstype <- "intercept"
  }

  # Generate formula for the internal `lm`
  if (rhstype == "intercept" | rhs == var_names(specification, "t")) {
    new.form <- stats::reformulate(paste0("a.(dichotomy=", deparse1(dichotomy), ")"))
    moderator <- character()
    ctrl_means_form <- stats::reformulate("1", response = "lhs")
  } else if (rhstype == "categorical") {
    new.form <- stats::reformulate(paste0("a.(dichotomy=", deparse1(dichotomy), "):",
                                          rhs, "+", rhs))
    moderator <- rhs
    ctrl_means_form <- stats::reformulate(rhs, response = "lhs", intercept = FALSE)
  } else {
    new.form <- stats::reformulate(paste0("a.(dichotomy=", deparse1(dichotomy), ") + ",
                                          "a.(dichotomy=", deparse1(dichotomy), "):",
                                          rhs, "+", rhs))
    moderator <- rhs
    ctrl_means_form <- stats::reformulate(rhs, response = "lhs")
  }
  ctrl_means_env <- new.env()
  assign("ctrl_means_data", data, envir = ctrl_means_env)
  environment(ctrl_means_form) <- ctrl_means_env
  
  mm.call <- lm.call
  mm.call[[2]] <- str2lang(paste(deparse1(new.form, width.cutoff = 500L), collapse = ""))
  # model.matrix.lm supports as `na.action` argument where
  # model.matrix.default doesn't
  mm.call[[1]] <- quote(stats::model.matrix.lm)
  mm.call$na.action <- "na.pass"
  names(mm.call)[2] <- "object"
  mm <- eval(mm.call, parent.frame())
  colnames(mm) <- gsub(paste0("dichotomy = ", deparse1(dichotomy)), "", colnames(mm),
                       fixed = TRUE)

  if (absorb) {
    mm <- areg.center(mm, as.factor(blocks), lm.call$weights)
  }

  # Strip intercept from data if it's in there
  mm <- mm[, !grepl("(Intercept)", colnames(mm)), drop = FALSE]
  # Make sure to keep it as a named matrix

  # Rename `a.()` to treatment named, adding a "." after to avoid conflict with
  # original treatment variable.
  colnames(mm) <- gsub("a\\.\\(\\)",
                       paste0(var_names(specification, "t"), "."),
                       colnames(mm))

  # Replace : with _ for interaction to try and avoid backticks
  colnames(mm) <- gsub("\\:", "_", colnames(mm))
  # This isn't foolproof as the sbgrp variable can cause them too
  # e.g. `as.factor(sbgrp)` forces backticks.

  # Rebuild RHS of formula: RHS is "txt_" or "txt_sbgrp", where "txt" has been
  # replaced with the actual treatment name above.
  lm_form <- eval(lm.call$formula, envir = parent.frame())
  lm.call$formula <- str2lang(paste(lm_form[[2]], " ~ ",
                                    paste(
                                      paste0("`", colnames(mm), "`"),
                                      collapse = "+"),
                                    collapse = ""))

  
  # If the moderators already exist in the passed-in data, remove them to avoid
  # either a) overloading variable names, or b) an error if data is a
  # `grouped_df` (see #137)
  data <- data[, !(names(data) %in% colnames(mm))]
  
  # Data for model should be original data, plus updated RHS (mm),
  lm.call$data <- cbind(data, mm)

  # restore subset
  lm.call$subset <- savedsubset

  model <- eval(lm.call, parent.frame())
  
  # set call's na.action to na.pass so expand.model.frame includes NA rows
  model$call$na.action <- "na.pass"
  
  # `&&` necessary to return FALSE immediately if not enough frames on stack
  if (sys.nframe() >= 2 &&
      !is.null(sys.call(-1)) &&
      sys.call(-1)[[1]] == as.name("lmitt")) {
    # If we're in `lmitt.formula()` via `lmitt()`, save that call.
    lmitt_call <- sys.call(-1)
  } else {
    # Otherwise save the `lmitt.formula()` call
    lmitt_call <- sys.call()
  }
  return(.convert_to_lmitt(model,
                           specification,
                           lmitt_fitted = TRUE,
                           moderator = moderator,
                           absorbed_intercepts = absorb,
                           ctrl_means_form = ctrl_means_form,
                           dichotomy = dichotomy,
                           lmitt_call = lmitt_call))
}

##' @export
##' @rdname lmitt
lmitt.lm <- function(obj,
                     specification = NULL,
                     ...) {
  return(as.lmitt(obj, specification))
}

if (requireNamespace("multcomp", quietly = TRUE)) {
  test_that(".make_uoa_ids for mmm object", {
    set.seed(39)
    fct_df <- data.frame(id = seq_len(13),
                         a = factor(c(rep(0, 3), rep(1, 6), rep(2, 4))),
                         x = rnorm(13),
                         y = rnorm(13))
    ca_df <- data.frame(id = seq(14, 26), x = rnorm(13), y = rnorm(13))
    
    ca <- lm(y ~ x, ca_df)
    spec <- rct_spec(a ~ unitid(id), fct_df)
    
    a1 <- lmitt(y ~ 1, spec, fct_df, dichotomy = a == 1 ~ a == 0)
    a2 <- lmitt(y ~ 1, spec, fct_df, dichotomy = a == 2 ~ a == 0)
    a1lm <- lmitt(lm(y ~ assigned(spec, fct_df, a == 1 ~ a == 0), fct_df), spec)
    a2lm <- lmitt(lm(y ~ assigned(spec, fct_df, a == 2 ~ a == 0), fct_df), spec)
    
    mods <- multcomp::mmm(a1, a2)
    expect_true(inherits(ids <- .make_uoa_ids(mods, "CR0"), "factor"))
    expect_equal(length(ids), 13)
    
    mods <- multcomp::mmm(a1lm, a2lm)
    expect_true(inherits(ids <- .make_uoa_ids(mods, "CR0"), "factor"))
    expect_equal(length(ids), 13)
    
    a1 <- lmitt(y ~ 1, spec, fct_df, offset = cov_adj(ca, newdata = fct_df), dichotomy = a == 1 ~ a == 0)
    a2 <- lmitt(y ~ 1, spec, fct_df, offset = cov_adj(ca, newdata = fct_df), dichotomy = a == 2 ~ a == 0)
    a1lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 1 ~ a == 0), fct_df,
         off = cov_adj(ca, newdata = fct_df, sp = spec)),
      spec)
    a2lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 2 ~ a == 0), fct_df,
         off = cov_adj(ca, newdata = fct_df, sp = spec)),
      spec)
    
    mods <- multcomp::mmm(a1, a2)
    expect_true(inherits(ids <- .make_uoa_ids(mods, "CR0"), "factor"))
    expect_equal(length(ids), 26)
    
    mods <- multcomp::mmm(a1lm, a2lm)
    expect_true(inherits(ids <- .make_uoa_ids(mods, "CR0"), "factor"))
    expect_equal(length(ids), 26)
  })
  
  test_that("factor treatment variables (#78) no weights, no cov adj", {
    set.seed(39)
    fct_df <- data.frame(id = seq_len(13),
                         a = factor(c(rep(0, 3), rep(1, 6), rep(2, 4))),
                         y = rnorm(13))
    spec <- rct_spec(a ~ unitid(id), fct_df)
    
    a1 <- lmitt(y ~ 1, spec, fct_df, dichotomy = a == 1 ~ a == 0)
    a2 <- lmitt(y ~ 1, spec, fct_df, dichotomy = a == 2 ~ a == 0)
    a1lm <- lmitt(lm(y ~ assigned(spec, fct_df, a == 1 ~ a == 0), fct_df), spec)
    a2lm <- lmitt(lm(y ~ assigned(spec, fct_df, a == 2 ~ a == 0), fct_df), spec)
    
    # check pt estimates of dichotomized mods
    grp_means <- rowsum(fct_df$y, fct_df$a) / rowsum(numeric(13) + 1, fct_df$a)
    expect_true(all.equal(a1$coefficients[2], grp_means[2,] - grp_means[1,], check.attributes = FALSE))
    expect_true(all.equal(a1$coefficients[2], a1lm$coefficients[2], check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], grp_means[3,] - grp_means[1,], check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], a2lm$coefficients[2], check.attributes = FALSE))
    
    # check vcov matrix for lmitt.formula
    mods <- multcomp::mmm(a1, a2)
    n <- nrow(fct_df)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
    
    # check vcov matrix for lmitt.lm
    mods <- multcomp::mmm(a1lm, a2lm)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
  })
  
  test_that("factor treatment variables (#78) weights, no cov adj", {
    set.seed(39)
    fct_df <- data.frame(id = seq_len(13),
                         a = factor(c(rep(0, 3), rep(1, 6), rep(2, 4))),
                         y = rnorm(13))
    spec <- rct_spec(a ~ unitid(id), fct_df)
    
    a1 <- lmitt(y ~ 1, spec, fct_df, w = ate(spec), dichotomy = a == 1 ~ a == 0)
    a2 <- lmitt(y ~ 1, spec, fct_df, w = ate(spec), dichotomy = a == 2 ~ a == 0)
    a1lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 1 ~ a == 0), fct_df,
         w = ate(spec, dichotomy = a == 1 ~ a == 0, data = fct_df)),
      spec)
    a2lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 2 ~ a == 0), fct_df,
         w = ate(spec, dichotomy = a == 2 ~ a == 0, data = fct_df)),
      spec)
    
    # check pt estimates of dichotomized mods
    grp_means_d1 <- rowsum(a1$model$y * a1$weights, a1$model$a.) / rowsum(a1$weights, a1$model$a.)
    grp_means_d2 <- rowsum(a2$model$y * a2$weights, a2$model$a.) / rowsum(a2$weights, a2$model$a.)
    expect_true(all.equal(a1$coefficients[2], drop(c(-1, 1) %*% grp_means_d1), check.attributes = FALSE))
    expect_true(all.equal(a1$coefficients[2], a1lm$coefficients[2], check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], drop(c(-1, 1) %*% grp_means_d2), check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], a2lm$coefficients[2], check.attributes = FALSE))
    
    # check vcov matrix for lmitt.formula
    mods <- multcomp::mmm(a1, a2)
    n <- nrow(fct_df)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
    
    # check vcov matrix for lmitt.lm
    mods <- multcomp::mmm(a1lm, a2lm)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
  })
  
  test_that("factor treatment variables (#78) no weights, cov adj", {
    set.seed(39)
    fct_df <- data.frame(id = seq_len(13),
                         a = factor(c(rep(0, 3), rep(1, 6), rep(2, 4))),
                         x = rnorm(13),
                         y = rnorm(13))
    ca_df <- data.frame(id = seq(14, 26), x = rnorm(13), y = rnorm(13))
    spec <- rct_spec(a ~ unitid(id), fct_df)
    
    ca <- lm(y ~ x, ca_df)
    a1 <- lmitt(y ~ 1, spec, fct_df, offset = cov_adj(ca, newdata = fct_df), dichotomy = a == 1 ~ a == 0)
    a2 <- lmitt(y ~ 1, spec, fct_df, offset = cov_adj(ca, newdata = fct_df), dichotomy = a == 2 ~ a == 0)
    a1lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 1 ~ a == 0), fct_df,
         off = cov_adj(ca, newdata = fct_df, sp = spec)),
      spec)
    a2lm <- lmitt(
      lm(y ~ assigned(spec, fct_df, a == 2 ~ a == 0), fct_df,
         off = cov_adj(ca, newdata = fct_df, sp = spec)),
      spec)
    
    # check pt estimates of dichotomized mods
    grp_means <- rowsum(fct_df$y - cov_adj(ca, newdata = fct_df, sp = spec), fct_df$a) / rowsum(numeric(13) + 1, fct_df$a)
    expect_true(all.equal(a1$coefficients[2], grp_means[2,] - grp_means[1,], check.attributes = FALSE))
    expect_true(all.equal(a1$coefficients[2], a1lm$coefficients[2], check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], grp_means[3,] - grp_means[1,], check.attributes = FALSE))
    expect_true(all.equal(a2$coefficients[2], a2lm$coefficients[2], check.attributes = FALSE))
    
    # check vcov matrix for lmitt.formula
    mods <- multcomp::mmm(a1, a2)
    n <- nrow(fct_df) + nrow(ca_df)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0", cov_adj_rcorrect = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
    
    # check vcov matrix for lmitt.lm
    mods <- multcomp::mmm(a1lm, a2lm)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0", cov_adj_rcorrect = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
  })
  
  test_that("combine lmitts (#79) for different DV's", {
    set.seed(39)
    Q_df <- data.frame(id = seq_len(13),
                       a = rbinom(13, 1, 0.5),
                       x = rnorm(13),
                       y1 = rnorm(13),
                       y2 = rnorm(13))
    ca_df <- data.frame(id = seq(14, 26), x = rnorm(13), y1 = rnorm(13), y2 = rnorm(13))
    spec <- rct_spec(a ~ unitid(id), Q_df)
    
    ca1 <- lm(y1 ~ x, ca_df)
    ca2 <- lm(y2 ~ x, ca_df)
    a1 <- lmitt(y1 ~ 1, spec, Q_df, offset = cov_adj(ca1, newdata = Q_df))
    a2 <- lmitt(y2 ~ 1, spec, Q_df, offset = cov_adj(ca2, newdata = Q_df))
    a1lm <- lmitt(
      lm(y1 ~ assigned(spec, Q_df), Q_df,
         off = cov_adj(ca1, newdata = Q_df, sp = spec)),
      spec)
    a2lm <- lmitt(
      lm(y2 ~ assigned(spec, Q_df), Q_df,
         off = cov_adj(ca2, newdata = Q_df, sp = spec)),
      spec)
    
    # check vcov matrix for lmitt.formula
    mods <- multcomp::mmm(a1, a2)
    n <- nrow(Q_df) + nrow(ca_df)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0", cov_adj_rcorrect = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
    
    # check vcov matrix for lmitt.lm
    mods <- multcomp::mmm(a1lm, a2lm)
    tsts <- multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1, 0, 0), nrow = 1)),
                           vcov = function(x) vcov_tee(x, type = "HC0", cov_adj_rcorrect = "HC0"))
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      n / (n-1) * (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
  })
  
  test_that("combine lmitts (#79) for different moderator variables", {
    set.seed(39)
    Q_df <- data.frame(id = seq_len(45),
                       b = rep(LETTERS[1:3], each = 15),
                       a = rep(c(rep(0, 8), rep(1, 7)), 3),
                       x1 = factor(sample(seq_len(3), 45, re = TRUE)),
                       x2 = factor(sample(letters[1:2], 45, re = TRUE)),
                       y = rnorm(45))
    ca_df <- data.frame(id = seq(46, 58),
                        x1 = factor(sample(seq_len(3), 13, re = TRUE)),
                        x2 = factor(sample(letters[1:2], 13, re = TRUE)),
                        y = rnorm(13))
    spec <- rct_spec(a ~ unitid(id) + block(b), Q_df)
    
    ca <- lm(y ~ x1 + x2, ca_df)
    a1 <- lmitt(y ~ x1, spec, Q_df, w = ate(spec), offset = cov_adj(ca, newdata = Q_df))
    a2 <- lmitt(y ~ x2, spec, Q_df, w = ate(spec), offset = cov_adj(ca, newdata = Q_df))
    
    # check vcov matrix for lmitt.formula
    mods <- multcomp::mmm(a1, a2)
    n <- nrow(Q_df) + nrow(ca_df)
    tsts <- multcomp::glht(
      mods, rbind(c(rep(0, 3), 1, rep(0, 16)), c(rep(0, 14), 1, rep(0, 5))),
      vcov. = function(x, ...) vcov_tee(x, type = "HC0", cov_adj_rcorrect = "HC0", ...), cadjust = FALSE)
    vmat <- summary(tsts)$vcov
    expect_true(all.equal(
      vmat,
      (1 / n) * (bread(mods) %*% (crossprod(estfun(mods)) / n) %*% bread(mods)),
      check.attributes = FALSE
    ))
  })
  
  test_that("error if Q or C df's don't match across models in mmm object", {
    on.exit(options("show.error.messages" = TRUE))
    options("show.error.messages" = FALSE)
    set.seed(39)
    Q_df1 <- data.frame(id = seq_len(45),
                        b = rep(LETTERS[1:3], each = 15),
                        a = rep(c(rep(0, 8), rep(1, 7)), 3),
                        x1 = factor(sample(seq_len(3), 45, re = TRUE)),
                        y1 = rnorm(45),
                        y2 = rnorm(45))
    Q_df2 <- data.frame(id = seq_len(60),
                        b = rep(LETTERS[1:3], each = 20),
                        a = rep(c(rep(0, 8), rep(1, 12)), 3),
                        x1 = factor(sample(seq_len(3), 60, re = TRUE)),
                        y1 = rnorm(60),
                        y2 = rnorm(60))
    ca_df1 <- data.frame(id = seq(46, 58),
                         x1 = factor(sample(seq_len(3), 13, re = TRUE)),
                         y1 = rnorm(13),
                         y2 = rnorm(13))
    ca_df2 <- data.frame(id = seq(46, 59),
                         x1 = factor(sample(seq_len(3), 14, re = TRUE)),
                         y1 = rnorm(14),
                         y2 = rnorm(14))
    spec1 <- rct_spec(a ~ unitid(id) + block(b), Q_df1)
    spec2 <- rct_spec(a ~ unitid(id) + block(b), Q_df2)
    
    a1 <- lmitt(y1 ~ 1, spec1, Q_df1, w = ate(spec1))
    a2 <- lmitt(y2 ~ 1, spec2, Q_df2, w = ate(spec2))
    mods <- multcomp::mmm(a1, a2)
    expect_error(
      multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1), nrow = 1)), vcov = vcov_tee, cadjust = FALSE)
    )
    
    ca1 <- lm(y1 ~ x1, ca_df1)
    ca2 <- lm(y2 ~ x1, ca_df2)
    a1 <- lmitt(y1 ~ 1, spec1, Q_df1, w = ate(spec1), offset = cov_adj(ca1))
    a2 <- lmitt(y2 ~ 1, spec1, Q_df1, w = ate(spec1), offset = cov_adj(ca2))
    mods <- multcomp::mmm(a1, a2)
    expect_error(
      multcomp::glht(mods, multcomp::mlf(matrix(c(0, 1), nrow = 1)), vcov = vcov_tee, cadjust = FALSE)
    )
  })
}

test_that("Treatment inputs", {
  data(simdata)

  # numeric 0/1 treatment
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.numeric(treatment(spec)[, 1]))
  expect_true(has_binary_treatment(spec))
  expect_silent(mod <- lmitt(y ~ 1, data = simdata, specification = spec))

  # numeric non-0/1
  spec <- rct_spec(o ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.numeric(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(mod <- lmitt(y ~ 1, data = simdata, specification = spec),
               "continuous treatment")

  # factor 0/1
  simdata$foo <- as.factor(simdata$z)
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Factor treatment")

  # factor numeric non 0/1
  simdata$foo <- as.factor(simdata$o)
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Factor treatment")

  # factor string levels
  simdata$foo <- factor(letters[simdata$o])
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Factor treatment")

  # ordinal 0/1
  simdata$foo <- as.ordered(simdata$z)
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_true(is.ordered(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Ordered treatment")

  # ordinal numeric non 0/1
  simdata$foo <- as.ordered(simdata$o)
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_true(is.ordered(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Ordered treatment")

  # ordinal string levels
  simdata$foo <- ordered(letters[simdata$o])
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.factor(treatment(spec)[, 1]))
  expect_true(is.ordered(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec),
               "Ordered treatment")

  # character
  simdata$foo <- letters[simdata$o]
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.character(treatment(spec)[, 1]))
  expect_false(has_binary_treatment(spec))

  # logical
  simdata$foo <- as.logical(simdata$z)
  spec <- rct_spec(foo ~ cluster(uoa1, uoa2), data = simdata)
  expect_true(is.logical(treatment(spec)[, 1]))
  expect_true(has_binary_treatment(spec))

  # conditional
  spec <- rct_spec(o > 2 ~ uoa(uoa1, uoa2), data = simdata)
  expect_true(is.logical(treatment(spec)[, 1]))
  expect_true(has_binary_treatment(spec))

  old_opt <- options()
  on.exit(options(old_opt))
  options(propertee_warn_on_conditional_treatment = FALSE)
  spec <- rct_spec(o > 2 ~ uoa(uoa1, uoa2), data = simdata)
  expect_true(is.logical(treatment(spec)[, 1]))
  expect_true(has_binary_treatment(spec))
  options(old_opt)

  # Non-conditional but bad variable name
  names(simdata)[5] <- "z==1"
  expect_error(spec <- rct_spec(`z==1`  ~ uoa(uoa1, uoa2),
                                data = simdata),
               "isn't logical")


  # tests for other things...?

  z <- c(10485849600, 10477641600, 10561104000, 10562745600)
  dat <- data.frame(z = as.Date(as.POSIXct(z, origin = "1582-10-14", tz = "GMT")),
                    unit = 1:4)

  expect_warning(rct_spec(z ~ unitid(unit), data = dat),
                 "STRONGLY")

})

test_that("Same result with multiple variable for blocks", {
  data(simdata)
  spec1 <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  wts1 <- ate(spec1, data = simdata)

  mod1 <- lmitt(y ~ x, data = simdata, spec = spec1, weights = wts1)
  coeff1 <- summary(mod1)$coeff


  simdata2 <- simdata
  simdata2$bid1 <- rep(0:1, times = c(34, 16))
  simdata2$bid2 <- rep(0:1, times = c(20, 30))

  # Make sure block info is same, only diagonal entries of cross table between
  # original blocks and new blocks should have values.
  expect_true(sum(diag(with(simdata2, table(bid, paste(bid1, bid2))))) == nrow(simdata2))

  spec2 <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid1, bid2), data = simdata2)
  expect_true(ncol(blocks(spec2)) == 2)

  wts2 <- ate(spec2, data = simdata2)

  expect_true(all(wts1 == wts2))

  mod2 <- lmitt(y ~ x, data = simdata2, spec = spec2, weights = wts2)
  coeff2 <- summary(mod2)$coeff

  all(coeff1 == coeff2)

  simdata3 <- simdata
  simdata3$bid1 <- simdata$bid == 1
  simdata3$bid2 <- simdata$bid == 2
  simdata3$bid3 <- simdata$bid == 3

  # Make sure block info is same, only diagonal entries of cross table between
  # original blocks and new blocks should have values.
  expect_true(sum(diag(with(simdata3, table(bid, paste(bid3, bid2, bid1))))) == nrow(simdata3))
  # Changed order of bid3-bid1 because of sorting

  spec3 <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid1, bid2, bid3), data = simdata3)
  expect_true(ncol(blocks(spec3)) == 3)

  wts3 <- ate(spec3, data = simdata3)

  expect_true(all(wts1 == wts3))

  mod3 <- lmitt(y ~ x, data = simdata3, spec = spec3, weights = wts3)
  coeff3 <- summary(mod3)$coeff

  all(coeff1 == coeff3)


})

test_that("find and validate dichotomies with no upstream dichotomies and list of NULLs argument", {
  expect_true(is.null(.validate_dichotomy(list(lmitt = NULL, weights = NULL))))
})

test_that("find and validate dichotomies with no upstream dichotomies and non-null argument", {
  expect_identical(deparse1(.validate_dichotomy(z ~ 1)), deparse1(z ~ 1))
})

test_that("find and validate dichotomies doesn't get valid option", {
  expect_error(.validate_dichotomy(character()), "formulas provided")
})

test_that("find and validate dichotomies with upstream dichotomy in lmitt.formula()", {
  data(simdata)
  spec <- rct_spec(dose ~ cluster(uoa1, uoa2), simdata)
  expect_warning(
    expect_warning(
      expect_warning(
        mod1 <- lmitt(y ~ 1, specification = spec, data = simdata, dichotomy = dose > 50 ~ dose == 50,
                      weights = ate(spec, dichotomy = dose <= 250 ~ dose > 250)),
        "passed to `.weights_calc()"
      ),
      "passed to `a.()"
    ),
    "passed to `a.()"
  )
  expect_equal(mod1$model$dose., as.numeric(simdata$dose > 50))
  expect_equal(mod1$model$`(weights)`, ate(spec, dose <= 250 ~ dose > 250, data = simdata))

  mod2 <- lmitt(y ~ 1, specification = spec, data = simdata, dichotomy = dose <= 250 ~ dose > 250,
                weights = ate(spec, dichotomy = dose <= 250 ~ dose > 250))
  expect_equal(mod2$model$dose., as.numeric(simdata$dose <= 250))
  expect_equal(mod2$model$`(weights)`, ate(spec, dose <= 250 ~ dose > 250, data = simdata))

  mod3 <- lmitt(y ~ 1, specification = spec, data = simdata, dichotomy = dose <= 250 ~ dose > 250,
                weights = ate(spec))
  expect_equal(mod3$model$dose., mod2$model$dose.)
  expect_equal(mod3$model$`(weights)`@.Data, mod2$model$`(weights)`@.Data)
})

test_that("find and validate dichotomies with no upstream lmitt.formula()", {
  data(simdata)
  spec <- rct_spec(dose ~ cluster(uoa1, uoa2), simdata)
  expect_warning(
    mod1 <- lm(y ~ assigned(dichotomy = dose > 50 ~ dose == 50), data = simdata,
               weights = ate(spec, dichotomy = dose <= 250 ~ dose > 250)),
    "passed to `assigned()"
  )
  expect_equal(mod1$model[[2]], as.numeric(simdata$dose > 50))
  expect_equal(mod1$model$`(weights)`@.Data,
               ate(spec, dose <= 250 ~ dose > 250, data = simdata)@.Data)
  expect_equal(mod1$model$`(weights)`@target,
               ate(spec, dose <= 250 ~ dose > 250, data = simdata)@target)
  expect_equal(deparse1(mod1$model$`(weights)`@dichotomy),
               deparse1(ate(spec, dose <= 250 ~ dose > 250, data = simdata)@dichotomy))


  mod2 <- lm(y ~ assigned(dichotomy = dose <= 250 ~ dose > 250), data = simdata,
             weights = ate(spec, dichotomy = dose <= 250 ~ dose > 250))
  expect_equal(mod2$model[[2]], as.numeric(simdata$dose <= 250))
  expect_equal(mod2$model$`(weights)`@.Data,
               ate(spec, dose <= 250 ~ dose > 250, data = simdata)@.Data)
  expect_equal(mod2$model$`(weights)`@target,
               ate(spec, dose <= 250 ~ dose > 250, data = simdata)@target)
  expect_equal(deparse1(mod2$model$`(weights)`@dichotomy),
               deparse1(ate(spec, dose <= 250 ~ dose > 250, data = simdata)@dichotomy))

  expect_error(
    mod3 <- lm(y ~ assigned(dichotomy = dose > 50 ~ dose == 50), data = simdata,
               weights = ate(spec)),
    "Must provide a dichotomy"
  )
})

test_that("no warning when dichotomy is passed as a stored object", {
  set.seed(33)
  uoadata <- data.frame(cid = seq_len(8),
                        bid = c(rep(seq_len(2), 2), rep(1, 2), rep(2, 2)),
                        dose = c(rep(0, 4), rep(1, 2), rep(2, 2)))
  individdata <- data.frame(cid = rep(seq_len(8), each = 10), y = rnorm(80))
  specification <- rct_spec(dose ~ uoa(cid) + block(bid), data = uoadata)
  
  # store the dichotomy in an object and create weights ahead of time with stored dichotomy
  dich <- dose == 1 ~ dose == 0
  wts <- ett(specification, data = individdata, dichotomy = dich)
  
  # first, check when dichotomy is passed directly and through the weights
  expect_silent(om1 <- lmitt(y~1, specification, individdata, weights = wts, dichotomy = dich))
  
  # then check when only passed through the weights
  expect_silent(om2 <- lmitt(y~1, specification, individdata,
                             weights = ett(specification, data = individdata, dichotomy = dich)))
  
  # also, check when the dichotomy or an object storing a dichotomy is passed to a function
  fit_lmitt <- function(data, spec, d) {
    wts <- ett(spec, data = data, dichotomy = d)
    m1 <- lmitt(y~1, spec, data, weights = wts, dichotomy = d)
    m1
  }
  expect_silent(om3 <- fit_lmitt(individdata, specification, dose == 1 ~ dose == 0))
  expect_silent(om4 <- fit_lmitt(individdata, specification, dich))
  
  expect_equal(om1$coef, om2$coef)
  expect_equal(om1$coef, om3$coef)
  expect_equal(om1$coef, om4$coef)
  
  # and finally, non-lmitt calls
  expect_silent(om5 <- lm(y ~ assigned(specification, individdata, dich), individdata, weights = wts))
  expect_equal(unname(om1$coef[1:2]), unname(om5$coef))
})

test_that("proceed with NULL dichotomy when passed as a named object", {
  fit_lmitt <- function(data, spec, dich) {
    m1 <- lmitt(y~1, spec, data, dichotomy = dich)
    m1
  }
  
  testdata <- data.frame(cid = 1:10, z = c(rep(1, 5), rep(0, 5)), y = rnorm(10))
  spec <- rct_spec(z ~ unit_of_assignment(cid), data = testdata)
  
  expect_silent(om1 <- lmitt(y~1, spec, testdata))
  expect_silent(om2 <- lmitt(y~1, spec, testdata, dichotomy = NULL))
  expect_silent(om3 <- fit_lmitt(testdata, spec, z == 1 ~ z == 0))
  expect_silent(om4 <- fit_lmitt(testdata, spec, NULL))
  expect_equal(om1$coefficients, om2$coefficients)
  expect_equal(om1$coefficients, om3$coefficients)
  expect_equal(om1$coefficients, om4$coefficients)
})

test_that("fail when dichotomy is a named but non-existent object", {
  testdata <- data.frame(cid = 1:10, z = c(rep(1, 5), rep(0, 5)), y = rnorm(10))
  spec <- rct_spec(z ~ unit_of_assignment(cid), data = testdata)
  expect_error(ett(spec, data = testdata, dichotomy = quote(d)),
               "Could not find d in call stack")
})

test_that("issue 60", {
  data(simdata)

  # First, subset that does NOT exclude any clusters
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata,
                    subset = 1:50 <= 47)
  expect_equal(nrow(get_structure(spec)), 10)
  # By passing the subset
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec)


  # Now, subset that excludes clusters
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata,
                    subset = 1:50 <= 33)
  # This removes 3 complete clusters, and 1 row from a clsuter that remains,
  # leaving 33 rows of data with information about 7 clusters
  expect_equal(nrow(unique(simdata[1:50 <= 33, c("uoa1", "uoa2")])),
               nrow(get_structure(spec)))

  # The model should use all rows related to those 7 clusters, which is 34.
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec)

  # This counts the number of rows in simdata whose clusters are found in the
  # specification. (Instead of hardcoding the number, in case simdata changes)
  num_rows_with_clusters_in_spec <-
    sum(apply(simdata[, c("uoa1", "uoa2")], 1, paste, collapse = "") %in%
          apply(units_of_assignment(spec), 1, paste, collapse = ""))

  expect_equal(num_rows_with_clusters_in_spec,
               nrow(mod1$model))

})


test_that("issue 60, with NAs", {
  data(simdata)
  simdata$uoa1[1:10] <- simdata$uoa2[1:10] <- NA_integer_
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata,
                    subset = !is.na(simdata$uoa1) & !is.na(simdata$uoa2))
  expect_silent(damod <- lmitt(y ~ x, data = simdata,
                               subset = !is.na(uoa1) & !is.na(uoa2),
                               specification = spec, weights = ate()))
  # Just testing that this doesn't error/warn
})

#' @include StudySpecification.R WeightedStudySpecification.R StudySpecificationAccessors.R SandwichLayerVariance.R confint_lm.R
NULL
# The above ensures that `StudySpecification`, `WeightedStudySpecification`, and
# `vcov_tee` are defined prior to `teeMod`

setClass("teeMod",
         contains = "lm",
         slots = c(StudySpecification = "StudySpecification",
                   lmitt_fitted = "logical",
                   absorbed_intercepts = "logical",
                   moderator = "character",
                   ctrl_means_model = "lm",
                   lmitt_call = "call"))

setValidity("teeMod", function(object) {
  if (length(object@lmitt_fitted) != 1) {
    return("@lmitt_fitted slot must be a single logical")
  }
  return(TRUE)
})

##' @title Show a \code{teeMod}
##' @description Display information about a \code{teeMod} object
##' @param object \code{teeMod} object, usually a result of a call to
##'   [lmitt()].
##' @return \code{object}, invisibly.
##' @export
setMethod("show", "teeMod", function(object) {
  coeffs <- object$coefficients
  ## Display only treatment effects, intercepts from supplementary
  ## regression of same y but w/o the covadj offset
  if (object@lmitt_fitted) {
    toprint <- (
  ## This should match any coefficients starting with the "txt." or "`txt."
      grepl(paste0("^\\`?", var_names(object@StudySpecification, "t"), "\\."), names(coeffs)) |
  ## This should match the coefficients of the supplementary regression
        grepl(paste0("^", formula(object)[[2L]], ":"), names(coeffs))
    )
    print(coeffs[toprint])
  } else {
    print(coeffs)
  }
  invisible(object)
})

##' @title Compute variance-covariance matrix for fitted \code{teeMod} model
##' @description An S3method for \code{stats::vcov} that computes standard
##'   errors for \code{teeMod} models using \code{vcov_tee()}.
##' @details \code{vcov.teeMod()} wraps around \code{vcov_tee()}, so additional
##'   arguments passed to \code{...} will be passed to the \code{vcov_tee()}
##'   call. See documentation for \code{vcov_tee()} for information about
##'   necessary arguments.
##' @param object a fitted \code{teeMod} model
##' @param ... additional arguments to \code{vcov_tee()}.
##' @inherit vcov_tee return
##' @exportS3Method
vcov.teeMod <- function(object, ...) {
  cl <- match.call()
  cl[[1L]] <-  quote(vcov_tee)
  names(cl)[2L] <- "x"

  vmat <- eval(cl, parent.frame())
  return(vmat)
}

##' @title Confidence intervals with standard errors provided by
##'   \code{vcov.teeMod()}
##' @description An S3method for \code{stats::confint} that uses standard errors
##'   computed using \code{vcov.teeMod()}. Additional arguments passed to this
##'   function, such as \code{cluster} and \code{type}, specify the arguments of
##'   the \code{vcov.teeMod()} call.
##' @details Rather than call \code{stats::confint.lm()},
##'   \code{confint.teeMod()} calls \code{.confint_lm()}, a function internal to
##'   the \code{propertee} package that ensures additional arguments in the
##'   \code{...} of the \code{confint.teeMod()} call are passed to the internal
##'   \code{vcov()} call.
##' @inheritParams .confint_lm
##' @inherit .confint_lm return
##' @exportS3Method
confint.teeMod <- function(object, parm, level = 0.95, ...) {
  cl <- match.call()
  cl[[1L]] <- quote(.confint_lm)
  return(eval(cl, parent.frame()))
}

##' @title Extract empirical estimating equations from a \code{teeMod} model fit
##' @description An S3method for \code{sandwich::estfun} for producing a matrix
##'   of contributions to the direct adjustment estimating equations.
##' @details If a prior covariance adjustment model has been passed to the
##'   \code{offset} argument of the \code{teeMod} model using \code{cov_adj()},
##'   \code{estfun.teeMod()} incorporates contributions to the estimating
##'   equations of the covariance adjustment model.\cr\cr The covariance
##'   adjustment sample may not fully overlap with the direct adjustment sample,
##'   in which case \code{estfun.teeMod()} returns a matrix with the same number
##'   of rows as the number of unique units of observation used to fit the two
##'   models. Uniqueness is determined by matching units of assignment used to
##'   fit the covariance adjustment model to units of assignment in the
##'   \code{teeMod} model's \code{StudySpecification} slot; units of observation
##'   within units of assignment that do not match are additional units that add
##'   to the row count.\cr\cr The working residuals (and, if applicable, working
##'   weights) from \code{values_from} will be used in the output matrix in
##'   place of those from \code{x}. In other words, the contributions to the
##'   the empirical estimating equations remain the same but are evaluated at
##'   parameter estimates from \code{values_from} rather than \code{x}.\cr\cr The \code{by} argument in \code{cov_adj()} can
##'   provide a column or a pair of columns (a named vector where the name
##'   specifies a column in the direct adjustment sample and the value a column
##'   in the covariance adjustment sample) that uniquely specifies units of
##'   observation in each sample. This information can be used to align each
##'   unit of observation's contributions to the two sets of estimating
##'   equations. If no \code{by} argument is provided and units of observation
##'   cannot be uniquely specified, contributions are aligned up to the unit of
##'   assignment level. If standard errors are clustered no finer than that,
##'   they will provide the same result as if each unit of observation's
##'   contributions were aligned exactly.\cr\cr This method incorporates bias
##'   corrections made to the residuals of \code{x} and, if applicable, the
##'   covariance model stored in its \code{offset}. When its crossproduct is taken
##'   (perhaps after suitable summing across rows within clusters), it provides
##'   a heteroskedasticity- (or cluster-) robust estimate of the meat matrix of
##'   the variance-covariance of the parameter estimates in \code{x}.
##'
##' @param x a fitted \code{teeMod} model
##' @param values_from optional, a fitted model object. Defaults to \code{x}.
##' @param ... arguments passed to methods, most importantly those that define
##'   the bias corrections for the residuals of \code{x} and, if applicable, a
##'   \code{fitted_covariance_model} stored in its offset
##' @return An \eqn{n\times k} matrix of empirical estimating equations for \code{x}.
##'   \code{k} includes the model intercept, main effects of treatment and
##'   moderator variables, any moderator effects, and marginal and conditional
##'   means of the outcome (and \code{offset}, if provided) in the control condition.
##'   See Details for definition of \eqn{n}.
##' @exportS3Method
estfun.teeMod <- function(x, values_from = x, ...) {
  # change model object's na.action to na.exclude so estfun returns NA rows
  if (!is.null(x$na.action)) class(x$na.action) <- "exclude"
  dots <- list(...)
  # this sets the default to model-based
  vcov_type <- match.arg(dots$vcov_type, c("MB", "HC", "CR", "DB"))
  # sandwich::sandwich calls `NROW(estfun(x))` to get the scaling factor, so to
  # allow users to use sandwich::sandwich, we need to accommodate that kind of
  # call to `estfun` where itt_rcorrect isn't provided. we'll set a default
  # option here (the correction doesn't change the dimension of the output) but
  # make sure the user-facing variance estimation function `vcov_tee()` passes
  # on an `itt_rcorrect` argument that correctly reflects the args provided there
  if (is.null(itt_rcorrect <- dots$itt_rcorrect)) itt_rcorrect <- "HC0"
  if (!is.null(values_from$na.action)) {
    class(values_from$na.action) <- "exclude"
  }

  # MB and DB estfun without covariance adjustment; below, this will be replaced
  # entirely if there's covariance adjustment
  mat <- .base_S3class_estfun(x) - .estfun_DB_blockabsorb(x, ...)
  
  if (vcov_type != "DB") {
    # get estfun for ctrl mean regression 
    cm_ef <- estfun(x@ctrl_means_model)
    q <- ncol(cm_ef)
    cm_ef[is.na(cm_ef)] <- 0
    mat <- cbind(mat, cm_ef)
  } else {
    q <- 0
    cm_ef <- NULL
  }

  ## if ITT model offset doesn't contain info about covariance model, estimating
  ## equations should be the ITT model estimating equations
  if (is.null(sl <- x$model$`(offset)`) | !inherits(sl, "SandwichLayer")) {
    # .base_s3_class_estfun() uses working weights and residuals; we replace
    # those with working weights and residuals from the `values_from` arg
    resids_to_replace <- stats::residuals(x, type = "working")
    replacement_resids <- stats::residuals(values_from, type = "working")
    if (length(resids_to_replace) != length(replacement_resids)) {
      stop("Lengths of residuals(x) and residuals(values_from) do not match")
    }
    if (is.null(wts_to_replace <- stats::weights(x, type = "working"))) {
      wts_to_replace <- rep(1, nrow(mat))
    }
    if (is.null(replacement_wts <- stats::weights(values_from,
                                                  type = "working"))) {
      replacement_wts <- rep(1, nrow(mat))
    }
    # even though `replacement_resids` may be from the model passed to
    # `values_from`, the `x` argument of `rcorrect` should still be `x`
    # and the `model` argument should still be `itt` because these inform the
    # CR1/CR2 corrections, which should be drawn from the original model but
    # for using the working weights from `values_from` in CR2 corrections
    replacement_resids <- do.call(.rcorrect,
                                  c(list(resids = replacement_resids, x = x,
                                         model = "itt", type = itt_rcorrect,
                                         cluster = dots$cls,
                                         values_from = values_from),
                                    dots[setdiff(names(dots), "cls")]))
    # after setting the na.action to "exclude" at the top of the
    # function, `residuals` and `weights` return NA's, so we make sure the
    # entries in the estfun are zeros rather than NA's
    return(
      mat * 
        (replace(replacement_resids, is.na(replacement_resids), 0) /
           replace(resids_to_replace, is.na(resids_to_replace), 1)) *
        (replace(replacement_wts, is.na(replacement_wts), 0) /
           replace(wts_to_replace,
                   is.na(wts_to_replace) |
                     (!is.na(wts_to_replace) & wts_to_replace == 0),
                   1))
    )
             
  }

  ## otherwise, extract/compute the rest of the relevant matrices/quantities
  estmats <- .align_and_extend_estfuns(x, values_from, cm_ef, ...)
  a11_inv <- .get_a11_inverse(x)
  a21 <- .get_a21(x, ...)

  ## get scaling constants
  # include NA's to match other routines
  nq <- nrow(stats::model.frame(x, na.action = "na.pass"))
  nc <- nrow(stats::model.frame(sl@fitted_covariance_model))
  n <- nrow(estmats[["psi"]])

  ## form matrix of estimating equations
  mat <- estmats[["psi"]] - nq / nc * estmats[["phi"]] %*% t(a11_inv) %*% t(a21)
  mat[,1:(ncol(estmats[["psi"]]) - q)] <- mat[,1:(ncol(estmats[["psi"]])-q)] -
    .estfun_DB_blockabsorb(x, ...)
  return(mat)
}

##' @title Extract bread matrix from a \code{teeMod} model fit
##' @description
##'   An S3method for \code{sandwich::bread} that extracts the bread of the
##'   direct adjustment model sandwich covariance matrix.
##'
##' @details This function is a thin wrapper around
##'   \code{.get_tilde_a22_inverse()}.
##' @param x a fitted \code{teeMod} model
##' @param values_from optional, a fitted model object. Defaults to \code{x}.
##' @param ... arguments passed to methods
##' @inherit vcov_tee return
##' @exportS3Method
bread.teeMod <- function(x, values_from = x, ...)
  .get_tilde_a22_inverse(x, values_from, ...)

##' @title (Internal) Bias correct residuals contributing to standard errors of
##'   a \code{teeMod}
##' @param resids numeric vector of residuals to correct
##' @param x teeMod object
##' @param model string indicating which model the residuals are from.
##' \code{"itt"} indicates correction to the residuals of \code{x}, and
##' \code{"cov_adj"} indicates correction to the residuals of the covariance
##' adjustment model. This informs whether corrections should use information
##' from \code{x} or the \code{fitted_covariance_model} slot of the
##' \code{SandwichLayer} object in the \code{offset} for corrections.
##' @param type string indicating the desired bias correction. Can be one of
##' \code{"(HC/CR/MB)0"}, \code{"(HC/CR/MB)1"}, or \code{"(HC/CR/MB)2"}
##' @param ... additional arguments passed from up the call stack; in
##' particular, the \code{cluster_cols} argument, which informs whether to
##' cluster and provide CR2 corrections instead of HC2 corrections, as well as
##' the correction for the number of clusters in the CR1 correction. This may
##' also include a \code{by} argument.
##' @keywords internal
.rcorrect <- function(resids, x, model, type, ...) {
  if (type %in% paste0(c("HC", "CR", "MB", "DB"), "0")) {
    cr <- resids
  } else if (type %in% c(paste0(rep(c("MB", "HC", "CR"), 2),
                                rep(c(1, 2), each = 3)))) {
    dots <- list(...)
    if (is.null(cluster_cols <- dots$cluster_cols)) {
      cluster_cols <- var_names(x@StudySpecification, "u")
    }
    sl <- x$model$`(offset)`
    cmod <- if (!is.null(sl) & inherits(sl, "SandwichLayer")) {
      sl@fitted_covariance_model
    } else NULL 
    
    if (substr(type, 3, 3) == 1) {
      # HC/CR1 correction is based on n and k from propertee's estfun matrix
      # for both sets of residuals
      cls <- if ("cluster" %in% names(dots)) dots$cluster else {
        if (is.null(by <- dots$by) & !is.null(cmod)) {
          by <- setdiff(colnames(sl@keys),
                        c(var_names(x@StudySpecification, "u"), "in_Q"))
          if (length(by) == 0) by <- NULL
        }
        if (is.null(by)) by <- cluster_cols
        do.call(".make_uoa_ids",
                list(x = x, vcov_type = substr(type, 1, 2),
                     cluster = cluster_cols, by = by))
      }
      g <- length(unique(cls))
      n <- length(cls)
      k <- x$rank + if (!is.null(cmod) & inherits(cmod, "lmrob")) {
        sum(!is.na(cmod$coefficients)) + 1
      } else if (!is.null(cmod)) sum(!is.na(cmod$coefficients)) else 0
      cr <- sqrt(g / (g-1) * (n-1) / (n-k)) * resids
    } else {
      # (HC/CR)2 correction; derived separately for the two regressions
      model <- match.arg(model, c("itt", "cov_adj"))
      if (model == "itt") {
        mod <- x 
        efm <- .base_S3class_estfun(mod)
        mf_data <- get("data", environment(formula(mod)))
      } else {
        if (is.null(cmod)) {
          stop("x must have a SandwichLayer object as an offset")
        }
        mod <- sl@fitted_covariance_model
        efm <- estfun(mod)
        mf_data <- eval(mod$call$data, environment(formula(mod)))
      }
      if ("cluster" %in% names(dots)) {
        cls <- dots$cluster
      } else if (cluster_cols[1] == "..uoa..") {
        cls <- row.names(stats::model.frame(mod, na.action = na.pass))
      } else {
        cls <- Reduce(
          function(l, r) paste(l, r, sep = "_"),
          as.list(stats::expand.model.frame(mod, cluster_cols)[, cluster_cols
                                                               , drop=FALSE])
        )
      }
      if (any(nas <- is.na(cls))) cls[nas] <- replicate(
        sum(nas),
        paste(sample(letters, 8, replace = TRUE), collapse = ""),
        simplify = TRUE
      )

      g <- length(unique(cls))
      n <- nrow(efm)
      k <- ncol(efm)
      
      if (g == n) {
        pii <- stats::hatvalues(mod)
        cr <- 1 / sqrt(1 - pii) * resids
      } else {
        if (inherits(mod, c("glmrob", "lmrob"))) stop(
          "CR2 correction not implemented for robust fits")
        wres <- stats::residuals(mod, type = "working")
        XW <- sweep(efm, 1, wres, FUN = "/")
        # if there's a values_from argument in estfun.teeMod, use its working
        # weights in corrections
        if (!is.null(dots$values_from)) {
          if (is.null(wts_to_replace <- stats::weights(mod,
                                                       type = "working"))) {
            wts_to_replace <- rep(1, nrow(XW))
          }
          if (is.null(replacement_wts <- stats::weights(dots$values_from,
                                                        type = "working"))) {
            replacement_wts <- rep(1, nrow(XW))
          }
          XW <- sweep(
            sweep(
              XW,
              1,
              replace(wts_to_replace,
                      is.na(wts_to_replace) |
                        (!is.na(wts_to_replace) & wts_to_replace == 0),
                      1),
              FUN = "/"
            ),
            1,
            replace(replacement_wts,
                    is.na(replacement_wts) |
                      (!is.na(replacement_wts) & replacement_wts == 0),
                    1),
            FUN = "*"
          )
        }
        XW[is.na(XW)] <- 0
        X <- stats::model.matrix(
          stats::delete.response(stats::terms(mod)),
          do.call("model.frame",
                  list(mod, mf_data, subset = eval(mod$call$subset, mf_data),
                       na.action = na.pass)),
          contrasts.arg = mod$contrasts,
          xlev = mod$xlevels
        )[,colnames(XW),drop=FALSE] # estfun drops NA coeffs for XW, so we have
                                    # to align with that
        XTWX_inv <- solve(crossprod(XW, X))
        cr <- numeric(length(cls))
        for (cl in unique(cls)) {
          cl_ix <- which(cls == cl)
          crc <- rep(NA_real_, length(cl_ix))
          nas <- stats::na.action(mod)
          ok <- setdiff(cl_ix, nas)
          if (length(ok) > 0) {
            if (inherits(mod, "teeMod")) {
              iss <- cluster_iss(mod, cluster_unit = cl, cluster_ids = cls, ...)
            } else {
              I_P_cc <- diag(length(ok)) - tcrossprod(
                tcrossprod(X[ok,,drop=FALSE], XTWX_inv), XW[ok,,drop=FALSE])
              schur <- eigen(I_P_cc)
              iss <- schur$vectors %*%
                (solve(schur$vectors) / sqrt(schur$values))
            }
            crc[!(cl_ix %in% nas)] <- drop(iss %*% resids[ok])
          }
          cr[cl_ix] <- crc
        }
      }
    }
  } else {
    stop(paste0("'", type, "' bias correction not available"))
  }
  
  return(cr)
}
##' @title (Internal) Align the dimensions and rows of direct adjustment and
##'   covariance adjustment model estimating equations matrices
##' @details \code{.align_and_extend_estfuns()} first extracts the matrices of
##'   contributions to the empirical estimating equations for the direct
##'   adjustment and covariance adjustment models; then, it pads the matrices
##'   with zeros to account for units of observation that appear in one
##'   model-fitting sample but not the other; finally it orders the matrices so
##'   units of observation (or if unit of observation-level ordering is
##'   impossible, units of assignment) are aligned.\cr\cr As in
##'   \code{estfun.teeMod()}, the working residuals (and, if applicable,
##'   weights) from \code{values_from} will be used in the output matrix in
##'   place of those from \code{x}.
##' @param x a fitted \code{teeMod} model
##' @param values_from optional, a fitted model object. Defaults to \code{x}.
##' @param ctrl_means_ef_mat optional, a matrix of estimating equations corresponding
##'   to the estimates of the marginal (and possibly conditional) means of the outcome
##'   and \code{offset} in the control condition. These are aligned and extended
##'   in the same way as the matrix of estimating equations for \code{x} and
##'   \code{cbind}ed to them
##' @param by optional, a character vector indicating columns that uniquely identify
##'   rows in the dataframe used for fitting \code{x} and the dataframe passed to the
##'   \code{data} argument of the covariance adjustment model fit. The default is
##'   \code{NULL}, in which case the unit of assignment columns specified in the
##'   \code{StudySpecification} slot of \code{x} are used.
##' @param ... mostly arguments passed to methods, but the special case is the argument
##'   \code{loco_residuals}, which indicates the offsets in the residuals of \code{x}
##'   should be replaced by versions that use leave-one-cluster-out estimates of
##'   the covariance model
##' @return A list of two matrices, one being the aligned contributions to the
##'   estimating equations for the direct adjustment model, and the other being
##'   the aligned contributions to the covariance adjustment model.
##' @keywords internal
.align_and_extend_estfuns <- function(x,
                                      values_from = x,
                                      ctrl_means_ef_mat = NULL,
                                      by = NULL,
                                      ...) {
  if (!inherits(x, "teeMod") | !inherits(x$model$`(offset)`, "SandwichLayer")) {
    stop("`x` must be a fitted teeMod object with a SandwichLayer offset")
  }
  dots <- list(...)
  dots$type <- NULL
  ## same in-line comment in `estfun.teeMod()` about setting itt_rcorrect applies
  ## to both args here
  if (is.null(itt_rcorrect <- dots$itt_rcorrect)) itt_rcorrect <- "HC0"
  if (is.null(cov_adj_rcorrect <- dots$cov_adj_rcorrect)) {
    cov_adj_rcorrect <- "HC0"
  }
  if (!is.null(values_from$na.action)) {
    class(values_from$na.action) <- "exclude"
  }
  
  uoa_cols <- var_names(x@StudySpecification, "u")
  cluster_cols <- if (is.null(dots$cluster_cols)) {
    uoa_cols
  } else dots$cluster_cols
  sl <- x$model$`(offset)`
  cmod <- sl@fitted_covariance_model
  # changing the na.action to exclude returns NA's where rows had NA's; see further details below
  if (!is.null(cmod$na.action)) class(cmod$na.action) <- "exclude"
  
  # the ordering output by `.order_samples()` is explained in that function's
  # documentation
  id_order <- .order_samples(x, by = by, verbose = FALSE)
  n <- length(c(id_order$Q_not_C, id_order$C_in_Q, id_order$C_not_Q))
  nq <- length(c(id_order$Q_not_C, id_order$Q_in_C))
  nc <- length(c(id_order$C_in_Q, id_order$C_not_Q))
  
  ## get the unaligned + unextended estimating equations
  # since we change the na.action to na.exclude for the covariance adjustment
  # model above, estfun() returns rows filled with NA's for rows with any NA's,
  # and similarly, residuals() returns NA's. if there are no rows with NA's,
  # estfun() and residuals() return their usual outputs
  phi_r <- stats::residuals(cmod, type = "working") # for teeMod, lm, lmrob this
                                                    # gives desired type =
                                                    # "response"
  phi <- sandwich::estfun(cmod)
  phi <- replace(phi, is.na(phi), 0) / replace(phi_r, is.na(phi_r), 1) *
    replace(do.call(.rcorrect,
                    c(list(resids = phi_r, x = x, model = "cov_adj",
                           type = cov_adj_rcorrect, by = by,
                           values_from = values_from),
                      dots)),
            is.na(phi_r),
            0)

  # use jackknife first-stage coefficient estimates if Q and C overlap
  psi_r_to_replace <- stats::residuals(x, type = "working")
  if (!is.null(dots$loco_residuals) & sum(sl@keys$in_Q) > 0) {
    # .compute_loo_resids uses residuals from `values_from` arg
    replacement_psi_r <- .compute_loo_resids(x, cluster_cols, ...)
  } else {
    replacement_psi_r <- stats::residuals(values_from, type = "working")
  }
  if (length(replacement_psi_r) != length(psi_r_to_replace)) {
    stop("Lengths of residuals(x) and residuals(values_from) do not match")
  }
  if (is.null(wts_to_replace <- stats::weights(x, type = "working"))) {
    wts_to_replace <- rep(1, length(psi_r_to_replace))
  }
  if (is.null(replacement_wts <- stats::weights(values_from,
                                                type = "working"))) {
    replacement_wts <- rep(1, length(replacement_psi_r))
  }
  # see in-line comments of estfun.teeMod() for verification of the .rcorrect
  # args
  replacement_psi_r <- do.call(.rcorrect,
                               c(list(resids = replacement_psi_r, x = x,
                                      model = "itt", type = itt_rcorrect,
                                      values_from = values_from, by = by),
                                 dots))
  psi <- .base_S3class_estfun(x) *
    (replace(replacement_psi_r, is.na(replacement_psi_r), 0) /
       replace(psi_r_to_replace, is.na(psi_r_to_replace), 1)) *
    (replace(replacement_wts, is.na(replacement_wts), 0) /
       replace(wts_to_replace,
               is.na(wts_to_replace) |
                 (!is.na(wts_to_replace) & wts_to_replace == 0),
               1))

  Q_order <- c(as.numeric(names(id_order$Q_not_C)),
               as.numeric(names(id_order$Q_in_C)))
  aligned_psi <- matrix(0, nrow = n, ncol = ncol(psi),
                        dimnames = list(seq(n), colnames(psi)))
  aligned_psi[1:nq,] <- psi[Q_order,,drop=FALSE]

  C_order <- c(as.numeric(names(id_order$C_in_Q)),
               as.numeric(names(id_order$C_not_Q)))
  aligned_phi <- matrix(0, nrow = n, ncol = ncol(phi),
                        dimnames = list(seq_len(n), colnames(phi)))
  aligned_phi[(n-nc+1):n,] <- phi[C_order,,drop=FALSE]
  
  out <- list(psi = aligned_psi, phi = aligned_phi)

  if (!is.null(ctrl_means_ef_mat)) {
    aligned_cm_ef <- matrix(0, nrow = n, ncol = ncol(ctrl_means_ef_mat),
                            dimnames = list(seq(n),
                                            colnames(ctrl_means_ef_mat)))
    aligned_cm_ef[1:nq,] <- ctrl_means_ef_mat[Q_order,,drop=FALSE]
    out[["psi"]] <- cbind(out[["psi"]], aligned_cm_ef)
  }

  return(out)
}

##' @title (Internal) Extract empirical estimating equations from a
##' \code{teeMod} model using the S3 method associated with its
##' \code{.S3Class} slot
##' @inheritParams estfun.teeMod
##' @return S3 method
##' @keywords internal
.base_S3class_estfun <- function(x) {
  ## this vector indicates the hierarchy of `sandwich::estfun` methods to use
  ## to extract ITT model's estimating equations
  x$coefficients <- replace(x$coefficients, is.na(x$coefficients), 0)
  valid_classes <- c("glm", "lmrob", "svyglm", "lm")
  base_class <- match(x@.S3Class, valid_classes)
  if (all(is.na(base_class))) {
    stop(paste("Direct adjustment model must have been fitted using a function from the",
               "`propertee`, `stats`, `robustbase`, or `survey` package"))
  }
  
  ef <- getS3method("estfun", valid_classes[min(base_class, na.rm = TRUE)])(x)
  ef[is.na(ef)] <- 0
  keep_coef <- if (is.null(aliased <- stats::alias(x)$Complete)) {
    colnames(ef)
  } else {
    if (is.null(nms <- colnames(aliased))) 1 else nms
  }
  return(ef[, keep_coef,drop=FALSE])
}

#' @title Compute residuals for a \code{teeMod} object with leave-one-out estimates
#' of the \code{offset}
#' @details
#' The residual for any observation also used for fitting the \code{fitted_covariance_model}
#'   stored in the \code{offset} of \code{x} is replaced by an estimated residual
#'   that uses a cluster leave-one-out estimate of the \code{fitted_covariance_model}
#'   for generating a value of the \code{offset}.
#' @param x a \code{teeMod} object
#' @param cluster vector of column names that identify clusters
#' 
#' @importFrom stats model.frame
#' @keywords internal
.compute_loo_resids <- function(x, cluster, ...) {
  ## what's in the call
  sl <- x$model$`(offset)`
  cmod <- sl@fitted_covariance_model
  
  ## get cluster ID's
  in_Q <- which(sl@keys$in_Q)
  if (cluster[1] == "..uoa..") {
    C_cls <- rownames(stats::model.frame(cmod, na.action = na.pass))
  } else {
    C_cls <- Reduce(
      function(l, r) paste(l, r, sep = "_"),
      as.list(stats::expand.model.frame(cmod, cluster)[, cluster, drop=FALSE])
    )
  }
  jk_units <- unique(C_cls[in_Q])

  ## get LOO coefficients for each overlapping unit in Q and C
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        cf <- eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
        matrix(cf, ncol = 1, dimnames = list(names(cf), NULL))
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  loo_cmod[is.na(loo_cmod)] <- 0 # replace NA coeffs due to singularity with 0's
  
  ## make preds for units in Q using LOO estimates of cov adj model
  ## (make sure to set treatment indicator to 0 if it was included in the model)
  tt <- stats::delete.response(stats::terms(cmod))
  Q_data <- get("data", envir = environment(formula(x)))
  mf <- call("model.frame",
             tt,
             Q_data,
             subset = eval(x@lmitt_call$subset, envir = Q_data),
             na.action = na.pass,
             xlev = cmod$xlevels)
  mf <- eval(mf)
  trt_name <- var_names(x@StudySpecification, "t")
  if (trt_name %in% colnames(mf)) {
    trts <- treatment(x@StudySpecification)[, 1]
    if (is.numeric(trts)) {
      mf[[trt_name]] <- min(abs(trts))
    } else if (is.logical(trts)) {
      mf[[trt_name]] <- FALSE
    } else if (is.factor(trts)) {
      mf[[trt_name]] <- levels(trts)[1]
    }
  }
  mm <- stats::model.matrix(tt, mf, contrasts.arg = cmod$contrasts)

  if (cluster[1] == "..uoa..") {
    Q_cls <- rownames(stats::model.frame(x, na.action = na.pass))
  } else {
    Q_cls <- Reduce(
      function(l, r) paste(l, r, sep = "_"),
      as.list(stats::expand.model.frame(x, cluster)[, cluster, drop=FALSE])
    )
  }
  Q_cls_in_C <- Q_cls %in% colnames(loo_cmod)
  loo_preds <- rowSums(mm[Q_cls_in_C,,drop=FALSE] *
                         t(loo_cmod[, as.character(Q_cls[Q_cls_in_C])]))
  if (!is.null(cmod$family)) loo_preds <- cmod$family$linkinv(loo_preds)

  ## slot them in to existing predictions
  all_preds <- sl@.Data
  all_preds[Q_cls_in_C] <- loo_preds
  
  ## make residuals (subtract original offset bc it's added in fitted values)
  if (is.null(values_from <- list(...)$values_from)) values_from <- x
  if (!is.null(values_from$na.action)) {
    class(values_from$na.action) <- "exclude"
  }
  resids <- stats::residuals(values_from, type = "working")
  os <- stats::model.offset(stats::model.frame(x, na.action = na.pass))
  
  return(resids + os - all_preds)
}

#' @title Make ID's to pass to the \code{cluster} argument of \code{vcov_tee()}
#' @description \code{.make_uoa_ids()} returns a factor vector of cluster ID's
#'   that align with the order of the units of observations' contributions in
#'   \code{estfun.teeMod()}. This is to ensure that when \code{vcov_tee()} calls
#'   \code{sandwich::meatCL()}, the \code{cluster} argument aggregates the
#'   correct contributions to estimating equations within clusters.
#' @param x a fitted \code{teeMod} object
#' @param vcov_type a string indicating model-based or design-based
#'   covariance estimation. Currently, "MB", "CR", and "HC" are the only strings
#'   registered as indicating model-based estimation.
#' @param cluster character vector or list; optional. Specifies column names
#'   that appear in both the covariance adjustment and direct adjustment model
#'   dataframes. Defaults to NULL, in which case unit of assignment columns
#'   indicated in the StudySpecification will be used for clustering. If there
#'   are multiple clustering columns, they are concatenated together for each
#'   row and separated by "_".
#' @param ... arguments passed to methods
#' @return A vector with length equal to the number of unique units of
#'   observation used to fit the two models. See Details of
#'   \code{estfun.teeMod()} for the method for determining uniqueness.
#' @keywords internal
.make_uoa_ids <- function(x, vcov_type, cluster = NULL, ...) {
  if (!inherits(x, c("teeMod", "mmm"))) {
    stop("Must be provided a teeMod or mmm object")
  } else if (inherits(x, "mmm")) {
    # just use the first model to generate the ID's; this will be the same for
    # mmm objects of models with the same Q (and possibly C) samples
    x <- x[[1L]]
  }
  mc <- match.call()

  # Must be a teeMod object for this logic to occur
  if (!inherits(cluster, "character")) {
    cluster <- var_names(x@StudySpecification, "u")
  }

  # get observation-level unit of assignment and cluster ID's for observations
  # in Q
  Q_obs <- .sanitize_Q_ids(x, id_col = cluster, ...)
  Q_obs_ids <- Q_obs$cluster

  # If there's no covariance adjustment info, return the ID's found in Q
  if (!inherits(ca <- x$model$`(offset)`, "SandwichLayer")) {
    return(factor(Q_obs_ids, levels = unique(Q_obs_ids)))
  }

  # `keys` may have columns in addition to "in_Q" and the uoa columns if a
  # `by` argument was specified in `cov_adj()` or `as.SandwichLayer()`. If it
  # does, use the columns exclusively specified in `by` to produce the order
  if (is.null(by <- mc$by)) by <- setdiff(colnames(ca@keys),
                                            c(var_names(x@StudySpecification, "u"), "in_Q"))
  if (length(by) == 0) {
    by <- cluster
  }

  ord_call <- call(".order_samples", quote(x), by = quote(by))
  id_order <- eval(ord_call)

  # if no "by" was specified in cov_adj(), cluster variable was used for ordering,
  # so we can take the names of the sorted vector. Otherwise, we need to get
  # ID's associated with the ordering.
  if (length(setdiff(cluster, by)) == 0) {
    ids <- Reduce(c, id_order[c("Q_not_C", "Q_in_C", "C_not_Q")])
  } else {
    C_ids <- .sanitize_C_ids(ca, cluster, verbose = FALSE, sorted = FALSE, ...)
    ids <- c(
      Q_obs_ids[as.numeric(names(id_order$Q_not_C))],
      Q_obs_ids[as.numeric(names(id_order$Q_in_C))],
      C_ids[as.numeric(names(id_order$C_not_Q))]
    )
  }

  na_ids <- is.na(ids)
  ids[na_ids] <- apply(
    matrix(sample(c(letters, LETTERS), 8 * sum(na_ids), replace = TRUE), ncol = 8),
    1, function(...) paste(..., collapse = "")
  )

  return(factor(ids, levels = unique(ids)))
}

#' @title (Internal) Order observations used to fit a \code{teeMod} model and a
#'   prior covariance adjustment model
#' @details \code{.order_samples()} underpins the ordering for
#'   \code{.make_uoa_ids()} and \code{estfun.teeMod()}. This function orders the
#'   outputs of those functions, but also informs how the original matrices of
#'   contributions to estimating equations need to be indexed to align units of
#'   observations' contributions to both sets of estimating equations.\cr\cr
#'   When a \code{by} argument is provided to \code{cov_adj()}, it is used to
#'   construct the order of \code{.order_samples()}.
#' @param x a fitted \code{teeMod} model
#' @param by character vector of columns to get ID's for ordering from. Default
#'   is NULL, in which case unit of assignment ID's are used for ordering.
#' @param ... arguments passed to methods
#' @return A list of four named vectors. The \code{Q_not_C} element holds the
#'   ordering for units of observation in the direct adjustment sample but not
#'   the covariance adjustment samples; \code{Q_in_C} and \code{C_in_Q}, the
#'   ordering for units in both; and \code{C_not_Q}, the ordering for units in
#'   the covariance adjustment sample only. \code{Q_in_C} and \code{C_in_Q}
#'   differ in that the names of the \code{Q_in_C} vector correspond to row
#'   indices of the original matrix of estimating equations for the direct
#'   adjustment model, while the names of \code{C_in_Q} correspond to row
#'   indices of the matrix of estimating equations for the covariance adjustment
#'   model. Similarly, the names of \code{Q_not_C} and \code{C_not_Q} correspond
#'   to row indices of the direct adjustment and covariance adjustment samples,
#'   respectively. Ultimately, the order of \code{.make_uoa_ids()} and
#'   \code{estfun.teeMod()} is given by concatenating the vectors stored in
#'   \code{Q_not_C}, \code{Q_in_C}, and \code{C_not_q}.
#' @keywords internal
.order_samples <- function(x, by = NULL, ...) {
  if (!inherits(x, "teeMod") | !inherits(ca <- x$model$`(offset)`, "SandwichLayer")) {
    stop(paste("x must be a teeMod object with a SandwichLayer offset or",
               "ca must be a SandwichLayer object to retrieve information about",
               "the covariance adjustment model"))
  }
  ## `keys` may have additional columns beyond "in_Q" and the uoa columns if a
  ## `by` argument was specified in `cov_adj()` or `as.SandwichLayer()`. If it
  ## does, use the columns exclusively specified in `by` to merge.
  if (is.null(by) | length(setdiff(colnames(ca@keys),
                                   c(var_names(x@StudySpecification, "u"), "in_Q"))) > 0) {
    by <- setdiff(colnames(ca@keys), c(var_names(x@StudySpecification, "u"), "in_Q"))
  }

  ## Should only hit this if a custom `cluster` argument hasn't been passed to
  ## vcov_tee or no `by` was specified in `cov_adj`
  if (length(by) == 0) {
    by <- var_names(x@StudySpecification, "u")
  }

  # The order, given by the names of the output vector, will be:
  # Q not in C --> Q in C --> C not in Q. The values in the vector correspond to
  # the rows to pull from the original estfun matrices

  # get all ID's in Q
  # Q_ids <- .sanitize_Q_ids(x, id_col = by, ...)[, "cluster"]
  if (x@StudySpecification@unit_of_assignment_type == "none") {
    Q_ids <- rownames(model.frame(x, na.action = NULL))
  } else {
    Q_ids <- stats::expand.model.frame(x, by)[, by, drop = FALSE]
    Q_ids <- apply(Q_ids, 1, function(...) paste(..., collapse = "_"))
  }

  # get all ID's in C and replace NA's with unique ID
  C_ids <- .sanitize_C_ids(ca, by, sorted = FALSE, ...)

  # need Q_in_C and C_in_Q to have the same order so contributions are aligned
  Q_in_C <- stats::setNames(Q_ids[which(Q_ids %in% C_ids)], which(Q_ids %in% C_ids))
  Q_in_C <- sort(Q_in_C)
  C_in_Q <- stats::setNames(C_ids[which(C_ids %in% Q_ids)], which(C_ids %in% Q_ids))
  C_in_Q <- sort(C_in_Q)

  if (length(Q_in_C) != length(C_in_Q)) {
    stop(paste("Contributions to covariance adjustment and/or effect estimation",
               "are not uniquely specified. Provide a `by` argument to `cov_adj()`",
               "or `vcov.teeMod()`"))
  }

  out <- list(
    Q_not_C = stats::setNames(Q_ids[!(Q_ids %in% C_ids)], which(!(Q_ids %in% C_ids))),
    Q_in_C = Q_in_C,
    C_in_Q = C_in_Q,
    C_not_Q = stats::setNames(C_ids[!(C_ids %in% Q_ids)], which(!(C_ids %in% Q_ids)))
  )

  return(out)
}

#' @title (Internal) Return ID's used to order observations in the direct
#'   adjustment sample
#' @param x a fitted \code{teeMod} model
#' @param id_col character vector; optional. Specifies column(s) whose ID's will
#'   be returned. The column must exist in the data that created the
#'   \code{StudySpecification} object. Default is NULL, in which case unit of
#'   assignment columns indicated in the specification will be used to generate
#'   ID's.
#' @param ... arguments passed to methods
#' @return A vector with length equal to the number of units of observation in
#'   the direct adjustment sample
#' @keywords internal
.sanitize_Q_ids <- function(x, id_col = NULL, ...) {
  # link the units of assignment in the StudySpecification with desired cluster ID's
  uoa_cls_df <- .make_uoa_cluster_df(x@StudySpecification, id_col)
  uoa_cols <- var_names(x@StudySpecification, "u")
  if (nrow(uoa_cls_df) == nrow(x$model)) {
    expand_cols <- unique(c(uoa_cols, id_col))
    by.y <- if (length(expand_cols) == length(uoa_cols)) uoa_cols else c(uoa_cols, "cluster")
  } else {
    expand_cols <- by.y <- uoa_cols
  }

  if (x@StudySpecification@unit_of_assignment_type == "none") {
    moddata <- x$call$data
    moddata$..uoa.. <- rownames(moddata)
    x$call$data <- moddata
  }
  obs_uoa_ids <- stats::expand.model.frame(x,
                                           expand_cols)[, expand_cols, drop = FALSE]

  out <- merge(cbind(obs_uoa_ids, .order_id = seq_len(nrow(obs_uoa_ids))),
               uoa_cls_df, by.x = expand_cols, by.y = by.y, all.x = TRUE, sort = FALSE)

  out <- out[sort(out$.order_id, index.return = TRUE)$ix,]
  out$.order_id <- NULL
  colnames(out) <- unique(c(by.y, "cluster"))
  out$cluster <- as.character(out$cluster)
  rownames(out) <- NULL

  return(out)
}

#' @exportS3Method getCall teeMod
#' @importFrom stats getCall
getCall.teeMod <- function(x, ...) {
  return(x$call)
}

#' @export
update.teeMod <- function(object, ...) {
  stop(paste("teeMod objects do not support an `update` method.\n",
             "You can use `update` on the formula object passed to `lmitt`",
             "instead."))
}

#' @title Design-based estimating equations contributions
#' @param x a fitted \code{teeMod} object
#' @param ... arguments passed to methods
#' @details calculate contributions to empirical estimating equations from a
#'   \code{teeMod} model with absorbed intercepts from the design-based
#'   perspective
#' @return An \eqn{n\times k} matrix
#' @keywords internal
.estfun_DB_blockabsorb <- function (x, by = NULL, ...){
  # return 0 if not asking for a design-based SE or DA does not absorb intercepts
  cl <- match.call()
  vcov_type <- match.arg(cl$vcov_type, c("MB", "HC", "CR", "DB"))
  if (!(db <- vcov_type == "DB") | !x@absorbed_intercepts) {
    return(0)
  }
  
  if (inherits(x$model$`(offset)`, "SandwichLayer")){
    # if the model involves SandwichLayer covariance adjustment
    temp <- .align_and_extend_estfuns(x, by = by, ...)[["psi"]]
  }
  else
    temp <- .base_S3class_estfun(x)

  phitilde <- .get_phi_tilde(x)
  appinv_atp <- .get_appinv_atp(x)

  if (!is.null(x$call$offset)){
    # if the model involves covariance adjustment
    # define the row ordering using .order_samples() and insert rows of 0's into
    # the matrices where necessary while maintaining observation alignment
    id_order <- .order_samples(x, by = by, verbose = FALSE)
    aligned_phitilde <- matrix(
      0, nrow = nrow(temp), ncol = ncol(phitilde),
      dimnames = list(seq_len(nrow(temp)), colnames(phitilde)))

    C_order <- c(as.numeric(names(id_order$C_in_Q)), as.numeric(names(id_order$C_not_Q)))
    aligned_phitilde[which(!is.na(C_order)),] <- phitilde[C_order[!is.na(C_order)], ]
    mat <- aligned_phitilde %*% appinv_atp
  }
  else{
    mat <- phitilde %*% appinv_atp
  }
  mat <- cbind(matrix(0, nrow = nrow(mat), ncol = ncol(temp) - ncol(mat)), mat)
  return(mat)
}

test_that(paste("teeMod object created correctly with weights and no",
                "SandwichLayer in the lm call"), {

  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa2, uoa1) + block(bid), data = simdata)

  sslm <- new("teeMod",
              lm(y ~ assigned(), data = simdata, weights = ate(spec)),
              StudySpecification = spec,
              ctrl_means_model = lm(y ~ 1, simdata),
              lmitt_fitted = TRUE)

  expect_s4_class(sslm, "teeMod")
  expect_true(inherits(sslm, "lm"))

  expect_identical(sslm$model$"(weights)"@StudySpecification, spec)
  expect_identical(sslm$model$"(weights)"@StudySpecification, sslm@StudySpecification)
})

test_that(paste("teeMod object created correctly with weights and ",
                "SandwichLayer in the lm call"), {
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa2, uoa1) + block(bid), data = simdata)
  cmod <- lm(y ~ x, data = simdata)
  sslm <- new("teeMod",
              lm(y ~ assigned(), data = simdata, weights = ate(spec),
                 offset = cov_adj(cmod)),
              StudySpecification = spec,
              ctrl_means_model = lm(y ~ 1, simdata),
              lmitt_fitted = FALSE)

  expect_s4_class(sslm, "teeMod")
  expect_true(inherits(sslm, "lm"))

  expect_equal(sslm$model$`(offset)`@.Data, as.numeric(cmod$fitted.values))
  expect_identical(sslm$model$`(offset)`@fitted_covariance_model, cmod)
  expect_equal(sslm$model$`(offset)`@prediction_gradient,
               stats::model.matrix(cmod))
  expect_identical(sslm$model$`(offset)`@keys,
                   cbind(simdata[,var_names(spec, "u")], in_Q = rep(TRUE, nrow(simdata))))
})

test_that("teeMod print/show", {

  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa2, uoa1) + block(bid), data = simdata)
  cmod <- lm(y ~ z, data = simdata)

  ######## Pretend it comes from `as.lmitt`
  sslm <- new("teeMod",
              lm(y ~ assigned(), data = simdata, weights = ate(spec),
                 offset = cov_adj(cmod)),
              StudySpecification = spec,
              ctrl_means_model = lm(y ~ 1, simdata),
              lmitt_fitted = FALSE)

  aslm <- as(sslm, "lm")

  expect_silent(invisible(capture.output(expect_identical(print(sslm), sslm))))
  expect_silent(invisible(capture.output(expect_identical(show(sslm), sslm))))

  # Expect "assigned()"
  expect_output(print(sslm), "assigned()")
  expect_output(show(sslm), "assigned()")

  ######## Now pretend it comes from `lmitt.formula`
  # Need `txt_` to match what comes out of a call like `lmitt(y ~ 1...)`
  simdata$z. <- simdata$z
  sslm <- new("teeMod",
              lm(y ~ z., data = simdata, weights = ate(spec),
                 offset = cov_adj(cmod)),
              StudySpecification = spec,
              ctrl_means_model = lm(y ~ 1, simdata),
              lmitt_fitted = TRUE)

  aslm <- as(sslm, "lm")

  expect_silent(invisible(capture.output(expect_identical(print(sslm), sslm))))
  expect_silent(invisible(capture.output(expect_identical(show(sslm), sslm))))

  # Expect "assigned()"
  expect_output(print(sslm), "z\\.")
  expect_output(show(sslm), "z\\.")
  
  # Expect new ctrl grp means
  suppressMessages(m1 <- lmitt(y ~ 1, spec, simdata, offset = cov_adj(cmod)))
  expect_output(show(m1), "y:\\(Intercept\\)")
})

test_that("lm to teeMod succeeds with weights and no SandwichLayer", {

  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  mod <- lm(y ~ assigned(), data = simdata, weights = ate(spec))

  mod_ss <- as.lmitt(mod)

  expect_s4_class(mod_ss, "teeMod")
  expect_true(inherits(mod_ss, "lm"))

  expect_identical(mod_ss$model$"(weights)"@StudySpecification, spec)

  mod_lmitt <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)

  expect_true(all(round(mod_lmitt$coef, 4) %in% round(mod_ss$coef, 4)))
  expect_identical(mod_ss@StudySpecification, mod_lmitt@StudySpecification)
})

test_that("lm to teeMod with weights and SandwichLayer", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  cmod <- lm(y ~ x, data = simdata)

  mod <- lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))

  mod_ss <- as.lmitt(mod)

  expect_s4_class(mod_ss, "teeMod")
  expect_true(inherits(mod_ss, "lm"))

  expect_equal(mod_ss$model$`(offset)`@.Data, as.numeric(cmod$fitted.values))
  expect_identical(mod_ss$model$`(offset)`@fitted_covariance_model, cmod)
  expect_equal(mod_ss$model$`(offset)`@prediction_gradient,
               stats::model.matrix(cmod))
  expect_identical(mod_ss$model$`(offset)`@keys,
                   cbind(simdata[,var_names(spec, "u")], in_Q = rep(TRUE, nrow(simdata))))

  mod_lmitt <- lmitt(y ~ 1, data = simdata, weights = ate(),
                     offset = cov_adj(cmod), specification = spec)

  expect_true(all(round(mod_lmitt$coef, 4) %in%  round(mod_ss$coef, 4)))
  expect_identical(mod_ss@StudySpecification, mod_lmitt@StudySpecification)
})

test_that("Conversion from lm to teeMod fails without an lm object", {
  expect_error(as.lmitt(1),
               "lm object")
})

test_that("lm to teeMod fails without a StudySpecification object", {
  data(simdata)

  expect_error(as.lmitt(lm(y ~ assigned(), data = simdata,
                                    weights = seq_len(nrow(simdata)))),
               "Unable to locate StudySpecification")
})

test_that("vcov, confint, etc", {
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa2, uoa1) + block(bid), data = simdata)

  sslm <- as.lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec)))

  expect_true(is.matrix(vcov(sslm)))
  expect_equal(dim(vcov(sslm)), c(3, 3))

  expect_true(is.matrix(confint(sslm)))
  expect_equal(dim(confint(sslm)), c(3, 2))


})

test_that("subsetting model with weights and/or cov_adj", {
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  mod1 <- lm(y ~ assigned(),
             data = simdata,
             weights = ate(spec),
             offset = cov_adj(lm(y ~ x, data = simdata)))

  expect_true(inherits(mod1$model$`(weights)`, "WeightedStudySpecification"))
  expect_true(inherits(mod1$model$`(offset)`, "SandwichLayer"))

  # add subsetting
  mod2 <- lm(y ~ assigned(),
             data = simdata,
             weights = ate(spec),
             offset = cov_adj(lm(y ~ x, data = simdata)),
             subset = simdata$dose < 300)

  expect_true(inherits(mod2$model$`(weights)`, "WeightedStudySpecification"))
  expect_true(inherits(mod2$model$`(offset)`, "SandwichLayer"))

})

test_that("Differing specifications found", {

  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  spec2 <- obs_spec(o ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  # This should error, since there's no `specification` argument to `lm`, `assigned()`
  # finds both the `spec` inside `ate()` and `spec2` inside `cov_adj()`
  ## expect_error(lm(y ~ assigned(),
  ##                 data = simdata,
  ##                 weights = ate(spec),
  ##                 offset = cov_adj(lm(y ~ x, data = simdata),
  ##                                  specification = spec2)),
  ##              "Multiple differing")

  expect_error(lmitt(y ~ 1,
                        data = simdata,
                        specification = spec,
                        weights = ate(),
                        offset = cov_adj(lm(y ~ x, data = simdata),
                                         specification = spec2)),
               "Multiple differing")

  expect_error(lmitt(y ~ 1,
                        data = simdata,
                        specification = spec,
                        weights = ate(),
                        offset = cov_adj(lm(y ~ x, data = simdata),
                                         specification = spec2)),
               "Multiple differing")


  mod <- lm(y ~ assigned(), data = simdata,
            offset = cov_adj(lm(y ~ x, data = simdata),
                             specification = spec2))

  expect_error(as.lmitt(mod, specification = spec), "Multiple differing")

  expect_error(lmitt(y ~ 1, data = simdata,
                     offset = cov_adj(lm(y ~ x, data = simdata),
                                      specification = spec2),
                     specification = spec),
               "Multiple differing")

  # Checking that things work when passing multiple of the same specifications
  mod1 <- lmitt(y ~ x, data = simdata, weights = ate(), specification = spec)
  mod2 <- lmitt(y ~ x, data = simdata, weights = ate(spec), specification = spec)
  expect_equal(mod1$coefficients, mod2$coefficients)

  mod3 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)
  mod4 <- lmitt(y ~ 1, data = simdata, weights = ate(spec), specification = spec)
  expect_equal(mod3$coefficients, mod4$coefficients)


  spec3 <- obs_spec(o ~ uoa(uoa1, uoa2), dat = simdata)

  # No error with a dichotomy
  expect_silent(mod <- lmitt(y ~ 1, data = simdata, specification = spec3,
                             weights = ate(spec3, dichotomy = o > 2 ~ .)))

})

test_that("teeMod object has its own evaluation environment", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec)
  mod2 <- lmitt(lm(y ~ assigned(), simdata, weights = ate(spec)))

  expect_false(identical(environment(), environment(formula(mod1))))
  expect_false(identical(environment(), environment(formula(mod2))))

  # `lmitt.formula` builds a data that incluspec the passed-in data, plus updated
  # RHS and LHS for centering and absorb.
  expect_true(all(simdata %in% environment(formula(mod1))$data))
  expect_true(all(environment(formula(mod2))$data %in% environment(formula(mod1))$data))
  expect_equal(environment(formula(mod1))$specification, spec)
  expect_equal(environment(formula(mod1))$specification, environment(formula(mod2))$specification)
})

test_that("vcov.teeMod handles vcov_tee arguments and non-SL offsets", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod1 <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec),
                     offset = cov_adj(cmod)))
  ssmod2 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)

  vmat1 <- vcov(ssmod1)
  vmat2 <- vcov(ssmod1, type = "HC0")
  vmat3 <- vcov(ssmod1, cadjust = FALSE)

  expect_error(vcov(ssmod1, type = "not_a_type"), "not defined")
  expect_identical(vmat1, vmat2)

  uoas <- apply(simdata[, c("uoa1", "uoa2")], 1, function(...) paste(..., collapse = "_"))
  vmat3 <- vcov(ssmod2, type = "CR0", cov_adj_rcorrect = "HC0")
  expect_true(all.equal(
    vmat3,
    sandwich::sandwich(ssmod2, meat. = sandwich::meatCL, cluster = uoas),
    check.attributes = FALSE))
})

test_that("confint.teeMod handles vcov_tee `type` arguments and non-SL offsets", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod1 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec,
                     offset = cov_adj(cmod))
  ssmod2 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)

  expect_error(confint(ssmod1, type = "not_a_type"), "not defined")

  # default CI
  vcov_tee_ci.95 <- ssmod1$coefficients[1:4] + sqrt(diag(vcov_tee(ssmod1))) %o%
    qt(c(0.025, 0.975), ssmod1$df.residual)
  dimnames(vcov_tee_ci.95) <- list(names(ssmod1$coefficients[1:4]), c("2.5 %", "97.5 %"))
  ci1 <- confint(ssmod1, type = "HC0")
  ci2 <- confint(ssmod1)
  expect_equal(ci1, ci2)
  expect_equal(ci1, vcov_tee_ci.95)

  # HC1 CI
  vcov_tee_HC1_ci.95 <- ssmod1$coefficients[1:4] + sqrt(diag(vcov_tee(ssmod1, type = "HC1"))) %o%
    qt(c(0.025, 0.975), ssmod1$df.residual)
  dimnames(vcov_tee_HC1_ci.95) <- list(names(ssmod1$coefficients)[1:4], c("2.5 %", "97.5 %"))
  ci_HC1 <- confint(ssmod1, type = "HC1")
  expect_equal(ci_HC1, vcov_tee_HC1_ci.95)

  # CI with different level
  vcov_tee_ci.9 <- ssmod1$coefficients[1:4] + sqrt(diag(
    vcov_tee(ssmod1, type = "CR0", cov_adj_rcorrect = "HC0"))) %o%
    qt(c(0.05, 0.95), ssmod1$df.residual)
  dimnames(vcov_tee_ci.9) <- list(names(ssmod1$coefficients)[1:4], c("5 %", "95 %"))
  ci1 <- confint(ssmod1, level = 0.9, type = "CR0", cov_adj_rcorrect = "HC0")
  expect_equal(ci1, vcov_tee_ci.9)

  # CI with lmitt.lm
  uoas <- apply(simdata[, c("uoa1", "uoa2")], 1,
                function(...) {
                  paste(..., collapse = "_")
                })

  vcovlm_z.95 <- ssmod2$coefficients[1:3] +
    sqrt(diag(sandwich::sandwich(ssmod2, meat. = sandwich::meatCL,
                            cluster = uoas, itt_rcorrect = "HC0", cov_adj_rcorrect = "HC0"))) %o%
    qt(c(0.025, 0.975), ssmod2$df.residual)
  ci1 <- confint(ssmod2, type = "HC0", cov_adj_rcorrect = "HC0")
  expect_true(all.equal(ci1, vcovlm_z.95, check.attributes = FALSE))
})

test_that("absorbed_intercepts", {
  data(simdata)

  blockedspec <- rct_spec(z ~ block(bid) + cluster(uoa1, uoa2), data = simdata)
  noblocksspec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  blocked_lmitt_fitted_absorbed <- lmitt(y ~ 1, data = simdata,
                                         specification = blockedspec, absorb = TRUE)
  expect_message(blocked_lmitt_fitted_not_absorbed <-
                   lmitt(y ~ 1, data = simdata, specification = blockedspec,
                         absorb = FALSE))
  blocked_not_lmitt_fitted <- as.lmitt(lm(y ~ assigned(blockedspec), data = simdata),
                                       specification = blockedspec)

  noblocks_lmitt_fitted_absorbed <- lmitt(y ~ 1, data = simdata,
                                          specification = noblocksspec, absorb = TRUE)
  noblocks_lmitt_fitted_not_absorbed <- lmitt(y ~ 1, data = simdata,
                                              specification = noblocksspec, absorb = FALSE)

  expect_true(blocked_lmitt_fitted_absorbed@absorbed_intercepts)
  expect_false(blocked_lmitt_fitted_not_absorbed@absorbed_intercepts)
  expect_false(blocked_not_lmitt_fitted@absorbed_intercepts)
  expect_false(noblocks_lmitt_fitted_not_absorbed@absorbed_intercepts)
  expect_true(noblocks_lmitt_fitted_absorbed@absorbed_intercepts)
})

test_that("@moderator slot", {
  data(simdata)

  blockedspec <- rct_spec(z ~ block(bid) + cluster(uoa1, uoa2), data = simdata)
  noblocksspec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  noblocks_lmitt_fittedsbgrp <- lmitt(y ~ force, data = simdata,
                                      specification = noblocksspec)
  expect_message(blocked_lmitt_fittedsbgrp <- lmitt(y ~ force, data = simdata,
                                                    specification = blockedspec))
  lmitt_fitted_nosbgrp <- lmitt(y ~ 1, data = simdata, specification = noblocksspec)
  expect_message(blocked_lmitt_fitted_nosbgrp <- lmitt(y ~ 1, data = simdata,
                                                       specification = blockedspec))
  not_lmitt_fitted <- as.lmitt(lm(y ~ assigned(noblocksspec), data = simdata),
                               specification = noblocksspec)

  expect_equal(noblocks_lmitt_fittedsbgrp@moderator, "force")
  expect_equal(blocked_lmitt_fittedsbgrp@moderator, "force")
  expect_equal(lmitt_fitted_nosbgrp@moderator, vector("character"))
  expect_equal(blocked_lmitt_fitted_nosbgrp@moderator, vector("character"))
  expect_equal(not_lmitt_fitted@moderator, vector("character"))
})

test_that("estfun.teeMod requires a certain model class", {
  data(simdata)

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec)
  mod@.S3Class <- "not_a_mod"

  expect_error(estfun(mod), "must have been fitted using")
})

test_that(paste("estfun.teeMod returns original psi if no offset or no",
                "SandwichLayer offset"), {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  ca <- cov_adj(cmod, newdata = simdata, specification = spec)

  nolmittmod1 <- lm(y ~ z, simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec)
  nolmittmod2 <- lm(y ~ z, data = simdata, offset = ca)
  mod2 <- lmitt(y ~ 1, data = simdata, specification = spec, offset = stats::predict(cmod))

  ef_expected1 <- cbind(estfun(nolmittmod1), estfun(mod1@ctrl_means_model))
  ef_expected2 <- cbind(estfun(nolmittmod2), estfun(mod2@ctrl_means_model))

  expect_true(all.equal(estfun(mod1), ef_expected1, check.attributes = FALSE))
  expect_true(all.equal(estfun(mod2), ef_expected2, check.attributes = FALSE))
})

test_that("estfun.teeMod gets scaling constants right with no overlap", {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata
  nc <- 30
  nq <- nrow(simdata_copy)
  n <- nc + nq
  cmod_ssta <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_ssta)

  simdata_copy$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), simdata_copy)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata_copy, offset = cov_adj(cmod))

  ef_pieces <- .align_and_extend_estfuns(dmod, by = "id", type_psi = "HC0", type_phi = "HC0")
  a11_inv <- .get_a11_inverse(dmod)
  a21 <- .get_a21(dmod)[1:2,]
  expect_equal(
    estfun(dmod, type_psi = "HC0", type_phi = "HC0")[,1:2],
    ef_pieces[["psi"]] - nq / nc * ef_pieces[["phi"]] %*% t(a11_inv) %*% t(a21)
  )
})

test_that("estfun.teeMod gets scaling constants right with partial overlap", {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata
  nq <- nrow(simdata_copy)
  simdata_copy$id <- seq(nq)

  aux_nc <- 30
  n <- nc <- nq + aux_nc
  aux_cmod_ssta <- data.frame(y = rnorm(aux_nc), x = rnorm(aux_nc), id = nq + seq(aux_nc))
  cmod <- lm(y ~ x, rbind(simdata_copy[, c("y", "x", "id")], aux_cmod_ssta))

  spec <- rct_spec(z ~ unitid(id), simdata_copy)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata_copy, offset = cov_adj(cmod))

  ef_pieces <- .align_and_extend_estfuns(dmod, by = "id", type_psi = "HC0", type_phi = "HC0")
  a11_inv <- .get_a11_inverse(dmod)
  a21 <- .get_a21(dmod)[1:2,]
  expect_equal(
    estfun(dmod, type_psi = "HC0", type_phi = "HC0")[,1:2],
    ef_pieces[["psi"]] - nq / nc * ef_pieces[["phi"]] %*% t(a11_inv) %*% t(a21)
  )
})

test_that("estfun.teeMod with missing values", {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata
  simdata_copy$y[1:3] <- NA_real_
  nc <- 30
  nq <- nrow(simdata_copy)
  n <- nc + nq
  cmod_ssta <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_ssta)
  
  simdata_copy$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), simdata_copy)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata_copy, offset = cov_adj(cmod))
  
  ef <- estfun(dmod, type_psi = "HC0", type_phi = "HC0")[,1:2]
  expect_equal(nrow(ef), nc + nq)
  expect_true(all.equal(ef[1:3,], matrix(0, nrow = 3, ncol = 2), check.attributes = FALSE))
  class(dmod$na.action) <- "exclude"
  ef_pieces <- .align_and_extend_estfuns(dmod, by = "id", type_psi = "HC0", type_phi = "HC0")
  a11_inv <- .get_a11_inverse(dmod)
  a21 <- .get_a21(dmod)[1:2,]
  expect_equal(
    ef,
    ef_pieces[["psi"]] - nq / nc * ef_pieces[["phi"]] %*% t(a11_inv) %*% t(a21)
  )
})

test_that(paste("estfun.teeMod returns correct dimensions and alignment",
                "when exact alignment between C and Q is possible"), {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata
  simdata_copy$uid <- rownames(simdata_copy)

  shuffled_simdata <- simdata_copy[sample(rownames(simdata_copy)),]
  cmod <- lm(y ~ x, simdata_copy)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  mod1 <- lmitt(y ~ 1, data = simdata_copy, specification = spec, offset = cov_adj(cmod, by = "uid"))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec, offset = cov_adj(cmod, by = "uid"))

  expect_equal(dim(estfun(mod1, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata_copy), 4))
  expect_equal(estfun(mod1, type_psi = "HC0", type_phi = "HC0"),
               estfun(mod2, type_psi = "HC0", type_phi = "HC0"))
})

test_that(paste("estfun.teeMod returns correct dimensions when only",
                "inexact alignment between C and Q is possible"), {
  set.seed(438)
  data(simdata)

  shuffled_simdata <- simdata[sample(rownames(simdata)),]
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec, offset = cov_adj(cmod))

  expect_equal(dim(estfun(mod1, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata), 4))
  expect_equal(dim(estfun(mod2, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata), 4))
})

test_that(paste("estfun.teeMod returns correct dimensions and alignment",
                "when exact alignment between C and Q is possible but they only",
                "partially overlap"), {
  set.seed(438)
  data(simdata)
  simdata$uid <- rownames(simdata)

  Q_data <- simdata[simdata$uid %in% seq_len(20),]
  shuffled_Q_data <- simdata[sample(rownames(Q_data)),]
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  mod1 <- lmitt(y ~ 1, data = Q_data, specification = spec, offset = cov_adj(cmod, by = "uid"))
  mod2 <- lmitt(y ~ 1, data = shuffled_Q_data, specification = spec, offset = cov_adj(cmod, by = "uid"))

  expect_equal(dim(estfun(mod1, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata), 4))
  expect_equal(estfun(mod1, type_psi = "HC0", type_phi = "HC0"),
               estfun(mod2, type_psi = "HC0", type_phi = "HC0"))
})

test_that(paste("estfun.teeMod returns correct dimensions for partial",
                "overlap of C and Q when only inexact alignment is possible"), {
  data(simdata)
  set.seed(401)
  cmod_ssta <- rbind(
    simdata[simdata$uoa1 == 1, c("x", "y", "uoa1", "uoa2")],
    data.frame("x" = rnorm(100), "y" = rnorm(100), "uoa1" = NA, "uoa2" = NA)
  )

  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))

  expect_equal(dim(estfun(mod, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata) + nrow(cmod_ssta[is.na(cmod_ssta$uoa1),]), 4))
})

test_that(paste("estfun.teeMod returns correct dimensions for no",
                "overlap of C and Q when only inexact alignment is possible"), {
  data(simdata)
  set.seed(401)
  cmod_ssta <- data.frame(
    "x" = rnorm(100), "y" = rnorm(100), "uoa1" = NA, "uoa2" = NA
  )

  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))

  expect_equal(dim(estfun(mod, type_psi = "HC0", type_phi = "HC0")),
               c(nrow(simdata) + nrow(cmod_ssta), 4))
})

test_that("estfun zeros out NA's from ctrl means regression", {
  moddata <- data.frame(a = c(rep(c(0, 1), each = 5), NA_real_), y = rnorm(11), id = seq_len(11))
  spec <- rct_spec(a ~ unitid(id), data = moddata)
  mod <- lmitt(y ~ 1, data = moddata, spec = spec)
  ef <- estfun(mod, type_psi = "HC0", type_phi = "HC0")
  expect_true(all(!is.na(ef)))
  expect_true(!all(ef == 0))
  expect_equal(ef[11,3], 0)
})

test_that("estfun with invalid values_from", {
  data(simdata)
  not_all_simdata <- simdata[seq_len(7),]
  values_from <- lm(y ~ x, not_all_simdata)
  tm <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata)
  expect_error(
    estfun(tm, type_psi = "HC0", values = values_from),
    "Lengths"
  )
})

test_that("estfun (no covariance adjustment) with values_from", {
  data(simdata)
  values_from <- lm(y ~ 1, simdata)
  tm <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata)
  
  ## no missing
  ef <- estfun(tm, type_psi = "HC0", values = values_from)[,1:2]
  expect_true(all.equal(
    ef,
    stats::model.matrix(tm) * stats::residuals(values_from, type = "working"),
    check.attributes = FALSE
  ))
  
  ## missing
  data_with_missingness <- simdata
  data_with_missingness$y[c(3, 9, 25, 42)] <- NA_real_
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data_with_missingness)
  values_from <- lm(y ~ 1, data_with_missingness)
  tm <- lmitt(y ~ 1, spec, data_with_missingness)
  class(attr(tm$model, "na.action")) <- "pass"
  class(values_from$na.action) <- "exclude"
  
  ef <- estfun(tm, type_psi = "HC0", values = values_from)[,1:2]
  resids <- stats::residuals(values_from, type = "working")
  expect_true(all.equal(
    ef,
    stats::model.matrix(stats::formula(tm),
                        stats::model.frame(tm, na.action = na.pass)) *
      replace(resids, is.na(resids), 0),
    check.attributes = FALSE
  ))
  
  ## resids_from is a glm
  simdata$ybinom <- rbinom(nrow(simdata), 1, 0.5)
  values_from <- glm(ybinom ~ 1, simdata, family = "binomial")
  tm <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata)
  
  ef <- estfun(tm, type_psi = "HC0", values = values_from)[,1:2]
  expect_true(all.equal(
    ef,
    stats::model.matrix(tm) *
      stats::residuals(values_from, type = "working") *
      stats::weights(values_from, type = "working"),
    check.attributes = FALSE
  ))
})

test_that(paste(".align_and_extend_estfuns (estfun with covariance adjustment)",
                "with values_from"), {
  ## no missing
  camod <- lm(y ~ x, simdata)
  values_from <- lm(y ~ 1, simdata)
  tm <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata, offset = cov_adj(camod))
  
  efs <- .align_and_extend_estfuns(tm,
                                   values = values_from,
                                   itt_rcorrect = "HC0",
                                   cov_adj_rcorrect = "HC0")
  expect_true(all.equal(
    efs$psi,
    stats::model.matrix(tm) * stats::residuals(values_from, type = "working"),
    check.attributes = FALSE
  ))
  
  ## missing
  data_with_missingness <- simdata
  data_with_missingness$uid <- seq_len(nrow(simdata))
  data_with_missingness$y[c(3, 9, 25, 42)] <- NA_real_
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data_with_missingness)
  camod <- lm(y ~ x, data_with_missingness)
  resids_from <- lm(y ~ 1, data_with_missingness)
  tm <- lmitt(y ~ 1, spec, data_with_missingness,
              offset = cov_adj(camod, by = "uid"))
  class(tm$na.action) <- "exclude"
  class(attr(tm$model, "na.action")) <- "pass"
  class(resids_from$na.action) <- "exclude"
  
  efs <- .align_and_extend_estfuns(tm,
                                   values = values_from,
                                   itt_rcorrect = "HC0",
                                   cov_adj_rcorrect = "HC0")
  resids <- stats::residuals(values_from, type = "working")
  X <- stats::model.matrix(stats::formula(tm),
                           stats::model.frame(tm, na.action = na.pass))
  expect_true(all.equal(
    efs$psi,
    rbind(matrix(0, nrow = length(tm$na.action), ncol = ncol(X),
                 dimnames = list(NULL, colnames(X))),
          X[sort(as.character(setdiff(seq_len(nrow(simdata)),
                                      tm$na.action))),]) *
      c(rep(0, length(tm$na.action)),
        replace(resids, is.na(resids), 0)[
          as.numeric(sort(as.character(setdiff(seq_len(nrow(simdata)),
                                               tm$na.action))))
          ]),
    check.attributes = FALSE
  ))
})

test_that(".compute_loo_resids with values_from", {
  ## no missing
  camod <- lm(y ~ x, simdata)
  values_from <- lm(y ~ 1, simdata)
  tm <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata, offset = cov_adj(camod))
  resids <- .compute_loo_resids(tm,
                                c("uoa1", "uoa2"),
                                values_from = values_from)
  
  sbst <- paste(simdata$uoa1, simdata$uoa2, sep = "_") != "1_1"
  loo_mod <- lm(y ~ x, simdata, subset = sbst)
  expect_true(
    all.equal(
      resids[!sbst],
      residuals(values_from, type = "working")[!sbst] +
        stats::model.offset(stats::model.frame(tm, na.action = na.pass))[!sbst] -
        predict(loo_mod, simdata[!sbst,]),
      check.attributes = FALSE
    )
  )
  
  ## missing
  data_with_missingness <- simdata
  data_with_missingness$uid <- seq_len(nrow(simdata))
  data_with_missingness$y[c(3, 9, 25, 42)] <- NA_real_
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data_with_missingness)
  camod <- lm(y ~ x, data_with_missingness)
  resids_from <- lm(y ~ 1, data_with_missingness)
  tm <- lmitt(y ~ 1, spec, data_with_missingness,
              offset = cov_adj(camod, by = "uid"))
  class(tm$na.action) <- "exclude"
  class(attr(tm$model, "na.action")) <- "pass"
  class(resids_from$na.action) <- "exclude"

  resids <- .compute_loo_resids(tm,
                                c("uoa1", "uoa2"),
                                values_from = values_from)
  sbst <- paste(simdata$uoa1, simdata$uoa2, sep = "_") != "1_1"
  loo_mod <- lm(y ~ x, data_with_missingness, subset = sbst)
  expect_true(
    all.equal(
      resids[!sbst],
      residuals(values_from, type = "working")[!sbst] +
        stats::model.offset(stats::model.frame(tm, na.action = na.pass))[!sbst] -
        predict(loo_mod, simdata[!sbst,]),
      check.attributes = FALSE
    )
  )
})

if (requireNamespace("robustbase", quietly = TRUE)) {
  test_that("estfun.teeMod returns correct dimensions for rectangular A11_inv", {
    data(simdata)
    cmod <- robustbase::lmrob(y ~ x, simdata)
    spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
    dmod <- lmitt(lm(y ~ z.(spec), simdata, offset = cov_adj(cmod)), specification = spec)

    out <- estfun(dmod, type_psi = "HC0", type_phi = "HC0")
    expect_equal(dim(out), c(50, 4))
  })
}

test_that(paste("bread.teeMod returns bread.lm for teeMod objects",
                "with non-SandwichLayer or NULL offsets"), {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  ca <- predict(cmod)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  m1 <- lmitt(y ~ 1, data = simdata, specification = spec)
  m2 <- lmitt(y ~ 1, data = simdata, specification = spec, offset = ca)

  expected_out1 <- matrix(0, nrow = 3, ncol = 3)
  expected_out1[1:2, 1:2] <- nrow(simdata) * chol2inv(m1$qr$qr)
  expected_out1[3, 3] <- nrow(simdata) / sum(weights(m1@ctrl_means_model))
  dimnames(expected_out1) <- list(names(m1$coefficients), names(m1$coefficients))
  expect_equal(bread(m1), expected_out1)

  expected_out2 <- matrix(0, nrow = 4, ncol = 4)
  expected_out2[1:2, 1:2] <- nrow(simdata) * chol2inv(m2$qr$qr)
  expected_out2[3:4, 3:4] <- diag(2) * nrow(simdata) / sum(weights(m2@ctrl_means_model))
  dimnames(expected_out2) <- list(names(m2$coefficients), names(m2$coefficients))
  expect_equal(bread(m2), expected_out2)
})

test_that(paste("bread.teeMod returns expected output for full overlap",
                "of C and Q"), {
  data(simdata)
  simdata$uid <- seq_len(nrow(simdata))

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ uoa(uid), simdata)
  m <- lmitt(y ~ 1, data = simdata, specification = spec, weights = ate(spec),
             offset = cov_adj(cmod))

  expected_out <- matrix(0, nrow = 4, ncol = 4)
  expected_out[1:2, 1:2] <- nrow(simdata) * chol2inv(m$qr$qr)
  expected_out[3:4, 3:4] <- diag(2) * nrow(simdata) / sum(weights(m@ctrl_means_model))
  dimnames(expected_out) <- list(names(m$coefficients), names(m$coefficients))
  expect_equal(bread(m), expected_out)
})

test_that(paste("bread.teeMod returns expected output for partial overlap",
                "of C and Q"), {
  set.seed(438)
  data(simdata)
  nq <- nrow(simdata)
  simdata$id <- seq(nq)

  aux_nc <- 30
  n <- nc <- nq + aux_nc
  aux_cmod_ssta <- data.frame(y = rnorm(aux_nc), x = rnorm(aux_nc), id = nq + seq(aux_nc))
  cmod <- lm(y ~ x, rbind(simdata[, c("y", "x", "id")], aux_cmod_ssta))

  spec <- rct_spec(z ~ unitid(id), simdata)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata, offset = cov_adj(cmod))

  expected_out <- matrix(0, nrow = 4, ncol = 4)
  expected_out[1:2, 1:2] <- n * chol2inv(dmod$qr$qr)
  expected_out[3:4, 3:4] <- diag(2) * n / sum(weights(dmod@ctrl_means_model))
  dimnames(expected_out) <- list(names(dmod$coefficients), names(dmod$coefficients))
  expect_equal(bread(dmod), expected_out)
})

test_that(paste("bread.teeMod returns expected output for no overlap",
                "of C and Q"), {
  set.seed(438)
  data(simdata)
  nc <- 30
  nq <- nrow(simdata)
  n <- nc + nq
  cmod_ssta <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_ssta)

  simdata$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), simdata)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata, offset = cov_adj(cmod))

  expected_out <- matrix(0, nrow = 4, ncol = 4)
  expected_out[1:2, 1:2] <- n * chol2inv(dmod$qr$qr)
  expected_out[3:4, 3:4] <- diag(2) * n / sum(weights(dmod@ctrl_means_model))
  dimnames(expected_out) <- list(names(dmod$coefficients), names(dmod$coefficients))
  expect_equal(bread(dmod), expected_out)
})

test_that("bread.teeMod handles model with less than full rank", {
  data(simdata)
  copy_simdata <- simdata
  copy_simdata$o_fac <- as.factor(copy_simdata$o)
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), copy_simdata)

  ### lmitt.formula
  ssmod <- lmitt(y ~ o_fac, data = copy_simdata, specification = spec, offset = cov_adj(cmod))
  p <- ssmod$rank
  keep_ix <- ssmod$qr$pivot[1L:p]
  cm_mm <- model.matrix(ssmod@ctrl_means_model)
  q <- ncol(cm_mm)
  
  expected_out <- matrix(0, nrow = p + 2 * q, ncol = p + 2 * q)
  expected_out[1:p, 1:p] <- nrow(simdata) * solve(crossprod(model.matrix(ssmod))[keep_ix,keep_ix])
  expected_out[(p+1):(p+q), (p+1):(p+q)] <- nrow(simdata) * chol2inv(
    ssmod@ctrl_means_model$qr$qr)
  expected_out[(p+q+1):(p+2*q), (p+q+1):(p+2*q)] <- nrow(simdata) * chol2inv(
    ssmod@ctrl_means_model$qr$qr)
  expect_true(all.equal(bread(ssmod), expected_out, check.attributes = FALSE))
})

test_that("rcorrect fail", {
  r <- rep(c(-1, 1), 20)
  expect_error(
    .rcorrect(r, x = "not a teeMod but it's ok", model = "not valid but it's ok",
             type = "not a bias correction"),
    "not available"
  )
  
  expect_error(
    .rcorrect(r, x = list(c(1, 2)), model = "cov_adj",  type = "HC2", cluster_cols = "uoa1"),
    "must have a SandwichLayer"
  )
})

test_that("rcorrect HC/CR/MB/DB0", {
  r <- rep(c(-1, 1), 20)
  expect_equal(
    cr <- .rcorrect(r, x = "not a teeMod but it's ok", model = "not valid but it's ok",
                   type = "HC0"),
    r
  )
  expect_equal(cr, .rcorrect(r, x = "not a teeMod but it's ok", model = "not valid but it's ok",
                            type = "CR0"))
  expect_equal(cr, .rcorrect(r, x = "not a teeMod but it's ok", model = "not valid but it's ok",
                            type = "MB0"))
  expect_equal(cr, .rcorrect(r, x = "not a teeMod but it's ok", model = "not valid but it's ok",
                            type = "DB0"))
})

test_that("rcorrect (HC/CR/MB)1", {
  set.seed(749)
  udata <- data.frame(cid = seq_len(10),
                      bid = rep(letters[1:2], 5),
                      x1 = rnorm(10),
                      x2 = factor(rep(seq_len(3), 4)[1:10]),
                      a = rep(c(0, 1), each = 5),
                      y = rnorm(10))
  idata <- data.frame(cid = rep(udata$cid, each = 10),
                      bid = rep(rep(letters[1:2], each = 10), 5),
                      x1 = rnorm(100),
                      x2 = factor(rep(seq_len(3), 34)[1:100]),
                      y = rnorm(100))
  
  speci <- rct_spec(a ~ unitid(cid), udata)
  cmod <- lm(y ~ x1 + x2, idata)
  xm <- lmitt(y ~ 1, speci, idata, offset = cov_adj(cmod))
  
  r <- xm$residuals
  g <- nrow(udata)
  n <- nrow(idata)
  k <- 6
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC1"),
    r * sqrt((n-1) / (n-k) * g / (g-1))
  )
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "CR1"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "MB1"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "cov_adj", type = "HC1"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "cov_adj", type = "CR1"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "cov_adj", type = "MB1"))
  
  # non-default clustering
  cluster <- "bid"
  cls <- .make_uoa_ids(xm, "MB", cluster)
  g <- 2
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC1", cluster_cols = cluster, cluster = cls),
    r * sqrt((n-1) / (n-k) * g / (g-1))
  )
  expect_equal(
    cr,
    .rcorrect(r, x = xm, model = "itt", type = "CR1", cluster_cols = cluster, cluster = cls)
  )
  expect_equal(
    cr,
    .rcorrect(r, x = xm, model = "itt", type = "MB1", cluster_cols = cluster, cluster = cls)
  )
  expect_equal(
    cr,
    .rcorrect(r, x = xm, model = "cov_adj", type = "MB1", cluster_cols = cluster, cluster = cls)
  )
})

if (requireNamespace("robustbase", quietly = TRUE)) {
  test_that("rcorrect (HC/CR/MB)1 with lmrob cov_adj", {
    set.seed(749)
    udata <- data.frame(cid = seq_len(10),
                        bid = rep(letters[1:2], 5),
                        x1 = rnorm(10),
                        x2 = factor(rep(seq_len(3), 4)[1:10]),
                        a = rep(c(0, 1), each = 5),
                        y = rnorm(10))
    idata <- data.frame(cid = rep(udata$cid, each = 10),
                        bid = rep(rep(letters[1:2], each = 10), 5),
                        x1 = rnorm(100),
                        x2 = factor(rep(seq_len(3), 34)[1:100]),
                        y = rnorm(100))
    
    speci <- rct_spec(a ~ unitid(cid), udata)
    cmod <- robustbase::lmrob(y ~ x1 + x2, idata)
    xm <- lmitt(y ~ 1, speci, idata, offset = cov_adj(cmod))
    
    r <- xm$residuals
    g <- nrow(udata)
    n <- nrow(idata)
    k <- 7
    expect_equal(
      cr <- .rcorrect(r, x = xm, model = "itt", type = "HC1"),
      r * sqrt((n-1) / (n-k) * g / (g-1))
    )
  })
}

test_that("rcorrect (HC/CR/MB)2, no clustering", {
  set.seed(749)
  udata <- data.frame(cid = seq_len(10),
                      x1 = rnorm(10),
                      x2 = factor(rep(seq_len(3), 4)[1:10]),
                      a = rep(c(0, 1), each = 5),
                      y = rnorm(10))
  speci <- rct_spec(a ~ unitid(cid), udata)
  cmod <- lm(y ~ x1 + x2, udata)
  xm <- lmitt(y ~ 1, speci, udata, offset = cov_adj(cmod))
  
  r <- xm$residuals
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC2"),
    r / sqrt(1-stats::hatvalues(xm))
  )
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "CR2"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "MB2"))
  
  r <- cmod$residuals
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "cov_adj", type = "HC2"),
    r / sqrt(1-stats::hatvalues(cmod))
  )
  expect_equal(cr, .rcorrect(r, x = xm, model = "cov_adj", type = "CR2"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "cov_adj", type = "MB2"))
  
  # subset
  xm <- lmitt(y ~ 1, speci, udata, subset = cid > 3, offset = cov_adj(cmod))
  r <- stats::residuals(xm, "working")
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC2"),
    r / sqrt(1-stats::hatvalues(xm))
  )
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "CR2"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "MB2"))
  
  # NA's
  udata$y[9:10] <- NA_real_
  cmod <- lm(y ~ x1 + x2, udata)
  xm <- lmitt(y ~ 1, speci, udata, offset = cov_adj(cmod))
  class(xm$na.action) <- "exclude"
  r <- stats::residuals(xm, "working")
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC2"),
    r / sqrt(1-stats::hatvalues(xm))
  )
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "CR2"))
  expect_equal(cr, .rcorrect(r, x = xm, model = "itt", type = "MB2"))
})

test_that("rcorrect (HC/CR/MB)2, clustering", {
  set.seed(749)
  udata <- data.frame(cid = seq_len(10), a = rep(c(0, 1), each = 5))
  idata <- data.frame(cid = c(rep(udata$cid, each = 10), rep(NA_real_, 50)),
                      x1 = rnorm(150),
                      x2 = factor(rep(seq_len(3), 50)),
                      y = rnorm(150))
  speci <- rct_spec(a ~ unitid(cid), udata)
  cmod <- lm(y ~ x1 + x2, idata)
  da_data <- idata[!is.na(idata$cid),,drop=FALSE]
  xm <- lmitt(y ~ 1, speci, da_data, weights = ate(speci),
              offset = cov_adj(cmod))

  r <- xm$residuals
  pm <- chol2inv(xm$qr$qr)
  X <- stats::model.matrix(xm)
  cids <- unique(udata$cid)
  
  expected <- Reduce(
    c,
    mapply(
      function(c, r, cls, w) {
        ix <- cls == c
        Xc <- X[ix,,drop=FALSE]
        Pc <- Xc %*% pm %*% t(Xc) %*% diag(w[ix], nrow = sum(ix))
        schur <- eigen(diag(nrow = sum(ix)) - Pc)
        schur$vector %*% diag(1/sqrt(schur$values),
                              nrow = sum(ix)) %*% solve(schur$vector) %*% r[ix]
      },
      cids,
      SIMPLIFY = FALSE,
      MoreArgs = list(r = r, cls = da_data$cid, w = xm$weights)
    )
  )
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC2"),
    expected
  )
  expect_equal(.rcorrect(r, x = xm, model = "itt", type = "CR2"), cr)
  expect_equal(.rcorrect(r, x = xm, model = "itt", type = "MB2"), cr)
  
  # subset
  xm <- lmitt(y ~ 1, speci, idata, subset = !is.na(cid), offset = cov_adj(cmod))
  
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "itt", type = "HC2"),
    expected
  )
  expect_equal(.rcorrect(r, x = xm, model = "itt", type = "CR2"), cr)
  expect_equal(.rcorrect(r, x = xm, model = "itt", type = "MB2"), cr)
  
  # cov adj
  r <- cmod$residuals
  pm <- chol2inv(cmod$qr$qr)
  X <- stats::model.matrix(cmod)
  cmod_cls <- c(idata$cid[1:sum(!is.na(idata$cid))],
                sum(!is.na(idata$cid)) + seq_len(sum(is.na(idata$cid))))
  cids <- unique(cmod_cls)
  
  expected <- Reduce(
    c,
    mapply(
      function(c, r, cls, w) {
        ix <- cls == c
        Xc <- X[ix,,drop=FALSE]
        Pc <- Xc %*% pm %*% t(Xc) %*% diag(w[ix], nrow = sum(ix))
        schur <- eigen(diag(nrow = sum(ix)) - Pc)
        schur$vector %*% diag(1/sqrt(schur$values),
                              nrow = sum(ix)) %*% solve(schur$vector) %*% r[ix]
      },
      cids,
      SIMPLIFY = FALSE,
      MoreArgs = list(r = r, cls = cmod_cls, w = rep(1, nrow(idata)))
    )
  )
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "cov_adj", type = "HC2"),
    expected
  )
  expect_equal(.rcorrect(r, x = xm, model = "cov_adj", type = "CR2"), cr)
  expect_equal(.rcorrect(r, x = xm, model = "cov_adj", type = "MB2"), cr)
  
  # NA's
  idata$y[1:3] <- NA_real_
  cmod <- lm(y ~ x1 + x2, idata)
  da_data <- idata[!is.na(idata$cid),,drop=FALSE]
  xm <- lmitt(y ~ 1, speci, da_data, weights = ate(speci),
              offset = cov_adj(cmod))
  
  r <- xm$residuals
  pm <- chol2inv(xm$qr$qr)
  X <- stats::model.matrix(xm)
  cids <- unique(udata$cid)
  
  expected <- c(
    rep(NA, 3),
    Reduce(
      c,
      mapply(
        function(c, r, cls, w) {
          ix <- cls == c
          Xc <- X[ix,,drop=FALSE]
          Pc <- Xc %*% pm %*% t(Xc) %*% diag(w[ix], nrow = sum(ix))
          schur <- eigen(diag(nrow = sum(ix)) - Pc)
          schur$vector %*%
            diag(1/sqrt(schur$values),
                 nrow = sum(ix)) %*% solve(schur$vector) %*%
            r[ix]
        },
        cids,
        SIMPLIFY = FALSE,
        MoreArgs = list(r = r,
                        cls = da_data$cid[4:nrow(da_data)],
                        w = xm$weights)
      )
    )
  )

  class(xm$na.action) <- "exclude"
  expect_equal(
    cr <- .rcorrect(c(rep(NA_real_, 3), r), x = xm, model = "itt",
                    type = "HC2"),
    expected
  )
  
  # NA's + subset (check the model.frame calls are correct)
  xm <- lmitt(y ~ 1, speci, idata, subset = !is.na(cid), offset = cov_adj(cmod))
  
  r <- xm$residuals
  class(xm$na.action) <- "exclude"
  expect_equal(
    cr <- .rcorrect(c(rep(NA_real_, 3), r), x = xm, model = "itt",
                    type = "HC2"),
    expected
  )
  
  # all NA's in a given cluster
  idata$y[idata$cid == 1] <- NA_real_
  cmod <- lm(y ~ x1 + x2, idata)
  class(cmod$na.action) <- "exclude"
  da_data <- idata[!is.na(idata$cid),,drop=FALSE]
  xm <- lmitt(y ~ 1, speci, da_data, weights = ate(speci),
              offset = cov_adj(cmod))
  class(xm$na.action) <- "exclude"
  expect_equal(
    .rcorrect(
      stats::residuals(cmod, type = "working"),
      x = xm,
      model = "cov_adj",
      type = "HC2"
    )[idata$cid == 1 & !is.na(idata$cid)],
    rep(NA_real_, sum(!is.na(idata$cid) & idata$cid == 1))
  )
  expect_true(all(!is.na(vcov_tee(x = xm, type = "CR2"))))
  
  # glm
  idata <- data.frame(cid = c(rep(udata$cid, each = 10), rep(NA_real_, 50)),
                      x1 = rnorm(150),
                      x2 = factor(rep(seq_len(3), 50)),
                      y = round(runif(150)))
  cmod <- glm(y ~ x1 + x2, idata, family = binomial())
  xm <- lmitt(y ~ 1, speci, idata, subset = !is.na(cid), offset = cov_adj(cmod))
  
  r <- stats::residuals(cmod, type = "working")
  pm <- chol2inv(cmod$qr$qr)
  X <- stats::model.matrix(cmod)
  cids <- unique(cmod_cls)

  expected <- Reduce(
    c,
    mapply(
      function(c, r, cls, w) {
        ix <- cls == c
        Xc <- X[ix,,drop=FALSE]
        Pc <- Xc %*% pm %*% t(Xc) %*% diag(w[ix], nrow = sum(ix))
        schur <- eigen(diag(nrow = sum(ix)) - Pc)
        schur$vector %*% diag(1/sqrt(schur$values),
                              nrow = sum(ix)) %*% solve(schur$vector) %*% r[ix]
      },
      cids,
      SIMPLIFY = FALSE,
      MoreArgs = list(r = r,
                      cls = cmod_cls,
                      w = weights(cmod, type = "working"))
    )
  )
  
  expect_equal(
    cr <- .rcorrect(r, x = xm, model = "cov_adj", type = "HC2"),
    expected
  )
  
  # ..uoa..
  sdat <- data.frame(b = rep(replicate(50,
                                       paste(sample(letters, 10, TRUE),
                                             collapse = "")),
                             2),
                     a = rep(c(0, 1), each = 50),
                     y = 2 * rep(c(0, 1), each = 50) + 0.35 * rnorm(100))
  suppressWarnings(next_spec <- rct_spec(a ~ block(b), sdat))
  next_xm <- lmitt(y ~ 1, next_spec, sdat, weights = "ate")
  
  r <- next_xm$residuals
  pm <- chol2inv(next_xm$qr$qr)
  X <- stats::model.matrix(next_xm)
  expected <- numeric(nrow(sdat))
  for (ix in seq_len(nrow(sdat))) {
    Xc <- X[ix,,drop=FALSE]
    Pc <- Xc %*% pm %*% t(Xc) %*% diag(next_xm$weights[ix], nrow = length(ix))
    schur <- eigen(diag(nrow = length(ix)) - Pc)
    crc <- schur$vector %*% diag(1/sqrt(schur$values), nrow = length(ix)) %*% 
      solve(schur$vector) %*% r[ix]
    expected[ix] <- crc
  }
  expect_equal(
    cr <- unname(.rcorrect(r, x = next_xm, model = "itt", type = "HC2")),
    expected
  )
})

test_that(paste(".align_and_extend_estfuns fails if not a teeMod object",
                "with a SandwichLayer offset"), {
  data(simdata)

  mod1 <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  mod2 <- lmitt(y ~ 1, specification = spec, data = simdata)
  expect_error(.align_and_extend_estfuns(mod1), "must be a fitted")
  expect_error(.align_and_extend_estfuns(mod2), "must be a fitted")
})

test_that(paste(".align_and_extend_estfuns with `by` and the samples fully",
                "overlap (uses jackknifed first-stage coefficients)"), {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata

  simdata_copy$obs_id <- seq_len(nrow(simdata_copy))
  shuffle_ix <- sample(rownames(simdata_copy))
  shuffled_simdata <- simdata_copy[shuffle_ix,]
  cmod1 <- lm(y ~ x, simdata_copy)
  cmod2 <- lm(y ~ x, shuffled_simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  mod1 <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
                offset = cov_adj(cmod1, by = "obs_id"))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec,
                offset = cov_adj(cmod2, by = "obs_id"))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  C_cls <- paste(simdata_copy$uoa1, simdata_copy$uoa2, sep = "_")
  jk_units <- unique(C_cls)
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)
  loo_preds <- rowSums(X * t(loo_cmod[, C_cls, drop=FALSE]))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata_copy$y - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, itt_rcorrect = "HC0",
                                   cov_adj_rcorrect = "HC0",
                                   loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, itt_rcorrect = "HC0",
                                   cov_adj_rcorrect = "HC0",
                                   loco_residuals = TRUE)

  # tests to run (for each .align_and_extend_estfuns() test):
  # 1) do we get a matrix of estimating equations of dimension n?
  # 2) are rows' contributions to the matrix of estimating equations aligned in
  #     phi and psi?
  # 3) are variance estimates invariant to the original order of the data,
  #     even when we change the clustering level?
  expect_equal(dim(ef1$phi), c(nrow(simdata_copy), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata_copy), 2))
  expect_true(all.equal(
    ef1$phi,
    cmod_ef[sort(as.character(seq(nrow(simdata_copy)))),],
    check.attributes = FALSE))
  expect_true(all.equal(
    ef1$psi,
    mod_lm_ef[sort(as.character(seq(nrow(simdata_copy)))),],
    check.attributes = FALSE))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
  expect_equal(vcov_tee(mod1, cluster = "bid", loco_residuals = TRUE),
               vcov_tee(mod2, cluster = "bid", loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns with `by` and Q is a subset of C",
                "(uses jackknifing)"), {
  set.seed(438)
  data(simdata)

  simdata$obs_id <- seq_len(nrow(simdata))
  simdata_shuffle_ix <- sample(rownames(simdata))
  shuffled_simdata <- simdata[simdata_shuffle_ix,]
  Q_ix <- 1L:20L
  Q_data <- simdata[Q_ix,]
  Q_shuffle_ix <- sample(rownames(Q_data))
  shuffled_Q_data <- Q_data[Q_shuffle_ix,]
  cmod1 <- lm(y ~ x, simdata)
  cmod2 <- lm(y ~ x, shuffled_simdata)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_Q_data)
  mod1 <- lmitt(y ~ 1, data = Q_data, specification = spec1,
                offset = cov_adj(cmod1, by = "obs_id"))
  mod2 <- lmitt(y ~ 1, data = shuffled_Q_data, specification = spec2,
                offset = cov_adj(cmod2, by = "obs_id"))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  C_cls <- paste(simdata$uoa1, simdata$uoa2, sep = "_")
  jk_units <- unique(C_cls[Q_ix])
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)[Q_ix,,drop=FALSE] 
  loo_preds <- rowSums(X * t(loo_cmod[, C_cls[Q_ix], drop=FALSE]))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata$y[Q_ix] - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, loco_residuals = TRUE)

  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  nonzero_ix <- as.numeric(sort(as.character(Q_ix)))
  zero_ix <- setdiff(seq_len(50), Q_ix)
  expect_true(all.equal(ef1$phi, cmod_ef[c(nonzero_ix, zero_ix),],
                        check.attributes = FALSE))
  expect_true(all.equal(ef1$psi[Q_ix,], mod_lm_ef[nonzero_ix,],
                        check.attributes = FALSE))
  expect_true(all(ef1$psi[zero_ix] == 0))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns with `by` and C is a subset of Q"), {
  set.seed(438)
  data(simdata)
  simdata_copy <- simdata

  simdata_copy$obs_id <- seq_len(nrow(simdata_copy))
  simdata_shuffle_ix <- sample(rownames(simdata_copy))
  shuffled_simdata <- simdata_copy[simdata_shuffle_ix,]
  C_ix <- seq_len(34L)
  C_data <- simdata_copy[C_ix,]
  C_shuffle_ix <- sample(rownames(C_data))
  shuffled_C_data <- C_data[C_shuffle_ix,]
  cmod1 <- lm(y ~ x, C_data)
  cmod2 <- lm(y ~ x, shuffled_C_data)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_simdata)
  mod1 <- lmitt(y ~ 1, data = simdata_copy, specification = spec1,
                offset = cov_adj(cmod1, by = "obs_id"))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec2,
                offset = cov_adj(cmod2, by = "obs_id"))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  Q_cls <- paste(simdata_copy$uoa1, simdata_copy$uoa2, sep = "_")
  C_cls <- Q_cls[C_ix]
  jk_units <- unique(C_cls)
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)
  loo_preds <- replace(mod1$offset,
                       C_ix,
                       rowSums(X * t(loo_cmod[, C_cls, drop=FALSE])))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata$y - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, loco_residuals = TRUE)

  expect_equal(dim(ef1$phi), c(nrow(simdata_copy), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata_copy), 2))
  zero_ix <- seq(nrow(simdata_copy)-length(C_ix))
  nonzero_ix <- setdiff(seq_len(nrow(simdata_copy)), zero_ix)
  expect_true(all(ef1$phi[zero_ix] == 0))
  expect_true(all.equal(
    ef1$phi[nonzero_ix,],
    cmod_ef[sort(as.character(C_ix)),],
    check.attributes = FALSE))
  expect_true(all.equal(
    ef1$psi,
    mod_lm_ef[c(setdiff(seq_len(nrow(simdata_copy)), C_ix),
                sort(as.character(C_ix))),],
    check.attributes = FALSE))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
  expect_equal(vcov_tee(mod1, cluster = "bid", loco_residuals = TRUE),
               vcov_tee(mod2, cluster = "bid", loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns with `by` and C and Q have no",
                "overlap (doesn't use jackknifed first-stage estimates)"), {
  set.seed(438)
  data(simdata)

  simdata$obs_id <- seq_len(nrow(simdata))
  Q_data <- simdata[21:50,]
  C_data <- simdata[1:20,]
  Q_shuffle_ix <- sample(rownames(Q_data))
  shuffled_Q_data <- Q_data[Q_shuffle_ix,]
  C_shuffle_ix <- sample(rownames(C_data))
  shuffled_C_data <- C_data[C_shuffle_ix,]
  cmod1 <- lm(y ~ x, C_data)
  cmod2 <- lm(y ~ x, shuffled_C_data)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_Q_data)
  mod1 <- lmitt(y ~ 1, data = Q_data, specification = spec1,
                offset = cov_adj(cmod1, by = "obs_id"))
  mod2 <- lmitt(y ~ 1, data = shuffled_Q_data, specification = spec2,
                offset = cov_adj(cmod2, by = "obs_id"))

  cmod_ef <- estfun(cmod1)
  mod_lm_ef <- estfun(as(mod1, "lm"))
  ef1 <- .align_and_extend_estfuns(mod1)
  ef2 <- .align_and_extend_estfuns(mod2)

  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  expect_true(all(ef1$phi[1L:30L,] == 0))
  expect_true(all.equal(ef1$phi[31L:50L,], cmod_ef[1L:20L,],
                        check.attributes = FALSE))
  expect_true(all(ef1$psi[31L:50L,] == 0))
  expect_true(all.equal(ef1$psi[1L:30L,], mod_lm_ef[1L:30L,],
                        check.attributes = FALSE))
  expect_equal(vcov_tee(mod1), vcov_tee(mod2))
  expect_equal(vcov_tee(mod1, cluster = "bid"), vcov_tee(mod2, cluster = "bid"))
})

test_that(paste(".align_and_extend_estfuns when the samples fully overlap",
                "(no `by`) (uses jackknifing)"), {
  set.seed(438)
  data(simdata)

  shuffled_simdata <- simdata[sample(rownames(simdata)),]
  cmod1 <- lm(y ~ x, simdata)
  cmod2 <- lm(y ~ x, shuffled_simdata)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec1,
                offset = cov_adj(cmod1))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec2,
                offset = cov_adj(cmod2))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  C_cls <- paste(simdata$uoa1, simdata$uoa2, sep = "_")
  jk_units <- unique(C_cls)
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)
  loo_preds <- rowSums(X * t(loo_cmod[, C_cls, drop=FALSE]))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata$y - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, loco_residuals = TRUE)
  by_ix <- sort(apply(simdata[, c("uoa1", "uoa2")], 1,
                      function(...) paste(..., collapse = "_")))

  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  expect_true(all.equal(ef1$phi, cmod_ef[sort(seq(50L)),],
                        check.attributes = FALSE))
  expect_true(all.equal(ef1$psi, mod_lm_ef[sort(seq(50L)),],
                        check.attributes = FALSE))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
  expect_equal(vcov_tee(mod1, cluster = "bid", loco_residuals = TRUE),
               vcov_tee(mod2, cluster = "bid", loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns when Q is a subset of C (no `by`)",
                "(uses jackknifing)"), {
  set.seed(438)
  data(simdata)

  shuffled_simdata <- simdata[sample(rownames(simdata)),]
  Q_ix <- seq_len(20L)
  Q_data <- simdata[Q_ix,]
  shuffled_Q_data <- Q_data[sample(rownames(Q_data)),]
  cmod1 <- lm(y ~ x, simdata)
  cmod2 <- lm(y ~ x, shuffled_simdata)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_Q_data)
  mod1 <- lmitt(y ~ 1, data = Q_data, specification = spec1,
                offset = cov_adj(cmod1))
  mod2 <- lmitt(y ~ 1, data = shuffled_Q_data, specification = spec2,
                offset = cov_adj(cmod2))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  C_cls <- paste(simdata$uoa1, simdata$uoa2, sep = "_")
  jk_units <- unique(C_cls[Q_ix])
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)[Q_ix,,drop=FALSE] 
  loo_preds <- rowSums(X * t(loo_cmod[, C_cls[Q_ix], drop=FALSE]))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata$y[Q_ix] - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, loco_residuals = TRUE)
  by_ix <- sort(apply(simdata[, c("uoa1", "uoa2")], 1,
                      function(...) paste(..., collapse = "_")))

  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  expect_true(all.equal(ef1$phi, cmod_ef, check.attributes = FALSE))
  expect_true(all.equal(ef1$psi[Q_ix,], mod_lm_ef[Q_ix,],
                        check.attributes = FALSE))
  expect_true(all(ef1$psi[setdiff(seq_len(nrow(simdata)), Q_ix)] == 0))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns when exact alignment of C and Q",
                "isn't possible and C is a subset of Q (uses jackknifing)"), {
  set.seed(438)
  data(simdata)

  shuffled_simdata <- simdata[sample(rownames(simdata)),]
  C_ix <- seq_len(34L)
  C_data <- simdata[C_ix,]
  shuffled_C_data <- C_data[sample(rownames(C_data)),]
  cmod1 <- lm(y ~ x, C_data)
  cmod2 <- lm(y ~ x, shuffled_C_data)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec1,
                offset = cov_adj(cmod1))
  mod2 <- lmitt(y ~ 1, data = shuffled_simdata, specification = spec2,
                offset = cov_adj(cmod2))

  # get actual values for matrices of estimating equations
  cmod_ef <- estfun(cmod1)
  Q_cls <- paste(simdata$uoa1, simdata$uoa2, sep = "_")
  C_cls <- Q_cls[C_ix]
  jk_units <- unique(C_cls)
  loo_cmod <- Reduce(
    cbind,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        eval(cmod_cl, envir = environment(formula(cmod)))$coefficients
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod1, cls = C_cls)
    )
  )
  colnames(loo_cmod) <- jk_units
  X <- model.matrix(cmod1)
  loo_preds <- replace(mod1$offset,
                       C_ix,
                       rowSums(X * t(loo_cmod[, C_cls, drop=FALSE])))
  mod_lm_ef <- estfun(as(mod1, "lm")) / stats::residuals(mod1) * (
    simdata$y - loo_preds - mod1$fitted.values + mod1$offset)
  ef1 <- .align_and_extend_estfuns(mod1, loco_residuals = TRUE)
  ef2 <- .align_and_extend_estfuns(mod2, loco_residuals = TRUE)
  by_ix <- sort(apply(simdata[, c("uoa1", "uoa2")], 1,
                      function(...) paste(..., collapse = "_")))

  zero_ix <- seq(nrow(simdata)-length(C_ix))
  nonzero_ix <- setdiff(seq_len(nrow(simdata)), zero_ix)
  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  expect_true(all.equal(ef1$phi[nonzero_ix,],
                        cmod_ef[C_ix,],
                        check.attributes = FALSE))
  expect_true(all(ef1$phi[zero_ix] == 0))
  expect_true(all.equal(ef1$psi,
                        mod_lm_ef[c(setdiff(seq_len(nrow(simdata)), C_ix),
                                    C_ix),],
                        check.attributes = FALSE))
  expect_equal(vcov_tee(mod1, loco_residuals = TRUE),
               vcov_tee(mod2, loco_residuals = TRUE))
  expect_equal(vcov_tee(mod1, cluster = "bid", loco_residuals = TRUE),
               vcov_tee(mod2, cluster = "bid", loco_residuals = TRUE))
})

test_that(paste(".align_and_extend_estfuns when the samples have no overlap",
                "(no `by`) (doesn't use jackknifing)"), {
  set.seed(438)
  data(simdata)

  Q_data <- simdata[21:50,]
  C_data <- simdata[1:20,]
  shuffled_Q_data <- Q_data[sample(rownames(Q_data)),]
  shuffled_C_data <- C_data[sample(rownames(C_data)),]
  cmod1 <- lm(y ~ x, C_data)
  cmod2 <- lm(y ~ x, shuffled_C_data)
  spec1 <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = shuffled_Q_data)
  mod1 <- lmitt(y ~ 1, data = Q_data, specification = spec1,
                offset = cov_adj(cmod1))
  mod2 <- lmitt(y ~ 1, data = shuffled_Q_data, specification = spec2,
                offset = cov_adj(cmod2))

  cmod_ef <- estfun(cmod1)
  mod_lm_ef <- estfun(as(mod1, "lm"))
  ef1 <- .align_and_extend_estfuns(mod1)
  ef2 <- .align_and_extend_estfuns(mod2)
  by_ix <- sort(apply(simdata[, c("uoa1", "uoa2")], 1,
                      function(...) paste(..., collapse = "_")))

  expect_equal(dim(ef1$phi), c(nrow(simdata), 2))
  expect_equal(dim(ef1$psi), c(nrow(simdata), 2))
  expect_true(all.equal(ef1$phi[31L:50L,], cmod_ef[1L:20L,],
                        check.attributes = FALSE))
  expect_true(all(ef1$phi[1L:30L,] == 0))
  expect_true(all.equal(ef1$psi[1L:30L,], mod_lm_ef[1L:30L,],
                        check.attributes = FALSE))
  expect_true(all(ef1$psi[31L:50L,] == 0))
  expect_equal(vcov_tee(mod1), vcov_tee(mod2))
  expect_equal(vcov_tee(mod1, cluster = "bid"), vcov_tee(mod2, cluster = "bid"))
})

test_that(".align_and_extend_estfuns with ctrl means estfun", {
  moddata <- data.frame(a = c(rep(c(0, 1), each = 5), NA_real_), x = rnorm(11),
                        y = rnorm(11), id = seq_len(11))
  cmod <- lm(y ~ x, moddata)
  spec <- rct_spec(a ~ unitid(id), data = moddata)
  mod <- lmitt(y ~ 1, data = moddata, spec = spec, offset = cov_adj(cmod))
  class(mod$na.action) <- "exclude"
  cm_ef <- estfun(mod@ctrl_means_model)
  cm_ef[is.na(cm_ef)] <- 0
  aligned1 <- .align_and_extend_estfuns(mod)
  aligned2 <- .align_and_extend_estfuns(mod, ctrl_means_ef_mat = cm_ef)
  expect_equal(length(aligned1), 2)
  expect_equal(length(aligned2), 2)
  expect_true(all.equal(cm_ef[c(1, 10:11, 2:9),], aligned2$psi[, 3:4],
                        check.attributes = FALSE))
})

test_that(".align_and_extend_estfuns converts NA's to 0's", {
  set.seed(249)
  dat <- data.frame(y = c(rep(NA_real_, 3), rnorm(27)),
                    a = rep(c(0, 1), each = 15),
                    x = rnorm(30))
  suppressWarnings(spec <- rct_spec(a ~ 1, dat))
  cmod <- lm(y ~ x, dat)
  xm <- lmitt(y ~ 1, spec, dat, offset = cov_adj(cmod))
  
  class(xm$na.action) <- "exclude"
  aligned <- .align_and_extend_estfuns(xm)
  expect_equal(dim(aligned$psi), c(30, 2))
  expect_equal(dim(aligned$phi), c(30, 2))
  expect_equal(aligned$psi[1:3,],
               matrix(0, nrow = 3, ncol = 2,
                      dimnames = list(seq_len(3), c("(Intercept)", "a."))))
  expect_equal(aligned$phi[1:3,],
               matrix(0, nrow = 3, ncol = 2,
                      dimnames = list(seq_len(3), c("(Intercept)", "x"))))
})

test_that(".make_uoa_ids fails without cluster argument or teeMod model", {
  data(simdata)
  mod <- lm(y ~ z, data = simdata)
  expect_error(.make_uoa_ids(mod), "Must be provided")
})

test_that(".make_uoa_ids returns correct ID's for non-SandwichLayer offset", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec)

  expected_out <- factor(
    apply(simdata[, c("uoa1", "uoa2"), drop = FALSE],
          1,
          function(...) paste(..., collapse = "_"))
  )
  expect_equal(.make_uoa_ids(mod, vcov_type = "CR"), expected_out)
})

test_that(".make_uoa_ids returns correct ID's for full overlap of C and Q", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  dmod <- lmitt(y ~ 1, data = simdata, specification = spec,
                offset = cov_adj(cmod))

  expected_out <- factor(
    apply(simdata[, c("uoa1", "uoa2"), drop = FALSE],
          1,
          function(...) paste(..., collapse = "_"))
  )
  out <- .make_uoa_ids(dmod, vcov_type = "CR")

  expect_true(all.equal(out, expected_out, check.attributes = FALSE))
})

test_that(".make_uoa_ids for no overlap of C and Q", {
  data(simdata)
  set.seed(300)
  cmod_ssta1 <- data.frame("y" = rnorm(10), "x" = rnorm(10),
                           "uoa1" = NA, "uoa2" = NA)
  cmod_ssta2 <- data.frame("y" = rnorm(10), "x" = rnorm(10),
                           "uoa1" = rep(c(1, 2), each = 5), "uoa2" = NA)

  cmod1 <- lm(y ~ x, cmod_ssta1)
  cmod2 <- lm(y ~ x, cmod_ssta2)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  dmod1 <- lmitt(y ~ 1, data = simdata, specification = spec,
                 offset = cov_adj(cmod1))
  dmod2 <- lmitt(y ~ 1, data = simdata, specification = spec,
                 offset = cov_adj(cmod2))

  Q_uoas <- apply(simdata[, c("uoa1", "uoa2"), drop = FALSE], 1,
                  function(...) paste(..., collapse = "_"))

  ids1 <- .make_uoa_ids(dmod1, vcov_type = "CR")
  ids2 <- .make_uoa_ids(dmod2, vcov_type = "CR")

  expect_true(is.factor(ids1))
  expect_true(is.factor(ids2))

  expect_equal(length(ids1), nrow(simdata) + 10)
  expect_equal(length(ids2), nrow(simdata) + 10)

  expect_true(all.equal(ids1[1:nrow(simdata)], factor(Q_uoas),
                        check.attributes = FALSE))
  expect_true(all.equal(ids2[1:nrow(simdata)], factor(Q_uoas),
                        check.attributes = FALSE))

  expect_equal(length(unique(ids1)), length(unique(Q_uoas)) + 10)
  expect_equal(length(unique(ids2)), length(unique(Q_uoas)) + 2)
})

test_that(".make_uoa_ids returns correct ID's for partial overlap of C and Q", {
  data(simdata)
  set.seed(300)
  C_not_Q <- data.frame("y" = rnorm(20), "x" = rnorm(20),
                        "uoa1" = NA, "uoa2" = NA)
  cmod_ssta <- rbind(simdata[, colnames(C_not_Q)], C_not_Q)

  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  dmod <- lmitt(y ~ 1, data = simdata, specification = spec,
                offset = cov_adj(cmod))

  Q_uoas <- apply(simdata[, c("uoa1", "uoa2"), drop = FALSE], 1,
                  function(...) paste(..., collapse = "_"))

  ids <- .make_uoa_ids(dmod, vcov_type = "CR")

  expect_true(is.factor(ids))

  expect_equal(length(ids), nrow(simdata) + 20)

  expect_true(all.equal(ids[1:nrow(simdata)], factor(Q_uoas),
                        check.attributes = FALSE))

  expect_equal(length(unique(ids)), length(unique(Q_uoas)) + 20)
})

test_that(paste(".make_uoa_ids returns correct ID's when cov_adj's 'by'",
                "argument provided a different ordering"), {
  set.seed(300)
  data(simdata)
  simdata_copy <- simdata
  simdata_copy$id <- sample(seq_len(nrow(simdata_copy)))

  C_not_Q <- data.frame("y" = rnorm(20), "x" = rnorm(20),
                        "uoa1" = NA, "uoa2" = NA, "id" = seq(51, 70))
  cmod_ssta <- rbind(simdata_copy[, colnames(C_not_Q)], C_not_Q)

  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata_copy)
  dmod <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
                offset = cov_adj(cmod, by = "id"))

  Q_uoas <- apply(simdata_copy[, c("uoa1", "uoa2"), drop = FALSE], 1,
                  function(...) paste(..., collapse = "_"))

  ids <- .make_uoa_ids(dmod, vcov_type = "CR")

  expect_true(is.factor(ids))
  expect_equal(length(ids), nrow(simdata_copy) + 20)
  expect_true(all.equal(
    ids[1L:nrow(simdata_copy)],
    factor(Q_uoas)[match(sort(as.character(simdata_copy$id)), simdata_copy$id)],
    check.attributes = FALSE))
  expect_equal(length(unique(ids)), length(unique(Q_uoas)) + 20)
})

test_that(".make_uoa_ids with `by` when Q ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(id = letters[1:15],
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  by_overlap <- sort(as.character(setdiff(seq_len(15), seq(1, 15, 3))))
  nonoverlap <- as.character(seq(1, 15, 3))
  overlap_Q_ix <- match(by_overlap, analysis_dat$by_col)
  overlap_C_ix <- match(by_overlap, cmod_ssta$by_col)

  out <- c(analysis_dat$id[overlap_Q_ix],
           cmod_ssta$id[match(nonoverlap, cmod_ssta$by_col)])
  expect_equal(length(ids <- .make_uoa_ids(mod,
                                           vcov_type = "HC0",
                                           cluster = "id",
                                           by = "by_col")),
               15)
  expect_equal(ids, factor(out, levels = unique(out)))
  expect_equal(ids, .make_uoa_ids(mod_w_by, vcov_type = "HC0", cluster = "id",
                                  by = "by_col"))
})

test_that(".make_uoa_ids with `by` when C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = letters[1:5],
                             x = rnorm(5),
                             a = c(rep(1, 3), rep(0, 2)),
                             y = rnorm(5),
                             by_col = seq(3, 15, 3))
  mod <- lmitt(y ~ 1, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ 1, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  by_overlap <- sort(as.character(seq(3, 15, 3)))
  nonoverlap <- as.character(setdiff(seq_len(15), seq(3, 15, 3)))
  overlap_Q_ix <- match(by_overlap, analysis_dat$by_col)
  overlap_C_ix <- match(by_overlap, cmod_ssta$by_col)

  out <- c(analysis_dat$id[overlap_Q_ix],
           cmod_ssta$id[match(nonoverlap, cmod_ssta$by_col)])
  expect_equal(length(ids <- .make_uoa_ids(mod,
                                           vcov_type = "HC0",
                                           cluster = "id",
                                           by = "by_col")),
               15)
  expect_equal(ids, factor(out, levels = unique(out)))
  expect_equal(ids, .make_uoa_ids(mod_w_by, vcov_type = "HC0", cluster = "id",
                                  by = "by_col"))
})

test_that(".make_uoa_ids with `by` when Q and C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat, 
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  by_overlap <- sort(as.character(setdiff(seq_len(15), seq(1, 15, 3))))
  nonoverlap <- as.character(seq(1, 15, 3))
  overlap_Q_ix <- match(by_overlap, analysis_dat$by_col)
  overlap_C_ix <- match(by_overlap, cmod_ssta$by_col)

  out <- c(analysis_dat$id[overlap_Q_ix],
           cmod_ssta$id[match(nonoverlap, cmod_ssta$by_col)])
  expect_equal(length(ids <- .make_uoa_ids(mod,
                                           vcov_type = "HC0",
                                           cluster = "id",
                                           by = "by_col")),
               15)
  expect_equal(ids, factor(out, levels = unique(out)))
  expect_equal(ids, .make_uoa_ids(mod_w_by, vcov_type = "HC0", cluster = "id",
                                  by = "by_col"))
})

test_that(".make_uoa_ids without `by` when Q and C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  expect_error(.make_uoa_ids(mod, vcov_type = "HC0", cluster = "id"),
               "not uniquely specified. Provide a `by` argument")
})

test_that(paste("unit-level clustering model-based .make_uoa_ids does not use",
                "block ID's for small blocks"), {
  data(simdata)

  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  suppressMessages(dmod <- lmitt(y ~ 1, specification = spec, data = simdata))

  uoa_ids <- factor(apply(simdata[, c("uoa1", "uoa2")],
                          1,
                          function(...) paste(..., collapse = "_")))
  expect_equal(
    out <- .make_uoa_ids(dmod, "CR"),
    uoa_ids
  )
  expect_equal(out, .make_uoa_ids(dmod, "MB"))
  expect_equal(out, .make_uoa_ids(dmod, "HC"))
  expect_equal(
    .make_uoa_ids(dmod, "DB"),
    out
  )
})

test_that(paste(".order_samples fails without a teeMod object or",
                "SandwichLayer offset"), {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  mod1 <- lm(y ~ z, simdata, offset = predict(cmod))
  mod2 <- lmitt(y ~ 1, data = simdata, specification = spec)

  expect_error(.order_samples(mod1), "must be a teeMod object")
  expect_error(.order_samples(mod2), "must be a teeMod object")
})

test_that(".order_samples when the samples fully overlap", {
  set.seed(300)
  data(simdata)
  simdata_copy <- simdata

  simdata_copy$uid <- seq_len(nrow(simdata_copy))
  cmod <- lm(y ~ x, simdata_copy)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  mod <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
               offset = cov_adj(cmod, by = "uid"))

  out <- .order_samples(mod)

  expect_equal(length(out$Q_not_C), 0)
  expect_equal(length(out$Q_in_C), nrow(simdata_copy))
  expect_equal(length(out$C_in_Q), nrow(simdata_copy))
  expect_equal(length(out$C_not_Q), 0)

  expect_equal(names(out$Q_in_C), ord <- sort(
    as.character(seq(nrow(simdata_copy)))))
  expect_equal(names(out$C_in_Q), ord)

  expect_true(all.equal(out$Q_in_C, ord, check.attributes = FALSE))
  expect_true(all.equal(out$C_in_Q, ord, check.attributes = FALSE))
})

test_that(".order_samples when Q is a subset of C", {
  set.seed(300)
  data(simdata)
  simdata_copy <- simdata

  simdata_copy$uid <- seq_len(nrow(simdata_copy))
  cmod_ssta <- rbind(
    simdata_copy[, c("x", "y", "uoa1", "uoa2", "uid")],
    data.frame(x = rnorm(30), y = rnorm(30), uoa1 = NA, uoa2 = NA,
               uid = seq_len(30) + nrow(simdata_copy))
  )
  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  mod <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
               offset = cov_adj(cmod, by = "uid"))

  out <- .order_samples(mod)

  expect_equal(length(out$Q_not_C), 0)
  expect_equal(length(out$Q_in_C), nrow(simdata_copy))
  expect_equal(length(out$C_in_Q), nrow(simdata_copy))
  expect_equal(length(out$C_not_Q), 30)

  expect_equal(names(out$Q_in_C), overlap_ord <- sort(
    as.character(seq(nrow(simdata_copy)))))
  expect_equal(names(out$C_in_Q), overlap_ord)
  expect_equal(names(out$C_not_Q), as.character(seq(51, 80)))

  expect_true(all.equal(out$Q_in_C, overlap_ord, check.attributes = FALSE))
  expect_true(all.equal(out$C_in_Q, overlap_ord, check.attributes = FALSE))
  expect_true(all.equal(out$C_not_Q, as.character(seq(51, 80)),
                        check.attributes = FALSE))
})


test_that(".order_samples when C is a subset of Q", {
  set.seed(300)
  data(simdata)
  simdata_copy <- simdata

  simdata_copy$uid <- seq_len(nrow(simdata_copy))
  cmod_ssta <- simdata_copy[1:20,]
  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata_copy)
  mod <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
               offset = cov_adj(cmod, by = "uid"))

  out <- .order_samples(mod)

  expect_equal(length(out$Q_not_C), 30)
  expect_equal(length(out$Q_in_C), 20)
  expect_equal(length(out$C_in_Q), 20)
  expect_equal(length(out$C_not_Q), 0)

  expect_equal(names(out$Q_not_C), as.character(seq(21, 50)))
  expect_equal(names(out$Q_in_C), overlap_ord <- sort(as.character(seq(20))))
  expect_equal(names(out$C_in_Q), overlap_ord)

  expect_true(all.equal(out$Q_not_C, as.character(seq(21, nrow(simdata_copy))),
                        check.attributes = FALSE))
  expect_true(all.equal(out$Q_in_C, overlap_ord, check.attributes = FALSE))
  expect_true(all.equal(out$C_in_Q, overlap_ord, check.attributes = FALSE))
})

test_that(".order_samples when the samples do not overlap", {
  set.seed(300)
  data(simdata)

  simdata$uid <- seq_len(nrow(simdata))
  cmod_ssta <- simdata[1:20,]
  Q_data <- simdata[21:50,]
  cmod <- lm(y ~ x, cmod_ssta)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  mod <- lmitt(y ~ 1, data = Q_data, specification = spec,
               offset = cov_adj(cmod, by = "uid"))

  out <- .order_samples(mod)

  expect_equal(length(out$Q_not_C), 30)
  expect_equal(length(out$Q_in_C), 0)
  expect_equal(length(out$C_in_Q), 0)
  expect_equal(length(out$C_not_Q), 20)

  expect_equal(names(out$Q_not_C), as.character(seq(30)))
  expect_equal(names(out$C_not_Q), as.character(seq(20)))

  expect_true(all.equal(out$Q_not_C, as.character(seq(21, 50)),
                        check.attributes = FALSE))
  expect_true(all.equal(out$C_not_Q, as.character(seq(20)),
                        check.attributes = FALSE))
})

test_that(".order_samples when no `by` argument provided", {
  set.seed(300)
  data(simdata)

  simdata$uid <- seq_len(nrow(simdata))
  C_data <- simdata[1:20,]
  Q_data <- simdata[21:50,]
  cmod <- lm(y ~ x, C_data)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_data)
  mod <- lmitt(y ~ 1, data = Q_data, specification = spec,
               offset = cov_adj(cmod))

  out <- .order_samples(mod)

  Q_uoas <- apply(Q_data[, c("uoa1", "uoa2")],
                  1,
                  function(...) paste(..., collapse = "_"))
  C_uoas <- apply(C_data[, c("uoa1", "uoa2")],
                  1,
                  function(...) paste(..., collapse = "_"))

  expect_equal(length(out$Q_not_C), 30)
  expect_equal(length(out$Q_in_C), 0)
  expect_equal(length(out$C_in_Q), 0)
  expect_equal(length(out$C_not_Q), 20)

  expect_true(all.equal(names(out$Q_not_C), as.character(seq(30)),
                        check.attributes = FALSE))
  expect_true(all.equal(names(out$C_not_Q), as.character(seq(20)),
                        check.attributes = FALSE))

  expect_true(all.equal(out$Q_not_C, Q_uoas, check.attributes = FALSE))
  expect_true(all.equal(out$C_not_Q, C_uoas, check.attributes = FALSE))
})

test_that(".order_samples with `by` when Q ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(id = letters[1:15],
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  overlap <- sort(as.character(setdiff(seq_len(15), seq(1, 15, 3))))
  nonoverlap <- as.character(seq(1, 15, 3))
  out <- list("Q_not_C" = setNames(character(0L), character(0L)),
              "Q_in_C" = setNames(overlap, match(overlap, analysis_dat$by_col)),
              "C_in_Q" = setNames(overlap, match(overlap, cmod_ssta$by_col)),
              "C_not_Q" = setNames(nonoverlap,
                                   which(cmod_ssta$by_col %in% nonoverlap)))
  expect_equal(ord <- .order_samples(mod, by = "by_col"), out)
  expect_equal(ord, .order_samples(mod_w_by))
})

test_that(".order_samples with `by` when C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = letters[1:5],
                             x = rnorm(5),
                             a = c(rep(1, 3), rep(0, 2)),
                             y = rnorm(5),
                             by_col = seq(3, 15, 3))
  mod <- lmitt(y ~ 1, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ 1, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  overlap <- sort(as.character(seq(3, 15, 3)))
  nonoverlap <- as.character(setdiff(seq_len(15), seq(3, 15, 3)))
  out <- list("Q_not_C" = setNames(character(0L), character(0L)),
              "Q_in_C" = setNames(overlap, match(overlap, analysis_dat$by_col)),
              "C_in_Q" = setNames(overlap, match(overlap, cmod_ssta$by_col)),
              "C_not_Q" = setNames(nonoverlap,
                                   which(cmod_ssta$by_col %in% nonoverlap)))
  expect_equal(ord <- .order_samples(mod, by = "by_col"), out)
  expect_equal(.order_samples(mod_w_by), ord)
})

test_that(".order_samples with `by` when both Q and C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  mod_w_by <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
                    offset = cov_adj(cmod, by = "by_col"))

  overlap <- sort(as.character(setdiff(seq_len(15), seq(1, 15, 3))))
  nonoverlap <- as.character(seq(1, 15, 3))
  out <- list("Q_not_C" = setNames(character(0L), character(0L)),
              "Q_in_C" = setNames(overlap, match(overlap, analysis_dat$by_col)),
              "C_in_Q" = setNames(overlap, match(overlap, cmod_ssta$by_col)),
              "C_not_Q" = setNames(nonoverlap,
                                   which(cmod_ssta$by_col %in% nonoverlap)))
  expect_equal(ord <- .order_samples(mod, by = "by_col"), out)
  expect_equal(.order_samples(mod_w_by), ord)
})

test_that(".order_samples without `by` when Q or C ID's aren't unique", {
  set.seed(23)
  cmod_ssta <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_ssta)

  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)

  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod))
  expect_error(.order_samples(mod),
               "not uniquely specified. Provide a `by` argument")
})

test_that("sanitize_Q_ids succeeds with valid `by` argument", {
  data(simdata)

  simdata_copy <- simdata
  simdata_copy$uid <- seq_len(nrow(simdata_copy))
  cmod <- lm(y ~ z, data = simdata_copy)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2, uid), simdata_copy)
  dmod <- lmitt(y ~ 1, specification = spec, data = simdata_copy,
                offset = cov_adj(cmod))
  out <- .sanitize_Q_ids(dmod, c("uoa1", "uoa2"))$cluster

  expected_out <- apply(simdata_copy[, c("uoa1", "uoa2"), drop = FALSE], 1,
                        function(...) paste(..., collapse = "_"))
  expect_equal(out, expected_out)
})

test_that(".base_S3class_estfun fails with invalid base S3 class", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec)
  mod@.S3Class <- "invalid_class"

  expect_error(.base_S3class_estfun(mod), "must have been fitted")
})

test_that("checking proper errors in conversion from lm to teeMod", {
  data(simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  expect_error(as.lmitt(lm(y ~ x, data = simdata), specification = spec),
               "aliases must be")


  # Take in either StudySpecification or WeightedStudySpecification
  mod1 <- propertee:::.convert_to_lmitt(lm(y ~ assigned(spec), data = simdata),
                                     spec,
                                     FALSE,
                                     TRUE,
                                     "x",
                                     lhs ~ 1,
                                     NULL,
                                     call("quote", call("ls")))
  mod2 <- propertee:::.convert_to_lmitt(lm(y ~ assigned(spec), data = simdata),
                                     ate(spec, data = simdata),
                                     FALSE,
                                     TRUE,
                                     "x",
                                     lhs ~ 1,
                                     NULL,
                                     call("quote", call("ls")))
  expect_identical(mod1@StudySpecification, mod2@StudySpecification)

  expect_error(propertee:::.convert_to_lmitt(lm(y ~ assigned(spec),
                                                data = simdata),
                                     1,
                                     FALSE,
                                     TRUE,
                                     "x",
                                     lhs ~ 1,
                                     NULL,
                                     call("quote", call("ls"))), "must be a")


})


test_that("disable update.teeMod", {
  data(simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  mod <- lmitt(y ~ 1, data = simdata, specification = spec)

  expect_error(update(mod),
               "teeMod objects do not support")
})

test_that("lmitt_call", {
  data(simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  # Make sure slot is actual call
  mod <- lmitt(y ~ 1, data = simdata, specification = spec)
  expect_true(is.call(mod@lmitt_call))

  # Call via `lmitt()` should match
  lmittcall <- str2lang("lmitt(y ~ 1, data = simdata, specification = spec)")
  mod <- eval(lmittcall)
  expect_equal(lmittcall, mod@lmitt_call)

  # Call via `lmitt.formula()` should match
  lmittformcall <- str2lang(paste("lmitt.formula(y ~ 1, data = simdata,",
                                  "specification = spec)"))
  mod <- eval(lmittformcall)
  expect_equal(lmittformcall, mod@lmitt_call)

  # as.lmitt, make sure is actual call
  lmmod <- lm(y ~ a.(spec), data = simdata)
  mod <- as.lmitt(lmmod, specification = spec)
  expect_true(is.call(mod@lmitt_call))

  # Call via `as.lmitt()` should match
  aslmittcall <- str2lang("as.lmitt(lmmod, specification = spec)")
  mod <- eval(aslmittcall)
  expect_equal(aslmittcall, mod@lmitt_call)

  # Call via `lmitt(lm` should match
  lmittlmcall <- str2lang("lmitt(lmmod, specification = spec)")
  mod <- eval(lmittlmcall)
  expect_equal(lmittlmcall, mod@lmitt_call)

  # Call via `lmitt.lm` should match
  lmittlmcall <- str2lang("lmitt.lm(lmmod, specification = spec)")
  mod <- eval(lmittlmcall)
  expect_equal(lmittlmcall, mod@lmitt_call)

})

test_that("lmitt_fitted object", {
  data(simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  mod <- lmitt(y ~ as.factor(o), data = simdata, specification = spec)
  expect_true(mod@lmitt_fitted)

  mod <- as.lmitt(lm(y ~ adopters(spec), data = simdata), specification = spec)
  expect_false(mod@lmitt_fitted)

  mod <- lmitt(lm(y ~ adopters(spec), data = simdata), specification = spec)
  expect_false(mod@lmitt_fitted)

})

test_that("printed effects aren't confused by bad naming", {
  data(simdata)
  simdata$abz.c <- as.factor(simdata$o)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  mod <- lmitt(y ~ abz.c, data = simdata, specification = spec)
  co <- capture.output(show(mod))
  cos <- strsplit(trimws(co), "[[:space:]]+")

  expect_true(
    all(sapply(
      levels(simdata$abz.c),
      function(lvl) {
        any(sapply(cos,
                   function(buf) any(grepl(paste0("z._abz.c", lvl), buf)))) &
          any(sapply(cos,
                     function(buf) any(grepl(paste0("y:abz.c", lvl), buf))))
      })))

  # to force ` in variable names via as.factor
  mod <- lmitt(y ~ as.factor(abz.c), data = simdata, specification = spec)
  co <- capture.output(show(mod))
  cos <- strsplit(trimws(co[c(1, 3)]), "[[:space:]]+")
  cos <- Reduce("c", cos)
  expect_length(cos, 6)
  expect_true(all(!isTRUE(grepl("^as\\.factor", cos))))
  expect_true(all(grepl("^(`z|y)", cos)))
})


test_that("as.lmitt call inside mapply", {
  min_df <- data.frame(schoolid = seq_len(20),
                       response_col = rnorm(20),
                       grade = rep(3:4, each = 10),
                       adsy = rep(rep(c(0, 1), each = 5), 2))

  min_spec <- rct_spec(adsy ~ unitid(schoolid), min_df)

  res <- mapply(function(fmla) {
    as.lmitt(lm(fmla, min_df, weights = ett(min_spec)))
  },
  list(response_col ~ z.(), response_col ~ grade + grade:z.()),
  SIMPLIFY = FALSE)


  expect_true(all(sapply(res, is, "teeMod")))
  expect_true(all(sapply(sapply(sapply(res,
                                       slot, "lmitt_call"),
                                "[[", 1),
                         deparse) == "as.lmitt"))

  res <- lapply(list(response_col ~ z.(), response_col ~ grade + grade:z.()),
                function(fmla) {
                  as.lmitt(lm(fmla, min_df, weights = ett(min_spec)))
                })

  expect_true(all(sapply(res, is, "teeMod")))
  expect_true(all(sapply(sapply(sapply(res,
                                       slot, "lmitt_call"),
                                "[[", 1),
                         deparse) == "as.lmitt"))

})

test_that("#81 continuous moderator shows appropriate coefficients", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) , data = simdata)

  mod <- lmitt(y ~ x, data = simdata, specification = spec)
  expect_length(mod$coeff, 6)
  coefnames <- strsplit(capture.output(show(mod))[1], " +")[[1]]
  coefnames <- coefnames[nchar(coefnames) > 0]
  expect_true(all(grepl("(z\\.|y:)", coefnames)))
})

test_that("Invalid input to .convert_to_lmitt", {
  data(simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  expect_error(.convert_to_lmitt(1, spec, FALSE, TRUE, "a",
                                 call("quote", call("ls"))))
})

test_that(".estfun_DB_blockabsorb returns 0 if not asking for
          specification-based SE or tee model does not absorb intercepts", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod <- lmitt(y ~ 1, specification = spec, data = simdata,
                 weights = ate(spec))
  ssmod_abs <- lmitt(y ~ 1, specification = spec, data = simdata,
                     weights = ate(spec),
                     absorb = TRUE)

  expect_true(all(.estfun_DB_blockabsorb(ssmod) == 0))
  expect_true(all(.estfun_DB_blockabsorb(ssmod, db = TRUE) == 0))
  expect_true(all(.estfun_DB_blockabsorb(ssmod_abs) == 0))
})

test_that(".estfun_DB_blockabsorb returns correct value", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod_abs <- lmitt(y ~ 1, specification = spec, data = simdata,
                     weights = ate(spec),
                     absorb = TRUE)

  expect_false(all(.estfun_DB_blockabsorb(ssmod_abs, vcov_type = "DB") == 0))

  phi <- .get_phi_tilde(ssmod_abs, vcov_type = "DB")
  aa <- .get_appinv_atp(ssmod_abs, vcov_type = "DB")
  expect_true(all.equal(
    cbind(0, phi %*% aa),
    .estfun_DB_blockabsorb(ssmod_abs, vcov_type = "DB")
  ))
})

##' @title Obtain Treatment from StudySpecification
##'
##' @description When passing a \code{lm} object to [lmitt()], extract and use
##'   the treatment variable specified in the \code{StudySpecification}.
##'
##' @details When passing a \code{lm} object to [lmitt()], the treatment
##'   variable in the \code{formula} passed to [lm()] needs to be identifiable.
##'   Rather than placing the treatment variable directly in the \code{formula},
##'   use one of these functions, to allow [lmitt()] to identify the treatment
##'   variable.
##'
##'   To keep the formula in the [lm()] call concise, instead of passing
##'   \code{specification} and \code{data} arguments to these functions, one can
##'   pass a \code{WeightedStudySpecification} object to the \code{weights}
##'   argument of the [lm()] call or a \code{SandwichLayer} object to the
##'   \code{offset} argument.
##'
##'   Alternatively, you can pass the \code{specification} and \code{data}
##'   arguments.
##'
##'   While \code{assigned()} can be used in any situation, it is most useful
##'   for scenarios where the treatment variable is non-binary and the
##'   \code{StudySpecification} contains a \code{Dichotomy}. For example, say
##'   \code{q} is a 3-level ordinal treatment variable, and the binary
##'   comparison of interest is captured in \code{dichotomy = q == 3 ~ q < 3}.
##'   If you were to fit a model including \code{q} as a predictor, e.g.
##'   \code{lm(y ~ q, ...)}, \code{lm} would treat \code{q} as the full ordinal
##'   variable. On the other hand, by calling \code{lm(y ~ assigned(), weights =
##'   ate(spec), ...)}, \code{assigned()} will generate the appropriate binary
##'   variable to allow estimation of treatment effects.
##'
##'   If called outside of a model call and without a \code{data} argument, this
##'   will extract the treatment from the \code{specification}. If this is the
##'   goal, the [treatment()] function is better suited for this purpose.
##'
##' @param specification Optional \code{StudySpecification}. If the
##'   \code{StudySpecification} can't be identified in the model (usually
##'   because neither weights (\code{ate()} or \code{ett()}) nor a covariate
##'   adjustment model (\code{cov_adj()}) are found), the
##'   \code{StudySpecification} can be passed diretly.
##' @param data Optional data set. By default [assigned()] will attempt to
##'   identify the appropriate data, if this fails (or you want to overwrite
##'   it), you can pass the data here.
##' @inheritParams ett
##' @return The treatment variable to be placed in the regression formula.
##' @rdname AssignedAliases
##' @export
##' @examples
##' data(simdata)
##' spec <- obs_spec(z ~ uoa(uoa1, uoa2), data = simdata)
##' mod <- lm(y ~ assigned(), data = simdata, weights = ate(spec))
##' lmittmod <- lmitt(mod)
##' summary(lmittmod, vcov.type = "CR0")
assigned <- function(specification = NULL, data = NULL, dichotomy = NULL) {
  if (is.null(specification)) {
    specification <- .get_spec()
  }

  if (is.null(data)) {
    # Get uoa info to enable merge in .expand_txt, and get non-treatment variables
    # in the dichotomy for .apply_dichotomy
    divars <- setdiff(all.vars(dichotomy), c(var_names(specification, "t"), "."))
    form <- as.formula(paste("~", paste(c(divars, var_names(specification, "u")),
                                        collapse = "+")))
    suppressWarnings(data <- try(.get_data_from_model("assigned", form),
                                 silent = TRUE))
    if (is(data, "try-error")) {
      warning(paste("`data` cannot be found. Extracting treatment",
                    "from `specification` instead."))
      return(treatment(specification)[, 1])
    }
  }

  if (specification@unit_of_assignment_type == "none") {
    data[["..uoa.."]] <- rownames(data)
  }


  tt <- treatment(specification, binary = FALSE)
  tt <- .expand_txt(tt, data, specification)
  if (is.null(dichotomy)) {
    possible_dichotomies <- .find_dichotomies()
    dichotomy <- .validate_dichotomy(possible_dichotomies)
  } else {
    dichotomy <- .validate_dichotomy(dichotomy)
  }


  if (!is.null(dichotomy)) {
    return(.apply_dichotomy(cbind(tt, data), dichotomy))
  } else {
    return(tt[,1])
  }
}

##' @rdname AssignedAliases
##' @export
adopters <- assigned

##' @rdname AssignedAliases
##' @export
a. <- assigned

##' @rdname AssignedAliases
##' @export
z. <- assigned

#' @include SandwichLayer.R
NULL

##' @title Prepare prognostic regression offset for \code{lmitt()} or \code{lm()}
##' @description \code{cov_adj()} takes a fitted model object and predicts
##'   outcomes under the counterfactual of being in the control condition.
##'   Using these predictions as offsets in a linear model with just the
##'   treatment variable as a regressor estimates the treatment effect with
##'   covariance adjustments as provided by the previously fitted model.
##'   In the special case that this covariance model was fit using only
##'   (quasi-)experimental control group observations, the method is known
##'   as Peters-Belson regression. \code{cov_adj()} assists the process by
##'   reporting its predictions on the scale of the response (in contrast to
##'   \code{predict.glm()}) and by retaining certain artifacts from
##'   covariance model fitting.  These artifacts are used by [vcov_tee()],
##'   and by [stats::vcov()] as applied to a \code{teeMod} object, so
##'   that the covariance model's contribution to sampling variability
##'   is reflected in treatment coefficient standard errors.
##'   \code{cov_adj()} can be used with \code{lm()}, but is better supported
##'   in combination with propertee's \code{lm}-wrapper, [lmitt()].
##' @details If \code{specification} is provided or can be found in the call
##'   stack, then prior to generating adjustments, \code{cov_adj()} will
##'   identify the treatment variable and replace all values with a reference
##'   level. If the treatment has logical type, this reference level is
##'   \code{FALSE}, and if it has numeric type, this is the smallest
##'   non-negative value (which means 0 for 0/1 binary). (Factor treatments are
##'   not currently supported.)\cr\cr More
##'   broadly, the \code{set_to_reference} argument can be used to set columns
##'   to values specified by the user prior to generating predictions. The
##'   argument takes a named list where the names specify the columns whose
##'   values should be replaced, and the entries specify the replacement values.
##'   If an entry is given as \code{default()}, internal routines will determine
##'   the reference level (the first level of a factor/factorized character
##'   variable or the minimum of a numeric/logical variable). \cr\cr If
##'   \code{newdata} is provided, the values of the output vector represent
##'   adjustments for the outcomes in \code{newdata}. If \code{cov_adj()} is
##'   called in the \code{offset} argument of \code{lmitt()}, predictions will
##'   correspond to rows in the \code{data} argument of \code{lmitt()}.
##'   Otherwise, adjustments are made based on the data to which \code{model} is
##'   fit. \cr\cr The \code{by} argument specifies columns for merging datasets.
##'   The names of the argument specify columns in the dataframe for building
##'   \code{specification}, while the entries specify columns in the dataframe
##'   for fitting \code{model}.
##' @param model any model that inherits from a \code{glm}, \code{lm}, or
##'   \code{robustbase::lmrob} object.
##' @param newdata a dataframe; optional. See Details for how \code{newdata}
##' is found in the call stack if none is provided.
##' @param specification a \code{StudySpecification} object; optional. If not
##' provided, it can be retrieved from the call stack if passed to suitable
##' calls.
##' @param by named vector; optional.
##' @param set_to_reference named list; optional.
##' @return A \code{SandwichLayer} if \code{specification} is not NULL or a
##'   \code{StudySpecification} object is found in the call stack, otherwise a
##'   \code{PreSandwichLayer} object
##' @export
##' @example inst/examples/cov_adj.R
cov_adj <- function(model,
                    newdata = NULL,
                    specification =  NULL,
                    by = NULL,
                    set_to_reference = NULL) {
  if (is.null(specification)) {
    specification <- .get_spec(NULL_on_error = TRUE)
  }

  if (is.null(newdata)) {
    form <- .update_ca_model_formula(model, by, specification)
    newdata <- tryCatch(
      .get_data_from_model("cov_adj", form),
      error = function(e) {
        warning(paste("Could not find direct adjustment data in the call",
                      "stack, or it did not contain the columns specified in",
                      "`by`. Using the covariance adjustment data to generate",
                      "the covariance adjustments"), call. = FALSE)
        tryCatch({
          data_call <- model$call$data
          if (is.null(data_call)) {
            stop("`model` must be fit using a `data` argument")
          }
          data <- eval(data_call, envir = environment(formula(model)))
          stats::model.frame(form, data, na.action = na.pass)
        }, error = function(e) {
          stop(paste(e$message, "in covariance adjustment data"), call. = FALSE)
        })
      })
  } else {
    newdata <- .as_data_frame(newdata)
  }


  if (!is.null(set_to_reference)) {
    if (is.null(names(set_to_reference))) stop(
      "`set_to_reference` must be a named list"
    )
    missing_cols <- setdiff(names(set_to_reference), names(newdata))
    if (length(missing_cols) > 0) {
      warning(
        warningCondition(paste("Columns",
                               paste(missing_cols, collapse = ", "),
                               "from `set_to_reference` not found in newdata"),
                         call = "cov_adj")
      )
    }
    for (c in seq_along(set_to_reference)) {
      col <- names(set_to_reference)[c]
      val <- set_to_reference[[c]]
      if (inherits(val, "set_to_reference_default")) {
        if (is.numeric(newdata[[col]])) {
          val <- min(newdata[[col]], na.rm = TRUE)
        } else if (is.logical(newdata[[col]])) {
          val <- as.logical(min(newdata[[col]], na.rm = TRUE))
        } else if (!inherits(newdata[[col]], "factor")) {
          val <- levels(factor(newdata[[col]]))[1L]
        } else {
          val <- levels(newdata[[col]])[1L]
        }
      }
      newdata[[col]] <- val
    }
  }
  if (!is.null(specification)) {
    trt_name <- var_names(specification,'t')
    if (trt_name %in% names(newdata))
      if (is.numeric(treatment(specification)[, 1])) {
        newdata[[trt_name]] <- min(abs(treatment(specification)[, 1]))
      } else if (is.logical(treatment(specification)[, 1])) {
        newdata[[trt_name]] <- FALSE
      } else if (is.factor(treatment(specification)[, 1])) {
        newdata[[trt_name]] <- levels(treatment(specification)[, 1])[1]
      } else {
        warning(
          paste("If treatment is an independent variable of the covariance",
                "model, predictions may (incorrectly) include treatment",
                "contributions. For now, setting these contributions to 0 is",
                "implemented only for logical or numeric treatments, but this",
                "treatment is a factor.")
        )
      }
    if (specification@unit_of_assignment_type == "none") {
      newdata$..uoa.. <- rownames(newdata)
    }
  } else if (is.null(set_to_reference)) {
    warning(
      paste("Without a specification, post-treatment variables in the",
            "covariance adjustment model cannot be ensured not to contribute",
            "to predictions that offset study outcomes. Use the",
            "set_to_reference argument or pass the StudySpecification to",
            "cov_adj() to avoid this warning.")
    )
  }

  psl <- .make_PreSandwichLayer(model, newdata)

  if (is.null(specification)) {
    return(psl)
  } else {
    sl <- as.SandwichLayer(psl, specification, by = by, Q_data = newdata)
    return(sl)
  }
}

##' Instruct \code{cov_adj()} to find a default reference value for columns in
##' a covariance adjustment model
##' @details
##' When set as the entry of a list passed to the \code{set_to_reference}
##' argument of \code{cov_adj()}, \code{cov_adj()} uses the following
##' replacement values depending on the column type:
##'  - factor/character: minimum level
##'  - numeric: minimum value
##'  - logical: FALSE
##' @return empty list that inherits from class \code{set_to_reference_default}
##' @export
default <- function() {
  structure(list(), class = "set_to_reference_default")
}

##' @title (Internal) Add columns for merging covariance adjustment and direct
##'   adjustment samples to model formula
##' @details This function is typically used prior to
##'   \code{.get_data_from_model()} and incorporates information provided in a
##'   \code{by} argument to ensure the necessary columns for merging the two
##'   samples are included in any \code{model.frame()} calls.
##' @inheritParams cov_adj
##' @return formula
##' @keywords internal
.update_ca_model_formula <- function(model, by = NULL, specification = NULL) {
  form <- deparse1(formula(model))
  if (!is.null(specification)) {
    form <- paste(
      c(form, paste(var_names(specification, "u"), collapse = " + ")),
      collapse = " + "
    )
  }
  if (!is.null(by) && is.null(names(by))) {
    form <- paste(c(form, paste(by, collapse = " + ")), collapse = " + ")
  } else if (!is.null(names(by))) {
    names(by)[names(by) == ""] <- by[names(by) == ""]
    form <- paste(c(form, paste(names(by), collapse = " + ")), collapse = " + ")
  }

  form <- as.formula(form)
  return(form)
}

setClass("StudySpecification",
         slots = c(structure = "data.frame",
                   column_index = "character",
                   type = "character",
                   unit_of_assignment_type = "character",
                   call = "call"))

setValidity("StudySpecification", function(object) {
  if (any(dim(object@structure) == 0)) {
    return("@structure must have positive dimensions")
  }
  if (any(duplicated(colnames(object@structure)))) {
    return("variables cannot be used more than once")
  }
  tr <- object@structure[object@column_index == "t"]
  if (ncol(tr) == 0) {
    return("Missing treatment index")
  }
  if (ncol(tr) > 1) {
    return("Only one treatment variable allowed")
  }
  tr <- tr[, 1]
  if (is.null(tr) ||
        (!is.factor(tr) && !is.numeric(tr) &&
         !is.character(tr)) && !is.logical(tr)) {
    return("Invalid treatment; must be factor, numeric or character")
  }
  if (length(table(tr)) < 2) {
    return("Invalid treatment; treatment can not be constant")
  }
  if (ncol(object@structure) != length(object@column_index)) {
    return("@column_index does not agree with number of columns in @structure")
  }
  if (any(colnames(object@structure) != names(object@column_index))) {
    return("name disagree between @structure and @column_index")
  }
  if (!all(object@column_index %in% c("t", "u", "b", "f"))) {
    wrong <- object@column_index[!object@column_index %in%
                                   c("t", "u", "b", "f")]
    return(paste("@column_index specification elements must be [t,u,b,f].",
                 "unknown elements:", paste(wrong, collapse = ", ")))
  }
  if (!object@type %in% c("RCT", "RD", "Obs")) {
    return(paste("@type must be one of [RCT,RD,Obs]. unknown @type:",
                 object@type))
  }
  if (object@type != "RD" && any(object@column_index == "f")) {
    return("Forcing variables only valid in RD")
  }
  if (object@type == "RD" && !any(object@column_index == "f")) {
    return("RD specifications must include at least one forcing variables")
  }
  if (!object@unit_of_assignment_type %in%
        c("cluster", "unitid", "unit_of_assignment", "none")) {
    return(paste('valid `unit_of_assignment_type`s are "unit_of_assignment",',
                 '"cluster" or "unitid"'))
  }
  return(TRUE)
})


##' Helper function to create a new \code{StudySpecification}. Called internally
##' from \code{rct_spec()}, \code{rd_spec()} or \code{obs_spec()}.
##' @title (Internal) Create a new \code{StudySpecification} object.
##' @param form Formula to create StudySpecification, see help for
##'   \code{rcr_spec()}, \code{rd_spec()} or \code{obs_spec()} for details.
##' @param data The data set
##' @param type One of "RCT", "RD", or "Obs"
##' @param subset Any subset information
##' @param call The call generating the \code{StudySpecification}.
##' @param na.fail Should it error on NA's (\code{TRUE}) or remove them
##'   (\code{FALSE})?
##' @param called_from_lmitt Logical; was this called inside \code{lmitt()}, or
##'   was it called from \code{*_spec()} (default).
##' @return A new StudySpecification object
##' @importFrom stats formula complete.cases terms
##' @importFrom utils capture.output
##' @keywords internal
.new_StudySpecification <- function(form,
                       data,
                       type,
                       subset = NULL,
                       call = NULL,
                       na.fail = TRUE,
                       called_from_lmitt = FALSE) {

  if (is.null(call) | !is.call(call)) {
    call <- match.call()
    warning(paste("Invalid call passed to `.new_StudySpecification`, using default.",
                  "Please use rd_spec, rct_spec, or obs_spec instead ",
                  "of `.new_StudySpecification` directly."))
  }

  if (!is.null(subset)) {
    data <- subset(data, subset = subset)
  }

  ## #174 convert all data.frames
  datadf <- .as_data_frame(data)
  # Moved this prior to `environment` call to hopefully detect more
  # bad input.

  ## keep formula's environment
  env <- environment(terms(form, data = data))
  environment(form) <- env
  call$formula <- form

  data <- datadf


  ### Track whether StudySpecification uses uoa/cluster/unitid for nicer output
  ### later

  if (grepl("unit_of_assignment\\([a-zA-Z]", deparse1(form)) |
        grepl("uoa\\([a-zA-Z]", deparse1(form))) {
    autype <- "unit_of_assignment"
  } else if (grepl("cluster\\([a-zA-Z]", deparse1(form))) {
    autype <- "cluster"
  } else if (grepl("unitid\\([a-zA-Z]", deparse1(form))) {
    autype <- "unitid"
  } else {
    autype <- "none"
    if (options()$propertee_warn_on_no_unit_of_assignment &
                  !called_from_lmitt) {
      warning(paste("The StudySpecification was created without an explicit",
                    "unit of assignment/unit ID. Merges going forward will",
                    "be done by row. It is up to the user to ensure that",
                    "row order is not modified.\nTo prevent this warning,",
                    "provide an explicit unit of assignment/unit ID."))
    }

    data[["..uoa.."]] <- rownames(data)
    form <- update(form, . ~ . + unit_of_assignment(..uoa..))
  }

  # Ensure whichever unit of assignment function is used, `unit_of_assignment`
  # is called
  form <- .update_form_to_unit_of_assignment(form)

  m <- do.call(data.frame,
               c(model.frame(form, data, na.action = na.pass),
                 check.names = FALSE))

  cd <- .rename_model_frame_columns(m)
  m <- cd[["renamedModelFrame"]]
  index <- cd[["index"]]

  # #94 handling NA's in non-treatment columns
  completecases <- stats::complete.cases(m[, index != "t"])
  na_tx  <- is.na(m[, index == "t", drop=TRUE])
  if (!all(completecases)) {
    if ( na.fail & !all(completecases | na_tx) ) {
      stop(paste("Missing values cannot be found in unit of assignment,",
                 "block or cluster variables (unless treatment is also NA).",
                 "Use option `na.fail = FALSE` for automatic removal of",
                 "incomplete cases."))
    } else {
      m <- m[completecases, ]
    }
  }





  m_collapse <- unique(m)

  rownames(m_collapse) <- NULL

  ########### Examine treatment variable.
  treatment <- m_collapse[, index == "t"]

  if (options()$propertee_warn_on_conditional_treatment &
                grepl("[<>=]", deparse1(form[[2]]))) {
    # If the user is using conditionals, we'll be converting logical to numeric
    # later but don't need to the add'l warning message.
    if (!is.logical(treatment)) {
      stop(paste("treatment has conditional but isn't logical.\n",
                 "This may happen if treatment variable name has\n",
                 "'<', '>', or '=' in it. Rename variable to proceed."))
    }
  } else if (!is.factor(treatment) &
             !is.numeric(treatment) &
             !is.character(treatment) &
             !is.logical(treatment)) {
    # Case 3: Treatment is not a factor, a conditional, a numeric, a logical or
    # a character. It is some other type. Warn user before converting to
    # numeric.
    warning(paste("Treatment variables which are not factors, numeric or",
                  " character are converted into numeric.\nIt is STRONGLY",
                  "recommended to do this conversion yourself to ensure it",
                  "proceeds as you expect."))
    treatment <- as.numeric(treatment)
  }
  # Case 0: treatment is numeric, logical, or character. No additional steps
  # needed.

  m_collapse[, index == "t"] <- treatment

  differing <- duplicated(m_collapse[, index == "u"])
  if (any(differing)) {
    noncon <- m_collapse[differing, index == "u", drop = FALSE]

    # Format data using utils::str or base formatting
    if (nrow(noncon) >= 6) {
      data_text <- paste(capture.output(noncon[1:5, , drop = FALSE]), collapse = "\n")
      data_text <- paste0(data_text, "\n...")
    } else {
      data_text <- paste(capture.output(noncon), collapse = "\n")
    }

    stop(paste0("Units of assignment with non-constant treatment, block or forcing:\n",
                data_text, "\n",
                "Each of treatment assignment, block and forcing must be ",
                "constant within unit of assignment."))
  }

  return(new("StudySpecification",
             structure = m_collapse,
             column_index = index,
             type = type,
             unit_of_assignment_type = autype,
             call = call))
}

##' @title Generates a \code{StudySpecification} object with the given
##'   specifications.
##'
##' @description Generate a randomized control treatment StudySpecification
##'   ([rct_spec()]), or an observational StudySpecification ([obs_spec()]), or
##'   a regression discontinuity StudySpecification ([rd_spec()]).
##'
##' @details The formula should include exactly one [unit_of_assignment()] to
##'   identify the units of assignment (one or more variables). (\code{uoa},
##'   \code{cluster}, or \code{unitid} are synonyms for
##'   \code{unit_of_assignment}; the choice of which has no impact on the
##'   analysis. See below for a limited exception in which the
##'   \code{unit_of_assignment} specification may be omitted.) If defining an
##'   \code{rd_spec}, the formula must also include a [forcing()] entry. The
##'   formula may optionally include a [block()] as well. Each of these can take
##'   in multiple variables, e.g. to pass both a household ID and individual ID
##'   as unit of assignment, use \code{uoa(hhid, iid)} and not \code{uoa(hhid) +
##'   uoa(iid)}.
##'
##'   The treatment variable passed into the left-hand side of \code{formula}
##'   can either be \code{logical}, \code{numeric}, or \code{character}. If it
##'   is anything else, it attempts conversion to one of those types (for
##'   example, \code{factor} and \code{ordered} are converted to \code{numeric}
##'   if the levels are \code{numeric}, otherwise to \code{character}). If the
##'   treatment is not \code{logical} or \code{numeric} with only values 0 and
##'   1, in order to generate weights with [ate()] or [ett()], the
##'   \code{dichotomy} argument must be used in those functions to identify the
##'   treatment and control groups. See [ett()] for more details on specifying a
##'   \code{dichotomy}.
##'
##'   There are a few aliases for each version.
##'
##'   If the formula excludes a \code{unit_of_assignment()}, data merges are
##'   performed on row order. Such formulas can also be passed as the
##'   specification argument to lmitt(), and that is their primary intended use
##'   case. It is recommended that each formula argument passed to
##'   *_specification() include a \code{unit_of_assignment()}, \code{uoa()} or
##'   \code{cluster()} term identifying the key variable(s) with which
##'   \code{StudySpecification} data is to be merged with analysis data.
##'   Exceptions to this rule will be met with a warning. To disable the
##'   warning, run \code{options("propertee_warn_on_no_unit_of_assignment" =
##'   FALSE)}.
##'
##'   The units of assignment, blocks, and forcing variables must be
##'   \code{numeric} or \code{character}. If they are otherwise, an attempt is
##'   made to cast them into \code{character}.
##'
##' @param formula a \code{formula} defining the \code{StudySpecification}
##'   components. See `Details` for specification.
##' @param data the data set from which to build the StudySpecification. Note
##'   that this data need not be the same as used to estimate the treatment
##'   effect; rather the \code{data} passed should contain information about the
##'   units of treatment assignment (as opposed to the units of analysis).
##' @param subset optional, subset the data before creating the
##'   \code{StudySpecification} object
##' @param na.fail If \code{TRUE} (default), any missing data found in the
##'   variables specified in \code{formula} (excluding treatment) will trigger
##'   an error. If \code{FALSE}, non-complete cases will be dropped before the
##'   creation of the \code{StudySpecification}
##' @return a \code{StudySpecification} object of the requested type for use in
##'   further analysis.
##' @export
##' @rdname StudySpecification_objects
##' @examples
##' data(simdata)
##' spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid),
##'                   data = simdata)
##'
##' data(schooldata)
##' spec <- obs_spec(treatment ~ unit_of_assignment(schoolid) + block(state),
##'                   data = schooldata)
rct_spec <- function(formula,
                       data,
                       subset = NULL,
                       na.fail = TRUE) {
  .check_spec_formula(formula)
  if (!is.null(substitute(subset))) {
    # Moving this inside .new_StudySpecification causes some scoping issues; it
    # should probably be inside there, but this works fine and is easier to
    # handle #209
    subset <- eval(substitute(subset), data, parent.frame())
  }
  return(.new_StudySpecification(form = formula,
                     data = data,
                     type = "RCT",
                     subset = subset,
                     call = match.call(),
                     na.fail = na.fail))
}

##' @export
##' @rdname StudySpecification_objects
rd_spec <- function(formula,
                      data,
                      subset = NULL,
                      na.fail = TRUE) {
  .check_spec_formula(formula, allow_forcing = TRUE)
  if (!is.null(substitute(subset))) {
    # Moving this inside .new_StudySpecification causes some scoping issues; it
    # should probably be inside there, but this works fine and is easier to
    # handle #209
    subset <- eval(substitute(subset), data, parent.frame())
  }
  return(.new_StudySpecification(form = formula,
                     data = data,
                     type = "RD",
                     subset = subset,
                     call = match.call(),
                     na.fail = na.fail))
}

##' @export
##' @rdname StudySpecification_objects
obs_spec <- function(formula,
                       data,
                       subset = NULL,
                       na.fail = TRUE) {
  .check_spec_formula(formula)
  if (!is.null(substitute(subset))) {
    # Moving this inside .new_StudySpecification causes some scoping issues; it
    # should probably be inside there, but this works fine and is easier to
    # handle #209
    subset <- eval(substitute(subset), data, parent.frame())
  }
  return(.new_StudySpecification(form = formula,
                     data = data,
                     type = "Obs",
                     subset = subset,
                     call = match.call(),
                     na.fail = na.fail))
}


################### Aliases
##' @export
##' @rdname StudySpecification_objects
rct_specification <- rct_spec

##' @export
##' @rdname StudySpecification_objects
rd_specification <- rd_spec

##' @export
##' @rdname StudySpecification_objects
obs_specification <- obs_spec

##' @export
##' @rdname StudySpecification_objects
obsstudy_spec <- obs_spec

##' @export
##' @rdname StudySpecification_objects
obsstudy_specification <- obs_spec


##' @title Show a \code{StudySpecification}
##' @description Display information about a \code{StudySpecification} object
##' @param object \code{StudySpecification} object, usually a result of a call
##'   to [rct_spec()], [obs_spec()], or [rd_spec()].
##' @return \code{object}, invisibly.
##' @export
setMethod("show", "StudySpecification", function(object) {
  spectype <- switch(object@type,
                     "RCT" = "Randomized Control Trial",
                     "RD" = "Regression Discontinuity StudySpecification",
                     "Obs" = "Observational Study")
  uoatype <- switch(object@unit_of_assignment_type,
                    "unit_of_assignment" = "Unit of Assignment",
                    "cluster" = "Cluster",
                    "unitid"  = "Unitid")

  cat(spectype)
  cat("\n\n")


  vartab <- var_table(object)
  # Add a nice separating line for printing
  vartab <- rbind(c("---------", "---------"),
                  vartab)
  print(data.frame(vartab), row.names = FALSE, right = FALSE)

  cat("\n")
  invisible(object)
})

##' After calling \code{model.frame()} on the formula input to
##' \code{.new_StudySpecification()}, the names of the columns will include
##' function names, e.g. "block(blockvar)". This function strips all these.
##'
##' @title (Internal) Rename columns to strip function calls
##' @param modframe A \code{data.frame}.
##' @return The \code{data.frame} with function calls removed
##' @keywords internal
.rename_model_frame_columns <- function(modframe) {

  index <- rep("t", ncol(modframe))

  rename_vars <- function(modelframe, index, type) {
    pos <- grepl(paste0("^", type), colnames(modelframe))
    if (any(pos)) {
      index[pos] <- substr(type, 1, 1)
      vars <- colnames(modelframe)[pos][1]
      vars <- sub(paste0("^", type, "\\("), "", vars)
      if (grepl("^cbind", vars)) {
        # If user called something like `uoa(cbind(a, b))`
        vars <- sub("^cbind\\(", "", vars)
        vars <- sub("\\)\\).*$", "", vars)
      } else {
        # If user just called `uoa(a, b)`
        vars <- sub("\\)[\\.0-9]*$", "", vars)
      }
      vars <- strsplit(gsub(" ", "", vars), ",")[[1]]
      colnames(modelframe)[pos] <- vars
    }
    return(list(modelframe, index))
  }

  o <- rename_vars(modframe, index, "unit_of_assignment")
  modframe <- o[[1]]
  index <- o[[2]]
  o <- rename_vars(modframe, index, "block")
  modframe <- o[[1]]
  index <- o[[2]]
  o <- rename_vars(modframe, index, "forcing")
  modframe <- o[[1]]
  index <- o[[2]]

  return(list(renamedModelFrame = modframe,
              index = index))
}

##' @title Extract Variable Names from \code{StudySpecification}
##'
##' @description Methods to extract the variable names to the elements of the
##'   structure of the \code{StudySpecification} (e.g. treatment, unit of
##'   analysis, etc)
##'
##' @details When \code{compress} is \code{TRUE}, the result will always have
##'   two columns. When \code{FALSE}, the result will have number of columns
##'   equal to the largest number of variables in a particular role, plus one.
##'   E.g., a call such as \code{rct_spec(z ~ unitid(a, b, c, d) ...} will have
##'   4+1=5 columns in the output matrix with \code{compress = FALSE}.
##'
##'   When \code{report_all} is \code{TRUE}, the matrix is guaranteed to have 3
##'   rows (when the \code{specification} is an RCT or Obs) or 4 rows (when the
##'   \code{specification} is a RD), with empty variable entries as appropriate.
##'   When \code{FALSE}, the matrix will have minimum 2 rows (treatment and unit
##'   of assignment/unitid/cluster), with additional rows for blocks and forcing
##'   if included in the \code{StudySpecification}.
##' @param specification a \code{StudySpecification} object
##' @param compress should multiple variables be compressed into a
##'   comma-separated string? Default \code{TRUE}. If \code{FALSE}, multiple
##'   columns can be created instead.
##' @param report_all should we report all possible structures even if they
##'   don't exist in the \code{StudySpecification}? Default \code{FALSE}.
##' @return \code{var_table} returns the requested table. \code{var_names}
##'   returns a vector of variable names.
##' @export
##' @rdname StudySpecification_var_names
##' @order 1
##' @examples
##' spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
##' var_table(spec)
##' var_table(spec, compress = FALSE)
##' var_names(spec, "t")
##' var_names(spec, "u")
##' var_names(spec, "b")
var_table <- function(specification, compress = TRUE, report_all = FALSE) {
  uoatype <- switch(specification@unit_of_assignment_type,
                    "unit_of_assignment" = "Unit of Assignment",
                    "cluster" = "Cluster",
                    "unitid"  = "Unitid")

  # Start with the "table" as a list; easier to handle non-equal number of
  # elements.
  rows <- list()
  rows[["t"]] <- c("Treatment", var_names(specification, "t"))
  rows[["u"]] <- c(uoatype    , var_names(specification, "u"))
  rows[["b"]] <- c("Block"    , var_names(specification, "b"))
  if (specification@type == "RD") {
    rows[["f"]] <- c("Forcing"  , var_names(specification, "f"))
  }

  # Identify if we have more than one variable specified in a specific
  # structure, and if so how ,long
  maxvar <- max(vapply(rows, length, 1))
  if (maxvar < 2) {
    # Should never hit this error
    stop("Internal error: No variables identified!")
  }

  if (!report_all) {
    # Drop any rows we don't need
    rows <- rows[vapply(rows, length, 1) > 1]
  }

  # add NA padding to make all rows the same length so the `rbind` that follows
  # doesn't need to recycle length
  rows <- lapply(rows, function(x) {
    if (length(x) < maxvar) {
      x <- c(x, rep(NA, maxvar - length(x)))
    }
    return(x)
  })

  # list to data.frame
  out <- do.call(rbind, rows)

  if (compress == TRUE) {
    # If compressing, collapse any repeats if necessary
    rows <- lapply(rows, function(x) {
      x <- x[!is.na(x)]
      c(x[1], paste(x[-1], collapse = ", "))
    })
    out <- do.call(rbind, rows)
    colnames(out) <- c("Structure", "Variables")
    out[out == ""] <- NA
  } else {
    if (maxvar == 2) {
      colnames(out) <- c("Structure", "Variables")
    } else {
      colnames(out) <- c("Structure", paste("Variable", seq_len(maxvar - 1)))
    }
  }
  rownames(out) <- NULL
  return(out)
}

##' @param type one of "t", "u", "b", "f"; for "treatment",
##'   "unit_of_assignment", "block", and "forcing" respectively
##' @param implicitBlocks If the \code{StudySpecification} is created without
##'   blocks, setting this to \code{TRUE} will return "\code{.blocks_internal}"
##'   as the variable name corresponding to the blocks.
##' @export
##' @rdname StudySpecification_var_names
##' @order 2
var_names <- function(specification, type, implicitBlocks = FALSE) {
  stopifnot(inherits(specification, "StudySpecification"))
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("t", "u", "b", "f"))
  if (type == "b" && !("b" %in% specification@column_index) && implicitBlocks) {
    return(".blocks_internal")
  }
  return(names(specification@structure)[specification@column_index == type])
}

##' @title Check for variable agreement within units of assignment
##'
##' @description Useful for debugging purposes to ensure that there is
##'   concordance between variables in the \code{StudySpecification} and data.
##'
##' @details Consider the following scenario: A \code{StudySpecification} is
##'   generated from some dataset, "data1", which includes a block variable
##'   "b1". Within each unique unit of assignment/unitid/cluster of "data1", it
##'   must be the case that "b1" is constant. (Otherwise the creation of the
##'   \code{StudySpecification} will fail.)
##'
##'   Next, a model is fit which includes weights generated from the
##'   \code{StudySpecification}, but on dataset "data2". In "data2", the block
##'   variable "b1" also exists, but due to some issue with data cleaning, does
##'   not agree with "b1" in "data1".
##'
##'   This could cause errors, either directly (via actual error messages) or
##'   simply produce nonsense results. [specification_data_concordance()] is
##'   specificationed to help debug these scenarios by providing information on
##'   whether variables in both the data used in the creation of
##'   \code{specification} ("data1" in the above example) and some new dataset,
##'   \code{data}, ("data2" in the above example) have any inconsistencies.
##'
##' @param specification a \code{StudySpecification} object
##' @param data a new data set, presumably not the same used to create
##'   \code{specification}.
##' @param by optional; named vector or list connecting names of variables in
##'   \code{specification} to variables in \code{data}. Names represent
##'   variables in \code{specification}; values represent variables in
##'   \code{data}. Only needed if variable names differ.
##' @param warn_on_nonexistence default \code{TRUE}. If a variable does not
##'   exist in \code{data}, should this be flagged? If \code{FALSE}, silently
##'   move on if a variable doesn't exist in \code{data}.
##' @return invisibly \code{TRUE} if no warnings are produced, \code{FALSE} if
##'   any warnings are produced.
##' @export
specification_data_concordance <- function(specification,
                                    data,
                                    by = NULL,
                                    warn_on_nonexistence = TRUE) {
  if (!is.null(by)) {
    specification <- .update_by(specification, data, by)
  }

  if (!all(var_names(specification, "u" ) %in% names(data))) {
    stop(paste("Missing unit of assignment/unitid/cluster variables",
               "in data set.\nAre you missing a `by=` argument?"))
  }

  merged <- merge(specification@structure, data,
                  all = TRUE,
                  by = var_names(specification, "u"))

  # Returns TRUE if any warnings were printed, FALSE otherwise
  .check <- function(type, specification, data) {
    anywarnings <- FALSE

    vnames <- var_names(specification, type)
    # If this particular type of variable exists in the specification...
    if (length(vnames) > 0) {
      # Loop over all variables of `type`
      for (var in vnames) {
        # If variable name exists in the data ...
        xynames <- paste0(var, c(".x", ".y"))
        if (all(xynames %in% names(data))) {
          varcompare <- merged[, xynames]
          # Ensure differences are 0
          if (!all(apply(varcompare[, xynames], 1, diff) == 0)) {
            warning(paste0("Inconsistencies in variable `", var, "`"))
            anywarnings <- TRUE
          }
        } else {
          if (warn_on_nonexistence) {
            warning(paste0("Variable `", var, "` not found in data"))
            anywarnings <- TRUE
          }
        }
      }
    }
    if (anywarnings) {
      return(invisible(TRUE))
    } else {
      return(invisible(FALSE))
    }
  }

  treat_had_warnings <- .check("t", specification, merged)
  block_had_warnings <- .check("b", specification, merged)
  force_had_warnings <- .check("f", specification, merged)

  if (any(treat_had_warnings,
          block_had_warnings,
          force_had_warnings)) {
    return(invisible(FALSE))
  } else {
    return(invisible(TRUE))
  }
}

test_that("absorb= argument", {

  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa2, uoa1) + block(bid) + forcing(force),
                   data = simdata)

  mod <- lmitt(y ~ 1, weights = ate(), data = simdata, specification = spec)
  expect_length(coefficients(mod), 3)

  mod <- lmitt(y ~ 1, weights = ate(), data = simdata, absorb = TRUE, specification = spec)
  expect_length(coefficients(mod), 3)

  # subgroup effects
  mod <- lmitt(y ~ as.factor(o), weights = ate(), data = simdata, specification = spec)
  expect_length(coefficients(mod), 12)

  mod <- lmitt(y ~ as.factor(o), weights = ate(), data = simdata, absorb = TRUE, specification = spec)
  expect_length(coefficients(mod), 12)

})

test_that("multiple variables in blocks", {
  data(simdata)
  simdata$bid1 <- (simdata$bid > 1) + 1
  simdata$bid2 <- (simdata$bid != 2) + 2

  spec <- rct_spec(z ~ cluster(uoa2, uoa1) + block(bid1, bid2),
                   data = simdata)

  mod <- lmitt(y ~ dose, data = simdata, absorb = TRUE, specification = spec)
  expect_length(coefficients(mod), 6)

})

# fake call
fc <- call("ls")

identical_specs <- function(a, b) {
  expect_identical(a, b)
}

test_that("StudySpecification creation", {

  tests <- function(d) {
      expect_s4_class(d, "StudySpecification")
      expect_s3_class(d@structure, "data.frame")
      expect_type(d@column_index, "character")
      expect_type(d@type, "character")
      expect_type(d@unit_of_assignment_type, "character")
      expect_type(d@call, "language")
      expect_true(inherits(d@call, "call"))

      expect_equal(ncol(d@structure), length(d@column_index))
      expect_length(d@type, 1)
  }

  # binary treatment
  d <- new("StudySpecification",
           structure = data.frame(a = 0:1, b = c(2, 4)),
           column_index = c("t", "u"),
           type = "RCT",
           unit_of_assignment_type = "cluster",
           call = fc)

  tests(d)

  # logical treatment
  d <- new("StudySpecification",
           structure = data.frame(a = c(TRUE, FALSE), b = c(2, 4)),
           column_index = c("t", "u"),
           type = "RCT",
           unit_of_assignment_type = "cluster",
           call = fc)

  tests(d)

  # >2 treatment levels
  d <- new("StudySpecification",
           structure = data.frame(a = 0:2, b = c(2, 4, 6)),
           column_index = c("t", "u"),
           type = "RCT",
           unit_of_assignment_type = "cluster",
           call = fc)

  tests(d)

  # character treatment

  d <- new("StudySpecification",
           structure = data.frame(a = letters[1:2], b = c(2, 4)),
           column_index = c("t", "u"),
           type = "RCT",
           unit_of_assignment_type = "cluster",
           call = fc)

  tests(d)

  # >2 character treatment levels
  d <- new("StudySpecification",
           structure = data.frame(a = letters[1:3], b = c(2, 4, 6)),
           column_index = c("t", "u"),
           type = "RCT",
           unit_of_assignment_type = "cluster",
           call = fc)

  tests(d)

})

test_that("StudySpecification validity", {

  expect_error(new("StudySpecification",
                   structure = data.frame(),
                   column_index = "t",
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "positive dimensions")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = 0:1,
                                          b = 0:1),
                   column_index = c("t", "t"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "one treatment")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = c(1, 0), a = c(1, 0)),
                   column_index = c("u", "b"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "Missing treatment")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = c(0, 0), a = c(1, 0)),
                   column_index = c("t", "u"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "treatment can not be constant")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = 0:1,
                                          b = c(2, 4)),
                   column_index = c("t", "b"),
                   type = "abc",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "unknown @type")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = c(1, 2),
                                          b = as.factor(c(1, 0)),
                                          c = c(2, 0)),
                   column_index = c("t", "u", "u", "u"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "number of columns")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = 0:1,
                                          b = c(2, 4)),
                   column_index = c("t", "k"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "unknown elements")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = 0:1,
                                          b = c(1, 0)),
                   column_index = c("t", "f"),
                   type = "RCT",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "Forcing variables only valid")

  expect_error(new("StudySpecification",
                   structure = data.frame(a = 0:1,
                                          b = c(1, 0)),
                   column_index = c("t", "u"),
                   type = "RD",
                   unit_of_assignment_type = "cluster",
                   call = fc),
               "at least one forcing")

})

test_that("StudySpecification creation", {
  data(mtcars)
  mtcars <- mtcars[-c(5, 11), ]

  spec_rct <- .new_StudySpecification(vs ~ cluster(qsec), data = mtcars,
                      type = "RCT",
                      call = quote(rct_spec(formula = vs ~ cluster(qsec), data = mtcars)))

  expect_s4_class(spec_rct, "StudySpecification")
  expect_s3_class(spec_rct@structure, "data.frame")
  expect_type(spec_rct@column_index, "character")
  expect_type(spec_rct@type, "character")
  expect_type(spec_rct@unit_of_assignment_type, "character")
  expect_type(spec_rct@call, "language")
  expect_true(inherits(spec_rct@call, "call"))

  expect_equal(dim(spec_rct@structure), c(30, 2))
  expect_length(spec_rct@column_index, 2)
  expect_length(spec_rct@type, 1)

  mtcars_subset <- subset(mtcars, select = c("vs", "qsec"))
  mtcars_subset$vs <- as.factor(mtcars_subset$vs)
  expect_true(all(spec_rct@structure == mtcars_subset))
  expect_equal(spec_rct@column_index, c("t", "u"), ignore_attr = TRUE)
  expect_equal(spec_rct@unit_of_assignment_type, "cluster")
  expect_equal(spec_rct@type, "RCT")

  # subset

  spec_obs <- .new_StudySpecification(
    vs ~ cluster(qsec), data = mtcars, type = "Obs", subset = mtcars$mpg > 17,
    call = quote(obs_spec(formula = vs ~ cluster(qsec), data = mtcars, subset = mtcars$mpg > 17))
  )

  expect_equal(dim(spec_obs@structure), c(sum(mtcars$mpg >  17), 2))
  expect_length(spec_obs@column_index, 2)
  expect_length(spec_obs@type, 1)

  mtcars_subset <- subset(mtcars, select = c("vs", "qsec"),
                          subset = mtcars$mpg > 17)
  mtcars_subset$vs <- as.factor(mtcars_subset$vs)
  expect_true(all(spec_obs@structure == mtcars_subset))
  expect_equal(spec_obs@column_index, c("t", "u"), ignore_attr = TRUE)
  expect_equal(spec_obs@unit_of_assignment_type, "cluster")
  expect_equal(spec_obs@type, "Obs")

  ### Complex specification
  spec_rd <- .new_StudySpecification(
    vs ~ block(disp, gear) + forcing(wt, cyl) + cluster(mpg, qsec),
    data = mtcars, type = "RD",
    call = quote(rd_spec(formula = vs ~ block(disp, gear) + forcing(wt, cyl) + cluster(mpg, qsec), data = mtcars))
  )

  mtcars_subset <- subset(mtcars, select = c("vs", "disp", "gear",
                                             "wt", "cyl", "mpg", "qsec"))
  mtcars_subset$vs <- as.factor(mtcars_subset$vs)
  expect_true(all(spec_rd@structure == mtcars_subset))
  expect_equal(spec_rd@column_index, c("t", "b", "b", "f", "f", "u", "u"),
               ignore_attr = TRUE)
  expect_equal(spec_rd@unit_of_assignment_type, "cluster")
  expect_equal(spec_rd@type, "RD")

  ### Complex specification with unitid
  spec_rd2 <- .new_StudySpecification(vs ~ block(disp, gear) + forcing(wt, cyl) +
                        unitid(mpg, qsec),
                  data = mtcars, type = "RD", call = fc)

  mtcars_subset <- subset(mtcars, select = c("vs", "disp", "gear",
                                             "wt", "cyl", "mpg", "qsec"))
  mtcars_subset$vs <- as.factor(mtcars_subset$vs)
  expect_true(all(spec_rd2@structure == mtcars_subset))
  expect_equal(spec_rd2@column_index, c("t", "b", "b", "f", "f", "u", "u"),
               ignore_attr = TRUE)
  expect_equal(spec_rd2@unit_of_assignment_type, "unitid")
  expect_equal(spec_rd2@type, "RD")

  ### Specific specifications
  rct_spec <- rct_spec(vs ~ cluster(qsec), data = mtcars)
  identical_specs(spec_rct, rct_spec)

  obs_spec <- obs_spec(vs ~ cluster(qsec), data = mtcars,
                        subset = mtcars$mpg > 17)
  identical_specs(spec_obs, obs_spec)

  rd_spec <- rd_spec(vs ~ block(disp, gear) + forcing(wt, cyl) +
                        cluster(mpg, qsec),
                      data = mtcars)
  identical_specs(spec_rd, rd_spec)


  # Missing call

  expect_warning(spec <- .new_StudySpecification(vs ~ cluster(qsec),
                                                 data = mtcars,
                                                 type = "RCT"),
                 "Invalid call")
  expect_identical(spec@call[[1]], as.name(".new_StudySpecification"))

  expect_warning(spec <- .new_StudySpecification(vs ~ cluster(qsec),
                                                 data = mtcars,
                                                 type = "RCT", call = 1),
                 "Invalid call")
  expect_identical(spec@call[[1]], as.name(".new_StudySpecification"))
})

test_that("unit of assignment differs from unit of analysis", {

  data(simdata)

  specrct <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  expect_s4_class(specrct, "StudySpecification")
  expect_equal(nrow(specrct@structure), 10)

  expect_error(rct_spec(z ~ cluster(uoa1) + block(bid), data = simdata),
               "uoa1")

  data(mtcars)
  mtcars$prop <- rep(1:8, 4)

  expect_error(rct_spec(vs ~ cluster(prop), data = mtcars),
               "prop")
  expect_error(rct_spec(vs ~ cluster(prop), data = mtcars),
               "...")

  data(simdata)
  simdata$z[1] <- 1
  expect_error(rct_spec(z ~ cluster(uoa1, uoa2), data = simdata),
               "uoa1")

  simdata$z[1] <- NA
  expect_error(rct_spec(z ~ cluster(uoa1, uoa2), data = simdata),
               "uoa1")
})

test_that("StudySpecification printing", {
  data(simdata)

  specrct <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  specrd  <-  rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                       data = simdata)
  specobs <- obs_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  expect_output(print(specrct), "Randomized")
  expect_output(show(specrct),  "Randomized")

  expect_silent(invisible(capture.output(expect_identical(specrct,
                                                          show(specrct)))))
  expect_silent(invisible(capture.output(expect_identical(specrct,
                                                          print(specrct)))))

  expect_output(print(specrd), "Discontinuity")
  expect_output(show(specrd),  "Discontinuity")
  expect_output(print(specobs), "Observational")
  expect_output(show(specobs),  "Observational")

  expect_output(show(specrct), "z")
  expect_output(show(specrct), "uoa1")
  expect_output(show(specrct), "bid")

  expect_output(show(specobs), "z")
  expect_output(show(specobs), "uoa1")
  expect_output(show(specobs), "bid")

  expect_output(show(specrd), "z")
  expect_output(show(specrd), "uoa1")
  expect_output(show(specrd), "bid")
  expect_output(show(specrd), "force")
})


test_that("StudySpecification type conversions", {

  # Helper function to check equivalence of StudySpecifications without worrying about some
  # oddities in @call
  check_equiv_spec <- function(d1, d2) {
    expect_identical(d1@type, d2@type)
    expect_identical(d1@structure, d2@structure)
    expect_identical(d1@column_index, d2@column_index)
    expect_identical(d1@unit_of_assignment_type, d2@unit_of_assignment_type)

    # Due to formula hacking in the @call, the calls can be non-identical.
    # However, they should both produce the same StudySpecification if `eval`'d.
    d1redux <- eval(d1@call)
    d2redux <- eval(d2@call)
    expect_identical(d1redux@type, d2redux@type)
    expect_identical(d1redux@structure, d2redux@structure)
    expect_identical(d1redux@column_index, d2redux@column_index)
    expect_identical(d1redux@unit_of_assignment_type,
                     d2redux@unit_of_assignment_type)

    invisible(TRUE)
  }

  data(simdata)

  specrct <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  specrd  <-  rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                       data = simdata)
  specobs <- obs_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  # Obs <-> RCT
  expect_identical(specrct, as_rct_spec(specobs))
  expect_identical(specobs, as_obs_spec(specrct))

  # RD -> Obs/RCT
  expect_error(as_rct_spec(specrd), "will be dropped")
  expect_error(as_obs_spec(specrd), "will be dropped")
  check_equiv_spec(specrct, as_rct_spec(specrd, loseforcing = TRUE))
  check_equiv_spec(specobs, as_obs_spec(specrd, loseforcing = TRUE))

  # RCT/RD -> Obs
  expect_error(as_rd_spec(specrct))
  check_equiv_spec(specrd, as_rd_spec(specrct, data = simdata,
                                      forcing = ~ . + forcing(force)))
  check_equiv_spec(specrd, as_rd_spec(specrct, data = simdata,
                                      forcing = . ~ . + forcing(force)))
  check_equiv_spec(specrd, as_rd_spec(specobs, data = simdata,
                                      forcing = ~ . + forcing(force)))
  check_equiv_spec(specrd, as_rd_spec(specobs, data = simdata,
                                       forcing = . ~ . + forcing(force)))
  expect_error(as_rd_spec(specobs, data = simdata, forcing = simdata$force),
               "as a formula")

})


test_that("variable transformations in StudySpecification", {

  data(simdata)

  spec <- rct_spec(dose + 3 ~ uoa(uoa1, uoa2), data = simdata)
  expect_true(all(treatment(spec)[, 1] %in% (simdata$dose + 3)))

  spec <- rd_spec(force > 4 & bid < 2 ~ cluster(uoa1, uoa2) + forcing(force),
                   data = simdata)
  expect_true(is.logical(treatment(spec)[, 1]))
})

test_that("#26 cbind in identification variables", {
  data(simdata)
  spec1 <- obs_spec(o ~ cluster(cbind(uoa1, uoa2)), data = simdata)
  spec2 <- obs_spec(o ~ cluster(uoa1, uoa2), data = simdata)

  spec1@call <- fc
  spec2@call <- fc
  expect_identical(spec1, spec2)

})

test_that("variable_concordance", {

  data(simdata)

  # No issues
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_silent(expect_true(specification_data_concordance(spec, simdata)))

  # inconsistent block
  sd <- simdata
  sd$bid[1] <- 2

  expect_warning(a <- specification_data_concordance(spec, sd),
                 "Inconsistencies.*`bid`")
  expect_false(a)

  # Inconsistent block & force
  sd$force[2] <- 2

  expect_warning(expect_warning(a <- specification_data_concordance(spec, sd),
                                "Inconsistencies.*`bid`"),
                  "Inconsistencies.*`force`")
  expect_false(a)

  # inconsistent treatment
  sd <- simdata
  sd$z[1] <- 1

  expect_warning(a <- specification_data_concordance(spec, sd),
                 "Inconsistencies.*`z`")

  # Missing variables
  sd <- simdata
  sd$z <- NULL
  expect_warning(a <- specification_data_concordance(spec, sd),
                 "`z` not found")
  expect_false(a)
  expect_silent(a <- specification_data_concordance(spec, sd,
                                             warn_on_nonexistence = FALSE))
  expect_true(a)

  sd <- simdata
  sd$bid <- NULL
  expect_warning(a <- specification_data_concordance(spec, sd),
                 "`bid` not found")
  expect_false(a)
  expect_silent(a <- specification_data_concordance(spec, sd,
                                             warn_on_nonexistence = FALSE))
  expect_true(a)

  sd <- simdata
  sd$force <- NULL
  expect_warning(a <- specification_data_concordance(spec, sd),
                 "`force` not found")
  expect_false(a)
  expect_silent(a <- specification_data_concordance(spec, sd,
                                             warn_on_nonexistence = FALSE))
  expect_true(a)

  sd$uoa1 <- NULL
  expect_error(specification_data_concordance(spec, sd),
               "missing a `by=`")

  # No block
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force),
                   data = simdata)
  expect_silent(expect_true(specification_data_concordance(spec, simdata)))

  # multiple variables
  sd <- simdata
  sd$force2 <- sd$force + 1
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force, force2),
                   data = sd)
  expect_silent(expect_true(specification_data_concordance(spec, sd)))

  expect_warning(specification_data_concordance(spec, simdata),
                 "`force2` not found")

  sd$force[1] <- 1
  expect_warning(specification_data_concordance(spec, sd),
                 "Inconsistencies.*`force`$")

  sd2 <- simdata
  sd2$force[1] <- 1
  expect_warning(expect_warning(specification_data_concordance(spec, sd2),
                                "Inconsistencies.*`force`$"),
                 "`force2` not found")

  sd$force2[1] <- 2
  expect_warning(expect_warning(specification_data_concordance(spec, sd),
                                "Inconsistencies.*`force`$"),
                 "Inconsistencies.*`force2`$")

  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  names(simdata)[names(simdata) == "uoa1"] <- "ccc"
  names(simdata)[names(simdata) == "force"] <- "fff"

  expect_silent(expect_true({
    specification_data_concordance(spec, simdata,
                            by = list(force = "fff",
                                      uoa1 = "ccc"))
  }))

})

test_that("NA's in data to create StudySpecification", {
  data(simdata)
  sd2 <- simdata
  sd2$uoa1[1] <- NA

  expect_error(obs_spec(z ~ unitid(uoa1, uoa2), data = sd2),
               "Missing values cannot be found")

  spec1 <- obs_spec(z ~ unitid(uoa1, uoa2), data = simdata)
  spec2 <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd2, na.fail = FALSE)

  spec1@call <- call("c")
  spec2@call <- call("c")

  expect_identical(spec1, spec2)

  sd2 <- simdata
  sd2$bid <- NA

  expect_silent(spec <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd2))
  expect_error(obs_spec(z ~ unitid(uoa1, uoa2) + block(bid), data = sd2),
               "Missing values cannot be found")

  # treatment NAs allowed & passed forward
  sd2 <- simdata
  sd2$z[1:10] <- NA
  expect_silent(spec <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd2,
                                  na.fail = TRUE))
  expect_equal(nrow(spec@structure), 10)
  expect_silent(spec <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd2,
                                  na.fail = FALSE))
  expect_equal(nrow(spec@structure), 10)

  sd2 <- simdata
  sd2$uoa1[1:4] <- NA
  spec <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd2, na.fail = FALSE)
  expect_equal(nrow(spec@structure), 9)

  # Cases with NAs in treatment and additional structure
  # variable(s) silently omitted
  sd3  <- sd2
  sd3$z[1:4]  <- NA
  expect_silent(spec_  <- obs_spec(z ~ unitid(uoa1, uoa2), data = sd3,
                                   na.fail = TRUE))
  expect_equal(nrow(spec_@structure), 9)

  spec@call <- spec_@call <- call("c")
  expect_identical(spec, spec_)
})

test_that("column name 'cluster' doesn't cause issues", {
  data(simdata)
  simdata$cluster <- simdata$uoa1

  expect_no_error(rct_spec(z ~ cluster(cluster, uoa2), data = simdata))
  expect_true(TRUE) # Avoid an empty test warning
})

test_that("#209 fix subsetting", {
  data(simdata)
  spec1 <- rct_spec(z ~ uoa(uoa1,uoa2), subset= simdata$bid != 1, data=simdata)
  spec2 <- rct_spec(z ~ uoa(uoa1,uoa2), subset= bid != 1, data=simdata)
  simdatasub <- subset(simdata, bid != 1)
  spec3 <- rct_spec(z ~ uoa(uoa1,uoa2), data=simdatasub)
  expect_identical(spec1@structure,
                   spec2@structure)
  expect_identical(spec1@structure,
                   spec3@structure)

  spec1 <- obs_spec(z ~ uoa(uoa1,uoa2), subset= simdata$bid != 1, data=simdata)
  spec2 <- obs_spec(z ~ uoa(uoa1,uoa2), subset= bid != 1, data=simdata)
  simdatasub <- subset(simdata, bid != 1)
  spec3 <- obs_spec(z ~ uoa(uoa1,uoa2), data=simdatasub)
  expect_identical(spec1@structure,
                   spec2@structure)
  expect_identical(spec1@structure,
                   spec3@structure)
})

test_that("#222 factor UOA", {

  example <- data.frame(schf = as.factor(letters),
                        schn = 1:26,
                        trt = rep(c(0,1),each = 13),
                        Y = rnorm(26))
  sp1 <- rct_spec(trt ~ unitid(schf), data = example)
  sp2 <- rct_spec(trt ~ unitid(schn), data = example)
  mod1 <- lmitt(Y ~ 1,
                specification = sp1,
                data = example,weights = "ate")
  mod2 <- lmitt(Y ~ 1,
                specification = sp2,
                data = example,weights = "ate")
  expect_identical(vcov(mod1), vcov(mod2))

})

test_that("get StudySpecification data from call's formula environment", {
  make_spec <- function(udata, zcol, ucol) {
    spec_form <- reformulate(paste0("unitid(", ucol, ")"), response = zcol)
    rct_specification(spec_form, udata)
  }
  
  udata <- data.frame(uid = seq_len(3), z = c(0, 1, 0))
  m1 <- make_spec(udata, "z", "uid")
  expect_identical(udata, get("udata", environment(m1@call$formula)))
  expect_silent(.make_uoa_cluster_df(m1, "uid"))
  
})

test_that("summary.teeMod basics", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification = spec, weights = ate())
  summod <- summary(mod)
  expect_s3_class(summod, "summary.teeMod")

  out <- capture_output(print(summod))

  expect_match(out, "Estimate")

  expect_no_match(out, "Residuals")
  expect_no_match(out, "R-squared")

})


test_that("teeMod with SandwichLayer offset summary uses vcov_tee SE's", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec),
                    offset = cov_adj(cmod)))

  s <- summary(ssmod, vcov.type = "CR0", dof.type = "stata")
  vc <- vcov_tee(ssmod, type = "CR0")
  dof <- length(unique(paste(simdata$uoa1, simdata$uoa2, sep = "_")))-1
  ix <- row.names(vc) != "offset:(Intercept)"
  expect_equal(s$coefficients[, 2L], sqrt(diag(vc)[ix]))
  expect_equal(s$coefficients[, 3L], ssmod$coefficients[ix] / sqrt(diag(vc)[ix]))
  expect_equal(s$coefficients[, 4L],
               2 * pt(abs(ssmod$coefficients[ix] / sqrt(diag(vc)[ix])),
                      dof,
                      lower.tail = FALSE))

  cmod <- lm(y ~ x, simdata, subset = z == 0)
  # Removing intercept which is otherwise estimated at 0, causing warnings in
  # sqrt(diag(vcov)) calculations below
  ssmod <- lmitt(lm(y ~ assigned() + 0, data = simdata, weights = ate(spec),
                    offset = cov_adj(cmod)))

  s <- summary(ssmod, vcov.type = "CR0", dof.type = "stata")
  vc <- vcov_tee(ssmod, type = "CR0")
  ix <- row.names(vc) != "offset:(Intercept)"
  expect_equal(s$coefficients[, 2L], sqrt(diag(vc)[ix]))
  expect_equal(s$coefficients[, 3L], ssmod$coefficients[ix] / sqrt(diag(vc)[ix]))
  expect_equal(s$coefficients[, 4L],
               2 * pt(abs(ssmod$coefficients[ix] / sqrt(diag(vc)[ix])),
                      dof,
                      lower.tail = FALSE))
})

test_that("teeMod w/o SandwichLayer offset summary uses sandwich SE's", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec)))

  uoas <- apply(simdata[, c("uoa1", "uoa2")],
                1,
                function(...) paste(..., collapse = "_"))
  s <- summary(ssmod, vcov.type = "CR0", cadjust = FALSE)
  expect_equal(
    s$coefficients[, 2],
    sqrt(diag(sandwich::sandwich(ssmod,
                                 meat. = sandwich::meatCL,
                                 cluster = uoas,
                                 cadjust = FALSE))))
  # IK degrees of freedom lead to larger p-values
  sIK <- summary(ssmod, vcov.type = "CR0", dof.type = "IK")
  expect_true(all(s$coefficients[,4] < sIK$coefficients[,4]))
})

test_that("vcov.type argument", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec)))
  expect_equal(summary(ssmod)$vcov.type, "HC0")
  expect_equal(summary(ssmod, vcov.type = "CR0")$vcov.type, "CR0")
  expect_equal(summary(ssmod, vcov.type = "MB0")$vcov.type, "MB0")
  expect_equal(summary(ssmod, vcov.type = "HC0")$vcov.type, "HC0")

  expect_equal(sum(grepl("HC0", capture.output(summary(ssmod)))), 1)
  expect_equal(sum(grepl("CR0",
                         capture.output(summary(ssmod, vcov.type = "CR0")))),
               1)

  expect_equal(sum(grepl("MB0",
                         capture.output(summary(ssmod, vcov.type = "MB0")))),
               1)
  expect_equal(sum(grepl("HC0",
                         capture.output(summary(ssmod, vcov.type = "HC0")))),
               1)

})

test_that("lmitt.form vs as.lmitt", {
  # `lmitt.formula` should report only "Treatment Effects", whereas
  # as.`lmitt`/`lmitt(lm(...` report coefficients since we don't have control
  # over what is reported.
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)

  co <- capture.output(summary(lmitt(y ~ 1, data = simdata, specification = spec)))
  expect_true(any(grepl("Treatment Effects", co)))
  expect_false(any(grepl("Coefficients", co)))

  co <- capture.output(summary(as.lmitt(lm(y ~ z.(spec), data = simdata),
                                        specification = spec)))
  expect_false(any(grepl("Treatment Effects", co)))
  expect_true(any(grepl("Coefficients", co)))

  co <- capture.output(summary(lmitt(lm(y ~ z.(spec), data = simdata),
                                        specification = spec)))
  expect_false(any(grepl("Treatment Effects", co)))
  expect_true(any(grepl("Coefficients", co)))


  ### With interactions, `as.lmitt` should report all coefficients
  mod <- as.lmitt(lm(
    y ~ as.factor(o) + as.factor(o):z.(spec), data = simdata),
    specification = spec)
  suppressWarnings(co <- capture.output(summary(mod), "NaNs"))
  expect_equal(sum(grepl("as.factor(o)", co, fixed = TRUE)),
               8)

})

test_that("issues with coefficients", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)
  mod <- lmitt(y ~ 1, data = simdata, specification =spec, subset = simdata$z == 0)
  s <- summary(mod, vcov.type = "CR0")
  co <- capture.output(s)
  expect_true(any(grepl("Treatment Effects", co)))

  expect_true(any(grepl("calculated via type", co)))

  sslm <- new("teeMod",
              lm(y ~ 0, data = simdata, weights = ate(spec)),
              StudySpecification = spec,
              ctrl_means_model = lm(1~0),
              lmitt_fitted = TRUE)

  expect_true(any(grepl("No Coefficients", capture.output(summary(sslm)))))

})

test_that("catching bug with summary(as.lmitt", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), simdata)

  mod <- lm(y ~ assigned(spec), data = simdata)
  ss <- summary(as.lmitt(mod, spec))
  expect_true(is(ss, "summary.teeMod"))
})

test_that("print.summary isn't confused by bad naming", {
  data(simdata)
  simdata$abz.c <- as.factor(simdata$o)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)

  mod <- lmitt(y ~ abz.c, data = simdata, specification = spec)
  expect_warning(co <- capture.output(summary(mod)), "sufficient degrees of freedom")
  cos <- strsplit(trimws(co), "[[:space:]]+")

  expect_true(!any(grepl("^abz\\.", co)))
  expect_equal(sum(grepl("^z\\.", co)), 4)

  # to force ` in variable names via as.factor
  mod <- lmitt(y ~ as.factor(abz.c), data = simdata, specification = spec)
  expect_warning(co <- capture.output(summary(mod)), "sufficient degrees of freedom")
  expect_true(!any(grepl("^`abz\\.", co)))
  expect_equal(sum(grepl("^`z\\.", co)), 4)

})

save_options <- options()
options("propertee_message_on_unused_blocks" = FALSE)

test_that("lmitt", {

  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa2, uoa1) + block(bid) + forcing(force),
                   data = simdata)

  mod <- lmitt(y ~ 1, weights = ate(), data = simdata, specification = spec)

  expect_s4_class(mod, "teeMod")
  expect_true(inherits(mod, "lm"))

  mod_ett <- lmitt(y ~ 1, weights = ett(), data = simdata, specification = spec)

  expect_s4_class(mod_ett, "teeMod")
})

test_that("lmitt and lm return the same in simple cases", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid),
                   data = simdata)

  ml <- lm(y ~ assigned(), data = simdata, weights = ate(spec))
  ssmod <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)
  ml2ssmod <- lmitt(ml)

  expect_true(all.equal(ssmod$coef[1:2], ml$coef, check.attributes = FALSE))
  expect_true(all.equal(ssmod$coef, ml2ssmod$coef, check.attributes = FALSE))
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(ssmod$model$`(weights)`, slot)),
                           deparse1(methods::slot(ml$model$`(weights)`, slot)))
               } else {
                 identical(methods::slot(ssmod$model$`(weights)`, slot),
                           methods::slot(ml$model$`(weights)`, slot))
               },
               logical(1L)))
  )
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(ssmod$model$`(weights)`, slot)),
                           deparse1(methods::slot(ml2ssmod$model$`(weights)`, slot)))
               } else {
                 identical(methods::slot(ssmod$model$`(weights)`, slot),
                           methods::slot(ml2ssmod$model$`(weights)`, slot))
               },
               logical(1L)))
  )

  ml <- lm(y ~ z, data = simdata, weights = ett(spec))
  ssmod <- lmitt(y ~ 1, data = simdata, weights = ett(), specification = spec)

  expect_true(all.equal(ssmod$coef[1:2], ml$coef, check.attributes = FALSE))
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(ssmod$model$`(weights)`, slot)),
                           deparse1(methods::slot(ml$model$`(weights)`, slot)))
               } else {
                 identical(methods::slot(ssmod$model$`(weights)`, slot),
                           methods::slot(ml$model$`(weights)`, slot))
               },
               logical(1L)))
  )

})

test_that("covariate adjustment", {
  data(simdata)
  spec <- rd_spec(z ~ cluster(uoa2, uoa1) + block(bid) + forcing(force),
                   data = simdata)

  camod <- lm(y ~ x, data = simdata)

  ssmod <- lmitt(y ~ 1, data = simdata, weights = ate(),
              offset = cov_adj(camod), specification = spec)
  expect_true(!is.null(ssmod$model$"(offset)"))
  expect_true(inherits(ssmod$model$"(offset)", "SandwichLayer"))

})


test_that("StudySpecification argument", {
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec)
  mod2 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)
  mod3 <- lmitt(y ~ 1, data = simdata, specification = z ~ cluster(uoa1, uoa2))
  expect_true(mod3@StudySpecification@type == "Obs")
  expect_identical(mod1@StudySpecification, mod2@StudySpecification)
  expect_identical(mod1@StudySpecification@structure, mod3@StudySpecification@structure)
  expect_identical(mod1@StudySpecification@unit_of_assignment_type, mod3@StudySpecification@unit_of_assignment_type)
  expect_identical(deparse1(mod1@StudySpecification@call), deparse1(mod3@StudySpecification@call))

  spec2 <- rd_spec(z ~ cluster(uoa1, uoa2) + forcing(force), data = simdata)

  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec2)
  mod2 <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec2)
  mod3 <- lmitt(y ~ 1, data = simdata,
                specification = z ~ cluster(uoa1, uoa2) + forcing(force))
  expect_true(mod3@StudySpecification@type == "RD")
  expect_identical(mod1@StudySpecification, mod2@StudySpecification)
  expect_identical(mod1@StudySpecification@structure, mod3@StudySpecification@structure)
  expect_identical(mod1@StudySpecification@unit_of_assignment_type, mod3@StudySpecification@unit_of_assignment_type)
  expect_identical(deparse1(mod1@StudySpecification@call), deparse1(mod3@StudySpecification@call))

  spec3 <- obs_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec3,
                subset = simdata$dose < 300)
  mod2 <- lmitt(y ~ 1, data = simdata, weights = ate(),
                subset = simdata$dose < 300, specification = spec3)
  mod3 <- lmitt(y ~ 1, data = simdata,
                specification = z ~ cluster(uoa1, uoa2) + block(bid),
                subset = simdata$dose < 300)
  expect_true(mod3@StudySpecification@type == "Obs")
  expect_identical(mod1@StudySpecification, mod2@StudySpecification)
  expect_identical(mod1@StudySpecification@structure, mod3@StudySpecification@structure)
  expect_identical(mod1@StudySpecification@unit_of_assignment_type, mod3@StudySpecification@unit_of_assignment_type)
  expect_identical(deparse1(mod1@StudySpecification@call), deparse1(mod3@StudySpecification@call))

  expect_error(lmitt(y ~ 1, data = simdata, specification = 4),
               "formula specifying such a specification")

})

test_that("Dichotomy argument", {

  data(simdata)
  spec <- obs_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec, dichotomy = dose > 200 ~ .)
  mod2 <- lmitt(y ~ 1, data = simdata,
                specification = dose ~ cluster(uoa1, uoa2),
                dichotomy = dose > 200 ~ .)
  expect_identical(mod1@StudySpecification@structure, mod2@StudySpecification@structure)
  expect_identical(mod1@StudySpecification@unit_of_assignment_type, mod2@StudySpecification@unit_of_assignment_type)
  expect_identical(deparse1(mod1@StudySpecification@call), deparse1(mod2@StudySpecification@call))
  expect_true(all.equal(mod1$coefficients, mod2$coefficients,
                        check.attributes =  FALSE))

})

test_that("Dichotomy from weights", {
  data(simdata)
  spec <- obs_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, specification = spec,
                weights = ate(dichotomy = dose > 200 ~ .))
  mod2 <- lmitt(y ~ 1, data = simdata, specification = spec, weights = "ate",
                dichotomy = dose > 200 ~ .)
  mod3 <- lmitt(y ~ 1, data = simdata, specification = spec, weights = ate(spec),
                dichotomy = dose > 200 ~ .)
  expect_identical(mod1@StudySpecification, mod2@StudySpecification)
  expect_identical(mod1@StudySpecification, mod3@StudySpecification)
  expect_true(all.equal(mod1$coefficients, mod2$coefficients,
                        check.attributes =  FALSE))
  expect_true(all.equal(mod1$coefficients, mod3$coefficients,
                        check.attributes =  FALSE))
  expect_true(
    all.equal(mod2$model$`(weights)`, ate(spec, data = simdata, dichotomy = dose > 200 ~ .))
  )
  expect_true(
    all.equal(mod3$model$`(weights)`, ate(spec, data = simdata, dichotomy = dose > 200 ~ .))
  )
})

test_that("Minimal support for continuous treatments", {

  data(simdata)
  spec1 <- obs_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  expect_error(mod1 <- lmitt(y ~ 1, data = simdata, specification = spec1, absorb=FALSE),
               "continuous treatment")
  spec2 <- obs_spec(dose ~ cluster(uoa1, uoa2) +block(bid), data = simdata)
  expect_error(mod2 <- lmitt(y ~ 1, data = simdata, specification = spec2, absorb=TRUE),
               "continuous treatment")

  ## But we don't also allow factor or ordinal treatments
  simdata_ <-
      transform(simdata, dosef=cut(dose, breaks=c(50, 100, 200, 300),
                          labels=c("lo", "med", "hi"),
                          include.lowest=TRUE)
                )
  spec3 <-
      obs_spec(dosef ~ cluster(uoa1, uoa2), data = simdata_)
  expect_error(lmitt(y ~ 1, data = simdata, specification = spec3),
               "not supported")
})

test_that("control means for intercept only", {
  set.seed(93)
  pairs <- data.frame(b = c(rep(seq_len(2), each = 3), rep(3, 2)),
                      new_trt = rep(c(1, 0), 4),
                      x = rnorm(8),
                      dv = rnorm(8),
                      id = seq_len(8))
  spec <- rct_spec(new_trt ~ block(b) + unitid(id), pairs)
  
  covadj_data <- data.frame(dv = rnorm(30), x = rnorm(30), id = NA, b = NA)
  cmod <- lm(dv ~ x, covadj_data)
  cad <- cov_adj(cmod, newdata = pairs, spec = spec)
  
  # lmitt
  suppressMessages(m1 <- lmitt(dv ~ 1, spec, pairs))
  suppressMessages(m2 <- lmitt(lm(dv ~ z.(spec, pairs), pairs), spec))
  expect_equal(length(m1$coef), 3)
  expect_equal(m1$coef[3], c("dv:(Intercept)" = mean(pairs$dv[pairs$new_trt == 0])))
  expect_equal(m2$coef[3], m1$coef[3])

  # cov adj
  suppressMessages(m3 <- lmitt(dv ~ 1, spec, pairs, offset = cad))
  suppressMessages(m4 <- lmitt(lm(dv ~ a.(spec, pairs), pairs, offset = cad), spec))
  expect_equal(length(m3$coef), 4)
  expect_equal(m3$coef[3:4],
               c("dv:(Intercept)" = mean(pairs$dv[pairs$new_trt == 0]),
                 "cov_adj:(Intercept)" = mean(cad[pairs$new_trt == 0])))
  expect_equal(m3$coef[3:4], m4$coef[3:4])
  
  # case weights
  cw <- rep(c(10, 12), each = 4) 
  suppressMessages(m5 <- lmitt(dv ~ 1, spec, pairs, w = cw, offset = cad))
  suppressMessages(m6 <- lmitt(lm(dv ~ adopters(spec, pairs), pairs, w = cw, off = cad), spec))
  expect_equal(m5$coef[3:4],
               c("dv:(Intercept)" = sum(cw * pairs$dv * (1-pairs$new_trt)) / sum(cw * (1-pairs$new_trt)),
                 "cov_adj:(Intercept)" = sum(cw * cad * (1-pairs$new_trt)) / sum(cw * (1-pairs$new_trt))))
  expect_equal(m5$coef[3:4], m6$coef[3:4])
  
  # absorb = TRUE
  m7 <- lmitt(dv ~ 1, spec, pairs, offset = cad, absorb = TRUE)
  pis <- c(rep(2/3, 3), rep(1/3, 3), rep(1/2, 2))
  expect_equal(m7$coef[3:4],
               c("dv:(Intercept)" = sum(pis * pairs$dv * (1-pairs$new_trt)) / sum(pis * (1-pairs$new_trt)),
                 "cov_adj:(Intercept)" = sum(pis * cad * (1-pairs$new_trt)) / sum(pis * (1-pairs$new_trt))))
  
  # case weights and absorb = TRUE
  m8 <- lmitt(dv ~ 1, spec, pairs, w = cw, offset = cad, absorb = TRUE)
  cw_pis <- cw * (rowsum(cw * pairs$new_trt, pairs$b) / rowsum(cw, pairs$b))[pairs$b]
  expect_equal(m8$coef[3:4],
               c("dv:(Intercept)" = sum(cw_pis * pairs$dv * (1-pairs$new_trt)) / sum(cw_pis * (1-pairs$new_trt)),
                 "cov_adj:(Intercept)" = sum(cw_pis * cad * (1-pairs$new_trt)) / sum(cw_pis * (1-pairs$new_trt))))

  # missing data--stratum where trt is missing outcome still contributes to ctrl mean estimation
  pairs <- data.frame(b = c(rep(seq_len(2), each = 3), rep(3:4, each = 2)),
                      new_trt = rep(c(1, 0), 5),
                      x = rnorm(10),
                      dv = c(rnorm(8), NA_real_, rnorm(1)),
                      id = seq_len(10))
  spec <- rct_spec(new_trt ~ block(b) + unitid(id), pairs)
  m9 <- lmitt(dv ~ 1, spec, pairs, absorb = TRUE)
  pis <- c(pis, rep(1/2, 2))
  expect_equal(
    m9$coef[3],
    c("dv:(Intercept)" = sum(pis * pairs$dv * (1-pairs$new_trt), na.rm = TRUE) /
        sum(pis * (1-pairs$new_trt), na.rm = TRUE)))
})

test_that("control means for continuous moderator", {
  set.seed(93)
  pairs <- data.frame(b = c(rep(seq_len(2), each = 3), rep(3, 2)),
                      new_trt = rep(c(1, 0), 4),
                      x = rnorm(8),
                      mvar = sample(3, 8, replace = TRUE),
                      dv = rnorm(8),
                      id = seq_len(8),
                      pis = c(rep(2/3, 3), rep(1/3, 3), rep(1/2, 2)),
                      cw = rep(c(10, 12), each = 4))
  spec <- rct_spec(new_trt ~ block(b) + unitid(id), pairs)
  
  covadj_data <- data.frame(dv = rnorm(30), x = rnorm(30), id = NA, b = NA)
  cmod <- lm(dv ~ x, covadj_data)
  cad <- cov_adj(cmod, newdata = pairs, spec = spec)
  
  unwtd.ctrl.reg <- lm(cbind(dv, cad) ~ mvar, pairs, w = 1 - new_trt)
  cw.ctrl.reg <- lm(cbind(dv, cad) ~ mvar, pairs, w = cw * (1 - new_trt))
  pis.ctrl.reg <- lm(cbind(dv, cad) ~ mvar, pairs, w = pis * (1 - new_trt))
  cw_pis <- (rowsum(pairs$cw * pairs$new_trt, pairs$b) / rowsum(pairs$cw, pairs$b))[pairs$b]
  cw_pis.ctrl.reg <- lm(cbind(dv, cad) ~ mvar, pairs, w = pairs$cw * cw_pis * (1 - new_trt))
  
  # no weights
  suppressMessages(m1 <- lmitt(dv ~ mvar, spec, pairs))
  expect_equal(length(m1$coef), 6)
  expect_equal(
    m1$coef[5:6],
    setNames(unwtd.ctrl.reg$coef[,1], c("dv:(Intercept)", "dv:mvar"))
  )
  
  # cov adj
  suppressMessages(m2 <- lmitt(dv ~ mvar, spec, pairs, offset = cad))
  expect_equal(length(m2$coef), 8)
  expect_equal(
    m2$coef[5:8],
    setNames(c(unwtd.ctrl.reg$coef),
             c("dv:(Intercept)", "dv:mvar", "cov_adj:(Intercept)", "cov_adj:mvar"))
  )
  
  # case weights
  suppressMessages(m3 <- lmitt(dv ~ mvar, spec, pairs, weights = pairs$cw, offset = cad))
  expect_equal(length(m3$coef), 8)
  expect_equal(
    m3$coef[5:8],
    setNames(c(cw.ctrl.reg$coef),
             c("dv:(Intercept)", "dv:mvar", "cov_adj:(Intercept)", "cov_adj:mvar"))
  )
  
  # absorb = TRUE
  m4 <- lmitt(dv ~ mvar, spec, pairs, offset = cad, absorb = TRUE)
  expect_equal(length(m4$coef), 8)
  expect_equal(
    m4$coef[5:8],
    setNames(c(pis.ctrl.reg$coef),
             c("dv:(Intercept)", "dv:mvar", "cov_adj:(Intercept)", "cov_adj:mvar"))
  )
  
  # case weights and absorb = TRUE
  m5 <- lmitt(dv ~ mvar, spec, pairs, weights = pairs$cw, offset = cad, absorb = TRUE)
  expect_equal(length(m5$coef), 8)
  expect_equal(
    m5$coef[5:8],
    setNames(c(cw_pis.ctrl.reg$coef),
             c("dv:(Intercept)", "dv:mvar", "cov_adj:(Intercept)", "cov_adj:mvar"))
  )
})

test_that("control means for categorical moderator", {
  set.seed(93)
  # pairs <- data.frame(b = c(rep(seq_len(2), each = 3), rep(3, 2)),
  #                     new_trt = rep(c(1, 0), 4),
  #                     x = rnorm(8),
  #                     mvar = letters[rep(seq_len(3), 3)[-9]],
  #                     dv = rnorm(8),
  #                     id = seq_len(8),
  #                     pis = c(rep(2/3, 3), rep(1/3, 3), rep(1/2, 2)),
  #                     cw = rep(c(10, 12), each = 4))
  pairs <- data.frame(b = c(rep(seq_len(2), each = 7), rep(3, 10)),
                      new_trt = rep(c(1, 0), 12),
                      x = rnorm(24),
                      mvar = c(rep("a", 3), rep("b", 4), rep("a", 7), rep(c("a", "b"), each = 5)),
                      dv = rnorm(24),
                      id = seq_len(24),
                      pis = c(rep(2/3, 3), rep(1/2, 4), rep(3/7, 7), rep(3/5, 5), rep(2/5, 5)),
                      cw = rep(c(10, 12), each = 12))
  spec <- rct_spec(new_trt ~ block(b) + unitid(id), pairs)
  
  covadj_data <- data.frame(dv = rnorm(30), x = rnorm(30), id = NA, b = NA)
  cmod <- lm(dv ~ x, covadj_data)
  cad <- cov_adj(cmod, newdata = pairs, spec = spec)
  
  unwtd.ctrl.reg <- lm(cbind(dv, cad) ~ 0+mvar, pairs, w = 1 - new_trt)
  cw.ctrl.reg <- lm(cbind(dv, cad) ~ 0+mvar, pairs, w = cw * (1 - new_trt))
  pis.ctrl.reg <- lm(cbind(dv, cad) ~ 0+mvar, pairs, w = pis * (1 - new_trt))
  cw_pis <- (rowsum(pairs$cw * pairs$new_trt, paste(pairs$b, pairs$mvar, sep = "_")) /
               rowsum(pairs$cw, paste(pairs$b, pairs$mvar, sep = "_")))[paste(pairs$b, pairs$mvar, sep = "_"),]
  cw_pis.ctrl.reg <- lm(cbind(dv, cad) ~ 0+mvar, pairs, w = pairs$cw * cw_pis * (1 - new_trt))
  
  # no weights
  suppressMessages(m1 <- lmitt(dv ~ mvar, spec, pairs))
  expect_equal(length(m1$coef), 6)
  expect_equal(
    m1$coef[5:6],
    setNames(unwtd.ctrl.reg$coef[,1], c("dv:mvara", "dv:mvarb"))
  )
  
  # cov adj
  suppressMessages(m2 <- lmitt(dv ~ mvar, spec, pairs, offset = cad))
  expect_equal(length(m2$coef), 8)
  expect_equal(
    m2$coef[5:8],
    setNames(c(unwtd.ctrl.reg$coef),
             c("dv:mvara", "dv:mvarb", "cov_adj:mvara", "cov_adj:mvarb"))
  )
  
  # case weights
  suppressMessages(m3 <- lmitt(dv ~ mvar, spec, pairs, weights = pairs$cw, offset = cad))
  expect_equal(length(m3$coef), 8)
  expect_equal(
    m3$coef[5:8],
    setNames(c(cw.ctrl.reg$coef),
             c("dv:mvara", "dv:mvarb", "cov_adj:mvara", "cov_adj:mvarb"))
  )
  
  # absorb = TRUE
  suppressMessages(m4 <- lmitt(dv ~ mvar, spec, pairs, offset = cad, absorb = TRUE))
  expect_equal(length(m4$coef), 8)
  expect_equal(
    m4$coef[5:8],
    setNames(c(pis.ctrl.reg$coef),
             c("dv:mvara", "dv:mvarb", "cov_adj:mvara", "cov_adj:mvarb"))
  )
  
  # case weights and absorb = TRUE
  suppressMessages(m5 <- lmitt(dv ~ mvar, spec, pairs, w = pairs$cw, offset = cad, absorb = TRUE))
  expect_equal(length(m5$coef), 8)
  expect_equal(
    m5$coef[5:8],
    setNames(c(cw_pis.ctrl.reg$coef),
             c("dv:mvara", "dv:mvarb", "cov_adj:mvara", "cov_adj:mvarb"))
  )
})

test_that("control means for dichotomized treatment", {
  set.seed(93)
  strata <- data.frame(b = rep(seq_len(2), each = 10),
                       trt2dich = letters[rep(rep(seq_len(3), 4)[1:10], 2)],
                       x = rnorm(20),
                       dv = rnorm(20),
                       id = seq_len(20))
  spec <- rct_spec(trt2dich ~ block(b) + unitid(id), strata)
  
  covadj_data <- data.frame(dv = rnorm(30), x = rnorm(30), id = NA, b = NA)
  cmod <- lm(dv ~ x, covadj_data)
  expect_warning(cad <- cov_adj(cmod, newdata = strata, spec = spec),
                 "treatment is an independent variable")
  
  suppressMessages(m1 <- lmitt(dv ~ 1, spec, strata, offset = cad,
                               dichotomy = trt2dich %in% c("a", "b") ~ trt2dich == "c"))
  expect_equal(length(m1$coef), 4)
  expect_equal(
    m1$coef[3:4],
    c("dv:(Intercept)" = mean(strata$dv[strata$trt2dich == "c"]),
      "cov_adj:(Intercept)" = mean(cad[strata$trt2dich == "c"]))
  )
  
  m2 <- lmitt(dv ~ 1, spec, strata, offset = cad, absorb = TRUE,
              dichotomy = trt2dich %in% c("a", "b") ~ trt2dich == "c")
  pis <- tapply(strata$trt2dich %in% c("a", "b"), strata$b, mean)[strata$b]
  expect_equal(length(m2$coef), 4)
  expect_equal(
    m2$coef[3:4],
    c("dv:(Intercept)" = sum(pis * strata$dv * (strata$trt2dich == "c")) / sum(pis * (strata$trt2dich == "c")),
      "cov_adj:(Intercept)" = sum(pis * cad * (strata$trt2dich == "c")) / sum(pis * (strata$trt2dich == "c")))
  )
})

test_that("lmitt finds StudySpecification wherever it's stored", {
  data(simdata)
  spec_form <- z ~ uoa(uoa1, uoa2) + block(bid)
  spec <- obs_spec(spec_form, data = simdata)
  camod <- lm(y ~ x, data = simdata)

  mod1 <- lmitt(y ~ 1, data = simdata, weights = ate(),
                offset = cov_adj(camod, specification = spec),
                specification = spec)
  mod2 <- lmitt(y ~ 1, specification = spec_form, data = simdata, weights = ate(),
                offset = cov_adj(camod))
  mod3 <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(),
                offset = cov_adj(camod))

  expect_equal(coef(mod1), coef(mod2))
  expect_equal(coef(mod1), coef(mod3))
  expect_equal(vcov_tee(mod1), vcov_tee(mod2))
  expect_equal(vcov_tee(mod1), vcov_tee(mod3))
})

test_that("Allowed inputs to lmitt #73", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  ### Allowed versions

  expect_no_error(l1 <- lmitt(y ~ 1, specification = spec, data = simdata))
  expect_no_error(l2 <- lmitt(y ~ dose, specification = spec, data = simdata))

  ### Disallowed versions

  expect_error(lmitt(y ~ 0, specification = spec, data = simdata))
  expect_error(lmitt(y ~ dose + 0, specification = spec, data = simdata))
  expect_error(lmitt(y ~ 0 + dose, specification = spec, data = simdata))
  expect_error(lmitt(y ~ dose + bid, specification = spec, data = simdata))
  expect_error(lmitt(y ~ 1 + dose + bid, specification = spec, data = simdata))
  expect_error(lmitt(y ~ 1 + x, specification = spec, data = simdata))
  expect_error(lmitt(y ~ x + 1, specification = spec, data = simdata))

  ### Main + Subgroup effects.
  ### NYI - see #73
  expect_error(lmitt(y ~ 1 + dose, specification = spec, data = simdata),
               "To estimate subgroup effects") # Remove this once #73 is fixed

  #expect_no_error(l3 <- lmitt(y ~ 1 + dose, specification = spec, data = simdata))
  #expect_no_error(l4 <- lmitt(y ~ dose + 1, specification = spec, data = simdata))

  #expect_identical(l3$coeff, l4$coeff)

})

test_that("weights argument can be string", {

  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  l1 <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate())
  l2 <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ate")

  l3 <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ett())
  l4 <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ett")

  l1@lmitt_call <- call("ls")
  l2@lmitt_call <- call("ls")
  l3@lmitt_call <- call("ls")
  l4@lmitt_call <- call("ls")

  expect_true(all.equal(l1, l2))
  expect_true(all.equal(l3, l4))

  # all.equal returns strings detailing differences
  expect_true(is.character(all.equal(l1, l3)))

  # Passing a different string warns users, but allows `lm()` to error
  # in case user is trying to pass something to someplace else
  expect_error(expect_warning(lmitt(y ~ 1, specification = spec, data =simdata,
                                    weights = "abc"), "other than"))

})

test_that("Regular weights can still be used", {

  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  mod <- lmitt(y ~ 1, specification = spec, weights = simdata$dose, data = simdata)
  expect_true(is(mod, "teeMod"))

})

test_that("Aliases for assigned() aren't allowed either", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  expect_error(lmitt(y ~ assigned(), specification = spec, data = simdata),
               "Do not specify")

  expect_error(lmitt(y ~ adopters(), specification = spec, data = simdata),
               "Do not specify")

  expect_error(lmitt(y ~ a.(), specification = spec, data = simdata),
               "Do not specify")


  expect_error(lmitt(y ~ z.(), specification = spec, data = simdata),
               "Do not specify")

  #### because of the `.` in `a.()`, let's make sure the regexp isn't breaking
  ad <- assigned
  # Passing in a different weight to ensure we get a predictable error that
  # ocurs *after* checking the formula
  spec2 <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  expect_error(lmitt(y ~ ad(), specification = spec, data = simdata,
                     weights = ate(spec2)),
               "Multiple differing")

  # Just to ensure that `lmitt.formula()` doesn't get changed in a way such that
  # checking for differing specifications happens prior:
  expect_error(lmitt(y ~ a.(), specification = spec, data = simdata,
                     weights = ate(spec2)),
               "Do not specify")

  rm(ad)

})

test_that("User can pass WeightedStudySpecification", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  mod <- lmitt(y ~ 1, data = simdata, specification = ate(spec))

  expect_identical(spec, mod@StudySpecification)

})

test_that("#82 Informative error if specification in weight/offset but not lmitt", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  expect_error(lmitt(y ~ 1, data = simdata, weights = ate(spec)),
               "into the weight function")

  cmod <- lm(y ~ x, data = simdata)
  expect_error(lmitt(y ~ 1, data = simdata, offset = cov_adj(cmod, specification = spec)),
               "into the offset function")

})

test_that("#81 continuous moderator", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) , data = simdata)

  mod <- lmitt(y ~ x, data = simdata, specification = spec)
  expect_length(mod$coeff, 6)

  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  mod <- lmitt(y ~ x, data = simdata, specification = spec, absorb = TRUE)
  expect_length(mod$coeff, 6)

})

test_that("non-integer units of assignment", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid) , data = simdata)

  expect_no_error(lmitt(y ~ x, data = simdata, specification = spec, absorb = TRUE))

  simdata$uoa1 <- as.character(simdata$uoa1)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid) , data = simdata)

  expect_no_error(lmitt(y ~ x, data = simdata, specification = spec, absorb = TRUE))

  simdata$uoa1 <- as.factor(simdata$uoa1)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid) , data = simdata)

  expect_no_error(lmitt(y ~ x, data = simdata, specification = spec, absorb = TRUE))

  expect_true(TRUE) # Avoid an empty test warning

})

test_that("#137 ensure absorb is using the correct moderator", {
  data(simdata)

  mod1 <- lmitt(y ~ o, data = simdata,
                specification = z ~ cluster(uoa1, uoa2) + block(bid))
  mod2 <- lmitt(y ~ o, data = simdata,
                specification = z ~ cluster(uoa1, uoa2) + block(bid),
                absorb = TRUE)

  o1 <- model.matrix(mod1)[, "o"]
  o2 <- model.matrix(mod2)[, "o"]

  # o1 and o2 should be different, since the latter should have been
  # group-centered due to `absorb = TRUE`. Due to a bug discovered in #137, the
  # un-centered version of the continuous moderator was being using previously,
  # leading o1 and o2 to be identical (incorrectly).

  expect_true(!(all.equal(o1, o2, check.attributes = FALSE) == TRUE))


})

test_that("#140 handling 0 weights", {

  # A single 0 weight observation in a larger block - that observation's
  # conbribution to the estimating equation should be 0
  data(simdata)
  simdata$weight <- 1
  simdata$weight[1] <- 0

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  mod <- lmitt(y ~ x, data = simdata, specification = spec,
               weights = simdata$weight)
  ee <- propertee:::estfun.teeMod(mod)
  expect_true(all(ee[1, ] == 0))

  # A block with only a single non-zero weight should contribute nothing to the
  # estimating equation
  data(simdata)
  simdata$weight <- 1
  simdata$weight[simdata$bid == 1] <- 0
  simdata$weight[simdata$bid == 1][1] <- 1

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  mod <- lmitt(y ~ x, data = simdata, specification = spec,
               absorb = TRUE, weights = simdata$weight)
  ee <- propertee:::estfun.teeMod(mod)
  expect_true(all(ee[simdata$bid == 1, 2:ncol(ee)] == 0))
  expect_true(all(ee[simdata$bid == 1 & simdata$weight == 0, 1] == 0))

  data(simdata)
  simdata$z[simdata$bid == 1] <- 1
  simdata$weight <- 1
  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  # Block 1 has 0 variance in treatment
  mod <- lmitt(y ~ x, data = simdata, specification = spec,
               absorb = TRUE, weights = simdata$weight)
  ee <- estfun.teeMod(mod)
  expect_true(all(ee[simdata$bid == 1, "z."] == 0))
  expect_true(all(
    sapply(ee[simdata$bid == 1, c("(Intercept)", "x", "z._x")],
           function(vals) any(vals != 0))
  ))

})


options(save_options)
#### !!!!!!!!!!!NOTE!!!!!!!!!!!!!
# This test below should NOT have `options()$propertee_message_on_unused_blocks`
# set to FALSE. So it needs to stay below the restoration of options line above.
# Other tests should probably go above the restoration of options line.

test_that("Message if specification has block info but isn't used in lmitt", {
  data(simdata)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  expect_message(lmitt(y ~ 1, data = simdata, specification = spec),
                 "is not used in this model")
  expect_message(lmitt(y ~ 1, data = simdata, specification = ate(spec)),
                 "is not used in this model")
  expect_message(lmitt(y ~ 1, data = simdata, specification = ate(spec),
                       weights = abs(simdata$x)),
                 "is not used in this model")

  expect_silent(x <- lmitt(y ~ 1, data = simdata, absorb = TRUE, specification = spec))
  expect_silent(x <- lmitt(y ~ 1, data = simdata, weights = "ate", specification = spec))
  expect_silent(x <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec))
  expect_silent(x <- lmitt(y ~ 1, data = simdata, weights = ett()*3, specification = spec))

})

### READ COMMENT ABOUT LAST TEST!

test_that(paste("absorb=TRUE doesn't drop rows when some strata have weights=0",
                "(some strata only have treated clusters)"), {
  data(simdata)
  simdata[simdata$uoa1 == 5 & simdata$uoa2 == 2, "bid"] <- 4

  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  absorb_mod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ate", absorb = TRUE)
  no_absorb_mod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ate", absorb = FALSE)

  expect_equal(dim(model.matrix(absorb_mod)), dim(model.matrix(no_absorb_mod)))
})

test_that(paste("absorb=TRUE doesn't drop rows when some strata have weights=0",
                "(some strata only have untreated clusters)"), {
  data(simdata)
  simdata[simdata$uoa1 == 4 & simdata$uoa2 == 2, "bid"] <- 4

  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  absorb_mod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ate", absorb = TRUE)
  no_absorb_mod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = "ate", absorb = FALSE)

  expect_equal(dim(model.matrix(absorb_mod)), dim(model.matrix(no_absorb_mod)))
})

test_that("#147 lmitt.formula accepts references to formula objects", {
  data(simdata)
  lmitt_form <- y ~ 1
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  expect_equal(
    co <- capture.output(lmitt(lmitt_form, data = simdata, specification = spec)),
    capture.output(lmitt(y ~ 1, data = simdata, specification = spec))
  )

  make_lmitt <- function(data, dv_col) {
    lmitt_form <- as.formula(paste0(dv_col, "~1"))
    spec <- rct_spec(z ~ unitid(uoa1, uoa2), data)
    lmitt(lmitt_form, specification = spec, data = data)
  }
  expect_equal(capture.output(make_lmitt(simdata, "y")), co)

})

test_that("get_structure() fn", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  ss <- get_structure(spec)

  expect_true(is(ss, "StudySpecificationStructure"))
  expect_identical(data.frame(ss), spec@structure)
  expect_identical(ss@StudySpecification, spec)

  ss2 <- get_structure(spec)
  # No dichotomy so should be identical
  expect_identical(data.frame(ss), data.frame(ss2))
  expect_identical(ss2@StudySpecification, spec)


  spec <- rct_spec(o ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  ss <- get_structure(spec)

  expect_true(is(ss, "StudySpecificationStructure"))
  expect_identical(data.frame(ss), spec@structure)
  expect_identical(ss@StudySpecification, spec)
})

test_that("show.StudySpecification.Structure", {

  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  ss <- get_structure(spec)
  expect_output(show(ss), "Treatment")
  expect_output(show(ss), "Cluster")
  expect_output(show(ss), "Block")

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)

  cap <- capture_output(show(get_structure(spec)))
  expect_match(cap, "Treatment")
  expect_match(cap, "Unit of Assignment", fixed = TRUE)
  expect_no_match(cap, "Block")

  spec <- rd_spec(z ~ unitid(uoa1, uoa2) + forcing(force), data = simdata)

  cap <- capture_output(show(get_structure(spec)))

  expect_match(cap, "Treatment")
  expect_match(cap, "Unit ID", fixed = TRUE)
  expect_no_match(cap, "Block")
  expect_match(cap, "Forcing")

})

test_that("treatments with `newdata`", {
  data(simdata)

  # Case 1: binary = FALSE
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid),
                    data = simdata)

  expect_true(all.equal(treatment(spec, newdata = simdata),
                        simdata[, "z", drop = FALSE]))

  # Case 2b: binary = TRUE, stored treatment is non-binary but has dichotomy
  spec <- rct_spec(o ~ cluster(uoa1, uoa2) + block(bid),
                    data = simdata)

  expect_equal(nrow(treatment(spec, dichotomy = o >= 3 ~ ., newdata = simdata)),
               50)
  expect_true(all(treatment(spec, dichotomy = o >= 3 ~ ., newdata = simdata)[,1] %in% 0:1))

})



test_that("units of assignment and clusters with `newdata`", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_true(all.equal(clusters(spec, newdata = simdata), simdata[, c("uoa1", "uoa2")]))

  clusters(spec) <- 1:10
  expect_length(clusters(spec, newdata = simdata)[, 1], 50)

  # This is only 5 since simdata doesn't have the same modification as two lines
  # above.
  expect_equal(length(unique(clusters(spec, newdata = simdata)[, 1])), 5)


  spec <- rd_spec(z ~ uoa(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_true(all.equal(units_of_assignment(spec, newdata = simdata), simdata[, c("uoa1", "uoa2")]))

  units_of_assignment(spec) <- 1:10
  expect_length(units_of_assignment(spec, newdata = simdata)[, 1], 50)

  # This is only 5 since simdata doesn't have the same modification as two lines
  # above.
  expect_equal(length(unique(units_of_assignment(spec, newdata = simdata)[, 1])), 5)

})

test_that("blocks with `newdata`", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_true(all.equal(blocks(spec, newdata = simdata), simdata[, "bid", drop = FALSE]))

  blocks(spec) <- 1:10
  expect_length(blocks(spec, newdata = simdata)[, 1], 50)
  expect_equal(length(unique(blocks(spec, newdata = simdata)[, 1])), 10)

  # More blocks
  df <- data.frame(bid1 = 1:10,
                   bid2 = 10:1)
  blocks(spec) <- df
  expect_identical(dim(blocks(spec, newdata = simdata)), c(50L, 2L))

  ### multi-dimensional blocks

  simdata2 <- simdata
  simdata2$uoa1 <- 1:50
  simdata2 <- rbind(simdata2, simdata2) # Duplicate rows to make data different
  names(simdata2)[1:2] <- c("cid", "bida")
  spec <- rd_spec(z ~ cluster(cid) + block(bid, bida) + forcing(force),
                   data = simdata2)

  expect_identical(dim(blocks(spec, newdata = simdata2)), c(100L, 2L))
})

test_that("forcing with `newdata`", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_true(all.equal(forcings(spec, newdata = simdata), simdata[, "force", drop = FALSE]))

  forcings(spec) <- 1:10
  expect_length(forcings(spec, newdata = simdata)[, 1], 50)
  expect_equal(length(unique(forcings(spec, newdata = simdata)[, 1])), 10)

  # More forcings
  df <- data.frame(f1 = 1:10,
                   f2 = 10:1)
  forcings(spec) <- df
  expect_identical(dim(forcings(spec, newdata = simdata)), c(50L, 2L))

  ### multi-dimensional forcings

  simdata2 <- simdata
  simdata2$force2 <- rep(rnorm(10, mean = 5), times = rep(c(4, 6), times = 5))
  simdata2 <- rbind(simdata2, simdata2) # Duplicate rows to make data different
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) +
                     forcing(force, force2), data = simdata2)


  expect_identical(dim(forcings(spec, newdata = simdata2)), c(100L, 2L))
})

test_that(".specification_accessors_newdata_validate", {
  expect_warning(.specification_accessors_newdata_validate(1, NULL))

  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  names(simdata)[names(simdata) == "uoa1"] <- "ccc"
  names(simdata)[names(simdata) == "force"] <- "fff"

  expect_true(.specification_accessors_newdata_validate(simdata,
                                                 by = list(force = "fff",
                                                           uoa1 = "ccc")))

})

test_that(".get_col_from_new_data", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  simdata2 <- simdata
  names(simdata2)[names(simdata2) == "uoa1"] <- "ccc"
  names(simdata2)[names(simdata2) == "force"] <- "fff"

  expect_identical(.get_col_from_new_data(spec, simdata, type = "b"),
                   .get_col_from_new_data(spec, simdata2, type = "b",
                                          by = list(force = "fff",
                                                    uoa1 = "ccc")))


})

old_opt <- options()
on.exit(options(old_opt))
options(propertee_warn_on_no_unit_of_assignment = FALSE)

test_that("Support lack of UOA", {
  data(simdata)
  # I was having weird issues with testthat when I worked with simdata directly;
  # so that's why I'm making a copy here - JE
  sd_ided <- simdata
  sd_ided$id <- rownames(sd_ided)

  spec1 <- rct_spec(z ~ uoa(id), data = sd_ided)
  spec2 <- rct_spec(z ~ 1, data = sd_ided)
  expect_true(all.equal(spec1@structure,
                        spec2@structure,
                        check.attributes = FALSE))
  expect_identical(spec1@column_index,
                   spec2@column_index)
  expect_equal(spec1@unit_of_assignment_type, "unit_of_assignment")
  expect_equal(spec2@unit_of_assignment_type, "none")

  mod1 <- lmitt(y ~ 1, spec = spec1, data = sd_ided)
  mod2 <- lmitt(y ~ 1, spec = spec2, data = sd_ided)
  expect_true(isTRUE(all.equal(coefficients(mod1),
                               coefficients(mod2))))

  smod1 <- summary(mod1)
  smod2 <- summary(mod2)
  expect_true(isTRUE(all.equal(coefficients(smod1),
                               coefficients(smod2))))

  ## Shuffle
  sd_ided_shuffled <- sd_ided[sample(1:50), ]
  spec3 <- rct_spec(z ~ 1, data = sd_ided_shuffled)
  mod3 <- lmitt(y ~ 1, spec = spec3, data = sd_ided_shuffled)
  expect_true(isTRUE(all.equal(coefficients(mod1),
                               coefficients(mod3))))
  smod3 <- summary(mod3)
  expect_true(isTRUE(all.equal(coefficients(smod1),
                               coefficients(smod3))))


})

test_that("covadj", {
  data(simdata)
  sd_ided <- simdata
  sd_ided$id <- rownames(sd_ided)

  spec1 <- rct_spec(z ~ uoa(id), data = sd_ided)
  spec2 <- rct_spec(z ~ 1, data = sd_ided)

  camod <- lm(y ~ x, data = sd_ided, subset = z == 0)

  mod1 <- lmitt(y ~ 1, spec = spec1, offset = cov_adj(camod), data = sd_ided)
  mod2 <- lmitt(y ~ 1, spec = spec2, offset = cov_adj(camod), data = sd_ided)
  expect_true(isTRUE(all.equal(coefficients(mod1),
                               coefficients(mod2))))

  smod1 <- summary(mod1)
  smod2 <- summary(mod2)
  expect_true(isTRUE(all.equal(coefficients(smod1),
                               coefficients(smod2))))



  data(STARplus)
  st <- STARplus
  st$cond_small <- st$cond_at_entry == "small"
  spec1 <- obs_spec(cond_small ~ unit_of_assignment(stdntid) +
                      block(school_at_entry),
                   data = st, na.fail = FALSE)
  ca <- lm(read_yr1 ~ dob + gender, data = st, subset = !(cond_small))

  mod1 <- lmitt(read_yr1 ~ 1, spec1, data = st, absorb = TRUE,
                                 offset = cov_adj(ca))

  spec2 <- obs_spec(cond_small ~ block(school_at_entry),
                   data = st, na.fail = FALSE)
  suppressWarnings(mod2 <- lmitt(read_yr1 ~ 1, spec2, absorb = TRUE, data = st,
                                 offset = cov_adj(ca)))

  expect_true(isTRUE(all.equal(coefficients(mod1),
                               coefficients(mod2))))
  smod1 <- summary(mod1, vcov.type = "CR0")
  smod2 <- summary(mod2, vcov.type = "CR0")
  expect_true(isTRUE(all.equal(coefficients(smod1),
                               coefficients(smod2))))
})


options(old_opt)

test_that("warning message", {
  data(simdata)
  # Running each a few times to ensure options are reset appropriately
  expect_warning(rct_spec(z ~ 1, data = simdata))
  expect_no_warning(lmitt(y ~ 1, spec = z ~ 1, data = simdata))
  expect_warning(rct_spec(z ~ 1, data = simdata))
  expect_no_warning(lmitt(y ~ 1, spec = z ~ 1, data = simdata))
  expect_warning(rct_spec(z ~ 1, data = simdata))
})

test_that("Combining weighted specifications", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)

  w1 <- ate(spec, data = simdata[1:30,])
  w2 <- ate(spec, data = simdata[31:40,])
  w3 <- ate(spec, data = simdata[41:50,])

  c_w <- c(w1, w2, w3)
  expect_true(inherits(c_w, "numeric"))
  expect_length(c_w, 50)

  w1e <- ett(spec, data = simdata[1:30,])
  w2e <- ett(spec, data = simdata[31:40,])
  w3e <- ett(spec, data = simdata[41:50,])

  c_we <- c(w1e, w2e, w3e)
  expect_true(inherits(c_we, "numeric"))
  expect_length(c_we, 50)

  expect_message(c(w1, 1:5), "with a non-WeightedStudySpecification")
  expect_error(c(w1, w1e), "same target")

  spec2 <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  alt_w1 <- ate(spec2, data = simdata)

  expect_error(c(w1, alt_w1), "differing StudySpecification")

  # if the first argument is compatible with WeightedStudySpecification but isn't one (e.g.
  # numeric vector), c() will return a numeric vector
  #expect_true(inherits(c(1:5, w1), "WeightedStudySpecification"))
})

test_that("Combining weighted specifications with different dichotomys ", {
  spec <- rct_spec(dose ~ uoa(uoa1, uoa2), data = simdata)

  w1 <- ate(spec, data = simdata[1:10, ], dichotomy = dose >= 300 ~ .)
  w2 <- ate(spec, data = simdata[11:30, ], dichotomy = dose >= 200 ~ .)
  w3 <- ate(spec, data = simdata[31:50, ], dichotomy = dose >= 100 ~ .)

  c_w <- c(w1, w2, w3)
  expect_true(inherits(c_w, "numeric"))
  expect_length(c_w, 50)

  expect_warning(c(w1, w2, w3, warn_dichotomy_not_equal = TRUE),
                 "Concatenating")

  data_w  <- rbind(cbind(simdata[1:10, ], w=w1),
                   cbind(simdata[11:30, ], w=w2),
                   cbind(simdata[31:50, ], w=w3))
  expect_s4_class(data_w$w, "numeric")
  expect_true(inherits(data_w$w, "WeightedStudySpecification"))
})

test_that("Combine WeightedStudySpecifications & align weights with analysis data",{
    set.seed(202021)
    analysis_dat  <- data.frame(
        id = rep(letters[1:6], each = 2),
        year = ordered(rep(c("AY20", "AY21"), 6), levels=c("AY20", "AY21", ".")),
        y = rnorm(12)
    )
    specification_dat  <- data.frame(id=letters[1:6], blk=character(6),
                              year_trt = ordered(rep(c("AY21", "AY20", "."),
                                                     each = 2),
                                                 levels=c("AY20", "AY21", ".")
                                                 )
                              )
    specification_dat[   specification_dat$id %in% c("a", "e") , "blk"]  <- "A"
    specification_dat[ !(specification_dat$id %in% c("a", "e")), "blk"]  <- "B"
    spec  <- obs_spec(year_trt~uoa(id)+block(blk), data=specification_dat)

    w20  <- ett(spec, data=subset(analysis_dat,year=="AY20"),
                dichotomy= year_trt<="AY20" ~ .)
    w20[which(subset(analysis_dat,year=="AY20")$id %in% c("a", "e"))] |>
    as.numeric() |> expect_equal(rep(0,2))
    ## ...confirming that blocks without treatment variation
    ## receive 0 weight. This is also tested by the test that
    ## "#130 zero weights with non-varying treatment in a block",
    ## in test.WeightedStudySpecification, but here we go on to check that
    ## in the next year, when there is variation in treatment status,
    ## the same pair has nonzero weights:
    w21  <- ett(spec, data=subset(analysis_dat,year=="AY21"),
               dichotomy= year_trt<="AY21" ~ .)
    w21[which(subset(analysis_dat,year=="AY21")$id %in% c("a", "e"))] |>
    as.numeric() |> expect_equal(rep(1,2))

    w0 <- c(w20, w21)
    expect_length(w0, length(w20)+length(w21))
    expect_true(inherits(w0, "numeric"))
    expect_false(inherits(w0, "WeightedStudySpecification"))

### Bringing the weights back into the data frame is easier
### if you're happy to reorder the data.
    mf_dat <- cbind(analysis_dat[order(analysis_dat$year),], w0)
    expect_identical(mf_dat$w0, w0)

### Users might try the following to put w0 into the same order as the
### rows of analysis_dat.
    analysis_dat$w0 <- numeric(12)
    analysis_dat[analysis_dat$year=="AY20","w0"] <- w20
    analysis_dat[analysis_dat$year=="AY21","w0"] <- w21
    expect_true(inherits(analysis_dat$w0, "numeric"))
    expect_false(inherits(analysis_dat$w0, "WeightedStudySpecification"))

    analysis_dat$w0 <- w0
    expect_true(inherits(analysis_dat$w0, "numeric"))
    analysis_dat[which(analysis_dat$year=="AY20"), "w0"]  <- w20
    analysis_dat[which(analysis_dat$year=="AY21"), "w0"]  <- w21
    expect_true(inherits(analysis_dat$w0, "numeric"))

### Alternatively, use lapply and unsplit:
    analysis_dat$w1 <-
        lapply(c("AY20", "AY21"),
    {\(yr) ett(spec, data=subset(analysis_dat,year==yr),
               dichotomy= as.formula(paste0("year_trt<=\"", yr, "\"~."))) }
    ) |> unsplit(analysis_dat$year)
    expect_s4_class(analysis_dat$w1, "WeightedStudySpecification")
    expect_equal(analysis_dat$w1@.Data, analysis_dat$w0@.Data)

### (One might hope to use tapply instead of lapply, but I couldn't find
###    a `dichotomy=` formula that's compatible with that mechanism.)
### Doesn't work:
###    tapply(analysis_dat, ~year,
###    {\(dat) ett(spec, data=dat, dichotomy = year_trt <= dat[1L, "year"] ~ .)}
###    )

})

test_that("Accessing and replacing treatment", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_equal(dim(treatment(spec)), c(10, 1))
  expect_true(is.numeric(treatment(spec)[, 1]))
  expect_identical(treatment(spec), spec@structure[, 1, drop = FALSE])

  treatment(spec) <- rep(0:1, each = 5)
  expect_equal(treatment(spec), data.frame(z = rep(0:1, each = 5)))

  df <- data.frame(z = rep(0:1, each = 5))
  treatment(spec) <- df
  expect_equal(treatment(spec), data.frame(z = rep(0:1, each = 5)))

  treatment(spec)[1:2, 1] <- 1
  expect_equal(treatment(spec),
               data.frame(z = rep(c(1, 0, 1),
                                  times = c(2, 3, 5))))

  expect_error(treatment(spec) <- 1:5,
               "same number")

  expect_error(treatment(spec) <- data.frame(a = c(1, 0, 1, 0, 1)),
               "same number")

  # Continuous treatment, no dichotomy
  spec <- rd_spec(dose ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(treatment(spec), spec@structure[, 1, drop = FALSE])

  # Continuous treatment, dichotomy
  spec <- rd_spec(dose ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(treatment(spec), spec@structure[, 1, drop = FALSE])
  tt <- treatment(spec, dichotomy = dose <= 100 ~ dose == 200)

  expect_equal(dim(tt), c(10, 1))
  expect_true(is.numeric(tt[, 1]))
  expect_true(all(tt[, 1] %in% c(0, 1, NA)))
  expect_equal(colnames(tt), "z__")

  ############ .bin_txt

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(treatment(spec)[, 1], .bin_txt(spec))

  spec <- rd_spec(o ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_true(all(.bin_txt(spec, dichotomy = o >= 3 ~ .) %in% c(0, 1, NA)))
  expect_identical(.bin_txt(spec, dichotomy = o >= 3 ~ .),
                   treatment(spec, dichotomy = o >= 3 ~ .)[, 1])

  spec <- rd_spec(o ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_error(.bin_txt(spec), "Must provide a dichotomy")

})

test_that("New .bin_txt, .expand_txt, and .apply_dichotomy errors", {
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)
  expect_error(.bin_txt(spec, data.frame("uoa1" = seq(100, 110))),
               "Not all unit of assignment variables")

  expect_error(.apply_dichotomy(simdata$dose, dose > 50 ~ dose == 50),
               "expected to be a named `data.frame`")
  expect_error(.apply_dichotomy(treatment(spec), dichotomy = dose > 50 ~ dose == 50),
               "Could not find variables specified in dichotomy")
  expect_error(.apply_dichotomy(simdata, . ~ .),
               "At least one side")
  expect_error(.apply_dichotomy(simdata, dichotomy = dose >= 50 ~ dose == 50),
               "dichotomy overlaps")
})

test_that("Accessing and replacing unit of assignment", {
  data(simdata)

  spec <- rd_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid) +
                     forcing(force),
                   data = simdata)

  expect_identical(units_of_assignment(spec), spec@structure[, 2:3])

  units_of_assignment(spec) <- data.frame(uoa1 = 10:1, uoa2 = 1:10)
  expect_identical(var_names(spec, "u"), c("uoa1", "uoa2"))
  expect_true(all(data.frame(uoa1 = 10:1, uoa2 = 1:10) ==
                    units_of_assignment(spec)))

  units_of_assignment(spec) <- data.frame(abc1 = 10:1, abc2 = 1:10)
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 10:1, abc2 = 1:10) ==
                    units_of_assignment(spec)))

  m <- matrix(1:20, ncol = 2)
  units_of_assignment(spec) <- m
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 1:10, abc2 = 11:20) ==
                    units_of_assignment(spec)))

  colnames(m) <- c("qwe", "asd")
  units_of_assignment(spec) <- m
  expect_identical(var_names(spec, "u"), colnames(m))
  expect_true(all(data.frame(qwe = 1:10, asd = 11:20) ==
                    units_of_assignment(spec)))

  units_of_assignment(spec)[1, 1:2] <- 100
  expect_true(all(data.frame(qwe = c(100, 2:10),
                             asd = c(100, 12:20)) ==
                    units_of_assignment(spec)))

  # less units_of_assignment
  spec2 <- spec
  units_of_assignment(spec2) <- m[, 1]
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 1:10) == units_of_assignment(spec2)))

  spec2 <- spec
  units_of_assignment(spec2) <- 10:1
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 10:1) == units_of_assignment(spec2)))

  df <- data.frame(abc = 3:12)
  spec2 <- spec
  units_of_assignment(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == units_of_assignment(spec2)))

  ########## UnitsOfAssignment, reduce duplicates

  treatment(spec) <- rep(0:1, each = 5)

  expect_warning(units_of_assignment(spec) <- data.frame(c1 = 1, c2 = c(1, 1:9)),
                 "collapsing")
  expect_equal(nrow(spec@structure), 9)

  expect_error(units_of_assignment(spec) <- data.frame(c1 = 1, c2 = c(1:8, 1)),
               "non-constant")

  # more units_of_assignment
  df <- data.frame(abc = 3:12, def = 4:13, efg = 5:14)
  spec <- rd_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid) +
                     forcing(force), data = simdata)
  spec2 <- spec
  units_of_assignment(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == units_of_assignment(spec2)))

  m <- matrix(1:40, ncol = 4)
  spec2 <- spec
  expect_error(units_of_assignment(spec2) <- m,
               "must be named")
  colnames(m) <- letters[1:4]
  units_of_assignment(spec2) <- m
  expect_identical(var_names(spec2, "u"), colnames(m))
  expect_true(all(m == units_of_assignment(spec2)))

  # uoa
  spec <- rd_spec(z ~ uoa(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)
  expect_identical(units_of_assignment(spec), spec@structure[, 2:3])

})

test_that("Accessing and replacing clusters", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(clusters(spec), spec@structure[, 2:3])

  clusters(spec) <- data.frame(uoa1 = 10:1, uoa2 = 1:10)
  expect_identical(var_names(spec, "u"), c("uoa1", "uoa2"))
  expect_true(all(data.frame(uoa1 = 10:1, uoa2 = 1:10) == clusters(spec)))

  clusters(spec) <- data.frame(abc1 = 10:1, abc2 = 1:10)
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 10:1, abc2 = 1:10) == clusters(spec)))

  m <- matrix(1:20, ncol = 2)
  clusters(spec) <- m
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 1:10, abc2 = 11:20) == clusters(spec)))

  colnames(m) <- c("qwe", "asd")
  clusters(spec) <- m
  expect_identical(var_names(spec, "u"), colnames(m))
  expect_true(all(data.frame(qwe = 1:10, asd = 11:20) == clusters(spec)))

  clusters(spec)[1, 1:2] <- 100
  expect_true(all(data.frame(qwe = c(100, 2:10),
                             asd = c(100, 12:20)) == clusters(spec)))

  # less clusters
  spec2 <- spec
  clusters(spec2) <- m[, 1]
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 1:10) == clusters(spec2)))

  spec2 <- spec
  clusters(spec2) <- 10:1
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 10:1) == clusters(spec2)))

  df <- data.frame(abc = 3:12)
  spec2 <- spec
  clusters(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == clusters(spec2)))

  ########## Clusters, reduce duplicates

  treatment(spec) <- rep(0:1, each = 5)

  expect_warning(clusters(spec) <- data.frame(c1 = 1, c2 = c(1, 1:9)),
                 "collapsing")
  expect_equal(nrow(spec@structure), 9)

  expect_error(clusters(spec) <- data.frame(c1 = 1, c2 = c(1:8, 1)),
               "non-constant")

  # more clusters
  df <- data.frame(abc = 3:12, def = 4:13, efg = 5:14)
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)
  spec2 <- spec
  clusters(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == clusters(spec2)))

  m <- matrix(1:40, ncol = 4)
  spec2 <- spec
  expect_error(clusters(spec2) <- m,
               "must be named")
  colnames(m) <- letters[1:4]
  clusters(spec2) <- m
  expect_identical(var_names(spec2, "u"), colnames(m))
  expect_true(all(m == clusters(spec2)))
})

test_that("Accessing and replacing unitids", {
  data(simdata)

  spec <- rd_spec(z ~ unitid(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(unitids(spec), spec@structure[, 2:3])

  unitids(spec) <- data.frame(uoa1 = 10:1, uoa2 = 1:10)
  expect_identical(var_names(spec, "u"), c("uoa1", "uoa2"))
  expect_true(all(data.frame(uoa1 = 10:1, uoa2 = 1:10) == unitids(spec)))

  unitids(spec) <- data.frame(abc1 = 10:1, abc2 = 1:10)
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 10:1, abc2 = 1:10) == unitids(spec)))

  m <- matrix(1:20, ncol = 2)
  unitids(spec) <- m
  expect_identical(var_names(spec, "u"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 1:10, abc2 = 11:20) == unitids(spec)))

  colnames(m) <- c("qwe", "asd")
  unitids(spec) <- m
  expect_identical(var_names(spec, "u"), colnames(m))
  expect_true(all(data.frame(qwe = 1:10, asd = 11:20) == unitids(spec)))

  unitids(spec)[1, 1:2] <- 100
  expect_true(all(data.frame(qwe = c(100, 2:10),
                             asd = c(100, 12:20)) == unitids(spec)))

  # less unitids
  spec2 <- spec
  unitids(spec2) <- m[, 1]
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 1:10) == unitids(spec2)))

  spec2 <- spec
  unitids(spec2) <- 10:1
  expect_identical(var_names(spec2, "u"), "qwe")
  expect_true(all(data.frame(qwe = 10:1) == unitids(spec2)))

  df <- data.frame(abc = 3:12)
  spec2 <- spec
  unitids(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == unitids(spec2)))

  ########## Unitids, reduce duplicates

  treatment(spec) <- rep(0:1, each = 5)

  expect_warning(unitids(spec) <- data.frame(c1 = 1, c2 = c(1, 1:9)),
                 "collapsing")
  expect_equal(nrow(spec@structure), 9)

  expect_error(unitids(spec) <- data.frame(c1 = 1, c2 = c(1:8, 1)),
               "non-constant")

  # more unitids
  df <- data.frame(abc = 3:12, def = 4:13, efg = 5:14)
  spec <- rd_spec(z ~ unitid(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)
  spec2 <- spec
  unitids(spec2) <- df
  expect_identical(var_names(spec2, "u"), colnames(df))
  expect_true(all(df == unitids(spec2)))

  m <- matrix(1:40, ncol = 4)
  spec2 <- spec
  expect_error(unitids(spec2) <- m,
               "must be named")
  colnames(m) <- letters[1:4]
  unitids(spec2) <- m
  expect_identical(var_names(spec2, "u"), colnames(m))
  expect_true(all(m == unitids(spec2)))

})

test_that("Accessing and replacing uoa vs unitid vs cluster", {
  data(simdata)

  spec <- rd_spec(z ~ unitid(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  # created with unitid
  expect_error(clusters(spec), "not `cluster")
  expect_error(units_of_assignment(spec), "not `unit_of_assignment")

  # created with cluster
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)
  expect_error(unitids(spec), "not `unitid")
  expect_error(units_of_assignment(spec), "not `unit_of_assignment")

  # created with unit_of_assignment
  spec <- rd_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid) +
                     forcing(force), data = simdata)
  expect_error(unitids(spec), "not `unitid")
  expect_error(clusters(spec), "not `cluster")

  # created with uoa
  spec <- rd_spec(z ~ uoa(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)
  expect_error(unitids(spec), "not `unitid")
  expect_error(clusters(spec), "not `cluster")
})

test_that("Accessing and replacing blocks", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(blocks(spec), spec@structure[, 4, drop = FALSE])

  blocks(spec) <- 1:10
  expect_equal(blocks(spec), data.frame(bid = 1:10))

  mat <- matrix(10:1)
  blocks(spec) <- mat
  expect_equal(blocks(spec), data.frame(bid = 10:1))

  mat <- matrix(1:10)
  colnames(mat) <- "bid"
  blocks(spec) <- mat
  expect_equal(blocks(spec), data.frame(bid = 1:10))

  mat <- matrix(10:1)
  colnames(mat) <- "abc"
  blocks(spec) <- mat
  expect_equal(blocks(spec), data.frame(abc = 10:1))

  df <- data.frame(bid = 1:10)
  blocks(spec) <- df
  expect_equal(blocks(spec), df)

  blocks(spec)[1:2, 1] <- 1
  expect_equal(blocks(spec), data.frame(bid = c(1, 1, 3:10)))

  # More blocks
  df <- data.frame(bid1 = 1:10,
                   bid2 = 10:1)
  blocks(spec) <- df
  expect_equal(blocks(spec), df)

  expect_error(blocks(spec) <- data.frame(a = 1:5),
               "same number")

  expect_error(blocks(spec) <- 1:5,
               "same number")


  ### multi-dimensional blocks

  simdata2 <- simdata
  simdata2$uoa1 <- 1:50
  names(simdata2)[1:2] <- c("cid", "bida")
  spec <- rd_spec(z ~ cluster(cid) + block(bid, bida) + forcing(force),
                   data = simdata2)

  expect_identical(blocks(spec), spec@structure[, 3:4])

  blocks(spec) <- data.frame(bid = 50:1, bida = 1:50)
  expect_identical(var_names(spec, "b"), c("bid", "bida"))
  expect_true(all(data.frame(bid = 50:1, bida = 1:50) == blocks(spec)))

  blocks(spec) <- data.frame(abc1 = 50:1, abc2 = 1:50)
  expect_identical(var_names(spec, "b"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 50:1, abc2 = 1:50) == blocks(spec)))

  m <- matrix(1:100, ncol = 2)
  blocks(spec) <- m
  expect_identical(var_names(spec, "b"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 1:50, abc2 = 51:100) == blocks(spec)))

  colnames(m) <- c("qwe", "asd")
  blocks(spec) <- m
  expect_identical(var_names(spec, "b"), c("qwe", "asd"))
  expect_true(all(data.frame(qwe = 1:50, asd = 51:100) == blocks(spec)))

  blocks(spec)[1, 1:2] <- 100
  expect_true(all(data.frame(qwe = c(100, 2:50),
                             asd = c(100, 52:100)) == blocks(spec)))

  # less blocks
  spec2 <- spec
  blocks(spec2) <- 1:50
  expect_identical(var_names(spec2, "b"), "qwe")
  expect_true(all(blocks(spec2) == 1:50))

  spec2 <- spec
  blocks(spec2) <- matrix(50:1, ncol = 1)
  expect_identical(var_names(spec2, "b"), "qwe")
  expect_true(all(blocks(spec2) == 50:1))

  spec2 <- spec
  df <- data.frame(bbb = 1:50)
  blocks(spec2) <- df
  expect_identical(var_names(spec2, "b"), "bbb")
  expect_true(all(blocks(spec2) == df))

  # more blocks

  spec2 <- spec
  df <- data.frame(bbb = 1:50, ccc = 50:1, ddd = 1:50)
  blocks(spec2) <- df
  expect_identical(var_names(spec2, "b"), colnames(df))
  expect_true(all(blocks(spec2) == df))
})


test_that("Accessing and replacing forcing", {
  data(simdata)

  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                   data = simdata)

  expect_identical(forcings(spec), spec@structure[, 5, drop = FALSE])

  forcings(spec) <- 1:10
  expect_equal(forcings(spec), data.frame(force = 1:10))

  mat <- matrix(10:1)
  forcings(spec) <- mat
  expect_equal(forcings(spec), data.frame(force = 10:1))

  mat <- matrix(1:10)
  colnames(mat) <- "fvar"
  forcings(spec) <- mat
  expect_equal(forcings(spec), data.frame(fvar = 1:10))

  mat <- matrix(10:1)
  colnames(mat) <- "abc"
  forcings(spec) <- mat
  expect_equal(forcings(spec), data.frame(abc = 10:1))

  df <- data.frame(fvar = 1:10)
  forcings(spec) <- df
  expect_equal(forcings(spec), data.frame(fvar = 1:10))

  forcings(spec)[1:2, 1] <- 1
  expect_equal(forcings(spec), data.frame(fvar = c(1, 1, 3:10)))

  expect_error(forcings(spec) <- data.frame(a = 1:5),
               "same number")

  # more forcing

  df <- data.frame(fvar = 1:10, fvar2 = 10:1)
  forcings(spec) <- df
  expect_equal(var_names(spec, "f"), colnames(df))
  expect_equal(forcings(spec), df)

  ### multi-dimensional forcings

  simdata2 <- simdata
  simdata2$force2 <- rep(rnorm(10, mean = 5), times = rep(c(4, 6), times = 5))
  spec <- rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) +
                     forcing(force, force2), data = simdata2)

  expect_identical(forcings(spec), spec@structure[, 5:6])

  forcings(spec) <- data.frame(force = 10:1, force2 = 1:10)
  expect_identical(var_names(spec, "f"), c("force", "force2"))
  expect_true(all(data.frame(force = 10:1, force2 = 1:10) == forcings(spec)))

  forcings(spec) <- data.frame(abc1 = 10:1, abc2 = 1:10)
  expect_identical(var_names(spec, "f"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 10:1, abc2 = 1:10) == forcings(spec)))

  m <- matrix(1:20, ncol = 2)
  forcings(spec) <- m
  expect_identical(var_names(spec, "f"), c("abc1", "abc2"))
  expect_true(all(data.frame(abc1 = 1:10, abc2 = 11:20) == forcings(spec)))

  colnames(m) <- c("qwe", "asd")
  forcings(spec) <- m
  expect_identical(var_names(spec, "f"), c("qwe", "asd"))
  expect_true(all(data.frame(qwe = 1:10, asd = 11:20) == forcings(spec)))

  forcings(spec)[1, 1:2] <- 100
  expect_true(all(data.frame(qwe = c(100, 2:10),
                             asd = c(100, 12:20)) == forcings(spec)))

  # less forcings
  spec2 <- spec
  forcings(spec2) <- 1:10
  expect_identical(var_names(spec2, "f"), "qwe")
  expect_true(all(forcings(spec2) == 1:10))

  spec2 <- spec
  forcings(spec2) <- matrix(10:1, ncol = 1)
  expect_identical(var_names(spec2, "f"), "qwe")
  expect_true(all(forcings(spec2) == 10:1))

  spec2 <- spec
  df <- data.frame(fff = 1:10)
  forcings(spec2) <- df
  expect_identical(var_names(spec2, "f"), "fff")
  expect_true(all(forcings(spec2) == df))

  # more forcings

  spec2 <- spec
  df <- data.frame(fff = 1:10, ggg = 10:1, hhh = 1:10)
  forcings(spec2) <- df
  expect_identical(var_names(spec2, "f"), colnames(df))
  expect_true(all(forcings(spec2) == df))


  # Forcing for non-RD

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  expect_error(forcings(spec),
               "only used")
  expect_error(forcings(spec) <- rnorm(1:10),
               "only used")

  # duplicate variable names
  expect_error(rct_spec(z ~ cluster(uoa1, uoa2) + block(uoa1),
                          data = simdata),
               "more than once")
  expect_error(rct_spec(z ~ cluster(uoa1, uoa1, uoa2), data = simdata),
               "more than once")

})

test_that("Updating call works", {

  data(simdata)

  spec <- rct_spec(dose ~ unitid(uoa1, uoa2), data = simdata)

  expect_true(any(grepl("dose", spec@call$formula[[2]])))
  treatment(spec) <- data.frame(newt = 1:10)
  expect_false(all(grepl("dose", spec@call$formula[[2]])))
  expect_true(any(grepl("newt", spec@call$formula[[2]])))

  expect_true(any(grepl("uoa1", spec@call$formula[[3]])))
  unitids(spec) <- data.frame(newua = 1:10)
  expect_false(all(grepl("uoa1", spec@call$formula[[3]])))
  expect_true(any(grepl("newua", spec@call$formula[[3]])))

  spec2 <- rd_spec(dose ~ unitid(uoa1, uoa2) + block(bid) +
                      forcing(force), data = simdata)

  expect_true(any(grepl("bid", spec2@call$formula[[3]])))
  blocks(spec2) <- data.frame(newb = 1:10)
  expect_false(all(grepl("bid", spec2@call$formula[[3]])))
  expect_true(any(grepl("newb", spec2@call$formula[[3]])))

  expect_true(any(grepl("force", spec2@call$formula[[3]])))
  forcings(spec2) <- data.frame(newf = 1:10)
  expect_false(all(grepl("force", spec2@call$formula[[3]])))
  expect_true(any(grepl("newf", spec2@call$formula[[3]])))


})

test_that("treatment extraction with NA", {
  data(simdata)
  simdata$z[1:4] <- NA
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  expect_identical(.bin_txt(spec), spec@structure$z)
  expect_identical(.bin_txt(spec), treatment(spec)[, 1])


  simdata$dose[1:4] <- NA
  spec <- rct_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  expect_error(.bin_txt(spec), "Must provide a dichotomy")
  o <- treatment(spec)[, 1]
  expect_true(is.na(o[1]))
  expect_true(all(!is.na(o[-1])))


  z <- .bin_txt(spec, dichotomy = dose >= 250 ~ .)
  expect_true(is.na(z[1]))
  expect_true(all(z[-1] %in% 0:1))
  expect_identical(z, treatment(spec, dichotomy = dose >= 250 ~ .)[, 1])
  expect_identical(o, treatment(spec)[, 1])


})

test_that("convert_to_data.frame add'l testing", {
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  ##### Replacing existing component
  # vector
  cdf <- .convert_to_data.frame(1:10, spec, "t")
  expect_s3_class(cdf, "data.frame")
  expect_equal(dim(cdf), c(10, 1))
  expect_equal(names(cdf), "z")

  # unnamed matrix
  cdf <- .convert_to_data.frame(matrix(1:10, nrow = 10), spec, "t")
  expect_equal(names(cdf), "z")

  # named matrix
  mm <- matrix(1:10, nrow = 10)
  colnames(mm) <- "w"
  cdf <- .convert_to_data.frame(mm, spec, "t")
  expect_equal(names(cdf), "w")

  # unnamed data.frame
  df <- data.frame(1:10)
  colnames(df) <- NULL
  cdf <- .convert_to_data.frame(df, spec, "t")
  expect_equal(names(cdf), "z")

  # named data.frame
  cdf <- .convert_to_data.frame(data.frame(a = 1:10), spec, "t")
  expect_equal(names(cdf), "a")

  ##### Entering new components
  # vector
  expect_error(.convert_to_data.frame(1:10, spec, "b"),
               "requires a named")

  # unnamed matrix
  expect_error(.convert_to_data.frame(matrix(1:10, nrow = 10), spec, "b"),
               "requires a named")

  # named matrix
  mm <- matrix(1:10, nrow = 10)
  colnames(mm) <- "w"
  cdf <- .convert_to_data.frame(mm, spec, "b")
  expect_equal(names(cdf), "w")

  # unnamed data.frame
  df <- data.frame(1:10)
  colnames(df) <- NULL
  expect_error(.convert_to_data.frame(df, spec, "b"),
               "requires a named")

  # named data.frame
  cdf <- .convert_to_data.frame(data.frame(a = 1:10), spec, "b")
  expect_equal(names(cdf), "a")

  ##### Multiple columns

  # Matrix
  mm <- matrix(1:20, nrow = 10)
  cdf <- .convert_to_data.frame(mm, spec, "u")
  expect_s3_class(cdf, "data.frame")
  expect_equal(dim(cdf), c(10, 2))
  expect_equal(names(cdf), c("uoa1", "uoa2"))

  expect_error(.convert_to_data.frame(mm, spec, "b"),
               "requires a named")

  colnames(mm) <- c("a", "b")
  cdf <- .convert_to_data.frame(mm, spec, "u")
  expect_equal(names(cdf), c("a", "b"))
  cdf <- .convert_to_data.frame(mm, spec, "b")
  expect_equal(names(cdf), c("a", "b"))

  # data.frame
  df <- data.frame(a = 1:10, b = 1:10)
  cdf <- .convert_to_data.frame(df, spec, "u")
  expect_s3_class(cdf, "data.frame")
  expect_equal(dim(cdf), c(10, 2))
  expect_equal(names(cdf), c("a", "b"))
  cdf <- .convert_to_data.frame(df, spec, "b")
  expect_equal(names(cdf), c("a", "b"))

  colnames(df) <- NULL
  cdf <- .convert_to_data.frame(df, spec, "u")
  expect_equal(names(cdf), c("uoa1", "uoa2"))

  expect_error(.convert_to_data.frame(df, spec, "b"),
               "requires a named")


  ##### Extra columns

  # Matrix
  mm <- matrix(1:30, nrow = 10)
  expect_error(.convert_to_data.frame(mm, spec, "u"),
               "must be named")

  expect_error(.convert_to_data.frame(mm, spec, "b"),
               "requires a named")

  colnames(mm) <- c("a", "b", "c")
  cdf <- .convert_to_data.frame(mm, spec, "u")
  expect_s3_class(cdf, "data.frame")
  expect_equal(dim(cdf), c(10, 3))
  expect_equal(names(cdf), c("a", "b", "c"))
  cdf <- .convert_to_data.frame(mm, spec, "b")
  expect_equal(names(cdf), c("a", "b", "c"))

  # data.frame
  df <- data.frame(a = 1:10, b = 1:10, c = 1:10)
  cdf <- .convert_to_data.frame(df, spec, "u")
  expect_equal(names(cdf), c("a", "b", "c"))
  cdf <- .convert_to_data.frame(df, spec, "b")
  expect_equal(names(cdf), c("a", "b", "c"))

  colnames(df) <- NULL
  expect_error(.convert_to_data.frame(df, spec, "u"),
               "must be named")
  expect_error(.convert_to_data.frame(df, spec, "b"),
               "requires a named")



  expect_error(.convert_to_data.frame(1:10, spec, "q"),
               "Invalid type")

  expect_error(.convert_to_data.frame(lm(1~1), spec, "u"))
})

test_that("has_blocks", {
  data(simdata)

  blocked_spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  no_blocks_spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)

  expect_true(has_blocks(blocked_spec))
  expect_false(has_blocks(no_blocks_spec))
})

test_that("dichotomy args where specification has no blocks, no time-varying assignment", {
  data(simdata)
  spec1 <- rct_spec(dose ~ cluster(uoa1, uoa2), simdata)

  lm1 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata)
  expect_equal(lm1$model[[2]], as.numeric(simdata$dose>250))

  lm2 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata,
            weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lm3 <- lm(y ~ assigned(spec1), simdata,
            weights = ate(spec1, dichotomy = dose >250~dose<=250))
  wts <- ate(spec1, data = simdata, dichotomy = dose >250~dose<=250)
  lm4 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata,
            weights = wts)

  expect_true(all.equal(lm2$coefficients, lm3$coefficients, check.attributes = FALSE))
  expect_true(all.equal(lm2$coefficients, lm4$coefficients, check.attributes = FALSE))
  expect_equal(lm2$model[[2]], lm1$model[[2]])
  expect_equal(lm3$model[[2]], lm1$model[[2]])
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(lm2$model$`(weights)`, slot)),
                           deparse1(methods::slot(wts, slot)))
               } else {
                 identical(methods::slot(lm2$model$`(weights)`, slot),
                           methods::slot(wts, slot))
               },
               logical(1L)))
  )
  expected_wts <- numeric(nrow(simdata))
  expected_wts[simdata$dose >250] <- 1 / mean(simdata$dose>250)
  expected_wts[simdata$dose <=250] <- 1 / mean(simdata$dose<=250)
  expect_equal(lm2$model$`(weights)`@.Data, expected_wts)

  lmitt1 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose>250~dose<=250)
  expect_equal(lmitt1$model$dose., as.numeric(simdata$dose>250))
  expect_true(all.equal(lmitt1$coefficients[!grepl("^y:", names(lmitt1$coefficients))],
                        lm1$coefficients, check.attributes = FALSE))

  lmitt2 <- lmitt(y ~ 1, specification = spec1, simdata,
                  weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lmitt3 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lmitt4 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = ate(spec1))
  lmitt5 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = wts)

  expect_equal(lmitt2$coefficients, lmitt3$coefficients)
  expect_equal(lmitt2$coefficients, lmitt4$coefficients)
  expect_equal(lmitt2$coefficients, lmitt5$coefficients)
  expect_true(all.equal(lmitt2$coefficients[!grepl("^y:", names(lmitt2$coefficients))],
                        lm2$coefficients, check.attributes = FALSE))
  expect_equal(lmitt2$model[[2]], lmitt1$model[[2]])
  expect_equal(lmitt3$model[[2]], lm1$model[[2]])
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(lmitt2$model$`(weights)`, slot)),
                           deparse1(methods::slot(wts, slot)))
               } else {
                 identical(methods::slot(lmitt2$model$`(weights)`, slot),
                           methods::slot(wts, slot))
               },
               logical(1L)))
  )
  expect_equal(lmitt2$model$`(weights)`@.Data, expected_wts)

})

test_that("dichotomy args where specification has blocks, no time-varying assignment", {
  data(simdata)
  spec1 <- rct_spec(dose ~ cluster(uoa1, uoa2) + block(bid), simdata)

  lm1 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata)
  expect_equal(lm1$model[[2]], as.numeric(simdata$dose>250))

  lm2 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata,
            weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lm3 <- lm(y ~ assigned(spec1), simdata,
            weights = ate(spec1, dichotomy = dose >250~dose<=250))
  wts <- ate(spec1, data = simdata, dichotomy = dose >250~dose<=250)
  lm4 <- lm(y ~ assigned(spec1, dichotomy = dose >250~dose<=250), simdata,
            weights = wts)

  expect_true(all.equal(lm2$coefficients, lm3$coefficients, check.attributes = FALSE))
  expect_true(all.equal(lm2$coefficients, lm4$coefficients, check.attributes = FALSE))
  expect_equal(lm2$model[[2]], lm1$model[[2]])
  expect_equal(lm3$model[[2]], lm1$model[[2]])
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(lm2$model$`(weights)`, slot)),
                           deparse1(methods::slot(wts, slot)))
               } else {
                 identical(methods::slot(lm2$model$`(weights)`, slot),
                           methods::slot(wts, slot))
               },
               logical(1L)))
  )
  expected_wts <- numeric(nrow(simdata))
  # in block 1, no uoa's have dose > 250, while in blocks 2 and 1/3 do
  expected_wts[simdata$bid == 1] <- 0 #
  expected_wts[simdata$dose >250 & simdata$bid %in% c(2, 3)] <- 3
  expected_wts[simdata$dose <=250 & simdata$bid %in% c(2, 3)] <- 3/2
  expect_equal(lm2$model$`(weights)`@.Data, expected_wts)

  suppressMessages(
    lmitt1 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose>250~dose<=250)
  )
  expect_equal(lmitt1$model$dose., as.numeric(simdata$dose>250))

  lmitt2 <- lmitt(y ~ 1, specification = spec1, simdata,
                  weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lmitt3 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = ate(spec1, dichotomy = dose >250~dose<=250))
  lmitt4 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = ate(spec1))
  lmitt5 <- lmitt(y ~ 1, specification = spec1, simdata, dichotomy = dose >250~dose<=250,
                  weights = wts)

  expect_equal(lmitt2$coefficients, lmitt3$coefficients)
  expect_equal(lmitt2$coefficients, lmitt4$coefficients)
  expect_equal(lmitt2$coefficients, lmitt5$coefficients)
  expect_equal(lmitt2$model[[2]], lmitt1$model[[2]])
  expect_equal(lmitt3$model[[2]], lm1$model[[2]])
  expect_true(
    all(vapply(c(".Data", "StudySpecification", "target", "dichotomy"),
               function(slot) if (slot == "dichotomy") {
                 identical(deparse1(methods::slot(lmitt2$model$`(weights)`, slot)),
                           deparse1(methods::slot(wts, slot)))
               } else {
                 identical(methods::slot(lmitt2$model$`(weights)`, slot),
                           methods::slot(wts, slot))
               },
               logical(1L)))
  )
  expect_equal(lmitt2$model$`(weights)`@.Data, expected_wts)

})

test_that("dichotomy args where specification has blocks, time-varying assignment", {
  set.seed(202021)
  analysis_dat  <- data.frame(
    id = rep(letters[1:6], each = 2),
    year = ordered(rep(c("AY20", "AY21"), 6), levels=c("AY20", "AY21", ".")),
    y = rnorm(12)
  )
  specification_dat  <- data.frame(id=letters[1:6], blk=character(6),
                            year_trt = ordered(rep(c("AY21", "AY20", "."),
                                                   each = 2),
                                               levels=c("AY20", "AY21", ".")
                            )
  )
  specification_dat[   specification_dat$id %in% c("a", "e") , "blk"]  <- "A"
  specification_dat[ !(specification_dat$id %in% c("a", "e")), "blk"]  <- "B"
  spec  <- obs_spec(year_trt~uoa(id)+block(blk), data=specification_dat)

  wts <-
    lapply(c("AY20", "AY21"),
           {\(yr) ett(spec, data=subset(analysis_dat,year==yr),
                      dichotomy= as.formula(paste0("year_trt<=\"", yr, "\"~."))) }
    ) |> unsplit(analysis_dat$year)

  lm1 <- lm(y ~ assigned(spec, dichotomy = year_trt<=year~.), data = analysis_dat)
  analysis_dat_year_trts <- setNames(specification_dat$year_trt, specification_dat$id)[analysis_dat$id]
  expect_equal(lm1$model[[2]], as.numeric(analysis_dat_year_trts <= analysis_dat$year))

  lm2 <- lm(y ~ assigned(spec, dichotomy = year_trt<=year~.), data = analysis_dat,
            weights = wts)
  expected_wts <- numeric(nrow(analysis_dat))
  expected_wts[analysis_dat$id %in% c("a", "e") & analysis_dat$year == "AY20"] <- 0
  expected_wts[analysis_dat$id %in% c("a", "e") & analysis_dat$year == "AY21"] <- 1
  expected_wts[analysis_dat$id %in% c("c", "d") & analysis_dat$year == "AY20"] <- 1
  expected_wts[analysis_dat$id %in% c("b", "f") & analysis_dat$year == "AY20"] <- 1
  expected_wts[analysis_dat$id %in% c("b", "c", "d") & analysis_dat$year == "AY21"] <- 1
  expected_wts[analysis_dat$id %in% c("f") & analysis_dat$year == "AY21"] <- 3
  expect_equal(wts@.Data, expected_wts)

  suppressMessages(
    lmitt1 <- lmitt(y ~ 1, specification = spec, analysis_dat, dichotomy = year_trt<=year ~.)
  )
  expect_equal(lmitt1$model$year_trt., as.numeric(analysis_dat$year>=analysis_dat_year_trts))

  lmitt2 <- lmitt(y ~ 1, specification = spec, analysis_dat, dichotomy = year_trt<=year ~.,
                  weights = wts)
  expect_equal(lmitt2$model$year_trt., as.numeric(analysis_dat$year>=analysis_dat_year_trts))
})

test_that("summary.StudySpecification", {
  data(simdata)

  specrct <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  specrd  <-  rd_spec(z ~ cluster(uoa1, uoa2) + block(bid) + forcing(force),
                       data = simdata)
  specobs <- obs_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  expect_s3_class(summary(specrct), "summary.StudySpecification")
  expect_s3_class(summary(specrd) , "summary.StudySpecification")
  expect_s3_class(summary(specobs), "summary.StudySpecification")

  expect_s4_class(summary(specrct)$StudySpecification, "StudySpecification")
  expect_s4_class(summary(specrd)$StudySpecification , "StudySpecification")
  expect_s4_class(summary(specobs)$StudySpecification, "StudySpecification")

  expect_output(print(summary(specrct)), "Randomized")
  expect_output(print(summary(specrd)) , "Discontinuity")
  expect_output(print(summary(specobs)), "Observational")

  specrct <- rct_spec(o ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  expect_output(print(summary(specrct)), "...")
  expect_output(print(summary(specrct)), "excluded")
  expect_identical(specrct, summary(specrct)$StudySpecification)
  expect_s3_class(summary(specrct)$treatment_table, "table")

  expect_output(print(summary(specrct)),
                "group excluded")
  expect_output(print(summary(specrct), max_unit_print = 2),
                "groups excluded")

})

test_that(".get_spec", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  mod <- lm(y ~ x, data = simdata)

  mod1 <- lm(y ~ z, data = simdata, weights = ate(spec),
             offset = cov_adj(mod))
  mod2 <- lm(y ~ z, data = simdata, weights = ate(),
             offset = cov_adj(mod, specification = spec))
  mod3 <- lm(y ~ z, data = simdata, weights = ate(spec),
             offset = cov_adj(mod, specification = spec))
  mod4 <- lm(y ~ assigned(), data = simdata, weights = ate(spec),
             offset = cov_adj(mod))
  mod5 <- lm(y ~ assigned(), data = simdata, weights = ate(),
             offset = cov_adj(mod, specification = spec))
  mod6 <- lm(y ~ assigned(), data = simdata, weights = ate(spec),
             offset = cov_adj(mod, specification = spec))
  mod7 <- lm(y ~ assigned(), data = simdata, weights = ate(spec),
             offset = propertee::cov_adj(mod))
  expect_true(all(mod1$coef == mod2$coef))
  expect_true(all(mod1$coef == mod3$coef))
  expect_true(all(mod1$coef == mod4$coef))
  expect_true(all(mod1$coef == mod5$coef))
  expect_true(all(mod1$coef == mod6$coef))
  expect_true(all(mod1$coef == mod7$coef))

  spec2 <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  expect_error(lm(y ~ assigned(), data = simdata, weights = ate(spec),
                  offset = cov_adj(mod, specification = spec2)),
               "differing `StudySpecification`")

  expect_error(lm( y ~ assigned(), data = simdata),
               "Unable to locate")

  expect_error(lm(y ~ assigned(), data = simdata, weights = ate(),
                  offset = cov_adj(mod)),
               "Inf Recusion")

  # #37 offset in formula isntead of argument
  mod7 <- lm(y ~ z + offset(cov_adj(mod)), data = simdata, weights = ate(spec))
  expect_true(all(mod1$coef == mod7$coef))
})

test_that(".get_spec returns NULL with NULL_on_error = TRUE", {
  on.exit(cov_adj <- propertee::cov_adj)

  cov_adj <- function(model, newdata = NULL) {
    specification <- .get_spec(TRUE)

    return(specification)
  }

  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  mod <- lm(y ~ x, data = simdata)
  expect_true(is.null(cov_adj(mod, newdata = simdata)))
  mod1 <- lm(y ~ z, data = simdata, offset = cov_adj(mod))
  expect_true(is.null(mod1$offset))
})

test_that(".get_spec finds specification in expand.model.frame call", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)

  # ensure we're taking the StudySpecification object from the model not the existing object
  # in the environment
  uoanames <- var_names(spec, "u")
  spec <- rct_spec(z ~ cluster(uoa1), data = simdata,
                    subset = simdata$uoa1 %in% c(1, 3, 5))
  dat <- .expand.model.frame_teeMod(ssmod, uoanames)
  expect_true(is.data.frame(dat))
  expect_true(all(uoanames %in% colnames(dat)))
})

test_that(".get_spec finds specification in model.frame call", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, weights = ate(), specification = spec)

  mf1 <- stats::model.frame(ssmod)
  mf2 <- stats::model.frame(formula(ssmod), ssmod$call$data)
  mf3 <- stats::model.frame(terms(ssmod), ssmod$call$data)

  expect_equal(mf2, mf3)
  expect_equal(lapply(mf1@.Data, as.numeric),
               lapply(as.data.frame(cbind(mf2, weights = ssmod$weights))@.Data,
                      as.numeric))
})

test_that("var_table", {
  data(simdata)
  # RCT and Obs
  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)

  tbl <- var_table(spec)

  expect_true(inherits(tbl, "matrix"))
  expect_type(tbl, "character")
  expect_equal(dim(tbl), c(3, 2))

  expect_equal(tbl[, 1], c("Treatment", "Unit of Assignment", "Block"))

  # Only find multiple variables in cluster
  expect_equal(grepl(",", tbl[, 2]), c(FALSE, TRUE, FALSE))

  tbl2 <- var_table(spec, compress = FALSE)
  expect_true(inherits(tbl2, "matrix"))
  expect_type(tbl2, "character")
  expect_equal(colnames(tbl2)[-1], c("Variable 1", "Variable 2"))
  expect_equal(dim(tbl2), c(3, 3))
  expect_true(is.na(tbl2[1, 3]))
  expect_true(!is.na(tbl2[2, 3]))
  expect_true(is.na(tbl2[3, 3]))

  tbl3 <- var_table(spec, report_all = TRUE)
  expect_identical(tbl, tbl3)

  tbl4 <- var_table(spec, compress = FALSE, report_all = TRUE)
  expect_identical(tbl2, tbl4)

  # RD
  spec <- rd_spec(z ~ uoa(uoa1, uoa2) + forcing(force), data = simdata)

  tbl <- var_table(spec)

  expect_true(inherits(tbl, "matrix"))
  expect_type(tbl, "character")
  expect_equal(dim(tbl), c(3, 2))

  expect_equal(tbl[, 1], c("Treatment", "Unit of Assignment", "Forcing"))

  # Only find multiple variables in cluster
  expect_equal(grepl(",", tbl[, 2]), c(FALSE, TRUE, FALSE))

  tbl2 <- var_table(spec, compress = FALSE)
  expect_true(inherits(tbl2, "matrix"))
  expect_type(tbl2, "character")
  expect_equal(dim(tbl2), c(3, 3))
  expect_true(is.na(tbl2[1, 3]))
  expect_true(!is.na(tbl2[2, 3]))
  expect_true(is.na(tbl2[3, 3]))

  tbl3 <- var_table(spec, report_all = TRUE)
  expect_true(inherits(tbl3, "matrix"))
  expect_type(tbl3, "character")
  expect_equal(dim(tbl3), c(4, 2))
  expect_true(!is.na(tbl3[1, 2]))
  expect_true(!is.na(tbl3[2, 2]))
  expect_true(is.na(tbl3[3, 2]))
  expect_true(!is.na(tbl3[4, 2]))

  tbl4 <- var_table(spec, compress = FALSE, report_all = TRUE)
  expect_true(inherits(tbl4, "matrix"))
  expect_type(tbl4, "character")
  expect_equal(dim(tbl4), c(4, 3))
  expect_true(!is.na(tbl4[1, 2]))
  expect_true(!is.na(tbl4[2, 2]))
  expect_true(is.na(tbl4[3, 2]))
  expect_true(!is.na(tbl4[4, 2]))
  expect_true(is.na(tbl4[1, 3]))
  expect_true(!is.na(tbl4[2, 3]))
  expect_true(is.na(tbl4[3, 3]))
  expect_true(is.na(tbl4[4, 3]))

  # No double-variables
  simdata$cid <- paste(simdata$uoa1, simdata$uoa2)
  spec <- rct_spec(z ~ uoa(cid), data = simdata)

  expect_identical(var_table(spec),
                   var_table(spec, compress = FALSE))

})

#' @include StudySpecification.R
NULL

############### Treatment

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("treatment",
           function(x, newdata = NULL, dichotomy = NULL, by = NULL, ...) {
  standardGeneric("treatment")
})

##' @title Accessors and Replacers for \code{StudySpecification} objects
##'
##' @description Allows access to the elements which define a
##'   \code{StudySpecification}, enabling their extraction or replacement.
##'
##' @details For [treatment()], when argument \code{binary} is \code{FALSE}, the
##'   treatment variable passed into the \code{StudySpecification} is returned
##'   as a one-column \code{data.frame} regardless of whether it is binary or
##'   \code{x} has a \code{dichotomy}
##'
##'   If a \code{dichotomy} is passed, a binary one-column \code{data.frame}
##'   will be returned. If not and \code{binary} is \code{TRUE}, unless the
##'   \code{StudySpecification} has a binary treatment, [treatment()] will
##'   error. If \code{binary} is \code{"ifany"}, it will return the original
##'   treatment in this case.
##'
##'   The one-column \code{data.frame} returned by [treatment()] is named as
##'   entered in the \code{StudySpecification} creation, but if a
##'   \code{dichotomy} is passed, the column name is \code{"__z"} to try and
##'   avoid any name conflicts.
##'
##'   For the \code{value} when using replacers, the replacement must have the
##'   same number of rows as the \code{StudySpecification} (the same number of
##'   units of assignment). The number of columns can differ (e.g. if the
##'   \code{StudySpecification} were defined with two variable uniquely
##'   identifying blocks, you can replace that with a single variable uniquely
##'   identifying blocks, as long as it respects other restrictions.)
##'
##'   If the replacement value is a \code{data.frame}, the name of the columns
##'   is used as the new variable names. If the replacement is a \code{matrix}
##'   or \code{vector}, the original names are retained. If reducing the number
##'   of variables (e.g., moving from two variables uniquely identifying to a
##'   single variable), the appropriate number of variable names are retained.
##'   If increasing the number of variables, a \code{data.frame} with names must
##'   be provided.
##'
##' @param x a \code{StudySpecification} object
##' @param newdata optional; an additional \code{data.frame}. If passed, and the
##'   unit of assignment variable is found in \code{newdata}, then the requested
##'   variable type for each unit of \code{newdata} is returned. See \code{by}
##'   argument if the name of the unit of assignment differs.
##' @param dichotomy optional; a formula specifying how to dichotomize a
##'   non-binary treatment variable. See the Details section of the \code{ett()}
##'   or \code{att()} help pages for information on specifying this formula
##' @param by optional; named vector or list connecting names of unit of
##'   assignment/unitid/cluster variables in \code{x} to unit of
##'   assignment/unitid/cluster variables in \code{data}. Names represent
##'   variables in \code{x}; values represent variables in \code{newdata}. Only
##'   needed if variable names differ.
##' @param implicit Should a block-less \code{StudySpecification} return a
##'   constant 1 when extracting \code{blocks}?
##' @param ... ignored.
##' @return \code{data.frame} containing requested variable, or an updated
##'   \code{StudySpecification}. [treatment()] works slightly differently, see
##'   \code{Details}.
##' @export
##' @rdname StudySpecification_extractreplace
##' @examples
##' data(simdata)
##' spec <- obs_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
##' blocks(spec) # empty
##' blocks(spec) <- data.frame(blks = c(1, 1, 2, 2, 3, 3, 4, 4, 5, 5))
##' blocks(spec)
##' blocks(spec) <- c(5, 5, 4, 4, 3, 3, 2, 2, 1, 1)
##' blocks(spec) # notice that variable is not renamed
setMethod("treatment", "StudySpecification",
          function(x, newdata = NULL, dichotomy = NULL, by = NULL, ...) {
  .specification_accessors_newdata_validate(newdata, by)

  if (is.null(dichotomy)) {
    if (!is.null(newdata)) {
      return(.get_col_from_new_data(x, newdata, type = "t", by))
    } else {
      return(x@structure[x@column_index == "t"])
    }
  }

  if (!is.null(dichotomy)) {
    return(data.frame(z__ = .bin_txt(x, newdata, dichotomy)))
  } else if (!has_binary_treatment(x)) {
    stop(paste("No binary treatment can be produced. Treatment is",
               "non-binary and no `dichotomy` provided."))
  }
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("treatment<-", function(x, value) standardGeneric("treatment<-"))

##' @param value replacement. Either a \code{vector}/\code{matrix} of
##'   appropriate dimension, or a named \code{data.frame} if renaming variable
##'   as well. See \code{Details}.
##' @export
##' @rdname StudySpecification_extractreplace
setMethod("treatment<-", "StudySpecification", function(x, value) {
  value <- .convert_to_data.frame(value, x, "t")

  x@structure[x@column_index == "t"] <- value
  names(x@structure)[x@column_index == "t"] <- colnames(value)
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

##' @title (Internal) Extracts treatment as binary \code{vector}
##' @details If a \code{dichotomy} is specified or the \code{StudySpecification}
##'   has a treatment variable consisting only of 0/1 or \code{NA}, then returns
##'   the binary treatment. Otherwise (it has a non-binary treatment and lacks a
##'   dichotomy) it errors.
##' @param spec A \code{StudySpecification}, used to get treatment assignment
##'   information
##' @param data A dataframe with unit of assignment information and, if a
##'   dichotomy is provided, columns specified therein
##' @param dichotomy Optional, a formula. See the Details section of the
##'   \code{ett()} or \code{att()} help pages for information on specifying the
##'   formula
##' @return A \code{vector} of binary treatments
##' @keywords internal
.bin_txt <- function(spec, data = NULL, dichotomy = NULL) {
  # get treatment from the specification
  tt <- treatment(spec)

  if (!is.null(data)) tt <- .expand_txt(tt, data, spec)

  if (!is.null(dichotomy)) {
    treatment <- .apply_dichotomy(tt, dichotomy)
  } else {
    treatment <- tt[,1]
  }

  if (!all(treatment %in% c(0:1, NA))) {
    stop("Must provide a dichotomy if the `StudySpecification` has a non-binary treatment")
  }

  return(treatment)
}

##' @title (Internal) Expand treatment variable from a \code{StudySpecification}
##'   to a dataframe with unit of assignment information
##' @param txt A dataframe with one column corresponding to the treatment. Can
##'   be dichotomized or as it's stored in \code{spec}
##' @param data A dataframe with unit of assignment information
##' @param spec A \code{StudySpecification}, used to align unit of assignment
##'   information with \code{txt}
##' @keywords internal
.expand_txt <- function(txt, data, spec) {
  # if there's a `data` argument, merge it to the treatment info using the
  # units of assignment
  if (!all(var_names(spec, "u") %in% colnames(data))) {
    stop("Not all unit of assignment variables can be found in `data`")
  }
  treatment_uoa <- cbind(txt, spec@structure[, var_names(spec, "u"), drop = FALSE])
  txt <- .merge_preserve_order(data, treatment_uoa, by = var_names(spec, "u"), all.x = TRUE)

  txtname <- var_names(spec, "t")
  txt <- tryCatch(txt[, txtname, drop = FALSE],
                  error = function(e) {
                    # if treatment variable already exists in data, there
                    # will be a .x and .y version; e.g. z.x and z.y, so
                    # we'll extract the ".y" version (the second one)
                    # since the merge above has the treatment from the
                    # StudySpecification second.
                    txt[, paste0(txtname, ".y"), drop = FALSE]
                  })
  colnames(txt) <- txtname

  return(txt)
}

##' @title (Internal) Applies dichotomy to treatment
##' @description Given a dichotomy formula and a \code{data.frame} with a
##'   treatment variable and any variables in the formula, returns a
##'   \code{vector} containing only \code{0}, \code{1}, or \code{NA}.
##' @param txt A named \code{data.frame} containing a column of the treatment,
##'   such as that produed by `treatment(myspecification)`, and any variables
##'   specified in \code{dichotomy}.
##' @param dichotomy A formula specifying how to dichotomize the non-binary
##'   treatment column in \code{txt} (or a call that evaluates to a formula).
##'   See the Details section of the \code{ett()} or \code{att()} help pages for
##'   information on specifying this formula
##' @return A \code{vector} of binary treatments
##' @keywords internal
.apply_dichotomy <- function(txt, dichotomy) {
  if (!is.data.frame(txt)) stop("`txt` is expected to be a named `data.frame`")
  if (!all(setdiff(all.vars(dichotomy), ".") %in% colnames(txt))) {
    stop(paste("Could not find variables specified in dichotomy. Provide a",
               "data argument with these columns, or ensure provided data argument",
               "has them."))
  }

  lhs_dot <- rhs_dot <- FALSE
  if (dichotomy[[3]] == ".") {
    # control group is .
    # control goes first since txt switches LHS and RHS
    dichotomy[[3]] <- 1
    rhs_dot <- TRUE
  }
  if (dichotomy[[2]] == ".") {
    # treatment group is .
    dichotomy[[2]] <- dichotomy[[3]]
    dichotomy[[3]] <- 1
    lhs_dot <- TRUE
  }
  if (lhs_dot & rhs_dot) {
    stop("At least one side for dichotomy formula must not be `.`")
  }

  m <- model.frame(dichotomy, txt, na.action = na.pass)

  if (lhs_dot) {
    return(as.numeric(!m[,1]))
  } else if (rhs_dot) {
    return(as.numeric(m[,1]))
  } else {
    ditxt <- m[,2] + 2*m[,1] - 1
    ditxt[ditxt == -1] <- NA
    if (any(!is.na(ditxt) & ditxt > 1)) {
      stop("treatment dichotomy overlaps")
    }
    return(ditxt)
  }

}

############### Units of Assignment

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("units_of_assignment", function(x, newdata = NULL, by = NULL) {
  standardGeneric("units_of_assignment")
})

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("units_of_assignment", "StudySpecification", function(x, newdata = NULL, by = NULL) {
  if (x@unit_of_assignment_type == "none") {
    stop("StudySpecification specified without unit of assignment")
  }
  if (x@unit_of_assignment_type == "unitid") {
    stop("StudySpecification specified with `unitid()`, not `unit_of_assignment()`")
  }
  if (x@unit_of_assignment_type == "cluster") {
    stop("StudySpecification specified with `cluster()`, not `unit_of_assignment()`")
  }
  .specification_accessors_newdata_validate(newdata, by)
  if (!is.null(newdata)) {
    return(.get_col_from_new_data(x, newdata, type = "u", by))
  }
  return(x@structure[x@column_index == "u"])
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("units_of_assignment<-", function(x, value) {
  standardGeneric("units_of_assignment<-")
})

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("units_of_assignment<-", "StudySpecification", function(x, value) {

  value <- .convert_to_data.frame(value, x, "u")

  x <- .update_structure(x, value, "u")

  dupuoa <- duplicated(units_of_assignment(x))
  dupall <- duplicated(x@structure[x@column_index != "f"])
  if (any(dupuoa)) {

    if (sum(dupuoa) != sum(dupall)) {
      stop(paste("Fewer new units of assignment then original, but new",
                 "collapsed units would have non-constant treatment and/or",
                 "block structure"))
    }
    warning("Fewer new units of assignment then original, collapsing")

    x@structure <- x@structure[-dupuoa, ]
  }
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

############### Cluster

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("clusters", function(x, newdata = NULL, by = NULL) standardGeneric("clusters"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("clusters", "StudySpecification", function(x, newdata = NULL, by = NULL) {
  if (x@unit_of_assignment_type == "none") {
    stop("StudySpecification specified without clusters")
  }
  if (x@unit_of_assignment_type == "unitid") {
    stop("StudySpecification specified with `unitid()`, not `cluster()`")
  }
  if (x@unit_of_assignment_type == "unit_of_assignment") {
    stop("StudySpecification specified with `unit_of_assignment()`, not `cluster()`")
  }
  .specification_accessors_newdata_validate(newdata, by)
  if (!is.null(newdata)) {
    return(.get_col_from_new_data(x, newdata, type = "u", by))
  }
  return(x@structure[x@column_index == "u"])
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("clusters<-", function(x, value) standardGeneric("clusters<-"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("clusters<-", "StudySpecification", function(x, value) {

  value <- .convert_to_data.frame(value, x, "u")

  x <- .update_structure(x, value, "u")

  dupclust <- duplicated(clusters(x))
  dupall <- duplicated(x@structure[x@column_index != "f"])
  if (any(dupclust)) {

    if (sum(dupclust) != sum(dupall)) {
      stop(paste("Fewer new clusters then original, but new collapsed",
                 "clusters would have non-constant treatment and/or",
                 "block structure"))
    }
    warning("Fewer new clusters then original, collapsing")

    x@structure <- x@structure[-dupclust, ]
  }
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

############### Unitid

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("unitids", function(x) standardGeneric("unitids"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("unitids", "StudySpecification", function(x) {
  if (x@unit_of_assignment_type == "none") {
    stop("StudySpecification specified without unit IDs")
  }
  if (x@unit_of_assignment_type == "cluster") {
    stop("StudySpecification specified with `cluster()`, not `unitid()`")
  }
  if (x@unit_of_assignment_type == "unit_of_assignment") {
    stop("StudySpecification specified with `unit_of_assignment()`, not `unitid()`")
  }
  return(x@structure[x@column_index == "u"])
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("unitids<-", function(x, value) standardGeneric("unitids<-"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("unitids<-", "StudySpecification", function(x, value) {

  value <- .convert_to_data.frame(value, x, "u")

  x <- .update_structure(x, value, "u")

  dupids <- duplicated(unitids(x))
  dupall <- duplicated(x@structure[x@column_index != "f"])
  if (any(dupids)) {

    if (sum(dupids) != sum(dupall)) {
      stop(paste("Fewer new unitids then original, but new collapsed",
                 "units would have non-constant treatment and/or",
                 "block structure"))
    }
    warning("Fewer new unitids then original, collapsing")

    x@structure <- x@structure[-dupids, ]
  }
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

############### ..uoa..
# This is when a spec is specified without uoa/cluster/unitid, e.g. z ~ 1. See
# #193 for details

##' @title Return ..uoa.. column
##' @param spec A studyspecification
##' @return The ..uoa.. column
##' @keywords internal
..uoa.. <- function(spec) {
  if (spec@unit_of_assignment_type != "none") {
    stop("Internal error: ..uoa..() only valid for `unit_of_assignment_type` == 'none'")
  }
  return(spec@structure["..uoa.."])
}

############### Blocks

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("blocks", function(x,
                              newdata = NULL,
                              by = NULL,
                              ...) standardGeneric("blocks"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("blocks", "StudySpecification", function(x,
                                       newdata = NULL,
                                       by = NULL,
                                       ...,
                                       implicit = FALSE) {
  .specification_accessors_newdata_validate(newdata, by)
  if (!is.null(newdata)) {
    return(.get_col_from_new_data(x,
                                  newdata,
                                  type = "b",
                                  by,
                                  implicitBlock = implicit,
                                  ...))
  }
  if (!any(x@column_index == "b") & implicit) {
    return(data.frame(".blocks_internal" = rep(1, nrow(x@structure))))
  }
  return(x@structure[x@column_index == "b"])
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("blocks<-", function(x, value) standardGeneric("blocks<-"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("blocks<-", "StudySpecification", function(x, value) {
  value <- .convert_to_data.frame(value, x, "b")

  x <- .update_structure(x, value, "b")
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

##' @export
##' @rdname StudySpecification_extractreplace
has_blocks <- function(x) {
  return("b" %in% x@column_index)
}

############### Forcing

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("forcings", function(x, newdata = NULL, by = NULL) standardGeneric("forcings"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("forcings", "StudySpecification", function(x, newdata = NULL, by = NULL) {
  if (x@type != "RD") {
    stop("Forcing variable only used in RD specifications")
  }
  .specification_accessors_newdata_validate(newdata, by)
  if (!is.null(newdata)) {
    return(.get_col_from_new_data(x, newdata, type = "f", by))
  }
  return(x@structure[x@column_index == "f"])
})

##' @export
##' @rdname StudySpecification_extractreplace
setGeneric("forcings<-", function(x, value) standardGeneric("forcings<-"))

##' @export
##' @rdname StudySpecification_extractreplace
setMethod("forcings<-", "StudySpecification", function(x, value) {
  if (x@type != "RD") {
    stop("Forcing variable only used in RD specifications")
  }

  value <- .convert_to_data.frame(value, x, "f")

  x <- .update_structure(x, value, "f")
  x@call$formula <- .update_call_formula(x)
  validObject(x)
  return(x)
})

############### Helper Functions

##' Helper function for \code{StudySpecification} replacers to ensure
##' replacement is a properly named \code{data.frame}
##'
##' When given a replacement set of values (e.g \code{vector} or \code{matrix}),
##' this ensures that the replacement is a named \code{data.frame}.
##'
##' Input \code{vector}: Since it cannot be named, a vector can only be used to
##' replace an existing component. If the existing component has more than 1
##' column, uses the name of the first column.
##'
##' Input \code{matrix} or \code{data.frame}: If unnamed and replacing existing
##' component, must have no more columns than original component. (If less
##' columns, uses the name of the first few columns.) If named, can replace any
##' number of columns.
##'
##' @title (Internal) Ensures replacement column for \code{StudySpecification}
##'   is a \code{data.frame}.
##' @param value A \code{vector} or \code{data.frame} containing a replacement.
##' @param specification A \code{StudySpecification}
##' @param type One of "t", "f", "u" or "b"
##' @return \code{data.frame} containing named column(s)
##' @keywords internal
.convert_to_data.frame <- function(value, specification, type) {
  if (!type %in% c("t", "f", "u", "b")) {
    stop("Invalid type argument")
  }

  # colnames(value) will return NULL if passed a vector, or a matrix/df without
  # names
  if (is.null(colnames(value))) {
    null_name <- TRUE
  } else {
    null_name <- FALSE
  }

  if (!inherits(value, "data.frame")) {
    # Gives a nice error (cannot convert class to data.frame) so no need for
    # tryCatch here
    value <- as.data.frame(value)
  }

  if (nrow(specification@structure) != nrow(value)) {
    stop("replacement entries do not have same number of rows as current")
  }

  if (null_name) {
    old_names <- var_names(specification, type)
    if (length(old_names) > ncol(value)) {
      # If replacement has less columns, and we need to rename, take the first
      # few names from the original component.
      old_names <- old_names[seq_len(ncol(value))]
    } else if (length(old_names) < ncol(value)) {
      if (length(old_names) == 0) {
        # No original component in specification
        stop("Adding a new component requires a named `data.frame`")
      } else {
        # Original component is shorter
        stop(paste("Replacement component has more columns than original,",
                   "replacement must be named"))
      }
    }
    colnames(value) <- old_names
  }

  return(value)
}

##' Assumes \code{.convert_to_data.frame()} has already been called on
##' \code{new}
##' @title (Internal) Replaces \code{type} columns in \code{specification} with
##'   \code{new}
##' @param specification A \code{StudySpecification}
##' @param new A named \code{data.frame} with the replacement, should be the
##'   output of \code{.convert_to_data.frame()}.
##' @param type One of "t", "f", "u" or "b\".
##' @return The updated \code{StudySpecification}
##' @keywords internal
.update_structure <- function(specification, new, type) {
  specification@structure <-
    cbind.data.frame(specification@structure[specification@column_index != type], new)

  specification@column_index <- c(specification@column_index[specification@column_index != type],
                           rep(type, ncol(new)))
  names(specification@column_index) <- colnames(specification@structure)
  validObject(specification)
  return(specification)
  }

##' Helper function to update the \code{call} with the appropriate variable
##' names after they've been modified. Called within \code{StudySpecification}
##' replacers.
##'
##' It's return should be stuck into the specification via
##' \code{spec@call$formula <- .update_call_formula(spec)}
##' @title (Internal) Updates `spec@call`'s formula with the currently defined
##'   variable names.
##' @param specification A \code{StudySpecification}
##' @return An updated formula
##' @keywords internal
.update_call_formula <- function(specification) {

  .collapse <- function(d, type) {
    paste(var_names(d, type), collapse = ",")
  }

  # Get treatment, ~, and uoa/cluster
  form <- paste0(var_names(specification, "t"), "~", specification@unit_of_assignment_type,
                 "(", .collapse(specification, "u"), ")")

  # Get block if included
  if (length(var_names(specification, "b")) > 0) {
    form <- paste0(form, "+ block(", .collapse(specification, "b"), ")")
  }

  if (specification@type == "RD") {
    form <- paste0(form, "+ forcing(", .collapse(specification, "f"), ")")
  }
  return(formula(form))
}

##' @title (Internal) Extract specified \code{type} from new data set
##' @param specification A \code{StudySpecification}
##' @param newdata A \code{data.frame}, which may or may not be the one which
##'   was used to create \code{specification}. It must have the units of
##'   assignment variable(s) (though \code{by=} argument can be used if the name
##'   differ), and will appropriately merge with the \code{specification} the
##'   blocks, treatment or forcings.
##' @param type One of "t", "f", or "b".
##' @param by optional; named vector or list connecting names of unit of
##'   assignment/unitid/cluster variables in \code{specification} to unit of
##'   assignment/unitid/cluster variables in \code{data}. Names represent
##'   variables in the StudySpecification; values represent variables in the
##'   data. Only needed if variable names differ.
##' @param implicitBlock If the \code{StudySpecification} does not include a
##'   block, \code{TRUE} will return a constant 1 for the blocks if \code{type}
##'   requests it.
##' @param ... Additional arguments to \code{merge()}.
##' @return The column(s) belonging to the requested \code{type} in
##' @keywords internal
##' @importFrom stats model.matrix
.get_col_from_new_data <- function(specification,
                                   newdata,
                                   type,
                                   by = NULL,
                                   implicitBlock = FALSE,
                                   ...) {

  if (!is.null(by)) {
    specification <- .update_by(specification, newdata, by)
  }

  form_for_spec <- as.formula(paste("~",
                                      paste(c(var_names(specification, "u"),
                                              var_names(specification, type)),
                                            collapse = "+"),
                                      " - 1"))
  form_for_newdata <- as.formula(paste("~",
                                       paste(var_names(specification, "u"),
                                             collapse = "+"),
                                       " - 1"))

  specification_data <- stats::model.frame(form_for_spec, specification@structure)
  if (type == "b" &&
        var_names(specification, "b", implicitBlocks = TRUE)[1] == ".blocks_internal") {
    specification_data$.blocks_internal <- rep(1, nrow(specification_data))
  }

  if (specification@unit_of_assignment_type == "none") {
    newdata[["..uoa.."]] <- rownames(newdata)
  }



  newdata_data <- stats::model.frame(form_for_newdata, newdata, na.action = na.pass)

  merged <- .merge_preserve_order(newdata_data,
                                  specification_data,
                                  by = var_names(specification, "u"),
                                  sort = FALSE,
                                  ...)

  return(merged[var_names(specification, type, implicitBlocks = TRUE)])


}

##' @title (Internal) Checks newdata/by argument for specification accessors
##' @param newdata newdata argument from e.g. \code{treatment()},
##'   \code{blocks()}, etc
##' @param by from e.g. \code{treatment()}, \code{blocks()}, etc. See
##'   \code{.check_by()}
##' @return Invisibly \code{TRUE}. Warns or errors as appropriate.
##' @keywords internal
.specification_accessors_newdata_validate <- function(newdata, by) {

  if (!is.null(newdata)) {
    if (!is.data.frame(newdata)) {
      warning(paste("`newdata` is not a data.frame, errors or",
                    "unpredictable results may occur"))
    }
  }
  if (!is.null(by)) {
    .check_by(by)
  }

  invisible(TRUE)
}

test_that("apply dichotomization function", {
  zz <- data.frame(z = 1:10)

  bt <- .apply_dichotomy(zz, z > 5 ~ z == 2)
  expect_true(is.numeric(bt))
  expect_length(bt, nrow(zz))
  expect_true(all(bt %in% c(0, 1, NA)))

  expect_error(.apply_dichotomy(1, z > 5 ~ .),
               "to be a named")

  expect_error(.apply_dichotomy(zz, . ~ .),
               "least one side")

  expect_error(.apply_dichotomy(zz, z < 6 ~ z > 4),
               "overlaps")
})

test_that("apply dichotomy on factor", {
  zz <- data.frame(z = factor(sample(2:4, 20, TRUE)))

  expect_true(0 < mean(.apply_dichotomy(zz, z == "3" ~ .)))
  expect_true(mean(.apply_dichotomy(zz, z == "3" ~ .)) < 1)
  expect_identical(.apply_dichotomy(zz, z == 3 ~ .),
                   .apply_dichotomy(zz, z == "3" ~ .))

  # Ensure that we're comparing against the levels not the values - "2" is the
  # first level, but there is no level 1, so this should flag NO treated units
  expect_true(sum(.apply_dichotomy(zz, z == 1 ~ .)) == 0)

  expect_warning(.apply_dichotomy(zz, z > 3 ~ .),
                 "not meaningful")


  zz <- data.frame(z = factor(sample(c("a", "b", "c"), 20, TRUE)))

  expect_length(table(.apply_dichotomy(zz, z == "b" ~ z == "c"),
                      useNA = 'ifany'), 3)

  zz <- data.frame(z = ordered(sample(3:5, 20, TRUE)))

  expect_true(0 < mean(.apply_dichotomy(zz, z == "3" ~ .)))
  expect_true(mean(.apply_dichotomy(zz, z == "3" ~ .)) < 1)
  expect_identical(.apply_dichotomy(zz, z == 3 ~ .),
                   .apply_dichotomy(zz, z == "3" ~ .))

  expect_true(0 < mean(.apply_dichotomy(zz, z > "4" ~ .)))
  expect_true(mean(.apply_dichotomy(zz, z > "4" ~ .)) < 1)
  expect_identical(.apply_dichotomy(zz, z > 4 ~ .),
                   .apply_dichotomy(zz, z > "4" ~ .))

})

##' @title (Internal) Merge \code{data.frame}s ensuring order of first
##'   \code{data.frame} is maintained
##' @param x \code{data.frame} whose ordering is to be maintained
##' @param ... Additional arguments to \code{merge()}, particularly a second
##'   \code{data.frame} and a \code{by=} argument.
##' @return Merged \code{data.frame} with the same ordering as \code{x}.
##' @keywords internal
.merge_preserve_order <- function(x, ...) {
  x$..orig_ordering.. <- seq_len(nrow(x))
  x <- merge(x, ...)
  x <- x[order(x$..orig_ordering..), ]
  x$..orig_ordering.. <- NULL
  return(x)
}

#' @include StudySpecification.R StudySpecificationAccessors.R
NULL

#' (Internal) model predictions with some model artifacts, as S4 object
#' @slot .Data numeric vector of predictions
#' @slot fitted_covariance_model model standing behind the predictions
#' @slot prediction_gradient matrix, predictions gradient w/r/t model params
#' @keywords internal
setClass("PreSandwichLayer",
         contains = "numeric",
         slots = c(fitted_covariance_model = "ANY",
                   prediction_gradient = "matrix"))

setValidity("PreSandwichLayer", function(object) {
  if (!inherits(object@fitted_covariance_model$terms, "terms")) {
    return(paste("Fitted covariance adjustment model must have a valid 'terms'",
                 "attribute"))
  }

  tryCatch({
    sandwich::bread(object@fitted_covariance_model)
    sandwich::estfun(object@fitted_covariance_model)
    }, error = function(e) {
      stop(paste("Functions for extracting vcov elements not applicable to",
                 "fitted covariance adjustment model"))
  })

  if (!is.numeric(object@prediction_gradient)) {
    return("Prediction gradient must be a numeric matrix")
  }
  if (dim(object@prediction_gradient)[1] != length(object)) {
    msg <- paste0("Prediction gradient matrix (",
                  paste(dim(object@prediction_gradient), collapse = ", "),
                  ") does not have the same dimension along axis 1 as the ",
                  "covariance adjustment vector (", length(object), ")")
    return(msg)
  }
  if (dim(object@prediction_gradient)[2] !=
      qr(stats::model.matrix(object@fitted_covariance_model))$rank) {
    return(paste("Prediction gradient does not have the same number of columns",
                 "predictors in the covariance adjustment model"))
  }
  TRUE
})

#' (Internal) model predictions with more model artifacts, as S4 object
#'
#' Contains PreSandwichLayer class.  Only additional slots listed here.
#' @slot keys a data.frame
#' @slot StudySpecification a StudySpecification
#' @keywords internal
setClass("SandwichLayer",
         contains = "PreSandwichLayer",
         slots = c(keys = "data.frame",
                   StudySpecification = "StudySpecification"))

setValidity("SandwichLayer", function(object) {
  if (nrow(object@keys) !=
      nrow(stats::model.matrix(object@fitted_covariance_model))) {
    return(paste("Keys does not have the same number of rows as the dataset",
                 "used to fit the covariance adjustment model"))
  }

  if (any(is.na(object))) {
    msg <- paste("Some covariance adjustments are NA; be careful of dropping",
                 "these observations when fitting the ITT effect model")
    warning(msg)
  }
  TRUE
})

##' @title (Internal) \code{show} helper for
##'   \code{PreSandwichLayer}/\code{SandwichLayer}
##' @param object \code{PreSandwichLayer} or \code{SandwichLayer}
##' @return \code{object}, invisibly
##' @keywords internal
.show_layer <- function(object) {
  print(object@.Data)
  invisible(object)
}

##' @title Show a \code{PreSandwichLayer} or \code{SandwichLayer}
##'
##' @description Display information about a \code{PreSandwichLayer} or
##'   \code{SandwichLayer} object
##'
##' @param object \code{PreSandwichLayer} or \code{SandwichLayer} object
##' @return an invisible copy of \code{object}
##' @export
##' @rdname PreSandwichLayer.show
setMethod("show", "PreSandwichLayer", .show_layer)

setGeneric("subset")

##' @title \code{PreSandwichLayer} and \code{SandwichLayer} subsetting
##'
##' @description Return subset of a \code{PreSandwichLayer} or
##'   \code{SandwichLayer} which meets conditions.
##'
##' @param subset Logical vector identifying values to keep or drop
##' @return \code{x} subset by \code{subset} or \code{i}
##' @export
##' @rdname PreSandwichLayer.subset
setMethod("subset", "PreSandwichLayer", function(x, subset) {
  x@.Data <- subset(x@.Data, subset = subset)
  x@prediction_gradient <- subset(x@prediction_gradient, subset = subset)
  return(x)
})

setGeneric("[")

##' @param x \code{PreSandwichLayer} or \code{SandwichLayer} object
##' @param i indices specifying elements to extract or replace. See
##'   \code{help("[")} for further details.
##' @export
##' @importFrom methods callNextMethod
##' @rdname PreSandwichLayer.subset
setMethod("[", "PreSandwichLayer",
          function(x, i) {
            idx <- rep(FALSE, length(x@.Data))
            idx[i] <- TRUE
            dat <- methods::callNextMethod()
            x@.Data <- dat
            x@prediction_gradient <- subset(x@prediction_gradient, subset = idx)
            return(x)

          })

##' @title (Internal) Get covariance adjustments and their gradient with respect
##'   to covariance adjustment model coefficients
##' @description \code{.make_PreSandwichLayer()} takes a fitted covariance
##'   adjustment model passed to the \code{model} argument and generates
##'   adjustments to outcomes for observations in the \code{newdata} argument.
##'   It also evaluates the gradient of the adjustments taken with respect to
##'   the coefficients at the coefficient estimates.
##' @param model a fitted model to use for generating covariance adjustment
##'   values
##' @param newdata a dataframe with columns called for in \code{model}
##' @param ... additional arguments to pass on to \code{model.frame} and
##'   \code{model.matrix}. These cannot include \code{na.action}, \code{xlev},
##'   or \code{contrasts.arg}: the former is fixed to be \code{na.pass}, while
##'   the latter two are provided by elements of the \code{model} argument.
##' @return A \code{PreSandwichLayer} object
##' @keywords internal
.make_PreSandwichLayer <- function(model, newdata = NULL, ...) {
  UseMethod(".make_PreSandwichLayer")
}

##' @keywords internal
.make_PreSandwichLayer.default <- function(model, newdata = NULL, ...) {
  model_terms <- tryCatch(
    terms(model),
    error = function(e) stop("`model` must have `terms` method", call. = FALSE)
  )

  if (is.null(newdata)) {
    mf <- .get_data_from_model("cov_adj", formula(model))
  } else if (!is.null(term <- attr(newdata, "terms"))) {
    mf <- newdata
  } else {
    mf <- stats::model.frame(stats::delete.response(terms(model)),
                             data = newdata,
                             na.action = na.pass,
                             xlev = model$xlevels,
                             ...)
  }

  # contrasts.arg also follows `predict.lm`
  X <- stats::model.matrix(stats::delete.response(terms(model)),
                           data = mf,
                           contrasts.arg = model$contrasts,
                           ...)

  # use the `stats` package's method for handling model fits not of full rank
  p <- model$rank
  p1 <- seq_len(p)
  piv <- if(p) model$qr$pivot[p1]
  if(p < ncol(X)) {
    X <- X[, piv, drop = FALSE]
  }

  # TODO: support predict(..., type = "response"/"link"/other?)
  xb <- drop(X %*% model$coefficients[piv])

  return(new("PreSandwichLayer",
             xb,
             fitted_covariance_model = model,
             prediction_gradient = X))
}
.S3method(".make_PreSandwichLayer", "default", .make_PreSandwichLayer.default)

##' @keywords internal
.make_PreSandwichLayer.glm <- function(model, newdata = NULL, ...) {
  model_terms <- tryCatch(
    terms(model),
    error = function(e) stop("`model` must have `terms` method", call. = FALSE)
  )

  if (is.null(newdata)) {
    mf <- .get_data_from_model("cov_adj", formula(model))
  } else if (!is.null(term <- attr(newdata, "terms"))) {
    mf <- newdata
  } else {
    mf <- stats::model.frame(stats::delete.response(terms(model)),
                             data = newdata,
                             na.action = na.pass,
                             xlev = model$xlevels,
                             ...)
  }

  X <- stats::model.matrix(stats::delete.response(terms(model)),
                           data = mf,
                           contrasts.arg = model$contrasts,
                           ...)

  p <- model$rank
  p1 <- seq_len(p)
  piv <- if(p) model$qr$pivot[p1]
  if(p < ncol(X)) {
    X <- X[, piv, drop = FALSE]
  }

  # TODO: support predict(..., type = "response"/"link"/other?)
  xb <- drop(X %*% model$coefficients[piv])

  return(new("PreSandwichLayer",
             model$family$linkinv(xb),
             fitted_covariance_model = model,
             prediction_gradient = model$family$mu.eta(xb) * X))
}
.S3method(".make_PreSandwichLayer", "glm", .make_PreSandwichLayer.glm)

##' @keywords internal
.make_PreSandwichLayer.glmrob <- function(model, newdata = NULL, ...) {
  model_terms <- tryCatch(
    terms(model),
    error = function(e) stop("`model` must have `terms` method", call. = FALSE)
  )

  if (is.null(newdata)) {
    mf <- .get_data_from_model("cov_adj", formula(model))
  } else if (!is.null(term <- attr(newdata, "terms"))) {
    mf <- newdata
  } else {
    mf <- stats::model.frame(stats::delete.response(terms(model)),
                             data = newdata,
                             na.action = na.pass,
                             xlev = model$xlevels,
                             ...)
  }

  X <- stats::model.matrix(stats::delete.response(terms(model)),
                           data = mf,
                           contrasts.arg = model$contrasts,
                           ...)

  QR <- qr(model$w.r * stats::model.matrix(model))
  p <- QR$rank
  p1 <- seq_len(p)
  piv <- if(p) QR$pivot[p1]
  if(p < ncol(X)) {
    X <- X[, piv, drop = FALSE]
  }

  # TODO: support predict(..., type = "response"/"link"/other?)
  xb <- drop(X %*% model$coefficients[piv])

  return(new("PreSandwichLayer",
             model$family$linkinv(xb),
             fitted_covariance_model = model,
             prediction_gradient = model$family$mu.eta(xb) * X))
}
.S3method(".make_PreSandwichLayer", "glmrob", .make_PreSandwichLayer.glmrob)


##' @title Convert a \code{PreSandwichLayer} to a \code{SandwichLayer} with a
##'   \code{StudySpecification} object
##' @description \code{as.SandwichLayer()} uses the \code{StudySpecification}
##'   object passed to the \code{specification} argument to populate the slots
##'   in a \code{SandwichLayer} object that a \code{PreSandwichLayer} does not
##'   have sufficient information for.
##' @param x a \code{PreSandwichLayer} object
##' @param specification a \code{StudySpecification} object
##' @param by optional; a string or named vector of unique identifier columns in
##'   the data used to create \code{specification} and the data used to fit the
##'   covariance adjustment model. Default is NULL, in which case unit of
##'   assignment columns are used for identification (even if they do not
##'   uniquely identify units of observation). If a named vector is provided,
##'   names should represent variables in the data used to create
##'   \code{specification}, while values should represent variables in the
##'   covariance adjustment data.
##' @param Q_data dataframe of direct adjustment sample, which is needed to
##'   generate the \code{keys} slot of the \code{SandwichLayer} object. Defaults
##'   to NULL, in which case if \code{by} is NULL, the data used to create
##'   \code{specification} is used, and if \code{by} is not NULL, appropriate
##'   data further up the call stack (passed as arguments to \code{cov_adj()} or
##'   \code{lmitt.formula()}, for example) is used.
##' @return a \code{SandwichLayer} object
##' @export
as.SandwichLayer <- function(x, specification, by = NULL, Q_data = NULL) {
  if (!inherits(x, "PreSandwichLayer")) {
    stop("x must be a `PreSandwichLayer` object")
  }

  if (is.null(by)) {
    by <- stats::setNames(var_names(specification, "u"),
                          var_names(specification, "u"))
    if (is.null(Q_data)) Q_data <- specification@structure
  } else {
    if (is.null(names(by))) {
      names(by) <- by
    } else if (!is.null(names(by))) {
      names(by)[names(by) == ""] <- by[names(by) == ""]
    }

    if (is.null(Q_data)) {
      form <- .update_ca_model_formula(x@fitted_covariance_model,
                                       by,
                                       specification)
      Q_data <- tryCatch(
        .get_data_from_model("cov_adj", form),
        error = function(e) {
          warning(paste("Could not find direct adjustment data in the call",
                        "stack, or it did not contain the columns specified in",
                        "`by`. Searching for `names(by)` in",
                        "specification@structure. Supply the direct",
                        "adjustment data to the `Q_data` argument when using",
                        "`by` to avoid this error."),
                  call. = FALSE)
          tryCatch({
            specification@structure[, names(by), drop = FALSE]
          }, error = function(e) {
            stop(paste("Could not find columns specified in `by` in",
                       "specification@structure"),
                 call. = FALSE)
          })

        })
    }
  }

  data_call <- x@fitted_covariance_model$call$data
  if (is.null(data_call)) {
    stop(paste("The fitted covariance adjustment model for x must be fit using",
               "a `data` argument"))
  }

  if (specification@unit_of_assignment_type == "none") {
    keys <- data.frame(..uoa.. = rownames(x@fitted_covariance_model$model))
  } else {
    keys <- tryCatch(
      stats::expand.model.frame(x@fitted_covariance_model,
                                by,
                                na.expand = TRUE)[by],
      error = function(e) {
        covmoddata <- eval(
          data_call,
          envir = environment(formula(x@fitted_covariance_model))
        )
        stop(paste("Columns",
                   paste(setdiff(by, colnames(covmoddata)), collapse = ", "),
                   "are missing from the covariance adjustment model dataset"),
           call. = FALSE)
      })
  }

  Q_by_vals <- unique(
    apply(
      Q_data[, names(by), drop = FALSE],
      1,
      function(r) {
        if (all(is.na(r))) return(NA_character_) else paste(r, collapse = "_")
      }
    )
  )
  keys$in_Q <- apply(
    keys[, unname(by), drop = FALSE], 1, function(r) paste(r, collapse = "_")
  ) %in% Q_by_vals

  return(new("SandwichLayer",
             x,
             keys = keys,
             StudySpecification = specification))
}

#' @title (Internal) Return ID's used to order observations in the covariance
#'   adjustment sample
#' @param x a \code{SandwichLayer} object
#' @param id_col character vector or list; optional. Specifies column names that
#'   appear in botn the covariance adjustment and direct adjustmet samples.
#'   Defaults to NULL, in which case unit of assignment columns in the
#'   \code{SandwichLayer}'s \code{StudySpecification} slot will be used to
#'   generate ID's.
#' @param ... arguments passed to methods
#' @return A vector of length equal to the number of units of observation used
#'   to fit the covariance adjustment model
#' @keywords internal
.sanitize_C_ids <- function(x, id_col = NULL, sorted = FALSE, ...) {
  if (!inherits(x, "SandwichLayer")) {
    stop("x must be a `SandwichLayer` object")
  }
  if (is.null(id_col)) {
    id_col <- var_names(x@StudySpecification, "u")
  }

  C_ids <- tryCatch(
    x@keys[, id_col, drop = FALSE],
    error = function(e) {
      tryCatch({
        camod <- x@fitted_covariance_model
        stats::expand.model.frame(camod, id_col, na.expand = TRUE)[id_col]
      },
      error = function(e) {
        camod <- x@fitted_covariance_model
        data <- eval(camod$call$data, envir = environment(formula(camod)))
        stop(paste("The columns",
                   paste(setdiff(id_col, colnames(data)), collapse = ", "),
                   "could not be found in either the SandwichLayer's `keys`",
                   "slot or the covariance adjustment model dataset"),
             call. = FALSE)
      })
    })

  nas <- apply(is.na(C_ids), 1, all)
  out <- apply(C_ids, 1, function(...) paste(..., collapse = "_"))
  out[nas] <- NA_character_
  names(out) <- NULL

  if (sorted) {
    if (suppressWarnings(any(is.na(as.numeric(out))))) {
      out <- sort(out, index.return = TRUE)
    } else {
      out <- sort(as.numeric(out), index.return = TRUE)
    }
  }

  return(out)
}

test_that("PreSandwichLayer covariance model incompatible with model.matrix", {
  N <- 100
  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  invalid_cmod <- list("a" = c(1, 2, 3),
                       "terms" = list("b" = c(4, 5, 6)))

  expect_error(new("PreSandwichLayer",
                   offset,
                   fitted_covariance_model = invalid_cmod,
                   prediction_gradient = pred_gradient),
               "must have a valid 'terms' attribute")
})

test_that("PreSandwichLayer covariance model incompatible with sandwich package", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x, df)
  class(cmod) <- "new_lm"

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  expect_error(new("PreSandwichLayer",
                   offset,
                   fitted_covariance_model = cmod,
                   prediction_gradient = pred_gradient),
               "extracting vcov elements not applicable")
})

test_that("PreSandwichLayer prediction gradient is not a numeric matrix", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x, df)

  offset <- rep(1, N)
  pred_gradient <- as.matrix(cbind(1, as.character(df$x)))

  expect_error(new("PreSandwichLayer",
                   offset,
                   fitted_covariance_model = cmod,
                   prediction_gradient = pred_gradient),
               "must be a numeric matrix")
})

test_that("PreSandwichLayer prediction gradient has invalid number of rows", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x, df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N - 1, ncol = 2)

  expect_error(new("PreSandwichLayer",
                   offset,
                   fitted_covariance_model = cmod,
                   prediction_gradient = pred_gradient),
               "same dimension along axis 1")
})

test_that("PreSandwichLayer prediction gradient has invalid number of columns", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x, df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 1)

  expect_error(new("PreSandwichLayer",
                   offset,
                   fitted_covariance_model = cmod,
                   prediction_gradient = pred_gradient),
               "same number of columns")
})

test_that("SandwichLayer has NA's", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unitid(uid), df)
  keys <- df[, "uid", drop = FALSE]

  offset <- rep(1, N)
  offset[N] <- NA_real_
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  pred_gradient[N,] <- NA_real_
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_warning(new("SandwichLayer",
                     psl,
                     keys = keys,
                     StudySpecification = spec),
                 "adjustments are NA")
})

test_that("SandwichLayer keys doesn't have the same row count as covariance model data", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unitid(uid), df)
  keys <- df[seq_len(N-1), "uid", drop = FALSE]

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_error(new("SandwichLayer",
                   psl,
                   keys = keys,
                   StudySpecification = spec),
               "to fit the covariance adjustment model")
})

test_that("SandwichLayer created correctly", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unitid(uid), df)
  keys <- df[, "uid", drop = FALSE]

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_true(inherits(new("SandwichLayer",
                     psl,
                     keys = keys,
                     StudySpecification = spec),
                 "SandwichLayer"))

  keys$uid <- NA_integer_

  expect_true(inherits(new("SandwichLayer",
                     psl,
                     keys = keys,
                     StudySpecification = spec),
                 "SandwichLayer"))
})

test_that("as.SandwichLayer not called with a PreSandwichLayer", {
  set.seed(20)
  N <- 100
  df <- data.frame("z" = rbinom(N, 1, 0.5), "uid" = seq_len(N))
  spec <- rct_spec(z ~ unitid(uid), df)

  expect_error(as.SandwichLayer(seq_len(N), spec),
               "must be a `PreSandwichLayer`")
})

test_that("as.SandwichLayer not fit with a data argument", {
  set.seed(20)
  N <- 100
  x <- rnorm(N); y <- rnorm(N)
  specification_df <- data.frame("z" = rbinom(N, 1, 0.5), "uid" = seq_len(N))
  cmod <- lm(y ~ x)
  spec <- rct_spec(z ~ unitid(uid), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_error(as.SandwichLayer(psl, spec),
               "must be fit using a `data` argument")
})

test_that("as.SandwichLayer missing specvar columns from covariance model data", {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  specification_df <- data.frame("z" = rbinom(N, 1, 0.5), "uid" = seq_len(N))
  cmod <- lm(y ~ x, cmod_df)
  spec <- rct_spec(z ~ unitid(uid), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_error(as.SandwichLayer(psl, spec),
               "Columns uid are missing")
})

test_that("as.SandwichLayer used correctly with NULL `by` and `Q_data`", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unit_of_assignment(uid), df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  expect_true(inherits(sl1 <- as.SandwichLayer(psl, spec), "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl1@keys), c("uid", "in_Q"))), 0)
  expect_true(all(sl1@keys$in_Q == 1))
  expect_equal(sl1@keys$uid, df$uid)

  spec <- rct_spec(z ~ cluster(uid), df)
  expect_true(inherits(sl2 <- as.SandwichLayer(psl, spec), "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl2@keys), c("uid", "in_Q"))), 0)
  expect_true(all(sl2@keys$in_Q == 1))
  expect_equal(sl2@keys$uid, df$uid)

  spec <- rct_spec(z ~ unitid(uid), df)
  expect_true(inherits(sl3 <- as.SandwichLayer(psl, spec), "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl3@keys), c("uid", "in_Q"))), 0)
  expect_true(all(sl3@keys$in_Q == 1))
  expect_equal(sl3@keys$uid, df$uid)
})

test_that("as.SandwichLayer matches on `by` column when uoa columns don't (`by` has no names)", {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame(x = rnorm(N), y = rnorm(N), uid = seq_len(N), clust = rep(NA_integer_, N))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame(uid = seq_len(N), clust = rep(c(1, 2), each = N/2),
                          z = rep(c(0, 1), each = N/2))
  spec <- rct_spec(z ~ unit_of_assignment(clust), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  sl <- as.SandwichLayer(psl, spec, "uid", specification_df)

  expect_true(inherits(sl, "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl@keys), c("clust", "uid", "in_Q"))), 0)
  expect_true(all(sl@keys$in_Q))
})

test_that("as.SandwichLayer matches on `by` column when uoa columns don't (`by` has names)", {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame(x = rnorm(N), y = rnorm(N), uid = seq_len(N), clust = rep(NA_integer_, N))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame(uoa1 = seq_len(N), clust = rep(c(1, 2), each = N/2),
                          z = rep(c(0, 1), each = N/2))
  spec <- rct_spec(z ~ cluster(clust), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  by <- c("uoa1" = "uid")
  sl <- as.SandwichLayer(psl, spec, by, specification_df)

  expect_true(inherits(sl, "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl@keys), c("clust", "uid", "in_Q"))), 0)
  expect_true(all(sl@keys$in_Q))
})

test_that("as.SandwichLayer used correctly with unnamed `by` and NULL `Q_data`", {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "uoa1" = seq_len(N))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame("uoa1" = seq_len(N), "z" = rbinom(N, 1, 0.5))
  spec <- rct_spec(z ~ unit_of_assignment(uoa1), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  expect_warning(
    expect_warning(
      expect_warning(
        sl <- as.SandwichLayer(psl, spec, "uoa1"),
        "No call to"
      ),
      "Unable to detect"
    ),
    "Could not find direct adjustment data"
  )

  expect_true(inherits(sl, "SandwichLayer"))
  expect_equal(length(setdiff(colnames(sl@keys), c("uoa1", "in_Q"))), 0)
  expect_true(all(sl@keys$in_Q == 1))
  expect_equal(sl@keys$uoa1, cmod_df$uoa1)
})

test_that("as.SandwichLayer failed by", {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "uoa1" = seq_len(N))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame("uoa1" = seq_len(N), "z" = rbinom(N, 1, 0.5))
  spec <- rct_spec(z ~ unit_of_assignment(uoa1), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  expect_error(
    expect_warning(
      expect_warning(
        expect_warning(
          as.SandwichLayer(psl, specification = spec, by = "not_uoa"),
          "No call to"
        ),
        "Unable to detect"
      ),
      "Could not find direct adjustment data"
    ),
    "Could not find columns"
  )
})

test_that(paste("as.SandwichLayer produces correct ID's for univariate uoa ID's",
                "not starting at 1"), {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame("x" = rnorm(N), "y" = rnorm(N),
                        "uid" = c(paste0("0400", seq_len(floor(N/2))),
                                  rep(NA_character_, N - floor(N/2))))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame("uid" = paste0("0400", seq_len(N)),
                          "z" = rbinom(N, 1, 0.5))
  spec <- rct_spec(z ~ unit_of_assignment(uid), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  sl <- as.SandwichLayer(psl, spec)

  expect_equal(nrow(sl@keys),
               nrow(sl@fitted_covariance_model$model))
  expect_equal(sl@keys[, 1],
               c(paste0("0400", seq_len(floor(N/2))),
                 rep(NA_character_, N - floor(N/2))))
})

test_that(paste("as.SandwichLayer produces correct ID's for univariate uoa ID's",
                "not starting at 1"), {
  set.seed(20)
  N <- 100
  cmod_df <- data.frame("x" = rnorm(N), "y" = rnorm(N),
                        "classid" = c(rep(c(1, 2), each = floor(N/4)),
                                      rep(NA_character_, N - floor(N/2))),
                        "schoolid" = c(rep("04001", 2 * floor(N/4)),
                                       rep(NA_character_, N - floor(N/2))))
  cmod <- lm(y ~ x, cmod_df)
  specification_df <- data.frame("classid" = rep(c(1, 2), each = floor(N/4)),
                          "schoolid" = rep("04001", 2 * floor(N/4)),
                          "z" = rep(c(0, 1), each = floor(N/4)))
  spec <- rct_spec(z ~ cluster(classid, schoolid), specification_df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)

  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  sl <- as.SandwichLayer(psl, spec)

  expected_keys <- c(rep(paste0("04001_", c(1, 2)), each = floor(N/4)),
                     rep(NA_character_, N - 2 * floor(N/4)))

  expect_equal(nrow(sl@keys),
               nrow(sl@fitted_covariance_model$model))
  expect_equal(sl@keys[, "classid"],
               c(rep(c(1, 2), each = floor(N/4)),
                 rep(NA_character_, N - floor(N/2))))
  expect_equal(sl@keys[, "schoolid"],
               c(rep("04001", 2 * floor(N/4)),
                 rep(NA_character_, N - floor(N/2))))
})

test_that("as.SandwichLayer correctly marks in_Q rows with multiple columns", {
  set.seed(42)
  N <- 20
  cmod_df <- data.frame(
    x = rnorm(N),
    y = rnorm(N),
    classid = rep(c(1, 1, 2, 2), each = N / 4),
    schoolid = rep(c("A", "B", "A", "B"), each = N / 4)
  )
  cmod <- lm(y ~ x, cmod_df)

  Q_data <- data.frame(
    classid = c(1, 2),
    schoolid = c("A", "B"),
    z = c(0, 1)
  )
  spec <- rct_spec(z ~ cluster(classid, schoolid), Q_data)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  sl <- as.SandwichLayer(psl, spec, Q_data = Q_data)

  # only rows with classid=1 and schoolid=A, or classid=2 and schoolid=B, should
  # be in_Q
  expected_in_Q <- (cmod_df$classid == 1 & cmod_df$schoolid == "A") |
    (cmod_df$classid == 2 & cmod_df$schoolid == "B")
  expect_equal(sl@keys$in_Q, expected_in_Q)

  expect_equal(sum(sl@keys$in_Q), N / 2)
})

test_that("show_sandwich_layer works", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unitid(uid), df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)

  out <- capture.output(show(as.vector(psl)))
  caout <- capture.output(show(psl))
  expect_identical(out, caout)

  sl <- as.SandwichLayer(psl, spec)
  out <- capture.output(show(as.vector(sl)))
  caout <- capture.output(show(sl))
  expect_identical(out, caout)
})

test_that("subsetting PreSandwich and SandwichLayer works", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N), "z" = rbinom(N, 1, 0.5),
                   "uid" = seq_len(N))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(z ~ unitid(uid), df)

  offset <- rep(1, N)
  pred_gradient <- matrix(1, nrow = N, ncol = 2)
  psl <- new("PreSandwichLayer",
             offset,
             fitted_covariance_model = cmod,
             prediction_gradient = pred_gradient)
  keys <- df[, "uid", drop = FALSE]
  sl <- new("SandwichLayer",
            psl,
            keys = keys,
            StudySpecification = spec)

  no_subset_psl <- subset(psl, rep(TRUE, length(offset)))
  no_subset_sl <- subset(sl, rep(TRUE, length(offset)))
  expect_identical(no_subset_psl@.Data, psl@.Data)
  expect_identical(no_subset_psl@prediction_gradient,
                   psl@prediction_gradient[1:100,])
  expect_identical(no_subset_sl@.Data, sl@.Data)
  expect_identical(no_subset_sl@prediction_gradient,
                   sl@prediction_gradient[1:100,])

  expect_true(inherits(no_subset_psl, "PreSandwichLayer"))
  expect_true(inherits(no_subset_sl, "SandwichLayer"))

  expect_identical(no_subset_psl@fitted_covariance_model,
                   psl@fitted_covariance_model)
  expect_identical(no_subset_sl@fitted_covariance_model,
                   sl@fitted_covariance_model)

  subset_length <- floor(length(offset) / 2)
  subset_psl <- subset(psl, c(rep(TRUE, subset_length),
                              rep(FALSE, length(offset) - subset_length)))
  subset_sl <- subset(sl, c(rep(TRUE, subset_length),
                            rep(FALSE, length(offset) - subset_length)))
  expect_identical(subset_psl@.Data, psl@.Data[1:subset_length])
  expect_identical(subset_sl@.Data, sl@.Data[1:subset_length])

  expect_true(inherits(subset_psl, "PreSandwichLayer"))
  expect_true(inherits(subset_sl, "SandwichLayer"))

  expect_identical(subset_psl@fitted_covariance_model,
                   psl@fitted_covariance_model)
  expect_identical(subset_psl@prediction_gradient,
                   psl@prediction_gradient[1:subset_length,])
  expect_identical(subset_sl@keys, sl@keys)
  expect_identical(subset_sl@StudySpecification, sl@StudySpecification)

  no_subset_psl <- psl[]
  no_subset_sl <- sl[]
  expect_identical(no_subset_psl@.Data, psl@.Data)
  expect_identical(no_subset_sl@.Data, sl@.Data)

  expect_true(inherits(no_subset_psl, "PreSandwichLayer"))
  expect_true(inherits(no_subset_sl, "SandwichLayer"))

  expect_identical(psl[1:10]@.Data, psl@.Data[1:10])
  expect_identical(sl[1:10]@.Data, sl@.Data[1:10])
  expect_identical(psl[1:10]@prediction_gradient,
                   psl@prediction_gradient[1:10,])
  expect_identical(sl[1:10]@prediction_gradient,
                   sl@prediction_gradient[1:10,])
})

test_that(".make_PreSandwichLayer newdata is a matrix", {
  set.seed(20)
  N <- 100
  df <- data.frame("x1" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x1, df)
  expect_error(.make_PreSandwichLayer(cmod, matrix(1)),
               "must be a data.frame")
})

test_that(".make_PreSandwichLayer model doesn't have a terms method", {
  expect_error(.make_PreSandwichLayer(list("coefficients" = c(1., 1.)),
                                               data.frame("x" = 1)),
               "must have `terms`")
})

test_that(".make_PreSandwichLayer model frame missing cmod columns", {
  set.seed(20)
  N <- 100
  df <- data.frame("x1" = rnorm(N), "x2" = rnorm(N), "y" = rnorm(N))
  pred_df <- data.frame("x1" = rnorm(N))
  cmod <- lm(y ~ x1 + x2, df)

  expect_error(.make_PreSandwichLayer(cmod, pred_df),
               "'x2' not found")
})

test_that(".make_PreSandwichLayer returns expected output for `lm` object", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod <- lm(y ~ x, df)
  psl <- .make_PreSandwichLayer(cmod, df)

  expect_true(all.equal(psl@.Data, cmod$fitted.values, check.attributes = FALSE))
  expect_true(all.equal(psl@prediction_gradient, stats::model.matrix(cmod),
                        check.attributes = FALSE))
})

test_that(".make_PreSandwichLayer returns expected output for `lm` object with new data", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  pred_df <- data.frame("x" = rnorm(N))
  cmod <- lm(y ~ x, df)
  psl <- .make_PreSandwichLayer(cmod, pred_df)

  expect_true(all.equal(
    psl@.Data,
    drop(stats::model.matrix(formula(stats::delete.response(terms(cmod))), pred_df) %*%
           cmod$coefficients),
    check.attributes = FALSE
  ))
  expect_true(all.equal(
    psl@prediction_gradient,
    stats::model.matrix(formula(stats::delete.response(terms(cmod))), pred_df)
  ))
})

test_that(".make_PreSandwichLayer returns expected output when formula is a symbol", {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  cmod_form <- y ~ x
  cmod <- lm(cmod_form, df)
  psl <- .make_PreSandwichLayer(cmod, df)

  expect_true(all.equal(psl@.Data, cmod$fitted.values, check.attributes = FALSE))
  expect_true(all.equal(psl@prediction_gradient, stats::model.matrix(cmod),
                        check.attributes = FALSE))
})

test_that(paste(".make_PreSandwichLayer returns expected output",
                "when variables are modified in the formula"), {
  set.seed(20)
  N <- 100
  df <- data.frame("x1" = rnorm(N), "x2" = rpois(N, 1) + 1, "y" = rnorm(N))
  cmod_form <- y ~ stats::poly(x1, 3) + log(x2)
  cmod <- lm(cmod_form, data = df)
  psl <- .make_PreSandwichLayer(cmod, df)

  expect_true(all.equal(psl@.Data, cmod$fitted.values, check.attributes = FALSE))
  expect_true(all.equal(psl@prediction_gradient, stats::model.matrix(cmod),
                        check.attributes = FALSE))
})

test_that(".make_PreSandwichLayer returns expected output for `glm` object", {
  set.seed(20)
  N <- 100
  df <- data.frame("x1" = rnorm(N), "x2" = rpois(N, 1) + 1, "y" = rnorm(N))
  cmod <- glm(x2 ~ x1, data = df, family = stats::poisson())
  mm <- stats::model.matrix(cmod)
  psl <- .make_PreSandwichLayer(cmod, df)

  expect_true(all.equal(psl@.Data, drop(exp(mm %*% cmod$coefficients)),
                        check.attributes = FALSE))
  expect_true(all.equal(psl@prediction_gradient,
                        cmod$family$mu.eta(drop(mm %*% cmod$coefficients)) * mm,
                        check.attributes = FALSE))
})

test_that(paste(".make_PreSandwichLayer returns expected output",
                "for `glm` object with new data"), {
  set.seed(20)
  N <- 100
  df <- data.frame("x1" = rnorm(N), "x2" = rpois(N, 1) + 1, "y" = rnorm(N))
  cmod <- glm(x2 ~ x1, data = df, family = stats::poisson())
  pred_df <- data.frame("x1" = rnorm(N))
  mm <- stats::model.matrix(formula(stats::delete.response(terms(cmod))),
                            data = pred_df)
  psl <- .make_PreSandwichLayer(cmod, pred_df)

  expect_true(all.equal(psl@.Data, drop(exp(mm %*% cmod$coefficients)),
                        check.attributes = FALSE))
  expect_true(all.equal(psl@prediction_gradient,
                        cmod$family$mu.eta(drop(mm %*% cmod$coefficients)) * mm,
                        check.attributes = FALSE))
})

if (requireNamespace("robustbase", quietly = TRUE)) {
  test_that(".make_PreSandwichLayer glmrob", {
    set.seed(30)
    x = sample(rep(c(-1, 1), each = 15))
    y = c(0.5 * x[1:29] - 2 + 1e-2 * rnorm(29), 12 * x[30] + 1e-2 * rnorm(1))
    moddata <- data.frame(x = x, y = y)

    suppressWarnings(
      mod <- robustbase::glmrob(y ~ x, moddata, family = stats::gaussian(),
                                control = robustbase::glmrobMqle.control(tcc = 1.5))
    )

    expect_true(!all(mod$w.r == 1.))
    expect_true(all.equal(.make_PreSandwichLayer(mod, moddata)@.Data,
                          mod$fitted.values,
                          check.attributes = FALSE))
  })
}

test_that(paste(".make_PreSandwichLayer returns expected output when",
                "NA's are present"), {
  set.seed(20)
  N <- 100
  df <- data.frame("x" = rnorm(N), "y" = rnorm(N))
  pred_df <- data.frame("x" = rnorm(N))
  cmod <- lm(y ~ x, df)

  pred_df[N, c("x")] <- NA_real_
  psl <- .make_PreSandwichLayer(cmod, pred_df)
  pred_gradient <- stats::model.matrix(
    formula(stats::delete.response(terms(cmod))),
    stats::model.frame(pred_df, na.action = na.pass))

  expect_equal(length(psl@.Data), N)
  expect_equal(sum(is.na(psl@.Data)), 1)
  expect_equal(dim(psl@prediction_gradient), dim(pred_gradient))
  expect_true(all.equal(psl@prediction_gradient[seq_len(N-1),],
                        pred_gradient[seq_len(N-1),],
                        check.attributes = FALSE))
})

test_that(".make_PreSandwichLayer with transformed columns in formula", {
  set.seed(39)
  cmod_dat <- data.frame(uid = NA_character_, x = rnorm(12), y = rnorm(12))
  cmod <- lm(y ~ scale(x, scale = FALSE), cmod_dat)
  spec_dat <- data.frame(uid = seq_len(10), x = rnorm(10),
                         a = rep(c(0, 1), 5), y = rnorm(10))
  spec <- rct_spec(a ~ unitid(uid), spec_dat)
  form <- .update_ca_model_formula(cmod, by = NULL, spec)
  newdata <- stats::model.frame(form, data = cmod_dat, na.action = na.pass)
  
  psl <- .make_PreSandwichLayer(cmod, newdata)
  expect_equal(colnames(psl@prediction_gradient),
               names(cmod$coefficients)[!is.na(cmod$coefficients)])
  expect_equal(psl@prediction_gradient,
               stats::model.matrix(stats::formula(cmod), cmod_dat))
  
  sl <- lmitt(y ~ 1, spec, spec_dat, offset = cov_adj(cmod))$model$`(offset)`
  expect_equal(colnames(sl@prediction_gradient),
               names(cmod$coefficients)[!is.na(cmod$coefficients)])
  expect_equal(sl@prediction_gradient,
               stats::model.matrix(formula(cmod), spec_dat))
  
  ## glm
  cmod_dat$y <- c(rep(0, 6), rep(1, 6))
  spec_dat$y <- c(rep(0, 5), rep(1, 5))
  cmod <- glm(y ~ scale(x, scale = FALSE), cmod_dat,
              family = stats::binomial())
  form <- .update_ca_model_formula(cmod, by = NULL, spec)
  newdata <- stats::model.frame(form, data = cmod_dat, na.action = na.pass)
  
  psl <- .make_PreSandwichLayer(cmod, newdata)
  expect_equal(colnames(psl@prediction_gradient),
               names(cmod$coefficients)[!is.na(cmod$coefficients)])
  sl <- lmitt(y ~ 1, spec, spec_dat, offset = cov_adj(cmod))$model$`(offset)`
  expect_equal(colnames(sl@prediction_gradient),
               names(cmod$coefficients)[!is.na(cmod$coefficients)])
  
})

test_that(".sanitize_C_ids fails with invalid `cluster` argument", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))

  expect_error(.sanitize_C_ids(ssmod$model$`(offset)`, id_col = "uid"),
               "uid could not be found")
})

test_that(".sanitize_C_ids with full UOA info", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))

  ids <- .sanitize_C_ids(ssmod$model$`(offset)`)
  expected_ids <- apply(simdata[, c("uoa1", "uoa2")], 1, function(...) paste(..., collapse = "_"))
  expect_equal(ids, expected_ids)
})

test_that(".sanitize_C_ids with partial UOA info", {
  data(simdata)
  cmod_data <- data.frame("x" = rnorm(10), "y" = rnorm(10),
                          "uoa1" = rep(c(1, 2), each = 5),  "uoa2" = NA)

  cmod <- lm(y ~ x, cmod_data)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, specification = spec,
                offset = cov_adj(cmod))

  ids <- .sanitize_C_ids(ssmod$model$`(offset)`)
  expect_equal(length(ids), nrow(cmod_data))
  expect_equal(length(unique(ids)), 2)
})

test_that(".sanitize_C_ids with no UOA info", {
  data(simdata)
  cmod_data <- data.frame("x" = rnorm(10), "y" = rnorm(10), "uoa1" = NA,  "uoa2" = NA)

  cmod <- lm(y ~ x, cmod_data)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, specification = spec,
                offset = cov_adj(cmod))

  ids <- .sanitize_C_ids(ssmod$model$`(offset)`)
  expect_equal(length(ids), nrow(cmod_data))
  expect_equal(length(unique(ids)), 1)
})

test_that(".sanitize_C_ids miscellaneous errors", {
  expect_error(.sanitize_C_ids(2), "x must be a `SandwichLayer`")

  n <- 10
  df <- data.frame("x" = rnorm(n), "a" = rep(c(0, 1), each = 5), "y" = rnorm(n),
                   "cid" = sample(seq_len(n)))
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(a ~ cluster(cid), df)
  sl <- cov_adj(cmod, newdata = df, specification = spec)
  num_C_ids <- .sanitize_C_ids(sl, sorted = TRUE)
  expect_true(all.equal(num_C_ids$x, seq_len(n), check.attributes = FALSE))

  df$cid <- sample(letters[1:n])
  cmod <- lm(y ~ x, df)
  spec <- rct_spec(a ~ cluster(cid), df)
  sl <- cov_adj(cmod, newdata = df, specification = spec)
  char_C_ids <- .sanitize_C_ids(sl, sorted = TRUE)
  expect_true(all.equal(char_C_ids$x, letters[1:n], check.attributes = FALSE))
})

##' \code{assigned()}/\code{ate()}/\code{ett()}/\code{cov_adj()} all need the
##' \code{StudySpecification} to operate. If any are called in the model without
##' a \code{specification=} argument, this function sees if it can find the
##' \code{StudySpecification} in another of these functions.
##'
##' Note that it will never look inside \code{assigned()} (gets complicated in
##' formulas), only in weights or \code{cov_adj()}. E.g.
##'
##' \code{lm(y ~ assigned(), weights = ate(spec), offest = cov_adj(mod1))}
##'
##' \code{lm(y ~ assigned(), weights = ate(), offest = cov_adj(mod1, specification = spec))}
##'
##' will both work, but
##'
##' \code{lm(y ~ assigned(spec), weights = ate(), offest = cov_adj(mod1))}
##'
##' will fail.
##' @title (Internal) Locate a \code{StudySpecification} in the call stack
##' @param NULL_on_error if \code{TRUE}, returns \code{NULL} if a
##'   \code{StudySpecification} object is not found rather than throwing an
##'   error.
##' @return A \code{StudySpecification}, or \code{NULL} if \code{NULL_on_error}
##'   is \code{TRUE} and the \code{StudySpecification} can't be found.
##' @keywords internal
.get_spec <- function(NULL_on_error = FALSE) {
  specification <- NULL

  # The plain `lmitt` call may contain a specification formula, only in `lmitt.formula`
  # will that be converted into an actual specification. Want to return this immediately
  # to prevent infinite recursion
  found_lmitt <- grepl("^lmitt\\.formula$",
                       lapply(sys.calls(), "[[", 1),
                       perl = TRUE)

  if (any(found_lmitt)) {
    specification <- get("specification", sys.frame(which(found_lmitt)[1]))
    if (inherits(specification, "StudySpecification")) {
      return(specification)
    } else {
      specification <- NULL
    }
  }

  # #100 Checking for recusion and exiting early.
  if (sum(lapply(sys.calls(), "[[", 1) == ".get_spec") > 2) {
    stop(paste("Unable to locate StudySpecification in call stack, please use the",
               "`specification` argument to pass a StudySpecification object.",
               "(Internal: Inf Recusion error)"))
  }


  # Searching for weights or cov_adj is basically the same, except for argument
  # type
  .find.specification <- function(type) {
    stopifnot(type %in% c("weights", "offset"))

    specification <- NULL

    # Identify all frames with the appropriate argument
    keyframes <- !vapply(lapply(sys.calls(), `[[`, type), is.null, logical(1))

    # Loop over each frame which has an `type` argument.
    # Its most likely the first frame, but perhaps not.
    for (i in which(keyframes)) {
      possible_spec_holder <- get(type, sys.frame(i))
      if (inherits(possible_spec_holder, "WeightedStudySpecification") ||
          inherits(possible_spec_holder, "SandwichLayer")) {
        # If we have a WeightedStudySpecification, save it and break
        specification <- possible_spec_holder@StudySpecification
        break()
      }
    }
    if (!inherits(specification, "StudySpecification")) {
      return(NULL)
    }
    return(specification)
  }

  weight_spec <- NULL
  covadj_spec <- NULL
  mf_spec <- NULL
  emf_spec <- NULL
  # This avoids infinite recursion; if we're in weights or in cov_adj, don't
  # look for it again. Only assigned will look for both.
  if (sys.call(-1)[[1]] != ".weights_calc") {
    weight_spec <- .find.specification("weights")
  }
  if (sys.call(-1)[[1]] != "cov_adj" && sys.call(-1)[[1]] != "propertee::cov_adj") {
    covadj_spec <- .find.specification("offset")
  }

  # `model.frame` may have a teeMod object we can extract from
  mf_calls <- grepl("model\\.frame$", lapply(sys.calls(), "[[", 1), perl = TRUE)

  mf_spec <- lapply(which(mf_calls), function(x) {
    form <- get("formula", sys.frame(x))
    found_spec <- if (inherits(form, "teeMod")) {
      form@StudySpecification
    } else if (inherits(form, "terms") | inherits(form, "formula")) {
      tt <- terms(form, specials = c("assigned", "adopters", "a.", "z."))
      trt_term_ix <- vapply(attr(tt, "specials"),
                            function(x) if (!is.null(x)) x else 0L,
                            integer(1L))
      if (any(trt_term_ix)) {
        trt_term <- match.call(
          get(names(trt_term_ix)[trt_term_ix != 0], environment(assigned)),
          attr(tt, "variables")[[trt_term_ix[trt_term_ix != 0]+1L]]
        )
        tryCatch(eval(trt_term$specification, environment(form)),
                 error = function(e) NULL)
      } else {
        NULL
      }
    } else {
      NULL
    }
    if (!inherits(found_spec, "StudySpecification")) {
      found_spec <- NULL
    }
    return(found_spec)
  })

  emf_calls <- grepl("expand\\.model\\.frame$",
                     lapply(sys.calls(), "[[", 1), perl = TRUE) |
    grepl("\\.expand\\.model\\.frame_teeMod$",
          lapply(sys.calls(), "[[", 1), perl = TRUE)

  emf_spec <- lapply(which(emf_calls), function(x) {
    mod <- get("model", sys.frame(x))
    if (inherits(mod, "teeMod")) mod@StudySpecification else NULL
  })

  # At this point, each *_spec is either NULL, or a StudySpecification (as enforced by
  # .find.specification() and the special lmitt case)
  potential_specs <- append(emf_spec,
                              append(mf_spec,
                                     c(weight_spec = weight_spec,
                                       covadj_spec = covadj_spec)))

  found_specs <- !vapply(potential_specs, is.null, logical(1))

  if (!any(found_specs)) {
    # Found nothing
    if (NULL_on_error) {
      return(NULL)
    }
    stop(paste("Unable to locate StudySpecification in call stack, please use the",
               " `specification` argument to pass a StudySpecification object."))
  }
  if (sum(found_specs) > 1) {
    # Found multiple StudySpecifications; ensure they're all the same
    if (sum(!duplicated(potential_specs[found_specs])) > 1) {
      stop("Multiple differing `StudySpecification` found")
    }
  }
  # At this point, we know at least one specification is non-NULL, and if there are
  # multiple, they're duplicates, so just take the first real StudySpecification.
  return(potential_specs[found_specs][[1]])

}

#' @include StudySpecification.R StudySpecificationAccessors.R
NULL

##' @title Special terms in \code{StudySpecification} creation formula
##'
##' @description These are special functions used only in the definition of
##'   \code{StudySpecification} objects. They identify the units of assignment,
##'   blocks and forcing variables. They should never be used outside of the
##'   \code{formula} argument to \code{obs_spec}, \code{rct_spec}, or
##'   \code{rd_spec}.
##'
##' @details These functions have no use outside of the formula in creating a
##'   \code{StudySpecification}.
##'
##'   \code{unit_of_assignment}, \code{uoa}, \code{cluster} and \code{unitid}
##'   are synonyms; you must include one and only one in each
##'   \code{StudySpecification}. The choice of which to use will have no impact
##'   on any analysis, only on some output and the name of the stored element in
##'   the \code{StudySpecification}. Accessors/ replacers
##'   (\code{units_of_assignment}, \code{unitids}, \code{clusters}) respect the
##'   choice made at the point of creation of the \code{StudySpecification}, and
##'   only the appropriate function will work.
##'
##'   See \code{rct_spec}, \code{obs_spec}, or \code{rd_spec} for examples of
##'   their usage.
##'
##' @param ... any number of variables of the same length.
##' @return the variables with appropriate labels. No use outside of their
##'   inclusion in the \code{formula} argument to \code{obs_spec},
##'   \code{rct_spec}, or \code{rd_spec}
##' @export
##' @rdname StudySpecificationSpecials
unit_of_assignment <- function(...) {
  allf <- list(...)
  results <- lapply(allf, function(x) {
    if (!is.numeric(x)) return(as.character(x))
    return(x)
  })
  return(do.call(cbind, results))
}

##' @rdname StudySpecificationSpecials
##' @export
unitid <- unit_of_assignment

##' @rdname StudySpecificationSpecials
##' @export
cluster <- unit_of_assignment

##' @rdname StudySpecificationSpecials
##' @export
uoa <- unit_of_assignment

##' @rdname StudySpecificationSpecials
##' @export
block <- unit_of_assignment

##' @rdname StudySpecificationSpecials
##' @export
forcing <- unit_of_assignment

#
##' Checks performed:
##' * Ensure presence of no more than one of \code{unit_of_assignment()},
##'   \code{cluster()} or \code{unitid()}.
##' * Disallow multiple \code{block()} or multiple \code{forcing()} terms.
##' * Disallow \code{forcing()} unless in RDD.
##'
##' @title (Internal) Perform checks on formula for creation of
##'   StudySpecification.
##' @param form A formula passed to \code{.new_StudySpecification()}
##' @param allow_forcing Binary whether \code{forcing()} is allowed (\code{TRUE}
##'   for RDD, \code{FALSE} for RCT and Obs).
##' @return \code{TRUE} if all checks pass, otherwise errors.
##' @keywords internal
.check_spec_formula <- function(form, allow_forcing = FALSE) {
  tt <- terms(form, c("unit_of_assignment", "uoa", "cluster",
                      "unitid", "block", "forcing"))
  specials <- attr(tt, "specials")

  if (attr(tt, "response") == 0) {
    stop("Must specify a treatment variable as the left side of the formula.")
  }

  spec_uas <- specials$unit_of_assignment
  spec_uoa <- specials$uoa
  spec_clu <- specials$cluster
  spec_uni <- specials$unitid

  len_uas <- length(spec_uas)
  len_uoa <- length(spec_uoa)
  len_clu <- length(spec_clu)
  len_uni <- length(spec_uni)

  if (len_uas + len_uoa + len_clu + len_uni > 1) {
    # there's 2+ entered; need to figure out what combination

    if ((len_uas >= 1) + (len_uoa >= 1) + (len_clu >= 1) + (len_uni >= 1) > 1) {
      # There's more than one specified.
      stop(paste("Only one of `unit_of_assignment()`, `cluster()` or",
                 "`unitid()` can be entered."))
    } else {
      which <- switch(    (len_uoa >= 1) +
                          (len_uas >= 1) +
                      2 * (len_clu >= 1) +
                      3 * (len_uni >= 1),
                      "`unit_of_assignment()`",
                      "`cluster()`",
                      "`unitid()`")
      stop(paste0("Only one instance of ", which, " can be entered. (",
                 which, " can take multiple variables)"))
    }
  }

  if (!is.null(specials$block) && length(specials$block) > 1) {
    stop("Specify only one block() (block() can take multiple variables).")
  }

  if (!allow_forcing && !is.null(attr(tt, "specials")$forcing)) {
    stop("forcing() only allowed in rd_spec")
  } else if (allow_forcing && length(specials$forcing) > 1) {
    stop("Specify only one forcing() (forcing() can take multiple variables).")
  }

  invisible(TRUE)
}


##' Internally, we always refer to uoa/cluster/unitid as "unit_of_assignment"
##' @title (Internal) Rename cluster/unitid/uoa in a formula to
##'   unit_of_assignment for internal consistency
##' @param form A formula passed to \code{.new_StudySpecification()}
##' @return The formula with "cluster"/"unitid"/"uoa" replace with
##'   "unit_of_assignment"
##' @keywords internal
.update_form_to_unit_of_assignment <- function(form) {
  form <- deparse1(form)
  form <- gsub("cluster\\(", "unit_of_assignment(", form)
  form <- gsub("uoa\\(", "unit_of_assignment(", form)
  form <- gsub("unitid\\(", "unit_of_assignment(", form)
  return(as.formula(form))
}

##' @title Check whether treatment stored in a \code{StudySpecification} object
##'   is binary
##' @param spec \code{StudySpecification} object
##' @return logical vector of length 1
##' @export
has_binary_treatment <- function(spec) {
  if (!inherits(spec, "StudySpecification")) {
    stop("spec must be a StudySpecification object.")
  }

  if (is.factor(treatment(spec)[, 1])) {
    # Short circuit if treatment is a factor
    return(FALSE)
  }

  return(all(treatment(spec)[, 1] %in% c(0, 1, TRUE, FALSE, NA)))
}

##' @title Test equality of two \code{StudySpecification} objects
##'
##' @description Check whether two \code{StudySpecification} objects are
##'   identical.
##'
##' @param x A \code{StudySpecification} object.
##' @param y A \code{StudySpecification} object.
##' @return Logical, are \code{x} and \code{y} identical?
##' @export
identical_StudySpecifications <- function(x, y) {
  return(identical(x, y))
}

##' @title Identify fine strata
##'
##' @description Identify blocks in a \code{StudySpecification} with exactly one
##'   treated or one control unit of assignment.
##' @param spec A \code{StudySpecification} object.
##' @return Logical vector with length given by the number of blocks in
##'   \code{StudySpecification}
##' @export
identify_small_blocks <- function(spec) {
  blk_txt_cts <- specification_table(spec, "t", "b")
  is_small_blk <- apply(blk_txt_cts, 1, function(blk) any(blk == 1))
  return(is_small_blk)
}

##' @title Make a dataframe that links units of assignment with clusters
##' @param spec A \code{StudySpecification} object.
##' @param cluster A character vector of column names to use as clusters.
##'   Columns must exist in the dataframe used to create the
##'   \code{StudySpecification} object. Defaults to NULL, in which case the
##'   column names specified in the \code{unitid()},
##'   \code{unit_of_assignment()}, or \code{cluster()} function in the
##'   \code{StudySpecification} formula will be used.
##' @return A dataframe where the number of rows coincides with the number of
##'   distinct unit of assignment or cluster combinations (depending on whether
##'   `cluster` is a more or less granular level than the assignment level) and
##'   the columns correspond to the unit of assignment columns and a "cluster"
##'   column
##' @keywords internal
.make_uoa_cluster_df <- function(spec, cluster = NULL) {
  if (!inherits(spec, "StudySpecification")) stop("Must be provided a valid `StudySpecification` object")
  uoa_cols <- var_names(spec, "u")
  q_df <- eval(spec@call$data, envir = environment(spec@call$formula))
  if (!is.null(spec@call$subset)) q_df <- subset(q_df, eval(spec@call$subset, envir = q_df))

  if (is.null(q_df)) {
    stop("Could not find specification data in the call stack")
  }

  if (spec@unit_of_assignment_type == "none") {
    q_df[["..uoa.."]] <- rownames(q_df)
  }

  if (!is.null(cluster) & !all(cluster %in% colnames(q_df))) {
    stop(paste("Could not find", cluster, "column in the specification data"))
  }

  q_df <- q_df[, unique(c(uoa_cols, cluster)), drop = FALSE]
  grab_uoas_fn <- switch(
    spec@unit_of_assignment_type,
    "unitid" = unitids,
    "unit_of_assignment" = units_of_assignment,
    "cluster" = clusters,
    "none" = ..uoa..
  )
  uoas <- grab_uoas_fn(spec)

  out <- unique(merge(uoas, q_df, by = uoa_cols, all.y = TRUE))
  rownames(out) <- NULL
  if (nrow(out) < nrow(uoas)) {
    warning(paste("Some units of assignment in the StudySpecification",
                  "were not found in the data used to",
                  "create the StudySpecification object. Ensure",
                  "the original data has not been",
                  "modified."))
  }

  if (is.null(cluster_cols <- cluster)) cluster_cols <- uoa_cols
  out$cluster <- apply(
    out[, cluster_cols, drop = FALSE],
    1,
    function(...) paste(..., collapse = "_")
  )

  return(out[, c(uoa_cols, "cluster")])
}

test_that("StudySpecification formula checking", {
  expect_true(.check_spec_formula(y ~ cluster(x)))
  expect_true(.check_spec_formula(y ~ cluster(x, z, q, r)))
  expect_true(.check_spec_formula(y ~ unitid(x)))
  expect_true(.check_spec_formula(y ~ unitid(x, z, q, r)))
  expect_true(.check_spec_formula(y ~ unit_of_assignment(x)))
  expect_true(.check_spec_formula(y ~ unit_of_assignment(x, z, q, r)))
  expect_true(.check_spec_formula(y ~ uoa(x)))
  expect_true(.check_spec_formula(y ~ uoa(x, z, q, r)))

  expect_error(.check_spec_formula(~ cluster(x)),
               "treatment")

  expect_error(.check_spec_formula(y ~ cluster(x) + unitid(z)),
               "Only one of")

  expect_error(.check_spec_formula(y ~ cluster(x) + cluster(z)),
               "Only one instance of `cluster")

  expect_error(.check_spec_formula(y ~ unitid(x) + unitid(z)),
               "Only one instance of `unitid")

  expect_error(.check_spec_formula(y ~ unit_of_assignment(x) +
                                       unit_of_assignment(z)),
               "Only one instance of `unit_of")

  expect_error(.check_spec_formula(y ~ uoa(x) + uoa(z)),
               "Only one instance of `unit_of")

  expect_true(.check_spec_formula(y ~ cluster(x) + block(z)))
  expect_true(.check_spec_formula(y ~ cluster(x) + block(z, a, b, c)))

  expect_error(.check_spec_formula(y ~ cluster(x) + block(z) + block(q)),
               "only one block")

  expect_error(.check_spec_formula(y ~ cluster(x) + forcing(z)),
               "only allowed")

  expect_true(.check_spec_formula(y ~ cluster(x) + forcing(z),
                                 allow_forcing = TRUE))

  expect_error(.check_spec_formula(y ~ cluster(x) + forcing(z) + forcing(q),
                                  allow_forcing = TRUE),
               "only one forcing")

})

test_that("binary treatment and dichotomy", {
  expect_error(has_binary_treatment(1),
               "must be")

  spec1 <- obs_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  spec2 <- obs_spec(o ~ uoa(uoa1, uoa2), data = simdata)

  expect_true(has_binary_treatment(spec1))

  expect_false(has_binary_treatment(spec2))
})

test_that("identical_StudySpecifications function", {
  data(simdata)

  spec1 <- rct_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  spec2 <- rct_spec(dose ~ cluster(uoa1, uoa2), data = simdata)
  spec3 <- rct_spec(dose ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  expect_true(identical_StudySpecifications(spec1, spec2))
  expect_false(identical_StudySpecifications(spec1, spec3))
  expect_false(identical_StudySpecifications(spec2, spec3))

})

test_that("identify_small_blocks", {
  spec_data <- data.frame(uid = letters[1:12],
                         bid = rep(LETTERS[1:3], each = 4),
                         a = c(rep(rep(c(0, 1), each = 2), 2), 0, rep(1, 3)))
  spec <- rct_spec(a ~ unitid(uid) + block(bid), spec_data)

  small_blocks <- identify_small_blocks(spec)
  expect_equal(length(small_blocks), 3)
  expect_equal(names(small_blocks), LETTERS[1:3])
  expect_true(all.equal(small_blocks, c(FALSE, FALSE, TRUE), check.attributes = FALSE))
})

test_that(".make_uoa_cluster_df errors", {
  expect_error(.make_uoa_cluster_df("spec"), "valid `StudySpecification` object")

  data(simdata)
  simdata_copy <- simdata
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata_copy)

  simdata_copy$uoa2[simdata_copy$uoa1 == 1 & simdata_copy$uoa2 == 2] <- 1
  expect_warning(.make_uoa_cluster_df(spec), "Some units of assignment")

  simdata_copy <- NULL
  expect_error(.make_uoa_cluster_df(spec), "specification data")

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  expect_error(.make_uoa_cluster_df(spec, "id"), "id column in the specification data")
})

test_that(".make_uoa_cluster_df", {
  spec_data <- data.frame(id = letters[1:20],
                         uid1 = rep(letters[1:4], each = 5),
                         uid2 = rep(letters[1:4], each = 5),
                         bid = rep(LETTERS[1:2], each = 10),
                         a = rep(rep(c(0, 1), each = 5), 2))

  # test specification creation using `cluster()`, `unitid()`, and `unit_of_assignment()`
  cluster_spec <- rct_spec(a ~ cluster(uid1), spec_data)
  unitid_spec <- rct_spec(a ~ unitid(uid1), spec_data)
  uoa_spec <- rct_spec(a ~ unit_of_assignment(uid1), spec_data)

  expect_true(all.equal(uc_df <- .make_uoa_cluster_df(cluster_spec),
                        .make_uoa_cluster_df(unitid_spec),
                        check.attributes = FALSE))
  expect_true(all.equal(uc_df,
                        .make_uoa_cluster_df(uoa_spec),
                        check.attributes = FALSE))
  expect_equal(uc_df[,1,drop=FALSE], clusters(cluster_spec))
  expect_equal(colnames(uc_df), c("uid1", "cluster"))
  expect_true(all.equal(uc_df$cluster, letters[1:4],
                        check.attributes = FALSE))
  
  # test when there's a subset
  subset_spec <- rct_spec(a ~ unitid(uid1, uid2), spec_data, subset = bid == "A")
  expect_equal(.make_uoa_cluster_df(subset_spec, c("uid1", "uid2")),
               data.frame(uid1 = letters[1:2], uid2 = letters[1:2],
                          cluster = paste(letters[1:2], letters[1:2], sep = "_")))

  # test cluster arg specification when level is coarser than assignment level
  expect_equal(dim(uc_df <- .make_uoa_cluster_df(cluster_spec, "bid")),
               c(4, 2))
  expect_equal(uc_df[,1,drop=FALSE], clusters(cluster_spec))
  expect_equal(colnames(uc_df), c("uid1", "cluster"))
  expect_true(all.equal(uc_df$cluster, rep(LETTERS[1:2], each = 2),
                        check.attributes = FALSE))

  # test cluster arg specification when level is finer than assignment level
  expect_equal(dim(uc_df <- .make_uoa_cluster_df(cluster_spec, "id")),
               c(20, 2))
  expect_true(all.equal(uc_df[,1], rep(clusters(cluster_spec)[,1], each = 5),
                        check.attributes = FALSE))
  expect_equal(colnames(uc_df), c("uid1", "cluster"))
  expect_true(all.equal(uc_df$cluster, letters[1:20],
                        check.attributes = FALSE))

  # test cluster arg specification with multiple columns
  expect_equal(dim(uc_df <- .make_uoa_cluster_df(cluster_spec, c("uid2", "bid"))),
               c(4, 2))
  expect_true(all.equal(uc_df$cluster,
                        paste(letters[1:4], rep(LETTERS[1:2], each = 2),
                              sep = "_"),
                        check.attributes = FALSE))

  # test specification with multiple unit of assignment columns
  cluster_spec <- rct_spec(a ~ cluster(uid1, uid2), spec_data)
  expect_equal(dim(uc_df <- .make_uoa_cluster_df(cluster_spec)),
               c(4, 3))
  expect_equal(uc_df[,1:2,drop=FALSE], clusters(cluster_spec))
  expect_equal(colnames(uc_df), c("uid1", "uid2", "cluster"))
  expect_true(all.equal(uc_df$cluster,
                        paste(letters[1:4], letters[1:4], sep = "_"),
                        check.attributes = FALSE))
})

#' @include StudySpecification.R
NULL
# The above ensures that `StudySpecification` is defined prior to
# `StudySpecificationStructure`

setClass("StudySpecificationStructure",
         contains = "data.frame",
         slots = c(StudySpecification = "StudySpecification"))

setValidity("StudySpecificationStructure", function(object) {
  validObject(as(object@StudySpecification, "StudySpecification"))

  # Todo: check validity of `data.frame`. Right dimensions?
  return(TRUE)
})
######### Structure

##' @title \code{StudySpecification} Structure Information
##'
##' @description Obtaining a \code{data.frame} which encodes the specification
##'   information.
##' @param specification a \code{StudySpecification} object
##' @return A \code{StudySpecificationStructure} object containing the structure
##'   of the \code{specification} as a \code{data.frame}.
##' @export
##' @rdname StudySpecification_structure
##' @examples
##' data(simdata)
##' spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
##' get_structure(spec)
get_structure <- function(specification) {

  struct <- specification@structure
  struct[, specification@column_index == "t"] <-
    treatment(specification)

  return(new("StudySpecificationStructure",
             struct,
             StudySpecification = specification))
}

##' @param object a \code{StudySpecificationStructure} object, typically the
##'   output of \code{get_structure}
##' @export
##' @rdname StudySpecification_structure
setMethod("show", "StudySpecificationStructure", function(object) {
  struct <- data.frame(object)
  old_row_names <- rownames(struct)

  var_id_row <- object@StudySpecification@column_index
  var_id_row <- gsub("^t$", "Treatment", var_id_row)
  uoatype <- switch(object@StudySpecification@unit_of_assignment_type,
                    "unit_of_assignment" = "Unit of Assignment",
                    "cluster" = "Cluster",
                    "unitid"  = "Unit ID")
  var_id_row <- gsub("^u$", uoatype, var_id_row)
  var_id_row <- gsub("^b$", "Block", var_id_row)
  var_id_row <- gsub("^f$", "Forcing", var_id_row)

  struct <- rbind(var_id_row, struct)
  rownames(struct) <- c("", old_row_names)


  show(struct)
  invisible(object)
})

##' @param treatment_binary Should the treatment be dichotomized if
##'   \code{object} contains a \code{dichotomy}? Ignored if \code{object} does
##'   not contain a \code{dichotomy}.
##' @export
##' @method summary StudySpecification
##' @rdname StudySpecification_summary
summary.StudySpecification <- function(object, ..., treatment_binary = TRUE) {
  out <- list()
  out$StudySpecification <- object
  out$treatment_table <- stable(object, "treatment", ...)
  class(out) <- "summary.StudySpecification"
  return(out)
}

##' @title Summarizing \code{StudySpecification} objects
##' @description [summary()] method for class \code{StudySpecification}.
##' @param object \code{StudySpecification} object, usually a result of a call
##'   to [rct_spec()], [obs_spec()], or [rd_spec()].
##' @param x \code{summary.StudySpecification} object, usually as a result of a
##'   call to [summary.StudySpecification()]
##' @param ... Ignored
##' @param max_unit_print Maximum number of treatment levels to print in
##'   treatment table
##' @return The \code{StudySpecification} or
##'   \code{summary.StudySpecification}object, invisibly
##' @export
##' @rdname StudySpecification_summary
print.summary.StudySpecification <- function(x, ..., max_unit_print = 3) {
  show(x$StudySpecification)

  cat("Number of units per Treatment group: \n")
  tt <- x$treatment_table
  # The value below defines the max number of treatment groups to print
  max_print_table <- min(max_unit_print, length(tt))
  tdf <- as.data.frame(tt[seq_len(max_print_table)])
  colnames(tdf) <- c("Txt Grp", "Num Units")
  # knitr::kable(tt, align = "cc", format = "simple") looks real nice if we want
  # to add that dependency
  if (length(tt) > max_print_table) {
    tdf[, 1] <- as.character(tdf[, 1])
    tdf <- rbind(tdf, c("...", ""))
  }
  print(tdf, row.names = FALSE)
  if (length(tt) > max_print_table) {
    if (length(tt) - max_print_table == 1) {
      group <- "group"
    } else {
      group <- "groups"
    }
    cat(paste0(length(tt) - max_print_table, " smaller treatment ",
               group, " excluded.\n"))
    cat("Use `specification_table` function to view full results.")
  }
  cat("\n")
  invisible(x)
}

#' @title Simulated data
#'
#' @description Simulated data to use with the \bold{propertee} package with unit
#'   of assignment level treatment assignment
#'
#' @format A data.frame with 100 rows and 7 columns.
#' \itemize{
#'   \item \code{uoa1} First level unit of assignment ID
#'   \item \code{uoa2} Second level unit of assignment ID
#'   \item \code{bid} Block ID
#'   \item \code{force} Forcing variable
#'   \item \code{z} Binary treatment indicator
#'   \item \code{o} 4-level ordered treatment variable
#'   \item \code{dose} Dose treatment variable
#'   \item \code{x} Some predictor
#'   \item \code{y} Some outcome
#' }
#' @keywords dataset
"simdata"

#' @title Synthethic Regression Discontinuity Data
#'
#' @description The data for this example were randomly simulated using the
#'   \bold{synthpop} package in R based on data originally collected by Lindo,
#'   Sanders, and Oreopoulos (2010).
#'
#' @details See the "Regression Discontinuity StudySpecifications" vignette on the
#'   \code{propertee} website for more details on the original data, a link to
#'   the code used to generate this synthethic data, and a detailed example.
#'
#' @format A \code{data.frame} with 40,403 rows and 11 columns.
#' \itemize{
#'   \item \code{R}
#'   \item \code{lhsgrade_pct}
#'   \item \code{nextGPA}
#'   \item \code{probation_year1}
#'   \item \code{totcredits_year1}
#'   \item \code{male}
#'   \item \code{loc_campus1}
#'   \item \code{loc_campus2}
#'   \item \code{bpl_north_america}
#'   \item \code{english}
#'   \item \code{age_at_entry}
#' }
#' @keywords dataset
"lsoSynth"

#' @title STAR participants plus nonexperimental controls
#'
#' @description Data from Tennessee’s Project STAR study.
#'   This data frame describes student participants in the Project
#'   STAR (Student-Teacher Achievement Ratio) field experiment
#'   conducted in Tennessee, USA beginning in the mid-1980s, as well
#'   as an external control group consisting of the contemporaneous
#'   cohort of students attending a matched sample of Tennessee
#'   schools that did not participate in the STAR experiment.
#'   Variables are as described in Project \code{STAR} data
#'   documentation (see references), with five exceptions.
#'   Three \code{*_at_entry} variables were constructed
#'   as follows: \code{grade_at_entry} indicates the grade of student's
#'   first participation, while \code{school_at_entry} and
#'   \code{cond_at_entry} reflect the school ID and classroom type corresponding
#'   to the student's grade at entry to the study. Additionally, \code{read_yr1}
#'   and \code{math_yr1} capture a student's scaled scores on the Scholastic
#'   Assessment Test (SAT) administered to them during their \code{grade_of_entry}, 
#'   i.e. their earliest available post-treatment SAT measurements.
#'
#' @format  A \code{data.frame} with 13,382 rows and 56 columns.
#' \itemize{
#'    \item {stdnid} Student ID
#'    \item {gender} Student gender
#'    \item {race} Student race
#'    \item {birthmonth} Student month of birth
#'    \item {birthday} Student day of birth
#'    \item {birthyear} Student year of birth
#'    \item {read_yr1} SAT reading scaled score from grade at which
#'          student entered the study
#'    \item {math_yr1} SAT math scaled score from grade at which 
#'          student entered the study
#'    \item {gktreadss} Kindergarten reading scaled score (RCT participants only)
#'    \item {gktmathss} Kindergarten math scaled score (RCT participants only)
#'    \item {gktlistss} Kindergarten listening scaled score (RCT participants only)
#'    \item {gkwordskillss} Kindergarten word study skills scaled score (RCT participants only)
#'    \item {g1schid} Grade 1 School ID
#'    \item {g1tchid} Grade 1 Teacher ID
#'    \item {g1classsize} Class size of Grade 1
#'    \item {g1treadss} Grade 1 SAT reading scaled score
#'    \item {g1tmathss} Grade 1 SAT math scaled score
#'    \item {g1tlistss} Grade 1 total listening scale score in SAT
#'    \item {g1wordskillss} Grade 1 word study skills scale score in SAT
#'    \item {g1readbsraw} Grade 1 reading raw score in Basic Skills First (BSF) tests
#'    \item {g1mathbsraw} Grade 1 math raw score in BSF
#'    \item {g1readbsobjpct} Grade 1 reading percent objectives mastered in BSF tests
#'    \item {g1mathbsobjpct} Grade 1 math percent objectives mastered in BSF tests
#'    \item {g2schid} Grade 2 School ID
#'    \item {g2tchid} Grade 2 Teacher ID
#'    \item {g2classsize} Class size of Grade 2
#'    \item {g2treadss} Grade 2 total reading scale score in SAT
#'    \item {g2tmathss} Grade 2 total math scale score in SAT
#'    \item {g2tlistss} Grade 2 total listening scale score in SAT
#'    \item {g2wordskillss} Grade 2 word study skills scale score in SAT
#'    \item {g2readbsraw} Grade 2 reading raw score in BSF tests
#'    \item {g2mathbsraw} Grade 2 math raw score in BSF test
#'    \item {g2readbsobjpct} Grade 2 reading percent objectives mastered in BSF tests
#'    \item {g3schid} Grade 3 School ID
#'    \item {g3tchid} Grade 3 Teacher ID
#'    \item {g3classsize} Class size of Grade 3
#'    \item {g3treadss} Grade 3 total reading scale score in SAT
#'    \item {g3tmathss} Grade 3 total math scale score in SAT
#'    \item {g3langss} Grade 3 total language scale score in SAT
#'    \item {g3tlistss} Grade 3 total listening scale score in SAT
#'    \item {g3socialsciss} Grade 3 social science scale score in SAT
#'    \item {g3spellss} Grade 3 spelling scale score in SAT
#'    \item {g3vocabss} Grade 3 vocabulary scale score in SAT
#'    \item {g3mathcomputss} Grade 3 math computation scale score in SAT
#'    \item {g3mathnumconcss} Grade 3 concept of numbers scale score in SAT
#'    \item {g3mathapplss} Grade 3 math applications scale score in SAT
#'    \item {g3wordskillss} Grade 3 word study skills scale score in SAT
#'    \item {g3readbsraw} Grade 3 reading raw score in BSF tests
#'    \item {g3mathbsraw} Grade 3 math raw score in BSF tests
#'    \item {g3readbsobjpct} Grade 3 reading percent objectives mastered in BSF tests
#'    \item {g3mathbsobjpct} Grade 3 math percent objectives mastered in BSF tests
#'    \item {dob} Date of birth (with NAs imputed RCT participant median)
#'    \item {dobNA} Dat of birth not recorded
#'    \item {grade_at_entry} Grade at which each student first entered the study
#'    \item {school_at_entry} School ID corresponding to the student's grade
#'          at entry into the study
#'    \item {cond_at_entry} Classroom type corresponding to the student's grade
#'          at entry into the study
#' }
#' @details Note: This dataset bears a Creative Commons Zero license (v1.0).
#'
#' @references C.M. Achilles; Helen Pate Bain; Fred Bellott; Jayne Boyd-Zaharias;
#' Jeremy Finn; John Folger; John Johnston; Elizabeth Word, 2008,
#' "Tennessee's Student Teacher Achievement Ratio (STAR) project",
#' Harvard Dataverse, V1, https://doi.org/10.7910/DVN/SIWH9F UNF:3:Ji2Q+9HCCZAbw3csOdMNdA
#'
#' @source \doi{doi:10.7910/DVN/SIWH9F}
#' @keywords dataset
"STARplus"

#' @title Student data
#'
#' @description An example of data sets stored at two levels.
#'
#' @details In this hypothetical data, schools were randomly assignment to
#'   treatment status, but the unit of analysis is students. Thus the two data
#'   sets, one encoding school information (including treatment status) and one
#'   encoding student information (which does not include treatment status).
#'
#' @rdname studentdata
#' @format Two \code{data.frame}s, one with school-level data (\code{schooldata})
#'   including treatment assignment and a second with student-level data
#'   (\code{studentdata}).
#' \code{schoolata}:
#' \itemize{
#'   \item {schoolid} Unique school ID variable.
#'   \item {treatment} Was this school in the intervention group?
#'   \item {state} State which the school is in.
#'   \item {pct}_disadvantage Percent of student body flagged as "disadvantaged".
#' }
#' \code{studentdata}:
#' \itemize{
#'   \item {id} Unique student ID.
#'   \item {schoolid} Unique school ID variable.
#'   \item {grade} Student's grade, 3-5.
#'   \item {gpa} Student GPA in prior year.
#'   \item {math} Standarized math score (out of 100).
#' }
#' @examples
#' soec <- obs_spec(treatment ~ uoa(schoolid), data = schooldata)
#'
#' # Treatment effect
#' mod1 <- lmitt(math ~ 1, specification = soec, data = studentdata)
#'
#' # Treatment effect by grade
#' mod2 <- lmitt(math ~ as.factor(grade), specification = soec, data = studentdata)
#' @keywords dataset
"schooldata"

#' @rdname studentdata
"studentdata"

#' @title Intervention data from a pair-matched study of schools in Michigan
#'
#' @description Michigan high schools, with a plausible cluster RCT
#'
#' @details
#'
#' Grade 11 demographics for all Michigan high schools in 2013,
#' with mock block and treatment assignments for 14 high schools within
#' a large county in the metro Detroit area.
#' These schools were selected for this demonstration based on their
#' similarity to the 14 high schools from an adjacent Michigan
#' county that participated in the Pane et al (2013) study.  As a result,
#' they serve as an example of what one might expect to find as the
#' state-specific school-level subsample in a multi-state
#' paired cluster randomized trial featuring random assignment at
#' the school level.
#'
#' The mock experimental schools were selected by optimal matching of
#' experimental schools to adjacent county schools, with substitute
#' schools grouped into the same pairs or triples (\sQuote{fine
#' strata}) as were their experimental counterparts.  The original
#' pairs and triples had been selected to reduce variation in baseline
#' variables predictive of outcomes, and the blocking structure the
#' substitute sample inherits may be expected to do this as well.  The
#' treatment/control distinction is also inherited from the
#' experimental sample, but there is of course no treatment effect
#' within the mock experiment.
#'
#' The selection of mock experimental schools was based on both demographic
#' and student achievement variables, but the present data frame
#' includes only the demographic variables (as sourced from the
#' Common Core of Data \[CCD; U.S. Department of Education\]).  School
#' average outcomes in student test scores are available separately, from
#' Michigan's Center for Education Performance Information.  See the
#' vignette \sQuote{Real-data demonstration with a finely stratified
#' cluster RCT and a broader administrative database}, available on
#' the package website.
#' @references Pane, John F., et al.
#'     "Effectiveness of cognitive tutor algebra I at scale."
#'     \emph{Educational Evaluation and Policy Analysis} 36.2 (2014):
#'     127-144.
#'
#' U.S. Department of Education. Public Elementary/Secondary School
#' Universe Survey Data, v.2a. Institute of Education Sciences,
#' National Center for Education Statistics.
#'
#' @format  A \code{data.frame} with 14 rows and 13 columns.
#' \itemize{
#'  \item{schoolid} school id
#'  \item{blk} block
#'  \item{z} treatment variable
#'  \item{MALE_G11_PERC} percentage of G11 male students
#'  \item{FEMALE_G11_PERC} percentage of G11 female students
#'  \item{AM_G11_PERC} percentage of G11 American Indian/Alaska Native students
#'  \item{ASIAN_G11_PERC} percentage of G11 Asian students
#'  \item{HISP_G11_PERC} percentage of G11 Hispanic students
#'  \item{BLACK_G11_PERC} percentage of G11 Black students
#'  \item{WHITE_G11_PERC} percentage of G11 White students
#'  \item{PACIFIC_G11_PERC} percentage of G11 Hawaiian Native/Pacific Islander students
#'  \item{TR_G11_PERC} percentage of G11 Two or More Races students
#'  \item{G11} Number of G11 students
#' }
#' @keywords dataset
#' @examples
#' data(michigan_school_pairs)
#' mi_spec <- rct_spec(z ~ uoa(schoolid)+block(blk),
#' data=michigan_school_pairs)
#' mi_spec
#' table(is.na(michigan_school_pairs$blk))
#' specification_table(mi_spec, "block", "treatment")
"michigan_school_pairs"

#' @title Cluster-randomized experiment data on voter turnout in cable system markets
#'
#' @description This dataset is a toy example derived from a cluster-randomized
#' field experiment that evaluates the effect of \dQuote{Rock the Vote} TV advertisements
#' on voter turnout rate. The original study included 23,869 first-time voters
#' across 85 cable television markets in 12 states. These markets were grouped
#' into matched sets based on their past voter turnout rates and then randomly
#' assigned to either a treatment or control condition. This toy dataset is
#' constructed by randomly sampling 10% of individuals from selected cable
#' television markets in the original dataset.
#'
#' @format A \code{data.frame} with 248 rows and 7 columns.
#' \itemize{
#'   \item {age} Age of participant
#'   \item {vote_04} Outcome variable indicating whether participant voted
#'   \item {tv_company} Cable system serving participant's residential area
#'   \item {treatment} Binary variable denoting treatment assignment
#'   \item {pairs} A numeric indicator for the strata or matched pair group to which a cable system belongs (1-3)
#'   \item {population_size} Total population size of residential area served by cable system
#'   \item {sample_size} Number of individuals sampled from the cable system cluster
#' }
#'
#' @details The original dataset was drawn from a randomized controlled trial in
#' which 85 cable system areas were first grouped into 40 matched sets based on
#' historical voter turnout. Within each matched set, one cable system area was
#' randomly assigned to the treatment condition, while the others served
#' as controls.
#'
#' This toy dataset includes a subset of the original replication data,
#' specifically individuals from matched sets 1–3, which encompass 7 of the 85
#' cable system areas. Within these selected clusters, a 10% random sample of
#' individuals was taken.
#'
#' The fuller Green-Vavreck dataset that this derives from bears a Creative
#' Commons BY-NC-ND license (v3.0) and is housed in Yale University's Institution
#' for Social and Policy Studies (ID: D005).
#'
#' @source <https://dataverse.yale.edu/dataset.xhtml?persistentId=doi:10.60600/YU/OYIPLJ>
#'
#' @references Green, Donald P. & Lynn Vavreck (2008) "Analysis of Cluster-Randomized
#' Experiments: A Comparison of Alternative Estimation Approaches." Political Analysis 16(2):138-152.
#'
#' @keywords dataset
"GV_data"

#' @title Adjust residuals for both-sides absorption
#' @param x a fitted \code{teeMod} model
#' @details
#' This function subtracts off the block residual mean function
#' \eqn{\hat \alpha(v_b, \theta)} for each observation from model residuals
#' @return the fitted \code{teeMod} with updated block center residuals.
#' @keywords internal
block_center_residuals <- function(x){
  blks <- stats::expand.model.frame(
    x, var_names(x@StudySpecification, "b")
    )[names(residuals(x)),var_names(x@StudySpecification, "b")]
  n <- length(blks)
  w <- if (is.null(weights(x))) rep(1, n) else weights(x)
  blk_means <- suppressWarnings(rowsum(
    residuals(x) * w, blks, na.rm = TRUE) / rowsum(w, blks, na.rm = TRUE))
  centered_u <- blk_means[blks]
  x$residuals <- residuals(x) - centered_u
  return(x)
}

# Supported weights (targets) refer to the actual calculations. All targets
# should have at minimum one alias, themselves.

##' @title Valid Weights
##'
##' @description These track and access the weights we support as well as any aliases.
##'
##' @details "target" refers to weight calculations we support.
##'
##' "alias" refers to all possible names. Every "target" has at least one
##' "alias" (itself) and may have more.
##'
##' \code{.isValidWeightTarget()} and \code{.isValidWeightAlias()} identify
##' whether a given input (from a user) is a value weighting name.
##'
##' \code{.listValidWeightTargets()} and \code{.listValidWeightAliases()} are
##' for returning nicely formatted strings for messages to users.
##'
##' IMPORTANT: Adding new aliases MUST correspond to new functions defined in
##' weights_exported.R, with possible adjustments to calculations in
##' weights_internal.R.
##' @param target String
##' @param alias String
##' @return Logical for \code{.isValid*}, and a string for \code{.listValid*}.
##' @keywords internal
##' @rdname validWeights
.validWeights <- list(
  targets = c("ate", "ett", "etc", "ato"),
  aliases = c("ate", "ett", "att", "etc", "atc", "ato", "olw",
              "owt", "pwt")
)

##' @keywords internal
##' @rdname validWeights
.isValidWeightTarget <- function(target) {
  stopifnot(is.character(target))
  return(target %in% .validWeights$targets)
}

##' @keywords internal
##' @rdname validWeights
.isValidWeightAlias <- function(alias) {
  stopifnot(is.character(alias))
  return(alias %in% .validWeights$aliases)
}

##' @keywords internal
##' @rdname validWeights
.listValidWeightTargets <- function() {
  return(paste(.validWeights$targets, collapse = ", "))
}


##' @keywords internal
##' @rdname validWeights
.listValidWeightAliases <- function() {
  return(paste(.validWeights$aliases, collapse = ", "))
}

test_that("validWeights object", {
  expect_true(is.logical(.isValidWeightAlias("ate")))
  expect_true(is.logical(.isValidWeightTarget("ate")))
  expect_true(is.logical(.isValidWeightAlias("xxx")))
  expect_true(is.logical(.isValidWeightTarget("xxx")))

  expect_true(is.character(.listValidWeightAliases()))
  expect_true(is.character(.listValidWeightTargets()))

  expect_true(length(.validWeights) == 2)
  expect_true(all(sapply(.validWeights, length) >= 1))
})

test_that("weight aliases", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  mod1 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = ate())
  mod2 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = ett())
  mod3 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = att())
  expect_false(isTRUE(all.equal(summary(mod1)$coeff, summary(mod2)$coeff)))
  expect_true(isTRUE(all.equal(summary(mod2)$coeff, summary(mod3)$coeff)))

  mod4 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = "ate")
  mod5 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = "ett")
  mod6 <- lmitt(y ~ 1, data = simdata, spec = spec, weights = "att")
  expect_false(isTRUE(all.equal(summary(mod4)$coeff, summary(mod5)$coeff)))
  expect_true(isTRUE(all.equal(summary(mod5)$coeff, summary(mod6)$coeff)))

})

#' @importFrom utils getS3method methods
#' @importFrom stats model.frame terms na.pass vcov confint
#' @importFrom methods as is new validObject show
#' @importFrom sandwich estfun bread
NULL
#> NULL


.onLoad <- function(lib, pkg) {
  options("propertee_warn_on_conditional_treatment" = TRUE)
  options("propertee_message_on_unused_blocks" = TRUE)
  options("propertee_warn_on_no_unit_of_assignment" = TRUE)
}

#' @include weights_internal.R
NULL

############## ***IMPORTANT*** ######################
# If adding new weights here (either direct targets or aliases, YOU MUST UPDATE
# the `.validWeights` object inside validWeights.R

##' @title Generate Direct Adjusted Weights for Treatment Effect Estimation
##'
##' @description These should primarily be used inside models. See Details.
##'
##' @details These functions should primarily be used in the \code{weight}
##'   argument of [lmitt()] or[lm()]. All arguments are optional if used within
##'   those functions. If used on their own, \code{specification} and
##'   \code{data} must be provided.
##'
##'   - \code{ate} - Average treatment effect. Aliases: \code{ate()}.
##'   - \code{ett} - Effect of treatment on the treated. Aliases: \code{ett()},
##'                        \code{att()}.
##'   - \code{etc} - Effect of treatment on controls. Aliases: \code{etc()},
##'                        \code{atc()}.
##'   - \code{ato} - Overlap-weighted average effect. Aliases: \code{ato()},
##'                        \code{olw}, \code{owt}, \code{pwt}.
##'
##'   In a \code{StudySpecification} with \code{block}s, the weights are
##'   generated as a function of the ratio of the number of treated units in a
##'   block versus the total number of units in a block.
##'
##'   In any blocks where that ratio is 0 or 1 (that is, all units in the block
##'   have the same treatment status), the weights will be 0. In effect this
##'   removes from the target population any block in which there is no basis
##'   for estimating either means under treatment or means under control.
##'
##'   If block is missing for a given observation, a weight of 0 is applied.
##'
##'   A \code{dichotomy} is specified by a \code{formula} consisting of a
##'   conditional statement on both the left-hand side (identifying treatment
##'   levels associated with "treatment") and the right hand side (identifying
##'   treatment levels associated with "control"). For example, if your
##'   treatment variable was called \code{dose} and doses above 250 are
##'   considered treatment, you might write:
##'
##'   \code{ate(..., dichotomy = dose > 250 ~ dose <= 250}
##'
##'   The period (\code{.}) can be used to assign all other units of assignment.
##'   For example, we could have written the same treatment regime as either
##'
##'   \code{etc(..., dichotomy = dose > 250 ~ .}
##'
##'   or
##'
##'   \code{olw(..., dichotomy = . ~ dose <= 250}
##'
##'   The \code{dichotomy} formula supports Relational Operators (see
##'   [Comparison]), Logical Operators (see [Logic]), and \code{%in%} (see
##'   [match()]).
##'
##'   The conditionals need not assign all values of treatment to control or
##'   treatment, for example, \code{dose > 300 ~ dose < 200} does not assign
##'   \code{200 <= dose <= 300} to either treatment or control. This would be
##'   equivalent to manually generating a binary variable with \code{NA}
##'   whenever \code{dose} is between 200 and 300. Standard errors will reflect
##'   the sizes of the comparison groups specified by the \code{dichotomy}.
##'
##'   Tim Lycurgus contributed code for the computation of weights. The
##'   \sQuote{overlap weight} concept is due to Li, Morgan and Zaslavsky
##'   (2018), although the current implementation differs from that
##'   discussed in their paper in that it avoids estimated propensity scores.
##'
##' @param specification optional; a \code{StudySpecification} object created by
##'   one of \code{rct_spec()}, \code{rd_spec()}, or \code{obs_spec()}.
##' @param dichotomy optional; a formula defining the dichotomy of the treatment
##'   variable if it isn't already \code{0}/\code{1}. See details for more
##'   information. If \code{ett()} or \code{ate()} is called within a
##'   \code{lmitt()} call that specifies a \code{dichotomy} argument, that
##'   \code{dichotomy} will be used if the argument here has not been specified.
##' @param by optional; named vector or list connecting names of unit of
##'   assignment/ variables in \code{specification} to unit of
##'   assignment/unitid/cluster variables in \code{data}. Names represent
##'   variables in the StudySpecification; values represent variables in the
##'   data. Only needed if variable names differ.
##' @param data optional; the data for the analysis to be performed on. May be
##'   excluded if these functions are included as the \code{weights} argument of
##'   a model.
##' @return a \code{WeightedStudySpecification} object, which is a vector of
##'   numeric weights
##' @export
##' @rdname WeightCreators
##' @references Li, Fan, Kari Lock Morgan, and Alan M. Zaslavsky.
##' "Balancing covariates via propensity score weighting." Journal of the American Statistical Association 113, no. 521 (2018): 390-400.
##' @examples
##' data(simdata)
##' spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
##' summary(lmitt(y ~ 1, data = simdata, specification = spec, weights = ate()), vcov.type = "CR0")
ett <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ett",
                       weightAlias = "ett",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
att <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ett",
                       weightAlias = "att",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
ate <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ate",
                       weightAlias = "ate",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
etc <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "etc",
                       weightAlias = "etc",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
atc <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "etc",
                       weightAlias = "atc",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}


##' @rdname WeightCreators
##' @export
ato <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ato",
                       weightAlias = "ato",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
olw <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ato",
                       weightAlias = "olw",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
owt <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ato",
                       weightAlias = "owt",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

##' @rdname WeightCreators
##' @export
pwt <- function(specification = NULL, dichotomy = NULL, by = NULL, data = NULL) {
  return(.weights_calc(specification = specification,
                       target = "ato",
                       weightAlias = "pwt",
                       dichotomy = dichotomy,
                       by = by,
                       data = data))
}

# Ensures for each combination of usages below, lm and lmitt produce the same
# point esimate.

### Dimensions:
# Weights Yes/No
# Subgroup Yes/No
# Absorb Yes/No
# CovAdj Yes/No
# Subset No/In `StudySpecification`/in `lm`

# To avoid a bunch of messages being printed:
save_options <- options()
options("propertee_message_on_unused_blocks" = FALSE)

test_that("equivalent of lm and lmitt calls - no subset", {
  data(simdata)
  simdata$o <- as.factor(simdata$o)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  cmod <- lm(y ~ x, data = simdata)


  test_coeffs <- function(lmcoef, lmittcoef, n_sg_levels, cov_adj) {
    ncoef <- if (cov_adj) 4 else 3
    if (n_sg_levels > 0 & !cov_adj) {
      ncoef <- ncoef + 3 * (n_sg_levels - 1)
    } else if (n_sg_levels > 0 & cov_adj) {
      ncoef <- ncoef + 4 * (n_sg_levels - 1)
    }
    expect_equal(length(lmittcoef), ncoef)
    lmittcoef <- lmittcoef[!grepl("(Intercept)", names(lmittcoef))]
    expect_true(all(round(lmittcoef[!grepl("^(y|cov_adj):", names(lmittcoef))], 4) %in% round(lmcoef, 4)))
  }

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |   No   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec), data = simdata)
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |   No   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec), data = simdata, weights = ate(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |   No   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata)
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |   No   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata, weights = ett(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |  Yes   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata)
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |  Yes   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               weights = ett(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
                     weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |  Yes   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o  + as.factor(bid), data = simdata)
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |  Yes   |   No   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               weights = ate(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |   No   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec), data = simdata, offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |   No   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(), data = simdata, offset = cov_adj(cmod),
               weights = ate(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec,
                     offset = cov_adj(cmod), weights = ate())
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |   No   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata, offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec,
                     offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |   No   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata,
               offset = cov_adj(cmod), weights = ett(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, offset = cov_adj(cmod),
                     weights = ett())
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |  Yes   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
               offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |  Yes   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               offset = cov_adj(cmod), weights = ett(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
               offset = cov_adj(cmod), weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |  Yes   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |  Yes   |  Yes   |   No   |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               offset = cov_adj(cmod), weights = ate(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     offset = cov_adj(cmod), weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)


})

test_that("subset in specification", {
  data(simdata)
  simdata$o <- as.factor(simdata$o)
  spec <- obs_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata,
                    subset = !(simdata$uoa1 == 5 & simdata$uoa2 == 2))
  cmod <- lm(y ~ x, data = simdata)

  test_coeffs <- function(lmcoef, lmittcoef, n_sg_levels, cov_adj) {
    ncoef <- if (cov_adj) 4 else 3
    if (n_sg_levels > 0 & !cov_adj) {
      ncoef <- ncoef + 3 * (n_sg_levels - 1)
    } else if (n_sg_levels > 0 & cov_adj) {
      ncoef <- ncoef + 4 * (n_sg_levels - 1)
    }
    expect_equal(length(lmittcoef), ncoef)
    lmittcoef <- lmittcoef[!grepl("(Intercept)", names(lmittcoef))]
    expect_true(all(round(lmittcoef[!grepl("^(y|cov_adj):", names(lmittcoef))], 4) %in% round(lmcoef, 4)))
  }



  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |   No   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec), data = simdata)
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |   No   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec), data = simdata, weights = ate(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |   No   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata)
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |   No   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata, weights = ett(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |  Yes   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata)
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |  Yes   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               weights = ett(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
                     weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |  Yes   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o  + as.factor(bid), data = simdata)
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE)
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |  Yes   |   No   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               weights = ate(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), FALSE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |   No   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec), data = simdata, offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |   No   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(), data = simdata, offset = cov_adj(cmod),
               weights = ate(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec,
                     offset = cov_adj(cmod), weights = ate())
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |   No   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata, offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec,
                     offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |   No   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o, data = simdata,
               offset = cov_adj(cmod), weights = ett(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, offset = cov_adj(cmod),
                     weights = ett())
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   No     |  Yes   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
               offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   No     |  Yes   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec) + as.factor(bid), data = simdata,
               offset = cov_adj(cmod), weights = ett(spec))
  mod_lmitt <- lmitt(y ~ 1, data = simdata, specification = spec, absorb = TRUE,
               offset = cov_adj(cmod), weights = "ett")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, 0, TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   No    |   Yes    |  Yes   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               offset = cov_adj(cmod))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     offset = cov_adj(cmod))
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)

  ## | Weights | Subgroup | Absorb | CovAdj | Subset |
  ## |---------|----------|--------|--------|--------|
  ## |   Yes   |   Yes    |  Yes   |  Yes   | StudySpecification |

  mod_lm <- lm(y ~ assigned(spec):o + o + as.factor(bid), data = simdata,
               offset = cov_adj(cmod), weights = ate(spec))
  mod_lmitt <- lmitt(y ~ o, data = simdata, specification = spec, absorb = TRUE,
                     offset = cov_adj(cmod), weights = "ate")
  test_coeffs(mod_lm$coeff, mod_lmitt$coeff, length(levels(simdata$o)), TRUE)
})

options(save_options)

# This isn't going to test whether anything is actually correct (see individiual
# tests for that), but instead will run each different variation of `lm` call
# and ensure that no messages, warnings, or errors are produced.

# ******************** IMPORTANT ********************
# If debugging things in here, REMOVE EXPECT_SILENT!!! Running it will suppress
# all output, including inside `browser()`
test_that("binary treatment, in all data", {

  # Treatment exists in data
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  camod <- lm(y ~ x, data = simdata)

  # Weight alone
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec))))
  expect_silent(as.lmitt(lm(y ~ a.(), data = simdata, weights = ate(spec))))
  expect_silent(as.lmitt(lm(y ~ z.(), data = simdata, weights = ate(spec))))
  expect_silent(as.lmitt(lm(y ~ adopters(), data = simdata, weights = ate(spec))))

  # Adopters alone
  expect_silent(a <- lm(y ~ assigned(spec), data = simdata))
  # teeMod doesnt look in assigned for StudySpecification
  # expect_silent(as.lmitt(a))

  # cov_adj alone
  expect_silent(as.lmitt(lm(y ~ a.(), data = simdata,
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ z.(spec) + offset(cov_adj(camod, specification = spec)),
                            data = simdata)))


  # Weight + assigned
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                            weights = ett(spec))))


  # weight + cov_adj
  expect_silent(as.lmitt(lm(y ~ z.(), data = simdata, weights = ett(spec),
                            offset = cov_adj(camod))))
  expect_silent(as.lmitt(lm(y ~ adopters(), data = simdata, weights = ett(),
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ a.(), data = simdata, weights = ett(spec),
                            offset = cov_adj(camod, specification = spec))))

  # weight + cov_adj in formula
  expect_silent(as.lmitt(lm(y ~ z.() + offset(cov_adj(camod)), data = simdata,
                            weights = ett(spec))))
  # Fails when trying to obtain StudySpecification from a cov_adj inside offset in formula
  #expect_silent(as.lmitt(lm(y ~ z.() + offset(cov_adj(camod, specification = spec)),
  #                      data = simdata, weights = ett())))
  expect_silent(as.lmitt(lm(y ~ z.() + offset(cov_adj(camod, specification = spec)),
                            data = simdata, weights = ett(spec))))

  # assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                            offset = cov_adj(camod, specification = spec))))

  # weights + assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec),
                            offset = cov_adj(camod))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(),
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec),
                            offset = cov_adj(camod, specification = spec))))

  # weight + adopter + cov_adj in formula
  expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod)),
                            data = simdata, weights = ett(spec))))
  # Fails when trying to obtain StudySpecification from a cov_adj inside offset in formula
  #expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod, specification = spec)),
  #                      data = simdata, weights = ett()))
  expect_silent(as.lmitt(lm(y ~ assigned() +
                              offset(cov_adj(camod, specification = spec)),
                            data = simdata, weights = ett(spec))))

})

test_that("binary treatment, not in data2", {
  # Treatment doesn't exist in data
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  camod <- lm(y ~ x, data = simdata)
  simdata$z <- NULL

  # Adopters alone
  expect_silent(a <- lm(y ~ assigned(spec), data = simdata))
  # teeMod doesnt look in assigned for StudySpecification
  # expect_silent(as.lmitt(a))

  # Weight + assigned
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                            weights = ett(spec))))

  # assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                            offset = cov_adj(camod, specification = spec))))

  # weights + assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                                     weights = ett(spec),
                                     offset = cov_adj(camod))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                                     weights = ett(),
                                     offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                                     weights = ett(spec),
                                     offset = cov_adj(camod, specification = spec))))

  # weight + adopter + cov_adj in formula
  expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod)),
                                     data = simdata, weights = ett(spec))))
  # Fails when trying to obtain StudySpecification from a cov_adj inside offset in formula
  #expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod, specification = spec)),
  #                      data = simdata, weights = ett()))
  expect_silent(as.lmitt(lm(y ~ assigned() +
                                       offset(cov_adj(camod, specification = spec)),
                   data = simdata, weights = ett(spec))))

})

test_that("non-binary treatment, in all data, dichotomization in specification", {


  # Treatment exists in data
  data(simdata)

  spec <- rct_spec(dose ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  camod <- lm(y ~ x, data = simdata)


  # Adopters alone
  expect_silent(a <- lm(y ~ assigned(spec, dichotomy = dose >= 200 ~ .), data = simdata))
  # teeMod doesnt look in assigned for StudySpecification
  # expect_silent(as.lmitt(a))

  # Weight + assigned
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                                     weights = ett(spec, dichotomy = dose >= 200 ~ .))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                                     weights = ett(spec, dichotomy = dose >= 200 ~ .))))

  # assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(dichotomy = dose >= 200 ~ .), data = simdata,
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec, dichotomy = dose >= 200 ~ .), data = simdata,
                            offset = cov_adj(camod, specification = spec))))

  # weights + assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod, specification = spec))))

  # weight + adopter + cov_adj in formula
  expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod)),
                            data = simdata, weights = ett(spec, dichotomy = dose >= 200 ~ .))))
  # Fails when trying to obtain StudySpecification from a cov_adj inside offset in formula
  #expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod, specification = spec)),
  #                      data = simdata, weights = ett()))
  expect_silent(as.lmitt(lm(y ~ assigned() +
                              offset(cov_adj(camod, specification = spec)),
                            data = simdata, weights = ett(spec, dichotomy = dose >= 200 ~ .))))

})

test_that("non-binary treatment, not in data2, dichotomization in specification", {


  # Treatment exists in data
  data(simdata)

  spec <- rct_spec(dose ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
  camod <- lm(y ~ x, data = simdata)
  simdata$dose <- NULL


  # Adopters alone
  expect_silent(a <- lm(y ~ assigned(spec, dichotomy = dose >= 200 ~ .), data = simdata))
  # teeMod doesnt look in assigned for StudySpecification
  # expect_silent(as.lmitt(a))

  # Weight + assigned
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .))))

  # assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(dichotomy = dose >= 200 ~ .), data = simdata,
                   offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ assigned(spec, dichotomy = dose >= 200 ~ .), data = simdata,
                   offset = cov_adj(camod, specification = spec))))

  # weights + assigned + cov_adj
  expect_silent(as.lmitt(lm(y ~ assigned(), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod))))
  expect_silent(as.lmitt(lm(y ~ adopters(), data = simdata,
                            weights = ett(dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod, specification = spec))))
  expect_silent(as.lmitt(lm(y ~ z.(), data = simdata,
                            weights = ett(spec, dichotomy = dose >= 200 ~ .),
                            offset = cov_adj(camod, specification = spec))))

  # weight + adopter + cov_adj in formula
  expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod)),
                            data = simdata, weights = ett(spec, dichotomy = dose >= 200 ~ .))))
  # Fails when trying to obtain StudySpecification from a cov_adj inside offset in formula
  #expect_silent(as.lmitt(lm(y ~ assigned() + offset(cov_adj(camod, specification = spec)),
  #                      data = simdata, weights = ett()))
  expect_silent(as.lmitt(lm(y ~ assigned() +
                              offset(cov_adj(camod, specification = spec)),
                            data = simdata, weights = ett(spec, dichotomy = dose >= 200 ~ .))))

})

test_that("issue 243, as.lmitt() finds StudySpecification in assigned()", {
  data(simdata)
  
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)
  lmod <- lm(y ~ assigned(spec), simdata)
  expect_true(inherits(as.lmitt(lmod), "teeMod"))
  lmod <- lm(y ~ a.(spec), simdata)
  expect_true(inherits(as.lmitt(lmod), "teeMod"))
  lmod <- lm(y ~ z.(spec), simdata)
  expect_true(inherits(as.lmitt(lmod), "teeMod"))
  lmod <- lm(y ~ adopters(spec), simdata)
  expect_true(inherits(as.lmitt(lmod), "teeMod"))
  
  teeMods <- lapply(
    seq_len(2),
    function(x) {
      lmod <- lm(y ~ adopters(spec), simdata)
      as.lmitt(lmod)
    }
  )
  expect_true(inherits(as.lmitt(teeMods[[1L]]), "teeMod"))
  expect_true(inherits(as.lmitt(teeMods[[2L]]), "teeMod"))
})

test_that("Obtaining data for weights", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), data = simdata)

  mod <- lm(y ~ x, data = simdata, weights = ate(spec))

  #Handling multiple models in the stack
  mod1 <- lm(predict(lm(y~x, data = simdata, weights = ate(spec))) ~ 1)
  mod2 <- lm(predict(lm(y~x, data = simdata)) ~ 1, weights = ate(spec),
             data = simdata)
  expect_true(mod1$coefficients[1] != mod2$coefficients[1])

  # linear glm and lm are equivalent
  mod3 <- glm(y ~ x, data = simdata, weights = ate(spec))
  expect_equal(mod$coefficients,
               mod3$coefficients)

  # Works using `with`
  mod4 <- with(simdata,
               lm(y ~ x, weights = ate(spec)))
  expect_equal(mod$coefficients,
               mod4$coefficients)

  # Works using `attach`
  attach(simdata)
  mod5 <- lm(y ~ x, weights = ate(spec))
  expect_equal(mod$coefficients,
               mod5$coefficients)
  detach(simdata)

  # Allow simple manipulation of weights
  mod6 <- lm(y ~ x, data = simdata, weights = sqrt(ate(spec)))
  mod7 <- lm(y ~ x, data = simdata, weights = simdata$o*ate(spec))

  # additional arguments in weights
  spec2 <- rct_spec(o ~ cluster(uoa1, uoa2), data = simdata)
  mod8 <- lm(y ~ x, data = simdata, weights = ett(spec2, dichotomy = o >= 3 ~ .))


  # Test for fallback if no model.frame call

  w <- capture_warnings(with(simdata, ate(spec)))
  expect_equal(length(w), 4)
  expect_true(any(sapply(w, grepl, pattern = "No call")))
  expect_true(any(sapply(w, grepl, pattern = "trying fallback")))

  expect_error(.get_data_from_model("weights", 1),
               "must be a formula")

  expect_error(.get_data_from_model("abc", 1))

  # This will error in a nonstandard way; just want to ensure it fails
  f <- function() {
    frm <- y ~ x
    .get_data_from_model("weights", str2lang("frm"))
  }
  capture_warnings(expect_error(f()))

  f2 <- function() {
    frm <- 1
    .get_data_from_model("weights", str2lang("frm"))
  }
  expect_error(f2(), "unable to convert")

})

test_that("get_data_from_model finds data in lmitt object's eval env", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, specification = spec, data = simdata)

  lmitt_func <- function(x, ...) {
    rhs <- formula(x)[[3L]]
    eval(rhs, environment(formula(x)))
  }

  #expect_warning(lmitt_func(ssmod), "No call to `model.frame`")
  out <- model.matrix(formula(ssmod), environment(formula(ssmod))$data)
  expect_true(is.numeric(out))
  expect_equal(dim(out), c(nrow(simdata), 2))
  expect_true(all(out[, 2] == ssmod$model[, 2]))
})

test_that(paste("get_data_from_model frame recursion doesn't return",
                "data from lmitt found in global env"), {
  rlang_trace_top_env <- getOption("rlang_trace_top_env")
  on.exit(options(rlang_trace_top_env = rlang_trace_top_env))

  options(rlang_trace_top_env = environment())

  data(simdata)
  data <- simdata
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data)
  ssmod <- lmitt(y ~ 1, specification = spec, data = data)

  non_lmitt_func <- function(data = NULL) {
    if (is.null(data)) data <- .get_data_from_model("weights",
                                                    y ~ assigned(specification = spec))
    data$newcol <- 5
    return(data)
  }

  expect_warning(expect_warning(non_lmitt_func(),
                                "No call to `model.frame`"),
                 "Unable to detect")
  out <- suppressWarnings(non_lmitt_func())
  expect_identical(out, cbind(simdata, newcol = 5))
})

#' @include StudySpecification.R WeightedStudySpecification.R StudySpecificationAccessors.R SandwichLayerVariance.R
#' @include teeMod.R
NULL

##' @title Convert \code{lm} object into \code{teeMod}
##'
##' @description Converts the output of [lm()] into a \code{teeMod}
##'   object, for standard errors that account for block and cluster
##'   information carried with the \code{lm}'s weights, and/or an
##'   offset incorporating predictions of the outcome from a
##'   separate model.
##'
##' @details The formula with which \code{x} was created must include
##'   a treatment identifier (e.g. [assigned()]).  If a model-based
##'   offset is incorporated, the model's predictions would have to
##'   have been extracted using [cov_adj()] as opposed to
##'   \code{predict{}} in order for \code{teeMod} standard error
##'   calculations to reflect propagation of error from these
##'   predictions. This mechanism only supports treatment main effects:
##'   to estimate interactions of treatment assignment with a moderator
##'   variable, use [lmitt()] instead of \code{lm()} followed by
##'   \code{as.lmitt()}.
##'
##' @param x \code{lm} object with weights containing a
##'   \code{WeightedStudySpecification}, or an offset from [cov_adj()].
##' @param specification Optional, explicitly specify the
##'   \code{StudySpecification} to be used. If the \code{StudySpecification} is
##'   specified elsewhere in \code{x} (e.g. passed as an argument to any of
##'   \code{ate()}, \code{ett()}, \code{cov_adj()} or \code{assigned()}) it will
##'   be found automatically and does not need to be passed here as well. If
##'   different \code{StudySpecification} objects are passed (either through the
##'   \code{lm} in weights or covariance adjustment, or through this argument),
##'   an error will be produced.
##' @return \code{teeMod} object
##' @rdname as_lmitt
##' @importFrom stats formula weights
##' @export
as.lmitt <- function(x, specification = NULL) {
  if (!inherits(x, "lm")) {
    stop("input must be lm object")
  }


  ### 10/31/22 JE - The below causes a ton of errors in the test suite that do
  ### NOT show up interactively and thus are extremely challenging to debug. It
  ### is supposed to replace the treatment variable name with `assigned()` and
  ### refit the model. I've spent a few days trying to debug this and failing,
  ### so going to leave it as unfixable for the moment, and instead simply error
  ### if the user does not include `assigned()` in the `lm`.
  ## # Update formula to use `assigned()` if needed
  ## ff <- stats::formula(x)
  ## newff <- formula(gsub(var_names(specification, "t"), "assigned()", deparse1(ff)))
  ## environment(newff) <- environment(ff)

  ## # Ensure updated model will be the same.

  ## newx <- update(x, newff)
  ## if (!isTRUE(all.equal(newx$coefficients, x$coefficients,
  ##                       check.attributes = FALSE))) {
  ##   stop(paste("Treatment variable found in model formula.",
  ##              "Updating model to use `assigned()` instead produces",
  ##              "different results. Please refit original model using",
  ##              "`assigned()` in place of treatment variable name."))
  ## }
  ## x <- newx

  tt <- terms(stats::formula(x),
              specials = c("assigned", "a.", "z.", "adopters"))

  if (all(missing_alias <- vapply(attr(tt, "specials"), is.null, logical(1))) |
      sum(!missing_alias) > 1) {
    stop(paste("Exactly one of `assigned()` or its aliases must be in the",
               "model formula in place of the treatmentvariable name. The",
               "`lmitt()` function may be used to avoid explicitly indicating",
               "`assigned()`."))
  }

  ## find the study specification
  # first search the assignment function located via the specials attribute;
  # it indicates where the assignment function is in the `variables` attribute
  # of `tt` (which is a call, so we add 1). if it isn't given by a named arg,
  # then it must be provided as the first unnamed arg.
  a.call <- attr(tt, "variables")[[attr(tt, "specials")[[
    names(which(!missing_alias))]]+1]]
  if (is.null(specification)) {
    if (!is.null(names(a.call))) {
      specification <- a.call$s
      if (is.null(specification) & max(which(names(a.call) == "")) > 1) {
        specification <- a.call[[max(which(names(a.call) == ""))]]
      }
    } else {
      if (length(a.call) > 1) specification <- a.call[[2L]]
    }
  }

  # check the weights
  if (is.null(specification) &
      (has_ws <- inherits(x$model$`(weights)`, "WeightedStudySpecification"))) {
    specification <- x$model$`(weights)`@StudySpecification
  }
  
  # then the offset
  if (is.null(specification) & inherits(x$model$`(offset)`, "SandwichLayer")) {
    specification <- x$model$`(offset)`@StudySpecification
  }

  if (!inherits(specification, "StudySpecification")) {
    specification <- eval(specification, parent.frame())
  }
  
  ## find the dichotomy to pass on
  dichotomy <- a.call$dichotomy
  if (is.null(dichotomy) & length(a.call) == 4) dichotomy <- a.call[[4L]]
  if (is.null(dichotomy) & has_ws) {
    dichotomy <- x$model$`(weights)`@dichotomy
  }
  if (inherits(dichotomy, "formula") & length(dichotomy) == 0) dichotomy <- NULL
  
  #### Obtain the proper call.
  # The first conditional is the user calls `lmitt()` and passes an `lm` object.
  # The second conditional if the user calls `lmitt.lm()` directly
  # The third conditional is if the user calls `as.lmitt()`.

  # `&&` necessary to return FALSE immediately if not enough frames on stack
  if (sys.nframe() >=3 &&
      !is.null(sys.call(-2)) &&
      sys.call(-2)[[1]] == as.name("lmitt")) {
    # If we're in `lmitt.lm()` via `lmitt()`, save the call to `lmitt`.
    lmitt_call <- sys.call(-2)
  } else if (sys.nframe() >= 2 &&
             is.name(sys.call(-1)[[1]]) && # Catches mapply issue
             !is.null(sys.call(-1)) &&
             sys.call(-1)[[1]] == as.name("lmitt.lm")) {
    # If we're in `lmitt.lm()` directly, save that call.
    lmitt_call <- sys.call(-1)
  } else {
    # Otherwise save the `as.lmitt()` call
    lmitt_call <- sys.call()
  }
  return(.convert_to_lmitt(x,
                           specification,
                           lmitt_fitted = FALSE,
                           absorbed_intercepts = FALSE,
                           moderator = vector("character"),
                           ctrl_means_form = lhs ~ 1,
                           dichotomy = dichotomy,
                           lmitt_call = lmitt_call))
}

##' @rdname as_lmitt
##' @export
as.teeMod <- as.lmitt

.convert_to_lmitt <- function(lm_model,
                              specification,
                              lmitt_fitted,
                              absorbed_intercepts,
                              moderator,
                              ctrl_means_form,
                              dichotomy,
                              lmitt_call) {
  if (!inherits(lm_model, "lm")) {
    stop("input must be lm object")
  }

  # Ensure `specification=` is a proper object
  if (!is.null(specification) & !is(specification, "StudySpecification")) {
    # Allow WeightedStudySpecification just in case
    if (is(specification, "WeightedStudySpecification")) {
      specification <- specification@StudySpecification
    } else {
      stop(paste("If provided, `specification` must be a `StudySpecification` or",
                 "`WeightedStudySpecification` object"))
    }
  }

  # Check if we can find a specification in either Weights (preferred) or cov_adj
  specification_weights <- tryCatch(lm_model$model$"(weights)"@StudySpecification,
                             error = function(e) NULL)
  specification_cov_adj <- tryCatch(.get_cov_adj(lm_model)@StudySpecification,
                             error = function(e) NULL)

  specifications <- list(specification, specification_weights, specification_cov_adj)
  specifications <- specifications[!vapply(specifications, is.null, logical(1))]

  # The list contains all specifications possible found (one passed in, and one in each
  # of weights and cov_adj). Passing `unique` removes any duplicates (since
  # duplicates are OK).
  unique_specs <- unique(specifications)

  # At this point, if the length of `unique_specs` is 1, we're done. More than
  # one is an error.
  if (length(unique_specs) == 1) {
    specification <- unique_specs[[1]]
  } else if (length(unique_specs) > 1) {
    stop("Multiple differing `StudySpecification` found in object.")
  } else {
    # This should never be hit
    stop("Cannot locate a `StudySpecification`, pass via it `specification=` argument")
  }

  # #123 Make PreSandwichLayer to SandwichLayer if necessary
  ca <- .get_cov_adj(lm_model)
  if (is(ca, "PreSandwichLayer") & !is(ca, "SandwichLayer")) {
    lm_model$model$`(offset)` <- as.SandwichLayer(.get_cov_adj(lm_model),
                                                  specification = specification)
  }

  eval_env <- new.env(parent = environment(formula(lm_model)))
  # Find data
  if (lmitt_fitted) {
    # If `lmitt.formula` is called, get the data from there directly (since
    # inside `lmitt.formula`, we pass in the data directly after appenindg on
    # the updated RHS and LHS).
    data <- lm_model$call$data
  } else {
    # If `as.lmitt` (or `lmitt.lm`), evaluate the lm call's data using a
    # `fallback_data_search`-esque approach
    data_call <- lm_model$call$data
    #for (i in c(-1L, seq_len(sys.nframe()))) {
    for (i in seq_len(length(sys.calls()))) {
      try(data <- eval(data_call, envir = parent.frame(i)),
          silent = TRUE)
      if (!is.null(data) && is.data.frame(data)) {
        break()
      }
    }
    if (!inherits(data, "data.frame")) {
      stop("Could not determine appropriate data")
    }
  }
  
  ## code block for getting control means and tacking them onto coefficients
  if (is.null(ctrl_means_data <- mget("ctrl_means_data", envir = environment(ctrl_means_form),
                                      ifnotfound = list(NULL))[[1]])) {
    ctrl_means_data <- data
  }
  blks <- blocks(specification, ctrl_means_data, all.x = TRUE, implicit = TRUE)[,1]
  a_col <- a.(specification = specification, dichotomy = dichotomy, data = data)
  keep_rows <- which(
    if (is.null(sbs <- eval(lm_model$call$subset, ctrl_means_data))) {
      rep(TRUE, length(blks))
    } else sbs)
  ctrl_means_wts <- numeric(length(a_col))
  ctrl_means_wts[keep_rows] <- if (is.null(
    wts <- with(mod_copy <- lm_model, {
      if (!is.null(mod_copy$na.action)) class(mod_copy$na.action) <- "exclude"
      stats::weights(mod_copy)
    }))) rep(1, length(keep_rows)) else wts

  if (absorbed_intercepts) {
    mods <- if (length(moderator)) eval(str2lang(moderator), ctrl_means_data) else moderator
    ix <- if (inherits(mods, c("factor", "character"))) {
      paste(mods, blks, sep = "_")
    } else {
      blks
    }
    pis <- rowsum(ctrl_means_wts * a_col, ix) / rowsum(ctrl_means_wts, ix)
    ctrl_means_wts <- ctrl_means_wts * pis[ix,]
    ctrl_means_wts[is.na(blks) | is.na(mods)] <- 0
  }
  ctrl_means_wts <- ctrl_means_wts * (1 - a_col)
  
  if (!is.null(os <- lm_model$model$`(offset)`)) {
    model_keep_rows <- setdiff(keep_rows, stats::na.action(lm_model))
    os <- replace(numeric(length(a_col)), model_keep_rows, os)
  }
  
  ctrl_means_env <- new.env()
  assign("data", ctrl_means_data, envir = ctrl_means_env)
  environment(ctrl_means_form) <- ctrl_means_env
  ctrl_means_form[[2L]] <- quote(
    do.call(cbind,
            stats::setNames(list(data[[stats::formula(lm_model)[[2]]]], os),
                            c(stats::formula(lm_model)[[2]], "cov_adj")))
  )
  ctrl_means_cl <- lm_model$call[c(1, match(c("formula", "subset"), names(lm_model$call), 0))]
  ctrl_means_cl[[2]] <- ctrl_means_form
  ctrl_means_cl$data <- ctrl_means_data
  ctrl_means_cl$weights <- ctrl_means_wts
  ctrl_means_cl$na.action <- stats::na.exclude

  ctrl_means_lm <- eval(ctrl_means_cl)
  ctrl_means <- ctrl_means_lm$coefficients
  ctrl_means_labels <- paste(
    rep(if (is.null(cnms <- colnames(ctrl_means))) deparse1(stats::formula(lm_model)[[2]]) else cnms,
        each = if (is.null(nr <- nrow(ctrl_means))) 1 else nr),
    if (is.null(rnms <- rep(row.names(ctrl_means), ncol(ctrl_means)))) names(ctrl_means) else rnms,
    sep = ":")
  lm_model$coefficients <- c(lm_model$coefficients,
                             stats::setNames(c(ctrl_means), ctrl_means_labels))

  lm_model$call$data <- data
  # set call's na.action to na.pass so expand.model.frame includes NA rows
  # as our internal API calls for maintaining NA rows as much as possible
  lm_model$call$na.action <- "na.pass"
  assign("data", data, envir = eval_env)
  assign("specification", specification, envir = eval_env)
  environment(lm_model$terms) <- eval_env
  if (inherits(lm_model, "glm")) {
    lm_model$formula <- as.formula(lm_model, env = eval_env)
  }

  return(new("teeMod",
             lm_model,
             StudySpecification = specification,
             absorbed_intercepts = absorbed_intercepts,
             moderator = moderator,
             ctrl_means_model = ctrl_means_lm,
             lmitt_call = call("quote", lmitt_call),
             lmitt_fitted = lmitt_fitted))

}

##' @title Extract empirical estimating equations from a \code{glmbrob} model
##'   fit
##' @param x a fitted \code{glmrob} object
##' @param ... arguments passed to methods
##' @return matrix, estimating functions evaluated at data points and fitted
##'   parameters
##' @importFrom stats family model.matrix
##' @importFrom utils getFromNamespace
##' @rdname glmrob_methods
##' @exportS3Method
estfun.glmrob <- function(x, ...) {

  ## numbers refer to which lines in
  ## https://github.com/cran/robustbase/blob/335b69f2310bd21ca4cdfc17a2a99ebbcad84017/R/glmrobMqle.R
  ## the expressions were derived or copied from

  ## this was copied verbatim from lines 121-155
  switch(family(x)$family,
         "binomial" = {
           Epsi.init <- utils::getFromNamespace("EpsiBin.init", "robustbase")
           Epsi <- utils::getFromNamespace("EpsiBin", "robustbase")
           EpsiS <- utils::getFromNamespace("EpsiSBin", "robustbase")
           Epsi2 <- utils::getFromNamespace("Epsi2Bin", "robustbase")
           phiEst <- phiEst.cl <- 1
         },
         "poisson" = {
           Epsi.init <- utils::getFromNamespace("EpsiPois.init", "robustbase")
           Epsi <- utils::getFromNamespace("EpsiPois", "robustbase")
           EpsiS <- utils::getFromNamespace("EpsiSPois", "robustbase")
           Epsi2 <- utils::getFromNamespace("Epsi2Pois", "robustbase")
           phiEst <- phiEst.cl <- expression({1})
         },
         "gaussian" = {
           Epsi.init <- utils::getFromNamespace("EpsiGaussian.init", "robustbase")
           Epsi <- utils::getFromNamespace("EpsiGaussian", "robustbase")
           EpsiS <- utils::getFromNamespace("EpsiSGaussian", "robustbase")
           Epsi2 <- utils::getFromNamespace("Epsi2Gaussian", "robustbase")
           phiEst.cl <- utils::getFromNamespace("phiGaussianEst.cl", "robustbase")
           phiEst <- utils::getFromNamespace("phiGaussianEst", "robustbase")
         },
         "Gamma" = { ## added by ARu
           # Epsi.init <- utils::getFromNamespace("EpsiGamma.init", "robustbase")
           #### The `Gmn` function inside `EpsiGamma.init` was not being
           #### properly found. Bringing it in here manually and using
           #### `getFromNamespace` to locate it. JE 3/2024.
           Epsi.init <- expression({
             nu <- 1/phi
             snu <- 1/sqrt(phi)
             pPtc <- pgamma(snu + c(-tcc, tcc), shape = nu, rate = snu)
             pMtc <- pPtc[1]
             pPtc <- pPtc[2]
             aux2 <- tcc * snu
             GLtcc <- getFromNamespace("Gmn", "robustbase")(-tcc, nu)
             GUtcc <- getFromNamespace("Gmn", "robustbase")(tcc, nu)
           })
           Epsi <- utils::getFromNamespace("EpsiGamma", "robustbase")
           EpsiS <- utils::getFromNamespace("EpsiSGamma", "robustbase")
           Epsi2 <- utils::getFromNamespace("Epsi2Gamma", "robustbase")
           phiEst.cl <- utils::getFromNamespace("phiGammaEst.cl", "robustbase")
           phiEst <- utils::getFromNamespace("phiGammaEst", "robustbase")
           },
         ## else
         stop(gettextf("family '%s' not yet implemented", family$family),
              domain=NA)
         )

  ### change names of returned objects to what they were in original robustbase
  ### code
  ni <- x$ni
  tcc <- x$tcc
  eta <- x$linear.predictors
  phi <- x$dispersion
  mu <- x$fitted.values

  ### get V(mu)
  rr <- x$y-mu
  sni <- sqrt(ni)
  sVF <- sni*rr/x$residuals ## 164 (inverted)
  sV <- sVF*sqrt(x$dispersion) #168
  Vmu <- sVF^2 ## 163 (inverted)


 ### robustbase code for calculating cpsi=psi_c - E[psi_c]
  ##177-78
  K <- floor(mu*ni + tcc* sni*sV)
  H <- floor(mu*ni - tcc* sni*sV)

  eval(Epsi.init)
  residPS <- x$residual/sqrt(phi) # scaled Pearson residuals (169)
  cpsi <- pmax.int(-tcc,pmin.int(residPS,tcc))-eval(Epsi) ## 210

  ### dPsi/dBeta=dPsi/d eta d eta/d Beta=dPsi/d eta X
  dmu.deta <- family(x)$mu.eta(eta)
  X <- model.matrix(x)

  cpsi*x$w.x*sni/sV*dmu.deta*X
}

##' @title Extract bread matrix from an \code{lmrob()} fit
##' @param x a fitted \code{lmrob} object
##' @param ... arguments passed to methods
##' @return matrix, inverse Hessian of loss as evaluated at fitted parameters
##' @rdname glmrob_methods
##' @exportS3Method
bread.glmrob <- function(x, ...) {
  return(solve(x$matM))
}

##' @title Produce confidence intervals for linear models
##' @details \code{.confint_lm()} is a copy of \code{stats::confint.lm} but
##'   passes arguments in \code{...} to the \code{vcov()} call. When called on a
##'   \code{teeMod} model, this produces confidence intervals where standard
##'   errors are computed based on the desired formulation of the
##'   \code{vcov_tee()} call.
##'
##' @param object a fitted \code{teeMod} model
##' @param parm a specification of which parameters are to be given confidence
##'   intervals, either a vector of numbers or a vector of names. If missing,
##'   all parameters are considered.
##' @param level the confidence level required.
##' @param ... additional arguments to pass to \code{vcov.teeMod()}
##' @return A matrix (or vector) with columns giving lower and upper confidence
##'   limits for each parameter. These will be labelled as (1-level)/2 and 1 -
##'   (1-level)/2 in % (by default 2.5% and 97.5%)
##' @keywords internal
.confint_lm <- function(object, parm, level = 0.95, ...) {
  cf <- stats::coef(object)
  ses <- sqrt(diag(vcov(object, ...)))
  pnames <- names(ses)
  if (is.matrix(cf))
    cf <- stats::setNames(as.vector(cf), pnames)
  if (missing(parm))
    parm <- pnames
  else if (is.numeric(parm))
    parm <- pnames[parm]
  a <- (1 - level)/2
  a <- c(a, 1 - a)
  fac <- stats::qt(a, object$df.residual)
  pct <- paste(round(a, 3) * 100, "%")
  ci <- array(NA_real_, dim = c(length(parm), 2L), dimnames = list(parm,
                                                                   pct))
  ci[] <- cf[parm] + ses[parm] %o% fac
  ci
}

#' @include StudySpecification.R SandwichLayer.R block_center_residuals.R
NULL

#' @title Variance/Covariance for \code{teeMod} objects
#'
#' @description Compute robust sandwich variance estimates with optional
#'   covariance adjustment
#'
#' @details Variance estimates will be clustered on the basis of the columns
#'  provided to \code{cluster} (or obtained by the default behavior). As a result,
#'  providing \code{"HCx"} or \code{"CRx"} to \code{type} will produce the same
#'  variance estimate given that \code{cluster} remains the same.
#'
#' With prior covariance adjustment, unless the \code{data} argument of the covariance
#'  model fit is the same as the \code{data} argument for fitting \code{x} and the
#'  \code{StudySpecification} of \code{x} has been created with a formula of the
#'  form `trt_col ~ 1`, the column(s) provided to \code{cluster} must appear in the
#'  dataframes in both \code{data} arguments, even if the clustering structure does
#'  not exist, per se, in the covariance adjustment sample. For instance, in a
#'  finely stratified randomized trial, one might desire standard errors clustered
#'  at the block level, but the covariance adjustment model may include auxiliary
#'  units that did not participate in the trial. In this case, in the \code{data}
#'  argument of the fitted covariance model, the column(s) passed to \code{cluster}
#'  should have the block ID's for rows overlapping with the \code{data} argument
#'  used for fitting \code{x}, and NA's for any auxiliary units. \code{vcov_tee()}
#'  will treat each row with an NA as its own cluster.
#'  
#' For ITT effect estimates without covariance adjustment, \code{type}
#'  corresponds to the variance estimate desired. Supported options include:
#'
#' - \code{"MB0"}, \code{"HC0"}, and \code{"CR0"} for model-based HC/CR0 standard errors
#' - \code{"MB1"}, \code{"HC1"}, and \code{"CR1"} for model-based HC/CR1 standard errors
#' (for \code{"MB1"} and \code{"HC1"}, this is \eqn{n/(n - 2)}, and for \code{"CR1"},
#' this is \eqn{g\cdot(n-1)/((g-1)\cdot(n-2))}, where \eqn{g} is the number of
#' clusters in the sample used for fitting \code{x})
#' - \code{"MB2"}, \code{"HC2"}, and \code{"CR2"} for model-based HC/CR2 standard errors
#' - \code{"DB0"} for design-based HC0 variance estimates
#' 
#' The \code{type} argument does not correspond to existing variance estimators in
#'   the literature in the case of prior covariance adjustment. It specifies the
#'   bias correction to the residuals of \code{x}, but the residuals of the covariance
#'  model are corrected separately based on the \code{cov_adj_rcorrect} argument.
#'  The \code{cov_adj_rcorrect} argument takes the same options as \code{type}
#'  except \code{"DB0"}. When the covariance model includes rows in the treatment
#'  condition for fitting, the residuals of \code{x} are further corrected by
#'  having the values of \code{offset} replaced by predictions that use coefficient
#'  estimates that leave out rows in the same cluster (as defined by the
#'  \code{cluster} argument).
#'
#' @param x a fitted \code{teeMod} model
#' @param type a string indicating the desired bias correction for the residuals
#'   of \code{x}. Default makes no bias correction. See Details for supported types
#' @param cluster a vector indicating the columns that define clusters. The default
#'   is the unit of assignment columns in the \code{StudySpecification} stored
#'   in \code{x}. These columns should appear in the dataframe used for fitting
#'   \code{x} as well as the dataframe passed to the covariance model fit in the case of
#'   prior covariance adjustment. See Details
#' @param ... arguments to be passed to the internal variance estimation
#'   function, such as \code{cov_adj_rcorrect} and \code{loco_residuals}. If
#'   \code{x} has a \code{SandwichLayer} object in its offset, The former
#'   specifies the bias correction to the residuals of the covariance model, and
#'   the latter indicates whether the offset should be replaced with predictions
#'   from leave-one-cluster-out fits of the covariance adjustment model.
#'   See Details
#' @return A variance-covariance matrix with row and column entries for the estimated
#'   coefficients in \code{x}, the marginal mean outcome in the control condition,
#'   the marginal mean \code{offset} in the control condition (if an \code{offset}
#'   is provided), and if a moderator variable is specified in the formula for \code{x},
#'   the mean interaction in the control condition of the outcome and \code{offset}
#'   with the moderator variable
#' @export
#' @rdname var_estimators
vcov_tee <- function(x, type = NULL, cluster = NULL, ...) {
  if (is.null(type)) type <- "HC0"

  vcov_type <- substr(type, 1, 2)
  if (type != "DB0" & !(vcov_type %in% c("HC", "CR", "MB"))) {
    stop("covariance function not defined.\n")
  }
  var_func <- get(paste0(".vcov_", ifelse(vcov_type == "DB", type, vcov_type)))
  
  args <- list(...)
  args$x <- x
  
  if (vcov_type == "DB") {
    args$vcov_type <- vcov_type
  } else {
    args$type <- type
  }
  
  if (is.null(args$cov_adj_rcorrect) & inherits(x, "teeMod") &
      inherits(x$model$`(offset)`, "SandwichLayer")) args$cov_adj_rcorrect <- "HC0"

  if (is.null(args$by)) args$by <- cluster # if cov_adj() was not fit with a "by" argument, this is passed to .order_samples() to order rows of estfun() output
  args$cluster_cols <- cluster
  args$cluster <- .make_uoa_ids(x, vcov_type = vcov_type, cluster, ...) # passed on to meatCL to aggregate SE's at the level given by `cluster`

  est <- do.call(var_func, args)

  return(est)
}

#' @keywords internal
.vcov_CR <- function(x, ...) {
  if (!inherits(x, c("teeMod", "mmm"))) {
    stop("x must be a teeMod or mmm object")
  }

  args <- list(...)
  args$x <- x
  args$itt_rcorrect <- args$type
  args$type <- "HC0" # sandwich::meatCL won't correct meat after we've corrected
  n <- length(args$cluster)
  args$cls <- args$cluster # pass to estfun.teeMod and internal calls for speed

  bread. <- sandwich::bread(x, ...)
  meat. <- do.call(sandwich::meatCL, args)
  vmat <- (1 / n) * bread. %*% meat. %*% t(bread.)

  # NA any invalid estimates due to degrees of freedom checks
  if (inherits(x, "teeMod")) {
    if (is.null(values_from <- args$values_from)) values_from <- x
    vmat <- .check_df_moderator_estimates(vmat, x, args$cluster_cols, values_from)
  } else {
    start_ix <- 0
    for (mod_ix in seq_along(x)) {
      mod <- x[[mod_ix]]
      if (is.null(values_from <- args$values_from)) values_from <- mod
      not_aliased <- !is.na(mod$coefficients)
      vmat_ix <- start_ix + seq_len(sum(not_aliased))
      vmat[vmat_ix, vmat_ix] <- .check_df_moderator_estimates(
        vmat, mod, args$cluster_cols, values_from)[vmat_ix, vmat_ix]
      start_ix <- start_ix + sum(not_aliased)
    }
    vmat[apply(is.na(vmat), 1, any),] <- NA_real_
    vmat[,apply(is.na(vmat), 2, any)] <- NA_real_
  }

  attr(vmat, "type") <- match.call()$type
  attr(vmat, "cov_adj_rcorrect") <- args$cov_adj_rcorrect
  return(vmat)
}

#' @keywords internal
.vcov_HC <- function(x, ...) {
  out <- .vcov_CR(x, ...)
  return(out)
}

#' @keywords internal
.vcov_MB <- function(x, ...) {
  out <- .vcov_CR(x, ...)
  return(out)
}

#' @title (Internal) Compute the degrees of freedom of a contrast of a sandwich
#'  variance estimate associated with a \code{teeMod}
#' @details With \code{dof_type="IK"}, one obtains a degrees of freedom estimate
#'  as given by the procedure propose in Imbens and Kolesár (2016). For more
#'  details see the documentation of \code{.compute_IK_dof()}. With 
#'  \code{dof_type="stata"}, one obtains degrees of freedom equal to the number
#'  of clusters less one, the default provided by STATA software (see Cameron
#'  and Miller, 2015 or Bell and McCaffrey, 2002).
#'  
#'  
#'  \code{ell} should be an integer specifying a location in the vector
#'  of estimated coefficients (ignoring the coefficients suffixed by
#'  \code{:(Intercept)}) or a vector of the same length as the vector of
#'  estimated coefficients (ignoring the coefficients suffixed by
#'  \code{:(Intercept)}).
#'  
#'  \code{vcov.type} takes the same arguments as
#'  the \code{type} argument in \code{vcov_tee()}.
#'  
#'  
#'  \code{cluster_ids} should be ordered in alignment with the dataframe passed
#'  to \code{lmitt()}. It should not exclude NA's because the function will
#'  exclude them where necessary.
#' @param x \code{teeMod} object.
#' @param ell numeric.
#' @param vcov.type character.
#' @param dof.type character, either \code{IK} or \code{stata}.
#' @param cluster_ids optional, vector of ID's for clustering degrees of freedom
#'  estimate. If not provided, default is the ID's associated with
#'  \code{cluster}.
#' @param cluster optional, character identifiying the clustering variable
#'  if \code{cluster_ids} is not provided. If not provided, defaults to the
#'  unit of assignment columns specified in the \code{StudySpecification}.
#' @param ... Additional arguments passed from calls higher up the stack. These
#'  arguments are not used within this function.
#' @keywords internal
#' @references Guido W. Imbens and Michael Kolesár. Robust Standard Errors in
#' Small Samples: Some Practical Advice".
#' *The Review of Economics and Statistics*, 98(4):701-712, October 2016.
#' 
#' 
#' A. Colin Cameron and Douglas L. Miller. A Practitioner's Guide to
#' Cluster-Robust Inference. *The Journal of Human Resources*, 50(2):317-372,
#' 2015.
#' 
#' 
#' Robert M. Bell and Daniel F. McCaffrey. Bias Reduction in
#'  Standard Errors for Linear Regression with Multi-Stage Samples.
#'  *Survey Methodology*, 28(2):169-181, December 2002.
.get_dof <- function(x,
                     ell,
                     vcov.type,
                     dof.type = c("stata", "IK"),
                     cluster_ids = NULL,
                     cluster = NULL,
                     ...) {
  if (!inherits(x, "teeMod")) stop("x must be a teeMod object")
  dof.type <- match.arg(dof.type)
  if (is.null(cluster_ids)) {
    if (is.null(cluster)) cluster <- var_names(x@StudySpecification, "u")
    cluster_ids <- .make_uoa_ids(x, substr(vcov.type, 1, 2), cluster)
    if (inherits(x$model$`(offset)`, "SandwichLayer")) {
      ord <- .order_samples(x)
      cluster_ids <- cluster_ids[seq_along(c(ord$Q_not_C, ord$Q_in_C))][
        match(seq_along(c(ord$Q_not_C, ord$Q_in_C)),
              as.numeric(c(names(ord$Q_not_C), names(ord$Q_in_C))))
      ]
    }
  }
  cluster_ids <- factor(cluster_ids)
  
  if (dof.type == "stata") {
    dof <- length(levels(
      droplevels(cluster_ids[setdiff(seq_along(cluster_ids), na.action(x))])
    )) - 1
  } else {
    # compute dof using IK approach
    k <- ncol(x$qr$qr)
    if (length(ell) == 1) {
      ell <- replace(numeric(k), ell, 1)
    } else if (length(ell) != k) {
      stop("ell must have the same length as the rank of the fitted model")
    } else if (length(setdiff(unique(ell), c(0, 1))) != 0) {
      stop("ell must have only zeros or ones")
    }
    dof <- .compute_IK_dof(x,
                           ell = ell,
                           vcov.type = vcov.type,
                           cluster_ids = cluster_ids)
  }
  
  return(dof)
}

#' @title (Internal) Compute the degrees of freedom of a contrast of a
#'  sandwich variance estimate as proposed in Imbens and Kolesár (2016)
#' @details \code{ell} should be a vector of length equal to the number of
#'  estimated coefficients in the \code{teeMod} object. This excludes
#'  coefficients printed in \code{show.teeMod} with \code{:(Intercept)}
#'  suffixes. The degrees of freedom for a single standard error will specify
#'  for \code{ell} a vector of all zeros except one element, which will have a
#'  1 in the location corresponding to the coefficient of interest.
#'  
#'  
#'  \code{cluster_ids} should be ordered in alignment with the dataframe passed
#'  to \code{lmitt()}. It should not exclude NA's because the function will
#'  exclude them where necessary.
#'  
#'  \code{vcov.type} takes the same arguments
#'  as the \code{type} argument in \code{vcov_tee()}.
#' @param tm \code{teeMod} object.
#' @param ell numeric vector.
#' @param vcov.type character.
#' @param cluster_ids optional, vector of ID's for clustering degrees of freedom
#'  estimate. If not provided, default is the ID's associated with
#'  \code{cluster}.
#' @param cluster optional, character identifiying the clustering variable
#'  if \code{cluster_ids} is not provided. If not provided, defaults to the
#'  unit of assignment columns specified in the \code{StudySpecification}.
#' @param tol optional, numeric. Should not be changed.
#' @importFrom stats na.action residuals
#' @keywords internal
#' @references Guido W. Imbens and Michael Kolesár. Robust Standard Errors in
#' Small Samples: Some Practical Advice".
#' *The Review of Economics and Statistics*, 98(4):701-712, October 2016.
.compute_IK_dof <- function(tm,
                            ell,
                            vcov.type,
                            cluster_ids = NULL,
                            cluster = NULL,
                            tol = 1e-09) {
  # make na.action = exclude so residuals.default() includes NA's
  if (!is.null(tm$na.action)) class(tm$na.action) <- "exclude"
  if (is.null(cluster_ids)) {
    # put the ID's in the order of the fitting dataframe if they've been
    # re-ordered due to the presence of a covariance adjustment model
    if (is.null(cluster)) cluster <- var_names(tm@StudySpecification, "u")
    cluster_ids <- .make_uoa_ids(tm, substr(vcov.type, 1, 2), cluster)
    if (inherits(tm$model$`(offset)`, "SandwichLayer")) {
      ord <- .order_samples(tm)
      cluster_ids <- cluster_ids[seq_along(c(ord$Q_not_C, ord$Q_in_C))][
        match(seq_along(c(ord$Q_not_C, ord$Q_in_C)),
              as.numeric(c(names(ord$Q_not_C), names(ord$Q_in_C))))
        ]
    }
  }
  cluster_ids <- factor(cluster_ids)
  # make a copy that drops excluded observations, but keep the full vector to
  # pass to cluster_iss()
  cls <- droplevels(cluster_ids[setdiff(seq_along(cluster_ids), na.action(tm))])

  lmitt_data <- get("data", environment(formula(tm)))
  if (!is.null(sbst <- tm@lmitt_call$subset)) lmitt_data <- subset(
    lmitt_data, eval(sbst, envir = lmitt_data))
  wc <- tm@lmitt_call$weights
  if (inherits(wc, "call")) {
    if (exists(deparse1(wc[[1L]]))) {
      if (length(wc) == 1) {
        wc$specification <- get("specification", environment(formula(tm)))
      } else if (!inherits(eval(wc[[2L]], environment(formula(tm))),
                           "StudySpecification")) {
        wc$specification <- get("specification", environment(formula(tm)))
      }
      if (is.null(wc$data)) wc$data <- lmitt_data
    }
  }
  trts <- if (is.null(wc <- eval(wc, envir = environment(formula(tm)))) |
               !inherits(wc, "WeightedStudySpecification")) {
    treatment(tm@StudySpecification, newdata = lmitt_data)[,1]
  } else {
    if (length(dich <- wc@dichotomy) == 0) {
      treatment(tm@StudySpecification, newdata = lmitt_data)[,1]
    } else {
      treatment(tm@StudySpecification,
                newdata = lmitt_data,
                dichotomy = dich)[,1]
    }
  }
  
  ## estimate rho
  r <- residuals(tm, type = "response")
  r <- r[setdiff(seq_along(r), na.action(tm))]
  if (length(cls) == length(unique(cls))) {
    # for nonclustered data rho = 0
    rho <- 0
  } else {
    # NOTE: in issue #247, we update this calculation to match what's in the
    # dfadjustse function in the dfadjust package
    rho <- (sum(tapply(r, cls, sum)^2) - sum(r^2)) /
      (sum(tapply(rep(1, length(cls)), cls, sum)^2) - length(cls))
    rho <- max(rho, 0) # IK code doesn't allow negative rho
  }
  
  ## estimate sigma
  sig2 <- mean(r^2) - rho
  
  ## this is IK code with: 1) our fast CR2 correction implemented and
  ## 2) calls to the collapse package replaced with calls to rowsum()
  piv <- seq_len(tm$rank)
  Q <- qr.Q(tm$qr)[, piv, drop=FALSE]
  # add row names so we can put the AQ matrix back in its original alignment
  row.names(Q) <- seq_len(nrow(Q))
  R <- qr.R(tm$qr)[piv, piv, drop=FALSE]
  if (vcov.type %in% paste0(c("MB", "CR", "HC"), 2)) {
    if (length(cls) == length(unique(cls))) {
      AQ <- Q / sqrt(1-stats::hatvalues(tm)[setdiff(seq_along(cluster_ids),
                                                    na.action(tm))])
    } else {
      AQ <- do.call(
        rbind,
        lapply(
          levels(cls),
          function(s, tm, cls, cluster_ids, trts = NULL) {
            Qs <- Q[cls == s,, drop=FALSE]
            AQs <- cluster_iss(tm,
                               cluster_unit = s,
                               cluster_ids = cluster_ids,
                               assigned_trt = trts) %*% Qs
            row.names(AQs) <- row.names(Qs)
            return(AQs)
          },
          tm, cls = cls, cluster_ids = cluster_ids, trts
        )
      )
      AQ <- AQ[as.character(seq_len(nrow(Q))),, drop = FALSE]
    }
  } else {
    AQ <- Q
  }
  
  # if there are insufficient degrees of freedom, AQ will be numerically 0.
  # this is also the case in the dfadjust package, but that package does not
  # proactively address numerical zeros, instead allowing the division to wash
  # out and return 1 DOF. we address that here by returning NaN to remain
  # consistent with other cases of insufficient dof
  if (isTRUE(
    all.equal(AQ, matrix(0, nrow = nrow(AQ), ncol = ncol(AQ)), tolerance = tol)
  )) {
    return(NaN)
  }
  a <- drop(AQ %*% backsolve(R, ell, transpose = TRUE))
  as <- rowsum(a^2, cls)[,1]
  B <- rowsum(a * Q, cls)
  D <- rowsum(a, cls)[,1]
  Fm <- rowsum(Q, cls)
  GG <- (diag(as, nrow = length(as)) - tcrossprod(B)) * sig2 + rho * 
    tcrossprod(diag(D, length(D)) - tcrossprod(B, Fm))
  sum(diag(GG))^2/sum(GG^2)
}

#' @title (Internal) Compute \eqn{(I_{i} - H_{ii})^{-1/2}} as part of CR2
#'  variance estimates
#' @details The notation \eqn{I_{i}} and \eqn{H_{ii}} comes from Bell and
#'  McCaffrey (2002). The matrix \eqn{I_{i}} is an identity matrix with number
#'  of rows equal to the number of observations in cluster \eqn{i}, and
#'  \eqn{H_{ii}} subsets the hat matrix associated with the regression fit
#'  stored in \code{tm} to the rows associated with observations in cluster
#'  \eqn{i}.
#'  
#'  When possible, the function uses the method in Wasserman (2026)
#'  to cheaply compute the inverse symmetric square root.
#'  
#'  \code{cluster_ids} should be ordered in alignment with the dataframe passed
#'  to \code{lmitt()}. It should not exclude NA's because the function will
#'  exclude them.
#' @param tm \code{teeMod} object.
#' @param cluster_unit cluster to subset observations to. Must be found in
#'  \code{cluster_ids}.
#' @param cluster_ids optional, ID's for clustering standard errors. If not
#'  provided, default is the ID's associated with \code{cluster} if provided,
#'  otherwise the unit of assignment ID's.
#' @param ... Additional arguments passed from calls higher in the stack. One
#'  that may be used is \code{cluster}, which identifies the clustering variable
#'  if \code{cluster_ids} is not provided. If \code{cluster} is not provided,
#'  the default is the unit of assignment variable specified in the
#'  \code{StudySpecification}.
#' @importFrom stats model.frame na.action weights
#' @keywords internal
#' @references Robert M. Bell and Daniel F. McCaffrey. Bias Reduction in
#'  Standard Errors for Linear Regression with Multi-Stage Samples.
#'  *Survey Methodology*, 28(2):169-181, December 2002.
#'  
#'  
#'  Joshua Wasserman. Methods for Causal Inference in Settings with Clustered
#'  Data Subject to Missingness and Measurement Error. Unpublished thesis,
#'  June 2026. 
cluster_iss <- function(tm,
                        cluster_unit,
                        cluster_ids = NULL,
                        ...) {
  if (!inherits(tm, "teeMod")) stop("Must provide a teeMod object")
  tol <- 1e-11
  dots <- list(...)
  if (is.null(cluster_ids)) {
    if (is.null(dots$cluster)) cluster <- var_names(tm@StudySpecification, "u")
    if (is.null(dots$vcov.type)) stop(
      "Pass a `cluster_ids` or `vcov.type` argument to cluster_iss()")
    cluster_ids <- .make_uoa_ids(tm, substr(dots$vcov.type, 1, 2), cluster)
    # put the ID's in the order of the fitting dataframe if they've been
    # re-ordered due to the presence of a covariance adjustment model
    if (inherits(tm$model$`(offset)`, "SandwichLayer")) {
      ord <- .order_samples(tm)
      cluster_ids <- cluster_ids[seq_along(c(ord$Q_not_C, ord$Q_in_C))][
        match(seq_along(c(ord$Q_not_C, ord$Q_in_C)),
              as.numeric(c(names(ord$Q_not_C), names(ord$Q_in_C))))
      ]
    }
  }
  ix <- setdiff(which(cluster_ids == cluster_unit), na.action(tm))
  
  lmitt_data <- get("data", environment(formula(tm)))
  if (!is.null(sbst <- tm@lmitt_call$subset)) {
    lmitt_data <- subset(lmitt_data, eval(sbst, envir = lmitt_data))
  }
  
  # the treatment assignments, design matrix, and weights will all start with
  # all rows in the fitting dataframe (NA's included). we will then index each
  # (by creating trtg, Ag, and wg) to only the rows in the cluster and that
  # are not in the na.action
  if (is.null(trts <- dots$assigned_trt)) {
    wc <- tm@lmitt_call$weights
    if (inherits(wc, "call")) {
      if (exists(deparse1(wc[[1L]]))) {
        if (length(wc) == 1) {
          wc$specification <- get("specification", environment(formula(tm)))
        } else if (!inherits(eval(wc[[2L]], environment(formula(tm))),
                             "StudySpecification")) {
          wc$specification <- get("specification", environment(formula(tm)))
        }
        if (is.null(wc$data)) wc$data <- lmitt_data
      }
    }
    trts <- if (is.null(wc <- eval(wc, envir = environment(formula(tm)))) |
                !inherits(wc, "WeightedStudySpecification")) {
      treatment(tm@StudySpecification, newdata = lmitt_data)[,1]
    } else {
      if (length(dich <- wc@dichotomy) == 0) {
        treatment(tm@StudySpecification, newdata = lmitt_data)[,1]
      } else {
        treatment(tm@StudySpecification,
                  newdata = lmitt_data,
                  dichotomy = dich)[,1]
      }
    }
  }

  trtg <- trts[ix]
  tt <- stats::delete.response(stats::terms(tm))
  # lmitt_data has already had the subset applied, so we don't pass it a second
  # time to the model.frame call
  mf <- call("model.frame",
             tt,
             lmitt_data,
             na.action = na.pass,
             xlev = tm$xlevels)
  mf <- eval(mf)
  K <- tm$qr$rank
  piv <- tm$qr$pivot[seq_len(K)]
  A <- stats::model.matrix(tt,
                           mf,
                           contrasts.arg = tm$contrasts)[,piv,drop=FALSE]
  Ag <- A[ix,,drop=FALSE]
  inv <- chol2inv(tm$qr$qr[piv,piv])
  if (is.null(wts <- weights(tm, type = "working"))) wts <- rep(1, nrow(A))
  wg <- wts[ix]

  cg <- NULL
  Mgg <- NULL
  if (length(unique(trtg)) == 1) {
    # first, check for a moderator variable
    if (length(tm@moderator) > 0) {
      xvar <- lmitt_data[[tm@moderator]]
      # make sure the moderator variable is invariant within the cluster
      if (length(x <- unique(xvar[ix])) == 1) {
        if (tm@absorbed_intercepts) {
          # with an invariant moderator variable, the 1st row of the cluster
          # model matrix is the same as the rest
          cg <- sum(wg) * drop(crossprod(Ag[1,], inv) %*% Ag[1,])
          Mgg <- tcrossprod(sqrt(wg)) / sum(wg)
        } else {
          if (inherits(xvar, c("factor", "character"))) {
            # control clusters and treated clusters have different values of cg
            if (unique(trtg) == 0) {
              xcols <- which(colnames(Ag) %in% paste0(tm@moderator, x))
            } else {
              xcols <- which(colnames(Ag) %in% paste0(
                c(paste0(var_names(tm@StudySpecification, "t"), "._"), ""),
                tm@moderator, x)
              )
            }
            cg <- sum(inv[c(1, xcols), c(1, xcols)]) * sum(wg)
            Mgg <- tcrossprod(sqrt(wg)) / sum(wg)
          } else {
            if (unique(trtg) == 0) {
              cg <- (inv[1,1] + Ag[1,3] * inv[1,3] * 2 + Ag[1,3]^2 * inv[3,3]) *
                sum(wg)
              Mgg <- tcrossprod(sqrt(wg)) / sum(wg) 
            } else {
              ul <- sum(inv[1:2, 1:2])
              br <- sum(inv[3:4, 3:4])
              cg <- (ul + br * Ag[1,3]^2 + (sum(inv) - ul - br) * Ag[1,3]) *
                sum(wg)
              Mgg <- tcrossprod(sqrt(wg)) / sum(wg) 
            }
          }
        }
      }
    } else {
      if (tm@absorbed_intercepts) {
        # the 1st row of the cluster model matrix is the same as the rest
        cg <- sum(wg) * (1 / sum(wts) + Ag[1,2]^2 / sum(wts * A[,2]^2))
        Mgg <- tcrossprod(sqrt(wg)) / sum(wg)
      } else {
        # control clusters and treated clusters have different values of cg
        if (unique(trtg) == 0) {
          cg <- sum(wg) / sum(wts[setdiff(which(A[,2] == 0), na.action(tm))])
          Mgg <- tcrossprod(sqrt(wg)) / sum(wg)
        } else {
          cg <- sum(wg) / sum(wts[setdiff(which(A[,2] == 1), na.action(tm))])
          Mgg <- tcrossprod(sqrt(wg)) / sum(wg)
        }
      }
    }
  } else {
    if (length(tm@moderator) == 0) {
      if (tm@absorbed_intercepts) {
        # if fixed effects have been absorbed and clusters have treated and
        # control units, a sufficient condition for a valid cg and Mgg is:
        # sum(wg)     sum(wg * Ag[,2]^2)
        # -------  =  ------------------
        # sum(wts)    sum(wts * A[,2]^2)
        if (abs(sum(wg * Ag[,2]^2) / sum(wts * A[,2]^2) -
                sum(wg) / sum(wts)) < tol) {
          Mgg <- tcrossprod(sqrt(wg)) / sum(wg) +
            tcrossprod(sqrt(wg) * Ag[,2]) / sum(wg * Ag[,2]^2)
          # using either side of the above equality will work for cg
          cg <- sum(wg) / sum(wts)
        }
      } else {
        # when clusters contain both treated and control units, a sufficient
        # condition for simple construction of cg and Mgg is a constant ratio of
        # weighted size of the treatment group to the weighted size of the
        # control group across blocks
        rts <- tapply(data.frame(w = wts, t = trts),
                      cluster_ids,
                      function(df) {
                        wga <- rowsum(df[,1], df[,2], na.rm = TRUE)
                        wga[2,]/wga[1,]
                      })
        if (all(rts == mean(rts))) {
          Mgg <- matrix(0, nrow = length(trtg), ncol = length(trtg))
          Mgg[trtg == 0,trtg == 0] <- tcrossprod(sqrt(wg[trtg == 0])) /
            sum(wg[trtg == 0])
          Mgg[trtg == 1,trtg == 1] <- tcrossprod(sqrt(wg[trtg == 1])) /
            sum(wg[trtg == 1])
          # under this condition, cg's below are the same whether we use treated
          # or control units
          cg <- sum(wg[trtg == 0]) / sum(wts[trts == 0])
        }
      }
    }
  }

  # cg == 1 indicates our speedup can't be accommodated, so fall back to
  # original CR2 computation. Additionally, we use all.equal because,
  # numerically, cg may just be close to 1, and we call all.equal within isTRUE
  # per the all.equal documentation.
  if ((!is.null(cg) && isTRUE(all.equal(cg, 1))) | is.null(cg) | is.null(Mgg)) {
    Pgg <- Ag %*% inv %*% t(Ag * wg)
    symm <- isSymmetric(diag(nrow = length(ix)) - Pgg)
    eg <- eigen(diag(nrow = length(ix)) - Pgg, symmetric = symm)
    if (symm) {
      LamVT <- (1/sqrt(replace(eg$val, eg$val < tol, 0))) * t(eg$vec)
    } else {
      if (inherits(Vinv <- try(solve(eg$vec), silent = TRUE), "try-error")) {
        # fall back to svd, as the sandwich package does
        Vinv <- svd(eg$vec)
        Vinv <- Vinv$v[,Vinv$d > tol,drop=FALSE] %*% (
          (1/Vinv$d[Vinv$d > tol]) * t(Vinv$u[,Vinv$d > tol,drop=FALSE])
        )
      }
      LamVT <- (1/sqrt(replace(eg$val, eg$val < tol, 0))) * Vinv
    }
    return(eg$vec %*% LamVT)
  } else {
    return(diag(nrow = length(ix)) + (-1 + sqrt(1 + cg / (1-cg))) * Mgg)
  }
}

#' @title (Internal) Replace standard errors for moderator effect estimates with
#'   insufficient degrees of freedom with \code{NA}
#' @param vmat output of \code{.vcov_XXX()} called with input to \code{model}
#'   argument below as the first argument
#' @param model a fitted \code{teeMod} model
#' @param cluster_cols a character vector indicating the column(s) defining
#'   cluster ID's
#' @param values_from optional, a fitted model object. Defaults to \code{x}.
#' @param model_data dataframe or name of dataframe used to fit \code{model}
#' @param envir environment to get \code{model_data} from if \code{model_data}
#'   has class \code{name}
#' @return A variance-covariance matrix with NA's for estimates lacking
#'   sufficient degrees of freedom
#' @keywords internal
.check_df_moderator_estimates <- function(vmat,
                                          model,
                                          cluster_cols,
                                          values_from = model,
                                          model_data = quote(data),
                                          envir = environment(formula(model))) {
  if (!inherits(model, "teeMod")) {
    stop("`model` must be a teeMod object")
  }

  # if no moderator variable, no need to check df
  if (length(model@moderator) == 0) return(vmat)

  # if values_from didn't include the moderator, don't check df
  rhs_values_from <- trimws(
    strsplit(deparse1(stats::formula(values_from)[[3L]]), "\\+")[[1L]]
  )
  if (!any(grepl(model@moderator, rhs_values_from, fixed = TRUE))) return(vmat)

  # get model_data as a dataframe
  if (inherits(model_data, "name")) {
    model_data <- get(as.character(model_data), envir)
  } else if (!inherits(model_data, "data.frame")) {
    stop("`model_data` must be a dataframe or a quoted object name")
  }
  if (!inherits(model_data, "data.frame")) {
    stop("Could not find argument passed to `model_data` in the given `envir`")
  }
  
  # set na.action to exclude to use napredict below to expand the weights vector
  if (!is.null(model$na.action)) class(model$na.action) <- "exclude"

  # For categorical moderators, count the clusters contributing to estimation
  # for each level of the moderator variable; for continuous moderators, just
  # count the number of clusters. The moderator variable (or any level of the
  # moderator variable) must have at least three clusters contributing to
  # estimation for valid SE estimation.
  cluster <- .sanitize_Q_ids(model, cluster_cols)$cluster
  mod_form <- as.formula(paste0("~-1+", model@moderator))
  mf <- stats::expand.model.frame(model, mod_form, na.expand = FALSE)
  mod_vars <- stats::model.matrix(mod_form, mf)
  if (is.null(wts <- model$weights)) wts <- rep(1, nrow(model$model))
  in_model_fit <- ifelse(stats::napredict(model$na.action, wts) > 0,
                         1,
                         0)
  in_model_fit[model$na.action] <- 0
  if (ncol(mod_vars) > 1) {
    mod_counts <- rowsum(mod_vars * in_model_fit, cluster, na.rm = TRUE)
    valid_mods <- colSums(mod_counts != 0) > 2
  } else {
    # for continuous moderators, there are 4 coefficients to estimate. need more
    # than that for a variance estimate
    valid_mods <- stats::setNames(length(unique(cluster[in_model_fit == 1])) > 4,
                                  model@moderator)
  }


  # Replace SE's that don't meet df requirements with NA's
  if (any(!valid_mods)) {
    invalid_mods <- gsub("\\)", "\\\\)", gsub("\\(", "\\\\(", names(valid_mods)[!valid_mods]))
    warning(paste("The following subgroups do not have sufficient degrees of",
                  "freedom for standard error estimates and will be returned",
                  "as NA:",
                  paste(names(valid_mods)[!valid_mods], collapse = ", ")),
            call. = FALSE)
    # find cells of the covariance matrix that correspond to txt/mod group
    # interactions
    if (!(valid_mods[1])) invalid_mods <- c("\\(Intercept\\)", invalid_mods)
    dims_to_na <- which(grepl(paste0("(", paste(invalid_mods, collapse = "|"), ")"),
                              colnames(vmat),
                              perl = TRUE))

    vmat[dims_to_na, ] <- NA_real_
    vmat[, dims_to_na] <- NA_real_
  }

  return(vmat)
}

#' @details \code{.get_b12()}: \eqn{B_{12}} is the covariance of the unit of
#'   assignment- or cluster-level estimating equations for the covariance
#'   adjustment and direct adjustment models. It has a row for each term in the
#'   former and a column for each term in the latter. Any unit of assignment or
#'   cluster that does not appear in both model fits makes no contribution to
#'   this matrix. If there is no overlap between the two datasets, this function
#'   will return a matrix of zeros of appropriate dimension.
#' @param x a fitted \code{teeMod} model
#' @param ... arguments to pass to internal functions, such as \code{cluster}
#' @return \code{.get_b12()}: A \eqn{p\times 2} matrix
#' @keywords internal
#' @noRd
.get_b12 <- function(x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  if (x@.S3Class[1] != "lm") {
    stop("x must be an `lm` object")
  }

  sl <- x$model$`(offset)`
  if (!is(sl, "SandwichLayer")) {
    stop(paste("teeMod model must have an offset of class `SandwichLayer`",
               "for direct adjustment standard errors"))
  }
  # If cluster argument is NULL, use the `keys` dataframe created at initialization,
  # otherwise recreate given the desired clustering columns
  dots <- list(...)
  if (is.null(dots$cluster)) {
    cluster_cols <- var_names(x@StudySpecification, "u")
    keys <- sl@keys
  } else if (inherits(dots$cluster, "character")) {
    cluster_cols <- dots$cluster
    wide_frame <- tryCatch(
      stats::expand.model.frame(sl@fitted_covariance_model, cluster_cols,
                                na.expand = TRUE)[cluster_cols],
      error = function(e) {
        data <- eval(sl@fitted_covariance_model$call$data,
                     envir = environment(formula(sl@fitted_covariance_model)))
        stop(paste("The columns",
                   paste(setdiff(cluster_cols, colnames(data)), collapse = ", "),
                   "are missing from the covariance adjustment model dataset"),
             call. = FALSE)
      })

    # Check to see if provided column names overlap with specification
    # missing_spec_cols <- setdiff(cov_adj_cluster_cols, colnames(x@StudySpecification@structure))
    missing_spec_cols <- setdiff(cluster_cols, colnames(x@StudySpecification@structure))
    if (length(missing_spec_cols) > 0) {
      stop(paste("The following columns in the `cluster` argument cannot be found",
                 "in the teeMod object's StudySpecification:",
                 paste(missing_spec_cols, collapse = ", ")))
    }

    # Re-create keys dataframe with the new clustering columns
    keys <- as.data.frame(
      sapply(cluster_cols, function(col) {
        unique(x@StudySpecification@structure[[col]])[
          match(wide_frame[[col]], unique(x@StudySpecification@structure[[col]]), incomparables = NA)
        ]
      })
    )
  } else {
    stop(paste("If overriding `cluster` argument for meat matrix calculations,",
               "must provide a character vector specifying column names that",
               "exist in both the direct adjustment and covariance adjustment samples"))
  }

  Q_uoas <- stats::expand.model.frame(x, cluster_cols, na.expand = TRUE)[cluster_cols]
  Q_uoas <- apply(Q_uoas, 1, function(...) paste(..., collapse = "_"))
  C_uoas <- apply(keys[, cluster_cols, drop = FALSE], 1, function(...) paste(..., collapse = "_"))
  C_uoas_in_Q <- C_uoas %in% unique(Q_uoas)

  message(paste(sum(C_uoas_in_Q),
                "rows in the covariance adjustment data",
                "joined to the direct adjustment data\n"))

  # Check number of overlapping clusters n_QC; if n_QC <= 1, return 0 (but
  # similarly to .get_b11(), throw a warning when only one cluster overlaps)
  uoas_overlap <- length(unique(C_uoas[C_uoas_in_Q]))
  if (uoas_overlap <= 1) {
    if (uoas_overlap == 1) {
      warning(paste("Covariance matrix between covariance adjustment and direct adjustment",
                    "model estimating equations is numerically indistinguishable",
                    "from 0"))
    }
    return(
      matrix(0,
             nrow = dim(stats::model.matrix(sl@fitted_covariance_model))[2],
             ncol = dim(stats::model.matrix(x))[2])
    )
  }

  # Sum est eqns to cluster level; since non-overlapping rows are NA in `keys`,
  # `by` call excludes them from being summed
  cmod_estfun <- sandwich::estfun(sl@fitted_covariance_model)[C_uoas_in_Q,]
  cmod_aggfun <- ifelse(dim(cmod_estfun)[2] > 1, colSums, sum)
  cmod_eqns <- Reduce(rbind, by(cmod_estfun, C_uoas[C_uoas_in_Q], cmod_aggfun))

  # get rows from overlapping clusters in experimental data
  Q_uoas_in_C <- Q_uoas %in% unique(C_uoas[C_uoas %in% Q_uoas])
  damod_estfun <- sandwich::estfun(as(x, "lm"))[Q_uoas_in_C, , drop = FALSE]
  damod_aggfun <- ifelse(dim(damod_estfun)[2] > 1, colSums, sum)
  damod_eqns <- Reduce(rbind, by(damod_estfun, Q_uoas[Q_uoas_in_C], damod_aggfun))

  return(crossprod(cmod_eqns, damod_eqns))
}

#' @title (Internal) Estimate components of the sandwich covariance matrix
#'   returned by \code{vcov_tee()}
#' @details \code{.get_a22_inverse()}/\code{.get_tilde_a22_inverse()}:
#'   \eqn{A_{22}^{-1}} is the "bread" of the sandwich covariance matrix returned
#'   by \code{vcov_tee()} whether one has fit a prior covariance adjustment
#'   model or not.
#' @param x a fitted \code{teeMod} object
#' @param values_from optional, a fitted model object. Defaults to \code{x}.
#' @param ... arguments passed to \code{bread} method
#' @return \code{.get_a22_inverse()}/\code{.get_tilde_a22_inverse()}: A
#'   \eqn{2\times 2} matrix corresponding to an intercept and the treatment
#'   variable in the direct adjustment model
#' @keywords internal
#' @rdname sandwich_elements_calc
.get_a22_inverse <- function(x, values_from = x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }
  mc <- match.call()
  vcov_type <- match.arg(mc$vcov_type, c("MB", "HC", "CR", "DB")) # this sets the default to model-based
  
  if (x@.S3Class == "glm") {
    suppressWarnings(
      x <- update(values_from,
                  mustart = predict(values_from, type = "response"),
                  control = stats::glm.control(maxit = 1))
    )
  }
  teeMod_bread <- utils::getS3method("bread", x@.S3Class)(x)
  if (vcov_type == "DB") {
    return(teeMod_bread)
  }

  cm_mod <- x@ctrl_means_model
  ctrl_means_bread <- bread(cm_mod)
  if (!inherits(cm_mod, "mlm")) {
    dimnames(ctrl_means_bread) <- lapply(
      dimnames(ctrl_means_bread),
      function(nms, x) paste(formula(x)[[2L]], nms, sep = ":"),
      x = x)
  }

  out <- matrix(0, nrow = nrow(teeMod_bread) + nrow(ctrl_means_bread),
                ncol = ncol(teeMod_bread) + ncol(ctrl_means_bread),
                dimnames = list(c(rownames(teeMod_bread), rownames(ctrl_means_bread)),
                                  c(colnames(teeMod_bread), colnames(ctrl_means_bread))))
  out[1:nrow(teeMod_bread), 1:ncol(teeMod_bread)] <- teeMod_bread
  out[(nrow(teeMod_bread)+1):nrow(out), (ncol(teeMod_bread)+1):ncol(out)] <- ctrl_means_bread

  return(out)
}

#' @param x a fitted \code{teeMod} model
#' @param ... arguments to pass to internal functions
#' @details \code{.get_b22()}: \eqn{B_{22}} is the covariance of the unit of
#'   assignment- or cluster-level estimating equations for the direct adjustment
#'   model. In the absence of covariance adjustment, this is the meat of the
#'   sandwich covariance matrix for the direct adjustment model.
#' @return \code{.get_b22()}: A \eqn{2\times 2} matrix corresponding to an
#' intercept and the treatment variable in the direct adjustment model
#' @keywords internal
#' @noRd
.get_b22 <- function(x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  nq <- nrow(sandwich::estfun(x))

  # Create cluster ID matrix depending on cluster argument (or its absence)
  dots <- list(...)
  if (is.null(dots$cluster)) {
    uoas <- .expand.model.frame_teeMod(x,
                         var_names(x@StudySpecification, "u"))[, var_names(x@StudySpecification, "u"),
                                                   drop = FALSE]
  } else if (inherits(dots$cluster, "character")) {
    uoas <- tryCatch(
      .expand.model.frame_teeMod(x, dots$cluster)[,
                                                        dots$cluster,
                                                        drop = FALSE],
      error = function(e) {
        data <- eval(x$call$data,
                     envir = environment(formula(x)))
        stop(paste("The columns",
                   paste(setdiff(dots$cluster, colnames(data)), collapse = ", "),
                   "are missing from the direct adjustment data"),
             call. = FALSE)
      })
  } else {
    stop(paste("If overriding `cluster` argument for meat matrix calculations,",
               "must provide a character vector specifying column names in the",
               "direct adjustment data"))
  }

  if (ncol(uoas) == 1) {
    uoas <- factor(uoas[,1])
  } else {
    uoas <- factor(Reduce(function(...) paste(..., sep = "_"), uoas))
  }
  dots$cluster <- uoas
  dots$x <- as(x, "lm")

  out <- do.call(sandwich::meatCL, dots) * nq

  return(out)
}

#' @details \code{.get_a11_inverse()}: \eqn{A_{11}^{-1}} is the "bread" of the
#'   sandwich covariance matrix for the covariance adjustment model. This matrix
#'   contributes to the meat matrix of the direct adjustment sandwich covariance
#'   matrix.
#' @param x a fitted \code{teeMod} model
#' @return \code{.get_a11_inverse()}: A \eqn{p\times p} matrix where \eqn{p} is
#'   the dimension of the covariance adjustment model, including an intercept
#' @keywords internal
#' @rdname sandwich_elements_calc
.get_a11_inverse <- function(x) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  sl <- x$model$`(offset)`
  if (!inherits(sl, "SandwichLayer")) {
    stop(paste("teeMod model must have an offset of class `SandwichLayer`",
               "for direct adjustment standard errors"))
  }

  out <- sandwich::bread(sl@fitted_covariance_model)

  return(out)
}

#' @param x a fitted \code{teeMod} model
#' @param ... arguments to pass to internal functions
#' @details \code{.get_b11()}: \eqn{B_{11}} is the covariance of the unit of
#'   assignment- or cluster-level estimating equations for the covariance
#'   adjustment model. This matrix contributes to the meat matrix of the direct
#'   adjustment sandwich covariance matrix.
#' @return \code{.get_b11()}: A \eqn{p\times p} matrix where \eqn{p} is
#'   the dimension of the covariance adjustment model, including an intercept
#' @keywords internal
#' @noRd
.get_b11 <- function(x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  sl <- x$model$`(offset)`
  if (!inherits(sl, "SandwichLayer")) {
    stop(paste("teeMod model must have an offset of class `SandwichLayer`",
               "for direct adjustment standard errors"))
  }

  cmod <- sl@fitted_covariance_model
  nc <- sum(summary(cmod)$df[1L:2L])
  if (nc==0) nc <- length(cmod$fitted.values)
  if (nc==0) stop("can't determine extent of covariance model fitting data")

  # Get units of assignment for clustering
  dots <- list(...)
  if (is.null(dots$cluster)) {
    keys <- sl@keys
    cluster_cols <- setdiff(colnames(keys), "in_Q")
  } else if (inherits(dots$cluster, "character")) {
    cluster_cols <- dots$cluster
    keys <- tryCatch(
      stats::expand.model.frame(cmod, cluster_cols)[, cluster_cols, drop = FALSE],
      error = function(e) {
        data <- eval(cmod$call$data,
                     envir = environment(formula(cmod)))
        stop(paste("The columns",
                   paste(setdiff(cluster_cols, colnames(data)), collapse = ", "),
                   "are missing from the covariance adjustment model dataset"),
             call. = FALSE)
      })

    # check for NA's in the clustering columns
    nas <- rowSums(is.na(keys[, cluster_cols, drop = FALSE]))
    if (any(nas == length(cluster_cols))) {
      warning(paste("Some or all rows in the covariance adjustment model dataset",
                    "are found to have NA's for the given clustering columns.",
                    "This is taken to mean these observations should be treated",
                    "as IID. To avoid this warning, provide unique non-NA cluster",
                    "ID's for each row."))
    } else if (any(nas > 0 & nas < length(cluster_cols))) {
      warning(paste("Some rows in the covariance adjustment model dataset have",
                    "NA's for some but not all clustering columns. Rows sharing",
                    "the same non-NA cluster ID's will be clustered together.",
                    "If this is not intended, provide unique non-NA cluster ID's",
                    "for these rows."))
    }
  } else {
    stop(paste("If overriding `cluster` argument for meat matrix calculations,",
               "must provide a character vector specifying column names in the",
               "covariance adjustment model dataset"))
  }

  # Replace NA's for rows not in the experimental specification with a unique cluster ID
  uoas <- Reduce(function(...) paste(..., sep = "_"), keys[, cluster_cols, drop = FALSE])

  any_nas <- apply(keys[, cluster_cols, drop = FALSE], 1, function(x) any(is.na(x)))
  uoas[any_nas] <- NA_character_

  nuoas <- length(unique(uoas))
  nas <- is.na(uoas)
  if (any(nas)) {
    uoas[nas] <- paste0(nuoas - 1 + seq_len(sum(nas)), "*")
  }
  uoas <- factor(uoas)
  nuoas <- length(unique(uoas))

  # if the covariance model only uses one cluster, produce warning
  if (nuoas == 1) {
    warning(paste("Covariance adjustment model has meat matrix numerically",
                  "indistinguishable from 0"))
    return(
      matrix(0, nrow = ncol(stats::model.matrix(cmod)), ncol = ncol(stats::model.matrix(cmod)))
    )
  }
  dots$cluster <- uoas
  dots$x <- cmod
  out <- do.call(sandwich::meatCL, dots) * nc
  return(out)
}

#' @details \code{.get_a21()}/\code{.get_tilde_a21()}: \eqn{A_{21}} is the
#'   gradient of the estimating equations for the direct adjustment model taken
#'   with respect to the covariance adjustment model parameters. This matrix is
#'   the crossproduct of the prediction gradient for the units of observation in
#'   \eqn{\mathcal{Q}} and the model matrix of the direct adjustment model.
#' @param x a fitted \code{teeMod} model
#' @return \code{.get_a21()}/\code{.get_tilde_a21()}: A \eqn{2\times p} matrix
#'   where the number of rows are given by the intercept and the treatment
#'   variable in the direct adjustment model, and the number of columns are
#'   given by the dimension of the covariance adjustment model
#' @importFrom stats weights formula
#' @keywords internal
#' @rdname sandwich_elements_calc
.get_a21 <- function(x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  sl <- x$model$`(offset)`
  if (!inherits(sl, "SandwichLayer")) {
    stop(paste("teeMod model must have an offset of class `SandwichLayer`",
               "to propagate covariance adjustment model error"))
  }
  
  # Get contribution to the estimating equation from the direct adjustment model
  damod_mm <- stats::model.matrix(
    formula(x), stats::model.frame(x, na.action = na.pass))
  msk <- (apply(!is.na(sl@prediction_gradient), 1, all) &
            apply(!is.na(damod_mm), 1, all))
  if (!is.null(x$na.action)) class(x$na.action) <- "exclude"
  w <- if (is.null(w <- stats::weights(x))) numeric(length(msk)) + 1 else replace(w, is.na(w), 0)
  wZ <- damod_mm[, x$qr$pivot[1L:x$rank], drop = FALSE] * w
  
  mc <- match.call()
  vcov_type <- match.arg(mc$vcov_type, c("MB", "HC", "CR", "DB")) # this sets the default to model-based
  if (vcov_type != "DB") {
    # get gradient for ctrl means model; only keep full-rank columns
    cm <- x@ctrl_means_model
    K <- cm$rank
    piv <- cm$qr$pivot[seq_len(K)]
    cm_mm <- stats::model.matrix(formula(cm),
                                 stats::model.frame(cm, na.action = na.pass))
    cm_grad <- matrix(0, nrow = nrow(cm_mm), ncol = ncol(cm_mm),
                      dimnames = list(NULL, paste(formula(x)[[2]],
                                                  colnames(cm_mm),
                                                  sep = ":")))
    colnames(cm_mm) <- paste("cov_adj", colnames(cm_mm), sep = ":")
    if (inherits(cm, "mlm")) cm_grad <- cbind(cm_grad[,piv,drop=FALSE],
                                              cm_mm[,piv,drop=FALSE])
    cm_wts <- replace(cm_wts <- stats::weights(cm), is.na(cm_wts), 0)
    wZ <- cbind(wZ, -cm_grad * cm_wts)
  }

  out <- crossprod(wZ[msk,], sl@prediction_gradient[msk, , drop = FALSE])
  # scale by nq and keep it consistent with other nq calculations with na.action = na.pass
  nq <- nrow(stats::model.frame(x, na.action = na.pass))

  return(out / nq)
}

##' @keywords internal
##' @rdname sandwich_elements_calc
.get_tilde_a22_inverse <- function(x, values_from = x, ...) {
  out <- .get_a22_inverse(x, values_from, ...)

  if (!inherits(ca <- x$model$`(offset)`, "SandwichLayer")) {
    return(out)
  }

  nq <- nrow(stats::model.frame(x, na.action = na.pass))
  nc_not_q <- sum(!ca@keys$in_Q)
  n <- nq + nc_not_q

  out <- out * n / nq

  return(out)
}

##' @keywords internal
##' @rdname sandwich_elements_calc
.get_tilde_a21 <- function(x) {
  out <- .get_a21(x)

  nq <- nrow(stats::model.frame(x, na.action = na.pass))
  sl <- x$model$`(offset)`
  nc_not_q <- sum(!sl@keys$in_Q)
  n <- nq + nc_not_q

  out <- nq / n * out
}

#' @title (Internal) Design-based variance estimates with HC0 adjustment
#' @param x a fitted \code{teeMod} model
#' @details Design-based variance estimation is implemented for
#' \code{teeMod} models satisfying requirements including:
#' - The model uses \code{rct_spec} as \code{StudySpecification}
#' - The model only estimates a main treatment effect
#'
#' @keywords internal
#' @rdname var_estimators
.vcov_DB0 <- function(x, ...) {
  if (!inherits(x, "teeMod")) {
    stop("x must be a teeMod model")
  }

  if (!x@lmitt_fitted){
    # x@lmitt_fitted is false if someone created x using as.lmitt
    stop("x must have been fitted using lmitt.formula")
  }

  if (inherits(os <- x$model$`(offset)`, "SandwichLayer"))
    if (!all(os@keys$in_Q)){
      stop(paste("Design-based standard errors are not supported for teeMod",
                 "models with external sample for covariance adjustment"))
      # Our theory does not support this calculation
    }

  if (x@StudySpecification@type != "RCT"){
    stop("Design-based standard errors implemented only for RCT specifications")
    # The theory for design-based SE has not been extended to other specifications
  }

  args <- list(...)
  if ("type" %in% names(args)) {
    stop(paste("Cannot override the `type` argument for meat",
               "matrix computations"))
  }
  
  args$x <- if (isTRUE(args$const_effect)) block_center_residuals(x) else x
  args$db <- TRUE
  
  n <- length(args$cluster)

  if (x@absorbed_intercepts) {
    a22inv <- sandwich::bread(x, ...)
    meat <- do.call(sandwich::meatCL, args)
    vmat <- (1 / n) * a22inv %*% meat %*% a22inv
  }
  else {
    if (length(x@moderator) > 0){
      stop(paste("Design-based standard errors are not supported for teeMod",
                 "models with moderators and without absorbed intercepts"))
      # The theory for such models has not been derived
    }
    
    # If model weights does not incorporate IPW, throw a warning
    if (!(inherits(x@lmitt_call$weights, "call") &
          sum(grepl("ate", x@lmitt_call$weights)) > 0)){
      warning(paste("When calculating design-based standard errors,",
                    "please ensure that inverse probability weights are applied.",
                    "This could be done by specifying weights = ate() in",
                    "lmitt() or lm()."))
    }

    if (!is.null(x$call$offset)){
      vmat <- .get_DB_covadj_se(x, ...)
    }
    else {
      vmat <- .get_DB_wo_covadj_se(x)
    }
    
    name <- paste0(var_names(x@StudySpecification, "t"), ".")
    colnames(vmat) <- name
    rownames(vmat) <- name
  }
  attr(vmat, "type") <- "DB0"
  return(vmat)
}

### (If/when DB variance estimators other than DB0 are
###  added, consider replacing this with a stop that
###  invites the user to specify DB0 instead. -BH)
#' @rdname var_estimators
#' @keywords internal
.vcov_DB <- .vcov_DB0

#' @title (Internal) Design-based variance for models with
#'   covariance adjustment
#' @param x a fitted \code{teeMod} model
#' @details Calculate design-based variance estimate for \code{teeMod}
#'   models with covariance adjustment and without absorbed effects
#' @return design-based variance estimate of the main treatment effect
#'   estimate
#' @importFrom stats formula
#' @keywords internal
.get_DB_covadj_se <- function(x, ...){
  if (x@absorbed_intercepts) {
    stop(paste("x must be a tee model with covariance adjustment and\n",
               "without absorbed intercepts"))
    # For tee models with both covariance adjustment and absorbed intercepts,
    # use vcov_tee or .vcov_DB0 for variance estimation
  }
  specification_obj <- x@StudySpecification
  name_trt <- var_names(specification_obj, "t")
  if (name_trt %in% all.vars(stats::formula(x$model$`(offset)`@fitted_covariance_model))) {
    stop(paste("Design-based standard errors are not supported for\n",
               "tee models with treatment in prior covariance adjustment"))
    # Our theory does not support this calculation
  }

  bread <- .get_DB_covadj_bread(x, ...)

  signs1 <- ifelse(t(t(bread$b1[2,])) %*% bread$b1[2,] > 0, 1, 0)
  signs2 <- ifelse(t(t(bread$b1[2,])) %*% bread$b2[2,] > 0, 1, 0)
  signs3 <- ifelse(t(t(bread$b2[2,])) %*% bread$b2[2,] > 0, 1, 0)

  meat <- .get_DB_covadj_meat(x)

  term1 <- bread$b1 %*% (meat$m1u * signs1 + meat$m1l * (1-signs1)) %*% t(bread$b1)
  term2 <- bread$b2 %*% t(meat$m2u * signs2 + meat$m2l * (1-signs2)) %*% t(bread$b1)
  term3 <- bread$b2 %*% (meat$m3u * signs3 + meat$m3l * (1-signs3)) %*% t(bread$b2)

  vmat <- term1 + 2*term2 + term3
  return(as.matrix(vmat[2,2]))
}

#' @title (Internal) Bread matrix of design-based variance
#' @param x a fitted \code{teeMod} model
#' @details Calculate bread matrix for design-based variance estimate for
#'  \code{teeMod} models with covariance adjustment and without absorbed effects
#' @return a list of bread matrices
#' @keywords internal
.get_DB_covadj_bread <- function(x, ...) {
  a11inv <- .get_a11_inverse(x)
  a21 <- .get_a21(x, vcov_type="DB")
  a22inv <- .get_a22_inverse(x, vcov_type="DB")
  
  # C adjusts for linear transformation of the stacked estimating equation
  # each entry of the estimating equation are transformed to the form: 
  # a_fixed_quantity (under the design-based perspective) * a_treatment_indicator
  C <- matrix(c(1,1,0,1), nrow = 2, byrow = TRUE)

  n <- nrow(x$model)
  bread1 <- a22inv %*% a21 %*% a11inv / n
  bread2 <- - a22inv %*% C / n

  return(list(b1 = bread1, b2 = bread2))
}

#' @title (Internal) Meat matrix of design-based variance
#' @param x a fitted \code{teeMod} model
#' @details Calculate upper and lower bound estimates of meat matrix for
#'   design-based variance estimate for \code{teeMod} models with
#'   covariance adjustment and without absorbed effects
#' @return a list of meat matrix bounds
#' @importFrom stats model.frame
#' @keywords internal
.get_DB_covadj_meat <- function(x, ...) {
  agg <- .aggregate_to_cluster(x)
  data <- agg$data
  bid <- data[, agg$block] # block ids
  zobs <- data[, agg$z] # treatment assignments

  # number of covariates in the covariance model
  p <- ncol(stats::model.matrix(x$model$`(offset)`@fitted_covariance_model))
  XX <- .prepare_spec_matrix(x)
  
  # estimated covariances of entries of \Phi involving indicators z=0 or z=1
  V00 <- .cov_mat_est(XX[zobs==0,], bid[zobs==0])
  V11 <- .cov_mat_est(XX[zobs==1,], bid[zobs==1])
  V01 <- .cov01_est(XX, zobs, bid)
  
  # indices of V
  idl <- (p+2):(2*p+1)
  # upper and lower bounds of the 1st meat matrix term
  meat1u <- V00[1:p, 1:p] + V01[1:p, 1:p] + t(V01[1:p, 1:p]) + V11[1:p, 1:p]
  meat1l <- V00[idl, 1:p] + V01[idl, 1:p] + t(V01[idl, 1:p]) + V11[idl, 1:p]
  
  # upper and lower bounds of the 2nd meat matrix term
  meat2u <- cbind(V00[1:p, p+1], V01[1:p, p+1]) + cbind(V01[p+1, 1:p], V11[1:p, p+1])
  meat2l <- cbind(V00[idl, p+1], V01[idl, p+1]) + cbind(V01[2*p+2, 1:p], V11[idl, p+1])

  # upper and lower bounds of the 3rd meat matrix term
  meat3u <- matrix(c(V00[p+1, p+1], V01[p+1, p+1], V01[p+1, p+1], V11[p+1, p+1]), ncol = 2)
  meat3l <- matrix(c(V00[2*p+2, p+1], V01[2*p+2, p+1], V01[2*p+2, p+1], V11[2*p+2, p+1]), ncol = 2)

  return(list(m1u = meat1u, m1l = meat1l,
              m2u = meat2u, m2l = meat2l,
              m3u = meat3u, m3l = meat3l))
}

#' @title (Internal) Design-based variance for models without
#'   covariance adjustment
#' @param x a fitted \code{teeMod} model
#' @details Calculate bread matrix for design-based variance estimate for
#'   \code{teeMod} models without covariance adjustment and without absorbed
#'   effects
#' @return design-based variance estimate of the main treatment effect
#'   estimate
#' @importFrom stats aggregate var
#' @keywords internal
.get_DB_wo_covadj_se <- function(x, ...){
  if (x@absorbed_intercepts) {
    stop("x should not have absorbed intercepts")
  }
  if (!is.null(x$call$offset)){
    stop("x should not have covariance adjustment")
  }
  # aggregate block ids and outcomes to the cluster level
  agg <- .aggregate_to_cluster(x)
  data <- agg$data
  block <- agg$block

  ws <- data$.w
  yobs <- data$.wy # observed outcomes by cluster
  zobs <- data[, agg$z] # observed treatments by cluster
  bid <- data[, block] # block ids of clusters

  if (length(unique(bid)) == 1) {
    nbk <- t(specification_table(specification=x@StudySpecification, x="treatment"))
  } else {
    nbk <- specification_table(specification=x@StudySpecification, x="treatment",y="block")
  }
  # number of units by block and treatment, B by K
  small_blocks <- apply(nbk, 1, min) == 1 # small block indicator
  nb <- rowSums(nbk) # number of clusters in each block

  rho <- c(sum(ws * yobs * (1-zobs)) / sum(ws * (1-zobs)),
           sum(ws * yobs * zobs) / sum(ws * zobs))

  gammas <- (nbk[bid,1]*(1-zobs) + nbk[bid,2]*zobs) * ws # pseudo outcome gamma
  gamsbk <- list()  # s^2_b,j, sample variance of gamma by block and treatment
  for (k in 1:2){
    indk <- zobs == (k-1)
    gammas[indk] <- gammas[indk] * (yobs[indk] - rho[k])
    gamsbk[[k]] <- stats::aggregate(gammas[indk], by = list(data[indk,block]), FUN = stats::var)
  }
  gamsbk <- merge(gamsbk[[1]], gamsbk[[2]], by = "Group.1")[,2:3]
  gamsbk[is.na(gamsbk)] <- 0
  gamsb <- stats::aggregate(gammas, by = list(bid), FUN = var)[,2] # B vector

  nu1 <- rowSums(gamsbk / nbk) # large block variance estimates
  # small block variance estimates
  nu2 <- 2 / nbk[,1] / nbk[,2] * choose(nb,2) * gamsb -
    (1/nbk[,1] + 1/nbk[,2]) * ((nbk[,1]-1) *gamsbk[,1] + (nbk[,2]-1) *gamsbk[,2])
  varest <- (sum(nu1[!small_blocks]) + sum(nu2[small_blocks])) / sum(data$.w0)^2
  return(as.matrix(varest))
}

#' @title (Internal) Aggregate weights and outcomes to cluster level
#' @param x a fitted \code{teeMod} model
#' @details aggregate individual weights and outcomes to cluster weighted sums
#' @return a list of
#' - a data frame of cluster weights, outcomes, treatments, and block ids;
#' - treatment id column name;
#' - block id column name
#' @keywords internal
.aggregate_to_cluster <- function(x, ...){
  ws <- if (is.null(stats::weights(x))) 1 else stats::weights(x)
  data_temp <- x$call$data
  name_y <- as.character(x$terms[[2]]) # the column of y
  data_temp <- cbind(data_temp, .w = ws, 
                     .w0 = ws / ate(x@StudySpecification, data=data_temp),
                     .wy = ws * data_temp[, name_y])

  specification_obj <- x@StudySpecification
  name_trt <- var_names(specification_obj, "t")
  name_blk <- var_names(specification_obj, "b")
  name_clu <- var_names(specification_obj, "u")

  if (length(name_blk) > 0) {
    data_temp <- .merge_block_id_cols(data_temp, name_blk)
    name_blk <- name_blk[1] # merged block id is stored at column name_blk[1]
  }
  else {
    data_temp$bid <- 1
    name_blk <- 'bid'
  }
  eq_clu <- paste(name_clu, collapse = " + ")
  form <- paste("cbind(", name_trt, ", ", name_blk, ") ~ ", eq_clu, sep = "")
  form2 <- paste("cbind(.wy, .w, .w0) ~ ", eq_clu, sep = "")

  # aggregate z and bid by mean, aggregate w, w0, and w*y by sum
  data_aggr <- cbind(
    do.call("aggregate", list(as.formula(form), FUN = mean, data = data_temp)),
    do.call("aggregate", list(as.formula(form2), FUN = sum, data = data_temp))
  )
  data_aggr$.wy <- data_aggr$.wy / data_aggr$.w
  data_aggr <- data_aggr[order(data_aggr[, name_blk]), ]
  return(list(data = data_aggr, block = name_blk, z = name_trt))
}

#' @title (Internal) Merge multiple block IDs
#' @param df a data frame
#' @param ids a vector of block IDs, column names of df
#' @details merge multiple block ID columns by the value combinations and store
#'   the new block ID in the column \code{ids[1]}
#' @return a data frame with a column that contains unique block number IDs
#' @keywords internal
.merge_block_id_cols <- function(df, ids){
  if (!all(ids %in% colnames(df))){
    stop("Some block IDs are not present in the dataframe")
  }
  df[[ids[1]]] <- as.numeric(factor(do.call(paste, c(df[ids], sep = "."))))
  return(df)
}

#' @title (Internal) Helper function for design-based meat matrix
#'   calculation
#' @param x a fitted \code{teeMod} model
#' @return a \eqn{m \times (p+2)} matrix of cluster sums of design-based
#'   estimating equations scaled by \eqn{\sqrt{m_{b0}m_{b1}}/m_{b}}. Here
#'   \eqn{m} is the number of clusters, \eqn{p} is the number of covariates used
#'   in the prior covariance adjustment (excluding intercept)
#' @importFrom stats aggregate
#' @keywords internal
.prepare_spec_matrix <- function(x, ...) {
  specification_obj <- x@StudySpecification
  data <- x$call$data
  name_clu <- var_names(specification_obj, "u")
  # merged cluster id is stored at column name_clu[1]
  cid <- .merge_block_id_cols(data, name_clu)[, name_clu[1]]

  agg <- .aggregate_to_cluster(x)
  bid <- agg$data[, agg$block]  # block ids

  name_y <- as.character(x$terms[[2]]) # name of the outcome column
  # design matrix of the fitted covariance model
  X1 <- model.matrix(x$model$`(offset)`@fitted_covariance_model)
  p <- ncol(X1)

  wc <- x$model$`(offset)`@fitted_covariance_model$weight
  if (is.null(wc)) wc <- 1
  X1 <- matrix(
    rep(wc * (data[, name_y] - x$offset), p), ncol = p
    ) * X1 # n by p, wic * residual * xi

  wi <- if (is.null(stats::weights(x))) 1 else stats::weights(x)
  X2 <- wi * residuals(x)  # n by 1, wi[z] * (residual - rhoz) * z

  XX <- cbind(X1, X2)
  XX <- stats::aggregate(XX, by = list(cid), FUN = sum)[,-1]

  # multiple sqrt(product of #treated divided by block sizes) to XX by group
  if (length(unique(bid)) == 1) {
    nbk <- t(specification_table(specification=x@StudySpecification, x="treatment"))
  } else {
    nbk <- specification_table(specification=x@StudySpecification, x="treatment",y="block")
  }
  const <- sqrt(nbk[, 1] * nbk[, 2] / rowSums(nbk))
  XX <- sweep(XX, 1, const[bid], '*')
  return(XX)
}

#' @title (Internal) Helper function for design-based meat matrix
#'   calculation
#' @details Diagonal elements are estimated by sample variances Off-diagonal
#'   elements are estimated using the Young's elementary inequality
#' @return estimated upper and lower bounds of covariance matrix of estimating
#'   function vectors under either treatment or control
#' @importFrom stats cov
#' @keywords internal
.cov_mat_est <- function(XXz, bidz){
  cov0 <- tapply(1:nrow(XXz), bidz, function(s) stats::cov(XXz[s,]))
  covuu <- tapply(1:nrow(XXz), bidz, function(s) .add_vec(XXz[s,]))
  covll <- tapply(1:nrow(XXz), bidz, function(s) .add_vec(XXz[s,], upper=FALSE))

  cov0u <- lapply(1:length(cov0),
                  function(s) if (is.na(cov0[[s]][1,1])) covuu[[s]] else cov0[[s]])
  cov0l <- lapply(1:length(cov0),
                  function(s) if (is.na(cov0[[s]][1,1])) covll[[s]] else cov0[[s]])

  V00u <- Reduce('+', cov0u)
  V00l <- Reduce('+', cov0l)
  return(rbind(V00u, V00l))
}

#' @title (Internal) Helper function for design-based meat matrix calculation
#' @keywords internal
.add_mat_diag <- function(A, B){
  d <- nrow(A)
  A <- matrix(rep(diag(A), d), nrow = d, byrow = FALSE)
  B <- matrix(rep(diag(B), d), nrow = d, byrow = TRUE)
  #return(sqrt(abs(A * B)))
  return((A + B) / 2)
}

#' @title (Internal) Helper function for design-based meat matrix calculation
#' @keywords internal
.add_vec <- function(a, upper = TRUE){
  if (nrow(a) > 1) return(0)
  a <- as.numeric(a)
  d <- length(a)
  A <- matrix(rep(a, d), nrow = d, byrow = FALSE)
  B <- matrix(rep(a, d), nrow = d, byrow = TRUE)
  #if (upper) return(A*B + sqrt(abs(A*B)))
  #else return(A*B - sqrt(abs(A*B)))
  if (upper) return((A + B)^2 / 2)
  else return(- (A - B)^2 / 2)
}

#' @title (Internal) Helper function for design-based meat matrix
#'   calculation
#' @details the Young's elementary inequality is used
#' @return estimated upper and lower bounds of covariance matrix of estimating
#'   function vectors under treatment and under control
#' @importFrom stats cov
#' @keywords internal
.cov01_est <- function(XX, zobs, bid){
  cov0 <- tapply(
    1:nrow(XX[zobs==0,]), bid[zobs==0],
    function(s) stats::cov(XX[zobs==0,][s,])
    )
  cov1 <- tapply(
    1:nrow(XX[zobs==1,]), bid[zobs==1],
    function(s) stats::cov(XX[zobs==1,][s,])
    )
  cov01 <- lapply(1:length(cov0), function(s) .add_mat_diag(cov0[[s]], cov1[[s]]))

  cov01u <- lapply(1:length(cov0),
                   function(s) if (!is.na(cov01[[s]][1,1])) cov01[[s]]
                   else .add_mat_sqdif(XX, zobs, bid, s))
  cov01l <- lapply(1:length(cov0),
                   function(s) if (!is.na(cov01[[s]][1,1])) -cov01[[s]]
                   else -.add_mat_sqdif(XX, zobs, bid, s, upper=FALSE))
  V01u <- Reduce('+', cov01u)
  V01l <- Reduce('+', cov01l)
  return(rbind(V01u, V01l))
}

#' @title (Internal) Helper function for design-based meat matrix
#'   calculation
#' @keywords internal
.add_mat_sqdif <- function(X, zobs, bid, b, upper = TRUE){
  A <- X[zobs==0 & bid==b, ]
  if (upper) B <- X[zobs==1 & bid==b, ] else B <- -X[zobs==1 & bid==b, ]
  cov01u <- matrix(0, nrow=ncol(A), ncol=ncol(A))
  for (j in 1:ncol(B)){
    for (h in 1:nrow(B)){
      cov01u[,j] <- cov01u[,j] + colSums((A + B[h,j])^2)
    }
  }
  cov01u <- cov01u / nrow(A) / nrow(B)
  return(cov01u)
}

#' @title (Internal) Calculate grave\{phi\}
#' @keywords internal
#' @param x a fitted \code{teeMod} model
.get_phi_tilde <- function(x, ...){
  specification_obj <- x@StudySpecification
  df <- x$call$data

  # treatment assignments
  name_trt <- var_names(specification_obj, "t")
  assignment <- df[[name_trt]]
  z_ind <- sapply(c(0,1), function(val) as.integer(assignment == val))

  # block ids
  name_blk <- var_names(specification_obj, "b")
  df <- .merge_block_id_cols(df, name_blk)
  name_blk <- name_blk[1]
  stratum <- df[[name_blk]]
  s <- length(unique(stratum)) # number of blocks
  b_ind <- sapply(unique(stratum), function(val) as.integer(stratum == val))

  # nuisance parameters p
  n <- length(stratum) # number of units in Q
  ws <- if (is.null(stats::weights(x))) rep(1,n) else stats::weights(x)
  wb <- matrix(replicate(s, ws), ncol = s) * b_ind
  p <- t(z_ind) %*% wb / (matrix(1, nrow = 2, ncol = n) %*% wb) # 2 is the no. of treatments
  p1 <- p[2, ]

  # calculate phi tilde
  phitilde <- matrix(nrow = n, ncol = s)
  for (j in 1:s){
    phitilde[,j] <- ws * residuals(x) * (z_ind[,2] - p1[j]) * b_ind[,j]
  }
  return(phitilde)
}

#' @title (Internal) Product of \eqn{A_{pp}^{-1} A_{\tau p}^T}
#' @keywords internal
#' @param x a fitted \code{teeMod} model
#' @return An \eqn{s\times k} matrix \eqn{A_{pp}^{-1} A_{\tau p}^T}
.get_appinv_atp <- function(x, ...){
  specification_obj <- x@StudySpecification
  df <- x$call$data

  name_trt <- var_names(specification_obj, "t")
  assignment <- df[[name_trt]]
  k <- length(unique(assignment))

  # stratum ids
  name_blk <- var_names(specification_obj, "b")
  df <- .merge_block_id_cols(df, name_blk)
  name_blk <- name_blk[1]
  stratum <- df[[name_blk]]
  s <- length(unique(stratum)) # number of blocks
  b_ind <- sapply(unique(stratum), function(val) as.integer(stratum == val))
  
  n <- length(stratum) # number of units in the experiment
  ws <- if (is.null(stats::weights(x))) rep(1,n) else stats::weights(x)
  
  app <- list()
  atp <- list()
  for (i in 1:s){
    atp[[i]] <- ws * residuals(x) * b_ind[,i]
    app[[i]] <- ws * b_ind[,i]
  }
  
  mat <- matrix(0, nrow = (k-1)*s, ncol = k-1)
  for (i in 1:s){
    mat[i,1] <- sum(atp[[i]]) / sum(app[[i]])
  }
  return(mat)
}

test_that("vcov_tee errors when provided an invalid type", {
  data(simdata)
  cmod <- lm(y ~ z + x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  expect_error(vcov_tee(ssmod, "not_a_type"))
})

test_that("vcov_tee correctly sets and passes on args", {
  set.seed(993)
  data(simdata)
  cmod <- lm(y ~ z + x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  vmat1 <- suppressMessages(vcov_tee(ssmod, type = "CR0", cov_adj_rcorrect = "CR0"))
  expect_equal(vmat1,
               suppressMessages(.vcov_CR(ssmod, cluster = .make_uoa_ids(ssmod, "CR"),
                                         type = "CR0", cov_adj_rcorrect = "CR0")))

  vmat2 <- suppressMessages(vcov_tee(ssmod, type = "MB0", cov_adj_rcorrect = "MB0"))
  expect_true(all.equal(vmat2, vmat1, check.attributes = FALSE))
  expect_true(attr(vmat1, "type") != attr(vmat2, "type"))
  expect_true(attr(vmat1, "cov_adj_rcorrect") != attr(vmat2, "cov_adj_rcorrect"))

  vmat3 <- suppressMessages(vcov_tee(ssmod, type = "HC0", cov_adj_rcorrect = "HC0"))
  expect_true(all.equal(vmat3, vmat1, check.attributes = FALSE))
  expect_true(attr(vmat1, "type") != attr(vmat3, "type"))
  expect_true(attr(vmat1, "cov_adj_rcorrect") != attr(vmat3, "cov_adj_rcorrect"))
  
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  ssmod <- lmitt(y ~ 1, specification = spec, data = simdata, 
                 absorb = TRUE, offset = cov_adj(cmod))
  vmat4 <- suppressMessages(vcov_tee(ssmod, type = "DB0", const_effect = FALSE))
  expect_true(is.matrix(vmat4))
  expect_identical(attr(vmat4, "type"), "DB0")
  
  vmat5 <- suppressMessages(vcov_tee(ssmod, type = "DB0", const_effect = TRUE))
  expect_true(is.matrix(vmat5))
  expect_identical(attr(vmat5, "type"), "DB0")
  expect_false(identical(vmat4, vmat5))
  
  vmat6 <- vcov_tee(ssmod, type = "HC1", cov_adj_rcorrect = "HC1")
  expect_equal(attr(vmat6, "type"), "HC1")
  expect_equal(attr(vmat6, "cov_adj_rcorrect"), "HC1")
  
  vmat7 <- vcov_tee(ssmod)
  expect_equal(attr(vmat7, "type"), "HC0")
  expect_equal(attr(vmat7, "cov_adj_rcorrect"), "HC0")
  
  first_obs_ix <- c(1, 1 + cumsum(c(t(table(simdata$uoa1, simdata$uoa2)))))
  uniqdata <- simdata[first_obs_ix[1:(length(first_obs_ix)-1)],,drop=FALSE]
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  imod <- lmitt(y ~ 1, spec, uniqdata)
  vmat8 <- vcov_tee(imod)
  expect_equal(attr(vmat8, "type"), "HC0")
  expect_true(is.null(attr(vmat8, "cov_adj_rcorrect")))

  values_from <- lm(y ~ 1, simdata)
  imod <- lmitt(y ~ 1, z ~ cluster(uoa1, uoa2), simdata)
  ix <- seq_len(imod$rank)
  vmat9 <- vcov_tee(imod, values_from = values_from, cadjust = FALSE)[ix, ix]
  br <- bread(imod)[ix, ix]
  expect_true(all.equal(
    vmat9,
    br %*% crossprod(
      rowsum(model.matrix(imod) * residuals(values_from),
             paste(simdata$uoa1, simdata$uoa2, sep = "_"))
    ) %*% (t(br) / length(residuals(values_from))^2),
    check.attributes = FALSE
  ))
})

if (requireNamespace("robustbase", quietly = TRUE)) {
  test_that("vcov_tee sets correct bias correction for lmrob cov adj model", {
    set.seed(993)
    data(simdata)
    cmod <- robustbase::lmrob(y ~ z + x, simdata)
    spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
    ssmod <- lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))
    
    vmat5 <- vcov_tee(ssmod)
    expect_equal(attr(vmat5, "cov_adj_rcorrect"), "HC0")
  })
}

test_that(paste("vcov_tee produces correct calculations with valid `cluster` arugment",
                "when cluster ID's have no NA's"), {
  data(simdata)
  simdata_copy <- simdata
  simdata_copy$uid <- seq_len(nrow(simdata_copy))
  cmod <- lm(y ~ x, simdata_copy)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2, uid) + block(bid), simdata_copy)
  ssmod <- lmitt(y ~ 1, data = simdata_copy, specification = spec,
                weights = ate(spec), offset = cov_adj(cmod))

  # check default clustering level is the same when specified using cluster arg
  expect_equal(suppressMessages(vcov_tee(ssmod)),
               suppressMessages(vcov_tee(ssmod, cluster = c("uid"))))
})

test_that(paste("vcov_tee produces correct calculations with valid `cluster` arugment",
                "when cluster ID's have NA's (must be via column name)"), {
  data(simdata)
  df <- rbind(simdata, simdata)
  df[1:50, c("uoa1", "uoa2", "bid", "z")] <- NA
  cmod <- lm(y ~ x, df[1:50,])
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), df[51:100,])
  ssmod <- lmitt(y ~ 1, data = df[51:100,], specification = spec,
                weights = ate(spec), offset = cov_adj(cmod))

  expect_equal(vcov_tee(ssmod, cluster = c("uoa1", "uoa2")), vcov_tee(ssmod))
})

test_that("variance helper functions fail without a teeMod model", {
  data(simdata)
  cmod <- lm(y ~ z, data = simdata)

  expect_error(propertee:::.vcov_CR(cmod), "must be a teeMod")
  expect_error(propertee:::.get_b12(cmod), "must be a teeMod")
  expect_error(propertee:::.get_a22_inverse(cmod), "must be a teeMod")
  expect_error(propertee:::.get_b22(cmod), "must be a teeMod")
  expect_error(propertee:::.get_a22_inverse(cmod), "must be a teeMod")
  expect_error(propertee:::.get_a11_inverse(cmod), "must be a teeMod")
  expect_error(propertee:::.get_b11(cmod), "must be a teeMod")
  expect_error(propertee:::.get_a21(cmod), "must be a teeMod")

})

test_that(".get_dof", {
  data(simdata)
  expect_error(.get_dof(lm(y~x, simdata), 0, "CR2"), "teeMod object")
  
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  tm <- lmitt(y ~ 1, spec, simdata)
  
  # non-CR2 dof
  expect_equal(.get_dof(tm, 1, "CR0", "stata"), nrow(spec@structure)-1)
  expect_equal(.get_dof(tm, 1, "DB0", "stata", cluster = "bid"), 2)
  
  # CR2 dof
  expect_error(.get_dof(tm, seq_len(7), "CR2", "IK"), "same length")
  expect_error(.get_dof(tm, seq_len(2), "CR2", "IK"), "zeros or ones")
  expect_true(inherits(dof <- .get_dof(tm, 2, "CR2", "IK"), "numeric"))
  expect_equal(length(dof), 1)
})

test_that(".compute_IK_dof no clustering", {
  data(simdata)

  ## no clustering, continuous y, sufficient dof
  simdata$id <- seq_len(nrow(simdata))
  idspec <- rct_spec(z ~ unitid(id), simdata)
  tm <- lmitt(y ~ 1, idspec, simdata)
  ell <- c(0, 1)
  dof <- .compute_IK_dof(tm, ell, vcov.type = "CR2")
  
  Q <- qr.Q(tm$qr)
  R <- qr.R(tm$qr)
  I_P <- diag(nrow = nrow(simdata)) - tcrossprod(Q)
  G <- I_P * drop((Q / sqrt(1-stats::hatvalues(tm))) %*% t(solve(R)) %*% ell)
  GoG <- crossprod(G, diag(mean(stats::residuals(tm)^2), nrow = nrow(Q))) %*% G
  expected_dof <- sum(diag(GoG))^2 / sum(diag(GoG %*% GoG))
  expect_equal(dof, expected_dof)
  
  ## not enough dof
  set.seed(103)
  ad2 <- data.frame(cid = c(rep(1, 4), rep(2, 6), rep(3, 5), rep(4, 9),
                            rep(5, 12), rep(6, 7)),
                    x = factor(c(rep(0,15), rep(1, 9+12+7))),
                    a = c(rep(0, 4), rep(1, 11), rep(0, 9), rep(1, 19)),
                    y = rnorm(4 + 5 + 6 + 9+19))
  spec <- rct_spec(a ~ unitid(cid), ad2, subset = cid %in% c(1, 2, 4, 5))
  tm <- lmitt(y ~ x, spec, ad2, subset = cid %in% c(1, 2, 4, 5))

  bad.ell1 <- c(0, 0, 1, 0)
  bad.ell2 <- c(0, 0, 0, 1)
  suppressWarnings(dof_a.x0 <- .compute_IK_dof(tm, bad.ell1, "CR2"))
  # returning NaN does not cause warnings for pt() in summary.teeMod()
  expect_equal(dof_a.x0, NaN)
  # returning NaN does not cause warnings for pt() in summary.teeMod()
  suppressWarnings(dof_a.x1 <- .compute_IK_dof(tm, bad.ell2, "CR2"))
  expect_equal(dof_a.x1, NaN)
})

test_that(".compute_IK_dof w/ clustering", {
  set.seed(103)
  ad2 <- data.frame(cid = c(rep(1, 4), rep(2, 6), rep(3, 5), rep(4, 9),
                            rep(5, 12), rep(6, 7)),
                    a = c(rep(0, 4), rep(1, 11), rep(0, 9), rep(1, 19)),
                    y = rnorm(4 + 5 + 6 + 9+19))
  spec <- rct_spec(a ~ unitid(cid), ad2)
  tm <- lmitt(y ~ 1, spec, ad2)
  
  expect_true(inherits(.compute_IK_dof(tm, c(0, 1), "CR0", ad2$cid), "numeric"))
  expect_true(inherits(.compute_IK_dof(tm, c(0, 1), "CR2"), "numeric"))
  
  ad2$y[1] <- NA_real_
  tm <- lmitt(y ~ 1, spec, ad2)
  
  expect_true(inherits(CR0_dof <- .compute_IK_dof(tm, c(0, 1), "CR0", ad2$cid),
                       "numeric"))
  expect_equal(.compute_IK_dof(tm, c(0, 1), "CR1", ad2$cid), CR0_dof)
  expect_true(inherits(.compute_IK_dof(tm, c(0, 1), "CR2"), "numeric"))
})

test_that("cluster_iss, no absorption", {
  set.seed(33)
  ad2 <- data.frame(cid = c(rep(1, 4), rep(2, 6), rep(3, 5), rep(4, 9),
                            rep(5, 12), rep(seq(6, 8), each = 7)),
                    x = factor(c(rep(0,15), rep(1, 9+12+7+7), rep(0, 7))),
                    a = c(rep(0, 4), rep(1, 11), rep(0, 9), rep(1, 19),
                          rep(0, 7+7)),
                    y = rnorm(4 + 5 + 6 + 9+19+7+7))

  ## no moderator
  spec <- rct_spec(a ~ unitid(cid), ad2)
  tm <- lmitt(y ~ 1, spec, ad2)
  ATWA_inv <- chol2inv(tm$qr$qr)
  
  # control unit
  ix <- ad2$cid == 1
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*% t(model.matrix(tm)[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 1, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors)
  ))
  # treated unit
  ix <- ad2$cid == 6
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*% t(model.matrix(tm)[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 6, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors)
  ))
  # make sure we order cluster ID's correctly by default if there's covariance
  # adjustment (this shouldn't change the output of cluster_iss() because it's
  # solely based on the second-stage regression)
  cm <- lm(y ~ x, ad2)
  tm_ca <- lmitt(y ~ 1, spec, ad2, offset = cov_adj(cm))
  expect_true(all.equal(
    cluster_iss(tm_ca, cluster_unit = 6, vcov.type = "MB"),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors)
  ))
  
  ## cluster-invariant binary moderator variable
  tm <- lmitt(y ~ x, spec, ad2, weights = "ate")
  ATWA_inv <- chol2inv(tm$qr$qr)
  
  # treated unit w/ x = 1
  ix <- ad2$cid == 6
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    # test NULL cluster_ids and NULL cluster and NULL trts 
    cluster_iss(tm, cluster_unit = 6, vcov.type = "MB"),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  # treated unit w/ x = 0
  ix <- ad2$cid == 2
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 2, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  # control unit w/ x = 1
  ix <- ad2$cid == 4
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 4, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  # control unit w/ x = 0
  ix <- ad2$cid == 1
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    # test NULL cluster_ids and NULL cluster and NULL trts 
    cluster_iss(tm, cluster_unit = 1, vcov.type = "MB"),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  
  ## cluster-invariant continuous moderator variable
  ad2$x <- runif(length(unique(ad2$cid)))[ad2$cid]
  tm <- lmitt(y ~ x, spec, ad2, weights = "ate")
  ATWA_inv <- chol2inv(tm$qr$qr)
  
  # control unit 
  ix <- ad2$cid == 1
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 1, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  # treated unit
  ix <- ad2$cid == 6
  Pgg <- model.matrix(tm)[ix,] %*% ATWA_inv %*%
    t((model.matrix(tm) * weights(tm))[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 6, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
})

test_that("cluster_iss, absorption", {
  # don't need to distinguish between treated and control units in these tests
  set.seed(499)
  adb <- data.frame(cid = c(rep(1, 6), rep(2, 5), rep(3, 4), rep(4, 5),
                            rep(5, 6), rep(6, 5), rep(7, 4), rep(8, 5)),
                    b = rep(c(1, 2), each = 20),
                    a = c(rep(0, 11), rep(1, 20), rep(0, 9)),
                    y = rnorm(40))
  adb$x <- runif(8)[adb$cid]
  specification <- rct_spec(a ~ cluster(cid) + block(b), adb)
  
  ## no moderator
  tm <- lmitt(y ~ 1, specification, adb, weights = "ate", absorb = TRUE)
  piv <- seq_len(2)
  ATWA_inv <- chol2inv(tm$qr$qr[piv,piv,drop=FALSE])
  
  ix <- adb$cid == 1
  Ag <- model.matrix(tm)[ix,piv,drop=FALSE]
  Pgg <- Ag %*% ATWA_inv %*% t(Ag * weights(tm)[ix])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, 1, adb$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  
  ## cluster-invariant continuous moderator
  tm <- lmitt(y ~ x, specification, adb, weights = "ate", absorb = TRUE)
  piv <- seq_len(tm$rank)
  ATWA_inv <- chol2inv(tm$qr$qr[piv,piv,drop=FALSE])
  
  Ag <- model.matrix(tm)[ix,piv,drop=FALSE]
  Pgg <- Ag %*% ATWA_inv %*% t(Ag * weights(tm)[ix])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, 1, adb$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  
  ## cluster-invariant factor moderator
  adb$x <- factor(round(adb$x))
  tm <- lmitt(y ~ x, specification, adb, weights = "ate", absorb = TRUE)
  piv <- seq_len(tm$rank)
  ATWA_inv <- chol2inv(tm$qr$qr[piv,piv,drop=FALSE])
  
  Ag <- model.matrix(tm)[ix,piv,drop=FALSE]
  Pgg <- Ag %*% ATWA_inv %*% t(Ag * weights(tm)[ix])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, 1, adb$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
})

test_that("cluster_iss with fine strata (clusters have trt and ctrl units)", {
  set.seed(496)
  ## balanced in each stratum, absorption
  sdata <- data.frame(id = seq_len(20),
                      uid = rep(seq_len(4), each = 5),
                      ai = rep(rep(c(0, 1), each = 5), 2),
                      bi = rep(c(1, 2), each = 10),
                      yi = rnorm(20))
  ats <- rct_spec(ai ~ unitid(uid) + block(bi), sdata)
  tm <- lmitt(yi ~ 1, ats, data = sdata, absorb = TRUE)
  ids <- propertee:::.make_uoa_ids(tm, "MB")
  A <- model.matrix(tm)
  ATWA_inv <- chol2inv(tm$qr$qr)
  ix <- ids == unique(ids)[1]
  Ag <- A[ix,,drop=FALSE]
  Pgg <- Ag %*% ATWA_inv %*% t(Ag)
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[1], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))

  ## weighted balance in each stratum, absorption
  wts <- rep(c(rep(2, 4), 1), 4)
  tm <- lmitt(yi ~ 1, ats, data = sdata, weights = wts, absorb = TRUE)
  ATWA_inv <- chol2inv(tm$qr$qr)
  wg <- wts[ix]
  Pgg <- (Ag * sqrt(wg)) %*% ATWA_inv %*% t(Ag * sqrt(wg))
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[1], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  
  ## weighted balance, no absorption
  tm <- lmitt(yi ~ 1, ats, data = sdata, weights = "att")
  A <- model.matrix(tm)
  ATWA_inv <- chol2inv(tm$qr$qr)
  wg <- weights(tm)[ix]
  Ag <- A[ix,,drop=FALSE]
  Pgg <- (Ag * sqrt(wg)) %*% ATWA_inv %*% t(Ag * sqrt(wg))
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[1], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))

  ## same imbalanced treatment assignment in each stratum, absorption
  sdata_imbal <- sdata
  sdata_imbal <- rbind(sdata_imbal,
                       data.frame(id = seq(21,30),
                                  uid = rep(c(5, 6), each = 5),
                                  ai = 1,
                                  bi = rep(c(1, 2), each = 5),
                                  yi = rnorm(10)))
  imb <- rct_spec(ai ~ unitid(uid) + block(bi), sdata_imbal)
  tm <- lmitt(yi ~ 1, imb, data = sdata_imbal, absorb = TRUE)
  ids <- propertee:::.make_uoa_ids(tm, "MB")
  A <- model.matrix(tm)
  ATWA_inv <- chol2inv(tm$qr$qr)
  ix <- ids == unique(ids)[2]
  Ag <- A[ix,,drop=FALSE]
  Pgg <- Ag %*% ATWA_inv %*% t(Ag)
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[2], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  
  ## same imbalanced treatment assignment in each stratum, no absorption
  tm <- lmitt(yi ~ 1, imb, data = sdata_imbal, weights = "att")
  wts <- weights(tm)
  ids <- propertee:::.make_uoa_ids(tm, "MB")
  X <- model.matrix(tm)
  ATWA_inv <- chol2inv(tm$qr$qr)
  ix <- ids == unique(ids)[2]
  A <- model.matrix(tm)
  Ag <- A[ix,,drop=FALSE]
  Pgg <- (Ag * sqrt(wts[ix])) %*% ATWA_inv %*% t(Ag * sqrt(wts[ix]))
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[2], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
  expect_message(tm <- lmitt(yi ~ 1, imb, data = sdata_imbal), "absorb")
  wts <- rep(1, nrow(sdata_imbal))
  ids <- propertee:::.make_uoa_ids(tm, "MB")
  X <- model.matrix(tm)
  ATWA_inv <- chol2inv(tm$qr$qr)
  ix <- ids == unique(ids)[2]
  A <- model.matrix(tm)
  Ag <- A[ix,,drop=FALSE]
  Pgg <- (Ag * sqrt(wts[ix])) %*% ATWA_inv %*% t(Ag * sqrt(wts[ix]))
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[2], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))

  ## same weighted imbalance in each strata
  sdata$uid[sdata$uid == 2] <- c(1, rep(2, 4))
  sdata$uid[sdata$uid == 3] <- c(rep(3, 2), rep(4, 3))
  sdata$ai[sdata$uid == 1] <- 0
  sdata$ai[sdata$uid == 4] <- 1
  imb <- rct_spec(ai ~ unitid(uid) + block(bi), sdata)
  wts <- c(rep(2/3, 6), rep(1, 4), rep(4, 2), rep(1, 8))
  tm <- lmitt(yi ~ 1, imb, data = sdata, weights = wts, absorb = TRUE)
  ids <- propertee:::.make_uoa_ids(tm, "MB")
  ix <- ids == "2"
  A <- model.matrix(tm)
  Ag <- A[ix,,drop=FALSE]
  Pgg <- (Ag * sqrt(wts[ix])) %*% chol2inv(tm$qr$qr) %*% t(Ag * sqrt(wts[ix]))
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, unique(ids)[2], ids),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors),
    check.attributes = FALSE
  ))
})

test_that("cluster_iss with NA's and subset argument", {
  set.seed(33)
  ad2 <- data.frame(cid = c(rep(1, 4), rep(2, 6), rep(3, 5), rep(4, 9),
                            rep(5, 12), rep(seq(6, 8), each = 7)),
                    x = factor(c(rep(0,15), rep(1, 9+12+7+7), rep(0, 7))),
                    a = c(rep(0, 4), rep(1, 11), rep(0, 9), rep(1, 19),
                          rep(0, 7+7)),
                    y = rnorm(4 + 5 + 6 + 9+19+7+7))
  
  # subset
  spec <- rct_spec(a ~ unitid(cid), ad2)
  tm <- lmitt(y ~ 1, spec, ad2, subset = x == 0)
  
  ix <- ad2$cid == 1 & ad2$x == 0
  mm_full <- model.matrix(
    tm$terms,
    model.frame(tm$terms, get("data", attr(tm$terms, ".Environment")))
  )
  Pgg <- mm_full[ix,] %*% chol2inv(tm$qr$qr) %*% t(mm_full[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 1, cluster_ids = ad2$cid[ad2$x == 0]),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors)
  ))
  
  # NA's
  ad2$y[1] <- NA_real_
  tm <- lmitt(y ~ 1, spec, ad2, subset = x == 0)
  
  ix <- ad2$cid == 1 & !is.na(ad2$y)
  mm_full <- model.matrix(
    tm$terms,
    model.frame(tm$terms,
                get("data", attr(tm$terms, ".Environment")),
                na.action = na.pass)
  )
  Pgg <- mm_full[ix,] %*% chol2inv(tm$qr$qr) %*% t(mm_full[ix,])
  eg <- eigen(diag(nrow = sum(ix)) - Pgg)
  expect_true(all.equal(
    cluster_iss(tm, cluster_unit = 1, cluster_ids = ad2$cid),
    eg$vectors %*% diag(1/sqrt(eg$values), nrow = length(eg$values)) %*%
      solve(eg$vectors)
  ))
})

test_that("cluster_iss and .compute_IK_dof with dichotomies", {
  data(simdata)
  
  # first, dichotomy is specified only in lmitt()
  spec <- rct_spec(dose ~ unitid(uoa1, uoa2), simdata)
  tm <- lmitt(y ~ 1, spec, simdata, weights = "ate",
              dichotomy = dose > 100 ~ dose <= 100)
  expect_true(inherits(cluster_iss(tm, "1_1", .make_uoa_ids(tm, "MB")),
                       "matrix"))
  dof <- .compute_IK_dof(tm, c(0, 1), "CR2")
  expect_true(inherits(dof, "numeric") & dof < Inf)
  
  # second, dichotomy given in the weights as well
  wts <- ate(spec, dose > 100 ~ dose <= 100, data = simdata)
  tm <- lmitt(y ~ 1, spec, simdata, weights = wts,
              dichotomy = dose > 100 ~ dose <= 100)
  expect_true(inherits(cluster_iss(tm, "1_1", .make_uoa_ids(tm, "MB")),
                       "matrix"))
  dof <- .compute_IK_dof(tm, c(0, 1), "CR2")
  expect_true(inherits(dof, "numeric") & dof < Inf)
})

test_that(paste(".get_b12, .get_a11_inverse, .get_b11, .get_a21 used with teeMod model",
                "without SandwichLayer offset"), {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  offset <- stats::predict(cmod, simdata)
  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = offset)
  )

  expect_error(.get_b12(m), "must have an offset of class")
  expect_error(.get_a11_inverse(m), "must have an offset of class")
  expect_error(.get_b11(m), "must have an offset of class")
  expect_error(.get_a21(m), "must have an offset of class")
})

test_that(".get_b12 fails if ITT model isn't strictl an `lm`", {
  return(expect_true(TRUE)) # this function is deprecated and now failing
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  offset <- stats::predict(cmod, simdata)
  m <- as.lmitt(
    glm(y ~ assigned(), data = simdata, weights = ate(spec), offset = offset)
  )

  expect_error(.get_b12(m), "x must be an `lm` object")
})

test_that(paste(".get_b12 returns expected B_12 for individual-level",
                "experimental data identical to cov model data"), {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )

  uoanames <- var_names(m@StudySpecification, "u")
  zname <- var_names(m@StudySpecification, "t")

  # est eqns for lm are wts * resids * dmatrix
  cmod <- m$model$`(offset)`@fitted_covariance_model
  cmod_eqns <- Reduce(
    rbind,
    by(cmod$residuals * model.matrix(cmod),
       lapply(uoanames, function(col) m$model$`(offset)`@keys[,col]),
       colSums))

  Q <- stats::expand.model.frame(m, uoanames)
  msk <- !is.na(propertee:::.merge_preserve_order(
    Q,
    merge(unique(m$model$`(offset)`@keys), m@StudySpecification@structure),
    by = uoanames,
    all.x = TRUE,
    sort = FALSE)[zname])
  m_eqns <- Reduce(
    rbind,
    by((m$weights * m$residuals * model.matrix(m))[msk, , drop = FALSE],
       lapply(uoanames, function(col) Q[msk, col]),
       colSums))
  expect_message(b12 <- propertee:::.get_b12(m), paste(nrow(simdata), "rows"))
  expect_equal(b12,
               t(cmod_eqns) %*% m_eqns)
})

test_that(paste(".get_b12 returns expected B_12 for cluster-level",
                "experimental data whose rows fully overlap with cov model data"), {
  data(simdata)
  cluster_ids <- unique(simdata[, c("uoa1", "uoa2")])
  Q_cluster <- data.frame(Reduce(
    rbind,
    by(simdata,
       list(simdata$uoa1, simdata$uoa2),
       function(x) {colMeans(x[, c("uoa1", "uoa2", "x", "y", "z")])}),
  ), row.names = NULL)

  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_cluster)

  m <- as.lmitt(
    lm(y ~ assigned(), data = Q_cluster, weights = ate(spec), offset = cov_adj(cmod))
  )

  uoanames <- var_names(m@StudySpecification, "u")
  zname <- var_names(m@StudySpecification, "t")

  # est eqns for lm are wts * resids * dmatrix
  cmod <- m$model$`(offset)`@fitted_covariance_model
  cmod_eqns <- Reduce(
    rbind,
    by(cmod$residuals * model.matrix(cmod),
       lapply(uoanames, function(col) m$model$`(offset)`@keys[,col]),
       colSums))

  Q <- stats::expand.model.frame(m, uoanames)
  msk <- !is.na(propertee:::.merge_preserve_order(
    Q,
    merge(unique(m$model$`(offset)`@keys), m@StudySpecification@structure),
    by = uoanames,
    all.x = TRUE,
    sort = FALSE)[zname])
  m_eqns <- Reduce(
    rbind,
    by((m$weights * m$residuals * model.matrix(m))[msk, , drop = FALSE],
       lapply(uoanames, function(col) Q[msk, col]),
       colSums))
  expect_message(b12 <- propertee:::.get_b12(m), paste(nrow(simdata), "rows"))
  expect_equal(b12, t(cmod_eqns) %*% m_eqns)
})

test_that(".get_b12 fails with invalid custom cluster argument", {
  data(simdata)
  cmod_data <- cbind(simdata, new_col = factor(rbinom(nrow(simdata), 1, 0.5)))
  cmod <- lm(y ~ x, cmod_data)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )

  expect_error(propertee:::.get_b12(m, cluster = c("cid3")),
               "cid3 are missing from the covariance adjustment model dataset")
  expect_error(propertee:::.get_b12(m, cluster = c(TRUE, FALSE)),
               "must provide a character vector")
  expect_error(.get_b12(m, cluster = "new_col"),
               "in the teeMod object's StudySpecification: new_col")
})

test_that(".get_b12 produces correct estimates with valid custom cluster argument", {
  data(simdata)
  simdata[simdata$uoa2 == 1, "z"] <- 0
  simdata[simdata$uoa2 == 2, "z"] <- 1

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa2), data = simdata)

  m <- lmitt(y ~ 1, data = simdata, specification = spec, offset = cov_adj(cmod))

  cmod_eqns <- Reduce(
    rbind,
    by(estfun(cmod), list(simdata$uoa2), colSums)
  )
  ssmod_eqns <- Reduce(
    rbind,
    by(estfun(as(m, "lm")), list(simdata$uoa2), colSums)
  )
  expected <- crossprod(cmod_eqns, ssmod_eqns)

  # default (columns specified in `cluster` argument of StudySpecification) matches expected
  expect_equal(suppressMessages(propertee:::.get_b12(m)), expected)
  expect_equal(suppressMessages(propertee:::.get_b12(m, cluster = "uoa2")), expected)
})

test_that("get_b12 handles custom cluster columns with only NA's", {
  data(simdata)
  set.seed(200)

  cmod_data <- data.frame("y" = rnorm(100), "x" = rnorm(100),
                          "uoa1" = NA_integer_, "uoa2" = NA_integer_)
  cmod <- lm(y ~ x, cmod_data)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata,
                      offset = cov_adj(cmod, specification = spec)))

  expect_message(b12 <- propertee:::.get_b12(ssmod, cluster = c("uoa1", "uoa2")), "0 rows")
  expect_equal(b12,
               matrix(0, nrow = 2, ncol = 2))
})

test_that(paste("get_b12 handles multiple custom cluster columns where one is",
                "only NA's and the other has no NA's"), {
  # first test is where the non-NA cluster ID's are distinct
  cmod_data <- data.frame("y" = rnorm(100), "x" = rnorm(100),
                          "uoa1" = rep(seq(6, 10), each = 20),
                          "uoa2" = NA_integer_)
  cmod <- lm(y ~ x, cmod_data)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata,
                      offset = cov_adj(cmod, specification = spec)))

  expect_message(b12 <- propertee:::.get_b12(ssmod, cluster = c("uoa1", "uoa2")), "0 rows")
  expect_equal(b12,
               matrix(0, nrow = 2, ncol = 2))

  # second test is where the non-NA cluster ID's overlap with specification
  cmod_data$uoa1 <- rep(seq(1, 5), each = 20)
  cmod <- lm(y ~ x, cmod_data)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata,
                      offset = cov_adj(cmod, specification = spec)))

  expect_message(b12 <- propertee:::.get_b12(ssmod, cluster = c("uoa1", "uoa2")), "0 rows")
  expect_equal(b12,
               matrix(0, nrow = 2, ncol = 2))
})

test_that(paste(".get_b12 handles multiple custom cluster columns where both",
                "have a mix of NA's and non-NA's"), {
  data(simdata)
  cmod_data <- rbind(simdata, simdata)
  cmod_data[(nrow(simdata)+1):(2*nrow(simdata)), c("uoa1", "uoa2")] <- NA_integer_
  cmod <- lm(y ~ x, cmod_data)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec),
                      offset = cov_adj(cmod)))

  cmod_eqns <- Reduce(
    rbind,
    by(sandwich::estfun(cmod), list(cmod_data$uoa1, cmod_data$uoa2), colSums)
  )
  ssmod_eqns <- Reduce(
    rbind,
    by(sandwich:::estfun.lm(ssmod), list(simdata$uoa1, simdata$uoa2), colSums)
  )
  expected <- crossprod(cmod_eqns, ssmod_eqns)

  expect_message(b12 <- propertee:::.get_b12(ssmod, cluster = c("uoa1", "uoa2")),
                 paste(nrow(simdata), "rows"))
  expect_equal(b12,
               expected)
})

test_that(paste(".get_b12 returns expected B_12 for individual-level",
                "experimental data that is a subset of cov model data"), {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata, subset = simdata$uoa2 == 1)
  weighted_spec <- ate(spec, data = simdata[simdata$uoa2 == 1,])
  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata[simdata$uoa2 == 1,],
       weights = weighted_spec, offset = cov_adj(cmod))
  )

  uoanames <- var_names(m@StudySpecification, "u")
  zname <- var_names(m@StudySpecification, "t")

  # est eqns for lm are wts * resids * dmatrix
  cmod <- m$model$`(offset)`@fitted_covariance_model
  cmod_eqns <- Reduce(
    rbind,
    by((cmod$residuals * model.matrix(cmod))[m$model$`(offset)`@keys$in_Q,],
       lapply(uoanames, function(col) m$model$`(offset)`@keys[m$model$`(offset)`@keys$in_Q,col]),
       colSums))

  Q <- stats::expand.model.frame(m, uoanames)
  msk <- !is.na(.merge_preserve_order(
    Q,
    merge(unique(m$model$`(offset)`@keys[, uoanames, drop = FALSE]), m@StudySpecification@structure),
    by = uoanames,
    all.x = TRUE,
    sort = FALSE)[zname])
  m_eqns <- Reduce(
    rbind,
    by((m$weights * m$residuals * model.matrix(m))[msk, , drop = FALSE],
       lapply(uoanames, function(col) Q[msk, col]),
       colSums))
  expect_message(b12 <- propertee:::.get_b12(m), paste(nrow(simdata[simdata$uoa2 == 1,]), "rows"))
  expect_equal(b12,
               crossprod(cmod_eqns, m_eqns))
})

test_that(paste(".get_b12 returns expected B_12 for cluster-level",
                "experimental data that is a subset of cov model data"), {
  data(simdata)
  subset_cluster_ids <- unique(simdata[simdata$uoa2 == 1, c("uoa1", "uoa2")])
  Q_cluster_subset <- data.frame(Reduce(
    rbind,
    by(simdata,
       list(simdata$uoa1, simdata$uoa2),
       function(x) {colMeans(x[, c("uoa1", "uoa2", "x", "y", "z")])}),
  ), row.names = NULL)

  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = Q_cluster_subset)
  weighted_spec <- ate(spec, data = Q_cluster_subset)

  m <- as.lmitt(
    lm(y ~ assigned(), data = Q_cluster_subset,
       weights = weighted_spec, offset = cov_adj(cmod))
  )

  uoanames <- var_names(m@StudySpecification, "u")
  zname <- var_names(m@StudySpecification, "t")

  # est eqns for lm are wts * resids * dmatrix
  cmod <- m$model$`(offset)`@fitted_covariance_model
  cmod_eqns <- Reduce(
    rbind,
    by(cmod$residuals * model.matrix(cmod),
       lapply(uoanames, function(col) m$model$`(offset)`@keys[,col]),
       colSums))

  Q <- stats::expand.model.frame(m, uoanames)
  msk <- !is.na(propertee:::.merge_preserve_order(
    Q,
    merge(unique(m$model$`(offset)`@keys), m@StudySpecification@structure),
    by = uoanames,
    all.x = TRUE,
    sort = FALSE)[zname])
  m_eqns <- Reduce(
    rbind,
    by((m$weights * m$residuals * model.matrix(m))[msk, , drop = FALSE],
       lapply(uoanames, function(col) Q[msk, col]),
       colSums))
  expect_equal(suppressMessages(propertee:::.get_b12(m)),
               t(cmod_eqns) %*% m_eqns)
})

test_that(paste(".get_b12 returns expected B_12 for experimental",
                "data that has no overlap with cov model data"), {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  C_no_cluster_ids <- simdata
  C_no_cluster_ids[, var_names(spec, "u")] <- NA
  cmod <- lm(y ~ x, data = C_no_cluster_ids)

  m <- as.lmitt(
    lm(y ~ assigned() + force, data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  expect_message(b12 <- propertee:::.get_b12(m), "0 rows")
  expect_equal(dim(b12), c(2, 3))
  expect_true(all(b12 == 0))
})

test_that(paste(".get_b12 returns B_12 with correct dimensions when only one",
                "cluster overlapps between the covariance and direct adjustment",
                "samples"), {
  data(simdata)
  set.seed(200)

  cmod_data <- data.frame("y" = rnorm(10), "x" = rnorm(10),
                          "uoa1" = rep(1, 10), "uoa2" = rep(1, 10))
  cmod <- lm(y ~ x, cmod_data)
  spec <- rct_spec(z ~ cluster(uoa1), simdata, subset = simdata$uoa1 %in% c(1, 5))

  msk <- simdata$uoa1 %in% c(1, 5)
  m <- as.lmitt(
    lm(y ~ assigned(), simdata[msk,],
       weights = ate(spec, data = simdata[msk,]),
       offset = cov_adj(cmod, newdata = simdata[msk,]))
  )

  expect_warning(
    expect_message(b12 <- propertee:::.get_b12(m), paste(nrow(cmod_data), "rows")),
    "numerically indistinguishable from 0"
  )
  expect_equal(dim(b12), c(2, 2))
  expect_equal(b12, matrix(0, nrow = 2, ncol = 2))
})

test_that(paste(".get_b12 returns expected matrix when some rows in cmod data",
                "have NA values for covariates"), {
  data(simdata)
  set.seed(96)

  # random rows in cmod data have NA covariate values
  rvals <- runif(n = length(which(simdata$uoa1 %in% c(2, 4))))
  names(rvals) <- which(simdata$uoa1 %in% c(2, 4))
  rvals <- sort(rvals)
  simdata[as.numeric(names(rvals)[seq_len(round(length(rvals) / 4))]),
          "x"] <- NA_real_

  cmod <- lm(y ~ x, simdata, subset = uoa1 %in% c(2, 4))
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  expect_warning(expect_warning(lmitt(lm(y ~ assigned(), simdata,
                                         offset = cov_adj(cmod, specification = spec))),
                                "adjustments are NA"), "adjustments are NA")
  ## warning procs twice
  m <- suppressWarnings(
    lmitt(lm(y ~ assigned(), simdata, offset = cov_adj(cmod, specification = spec)))
  )

  cmod_eqns <- Reduce(
    rbind,
    by(sandwich::estfun(cmod),
       list(uoa1 = simdata$uoa1[!is.na(simdata$x) & simdata$uoa1 %in% c(2, 4)],
            uoa2 = simdata$uoa2[!is.na(simdata$x) & simdata$uoa1 %in% c(2, 4)]),
       colSums))
  msk <- which(simdata$uoa1[!is.na(simdata$x)] %in% c(2, 4))
  ssmod_eqns <- Reduce(
    rbind,
    by(estfun(as(m, "lm"))[msk, , drop = FALSE],
       list(uoa1 = simdata$uoa1[!is.na(simdata$x)][msk],
            uoa2 = simdata$uoa2[!is.na(simdata$x)][msk]),
       colSums))
  b12 <- suppressMessages(propertee:::.get_b12(m))
  expect_equal(b12, crossprod(cmod_eqns, ssmod_eqns))
})

test_that(paste(".get_b12 returns expected value for B12 when no intercept is",
                "included in the direct adjustment model"), {
  data(simdata)
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)

  m <- as.lmitt(lm(y ~ assigned() - 1, simdata, weights = ate(spec),
                   offset = cov_adj(cmod)))

  uoanames <- var_names(m@StudySpecification, "u")
  zname <- var_names(m@StudySpecification, "t")

  # est eqns for lm are wts * resids * dmatrix
  cmod <- m$model$`(offset)`@fitted_covariance_model
  cmod_eqns <- Reduce(
    rbind,
    by(cmod$residuals * model.matrix(cmod),
       lapply(uoanames, function(col) m$model$`(offset)`@keys[,col]),
       colSums))

  Q <- stats::expand.model.frame(m, uoanames)
  msk <- !is.na(propertee:::.merge_preserve_order(
    Q,
    merge(unique(m$model$`(offset)`@keys), m@StudySpecification@structure),
    by = uoanames,
    all.x = TRUE,
    sort = FALSE)[zname])
  m_eqns <- Reduce(
    rbind,
    by((m$weights * m$residuals * model.matrix(m))[msk, , drop = FALSE],
       lapply(uoanames, function(col) Q[msk, col]),
       colSums))
  expect_equal(suppressMessages(propertee:::.get_b12(m)),
               t(cmod_eqns) %*% m_eqns)
})

test_that(".get_a22_inverse correct w/o covariance adjustment", {
  data(simdata)

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  m_as.lmitt <- as.lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec)))
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec, weights = ate(spec))

  nq <- nrow(stats::model.frame(m_as.lmitt))
  inv_fim <- nq * chol2inv(m_as.lmitt$qr$qr)

  a22inv <- .get_a22_inverse(m_as.lmitt)
  expect_true(all.equal(a22inv[1:2, 1:2], inv_fim, check.attributes = FALSE))
  expect_true(all.equal(a22inv[3, 3], nq / sum(weights(m_as.lmitt@ctrl_means_model)), check.attributes = FALSE))
  expect_true(all(a22inv[1:2, 3] == 0) & all(a22inv[3, 1:2] == 0))
  
  a22inv <- .get_a22_inverse(m_lmitt.form)
  expect_true(all.equal(a22inv[1:2, 1:2], inv_fim, check.attributes = FALSE))
  expect_true(all.equal(a22inv[3, 3], nq / sum(weights(m_as.lmitt@ctrl_means_model)), check.attributes = FALSE))
  expect_true(all(a22inv[1:2, 3] == 0) & all(a22inv[3, 1:2] == 0))
})

test_that(".get_a22_inverse correct w/ covariance adjustment", {
  set.seed(438)
  data(simdata)
  testdata <- simdata
  nc <- 30
  nq <- nrow(testdata)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)

  testdata$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), testdata)

  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), data = testdata, weights = ate(spec), offset = cov_adj(cmod)
  ))
  m_lmitt.form <- lmitt(y ~ 1, data = testdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  inv_fim <- nq * chol2inv(m_as.lmitt$qr$qr)

  a22inv <- .get_a22_inverse(m_as.lmitt)
  expect_true(all.equal(a22inv[1:2, 1:2], inv_fim, check.attributes = FALSE))
  expect_true(all.equal(a22inv[3:4, 3:4], diag(2), check.attributes = FALSE))
  expect_true(all(a22inv[1:2, 3:4] ==0) & all(a22inv[3:4, 1:2] == 0))
  
  a22inv <- .get_a22_inverse(m_lmitt.form)
  expect_true(all.equal(a22inv[1:2, 1:2], inv_fim, check.attributes = FALSE))
  expect_true(all.equal(a22inv[3:4, 3:4], diag(2), check.attributes = FALSE))
  expect_true(all(a22inv[1:2, 3:4] ==0) & all(a22inv[3:4, 1:2] == 0))
})

test_that(".get_a22_inverse with missing values", {
  set.seed(438)
  data(simdata)
  testdata <- simdata
  testdata$y[1:3] <- NA_real_
  nc <- 30
  nq <- nrow(testdata)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)
  
  testdata$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), testdata)
  
  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), data = testdata, weights = ate(spec), offset = cov_adj(cmod)
  ))
  m_lmitt.form <- lmitt(y ~ 1, data = testdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))
  
  a22inv <- .get_a22_inverse(m_as.lmitt)
  expect_true(all.equal(a22inv[1:2, 1:2], nq * chol2inv(m_as.lmitt$qr$qr), check.attributes = FALSE))
  expect_true(all.equal(a22inv[3:4, 3:4],
                        diag(2) * nq / sum(weights(m_as.lmitt@ctrl_means_model), na.rm = TRUE),
                        check.attributes = FALSE))
  
  a22inv <- .get_a22_inverse(m_lmitt.form)
  expect_true(all.equal(a22inv[1:2, 1:2], nq * chol2inv(m_lmitt.form$qr$qr), check.attributes = FALSE))
  expect_true(all.equal(a22inv[3:4, 3:4],
                        diag(2) * nq / sum(weights(m_as.lmitt@ctrl_means_model), na.rm = TRUE),
                        check.attributes = FALSE))
})

test_that(".get_tilde_a22_inverse correct w/o covariance adjustment", {
  data(simdata)
  new_data <- simdata

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = new_data)
  m_as.lmitt <- as.lmitt(lm(y ~ assigned(), data = new_data, weights = ate(spec)))
  m_lmitt.form <- lmitt(y ~ 1, data = new_data, specification = spec, weights = ate(spec))

  expect_equal(.get_tilde_a22_inverse(m_as.lmitt), .get_a22_inverse(m_as.lmitt))
  expect_equal(.get_tilde_a22_inverse(m_lmitt.form), .get_a22_inverse(m_lmitt.form))
})

test_that(".get_tilde_a22_inverse correct w/ covariance adjustment", {
  set.seed(438)
  data(simdata)
  new_data <- simdata
  nc <- 30
  nq <- nrow(new_data)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)

  new_data$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), new_data)

  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), data = new_data, weights = ate(spec), offset = cov_adj(cmod)
  ))
  m_lmitt.form <- lmitt(y ~ 1, data = new_data, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  expect_equal(.get_tilde_a22_inverse(m_as.lmitt), n / nq * .get_a22_inverse(m_as.lmitt))
  expect_equal(.get_tilde_a22_inverse(m_lmitt.form), n / nq * .get_a22_inverse(m_lmitt.form))
})

test_that(".get_tilde_a22_inverse with missing values", {
  set.seed(438)
  data(simdata)
  new_data <- simdata
  new_data$y[1:3] <- NA_real_
  nc <- 30
  nq <- nrow(new_data)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)
  
  new_data$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), new_data)
  
  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), data = new_data, weights = ate(spec), offset = cov_adj(cmod)
  ))
  m_lmitt.form <- lmitt(y ~ 1, data = new_data, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))
  
  expect_equal(.get_tilde_a22_inverse(m_as.lmitt), n / nq * .get_a22_inverse(m_as.lmitt))
  expect_equal(.get_tilde_a22_inverse(m_lmitt.form), n / nq * .get_a22_inverse(m_lmitt.form))
})

test_that(".get_b22 returns correct value for lm object w/o offset", {
  data(simdata)

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(lm(y ~ assigned(), data = simdata, weights = ate(spec)))
  nq <- nrow(sandwich::estfun(m))
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1L))
  expect_equal(propertee:::.get_b22(m, type = "HC1"),
               vmat * nuoas / (nuoas - 1L) * (nq - 1L) / (nq - 2L))
})

test_that(".get_b22 returns correct value for lm object w offset", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  nq <- nrow(simdata)
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0", type_psi = "HC0", type_phi = "HC0"),
               vmat * nuoas / (nuoas - 1L))
  expect_equal(propertee:::.get_b22(m, type = "HC1", type_psi = "HC1", type_phi = "HC0"),
               vmat * nuoas / (nuoas - 1L) * (nq - 1L) / (nq - 2L))
})

test_that(".get_b22 fails with invalid custom cluster argument", {
  data(simdata)
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )

  expect_error(propertee:::.get_b22(m, cluster = c("cid3")),
               "cid3 are missing from the direct adjustment")
  expect_error(propertee:::.get_b22(m, cluster = c(TRUE, FALSE)),
               "must provide a character vector")
})

test_that(".get_b22 produces correct estimates with valid custom cluster argument", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  nq <- nrow(simdata)
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")

  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, cluster = uoanames, type = "HC0"),
               vmat * nuoas / (nuoas - 1L))
})

test_that(".get_b22 with one clustering column", {
  data(simdata)

  simdata[simdata$uoa1 == 4, "z"] <- 0
  simdata[simdata$uoa1 == 2, "z"] <- 1
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ uoa(uoa1), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  uoas <- factor(simdata$uoa1)
  nuoas <- length(levels(uoas))

  nq <- nrow(simdata)
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoa_matrix <- stats::model.matrix(as.formula("~ -1 + as.factor(uoa1)"),
                                    stats::expand.model.frame(m, "uoa1")[, "uoa1", drop = FALSE])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1L))
})

test_that(".get_b22 returns corrrect value for glm fit with Gaussian family", {
  data(simdata)
  return(expect_true(TRUE)) # added 1/14/25 because it's deprecated and now failing

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(glm(y ~ assigned(), data = simdata, weights = ate(spec)))
  nq <- nrow(sandwich::estfun(m))
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1))
})

test_that(".get_b22 returns correct value for poisson glm", {
  data(simdata)
  return(expect_true(TRUE)) # added 1/14/25 because it's deprecated and now failing

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    glm(round(exp(y)) ~ assigned(), data = simdata, weights = ate(spec),
        family = stats::poisson())
  )
  nq <- nrow(sandwich::estfun(m))
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1))
})

test_that(".get_b22 returns correct value for quasipoisson glm", {
  data(simdata)
  return(expect_true(TRUE)) # added 1/14/25 because it's deprecated and now failing

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    glm(round(exp(y)) ~ assigned(), data = simdata, weights = ate(spec),
        family = stats::quasipoisson())
  )
  nq <- nrow(sandwich::estfun(m))
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1))
})

test_that(".get_b22 returns correct value for binomial glm", {
  data(simdata)
  return(expect_true(TRUE)) # added 1/14/25 because it's deprecated and now failing

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- suppressWarnings(
    as.lmitt(
      glm(round(exp(y) / (1 + exp(y))) ~ assigned(), data = simdata, weights = ate(spec),
          family = stats::binomial())
    ))

  nq <- nrow(sandwich::estfun(m))
  WX <- m$weights * m$residuals * stats::model.matrix(m)

  uoanames <- var_names(m@StudySpecification, "u")
  form <- paste0("~ -1 + ", paste("as.factor(", uoanames, ")", collapse = ":"))
  uoa_matrix <- stats::model.matrix(as.formula(form),
                                    stats::expand.model.frame(m, uoanames)[, uoanames])

  uoa_eqns <- crossprod(uoa_matrix, WX)
  vmat <- crossprod(uoa_eqns)
  expect_equal(propertee:::.get_b22(m, type = "HC0"),
               vmat * nuoas / (nuoas - 1))
})

test_that(".get_a11_inverse returns correct value for lm cmod", {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m_as.lmitt <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  nc <- nrow(stats::model.frame(cmod))
  fim <- crossprod(stats::model.matrix(cmod))
  expect_equal(propertee:::.get_a11_inverse(m_as.lmitt), nc * solve(fim))
  expect_equal(propertee:::.get_a11_inverse(m_lmitt.form), nc * solve(fim))
})

test_that(".get_a11_inverse returns correct value for Gaussian glm cmod", {
  data(simdata)
  cmod <- glm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m_as.lmitt <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  nc <- nrow(stats::model.frame(cmod))
  dispersion <- sum((cmod$weights * cmod$residuals)^2) / sum(cmod$weights)
  fim <- crossprod(stats::model.matrix(cmod), stats::model.matrix(cmod))
  expect_equal(propertee:::.get_a11_inverse(m_as.lmitt), nc * dispersion * solve(fim))
  expect_equal(propertee:::.get_a11_inverse(m_lmitt.form), nc * dispersion * solve(fim))
})

test_that(".get_a11_inverse returns correct value for poisson glm cmod", {
  data(simdata)
  cmod <- glm(round(exp(y)) ~ x, data = simdata, family = stats::poisson())
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m_as.lmitt <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  nc <- nrow(stats::model.frame(cmod))
  fim <- crossprod(stats::model.matrix(cmod) * exp(cmod$linear.predictors),
                   stats::model.matrix(cmod))
  # tol due to diffs from chol2inv vs. solve(crossprod)
  expect_equal(propertee:::.get_a11_inverse(m_as.lmitt), nc * solve(fim), tolerance = 1e-4)
  expect_equal(propertee:::.get_a11_inverse(m_lmitt.form), nc * solve(fim), tolerance = 1e-4)
})

test_that(".get_a11_inverse returns correct value for quasipoisson glm cmod", {
  data(simdata)
  cmod <- glm(round(exp(y)) ~ x, data = simdata, family = stats::quasipoisson())
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m_as.lmitt <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  nc <- nrow(stats::model.frame(cmod))
  dispersion <- sum((cmod$weights * cmod$residuals)^2) / sum(cmod$weights)
  fim <- crossprod(stats::model.matrix(cmod) * exp(cmod$linear.predictors),
                   stats::model.matrix(cmod))
  # tol due to diffs from chol2inv vs. solve(crossprod)
  expect_equal(propertee:::.get_a11_inverse(m_as.lmitt), nc * dispersion * solve(fim),
               tolerance = 1e-4)
  expect_equal(propertee:::.get_a11_inverse(m_lmitt.form), nc * dispersion * solve(fim),
               tolerance = 1e-4)
})

test_that(".get_a11_inverse returns correct value for binomial glm cmod", {
  data(simdata)
  cmod <- suppressWarnings(
    glm(round(exp(y) / (1 + exp(y))) ~ x, data = simdata, family = stats::binomial())
  )
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m_as.lmitt <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )
  m_lmitt.form <- lmitt(y ~ 1, data = simdata, specification = spec,
                        weights = ate(spec), offset = cov_adj(cmod))

  nc <- nrow(stats::model.frame(cmod))
  fim <- crossprod(stats::model.matrix(cmod) * cmod$fitted.values * (1 - cmod$fitted.values),
                   stats::model.matrix(cmod))
  # tol due to diffs from chol2inv vs. solve(crossprod)
  expect_equal(propertee:::.get_a11_inverse(m_as.lmitt), nc * solve(fim), tolerance = 1e-6)
  expect_equal(propertee:::.get_a11_inverse(m_lmitt.form), nc * solve(fim), tolerance = 1e-6)
})

test_that(".get_b11 returns correct B_11 for one cluster column", {
  data(simdata)

  simdata[simdata$uoa1 == 4, "z"] <- 0
  simdata[simdata$uoa1 == 2, "z"] <- 1
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ uoa(uoa1), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec),
       offset = cov_adj(cmod)))

  uoas <- factor(simdata$uoa1)
  nuoas <- length(levels(uoas))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L) * (nc - 1L) / (nc - 2L)
  )
  expect_equal(.get_b11(m), expected)
})

test_that(".get_b11 fails with invalid custom cluster argument", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  nuoas <- nrow(spec@structure)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec),
       offset = cov_adj(cmod))
  )

  expect_error(propertee:::.get_b11(m, cluster = c("cid3")),
               "cid3 are missing from the covariance adjustment model dataset")
  expect_error(propertee:::.get_b11(m, cluster = c(TRUE, FALSE)),
               "must provide a character vector")
})

test_that(".get_b11 produces correct estimates with valid custom cluster argument", {
  data(simdata)

  simdata[simdata$uoa1 == 4, "z"] <- 0
  simdata[simdata$uoa1 == 2, "z"] <- 1
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ uoa(uoa1), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  uoas <- factor(simdata$uoa1)
  nuoas <- length(levels(uoas))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L) * (nc - 1L) / (nc - 2L)
  )

  expect_equal(propertee:::.get_b11(m, cluster = "uoa1"), expected)

  # test different clustering level
  bids <- factor(simdata[, "bid"])
  nbids <- length(levels(bids))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), simdata[, "bid"], colSums))) *
      nbids / (nbids - 1L) * (nc - 1L) / (nc - 2L)
  )
  expect_equal(propertee:::.get_b11(m, cluster = "bid"), expected)
})

test_that(".get_b11 handles NA's correctly in custom clustering columns", {
  data(simdata)
  set.seed(200)

  # check case where all clustering columns only have NA's
  cmod_data <- data.frame("y" = rnorm(100), "x" = rnorm(100),
                          "uoa1" = NA_integer_, "uoa2" = NA_integer_)
  cmod <- lm(y ~ x, cmod_data)
  nc <- sum(summary(cmod)$df[1L:2L])

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata,
                      offset = cov_adj(cmod, specification = spec)))

  expect_warning(propertee:::.get_b11(ssmod, cluster = c("uoa1", "uoa2")),
                 "are found to have NA's")
  expect_equal(suppressWarnings(propertee:::.get_b11(ssmod, cluster = c("uoa1", "uoa2"),
                                         type = "HC0", cadjust = FALSE)),
               crossprod(sandwich::estfun(cmod))) # there should be no clustering

  # check case where one clustering column doesn't only have NA's
  cmod_data$uoa1 <- rep(seq(6, 10), each = 20)
  cmod <- lm(y ~ x, cmod_data)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  ssmod <- as.lmitt(lm(y ~ assigned(), data = simdata,
                      offset = cov_adj(cmod, specification = spec)))

  expect_warning(propertee:::.get_b11(ssmod, cluster = c("uoa1", "uoa2")),
                 "have NA's for some but not all")
})

test_that(".get_b11 returns correct B_11 for multiple cluster columns", {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  uoas <- factor(Reduce(function(x, y) paste(x, y, sep = "_"), simdata[, c("uoa1", "uoa2")]))
  nuoas <- length(levels(uoas))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L) * (nc - 1L) / (nc - 2L)
  )
  expect_equal(propertee:::.get_b11(m), expected)
})

test_that(".get_b11 returns correct B_11 for lm cmod (HC1)", {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod)))

  uoas <- factor(Reduce(function(x, y) paste(x, y, sep = "_"), simdata[, c("uoa1", "uoa2")]))
  nuoas <- length(levels(uoas))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L) * (nc - 1L) / (nc - 2L)
  )
  expect_equal(propertee:::.get_b11(m), expected)
})

test_that(".get_b11 returns correct B_11 for glm object (HC0)", {
  data(simdata)
  return(expect_true(TRUE)) # added 1/14/25 because it's deprecated and now failing
  cmod <- glm(round(exp(y)) ~ x, data = simdata, family = stats::poisson())
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)

  m <- as.lmitt(
    glm(round(exp(y)) ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod),
        family = stats::poisson()))

  uoas <- factor(Reduce(function(x, y) paste(x, y, sep = "_"), simdata[, c("uoa1", "uoa2")]))
  nuoas <- length(levels(uoas))

  nc <- sum(summary(cmod)$df[1L:2L])
  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L)
  )
  expect_equal(propertee:::.get_b11(m), expected)
})

test_that(paste(".get_b11 returns correct B_11 for experimental data that is a",
                "subset of cov model data (also tests NA's in some cluster",
                "but not all cluster columns)"), {
  data(simdata)
  cmod <- lm(y ~ x, data = simdata)
  nc <- sum(summary(cmod)$df[1L:2L])

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata, subset = simdata$uoa2 == 1)
  weighted_spec <- ate(spec, data = simdata[simdata$uoa2 == 1,])
  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata[simdata$uoa2 == 1,],
       weights = weighted_spec, offset = cov_adj(cmod))
  )

  # replace NA's with distinct uoa values and recalculate nuoas for small-sample adjustment
  uoas <- Reduce(function(x, y) paste(x, y, sep = "_"), m$model$`(offset)`@keys)
  uoas <- factor(uoas)
  nuoas <- length(levels(uoas))

  expected <- (
    crossprod(Reduce(rbind, by(sandwich::estfun(cmod), uoas, colSums))) *
      nuoas / (nuoas - 1L) * (nc - 1L) / (nc - 2L)
  )
  expect_equal(propertee:::.get_b11(m), expected)
})


test_that(paste(".get_b11 returns correct B_11 for experimental data that has",
                "no overlap with cov model data"), {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  C_no_cluster_ids <- simdata
  C_no_cluster_ids[, var_names(spec, "u")] <- NA
  cmod <- lm(y ~ x, data = C_no_cluster_ids)
  nc <- sum(summary(cmod)$df[1L:2L])

  m <- as.lmitt(
    lm(y ~ assigned(), data = simdata, weights = ate(spec), offset = cov_adj(cmod))
  )

  # replace NA's with distinct uoa values and recalculate nuoas for small-sample adjustment
  uoas <- Reduce(function(x, y) paste(x, y, sep = "_"), m$model$`(offset)`@keys)
  uoas[grepl("NA", uoas)] <- NA_integer_
  nuoas <- length(unique(uoas))
  nas <- is.na(uoas)
  uoas[nas] <- paste0(nuoas - 1 + seq_len(sum(nas)), "*")
  uoas <- factor(uoas)

  # no adjustment for cases where each row is its own cluster
  expected <- crossprod(sandwich::estfun(cmod)) * (nc - 1L) / (nc - 2L)
  expect_equal(propertee:::.get_b11(m, cadjust = FALSE), expected)
})

test_that(".get_b11 returns expected B_11 when cmod fit to one cluster", {
  data(simdata)
  cmod <- lm(y ~ x, simdata, subset = uoa1 == 1)
  spec <- rct_spec(z ~ cluster(uoa1), simdata, subset = simdata$uoa1 %in% c(1, 5))

  msk <- simdata$uoa1 %in% c(1, 5)
  m <- as.lmitt(
    lm(y ~ assigned(), simdata[msk,],
       weights = ate(spec, data = simdata[msk,]),
       offset = cov_adj(cmod, newdata = simdata[msk,]))
  )
  expect_warning(propertee:::.get_b11(m),
                 "meat matrix numerically indistinguishable")
  expect_equal(suppressWarnings(propertee:::.get_b11(m)),
               matrix(0, nrow = 2, ncol = 2))
})

test_that(".get_a21 returns correct matrix for lm cmod and lm ssmod", {
  data(simdata)

  new_df <- simdata
  new_df$uid <- seq_len(nrow(new_df))
  cmod <- lm(y ~ x, new_df)
  spec <- rct_spec(z ~ unitid(uid), data = new_df)

  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), new_df, weights = ate(spec), offset = cov_adj(cmod)
  ))
  m_lmitt.form <- lmitt(y ~ 1, specification = spec, data = new_df,
                        weights = ate(spec), offset = cov_adj(cmod))

  Qmat <- m_as.lmitt$weights * stats::model.matrix(m_as.lmitt)
  ctrl.means.grad <- cbind(matrix(0, nrow = nrow(Qmat), ncol = 1),
                           model.matrix(m_as.lmitt@ctrl_means_model) *
                             weights(m_as.lmitt@ctrl_means_model))
  colnames(ctrl.means.grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  Cmat <- stats::model.matrix(cmod)
  nq <- nrow(stats::model.frame(m_as.lmitt))

  a21_as.lmitt <- .get_a21(m_as.lmitt)
  expect_equal(dim(a21_as.lmitt), c(4, 2))
  expect_true(all.equal(a21_as.lmitt, crossprod(cbind(Qmat, -ctrl.means.grad), Cmat) / nq,
                        check.atrributes = FALSE))

  a21_lmitt.form <- .get_a21(m_lmitt.form)
  expect_equal(dim(a21_lmitt.form), c(4, 2))
  expect_true(all.equal(a21_lmitt.form, crossprod(cbind(Qmat, -ctrl.means.grad), Cmat) / nq,
                        check.attributes = FALSE))
})

test_that(".get_a21 returns correct matrix for glm cmod and lm ssmod", {
  data(simdata)
  new_df <- simdata
  new_df$id <- seq(nrow(new_df))
  new_df$bin_y <- rbinom(nrow(new_df), 1, round(1 / (1 + exp(-new_df$x))))

  cmod <- suppressWarnings(glm(bin_y ~ x + force, data = new_df, family = stats::binomial()))
  spec <- rct_spec(z ~ unitid(id), new_df)
  m_as.lmitt <- as.lmitt(lm(
    bin_y ~ assigned(), new_df, weights = ate(spec), offset = cov_adj(cmod, by = "id")
  ))
  m_lmitt.form <- lmitt(bin_y ~ 1, specification = spec, data = new_df,
                        weights = ate(spec), offset = cov_adj(cmod, by = "id"))

  Qmat <- stats::model.matrix(m_as.lmitt)
  ctrl.means.grad <- cbind(matrix(0, nrow = nrow(Qmat), ncol = 1),
                           model.matrix(m_as.lmitt@ctrl_means_model) *
                             weights(m_as.lmitt@ctrl_means_model))
  colnames(ctrl.means.grad) <- c("bin_y:(Intercept)", "cov_adj:(Intercept)")
  Cmat <- cmod$prior.weights * cmod$family$mu.eta(cmod$linear.predictors) * stats::model.matrix(cmod)
  nq <- nrow(stats::model.frame(m_as.lmitt))

  a21_as.lmitt <- .get_a21(m_as.lmitt)
  expect_equal(dim(a21_as.lmitt), c(4, 3))
  expect_true(all.equal(a21_as.lmitt, crossprod(cbind(Qmat, -ctrl.means.grad), Cmat) / nq,
                        check.atrributes = FALSE))

  a21_lmitt.form <- .get_a21(m_lmitt.form)
  expect_equal(dim(a21_lmitt.form), c(4, 3))
  expect_true(all.equal(a21_lmitt.form, crossprod(cbind(Qmat, -ctrl.means.grad), Cmat) / nq,
                        check.attributes = FALSE))
})

test_that(".get_a21 with a moderator", {
  data(simdata)
  
  new_df <- simdata
  new_df$uid <- seq_len(nrow(new_df))
  cmod <- lm(y ~ x, new_df)
  spec <- rct_spec(z ~ unitid(uid), data = new_df)

  m_lmitt.form <- lmitt(y ~ x, specification = spec, data = new_df,
                        weights = ate(spec), offset = cov_adj(cmod))
  
  Qmat <- m_lmitt.form$weights * stats::model.matrix(m_lmitt.form)
  ctrl.means.grad <- cbind(matrix(0, nrow = nrow(Qmat), ncol = 2),
                           model.matrix(m_lmitt.form@ctrl_means_model) *
                             weights(m_lmitt.form@ctrl_means_model))
  colnames(ctrl.means.grad) <- c("y:(Intercept)", "cov_adj:(Intercept)", "y:x", "cov_adj:x")
  Cmat <- stats::model.matrix(cmod)
  nq <- nrow(stats::model.frame(m_lmitt.form))

  a21_lmitt.form <- .get_a21(m_lmitt.form)
  expect_equal(dim(a21_lmitt.form), c(8, 2))
  expect_true(all.equal(a21_lmitt.form, crossprod(cbind(Qmat, -ctrl.means.grad), Cmat) / nq,
                        check.attributes = FALSE))
})

test_that(".get_tilde_a21 correct values", {
  set.seed(438)
  data(simdata)

  new_df <- simdata
  nc <- 30
  nq <- nrow(new_df)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)

  new_df$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), new_df)

  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), new_df, weights = ate(spec), offset = cov_adj(cmod, by = "id")
  ))
  m_lmitt.form <- lmitt(y ~ 1, specification = spec, data = new_df,
                        weights = ate(spec), offset = cov_adj(cmod, by = "id"))

  expect_equal(.get_tilde_a21(m_as.lmitt), nq / n * .get_a21(m_as.lmitt))
  expect_equal(.get_tilde_a21(m_lmitt.form), nq / n * .get_a21(m_lmitt.form))
})

test_that(".get_tilde_a21 with missing", {
  set.seed(438)
  data(simdata)
  
  new_df <- simdata
  new_df$y[1:3] <- NA_real_
  nc <- 30
  nq <- nrow(new_df)
  n <- nc + nq
  cmod_data <- data.frame(y = rnorm(nc), x = rnorm(nc), id = nq + seq(nc))
  cmod <- lm(y ~ x, cmod_data)
  
  new_df$id <- seq(nq)
  spec <- rct_spec(z ~ unitid(id), new_df)
  
  m_as.lmitt <- as.lmitt(lm(
    y ~ assigned(), new_df, weights = ate(spec), offset = cov_adj(cmod, by = "id")
  ))
  m_lmitt.form <- lmitt(y ~ 1, specification = spec, data = new_df,
                        weights = ate(spec), offset = cov_adj(cmod, by = "id"))
  
  expect_equal(.get_tilde_a21(m_as.lmitt), nq / n * .get_a21(m_as.lmitt))
  expect_equal(.get_tilde_a21(m_lmitt.form), nq / n * .get_a21(m_lmitt.form))
})

test_that(paste(".get_a21 returns correct matrix when data input for lmitt",
                "has NA values for some covariate values"), {
  data(simdata)
  simdata[simdata$uoa1 == 1, "x"] <- NA_real_
  cmod_data <- subset(simdata, uoa1 %in% c(2, 4))
  m_data <- subset(simdata, uoa1 %in% c(1, 3, 5))
  cmod <- lm(y ~ x, data = cmod_data)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), m_data)

  # warning procs twice
  expect_warning(expect_warning(
    m_as.lmitt <- lmitt(lm(y ~ assigned(), m_data, offset = cov_adj(cmod, specification = spec))),
    "covariance adjustments are NA"), "covariance adjustments are NA")
  expect_warning(
    m_lmitt.form <- lmitt(y ~ 1, specification = spec, data = m_data, offset = cov_adj(cmod)),
    "covariance adjustments are NA"
  )

  ssmod_mm <- suppressWarnings(stats::model.matrix(m_as.lmitt))
  ctrl_means_mm <- (
    model.matrix(m_lmitt.form@ctrl_means_model) *
      weights(m_lmitt.form@ctrl_means_model)[!is.na(weights(m_lmitt.form@ctrl_means_model))]
  )
  ctrl.means.grad <- cbind(matrix(0, nrow = nrow(ssmod_mm), ncol = 1),
                           ctrl_means_mm[
                                 row.names(ctrl_means_mm) %in% setdiff(
                                   row.names(ctrl_means_mm), stats::na.action(m_lmitt.form))
                               ])
  colnames(ctrl.means.grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  pg <- stats::model.matrix(formula(cmod), m_data)
  nq <- nrow(m_data)

  a21_as.lmitt <- suppressWarnings(.get_a21(m_as.lmitt))
  expect_true(all.equal(a21_as.lmitt, crossprod(cbind(ssmod_mm, -ctrl.means.grad), pg) / nq,
                        check.attributes = FALSE))
  a21_lmitt.form <- suppressWarnings(.get_a21(m_lmitt.form))
  expect_true(all.equal(a21_lmitt.form, crossprod(cbind(ssmod_mm, -ctrl.means.grad), pg) / nq,
                        check.attributes = FALSE))
})

test_that(paste(".get_a21 returns correct matrix when data input for lmitt has
                NA values for some treatment assignments"), {
  data(simdata)
  simdata[simdata$uoa1 %in% c(2, 4), "z"] <- NA_integer_
  cmod <- lm(y ~ x, data = simdata, subset = uoa1 %in% c(2, 4))
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  m_as.lmitt <- lmitt(
    lm(y ~ assigned(), simdata, w = ate(spec, data = simdata),
       offset = cov_adj(cmod, specification = spec)))
  m_lmitt.form <- lmitt(y ~ 1, simdata, specification = spec, w = ate(spec), offset = cov_adj(cmod))

  ssmod_mm <- stats::model.matrix(m_as.lmitt) * m_as.lmitt$weights
  ctrl.means.grad <- cbind(matrix(0, nrow = nrow(ssmod_mm), ncol = 1),
                           model.matrix(m_lmitt.form@ctrl_means_model) *
                             weights(m_lmitt.form@ctrl_means_model)[
                               !is.na(weights(m_lmitt.form@ctrl_means_model))])
  colnames(ctrl.means.grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  pg <- stats::model.matrix(formula(cmod), simdata)[!is.na(simdata$z),]
  nq <- nrow(simdata)

  a21_as.lmitt <- suppressWarnings(.get_a21(m_as.lmitt))
  expect_true(all.equal(a21_as.lmitt, crossprod(cbind(ssmod_mm, -ctrl.means.grad), pg) / nq,
                        check.attributes = FALSE))
  a21_lmitt.form <- suppressWarnings(.get_a21(m_lmitt.form))
  expect_true(all.equal(a21_lmitt.form, crossprod(cbind(ssmod_mm, -ctrl.means.grad), pg) / nq,
                        check.attributes = FALSE))
})

test_that(".get_a21 returns only full rank columns for less than full rank model", {
  data(simdata)
  copy_simdata <- simdata
  copy_simdata$o_fac <- as.factor(copy_simdata$o)
  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), copy_simdata)

  ### lmitt.formula
  ssmod <- lmitt(y ~ o_fac, data = copy_simdata, specification = spec, offset = cov_adj(cmod))
  ctrl.means.mm <- model.matrix(ssmod@ctrl_means_model)
  expect_equal(dim(a21 <- .get_a21(ssmod)),
               c(ssmod$rank + 2 * ncol(ctrl.means.mm),
                 ssmod$model$`(offset)`@fitted_covariance_model$rank))
  keep_ix <- ssmod$qr$pivot[1L:ssmod$rank]
  expect_equal(rownames(a21),
               c(colnames(model.matrix(ssmod))[keep_ix],
                 paste(rep(c("y", "cov_adj"), each = ncol(ctrl.means.mm)),
                       rep(colnames(ctrl.means.mm), 2), sep = ":")))
})

test_that(".vcov_CR with covariance adjustment and no moderator returns (p+2)x(p+2) matrix", {
  data(simdata)

  cmod <- lm(y ~ x, simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
  m <- as.lmitt(lm(y ~ assigned(), simdata, weights = ate(spec), offset = cov_adj(cmod)))

  vmat <- propertee:::.vcov_CR(m)
  expect_equal(dim(vmat), c(4, 4))
})

test_that(".vcov_CR returns teeMod model sandwich if it has no SandwichLayer", {
  data(simdata)

  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  m <- lmitt(y ~ 1, data = simdata, specification = spec, weights = ate(spec))

  uoas <- apply(simdata[, c("uoa1", "uoa2")], 1, function(...) paste(..., collapse = "_"))
  expect_true(all.equal(vcov_tee(m, type = "CR0"),
                        sandwich::sandwich(m,
                                           meat. = sandwich::meatCL,
                                           cluster = uoas),
                        check.attributes = FALSE))
})

test_that("test the output of vcov_tee is correct with bias corrections", {
  data(simdata)
  
  cmod <- lm(y ~ x + z, simdata)
  spec <- rct_spec(z ~ unitid(uoa1, uoa2), simdata)
  xm <- lmitt(y ~ 1, spec, simdata, offset = cov_adj(cmod))
  
  cls <- paste(simdata$uoa1, simdata$uoa2, sep = "_")
  jk_units <- unique(cls)
  X <- stats::model.matrix(cmod)
  X[,"z"] <- 0
  Z <- stats::model.matrix(xm)
  ZTZ_inv <- chol2inv(xm$qr$qr)
  new_r <- Reduce(
    c,
    mapply(
      function(loo_unit, cmod, cls) {
        cmod_cl <- stats::getCall(cmod)
        cmod_cl$subset <- eval(cls != loo_unit)
        loo_cmod <- eval(cmod_cl, envir = environment(formula(cmod)))
        cl_ix <- cls == loo_unit
        (simdata$y - xm$fitted.values + xm$offset)[cl_ix] - drop(X[cl_ix,,drop=FALSE] %*% loo_cmod$coefficients)
      },
      jk_units,
      SIMPLIFY = FALSE,
      MoreArgs = list(cmod = cmod, cls = cls)
    )
  )
  ef_psi <- estfun(as(xm, "lm")) / xm$residuals * new_r
  
  n <- nrow(simdata)
  ef_phi <- estfun(cmod)
  
  ef_ctrl_means_model <- estfun(xm@ctrl_means_model)
  br <- .get_tilde_a22_inverse(xm)
  vc <- br %*% (
    crossprod(rowsum(cbind(ef_psi, ef_ctrl_means_model) -
                       ef_phi %*% t(.get_a11_inverse(xm)) %*% t(.get_a21(xm)),
                     cls)) / n^2
  ) %*% t(br)
  expect_true(all.equal(vc, vcov_tee(xm, cadjust = FALSE, loco_residuals = TRUE),
                        check.attributes = FALSE))
})

test_that(paste("HC0 .vcov_CR lm w/o clustering",
                "balanced grps",
                "no cmod/ssmod data overlap", sep = ", "), {
  set.seed(50)
  N <- 60

  # trt variable
  z <- rep(c(0, 1), N / 2)

  # p and mu for one categorical and one continuous cov
  px1 <- round(stats::runif(1, 0.05, 0.94), 1)
  mux2 <- round(stats::runif(1, 55, 90))
  x1 <- rep(c(0, 1), each = N / 2)
  v <- stats::rgamma(N, 50) # simulate dirichlet to balance continuous cov
  x2 <- c(t(
    Reduce(cbind, by(v, z, function(vc) vc / sum(vc) * mux2 * N / 2))
  ))

  df1 <- data.frame("z" = z, "x1" = x1, "x2" = x2, "uid" = seq_len(N))

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df1$y <- drop(stats::model.matrix(~ x1 + x2 + z, df1) %*% theta + rnorm(N))
  df2 <- df1
  df2$uid <- NA_integer_
  df <- rbind(df1, df2)

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()
  cmod <- lm(cmod_form, df, subset = is.na(uid))
  spec <- rct_spec(z ~ uoa(uid), df, subset = !is.na(df$uid))
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, subset = !is.na(uid), weights = ate(spec),
       offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, specification = spec, data = df, subset = !is.na(uid),
                           weights = ate(spec), offset = cov_adj(cmod))
  onemod <- lm(y ~ x1 + x2 + z, data = df, subset = !is.na(uid),
               weights = ate(spec))

  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(cmod_form, df[!is.na(df$uid),])
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ssmod_as.lmitt@ctrl_means_model) *
                     weights(ssmod_as.lmitt@ctrl_means_model))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nc <- nrow(Xstar)
  nq <- nrow(Z)
  n <- nc + nq

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- sandwich::bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * mean(ssmod_as.lmitt$weights)))
  
  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  ef_ssmod <- rbind(ef_ssmod, matrix(0, nrow = nc, ncol = ncol(ef_ssmod)))
  ef_cmod <- estfun(cmod)
  ef_cmod <- rbind(matrix(0, nrow = nrow(Z), ncol = ncol(ef_cmod)), ef_cmod)
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  ef_ctrl_means <- rbind(ef_ctrl_means, matrix(0, nrow = nc, ncol = ncol(ef_ctrl_means)))
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(estfun(ssmod_as.lmitt)) / n,
               crossprod(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% t(a11inv) %*% t(a21)) / n)

  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meat(ssmod_as.lmitt, adjust = FALSE))

  # with perfect group balance, a22inv %*% a21 ((Z'WZ)^(-1)Z'WX) should be 0
  # for the trt row
  expect_equal((bread. %*% a21)[2,],
               setNames(rep(0, dim(a21)[2]), colnames(a21)))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- propertee:::.vcov_CR(ssmod_as.lmitt, cluster = seq_len(n), cadjust = FALSE)
  expect_true(all.equal(vmat, (1/n) * bread. %*% meat. %*% t(bread.),
                        check.attributes = FALSE))

  # vmat should be equal to the outputs of sandwich
  expect_true(all.equal(vmat, sandwich::sandwich(ssmod_as.lmitt),
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = seq_len(n), cadjust = FALSE),
                        check.attributes = FALSE))

  # var_hat(z) should be smaller than var_hat(z) from onemod
  expect_true(all(diag(vmat) <
                  diag(vcov(onemod))[c(1, 4)]))
})

test_that(paste("HC0 .vcov_CR lm w/o clustering",
                "imbalanced grps",
                "no cmod/ssmod data overlap", sep = ", "), {
  set.seed(50)
  N <- 60

  # trt variable
  z <- rep(c(0, 1), N / 2)

  # p and mu for one categorical and one continuous cov
  px1 <- round(stats::runif(1, 0.05, 0.94), 1)
  mux2 <- round(stats::runif(1, 55, 90))
  x1 <- rbinom(N, 1, px1)
  x2 <- stats::rgamma(N, mux2)

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2, "uid" = seq_len(N))

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df$y <- drop(stats::model.matrix(~ x1 + x2 + z, df) %*% theta + rnorm(N))
  df$uid[seq_len(3 * N / 4)] <- NA_integer_

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()
  cmod <- lm(cmod_form, df, subset = is.na(uid))
  spec <- rct_spec(z ~ uoa(uid), df, subset = !is.na(df$uid))
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, subset = !is.na(uid), weights = ate(spec),
       offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, specification = spec, data = df, subset = !is.na(uid),
                           weights = ate(spec), offset = cov_adj(cmod))
  onemod <- lm(y ~ x1 + x2 + z, data = df, subset = !is.na(uid),
               weights = ate(spec))

  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(cmod_form, df[!is.na(df$uid),])
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ssmod_as.lmitt@ctrl_means_model) *
                     weights(ssmod_as.lmitt@ctrl_means_model))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nc <- nrow(Xstar)
  nq <- nrow(Z)
  n <- nc + nq

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- sandwich::bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * (n / sum(weights(ssmod_as.lmitt@ctrl_means_model)))))

  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  ef_ssmod <- rbind(ef_ssmod, matrix(0, nrow = nc, ncol = ncol(ef_ssmod)))
  ef_cmod <- estfun(cmod)
  ef_cmod <- rbind(matrix(0, nrow = nrow(Z), ncol = ncol(ef_cmod)), ef_cmod)
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  ef_ctrl_means <- rbind(ef_ctrl_means, matrix(0, nrow = nc, ncol = ncol(ef_ctrl_means)))
  expect_true(all.equal(meat. <- crossprod(estfun(ssmod_as.lmitt)) / n,
                        crossprod(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% t(a11inv) %*% t(a21)) / n,
                        check.attributes = FALSE))


  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meat(ssmod_as.lmitt, adjust = FALSE))

  # with imperfect group balance, a22inv %*% a21 should not be 0 for the trt row
  expect_false(all((bread. %*% a21)[2,] == 0))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- propertee:::.vcov_CR(ssmod_as.lmitt, cluster = seq_len(n), cadjust = FALSE)
  expect_true(all.equal(vmat, (1/n) * bread. %*% meat. %*% bread.,
                        check.attributes = FALSE))

  # vmat should be equal the outputs of sandwich
  expect_true(all.equal(vmat, sandwich::sandwich(ssmod_as.lmitt),
                       check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = seq_len(n), cadjust = FALSE),
                        check.attributes = FALSE))

  # var_hat(z) should be smaller than var_hat(z) from onemod
  expect_true(all(diag(vmat) <
                    diag(vcov(onemod))[c(1, 4)]))
})

test_that(paste("HC0 .vcov_CR lm w/ clustering",
                "balanced grps",
                "no cmod/ssmod data overlap", sep = ", "), {
  set.seed(50)
  nclusts_C <- nclusts_Q <- 4
  MI <- 16

  # trt variable
  z <- c(rep(rep(c(0, 1), each = MI), nclusts_C / 2),
         rep(rep(c(0, 1), each = MI), nclusts_Q / 2))

  # p and mu for each matched pair's categorical and continuous cov
  px1 <- round(stats::runif((nclusts_C + nclusts_Q) / 2, 0.05, 0.94), 1)
  mux2 <- round(stats::runif((nclusts_C) / 2, 55, 90))
  x1 <- c(
    mapply(function(x) {
      rep(c(rep(0, round(MI * x)), rep(1, round(MI * (1 - x)))), 2)
    }, px1)
  )
  x2 <- c(
    Reduce(
      cbind,
      Map(function(x) {
        # simulate dirichlet to balance continuous cov
        v <- stats::rgamma(MI * 2, 50)
        Reduce(cbind, by(v, rep(seq_len(2), each = MI),
                         function(vc) vc / sum(vc) * x * MI))
      }, mux2)
    )
  )

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2)

  # generate clustered errors
  error_sd <- round(stats::runif(nclusts_C + nclusts_Q, 1, 3), 1)
  icc <- 0.2
  eps <- stats::rnorm(nrow(df))
  Sigma <- matrix(0, nrow = nrow(df), ncol = nrow(df))
  for (i in (seq_len(nclusts_C + nclusts_Q) - 1)) {
    msk <- (1 + i * MI):((i + 1) * MI)
    Sigma[msk, msk] <- diag(error_sd[i + 1] - icc, nrow = MI) + icc
  }
  A <- chol(Sigma)
  eps <- t(A) %*% eps

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df$y <- drop(stats::model.matrix(~ x1 + x2 + z, df) %*% theta + eps)
  df$cid <- c(rep(NA_integer_, nclusts_C * MI), rep(seq_len(nclusts_Q), each = MI))

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()

  cmod <- lm(cmod_form, df, subset = is.na(cid))
  spec <- rct_spec(z ~ cluster(cid), df, subset = !is.na(df$cid))
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, subset = !is.na(cid), weights = ate(spec),
       offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, specification = spec, data = df, subset = !is.na(cid),
                           weights = ate(spec), offset = cov_adj(cmod))

  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(cmod_form, df[!is.na(df$cid),])
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ssmod_as.lmitt@ctrl_means_model) *
                     weights(ssmod_as.lmitt@ctrl_means_model))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nq <- nrow(Z)
  nc <- nrow(Xstar)
  n <- nc + nq

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- sandwich::bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * n / sum(weights(ssmod_as.lmitt@ctrl_means_model))))

  ids <- c(df$cid[!is.na(df$cid)],
           paste0(length(df$cid[!is.na(df$cid)]) + seq_len(length(df$cid[is.na(df$cid)])),
                  "*"))
  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  ef_ssmod <- rbind(ef_ssmod, matrix(0, nrow = nc, ncol = ncol(ef_ssmod)))
  ef_cmod <- estfun(cmod)
  ef_cmod <- rbind(matrix(0, nrow = nrow(Z), ncol = ncol(ef_cmod)), ef_cmod)
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  ef_ctrl_means <- rbind(ef_ctrl_means, matrix(0, nrow = nc, ncol = ncol(ef_ctrl_means)))
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(Reduce(rbind, by(estfun(ssmod_as.lmitt), ids, colSums))) / n,
               crossprod(Reduce(
                 rbind,
                 by(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% t(a11inv) %*% t(a21), ids, colSums))) / n)

  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meatCL(ssmod_as.lmitt, cluster = ids, cadjust = FALSE))

  # with perfect group balance, a22inv %*% a21 should have 0's in the trt row
  expect_equal((bread. %*% a21)[2,], setNames(rep(0, dim(a21)[2]), colnames(a21)))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- propertee:::.vcov_CR(ssmod_as.lmitt, cluster = ids, cadjust = FALSE)
  expect_true(all.equal(vmat, (1 / n) * bread. %*% meat. %*% bread.,
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = ids, cadjust = FALSE),
                        check.attributes = FALSE))

  # vmat should be equal the outputs of sandwich
  expect_true(all.equal(vmat,
                        sandwich::sandwich(ssmod_as.lmitt,
                                           meat. = sandwich::meatCL,
                                           cluster = ids, cadjust = FALSE),
                        check.attributes = FALSE))
})

test_that(paste("HC0 .vcov_CR lm w/ clustering",
                "imbalanced grps",
                "no cmod/ssmod data overlap", sep = ", "), {
  set.seed(50)
  nclusts_C <- nclusts_Q <- 4
  MI <- 16

  # trt variable
  z <- c(rep(rep(c(0, 1), each = MI), nclusts_C / 2),
         rep(rep(c(0, 1), each = MI), nclusts_Q / 2))

  # p and mu for each matched pair's categorical and continuous cov
  px1 <- round(stats::runif((nclusts_C + nclusts_Q) / 2, 0.05, 0.94), 1)
  mux2 <- round(stats::runif((nclusts_C + nclusts_Q) / 2, 55, 90))
  x1 <- c(mapply(function(x) c(rbinom(MI, 1, x), rbinom(MI, 1, x)), px1))
  x2 <- c(Reduce(cbind, Map(function(x) stats::rgamma(MI * 2, x), mux2)))

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2)

  # generate clustered errors
  error_sd <- round(stats::runif(nclusts_C + nclusts_Q, 1, 3), 1)
  icc <- 0.2
  eps <- stats::rnorm(nrow(df))
  Sigma <- matrix(0, nrow = nrow(df), ncol = nrow(df))
  for (i in (seq_len(nclusts_C + nclusts_Q) - 1)) {
    msk <- (1 + i * MI):((i + 1) * MI)
    Sigma[msk, msk] <- diag(error_sd[i + 1] - icc, nrow = MI) + icc
  }
  A <- chol(Sigma)
  eps <- t(A) %*% eps

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df$y <- drop(stats::model.matrix(~ x1 + x2 + z, df) %*% theta + eps)
  df$cid <- c(rep(NA_integer_, nclusts_C * MI), rep(seq_len(nclusts_Q), each = MI))

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()

  cmod <- lm(cmod_form, df, subset = is.na(cid))
  spec <- rct_spec(z ~ cluster(cid), df, subset = !is.na(df$cid))
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, subset = !is.na(cid), weights = ate(spec),
       offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, data = df, specification = spec, subset = !is.na(cid),
                           weights = ate(spec), offset = cov_adj(cmod))
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model

  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(cmod_form, df[!is.na(df$cid),])
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ssmod_as.lmitt@ctrl_means_model) *
                     weights(ssmod_as.lmitt@ctrl_means_model))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nc <- nrow(Xstar)
  nq <- nrow(Z)
  n <- nc + nq


  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- sandwich::bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * n / sum(weights(ctrl_means_mod))))

  ids <- c(df$cid[!is.na(df$cid)],
           paste0(length(df$cid[!is.na(df$cid)]) + seq_len(length(df$cid[is.na(df$cid)])),
                  "*"))
  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  ef_ssmod <- rbind(ef_ssmod, matrix(0, nrow = nc, ncol = ncol(ef_ssmod)))
  ef_cmod <- estfun(cmod)
  ef_cmod <- rbind(matrix(0, nrow = nq, ncol = ncol(ef_cmod)), ef_cmod)
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  ef_ctrl_means <- rbind(ef_ctrl_means, matrix(0, nrow = nc, ncol = ncol(ef_ctrl_means)))
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(Reduce(rbind, by(estfun(ssmod_as.lmitt), ids, colSums))) / n,
               crossprod(Reduce(
                 rbind,
                 by(cbind(ef_ssmod, ef_ctrl_means) -  nq / nc * ef_cmod %*% t(a11inv) %*% t(a21), ids, colSums))) / n)

  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meatCL(ssmod_as.lmitt, cluster = ids, cadjust = FALSE))

  # with imperfect group balance, a22inv %*% a21 should not be 0 for the trt row
  expect_false(all((bread. %*% a21)[2,] == 0))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- .vcov_CR(ssmod_as.lmitt, cluster = ids, cadjust = FALSE)
  expect_true(all.equal(vmat, (1 / n) * bread. %*% meat. %*% bread.,
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = ids, cadjust = FALSE),
                        check.attributes = FALSE))

  # vmat should be the same as the outputs of sandwich
  expect_true(all.equal(vmat,
                        sandwich::sandwich(ssmod_as.lmitt,
                                           meat. = sandwich::meatCL,
                                           cluster = ids, cadjust = FALSE),
                        check.attributes = FALSE))
})

test_that(paste("HC0 .vcov_CR lm w/o clustering",
                "imbalanced grps",
                "cmod is a strict subset of ssmod data", sep = ", "), {
  set.seed(50)
  N <- 60

  # trt variable
  z <- rep(c(0, 1), N / 2)

  # p and mu for one categorical and one continuous cov
  px1 <- round(stats::runif(1, 0.05, 0.94), 1)
  mux2 <- round(stats::runif(1, 55, 90))
  x1 <- rbinom(N, 1, px1)
  x2 <- stats::rgamma(N, mux2)

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2, "uid" = seq_len(N))

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df$y <- drop(stats::model.matrix(~ x1 + x2 + z, df) %*% theta + rnorm(N))

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()
  cmod_idx <- df$z == 0
  cmod <- lm(cmod_form, df[cmod_idx,])
  spec <- rct_spec(z ~ uoa(uid), df)
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, weights = ate(spec), offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, data = df, specification = spec, weights = ate(spec),
                           offset = cov_adj(cmod))
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(as.formula(cmod_form[-2]), df)
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ctrl_means_mod) * weights(ctrl_means_mod))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nc <- nrow(Xstar)
  nq <- n <- nrow(Z)

  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * n / sum(weights(ctrl_means_mod))))

  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  nonzero_ef_cmod <- estfun(cmod)
  ef_cmod <- matrix(0, nrow = nrow(ef_ssmod), ncol = ncol(nonzero_ef_cmod))
  colnames(ef_cmod) <- colnames(nonzero_ef_cmod)
  ef_cmod[which(df$z == 0),] <- nonzero_ef_cmod
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(estfun(ssmod_as.lmitt)) / n,
               crossprod(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% a11inv %*% t(a21)) / n)

  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meat(ssmod_as.lmitt, adjust = FALSE))

  # with imperfect group balance, a22inv %*% a21 should not be 0 for the trt row
  expect_false(all((bread. %*% a21)[2,] == 0))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- .vcov_CR(ssmod_as.lmitt, cluster = df$uid, cadjust = FALSE)
  expect_true(all.equal(vmat, (1 / n) * bread. %*% meat. %*% t(bread.),
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = df$uid, cadjust = FALSE),
                        check.attributes = FALSE))

  # vmat should be the same as the outputs of sandwich
  expect_true(all.equal(vmat, sandwich::sandwich(ssmod_as.lmitt, adjust = FALSE),
                        check.attributes = FALSE))
})

test_that(paste("HC0 .vcov_CR lm w/ clustering",
                "imbalanced grps",
                "cmod is a strict subset of ssmod data", sep = ", "), {
  set.seed(50)
  nclusts_C <- nclusts_Q <- 4
  MI <- 16

  # trt variable
  z <- c(rep(rep(c(0, 1), each = MI), nclusts_C / 2),
         rep(rep(c(0, 1), each = MI), nclusts_Q / 2))

  # p and mu for each matched pair's categorical and continuous cov
  px1 <- round(stats::runif((nclusts_C + nclusts_Q) / 2, 0.05, 0.94), 1)
  mux2 <- round(stats::runif((nclusts_C) / 2, 55, 90))
  x1 <- c(mapply(function(x) c(rbinom(MI, 1, x), rbinom(MI, 1, x)), px1))
  x2 <- c(Reduce(cbind, Map(function(x) stats::rgamma(MI * 2, x), mux2)))

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2)

  # generate clustered errors
  error_sd <- round(stats::runif(nclusts_C + nclusts_Q, 1, 3), 1)
  icc <- 0.2
  eps <- stats::rnorm(nrow(df))
  Sigma <- matrix(0, nrow = nrow(df), ncol = nrow(df))
  for (i in (seq_len(nclusts_C + nclusts_Q) - 1)) {
    msk <- (1 + i * MI):((i + 1) * MI)
    Sigma[msk, msk] <- diag(error_sd[i + 1] - icc, nrow = MI) + icc
  }
  A <- chol(Sigma)
  eps <- t(A) %*% eps

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  df$y <- drop(stats::model.matrix(~ x1 + x2 + z, df) %*% theta + eps)
  df$cid <- rep(seq_len(nclusts_C + nclusts_Q), each = MI)

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()
  cmod_idx <- df$z == 0

  cmod <- lm(cmod_form, df[cmod_idx,])
  spec <- rct_spec(z ~ cluster(cid), df)
  ssmod_as.lmitt <- as.lmitt(
    lm(ssmod_form, data = df, weights = ate(spec), offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, data = df, specification = spec, weights = ate(spec),
                           offset = cov_adj(cmod))
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model

  Xstar <- stats::model.matrix(cmod)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(as.formula(cmod_form[-2]), df)
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ctrl_means_mod) * weights(ctrl_means_mod))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  nc <- nrow(Xstar)
  nq <- n <- nrow(Z)

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar)))
  expect_equal(a21 <- .get_a21(ssmod_as.lmitt), crossprod(cbind(Z * ssmod_as.lmitt$weights, -cm_grad), X) / nq)
  bread. <- sandwich::bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * ssmod_as.lmitt$weights, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * n / sum(weights(ctrl_means_mod))))

  ids <- df$cid
  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  loo_r <- .compute_loo_resids(ssmod_as.lmitt, "cid")
  ef_ssmod <- ef_ssmod / stats::residuals(ssmod_as.lmitt) * loo_r
  nonzero_ef_cmod <- estfun(cmod)
  ef_cmod <- matrix(0, nrow = nrow(ef_ssmod), ncol = ncol(nonzero_ef_cmod))
  colnames(ef_cmod) <- colnames(nonzero_ef_cmod)
  ef_cmod[which(df$z == 0),] <- nonzero_ef_cmod
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(Reduce(rbind, by(estfun(ssmod_as.lmitt, loco_residuals = TRUE),
                                                   ids, colSums))) / n,
               crossprod(Reduce(
                 rbind,
                 by(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% t(a11inv) %*% t(a21), ids, colSums))) / n)

  # meat should be the same as the output of sandwich::meatCL
  expect_equal(meat., sandwich::meatCL(ssmod_as.lmitt, cluster = ids, cadjust = FALSE,
                                       loco_residuals = TRUE))

  # with imperfect group balance, a22inv %*% a21 should not be 0 for the trt row
  expect_false(all((bread. %*% a21)[2,] == 0))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- .vcov_CR(ssmod_as.lmitt, cluster = ids, cadjust = FALSE, loco_residuals = TRUE)
  expect_true(all.equal(vmat, (1 / n) * bread. %*% meat. %*% t(bread.),
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = ids, cadjust = FALSE, loco_residuals = TRUE),
                        check.attributes = FALSE))

  # vmat should be the same as the outputs of sandwich
  expect_true(all.equal(vmat,
                        sandwich::sandwich(ssmod_as.lmitt,
                                           meat. = sandwich::meatCL,
                                           cluster = ids, cadjust = FALSE,
                                           loco_residuals = TRUE),
                        check.attributes = FALSE))
})

test_that(paste("HC0 .vcov_CR binomial glm cmod",
                "w/o clustering",
                "imbalanced grps",
                "cmod is a strict subset of ssmod data", sep = ", "), {
  set.seed(50)
  N <- 60

  # trt variable
  z <- rep(c(0, 1), N / 2)

  # p and mu for one categorical and one continuous cov
  px1 <- round(stats::runif(1, 0.05, 0.94), 1)
  mux2 <- round(stats::runif(1, 55, 90))
  x1 <- rbinom(N, 1, px1)
  x2 <- stats::rgamma(N, mux2)

  df <- data.frame("z" = z, "x1" = x1, "x2" = x2, "uid" = seq_len(N))

  # generate y
  theta <- c(65, 1.5, -0.01, -2) # intercept, x1, x2, and treatment coeffs
  py <- 1 / (1 + exp(-scale(stats::model.matrix(~ x1 + x2 + z, df),
                            scale = FALSE) %*% theta))
  df$y <- rbinom(N, 1, py)

  cmod_form <- y ~ x1 + x2
  ssmod_form <- y ~ assigned()
  cmod <- glm(cmod_form, data = df, subset = z == 0, family = stats::binomial())
  spec <- rct_spec(z ~ uoa(uid), df)
  ssmod_as.lmitt <- lmitt(
    lm(ssmod_form, data = df, weights = ate(spec), offset = cov_adj(cmod))
  )
  ssmod_lmitt.form <- lmitt(y ~ 1, data = df, specification = spec, weights = ate(spec),
                           offset = cov_adj(cmod))
  ctrl_means_mod <- ssmod_as.lmitt@ctrl_means_model

  ## COMPARE BLOCKS TO MANUAL DERIVATIONS
  Xstar <- stats::model.matrix(cmod)
  fit_wstar <- cmod$weights
  rstar <- cmod$residuals
  mu_etastar <- cmod$family$mu.eta(cmod$linear.predictors)
  Z <- stats::model.matrix(ssmod_as.lmitt)
  X <- stats::model.matrix(formula(stats::delete.response(terms(cmod))), df)
  cm_grad <- cbind(matrix(0, nrow = nrow(X), ncol = 1),
                   stats::model.matrix(ctrl_means_mod) * weights(ctrl_means_mod))
  colnames(cm_grad) <- c("y:(Intercept)", "cov_adj:(Intercept)")
  w <- ssmod_as.lmitt$weights
  r <- ssmod_as.lmitt$residuals
  mu_eta <- cmod$family$mu.eta(drop(X %*% cmod$coefficients))
  nc <- nrow(Xstar)
  nq <- n <- nrow(Z)

  expect_equal(a11inv <- .get_a11_inverse(ssmod_as.lmitt), nc * solve(crossprod(Xstar * sqrt(fit_wstar))))
  expect_true(all.equal(a21 <- .get_a21(ssmod_lmitt.form),
                        crossprod(cbind(Z * w, -cm_grad), X * mu_eta) / nq,
                        check.attributes = FALSE))
  bread. <- bread(ssmod_as.lmitt)
  expect_equal(bread.[1:2, 1:2], n * solve(crossprod(Z * w, Z)))
  expect_equal(bread.[3:4, 3:4],
               matrix(0, nrow = 2, ncol = 2, dimnames = list(colnames(cm_grad), colnames(cm_grad))) + (
                 diag(2) * n / sum(weights(ctrl_means_mod))))

  ef_ssmod <- utils::getS3method("estfun", "lm")(ssmod_as.lmitt)
  nonzero_ef_cmod <- estfun(cmod)
  ef_cmod <- matrix(0, nrow = nrow(ef_ssmod), ncol = ncol(nonzero_ef_cmod))
  colnames(ef_cmod) <- colnames(nonzero_ef_cmod)
  ef_cmod[which(df$z == 0),] <- nonzero_ef_cmod
  ef_ctrl_means <- sweep(residuals(ctrl_means_mod), # need sweep bc residuals has 2 columns
                         1,
                         weights(ctrl_means_mod) * model.matrix(ctrl_means_mod), # only 1 col in model matrix
                         FUN = "*")
  colnames(ef_ctrl_means) <- colnames(cm_grad)
  expect_equal(meat. <- crossprod(estfun(ssmod_as.lmitt)) / n,
               crossprod(cbind(ef_ssmod, ef_ctrl_means) - nq / nc * ef_cmod %*% t(a11inv) %*% t(a21)) / n)

  # meat should be the same as the output of sandwich::meat
  expect_equal(meat., sandwich::meat(ssmod_as.lmitt, adjust = FALSE))

  # with imperfect group balance, a22inv %*% a21 should not be 0 for the trt row
  expect_false(all((bread. %*% a21)[2,] == 0))

  # check .vcov_CR0 matches manual matrix multiplication
  vmat <- .vcov_CR(ssmod_as.lmitt, cluster = seq_len(n), cadjust = FALSE)
  expect_true(all.equal(vmat, (1 / n) * bread. %*% meat. %*% t(bread.),
                        check.attributes = FALSE))

  # vmat should be the same for both lmitt calls
  expect_true(all.equal(vmat,
                        .vcov_CR(ssmod_lmitt.form, cluster = seq(n), cadjust = FALSE),
                        check.attributes = FALSE))

  # vmat should be the same as the outputs of sandwich
  expect_true(all.equal(vmat, sandwich::sandwich(ssmod_as.lmitt, adjust = FALSE),
                        check.attributes = FALSE))
})

test_that("type attribute", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ 1, data = simdata, weights = "ate", specification = spec)
  expect_identical(attr(vcov(ssmod), "type"), "HC0")
  expect_identical(attr(vcov(ssmod, type = "CR2"), "type"), "CR2")
  expect_identical(attr(vcov(ssmod, type = "MB2"), "type"), "MB2")
  expect_identical(attr(vcov(ssmod, type = "HC2"), "type"), "HC2")
})

test_that("#119 flagging vcov_tee entries as NA", {
  ### factor moderator variable
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  mod <- lmitt(y ~ as.factor(o), data = simdata, specification = spec)
  expect_warning(vc <- vcov_tee(mod, type = "CR0"),
                 "will be returned as NA: as.factor(o)1, as.factor(o)3",
                 fixed = TRUE)

  # Issue is in subgroups w/ moderator=1:z=0, moderator=1:z=1, and
  # moderator=3, so .check_df_moderator_estimates should NA those vcov entries
  na_dim <- c(1, 3, 5, 8, 10)
  expect_true(all(is.na(vc[na_dim, ])))
  expect_true(all(is.na(vc[, na_dim])))
  expect_true(all(!is.na(vc[-na_dim, -na_dim])))

  ### different factor moderator variable (may not produce negative diagonals
  ### but subgroups with moderator=1:z=0 and moderator=1:z=1 should be
  ### numerically 0)
  set.seed(37)
  ddata <- data.frame(modr = factor(c(rep(1, 8), rep(2, 12))),
                      y = rnorm(20),
                      z = c(rep(rep(c(0, 1), each = 4), 2), rep(1, 4)),
                      ass_id = rep(seq(5), each = 4))
  dspec <- rct_spec(z ~ unitid(ass_id), ddata)
  ssmod <- lmitt(y ~ modr, specification = dspec, data = ddata)
  expect_warning(vc <- vcov_tee(ssmod, type = "CR0"),
                 "will be returned as NA: modr1",
                 fixed = TRUE)

  na_dim <- c(1, 3, 5)
  expect_true(all(
    abs(
      diag(sandwich::sandwich(ssmod, meat. = sandwich::meatCL,
                              cluster = .make_uoa_ids(ssmod, "CR")))[na_dim]
    ) < .Machine$double.eps)
  )
  expect_true(all(is.na(vc[na_dim, ])))
  expect_true(all(is.na(vc[, na_dim])))
  expect_true(all(!is.na(vc[-na_dim, -na_dim])))

  ### valid factor moderator variable; vcov diagonals shouldn't be numerically 0
  ### when using the sandwich package and shouldn't have NA's when using vcov_tee
  ddata <- data.frame(modr = factor(c(rep(1, 8), rep(2, 12))),
                      y = rnorm(20),
                      z = c(rep(rep(c(0, 1), each = 4), 2), rep(1, 4)),
                      ass_id = rep(seq(10), each = 2))
  dspec <- rct_spec(z ~ unitid(ass_id), ddata)
  ssmod <- lmitt(y ~ modr, specification = dspec, data = ddata)
  vc <- vcov_tee(ssmod, type = "CR0")
  expect_true(all(
    abs(diag(sandwich::sandwich(ssmod, meat. = sandwich::meatCL,
                                cluster = .make_uoa_ids(ssmod, "CR")))[na_dim])
    > .Machine$double.eps)
  )
  expect_true(all(!is.na(vc)))

  #### lmitt.lm
  ## we chose to implement special logic for subgroup-level d.f. only
  ## for lmitt objects created with lmitt.formula; the commented-out
  ## tests that follow would have tested similar logic for lmitt.lm
  ##objects. (this distinction is reflected in `lmitt.lm()`'s leaving
  ## the "moderator" slot empty for teeMod objects,
  ## so .check_df_moderator_estimates will return
  ## the initial vcov matrix, and we need not check it here.)
  # mod <- lmitt(lm(y ~ as.factor(o) + as.factor(o):assigned(spec), data = simdata), specification = spec)
  # vc <- vcov_tee(mod)[5:7, 5:7]
  #
  # #****************************************
  # ### Setting these to NA manually only for testing purposes!
  # vc[1, ] <- NA
  # vc[, 1] <- NA
  # ### Remove these once #119 is addressed!!!!!
  # #****************************************
  #
  # # Issue is in subgroup o_fac=1, so the first entry in the vcov matrix
  # expect_true(all(is.na(vc[1, ])))
  # expect_true(all(is.na(vc[, 1])))
  # expect_true(all(!is.na(vc[-1, -1])))
  #
  # ### valid continuous moderator variable
  ssmod <- lmitt(y ~ o, data = simdata, specification = spec)
  vc <- vcov(ssmod, type = "CR0")
  expect_true(all(!is.na(vc)))
  #
  # ### invalid continuous moderator variable
  simdata <- simdata[(simdata$uoa1 == 2 & simdata$uoa2 == 2) |
                      (simdata$uoa1 == 2 & simdata$uoa2 == 1),]
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ o, data = simdata, specification = spec)
  expect_warning(vc <- vcov(ssmod, type = "CR0"), "will be returned as NA: o")
  na_dim <- which(grepl("(Intercept)", colnames(vc)) |
                    grepl("z._o", colnames(vc)))
  expect_true(all(is.na(vc[na_dim, ])))
  expect_true(all(is.na(vc[, na_dim])))
  expect_true(all(!is.na(vc[-na_dim, -na_dim])))
})

test_that(".check_df_moderator_estimates other warnings", {
  data(simdata)

  # fail without a teeMod model
  nossmod <- lm(y ~ x, simdata)
  vmat <- vcov(nossmod)
  cluster_ids <- apply(simdata[, c("uoa1", "uoa2")], 1, function(...) paste(..., collapse = "_"))
  expect_error(.check_df_moderator_estimates(vmat, nossmod, cluster_ids),
               "must be a teeMod")

  # invalid `data` arg
  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  ssmod <- lmitt(y ~ factor(o), specification = spec, data = simdata)
  expect_error(.check_df_moderator_estimates(vmat, ssmod
                                             , cluster_cols = c("uoa1", "uoa2")
                                             , model_data = cbind(y = simdata$y,
                                                                  z = simdata$z)
                                             , envir = parent.frame()),
               "`model_data` must be a dataframe")
})

test_that(".check_df_moderator_estimates #246", {
  # observations that did not contribute to the model fit do not count towards
  # the degrees of freedom count for moderator estimates
  set.seed(341)
  Qdata <- data.frame(x = factor(rep(c(1, 2), each = 3)),
                      z = rep(c(0, 1), 3),
                      y = runif(6))
  wts <- c(0, rep(1, 5))
  expect_warning(spec <- obs_spec(z ~ 1, Qdata), "explicit unit")
  tm <- lmitt(y ~ x, spec, Qdata, weights = wts)
  expect_warning(
    vc <- vcov_tee(tm),
    "x1"
  )
  expect_true(all(is.na(vc["z._x1",])))
  expect_true(all(is.na(vc[,"z._x1"])))

  Qdata$x2 <- runif(nrow(Qdata))
  wts <- rep(c(0, 1), c(2, 4))
  tm <- lmitt(y ~ x2, spec, Qdata, weights = wts)
  expect_warning(
    vc <- vcov_tee(tm),
    "x2"
  )
  expect_true(all(is.na(vc["z._x2",])))
  expect_true(all(is.na(vc[,"z._x2"])))
  
  # if values_from is fit w/o the moderator, don't check df
  values_from <- lmitt(y ~ 1, spec, Qdata, weights = wts)
  vc <- vcov_tee(tm, values_from = values_from)
  expect_true(all(is.na(vc)))
})

if (requireNamespace("multcomp", quietly = TRUE)) {
  test_that(".check_df_moderator_estimates w/ mmm object and aliased coeffs", {
    set.seed(4953)
    dft <- data.frame(y = rnorm(20),
                      z = rep(c(0,1), each = 10),
                      x1 = factor(rep(letters[1:3], 7)[seq_len(20)]))
    dft$x2 <- dft$z
    expect_warning(dta <- rct_spec(z ~ 1, dft), "explicit unit")
    tm1 <- lmitt(y ~ 1, dta, dft)
    tm2 <- lmitt(y ~ x1, dta, dft)
    tm3 <- lmitt(y ~ x2, dta, dft)
    mco <- multcomp::mmm(main = tm1, x1 = tm2, x2 = tm3)
    expect_silent(invisible(vcov_tee(mco)))
  })
}

test_that("#123 ensure PreSandwich are converted to Sandwich", {
  data(simdata)
  spec <- rct_spec(z ~ uoa(uoa1, uoa2), data = simdata)
  cmod <- lm(y ~ x, data = simdata)
  # Make sure its PreSandwich prior
  expect_warning(ca <- cov_adj(cmod, newdata = simdata),
                 "Without a specification")
  expect_false(is(ca, "SandwichLayer"))
  ssmod <- as.lmitt(lm(y ~ a.(spec), data = simdata, offset = ca),
                    specification = spec)
  expect_true(is(ssmod$model$`(offset)`, "SandwichLayer"))


  copy_simdata <- simdata
  copy_simdata$schoolid <- seq(900, 949)
  C_df <- rbind(copy_simdata[, c("schoolid", "x", "y")],
                data.frame(schoolid = seq(1000, 1049),
                           x = rnorm(50),
                           y = rnorm(50)))
  cmod <- lm(y ~ x, C_df)
  spec <- rct_spec(z ~ uoa(schoolid), copy_simdata)
  lm1 <- lm(y ~ assigned(), copy_simdata, weights = ett(specification = spec),
            offset = cov_adj(cmod, NULL, NULL, "schoolid"))
  ssmod1 <- lmitt(lm1, specification = spec)
  expect_true(is(ssmod$model$`(offset)`, "SandwichLayer"))
  v1 <- vcov_tee(ssmod1)

  wts2 <- ett(spec, data = copy_simdata)
  expect_warning(offst2 <- cov_adj(cmod, copy_simdata, by = "schoolid"),
                 "Without a specification")
  lm2 <- lm(y ~ assigned(), copy_simdata, weights = wts2, offset = offst2)
  ssmod2 <- lmitt(lm2, specification = spec)
  expect_true(is(ssmod$model$`(offset)`, "SandwichLayer"))
  v2 <- vcov_tee(ssmod2)

  expect_true(all.equal(v1, v2))

})

test_that(paste("model-based SE's cluster units of assignment in small blocks",
                "at uoa level"), {
  specdata <- data.frame(bid = rep(c(1, 2), each = 20),
                        uoa_id = c(rep(c(1, 2), each = 10), rep(seq(3, 6), each = 5)),
                        a = c(rep(c(0, 1), each = 10), rep(rep(c(0, 1), each = 5), 2)))
  specdata$y <- rnorm(40)
  spec <- rct_spec(a ~ unitid(uoa_id) + block(bid), specdata)
  suppressMessages(mod <- lmitt(y ~ 1, specification = spec, data = specdata))
  vc_w_small_block_clusters <- vcov_tee(mod)
  vc_w_no_small_block_clusters <- .vcov_CR(mod,
                                           cluster = .make_uoa_ids(mod, "DB"),
                                           by = "uoa_id")
  expect_true(vc_w_small_block_clusters[2, 2] == vc_w_no_small_block_clusters[2, 2])
})

test_that("#177 vcov with by argument", {
  set.seed(23)
  cmod_data <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_data)
  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)
  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat, offset = cov_adj(cmod))
  expect_error(vcov(mod), "not uniquely specified. Provide a `by` argument")
  expect_silent(vcov_tee(mod, by = "by_col"))
})

test_that("#177 vcov with by", {
  set.seed(23)
  cmod_data <- data.frame(yr = rep(c("00", "01", "02"), 5),
                          id = rep(letters[1:5], each = 3),
                          x = rnorm(5 * 3),
                          y = rnorm(5 * 3),
                          by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_data)
  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)
  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                             yr = rep(c("01", "02"), 5),
                             x = rnorm(10),
                             a = rep(c(rep(1, 3), rep(0, 2)), 2),
                             y = rnorm(10),
                             by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat, offset = cov_adj(cmod, by = "by_col"))
  expect_equal(vcov(mod), vcov(mod, by = "by_col"))
})

test_that("vcov_tee does not error when asking for specification-based SE for model
          with covariance adjustment but without absorbed block effects",{
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod_off <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                     offset = cov_adj(cmod))
  expect_true(!is.na(vcov_tee(ssmod_off, type = "DB0")))
})

test_that("vcov_tee errors when asking for specification-based SE for teeMod with
           external sample for covariance adjustment",{
  # Borrowing code from the #177 test
  set.seed(23)
  cmod_data <- data.frame(yr = rep(c("00", "01", "02"), 5),
                         id = rep(letters[1:5], each = 3),
                         x = rnorm(5 * 3),
                         y = rnorm(5 * 3),
                         by_col = seq_len(15))
  cmod <- lm(y ~ x, cmod_data)
  specdat <- data.frame(id = letters[1:5], a = c(rep(1, 3), rep(0, 2)))
  newspec <- rct_spec(a ~ unitid(id), specdat)
  analysis_dat <- data.frame(id = rep(letters[1:5], each = 2),
                            yr = rep(c("01", "02"), 5),
                            x = rnorm(10),
                            a = rep(c(rep(1, 3), rep(0, 2)), 2),
                            y = rnorm(10),
                            by_col = setdiff(seq_len(15), seq(1, 15, 3)))
  mod <- lmitt(y ~ yr, specification = newspec, data = analysis_dat,
               offset = cov_adj(cmod, by = "by_col"))
  expect_error(
    vcov_tee(mod, type = "DB0"),
    "teeMod models with external sample"
  )
})

test_that("vcov_tee errors when asking for specification-based SE for a non-RCT specification",{
  data(simdata)
  spec <- obs_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  ssmod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec))
  expect_error(
    vcov_tee(ssmod, type = "DB0"),
    "only for RCT specifications"
  )
})

test_that("vcov_tee errors when asking for specification-based SE for a model
          with moderators",{
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  ssmod_sub <- lmitt(y ~ dose, specification = spec, data = simdata, weights = ate(spec))
  expect_error(
    vcov_tee(ssmod_sub, type = "DB0"),
    "are not supported for teeMod models with moderators"
  )
})

test_that("vcov_tee throws a warning when asking for specification-based SE without IPW",{
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  teemod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ett())
  teemod_abs <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ett(spec),
                      absorb = TRUE)
  expect_warning(
    vcov_tee(teemod, type = "DB0"),
    "inverse probability weights"
  )
  expect_silent(vcov_tee(teemod_abs, type = "DB0"))
})

test_that(".merge_block_id_cols correctly combines multiple block columns", {
  df1 <- data.frame(
    block1 = c("A", "A", "B", "B", "B", "B"),
    block2 = c(1, 1, 1, 1, 2, 2)
  )
  df2 <- data.frame(
    block1 = c("A", "A", "B", "B", "B", "B"),
    block2 = c(T, F, T, F, T, F)
  )
  df3 <- data.frame(
    block1 = c("A", "A", "B", "B", "C", "C"),
    block2 = c(T, F, T, T, T, T),
    block3 = c(2, 2, 3, 3, 4, 4)
  )
  expect_equal(
    propertee:::.merge_block_id_cols(df=df1, ids=c("block1", "block2"))$block1,
    c(1, 1, 2, 2, 3, 3)
  )
  expect_equal(
    propertee:::.merge_block_id_cols(df=df1, ids=c("block1"))$block1,
    c(1, 1, 2, 2, 2, 2)
  )
  expect_equal(
    propertee:::.merge_block_id_cols(df=df2, ids=c("block1", "block2"))$block1,
    c(2, 1, 4, 3, 4, 3)
  )
  expect_equal(
    .merge_block_id_cols(df=df3, ids=c("block1", "block2", "block3"))$block1,
    c(2, 1, 3, 3, 4, 4)
  )
})

test_that(".get_DB_wo_covadj_se returns correct value for specifications with small blocks",{
  # generate data
  nbs <- rep(2, 10)
  n <- sum(nbs) # sample size
  B <- length(nbs) # number of blocks
  ws <- round(rnorm(n=n, mean=50, sd=10))
  yobs <- rnorm(n=n) # observed y's
  zobs <- c() # treatment assignment, 1 or 2
  for (b in 1:B){
    zobs <- c(zobs, sample(c(1,2)))
  }

  # calculate variance
  pi_all <- matrix(rep(1/2,2*n), nrow=n) # assignment probabilities, n by 2
  nbk_all <- matrix(rep(1,2*n), nrow=n) # nbk for all units, n by 2
  gammas <- cbind(nbk_all[,1] * ws, nbk_all[,2] * ws) # gamma, n by K
  nu <- c() # nu_b

  for (k in 1:2){
    indk <- (zobs == k)
    thetak <- sum(ws[indk] * yobs[indk] / pi_all[indk, k]) /
      sum(ws[indk] / pi_all[indk, k]) # Hajek estimators
    gammas[indk,k] <- gammas[indk,k] / pi_all[indk, k] * (yobs[indk] - thetak)
    for (b in 1:B){
      in_b <- (sum(nbs[1:b])-nbs[b]+1):(sum(nbs[1:b])) # indices of block b
      indbk <- (zobs[in_b] == k)
      if (k > 1){
        indb0 <- (zobs[in_b] == 1)
        nu <- c(nu, (gammas[in_b,k][indbk] - gammas[in_b,1][indb0])^2)
      }
    }
  }
  data <- data.frame(cid = 1:n, bid = rep(1:B, each=2), y = yobs, z = zobs-1, w = ws)
  spec <- rct_spec(z ~ cluster(cid) + block(bid), data)
  ssmod <- lmitt(y ~ 1, specification = spec, data = data, weights = ate(spec) * data$w)
  expect_equal(vcov_tee(ssmod, type="DB0")[1,1], sum(nu) / sum(ws)^2)
})

test_that(".get_DB_wo_covadj_se returns correct value for specifications with a few large blocks",{
  # generate data
  nbs <- rep(10, 2)
  n <- sum(nbs) # sample size
  B <- length(nbs) # number of blocks
  ws <- round(rnorm(n=n, mean=50, sd=10))
  yobs <- rnorm(n=n) # observed y's
  zobs <- rep(1, n) # treatment assignment, 1 or 2
  zobs[c(sample(1:10, 5), sample(11:20, 4))] <- 2

  # calculate variance
  pi_all <- matrix(c(rep(1/2,n), rep(c(0.6,0.4),10)), nrow=n, byrow = TRUE)
  # assignment probabilities, n by 2
  nbk <- matrix(c(5,6,5,4), nrow=B)  # nbk, B by 2
  gammas <- cbind(c(rep(5,10),rep(6,10)) * ws,
                  c(rep(5,10),rep(4,10)) * ws) # gamma, n by K
  gamsbk <- matrix(nrow=B, ncol=2)  # s^2 b,j
  nu <- c()

  for (k in 1:2){
    indk <- (zobs == k)
    thetak <- sum(ws[indk] * yobs[indk] / pi_all[indk, k]) /
      sum(ws[indk] / pi_all[indk, k])  # Hajek estimator
    gammas[indk,k] <- gammas[indk,k] / pi_all[indk, k] * (yobs[indk] - thetak)
    for (b in 1:B){
      in_b <- (sum(nbs[1:b])-nbs[b]+1):(sum(nbs[1:b])) # indices of block b
      gamsbk[b,k] <- var(gammas[in_b,k][zobs[in_b] == k])
      if (k > 1){
        nu <- c(nu, gamsbk[b,1] / nbk[b,1] + gamsbk[b,k] / nbk[b,k])
      }
    }
  }
  data <- data.frame(cid = 1:n, bid = rep(1:B, each=10), y = yobs, z = zobs-1, w = ws)
  spec <- rct_spec(z ~ cluster(cid) + block(bid), data)
  ssmod <- lmitt(y ~ 1, specification = spec, data = data, weights = ate(spec) * data$w)

  #ainv <- .get_DB_a_inverse(ssmod)
  #meat <- .get_DB_meat(ssmod)
  #vmat <- as.matrix((ainv %*% meat %*% t(ainv))[3,3])

  expect_equal(vcov_tee(ssmod, type="DB0")[1,1], sum(nu) / sum(ws)^2)
  #expect_true(vmat[1,1] != sum(nu) / sum(ws)^2)
})

test_that("specification-based SE for tee models without absorption does not crash", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  teemod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate())
  teemod_off <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                     offset = cov_adj(cmod))
  expect_silent(vcov_tee(teemod, type = "DB0"))
  expect_silent(vcov_tee(teemod_off, type = "DB0"))

  spec <- rct_spec(z ~ cluster(uoa1, uoa2), simdata)
  teemod <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate())
  teemod_off <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                      offset = cov_adj(cmod))
  expect_silent(vcov_tee(teemod, type = "DB0"))
  expect_silent(vcov_tee(teemod_off, type = "DB0"))
})

test_that("specification-based SE for tee models with absorption does not crash", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  teemod_abs <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                      absorb = TRUE)
  teemod_abs_off <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                          offset = cov_adj(cmod), absorb = TRUE)
  expect_silent(vcov_tee(teemod_abs, type = "DB0"))
  expect_silent(vcov_tee(teemod_abs_off, type = "DB0"))
  
  ssmod_sub_abs <- lmitt(y ~ dose, specification = spec, data = simdata, absorb = TRUE)
  ssmod_sub_abs_off <- lmitt(y ~ dose, specification = spec, data = simdata,
                             offset = cov_adj(cmod), absorb = TRUE)
  expect_silent(vcov_tee(ssmod_sub_abs, type = "DB0"))
  expect_silent(vcov_tee(ssmod_sub_abs_off, type = "DB0"))
  expect_silent(vcov_tee(ssmod_sub_abs, type = "DB0", const_effect = TRUE))
  expect_silent(vcov_tee(ssmod_sub_abs_off, type = "DB0", const_effect = TRUE))
})

test_that(".get_appinv_atp returns correct (A_{pp}^{-1} A_{tau p}^T)
          for tee models with absorbed intercept", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod_abs <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                     absorb = TRUE)
  ssmod_abs_off <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                         offset = cov_adj(cmod), absorb = TRUE)

  bid <- simdata$bid
  B <- cbind(as.integer(bid == 1), as.integer(bid == 2), as.integer(bid == 3))
  goal <- matrix(0, nrow = 3, ncol = 1)
  for (blk in 1:3){
    goal[blk] <- sum(weights(ssmod_abs) * residuals(ssmod_abs) * B[,blk]) /
      sum(ssmod_abs$weights * B[,blk])
  }
  expect_true(all.equal(goal, .get_appinv_atp(ssmod_abs, db = TRUE)))

  for (blk in 1:3){
    goal[blk] <- sum(weights(ssmod_abs_off) * residuals(ssmod_abs_off) * B[,blk]) /
      sum(weights(ssmod_abs_off)* B[,blk])
  }
  expect_true(all.equal(goal, propertee:::.get_appinv_atp(ssmod_abs_off, db = TRUE)))
})

test_that(".get_phi_tilde returns correct grave{phi}
          for tee models with absorbed intercept", {
  data(simdata)
  spec <- rct_spec(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  cmod <- lm(y ~ x, simdata)
  ssmod_abs <- lmitt(y ~ 1, specification = spec, data = simdata, weights = ate(spec),
                     absorb = TRUE)

  bid <- simdata$bid
  B <- cbind(as.integer(bid == 1), as.integer(bid == 2), as.integer(bid == 3))
  Z <- cbind(as.integer(simdata$z == 0), as.integer(simdata$z == 1))

  ws <- stats::weights(ssmod_abs)
  p <- matrix(0, nrow = 2, ncol = 3)
  for (k in 1:2)
    for (j in 1:3){
      p[k, j] <- sum(ws * Z[,k] * B[,j]) / sum(ws * B[,j])
    }
  goal <- matrix(0, nrow = 50, ncol = 3)
  for (s in 1:3){
    goal[,s] <- ws * residuals(ssmod_abs) * (Z[,2] - p[2,s]) * B[,s]
  }
  expect_true(all.equal(goal, propertee:::.get_phi_tilde(ssmod_abs, db = TRUE)))
  
  ssmod_abs <- lmitt(y ~ 1, specification = spec, data = simdata,
                     absorb = TRUE)
  p <- matrix(0, nrow = 2, ncol = 3)
  for (k in 1:2)
    for (j in 1:3){
      p[k, j] <- sum(Z[,k] * B[,j]) / sum(B[,j])
    }
  goal <- matrix(0, nrow = 50, ncol = 3)
  for (s in 1:3){
    goal[,s] <- residuals(ssmod_abs) * (Z[,2] - p[2,s]) * B[,s]
  }
  expect_true(all.equal(goal, propertee:::.get_phi_tilde(ssmod_abs, db = TRUE)))
})

test_that("block_center_residuals correctly subtracts off block center residuals", {
  data(simdata)
  des <- rct_specification(z ~ cluster(uoa1, uoa2) + block(bid), simdata)
  damod_abs <- lmitt(y ~ 1, specification = des, data = simdata, absorb = TRUE)
  r <- residuals(damod_abs)
  
  # calculate block center residuals
  block_id <- as.factor(simdata$bid)
  model <- lm(r ~ block_id)
  intercept <- coef(model)[1]
  block_effects <- coef(model)[-1]
  block_means <- intercept + c(0, block_effects)
  names(block_means) <- levels(block_id)
  
  # subtract off block means
  block_means <- block_means[block_id]
  r <- r - block_means
  expect_true(all.equal(r, residuals(propertee:::block_center_residuals(damod_abs))))
})

test_that("block_center_residuals works for input containing NA block IDs", {
  data(simdata)
  simdata[1:3, 'bid'] <- NA
  des <- rct_specification(z ~ cluster(uoa1, uoa2) + block(bid), 
                           simdata, na.fail = FALSE)
  damod_abs <- lmitt(y ~ 1, specification = des, data = simdata, absorb = TRUE)
  r <- residuals(damod_abs)
  
  # calculate block center residuals
  block_id <- as.factor(simdata$bid)
  model <- lm(r ~ block_id)
  intercept <- coef(model)[1]
  block_effects <- coef(model)[-1]
  block_means <- intercept + c(0, block_effects)
  names(block_means) <- levels(block_id)
  
  # subtract off block means
  block_means <- block_means[block_id]
  r <- r - block_means
  expect_true(all.equal(r, residuals(propertee:::block_center_residuals(damod_abs))))
})

test_that("block_center_residuals works for teeMod having NA in fitted values", {
  data(simdata)
  simdata[1:3, 'x'] <- NA
  des <- rct_specification(z ~ cluster(uoa1, uoa2) + block(bid), 
                           simdata, na.fail = FALSE)
  cmod <- lm(y ~ x, simdata)
  teemod <- suppressWarnings(lmitt(
    y ~ 1, specification = des, data = simdata, absorb = TRUE, offset = cov_adj(cmod)
    ))
  r <- residuals(teemod)
  
  # calculate block center residuals
  block_id <- as.factor(simdata$bid)[4:50]
  model <- lm(r ~ block_id)
  intercept <- coef(model)[1]
  block_effects <- coef(model)[-1]
  block_means <- intercept + c(0, block_effects)
  names(block_means) <- levels(block_id)
  
  # subtract off block means
  block_means <- block_means[block_id]
  r <- r - block_means
  expect_true(all.equal(r, residuals(propertee:::block_center_residuals(teemod))))
})

test_that("issue #239", {
  # crux of the issue is that when all observations with a given value of a
  # moderator are given a weight of 0 in the regression, the lm/mlm for
  # computing means in the control condition has NA coefficients. the .get_a21()
  # and custom bread.mlm() functions did not handle this as they needed to.
  # here we make sure they only retain columns that yield full-rank matrices
  set.seed(298)
  odata <- data.frame(x1 = factor(rep(seq_len(3), each = 5)),
                      y = rnorm(15),
                      a = rep(c(0, 1), 8)[1:15],
                      uid = seq_len(15))
  cmod <- lm(y~x1, odata)
  spec <- rct_spec(a ~ unitid(uid), odata)
  tmod <- lmitt(y ~ x1, spec, odata, weights = c(rep(0, 5), rep(1, 10)),
                offset = cov_adj(cmod))
  
  # test bread. if there's an offset and the rank of the ctrl means model is K,
  # then there should be K columns corresponding to the regression for the
  # response and K columns corresponding to the regression for the offset for a
  # total of 2 * K columns. when we add those to the bread matrix for the
  # effect estimation regression, which we'll say has rank R, then the number
  # of columns in the bread matrix will be R + 2 * K.
  twoK <- tmod@ctrl_means_model$rank * 2
  R <- tmod$rank
  br_cm <- bread.mlm(tmod@ctrl_means_model)
  expect_equal(ncol(br_cm), twoK)
  br <- bread(tmod)
  expect_equal(ncol(br), R + twoK)
  
  # test A21. should have R + 2 * K rows. so when we use it to make the output
  # of estfun.teeMod(), we should have a matrix with R + 2 * K columns
  a21 <- .get_a21(tmod)
  expect_equal(nrow(a21), R + twoK)
  ef <- estfun(tmod)
  expect_equal(ncol(ef), R + twoK)
})

##' Group-center akin to Stata's \code{areg}
##'
##' From the Stata documentation: \code{areg} begins by recalculating Y and X
##' and to have mean 0 within the groups specified by \code{absorb()}. The
##' overall mean of each variable is then added back in.
##' @param mm Matrix of variables to center
##' @param grp Group to center on
##' @param wts Optional weights
##' @param grand_mean_center Optional center output at \code{mean(var)}
##' @return Vector of group-centered values
##' @keywords internal
areg.center <- function(mm, grp, wts = NULL, grand_mean_center = FALSE) {
  if (is.null(wts)) {
    wts <- rep(1, nrow(mm))
  }
  mm2 <- mm[!is.na(wts),,drop=FALSE]
  grp_keep <- !is.na(grp)
  group_means <- sweep(rowsum((wts * mm2)[grp_keep,], grp[grp_keep], na.rm = TRUE),
                       1, rowsum(wts[grp_keep], grp[grp_keep], na.rm = TRUE), FUN = "/")
  group_means[is.nan(group_means)] <- 0
  out <- mm2 - group_means[grp,]
  if (grand_mean_center) {
    grand_means <- colSums(wts * mm2, na.rm = TRUE) / sum(wts)
    out <- sweep(out, 2, grand_means, FUN = "+")
  }
  return(out)
}

test_that("Basic functionality, grand_mean_center=TRUE", {

  mm <- matrix(rep(seq(2), 6), ncol = 1)

  # Returns the same dimension
  expect_identical(dim(mm),
                   dim(center_mm <- areg.center(mm, rep(1, nrow(mm)),
                                                grand_mean_center = TRUE)))

  # In a single block, return original
  expect_true(all.equal(mm, center_mm))

  # correct values for one column with multiple groups but no weights
  grps <- rep(seq(3), each = 4)
  expect_equal(areg.center(mm, grps, grand_mean_center = TRUE)[,1],
               rep(c(-0.5 + 1.5, 0.5 + 1.5), 6))

  # correct values for one column with multiple groups and weights
  wts <- rep(seq(2), 6)
  expect_equal(areg.center(mm, grps, wts, grand_mean_center = TRUE)[,1],
               rep(c(-2/3 + 5/3, 1/3 + 5/3), 6))
  
  # correct values for multiple columns with multiple groups and weights
  mm <- cbind(mm, mm)
  expect_identical(dim(mm), dim(center_mm <- areg.center(mm, grps, wts, grand_mean_center = TRUE)))
  expect_true(
    all.equal(center_mm,
              cbind(rep(c(-2/3 + 5/3, 1/3 + 5/3), 6),
                    rep(c(-2/3 + 5/3, 1/3 + 5/3), 6)),
              check.attributes = FALSE)
  )
  
  # correct values when a group has 0 weights for all units
  wts <- c(rep(0, 4), rep(seq(2), 4))
  expect_true(
    all.equal(areg.center(mm, grps, wts, grand_mean_center = TRUE),
              cbind(c(rep(c(1 + 5/3, 2 + 5/3), 2), rep(c(-2/3 + 5/3, 1/3 + 5/3), 4)),
                    c(rep(c(1 + 5/3, 2 + 5/3), 2), rep(c(-2/3 + 5/3, 1/3 + 5/3), 4))),
              check.attributes = FALSE)
  )
})

test_that("Basic functionality, grand_mean_center=FALSE", {

  mm <- matrix(rep(seq(2), 6), ncol = 1)
  
  # Returns the same dimension
  expect_identical(dim(mm),
                   dim(center_mm <- areg.center(mm, rep(1, nrow(mm)))))
  
  # In a single block, return original with grand mean subtracted out
  expect_true(all.equal(mm - mean(mm), center_mm))
  
  # correct values for one column with multiple groups but no weights
  grps <- rep(seq(3), each = 4)
  expect_equal(areg.center(mm, grps)[,1],
               rep(c(-0.5, 0.5), 6))
  
  # correct values for one column with multiple groups and weights
  wts <- rep(seq(2), 6)
  expect_equal(areg.center(mm, grps, wts)[,1],
               rep(c(-2/3, 1/3), 6))
  
  # correct values for multiple columns with multiple groups and weights
  mm <- cbind(mm, mm)
  expect_identical(dim(mm),
                   dim(center_mm <- areg.center(mm, grps, wts)))
  expect_true(
    all.equal(center_mm,
              cbind(rep(c(-2/3, 1/3), 6), rep(c(-2/3, 1/3), 6)),
              check.attributes = FALSE)
  )
  
  # correct values when a group has 0 weights for all units
  wts <- c(rep(0, 4), rep(seq(2), 4))
  expect_true(
    all.equal(areg.center(mm, grps, wts),
              cbind(c(rep(c(1, 2), 2), rep(c(-2/3, 1/3), 4)),
                    c(rep(c(1, 2), 2), rep(c(-2/3, 1/3), 4))),
              check.attributes = FALSE)
  )
})

if (requireNamespace("robustbase", quietly = TRUE)) {

  test_that("basic tests just to touch glmrobMethods.R file", {
    data(simdata)

    cmod <- robustbase::glmrob(z ~ x, simdata, family = binomial)
    expect_true(is(estfun(cmod), "matrix"))
    expect_true(is(bread(cmod), "matrix"))
    cmod <- robustbase::glmrob(bid ~ x, simdata, family = poisson)
    expect_true(is(estfun(cmod), "matrix"))
    expect_true(is(bread(cmod), "matrix"))
    cmod <- robustbase::glmrob(y ~ x, simdata, family = gaussian)
    expect_true(is(estfun(cmod), "matrix"))
    expect_true(is(bread(cmod), "matrix"))
    cmod <- robustbase::glmrob(force ~ x, simdata, family = Gamma)
    expect_true(is(estfun(cmod), "matrix"))
    expect_true(is(bread(cmod), "matrix"))
  })
}

if (requireNamespace("robustbase", quietly = TRUE)) {

  test_that("invalid bread. and estfun.lmrob calls produce expected errors", {
    data(simdata)

    cmod <- robustbase::lmrob(y ~ x, simdata)
    cmod$control$method <- "invalid_method"
    expect_error(estfun(cmod), "supports only SM or MM")
    expect_error(bread(cmod), "supports only SM or MM")

    cmod$control$method <- "SM"
    cmod$control$psi <- NULL
    expect_error(estfun(cmod), "psi is not defined")
    expect_error(bread(cmod), "psi is not defined")

    cmod$control <- NULL
    expect_error(estfun(cmod), "object must have a `control` element")
    expect_error(bread(cmod), "object must have a `control` element")
  })

  test_that("estfun.lmrob produces expected output", {
    data(simdata)

    cmod <- robustbase::lmrob(y ~ x, simdata)
    ef <- estfun(cmod)

    expect_equal(dim(ef), c(nrow(simdata), 3))
    expect_equal(colnames(ef), c("sigma", colnames(stats::model.matrix(cmod))))
    expect_true(is.null(attr(ef, "assign")))
    expect_true(is.null(attr(ef, "contrasts")))
  })

  test_that("bread.lmrob produces expected output", {
    data(simdata)

    cmod <- robustbase::lmrob(y ~ x, simdata)
    ainv <- bread(cmod)

    expect_equal(dim(ainv), c(2, 3))
    expect_equal(dimnames(ainv),
                 list(colnames(stats::model.matrix(cmod)),
                      c("sigma", colnames(stats::model.matrix(cmod)))))
  })

}

if (requireNamespace("robustbase", quietly = TRUE)) {
  test_that("the columns of estfun.glmrob have approximately 0 mean",{
    data(lsoSynth)

    mod <- robustbase::glmrob(probation_year1~age_at_entry+lhsgrade_pct+male,
                  family=binomial,data=lsoSynth)


    ee <- estfun(mod)

    expect_true(all(abs(colMeans(ee))<0.001))
  })

  test_that("the sandwich vcov matrix for a glmrob object is approximately the same as the cov returned with the object",{
    data(lsoSynth)

    mod <- robustbase::glmrob(probation_year1~age_at_entry+lhsgrade_pct+male,
                  family=binomial,data=lsoSynth)

    expect_true(max(abs(sandwich::sandwich(mod)-mod$cov))<0.01)
  })
} else {
  expect_true(TRUE)
  # Dummy test to avoid empty test warnings
}

---
title: "Non-binary Treatment Specification"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Non-binary Treatment Specification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(propertee)
```

## Binary Treatment

When creating a StudySpecification, handling binary treatment variables is straightforward.
If the treatment variable is either `numeric` with only values 0/1, or is
`logical`, then `lmitt()` will estimate a treatment effect of the difference
between the outcome in the treated group (`1` or `TRUE`) versus the control
group (`0` or `FALSE`).

### Missing treatment status

In all cases (binary and non-binary), missing values are allowed and any units
of assignment with missing treatment values are excluded from models fit via
`lmitt()`.

## Non-binary Treatment

However, the [`_spec()`](../reference/StudySpecification_objects.html) functions can take
in any (reasonable) form of treatment assignment.

If the treatment variable is a `numeric` with non-binary values, it is treated
as a continuous treatment effect and `lmitt(y ~ 1, ...` will estimate a single
coefficient on treatment.

If the treatment variable is a `character`, it is treated as a multi-level
treatment variable and `lmitt(y ~ 1, ...` will estimate treatment effects
against a reference category. The reference category is the first level defined
according to [R's comparison of characters](https://rdrr.io/r/base/Comparison.html).

`factor` and `ordered` objects are tricky to deal with, so while a `StudySpecification` can
be created with `factor` or `ordered` treatment variables, `lmitt()` will refuse
to estimate a model unless it is also provided a
[`dichotomy`](#dichotomzing-a-non-binary-treatment) (see below).

### Dichotomzing a Non-binary Treatment

Studies may offer treatment to units at different times or provide treatment to
units in varying intensities. Researchers may be interested in estimating treatment
effects at different times or given a certain threshold of provided treatment, however.
`propertee` accommodates these wishes by storing the time or intensity of treatment
for treated units in the  [`StudySpecification`](../reference/StudySpecification_objects.html), then offering
a `dichotomy=` argument to the weights calculation functions `ett()`/`ate()` and the
assginment creation function `assigned()` A `dichotomy` is presented as a
formula, where the left-hand side is a logical statement defining inclusion in
the treatment group, and the right-hand side is a logical statement defining
inclusion in the control group. For example, if `dose` represents the intensity
of a given treatment, we could set a threshold of 200, say, mg:

```{r, eval = FALSE}
dose > 200 ~ dose <= 200
```

All units of assignment with `dose` above 200 are treated units,
and all units of assignment with `dose` of 200 or below are control units.

A `.` can be used to define either group as the inverse of the other. For
example, the above dichotomy could be defined as either of

```{r, eval = FALSE}
dose > 200 ~ .
. ~ dose <= 200
```

Any units of assignment not assigned to either treatment or control are assumed
to have `NA` for a treatment status and will be ignored in the estimation of
treatment effects.

```{r, eval = FALSE}
dose >= 300 ~ dose <= 100
```

In this `dichotomy`, units of assignment in the range (100,300) are ignored.

### An Example

```{r}
data(simdata)
table(simdata$dose)
spec1 <- rct_spec(dose ~ uoa(uoa1, uoa2), data = simdata)
summary(spec1)
```

```{r}
head(ate(spec1, data = simdata, dichotomy = dose >= 300 ~ dose <= 100))
```

```{r}
head(assigned(spec1, data = simdata, dichotomy = dose >= 300 ~ dose <= 100))
```

---
title: "Introduction to propertee"
author: "propertee Authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to propertee}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#devtools::load_all("~/repositories/_r/propertee/")
library(propertee)
```

## Main Features

The **propertee** package, Prognostic Regression Offsets with
Propagation of ERrors (for Treatment Effect Estimation), facilitates
direct adjustment or standardization for experiments and observational
studies, with the option of using a separately fitted prognostic
regression model for covariance adjustment. Propertee calls for a
self-standing specification of a study's design and treatment
allocations, using these to create weights needed to align treatment
and control samples for estimation of treatment effects averaged
across a common standard population, with standard errors that
appropriately reflect the study design. For covariance adjustment it
enables offsetting the outcome against predictions from a dedicated
covariance model, with standard error calculations propagating error
as appropriate from the covariance model.

The main workflow consists of two main steps and one optional step:

1. Encode the study's design, including the unit of assignment,
   treatment status of each unit of assignment, and block membership
   information as appropriate, in a `StudySpecification` object. This
   is accomplished with the `obs_spec()`, `rct_spec()` or `rd_spec()`
   functions.
2. Optionally, fit a covariate adjustment model.  This is done with
   any of a number of modeling functions external to `propertee`, for
   example `lm()` or `glm()` from the `stats` package, or `lmrob` from
   the `robustbase` package.
3. Use `lm()` to contrast the separately reweighted treatment and
   control groups, incorporating optional covariance adjustments as an
   `offset` to `lm()`. The `StudySpecification` is unobtrusively
   retained, as is error propagation from the optional covariance
   adjustment, for use in subsequent standard error calculations.
   This is done using the `lmitt()` function, which either takes and
   enriches a fitted `lm` or, in the more fully supported usage,
   internally calls `lm()` to itself create a directly adjusted
   contrast of treatment conditions.

## Example Data

The example dataset comes from the state of Tennessee's Student-Teacher
Achievement Ratio (STAR) experiment. Students were randomly assigned to three
possible classroom conditions: small (13 to 17 students per teacher), regular
class (22 to 25 students per teacher), and regular-with-aide class (22 to 25
students with a full-time teacher's aide).

```{r}
data("STARplus")
table(STARplus$cond_at_entry)
```

For simplicity for this first example, we will examine a single binary
treatment - "small" classrooms versus "regular" and "regular+aide" classrooms.

```{r}
STARplus$cond_small <- STARplus$cond_at_entry == "small"
table(STARplus$cond_small)
```

After this basic example, we will see how **propertee** makes it easy to handle
non-binary treatment variables by introducing `dichotomy`s.

The outcome of interest is a reading score at the end of kindergarten.

```{r}
summary(STARplus$g1treadss)
```



We first take the random assignment scheme to have been either complete or
Bernoulli randomization of students, separately by school. (Later we'll change this to the more realistic assumption of complete or Bernoulli randomization, separately by school *and* year of entry into the study.) Accordingly, experimental blocks are given by the `school_at_entry` variable.

```{r}
length(unique(STARplus$school_at_entry))
head(table(STARplus$school_at_entry))
```

We need a unique identifier for the unit by which treatment was allocated.  STAR treatments were assigned to students, as opposed to classrooms or other aggregates of students; accordingly we need a unique identifier per student. The  `stdntid` variable fills this role.

```{r}
length(unique(STARplus$stdntid))
head(STARplus$stdntid)
```

## A Basic Example

### Defining the `StudySpecification`

The three `_spec` functions (`rct_spec()`, `obs_spec()`, and `rd_spec()`)
operate similarly. The first argument is the most important, and encodes all the
specification information through the use of a formula. The left-hand side of
the formula identifies the treatment variable. The right-hand side of the
formula consists of the following potential pieces of information:

1. `unit_of_assignment()`: This identifies the variable(s) which indicate the
   units of assignment. This is required for all specifications. The alias
   `uoa()` can be used in its place.
2. `block()`: The identifies the variable(s) which contain block information.
   Optional.
3. `forcing()`: In regression discontinuity specifications (`rd_spec()`), this
   identifies the variable(s) which contain forcing information.

To define a `StudySpecification` in our example:

```{r}
spec <- obs_spec(cond_small ~ unit_of_assignment(stdntid) + block(school_at_entry),
                  data = STARplus, na.fail = FALSE)
summary(spec)
```

Should additional variables be needed to identify the unit of assignment,
block, or forcing, they can be included. Had the variable identifying the years in which participants entered the study been available, for example, we might have identified the experimental blocks as combinations of school and year of study entry, rather than as the school alone.  To do this we would pass `block(year_at_entry, school_at_entry)`, rather than `block(school_at_entry)`, in the above.

### Estimating the treatment effect

The main function for estimating treatment effects is the `lmitt()` function. It
takes in three main required arguments:

1. A formula specifying the outcome and the desired treatment effect.
2. The data set containing the outcome information.
3. A [`StudySpecification`](#defining-the-specification).

Note that the data set does **not** need to be the same data set which generated
the `StudySpecification`; it does however need to include the same variables to
identify the units of assignment. (If the variable names differ, the `by=`
argument can be used to link them, though we recommend renaming to reduce the
likelihood of issues.)

For example, you may have one dataset containing school-level information, and a
separate dataset containing student-level information. Assume school is the unit
of assignment. While you could of course merge those two data-sets,
**propertee** can instead use the school-level data to define the
`StudySpecification`, and the student-level data to estimate the treatment
effect.

The formula entering `lmitt()` can take on one of two forms:

```r
y ~ 1
```

will estimate the main treatment effect on outcome variable `y`, and

```r
y ~ x
```

will estimate subgroup specific treatment effects for each level of `x` for the
outcome `y`, if `x` is categorical. For continuous `x`, a main effect and a
treatment-`x` interaction effect is estimated.

Therefore, to estimate the treatment effect in our example, we can run:

```{r}
te <- lmitt(g1treadss ~ 1, data = STARplus, specification = spec)
summary(te)
```

The data includes ethnicity; we can estimate subgroup effects by ethnicity:

```{r}
te_s <- lmitt(g1treadss ~ race, data = STARplus, specification = spec)
summary(te_s)
```

### Including specification weights

Study specification weights can be easily included in this estimation.
**propertee** supports average treatment effect (ATE) and effect of the
treatment on the treated (ETT) weights.

To include one of the weights, simply include the `weights = "ate"` or
`weights = "ett"` argument to `lmitt()`:

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, weights = "ate")
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, weights = "ett")
```

Internally, these call the `ate()` or `ett()` functions which can be used
directly.

```{r}
head(ate(spec, data = STARplus))
```

When included inside `lmitt()`, you do not need to specify any additional
arguments to `ate()` or `ett()`, enabling easy functions of weights. For example
if you had some other weight variable, say `wgt`, you could include `weights =
wgt*ate()` in the `lmitt()` call.

### Covariance Adjustment models

By itself, `lmitt()` does not allow for other covariates; e.g. something like
`lmitt(y ~ 1 + control_var,...` will fail. To adjust for covariates, a separate
covariate model should be fit. Any model which supports a `predict()` function
should work.

```{r}
camod <- lm(g1treadss ~ gender * dob + race, data = STARplus)
```

The `cov_adj()` function can be used to process the covariance adjustment model
and produce the required values; and its output can be passed as an `offset=`.

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec,
      weights = "ate", offset = cov_adj(camod))
```

Similarly to the weight functions, `cov_adj()` attempts to locate the correct
arguments (in this case, mainly the `data=` argument) to use in the model
command; while `cov_adj()` does fall back to using the data which is in the
covariance model, its safer to use the `newdata=` argument if calling
`cov_adj()` outside of the model.

```{r}
head(cov_adj(camod, newdata = STARplus))
```

Also, similarly to weights, `cov_adj()` can be used in normal modeling commands
as well.

```{r}
lm(g1treadss ~ cond_small, data = STARplus, weights = ate(spec),
   offset = cov_adj(camod))
```

### Absorbing Blocks

If fixed effects for blocks are desired, which can be absorbed away to avoid
estimating, the `absorb=TRUE` argument can be passed.

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, absorb = TRUE)
```

---
title: "OLS effect and error estimation with an auxiliary sample and a separate, not-necessarily-linear covariance model"
subtitle: "Real-data demonstration with a finely stratified cluster RCT and a broader administrative database"
author: "Josh Wasserman, Ben B. Hansen"
date: "2025-06-27"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```


## Overview
The `propertee` package offers tools for enhancing evaluations of treatments, policies, and interventions that respect the statistical properties endowed by the study specification. One such offering is a routine for covariance adjustment that allows researchers to model exogenous variation in outcomes of interest using data from their study as well as available auxiliary data. `propertee` accommodates linear, generalized linear, and robust regression models, providing users flexibility in functional form, fitting procedure, and fitting sample. This vignette serves as a step-by-step walkthrough of how users can use a prior covariance  adjustment model fit to inform estimates of intervention effects--as well as associated standard errors--with the software in `propertee`.

Pane et al. (2014) mounted a large-scale cluster randomized trial in seven states to study the effectiveness of Cognitive Tutor, an online/in-person blended algebra learning program.  Their report assessed program effectiveness in terms of scores on a test administered as part of the study.  However, scores on prior and subsequent state achievement tests are available for study schools as well as others in their states and districts, and these could be used for a complementary, perhaps more precise, assessment of the program's effects. Here we demonstrate this idea using state and district data from Michigan, where the Cognitive Tutor study had a significant footprint and where rich school achievement data are available for download from state websites.

Pane and coauthors kindly shared with us the names, pre-randomization pairings and treatment assignments of their study's 14 Michigan schools, but were not at liberty to make this information public.  To create a functionally similar case study while maintaining the participating schools's anonymity, we optimally pair-matched them to schools from a large nearby county in Michigan, Oakland. The pseudo-RCT considered in this vignette replaces each Michigan Cognitive Tutor study school with the Oakland County school it was paired to, otherwise inheriting from the actual RCT salient specification characteristics, such as the composition of school pairs and triples within which randomization was conducted.

Valid analysis of an RCT calls for careful attention to such characteristics, both for selecting a compatible estimator and for correctly implementing it.  By introducing a dedicated S4 structure for such characteristics, the `StudySpecification`, along with StudySpecification-aware functions for such tasks as inverse probability weighting and effect estimation with optional stratum fixed effects, `propertee` helps the analyst stay on top of implementation details. Its `cov_adj()`, a specialized `predict()`, decouples effect estimation from covariance adjustment while continuing to track what's necessary for valid standard error estimation, significantly broadening the range of estimators that are compatible with a given `StudySpecification`.

## Data
To run this vignette, first download the necessary data. We will use school-level averages of student performance on the Michigan Merit Examination (MME) in 2014 to measure intervention effects, and we will use school-level averages of scores on the 2012 and 2013 tests as covariates in our covariance adjustment model. These scores can be downloaded as a zipped file from the [Michigan Department of Education website](https://mischooldata.org/historical-assessment-data-files/). After unzipping that file, convert the resulting .xls file to a .csv to facilitate the use of base R commands for loading it into an R session.

We also use school-level characteristics from the [Common Core of Data](https://nces.ed.gov/ccd/files.asp#Fiscal:2,LevelId:7,SchoolYearId:28,Page:1) in the covariance adjustment model. Click on the link provided here, and download the "Flat File" for the 2013-2014 Public Elementary/Secondary School Universe Survey under "Data File". We can use base R commands to load in the unzipped .txt file.

The last necessary file can be be loaded from the `propertee` package by calling `data(michigan_school_pairs)`. This dataframe tracks which schools were paired together in the study and which schools were assigned to intervention and control.

## Package Installation
This vignette requires the installation of three package in addition to `propertee`: `httr` and `readxl`, which we'll use to import the Michigan schools data; and `robustbase`, providing functionality for outlier-robust regression. The vignette uses `robustbase` to demonstrate how `propertee` can handle alternatives to ordinary least squares for covariance adjustment modeling.

## Walkthrough
After loading the installed packages and reading in the downloaded data files, we clean the MME scores and school characteristics datasets. The scores data has rows corresponding to state-, intermediate school district (ISD)-, district-, and campus-wide averages. In addition to averages taken over all students in these subpopulations, some rows correspond to averages taken within substrata formed by gender, race/ethnicity, learning ability, or economic background. In the provided cleaning script (`get_and_clean_external_data.R`), we create two cleaned datasets, one for an analysis of the marginal effect of the intervention and one for an analysis of the heterogeneity of the intervention effect. Both datasets keep only rows corresponding to schools where MME scores were reported campus-wide or for the particular substratum in each of 2012, 2013, and 2014.

The school characteristics data spans the universe of public schools in the United States, so to clean it for this vignette, we first limit it to schools relevant to the study. The MME is taken almost exclusively by 11th graders and, as the name suggests, only taken by students in Michigan, so we first subset the data to schools in Michigan serving 11th graders. Then, we perform feature generation, creating derived covariates such as demographic breakdowns by gender, race/ethnicity, and free- or reduced-price lunch eligibility at the school level and in the 11th grade specifically. (The provided cleaning script performs these steps also.)

```{r load_and_clean, message=FALSE, warning=FALSE}
if (!require("robustbase")) library(robustbase)
if (!require("readxl")) library(readxl)
if (!require("httr")) library(httr)
if (!require("propertee")) library(propertee)

extdataURLs <- list(
CCD="https://nces.ed.gov/ccd/data/zip/sc132a_txt.zip",
MME="https://www.michigan.gov/cepi/-/media/Project/Websites/cepi/MiSchoolData/historical/Historical_Assessments/2011-2014MME.zip"
)
data(michigan_school_pairs)
michigan_school_pairs <-
     subset(michigan_school_pairs, !is.na(blk), c(schoolid,blk,z))
source("get_and_clean_external_data.R")
```

#### Creating the `StudySpecification` Object
The first step in estimating intervention effects using `propertee` is to create a `StudySpecification` object. This will store the information from `michigan_school_pairs` in a way that will allow for quick calculation of inverse probability of assignment weights that attend to the pair-matched structure of the study.

In studies where units of observation (eg. individual students) within units of assignment (eg. schools or districts) are used to estimate intervention effects, the data structure should reflect this nesting. For example, if we had student-level scores data as the observed data within schools or districts, it is important to specify how these units of assignment are allocated to different treatment or control conditions. In such studies, the `StudySpecification` object would also facilitate the definition of a vector of assignment indicators at the unit of observation level we could use for estimating the intervention effect.

Should more than one variable be needed to identify the unit of assignment, block, or forcing, they can be included. For example, perhaps `schoolidk` may be unique within district, but potentially not unique across districts. Then we’d use something like `block(districtid, schoolidk)` in the `_spec` function.
```{r}
spec <- rct_spec(z ~ unitid(schoolid) + block(blk), michigan_school_pairs)
```
In this `rct_spec()` call, the lefthand side indicates the assignment variable, and the righthand side indicates the unit of assignment and, if applicable, variable that identify matched sets or strata.  There are also `rd_spec()` and `obs_spec()` constructors, for regression discontinuity and for observational studies/quasiexperimental specifications, respectively.

The `StudySpecification` object's `structure` slot lists units of assignment, their allocations to conditions and, if applicable, blocks within which they were allocated.
```{r}
spec@structure
```

#### Fitting the Covariance Adjustment Model
We now fit the covariance adjustment model. The `propertee` package will generate predictions from this regression to explain residual variation of the outcomes in the study. Often, this produces more accurate and precise effect estimates. As mentioned earlier, `propertee` accommodates a host of fitting procedures for estimating this model, and the regression may leverage data from available auxiliary sources. We demonstrate this flexibility by fitting two covariance adjustment models for each analysis, one with least squares and one with robust regression. We fit these models to a sample including all schools in the study and all schools in Oakland County whose outcomes and covariates are measured in the data we've downloaded. The exact specification for this school-level model is provided in Equation 1.

$$
\begin{align}
\text{Average_Score_14} &= \beta_{0} + \beta_{1}\text{Total_Enrollment_14} + \beta_{2}\text{Title_I_Status_14}
\\&\quad+
\beta_{3}\text{Magnet_Status_14} + \beta_{4}\text{Charter_Status_14} + \beta_{5}\text{Education_Type_14} \\&\quad +
\beta_{6}\text{Perc_Female_14} + \beta_{7}\text{Perc_Native_14} + \beta_{8}\text{Perc_Asian_14} \\&\quad +
\beta_{9}\text{Perc_Hispanic_14} + \beta_{10}\text{Perc_Black_14} + \beta_{11}\text{Perc_White_14} \\&\quad+
\beta_{12}\text{Perc_Pacific_Islander_14} + \beta_{13}\text{Perc_Econ_Disadvantaged_14} \\&\quad+
\beta_{14}\text{Perc_Female_Grade11_14} + \beta_{15}\text{Perc_Native_Grade11_14} \tag{1} \\ &\quad+
\beta_{16}\text{Perc_Asian_Grade11_14} + \beta_{17}\text{Perc_Hispanic_Grade11_14} \\ &\quad+
\beta_{18}\text{Perc_Black_Grade11_14} + \beta_{19}\text{Perc_White_Grade11_14} \\ &\quad+
\beta_{20}\text{Perc_Pacific_Islander_Grade11_14} + \beta_{21}\text{Average_Score_13} \\&\quad+
\beta_{22}\text{Average_Score_12}
\end{align}
$$

```{r}
coname <- "OAKLAND COUNTY"
RESPONSE_COL <- "Average.Scale.Score.2014"
MODELING_COLS <- c(
  "TOTAL_ENROLLMENT", setdiff(CCD_CAT_COLS, "TYPE"),
  setdiff(colnames(analysis1data)[grepl("_PERC$", colnames(analysis1data))],
          c("MALE_PERC", "TR_PERC", "MALE_G11_PERC", "TR_G11_PERC")),
  paste0("Average.Scale.Score.", c(2013, 2012))
)
```


```{r}
not_missing_resp <- !is.na(analysis1data[[RESPONSE_COL]])
not_missing_covs <- rowSums(is.na(analysis1data[, MODELING_COLS])) == 0
county_ix <- analysis1data$CONAME == coname
county_camod_dat <- analysis1data[not_missing_resp & not_missing_covs,]

camod_form <- as.formula(
  paste0(RESPONSE_COL, "~", paste(MODELING_COLS, collapse = "+")))
lm_county_camod <- lm(camod_form, county_camod_dat,
                      weights = county_camod_dat$Total.Tested.2014)
```

```{r}
set.seed(650)
rob_county_camod <- robustbase::lmrob(
  camod_form, county_camod_dat, weights = county_camod_dat$Total.Tested.2014,
  control = robustbase::lmrob.control(max.it = 500L))
```

#### Estimating Marginal Intervention Effects
With the `StudySpecification` object created and the covariance adjustment model fit, we're prepared to evaluate the intervention. `propertee` supports the calculations of inverse probability of assignment weights, which can be used to estimate either the average intervention effect (ATE) or the average effect for the treated (ETT). These weights can be combined with additional unit weights to reflect varying sizes of units in the sample. In this analysis and the one that follows, we estimate the student-level ATE by calculating inverse probability of assignment weights for each school using the `ate()` function, then multiplying those weights by the number of students at the corresponding school who took the test. We incorporate the prognostic model using the `cov_adj()` function, which generates model predictions for each school and identifies overlap between the prognostic sample and the study sample.
```{r}
study1data <-
     merge(michigan_school_pairs, analysis1data, by = "schoolid", all.x = TRUE)
ip_wts <- propertee::ate(spec, data = study1data) * study1data$Total.Tested.2014
lm_ca <- propertee::cov_adj(lm_county_camod, newdata = study1data, specification = spec)
```

Next, we estimate the intervention effect by using the `lmitt()` function. The only major difference between `lmitt()` and the base `lm()` function in the requirement to specify a `StudySpecification` object. In this `lmitt()` function, we pass the inverse probability of assignment weights to the `weights` argument and the adjusted predictions to the `offset` argument.
```{r}
main_effect_fmla <- as.formula(paste0(RESPONSE_COL, "~1"))
lm_ca_effect <- propertee::lmitt(
  main_effect_fmla, specification = spec, data = study1data, weights = ip_wts,
  offset = lm_ca
)
```

The summary of a fitted `lmitt()` model, which is called a `teeMod`, shows the estimated intervention effect and the estimated standard error that has propagated uncertainty from the covariance adjustment regression.
```{r}
summary(lm_ca_effect, vcov.type = "HC0")
```
Since no actual intervention was implemented in the pseudo-RCT, the result of no effect is as expected.

To explore different variance estimation techniques, we can specify a different `vcov.type` in the `summary()` function.

```{r}
summary(lm_ca_effect, vcov.type = "HC1")
```

If parts of the auxiliary sample (here, Oakland County other than the 14 study schools) follow a different pattern of association between covariates and response, the covariance adjustment might wind up doing more harm than good.  To increase robustness to such contamination, we can use robust linear regression for generating predictions.
```{r}
rob_ca <- propertee::cov_adj(rob_county_camod, newdata = study1data, specification = spec)
rob_ca_effect <- propertee::lmitt(
  main_effect_fmla, specification = spec, data = study1data, weights = ip_wts,
  offset = rob_ca
)
summary(rob_ca_effect, vcov.type = "HC1")
```

#### Estimating Heterogeneous Intervention Effects
We can estimate average intervention effects conditional on different race/ethinicity groups by using the second cleaned dataset. `propertee` allows us to interpret the heterogenous effect estimates as the average effect of the intervention on the average MME score for students within each race/ethnicity group. The process for estimating these heterogeneous effects follows a similar user experience to the estimation of the marginal effect, but with two notable exceptions.

##### Exception 1: Formula Specification
The first exception is general to heterogeneous effect estimation with the `propertee` package. When estimating these effects, users need to modify the formula passed to `lmitt()`. Instead of specifying a constant (i.e. 1) on the right-hand side, users should provide a formula that specifies the subgroup variable on the right-hand side.

The first part of the code prepares the data by filtering out any rows with missing values in the response variable or the covariates. We also isolate data for the Oakland County.
```{r}
not_missing_resp <- !is.na(analysis2data[[RESPONSE_COL]])
not_missing_covs <- rowSums(is.na(analysis2data[, MODELING_COLS])) == 0
county_ix <- analysis2data$CONAME == coname
county_mod_camod_dat <- analysis2data[not_missing_resp & not_missing_covs,]
```

Next, we update the model formula to include the subgroup variable `DemographicGroup`, which will be used to estimate heterogeneous effects. We then use this updated formula to fit a weighted linear regression model.
```{r}
mod_camod_form <- update(camod_form, . ~ . + factor(DemographicGroup))
lm_county_mod_camod <- lm(mod_camod_form, county_mod_camod_dat,
                          weights = county_mod_camod_dat$Total.Tested.2014)
```

Then, we prepare the study dataset by merging the cleaned dataset with `michigan_school_pairs` to ensure we have school-level data that allows for the estimation of treatment effects based on school and demographic group.
```{r}
study2data <- merge(michigan_school_pairs, analysis2data, by = "schoolid", all.x = TRUE)
study2data <- study2data[study2data$DemographicGroup %in%
                           c("White", "Black or African American"),]
```

Now, we compute the inverse probability weights using the `ate()` function, which estimates the average effect treatment. Then, we adjust for covariates using `cov_adj()`.
```{r}
ip_wts <- propertee::ate(spec, data = study2data) * study2data$Total.Tested.2014
lm_mod_ca <- propertee::cov_adj(lm_county_mod_camod, newdata = study2data,
                                specification = spec, by = "uniqueid")
```

Finally, we specify a formula for estimating heterogeneous effects and fit the model using the `lmitt()` function.
```{r}
mod_effect_fmla <- as.formula(paste0(RESPONSE_COL, "~ DemographicGroup"))
lm_ca_mod_effect <- propertee::lmitt(mod_effect_fmla, specification = spec,
                                     data = study2data, weights = ip_wts,
                                     offset = lm_mod_ca)
summary(lm_ca_mod_effect, vcov.type = "CR1", cluster = "schoolid")
```

##### Second Exception: Overlapping Rows
The second exception arises when units of assignment (in this case, schools) contribute multiple observations to the heterogeneous effect estimation. This can happen because schools may have students in multiple race/ethnicity groups, leading to overlap in the data.The `StudySpecification` object does not provide enough information to uniquely identify these rows, which causes an issue for standard error calculations that must determine the exact overlap between the covariance adjustment and effect estimation samples. To alleviate this issue, both dataframes must have a column that uniquely identifies each row. If the two dataframes have overlapping rows, these unique identifiers should match up.

#### Conclusion
This vignette has demonstrated how to use the `propertee` package to enhance the evaluation of treatment effects by incorporating covariance adjustment models. The following key concepts and commands were covered:

- Create a `StudySpecification` object which encodes the design, including the unit of assignment, treatment status of each unit of assignment, and optionally block information. This is done using the `rct_spec()` or optionally `obs_spec()` and `rd_spec()` functions.
- Fit both least squares and robust regression models to adjust for covariates and enhance the precision of treatment effect estimates.
- Use the `cov_adj()` function to process the covariate adjustment model.
- Fit a model using the `lmitt()` function to estimate treatment effect that accounts for the specification information and the covariate adjustment by including the `weights` and `offset` arguments in the function.

---
title: "Introduction to propertee"
author: "propertee Authors"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to propertee}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#devtools::load_all("~/repositories/_r/propertee/")
library(propertee)
```

## Main Features

The **propertee** package, Prognostic Regression Offsets with
Propagation of ERrors (for Treatment Effect Estimation), facilitates
direct adjustment or standardization for experiments and observational
studies, with the option of using a separately fitted prognostic
regression model for covariance adjustment. Propertee calls for a
self-standing specification of a study's design and treatment
allocations, using these to create weights needed to align treatment
and control samples for estimation of treatment effects averaged
across a common standard population, with standard errors that
appropriately reflect the study design. For covariance adjustment it
enables offsetting the outcome against predictions from a dedicated
covariance model, with standard error calculations propagating error
as appropriate from the covariance model.

The main workflow consists of two main steps and one optional step:

1. Encode the study's design, including the unit of assignment,
   treatment status of each unit of assignment, and block membership
   information as appropriate, in a `StudySpecification` object. This
   is accomplished with the `obs_spec()`, `rct_spec()` or `rd_spec()`
   functions.
2. Optionally, fit a covariate adjustment model.  This is done with
   any of a number of modeling functions external to `propertee`, for
   example `lm()` or `glm()` from the `stats` package, or `lmrob` from
   the `robustbase` package.
3. Use `lm()` to contrast the separately reweighted treatment and
   control groups, incorporating optional covariance adjustments as an
   `offset` to `lm()`. The `StudySpecification` is unobtrusively
   retained, as is error propagation from the optional covariance
   adjustment, for use in subsequent standard error calculations.
   This is done using the `lmitt()` function, which either takes and
   enriches a fitted `lm` or, in the more fully supported usage,
   internally calls `lm()` to itself create a directly adjusted
   contrast of treatment conditions.

## Example Data

The example dataset comes from the state of Tennessee's Student-Teacher
Achievement Ratio (STAR) experiment. Students were randomly assigned to three
possible classroom conditions: small (13 to 17 students per teacher), regular
class (22 to 25 students per teacher), and regular-with-aide class (22 to 25
students with a full-time teacher's aide).

```{r}
data("STARplus")
table(STARplus$cond_at_entry)
```

For this introductory example, we will examine a single binary
treatment - "small" classrooms versus "regular" and "regular+aide" classrooms.

```{r}
STARplus$cond_small <- STARplus$cond_at_entry == "small"
STARplus$cond_small[STARplus$cond_at_entry == "nonexperimental"] <- NA
table(STARplus$cond_small)
```

This is only for the introduction -- **propertee** can keep track of treatments
with multiple levels, and includes convenience functions to dichotomize them 
for outcome analyses.  (This `dichotomy` functionality is described in a separate
vignette, `non-binary-treatments`.)

The outcome of interest is a reading score at the end of kindergarten.

```{r}
summary(STARplus$g1treadss)
```



We first take the random assignment scheme to have been either complete or
Bernoulli randomization of students, separately by school. (Later we'll change this to the more realistic assumption of complete or Bernoulli randomization, separately by school *and* year of entry into the study.) Accordingly, experimental blocks are given by the `school_at_entry` variable.

```{r}
length(unique(STARplus$school_at_entry))
head(table(STARplus$school_at_entry))
```

We need a unique identifier for the unit by which treatment was allocated.  STAR treatments were assigned to students, as opposed to classrooms or other aggregates of students; accordingly we need a unique identifier per student. The  `stdntid` variable fills this role.

```{r}
length(unique(STARplus$stdntid))
head(STARplus$stdntid)
```

## A Basic Example

### Defining the `StudySpecification`

The three `_spec` functions (`rct_spec()`, `obs_spec()`, and `rd_spec()`)
operate similarly. The first argument is the most important, and encodes all the
specification information through the use of a formula. The left-hand side of
the formula identifies the treatment variable. The right-hand side of the
formula consists of the following potential pieces of information:

1. `unit_of_assignment()`: This identifies the variable(s) which indicate the
   units of assignment. This is required for all specifications. The alias
   `uoa()` can be used in its place.
2. `block()`: The identifies the variable(s) which contain block information.
   Optional.
3. `forcing()`: In regression discontinuity specifications (`rd_spec()`), this
   identifies the variable(s) which contain forcing information.

To define a `StudySpecification` in our example:

```{r}
spec <- obs_spec(cond_small ~ unit_of_assignment(stdntid) + block(school_at_entry),
                  data = STARplus, na.fail = FALSE)
summary(spec)
```

Should additional variables be needed to identify the unit of assignment,
block, or forcing, they can be included. Had the variable identifying the years in which participants entered the study been available, for example, we might have identified the experimental blocks as combinations of school and year of study entry, rather than as the school alone.  To do this we would pass `block(year_at_entry, school_at_entry)`, rather than `block(school_at_entry)`, in the above.

### Estimating the treatment effect

The main function for estimating treatment effects is the `lmitt()` function. It
takes in three main required arguments:

1. A formula specifying the outcome and the desired treatment effect.
2. The data set containing the outcome information.
3. A [`StudySpecification`](#defining-the-specification).

Note that the data set does **not** need to be the same data set which generated
the `StudySpecification`; it does however need to include the same variables to
identify the units of assignment. (If the variable names differ, the `by=`
argument can be used to link them, though we recommend renaming to reduce the
likelihood of issues.)

For example, you may have one dataset containing school-level information, and a
separate dataset containing student-level information. Assume school is the unit
of assignment. While you could of course merge those two data-sets,
**propertee** can instead use the school-level data to define the
`StudySpecification`, and the student-level data to estimate the treatment
effect.

The formula entering `lmitt()` can take on one of two forms:

```r
y ~ 1
```

will estimate the main treatment effect on outcome variable `y`, and

```r
y ~ x
```

will estimate subgroup specific treatment effects for each level of `x` for the
outcome `y`, if `x` is categorical. For continuous `x`, a main effect and a
treatment-`x` interaction effect is estimated.

Therefore, to estimate the treatment coefficient in our example, we can run:

```{r}
te <- lmitt(g1treadss ~ 1, data = STARplus, specification = spec)
summary(te)
```

(A warning indicates that we have yet to specify either a weighting or block fixed effect adjustment, as we need to do to make the coefficient interpretable as an estimate of a treatment effect.  We'll turn to this topic presently.) To estimate separate treatment coefficients for each `race` subgroup:

```{r}
te_s <- lmitt(g1treadss ~ race, data = STARplus, specification = spec)
summary(te_s)
```

### Including specification weights

Study specification weights can be easily included in this estimation.
**propertee** supports average treatment effect (ATE) and effect of the
treatment on the treated (ETT) weights.

To include one of the weights, simply include the `weights = "ate"` or
`weights = "ett"` argument to `lmitt()`:

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, weights = "ate")
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, weights = "ett")
```

Internally, these call the `ate()` or `ett()` functions which can be used
directly.

```{r}
head(ate(spec, data = STARplus))
```

When included inside `lmitt()`, you do not need to specify any additional
arguments to `ate()` or `ett()`, enabling easy functions of weights. For example
if you had some other weight variable, say `wgt`, you could include `weights =
wgt*ate()` in the `lmitt()` call.

### Covariance Adjustment models

By itself, `lmitt()` does not allow for other covariates; e.g. something like
`lmitt(y ~ 1 + control_var,...` will fail. To adjust for covariates, a separate
covariate model should be fit. Any model which supports a `predict()` function
should work.

```{r}
camod <- lm(g1treadss ~ gender * dob + race, data = STARplus)
```

The `cov_adj()` function can be used to process the covariance adjustment model
and produce the required values; and its output can be passed as an `offset=`.

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec,
      weights = "ate", offset = cov_adj(camod))
```

Similarly to the weight functions, `cov_adj()` attempts to locate the correct
arguments (in this case, mainly the `data=` argument) to use in the model
command; while `cov_adj()` does fall back to using the data which is in the
covariance model, its safer to use the `newdata=` argument if calling
`cov_adj()` outside of the model.

```{r}
head(cov_adj(camod, newdata = STARplus))
```

Also, similarly to weights, `cov_adj()` can be used in normal modeling commands
as well.

```{r}
lm(g1treadss ~ cond_small, data = STARplus, weights = ate(spec),
   offset = cov_adj(camod))
```

### Absorbing Blocks

If fixed effects for blocks are desired, which can be absorbed away to avoid
estimating, the `absorb=TRUE` argument can be passed.

```{r}
lmitt(g1treadss ~ 1, data = STARplus, specification = spec, absorb = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.cov01_est}
\alias{.cov01_est}
\title{(Internal) Helper function for design-based meat matrix
calculation}
\usage{
.cov01_est(XX, zobs, bid)
}
\value{
estimated upper and lower bounds of covariance matrix of estimating
function vectors under treatment and under control
}
\description{
(Internal) Helper function for design-based meat matrix
calculation
}
\details{
the Young's elementary inequality is used
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/areg.center.R
\name{areg.center}
\alias{areg.center}
\title{Group-center akin to Stata's \code{areg}}
\usage{
areg.center(mm, grp, wts = NULL, grand_mean_center = FALSE)
}
\arguments{
\item{mm}{Matrix of variables to center}

\item{grp}{Group to center on}

\item{wts}{Optional weights}

\item{grand_mean_center}{Optional center output at \code{mean(var)}}
}
\value{
Vector of group-centered values
}
\description{
From the Stata documentation: \code{areg} begins by recalculating Y and X
and to have mean 0 within the groups specified by \code{absorb()}. The
overall mean of each variable is then added back in.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assigned.R
\name{assigned}
\alias{assigned}
\alias{adopters}
\alias{a.}
\alias{z.}
\title{Obtain Treatment from StudySpecification}
\usage{
assigned(specification = NULL, data = NULL, dichotomy = NULL)

adopters(specification = NULL, data = NULL, dichotomy = NULL)

a.(specification = NULL, data = NULL, dichotomy = NULL)

z.(specification = NULL, data = NULL, dichotomy = NULL)
}
\arguments{
\item{specification}{Optional \code{StudySpecification}. If the
\code{StudySpecification} can't be identified in the model (usually
because neither weights (\code{ate()} or \code{ett()}) nor a covariate
adjustment model (\code{cov_adj()}) are found), the
\code{StudySpecification} can be passed diretly.}

\item{data}{Optional data set. By default \code{\link[=assigned]{assigned()}} will attempt to
identify the appropriate data, if this fails (or you want to overwrite
it), you can pass the data here.}

\item{dichotomy}{optional; a formula defining the dichotomy of the treatment
variable if it isn't already \code{0}/\code{1}. See details for more
information. If \code{ett()} or \code{ate()} is called within a
\code{lmitt()} call that specifies a \code{dichotomy} argument, that
\code{dichotomy} will be used if the argument here has not been specified.}
}
\value{
The treatment variable to be placed in the regression formula.
}
\description{
When passing a \code{lm} object to \code{\link[=lmitt]{lmitt()}}, extract and use
the treatment variable specified in the \code{StudySpecification}.
}
\details{
When passing a \code{lm} object to \code{\link[=lmitt]{lmitt()}}, the treatment
variable in the \code{formula} passed to \code{\link[=lm]{lm()}} needs to be identifiable.
Rather than placing the treatment variable directly in the \code{formula},
use one of these functions, to allow \code{\link[=lmitt]{lmitt()}} to identify the treatment
variable.

To keep the formula in the \code{\link[=lm]{lm()}} call concise, instead of passing
\code{specification} and \code{data} arguments to these functions, one can
pass a \code{WeightedStudySpecification} object to the \code{weights}
argument of the \code{\link[=lm]{lm()}} call or a \code{SandwichLayer} object to the
\code{offset} argument.

Alternatively, you can pass the \code{specification} and \code{data}
arguments.

While \code{assigned()} can be used in any situation, it is most useful
for scenarios where the treatment variable is non-binary and the
\code{StudySpecification} contains a \code{Dichotomy}. For example, say
\code{q} is a 3-level ordinal treatment variable, and the binary
comparison of interest is captured in \code{dichotomy = q == 3 ~ q < 3}.
If you were to fit a model including \code{q} as a predictor, e.g.
\code{lm(y ~ q, ...)}, \code{lm} would treat \code{q} as the full ordinal
variable. On the other hand, by calling \code{lm(y ~ assigned(), weights =
  ate(spec), ...)}, \code{assigned()} will generate the appropriate binary
variable to allow estimation of treatment effects.

If called outside of a model call and without a \code{data} argument, this
will extract the treatment from the \code{specification}. If this is the
goal, the \code{\link[=treatment]{treatment()}} function is better suited for this purpose.
}
\examples{
data(simdata)
spec <- obs_spec(z ~ uoa(uoa1, uoa2), data = simdata)
mod <- lm(y ~ assigned(), data = simdata, weights = ate(spec))
lmittmod <- lmitt(mod)
summary(lmittmod, vcov.type = "CR0")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.teeMod.R
\name{summary.teeMod}
\alias{summary.teeMod}
\alias{print.summary.teeMod}
\title{Summarizing \code{teeMod} objects}
\usage{
\method{summary}{teeMod}(object, vcov.type = "HC0", ...)

\method{print}{summary.teeMod}(
  x,
  digits = max(3L, getOption("digits") - 3L),
  signif.stars = getOption("show.signif.stars"),
  ...
)
}
\arguments{
\item{object}{\code{teeMod} object}

\item{vcov.type}{A string indicating the desired variance estimator. See
\code{\link[=vcov_tee]{vcov_tee()}} for details on accepted types.}

\item{...}{Additional arguments to \code{\link[=vcov_tee]{vcov_tee()}}, such as the desired finite
sample heteroskedasticity-robust standard error adjustment.}

\item{x}{\code{summary.teeMod} object}

\item{digits}{the number of significant digits to use when printing.}

\item{signif.stars}{logical. If ‘TRUE’, ‘significance stars’ are printed for
each coefficient.}
}
\value{
object of class \code{summary.teeMod}
}
\description{
\code{\link[=summary]{summary()}} method for class \code{teeMod}
}
\details{
If a \code{teeMod} object is fit with a \code{SandwichLayer}
offset, then the usual \code{stats::summary.lm()} output is enhanced by
the use of covariance-adjusted sandwich standard errors, with t-test
values recalculated to reflect the new standard errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confint_lm.R
\name{.confint_lm}
\alias{.confint_lm}
\title{Produce confidence intervals for linear models}
\usage{
.confint_lm(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{a fitted \code{teeMod} model}

\item{parm}{a specification of which parameters are to be given confidence
intervals, either a vector of numbers or a vector of names. If missing,
all parameters are considered.}

\item{level}{the confidence level required.}

\item{...}{additional arguments to pass to \code{vcov.teeMod()}}
}
\value{
A matrix (or vector) with columns giving lower and upper confidence
limits for each parameter. These will be labelled as (1-level)/2 and 1 -
(1-level)/2 in \% (by default 2.5\% and 97.5\%)
}
\description{
Produce confidence intervals for linear models
}
\details{
\code{.confint_lm()} is a copy of \code{stats::confint.lm} but
passes arguments in \code{...} to the \code{vcov()} call. When called on a
\code{teeMod} model, this produces confidence intervals where standard
errors are computed based on the desired formulation of the
\code{vcov_tee()} call.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{.make_uoa_cluster_df}
\alias{.make_uoa_cluster_df}
\title{Make a dataframe that links units of assignment with clusters}
\usage{
.make_uoa_cluster_df(spec, cluster = NULL)
}
\arguments{
\item{spec}{A \code{StudySpecification} object.}

\item{cluster}{A character vector of column names to use as clusters.
Columns must exist in the dataframe used to create the
\code{StudySpecification} object. Defaults to NULL, in which case the
column names specified in the \code{unitid()},
\code{unit_of_assignment()}, or \code{cluster()} function in the
\code{StudySpecification} formula will be used.}
}
\value{
A dataframe where the number of rows coincides with the number of
distinct unit of assignment or cluster combinations (depending on whether
\code{cluster} is a more or less granular level than the assignment level) and
the columns correspond to the unit of assignment columns and a "cluster"
column
}
\description{
Make a dataframe that links units of assignment with clusters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.bin_txt}
\alias{.bin_txt}
\title{(Internal) Extracts treatment as binary \code{vector}}
\usage{
.bin_txt(spec, data = NULL, dichotomy = NULL)
}
\arguments{
\item{spec}{A \code{StudySpecification}, used to get treatment assignment
information}

\item{data}{A dataframe with unit of assignment information and, if a
dichotomy is provided, columns specified therein}

\item{dichotomy}{Optional, a formula. See the Details section of the
\code{ett()} or \code{att()} help pages for information on specifying the
formula}
}
\value{
A \code{vector} of binary treatments
}
\description{
(Internal) Extracts treatment as binary \code{vector}
}
\details{
If a \code{dichotomy} is specified or the \code{StudySpecification}
has a treatment variable consisting only of 0/1 or \code{NA}, then returns
the binary treatment. Otherwise (it has a non-binary treatment and lacks a
dichotomy) it errors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{has_binary_treatment}
\alias{has_binary_treatment}
\title{Check whether treatment stored in a \code{StudySpecification} object
is binary}
\usage{
has_binary_treatment(spec)
}
\arguments{
\item{spec}{\code{StudySpecification} object}
}
\value{
logical vector of length 1
}
\description{
Check whether treatment stored in a \code{StudySpecification} object
is binary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block_center_residuals.R
\name{block_center_residuals}
\alias{block_center_residuals}
\title{Adjust residuals for both-sides absorption}
\usage{
block_center_residuals(x)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
the fitted \code{teeMod} with updated block center residuals.
}
\description{
Adjust residuals for both-sides absorption
}
\details{
This function subtracts off the block residual mean function
\eqn{\hat \alpha(v_b, \theta)} for each observation from model residuals
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.check_df_moderator_estimates}
\alias{.check_df_moderator_estimates}
\title{(Internal) Replace standard errors for moderator effect estimates with
insufficient degrees of freedom with \code{NA}}
\usage{
.check_df_moderator_estimates(
  vmat,
  model,
  cluster_cols,
  values_from = model,
  model_data = quote(data),
  envir = environment(formula(model))
)
}
\arguments{
\item{vmat}{output of \code{.vcov_XXX()} called with input to \code{model}
argument below as the first argument}

\item{model}{a fitted \code{teeMod} model}

\item{cluster_cols}{a character vector indicating the column(s) defining
cluster ID's}

\item{values_from}{optional, a fitted model object. Defaults to \code{x}.}

\item{model_data}{dataframe or name of dataframe used to fit \code{model}}

\item{envir}{environment to get \code{model_data} from if \code{model_data}
has class \code{name}}
}
\value{
A variance-covariance matrix with NA's for estimates lacking
sufficient degrees of freedom
}
\description{
(Internal) Replace standard errors for moderator effect estimates with
insufficient degrees of freedom with \code{NA}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationConverters.R
\name{as_rct_spec}
\alias{as_rct_spec}
\alias{as_obs_spec}
\alias{as_rd_spec}
\title{Convert \code{StudySpecification} between types}
\usage{
as_rct_spec(StudySpecification, ..., loseforcing = FALSE)

as_obs_spec(StudySpecification, ..., loseforcing = FALSE)

as_rd_spec(StudySpecification, data, ..., forcing)
}
\arguments{
\item{StudySpecification}{a \code{StudySpecification} to convert}

\item{...}{Ignored.}

\item{loseforcing}{converting from RD to another \code{StudySpecification}
type will error to avoid losing the forcing variable. Setting
\code{loseforcing = TRUE} allows the conversion to automatically drop the
forcing variable. Default \code{FALSE}.}

\item{data}{converting to an RD requires adding a \code{forcing} variable,
which requires access to the original data.}

\item{forcing}{converting to an RD requires adding a \code{forcing}
variable. This should be entered as a formula which would be passed to
\code{update}, e.g. \code{forcing = . ~ . + forcing(forcevar)}.}
}
\value{
\code{StudySpecification} of the updated type
}
\description{
Convert a \code{StudySpecification} between a observational
study, a randomized control trial, and a regression discontinuity (created
from \code{obs_spec}, \code{rct_spec} and \code{rd_spec} respectively).
}
\examples{
spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
spec
as_obs_spec(spec)
as_rd_spec(spec, simdata, forcing = ~ . + forcing(force))
spec2 <- rd_spec(o ~ uoa(uoa1, uoa2) + forcing(force), data = simdata)
spec2
# as_rct_spec(spec2) # this will produce an error
as_rct_spec(spec2, loseforcing = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{lsoSynth}
\alias{lsoSynth}
\title{Synthethic Regression Discontinuity Data}
\format{
A \code{data.frame} with 40,403 rows and 11 columns.
\itemize{
\item \code{R}
\item \code{lhsgrade_pct}
\item \code{nextGPA}
\item \code{probation_year1}
\item \code{totcredits_year1}
\item \code{male}
\item \code{loc_campus1}
\item \code{loc_campus2}
\item \code{bpl_north_america}
\item \code{english}
\item \code{age_at_entry}
}
}
\usage{
lsoSynth
}
\description{
The data for this example were randomly simulated using the
\bold{synthpop} package in R based on data originally collected by Lindo,
Sanders, and Oreopoulos (2010).
}
\details{
See the "Regression Discontinuity StudySpecifications" vignette on the
\code{propertee} website for more details on the original data, a link to
the code used to generate this synthethic data, and a detailed example.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{.make_PreSandwichLayer}
\alias{.make_PreSandwichLayer}
\title{(Internal) Get covariance adjustments and their gradient with respect
to covariance adjustment model coefficients}
\usage{
.make_PreSandwichLayer(model, newdata = NULL, ...)
}
\arguments{
\item{model}{a fitted model to use for generating covariance adjustment
values}

\item{newdata}{a dataframe with columns called for in \code{model}}

\item{...}{additional arguments to pass on to \code{model.frame} and
\code{model.matrix}. These cannot include \code{na.action}, \code{xlev},
or \code{contrasts.arg}: the former is fixed to be \code{na.pass}, while
the latter two are provided by elements of the \code{model} argument.}
}
\value{
A \code{PreSandwichLayer} object
}
\description{
\code{.make_PreSandwichLayer()} takes a fitted covariance
adjustment model passed to the \code{model} argument and generates
adjustments to outcomes for observations in the \code{newdata} argument.
It also evaluates the gradient of the adjustments taken with respect to
the coefficients at the coefficient estimates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dichotomy.R
\name{.find_dichotomies}
\alias{.find_dichotomies}
\title{(Internal) Find \code{dichotomy} formulas in the call stack}
\usage{
.find_dichotomies()
}
\value{
A list where elements are either formulas or NULL, depending on
whether a \code{dichotomy} argument was found in \code{lmitt.formula()} or
its \code{weights} argument
}
\description{
(Internal) Find \code{dichotomy} formulas in the call stack
}
\details{
\code{.find_dichotomies()} searches for \code{lmitt.formula()}
calls and their \code{weights} arguments for any \code{dichotomy}
arguments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cov_adj.R
\name{.update_ca_model_formula}
\alias{.update_ca_model_formula}
\title{(Internal) Add columns for merging covariance adjustment and direct
adjustment samples to model formula}
\usage{
.update_ca_model_formula(model, by = NULL, specification = NULL)
}
\arguments{
\item{model}{any model that inherits from a \code{glm}, \code{lm}, or
\code{robustbase::lmrob} object.}

\item{by}{named vector; optional.}

\item{specification}{a \code{StudySpecification} object; optional. If not
provided, it can be retrieved from the call stack if passed to suitable
calls.}
}
\value{
formula
}
\description{
(Internal) Add columns for merging covariance adjustment and direct
adjustment samples to model formula
}
\details{
This function is typically used prior to
\code{.get_data_from_model()} and incorporates information provided in a
\code{by} argument to ensure the necessary columns for merging the two
samples are included in any \code{model.frame()} calls.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightedStudySpecification.R
\docType{class}
\name{WeightedStudySpecification-class}
\alias{WeightedStudySpecification-class}
\title{(Internal) Modeling weights with an accompanying StudySpecification}
\description{
(Internal) Modeling weights with an accompanying StudySpecification
}
\details{
\verb{@target} is used for calculation purpose; defining what weight to calculate
\verb{@weightAlias} is only to store the alias used in creation of the
weights in case we want to report it later.
}
\section{Slots}{

\describe{
\item{\code{.Data}}{numeric vector of modeling weights}

\item{\code{StudySpecification}}{a StudySpecification}

\item{\code{target}}{character string, e.g. "ate"}

\item{\code{weightAlias}}{alias for target appearing in an originating call}

\item{\code{dichotomy}}{formula describing a treatment/comparison dichotomy}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationConverters.R
\name{.remove_forcing}
\alias{.remove_forcing}
\title{(Internal) Removes the forcing column entirely from a
\code{StudySpecification}}
\usage{
.remove_forcing(spec)
}
\arguments{
\item{spec}{A \code{StudySpecification}}
}
\value{
The \code{StudySpecification} without any forcing variable
}
\description{
In preparation for converting an RD \code{StudySpecification} to another
\code{StudySpecification}, this will strip the forcing variable entirely. It
is removed from the data (both \code{@structure} and \code{@column_index}),
as well as from the formula stored in \code{@call}.
}
\details{
Note that the output \code{StudySpecification} will fail a validity check
(with \code{validObject()}) due to an RD \code{StudySpecification} requiring
a forcing variable, so change the \code{@type} immediately.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lmrob_methods.R
\name{estfun.lmrob}
\alias{estfun.lmrob}
\alias{bread.lmrob}
\title{Generate matrix of estimating equations for \code{lmrob()} fit}
\usage{
\method{estfun}{lmrob}(x, ...)

\method{bread}{lmrob}(x, ...)
}
\arguments{
\item{x}{An \code{lmrob} object produced using an MM/SM estimator chain}

\item{...}{Additional arguments to be passed to \code{bread}}
}
\value{
A \eqn{n\times }(p+1) matrix where the first column corresponds to
the scale estimate and the remaining \eqn{p} colums correspond to the
coefficients

A \eqn{p\times }(p+1) matrix where the first column corresponds to
the scale estimate and the remaining \eqn{p} colums correspond to the
coefficients
}
\description{
Generate matrix of estimating equations for \code{lmrob()} fit

Extract bread matrix from an \code{lmrob()} fit
}
\details{
This is part of a workaround for an issue in the robustbase code
affecting sandwich covariance estimation. The issue in question is issue
#6471, robustbase project on R-Forge. This function contributes to
providing sandwich estimates of covariance-adjusted standard errors for
robust linear covariance adjustment models.

This is part of a workaround for an issue in the robustbase code
affecting sandwich covariance estimation. The issue in question is issue
#6471, robustbase project on R-Forge. This function contributes to
providing sandwich estimates of covariance-adjusted standard errors for
robust linear covariance adjustment models.
}
\author{
Ben B. Hansen
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightedStudySpecification.R
\name{subset,WeightedStudySpecification-method}
\alias{subset,WeightedStudySpecification-method}
\alias{[,WeightedStudySpecification-method}
\title{\code{WeightedStudySpecification} subsetting}
\usage{
\S4method{subset}{WeightedStudySpecification}(x, subset)

\S4method{[}{WeightedStudySpecification}(x, i)
}
\arguments{
\item{x}{\code{WeightedStudySpecification} object}

\item{subset}{Logical vector identifying values to keep or drop}

\item{i}{indices specifying elements to extract or replace. See
\code{help("[")} for further details.}
}
\value{
A \code{WeightedStudySpecification} object which is a subsetted
version of \code{x}.
}
\description{
Provides functionality to subset the weights of a
\code{WeightedStudySpecification} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{simdata}
\alias{simdata}
\title{Simulated data}
\format{
A data.frame with 100 rows and 7 columns.
\itemize{
\item \code{uoa1} First level unit of assignment ID
\item \code{uoa2} Second level unit of assignment ID
\item \code{bid} Block ID
\item \code{force} Forcing variable
\item \code{z} Binary treatment indicator
\item \code{o} 4-level ordered treatment variable
\item \code{dose} Dose treatment variable
\item \code{x} Some predictor
\item \code{y} Some outcome
}
}
\usage{
simdata
}
\description{
Simulated data to use with the \bold{propertee} package with unit
of assignment level treatment assignment
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lmitt.R
\name{lmitt}
\alias{lmitt}
\alias{lmitt.formula}
\alias{lmitt.lm}
\title{Linear Model for Intention To Treat}
\usage{
lmitt(obj, specification, data, ...)

\method{lmitt}{formula}(
  obj,
  specification,
  data,
  absorb = FALSE,
  offset = NULL,
  weights = NULL,
  ...
)

\method{lmitt}{lm}(obj, specification = NULL, ...)
}
\arguments{
\item{obj}{A \code{formula} or a \code{lm} object. See Details.}

\item{specification}{The \code{StudySpecification} to be used.
Alternatively, a formula creating a specification. (Of the type of that
would be passed as the first argument to \code{\link[=rd_spec]{rd_spec()}}, \code{\link[=rct_spec]{rct_spec()}}, or
\code{\link[=obs_spec]{obs_spec()}}, with the difference that \code{\link[=cluster]{cluster()}}, \code{\link[=uoa]{uoa()}}
and \code{\link[=unit_of_assignment]{unit_of_assignment()}} terms can be omitted when each row of
\code{data} represents a distinct unit of assignment.) If the formula includes a \code{\link[=forcing]{forcing()}} element, an RD
specification is created. Otherwise an observational specification is
created. An RCT specification must be created manually using \code{\link[=rct_spec]{rct_spec()}}.}

\item{data}{A \code{data.frame} such as would be passed into \code{\link[=lm]{lm()}}.}

\item{...}{Additional arguments passed to \code{\link[=lm]{lm()}} and other functions.
An example of the latter is \code{dichotomy=}, a formula passed to
\code{\link[=assigned]{assigned()}} and, as appropriate, \code{\link[=ate]{ate()}}, \code{\link[=att]{att()}}, \code{\link[=atc]{atc()}} or
\code{\link[=ato]{ato()}}. It is used to dichotomize a non-binary treatment
variable in \code{specification}. See the Details section of the
\code{ate()} help page for examples.}

\item{absorb}{If \code{TRUE}, fixed effects are included for blocks
identified in the \code{StudySpecification}. Excluded in \code{FALSE}.
Default is \code{FALSE}. The estimates of these fixed effects are
suppressed from the returned object.}

\item{offset}{Offset of the kind which would be passed into \code{\link[=lm]{lm()}}. Ideally,
this should be the output of \code{\link[=cov_adj]{cov_adj()}}.}

\item{weights}{Which weights should be generated? Options are \code{"ate"}
or \code{"ett"}. Alternatively, the output of a manually run \code{ate()}
or \code{ett()} can be used.}
}
\value{
\code{teeMod} object (see Details)
}
\description{
Generates a linear model object to estimate a treatment effect,
with proper estimation of variances accounting for the study
specification.
}
\details{
The first argument to \code{\link[=lmitt]{lmitt()}} should be a formula specifying the
outcome on the left hand side. The right hand side of the formula can be
any of the following:

\itemize{
\item \code{1}: Estimates a main treatment effect.
\item a subgroup variable: Estimates a treatment effect within each level
of your subgrouping variable.
\item a continuous moderator: Estimates a main treatment effect as well as
a treatment by moderator interaction. The moderator is not
automatically centered.
}

Alternatively, \code{obj} can be a pre-created \code{lm} object. No
modification is made to the formula of the object. See the help for
\code{as.lmitt()} for details of this conversion.

The \code{\link[=lmitt]{lmitt()}} function's \code{subset=} argument governs
the subsetting of \code{data} prior to model fitting, just as with \code{\link[=lm]{lm()}}.
Functions such as \code{\link[=rct_spec]{rct_spec()}} that create \code{StudySpecification}s
also take an optional \code{subset=} argument, but its role differs
from that of the \code{subset=} argument of \code{\link[=lm]{lm()}} or \code{\link[=lmitt]{lmitt()}}. The \code{subset=}
argument when creating a \code{StudySpecification} restricts the data used
to generate the \code{StudySpecification}, but has no direct impact on the
future \code{lm()} or \code{lmitt()} calls using that
\code{StudySpecification}. (It can have an indirect impact by excluding
particular units from receiving a treatment assignment or weight. When
treatment assignments or weights are reconstructed from the \code{StudySpecification},
these units will receive NAs, and will be excluded from the \code{lm()}
or \code{lmitt()} fit under typical \code{na.action} settings.)

To avoid variable name collision, the treatment variable defined in the
\code{specification} will have a "\code{.}" appended to it. For example, if
you request a main treatment effect (with a formula of \code{~ 1}) with a
treatment variable named "txt", you can obtain its estimate from the
returned \code{teeMod} object via \code{$coefficients["txt."]}.

\code{\link[=lmitt]{lmitt()}} will produce a message if the \code{StudySpecification}
designates treatment assignment by block but the blocking
structure appears not to be reflected in the \code{weights}, nor
in a block fixed effect adjustment (via \code{absorb=TRUE}). While
not an error, this is at odds with intended uses of
\code{propertee}, so \code{lmitt()} flags it as a potential
oversight on the part of the analyst. To disable this message, run
\code{options("propertee_message_on_unused_blocks" = FALSE)}.

\code{\link[=lmitt]{lmitt()}} returns objects of class \sQuote{\code{teeMod}}, for
Treatment Effect Estimate Model, extending the lm class to add a
summary of the response distribution under control (the
coefficients of a controls-only regression of the response on an
intercept and any moderator variable).  \code{teeMod} objects also
record the underlying \code{StudySpecification} and information
about any externally fitted models \code{mod} that may have been
used for covariance adjustment by passing
\code{offset=cov_adj(mod)}. In the latter case, responses are
offsetted by predictions from \code{mod} prior to treatment effect
estimation, but estimates of the response variable distribution
under control are calculated without reference to \code{mod}.

The response distribution under control is also characterized when
treatment effects are estimated with block fixed effects, i.e. for
\code{\link[=lmitt]{lmitt()}} with a \code{formula} first argument with option
\code{absorb=TRUE}.  Here as otherwise, the supplementary
coefficients describe a regression of the response on an intercept
and moderator variables, to which only control observations
contribute; but in this case the weights are modified for this
supplementary regression. The treatment effect estimates adjusted
for block fixed effects can be seen to coincide with estimates
calculated without block effect but with weights multiplied by an
additional factor specific to the combination of block and
treatment condition. For block \eqn{s} containing units with weights
\eqn{w_i} and binary treatment assignments \eqn{z_i}, define
\eqn{\hat{\pi}_s} by \eqn{\hat{\pi}_s\sum_sw_i=\sum_sz_iw_i}. If
\eqn{\hat{\pi}_s} is 0 or 1, the block doesn't contribute to
effect estimation and the additional weighting factor is 0; if
\eqn{0 < \hat{\pi}_s < 1}, the additional weighting factor is
\eqn{1 - \hat{\pi}_s} for treatment group members and
\eqn{\hat{\pi}_s} for controls. When estimating a main effect only
or a main effect with continuous moderator, supplementary
coefficients under option \code{absorb=TRUE} reflect regressions
with additional weighting factor equal to 0 or \eqn{\hat{\pi}_s},
respectively, for treatment or control group members of block
\eqn{s}. With a categorical moderator and \code{absorb=TRUE},
this additional weighting factor determining supplementary coefficients
is calculated separately for each level \eqn{\ell} of the moderator
variable, with the sums defining \eqn{\hat{\pi}_{s\ell}} restricted
not only to block \eqn{s} but also to observations with moderator
equal to \eqn{\ell}.
}
\examples{
data(simdata)
spec <- rct_spec(z ~ cluster(uoa1, uoa2), data = simdata)
mod1 <- lmitt(y ~ 1, data = simdata, specification = spec, weights = "ate")
mod2 <- lmitt(y ~ as.factor(o), data = simdata, specification = spec, weights = "ate")
### observational study with treatment z assigned row-wise within blocks:
mod3 <- lmitt(y ~ 1, data=simdata, specification=z ~ block(bid), weights="att")
### regression discontinuity study with units of assignment
### given by combinations of uoa1, uoa2:
mod4 <- lmitt(y ~ 1, data = simdata,
              specification = z ~ uoa(uoa1, uoa2) + forcing(force))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{subset,PreSandwichLayer-method}
\alias{subset,PreSandwichLayer-method}
\alias{[,PreSandwichLayer-method}
\title{\code{PreSandwichLayer} and \code{SandwichLayer} subsetting}
\usage{
\S4method{subset}{PreSandwichLayer}(x, subset)

\S4method{[}{PreSandwichLayer}(x, i)
}
\arguments{
\item{x}{\code{PreSandwichLayer} or \code{SandwichLayer} object}

\item{subset}{Logical vector identifying values to keep or drop}

\item{i}{indices specifying elements to extract or replace. See
\code{help("[")} for further details.}
}
\value{
\code{x} subset by \code{subset} or \code{i}
}
\description{
Return subset of a \code{PreSandwichLayer} or
\code{SandwichLayer} which meets conditions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.order_samples}
\alias{.order_samples}
\title{(Internal) Order observations used to fit a \code{teeMod} model and a
prior covariance adjustment model}
\usage{
.order_samples(x, by = NULL, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{by}{character vector of columns to get ID's for ordering from. Default
is NULL, in which case unit of assignment ID's are used for ordering.}

\item{...}{arguments passed to methods}
}
\value{
A list of four named vectors. The \code{Q_not_C} element holds the
ordering for units of observation in the direct adjustment sample but not
the covariance adjustment samples; \code{Q_in_C} and \code{C_in_Q}, the
ordering for units in both; and \code{C_not_Q}, the ordering for units in
the covariance adjustment sample only. \code{Q_in_C} and \code{C_in_Q}
differ in that the names of the \code{Q_in_C} vector correspond to row
indices of the original matrix of estimating equations for the direct
adjustment model, while the names of \code{C_in_Q} correspond to row
indices of the matrix of estimating equations for the covariance adjustment
model. Similarly, the names of \code{Q_not_C} and \code{C_not_Q} correspond
to row indices of the direct adjustment and covariance adjustment samples,
respectively. Ultimately, the order of \code{.make_uoa_ids()} and
\code{estfun.teeMod()} is given by concatenating the vectors stored in
\code{Q_not_C}, \code{Q_in_C}, and \code{C_not_q}.
}
\description{
(Internal) Order observations used to fit a \code{teeMod} model and a
prior covariance adjustment model
}
\details{
\code{.order_samples()} underpins the ordering for
\code{.make_uoa_ids()} and \code{estfun.teeMod()}. This function orders the
outputs of those functions, but also informs how the original matrices of
contributions to estimating equations need to be indexed to align units of
observations' contributions to both sets of estimating equations.\cr\cr
When a \code{by} argument is provided to \code{cov_adj()}, it is used to
construct the order of \code{.order_samples()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specification_table.R
\name{specification_table}
\alias{specification_table}
\alias{stable}
\title{Table of elements from a \code{StudySpecification}}
\usage{
specification_table(
  specification,
  x,
  y = NULL,
  sort = FALSE,
  decreasing = TRUE,
  use_var_names = FALSE,
  ...
)

stable(
  specification,
  x,
  y = NULL,
  sort = FALSE,
  decreasing = TRUE,
  use_var_names = FALSE,
  ...
)
}
\arguments{
\item{specification}{A \code{StudySpecification} object}

\item{x}{One of "treatment", "unit of assignment", (synonym "uoa"), "block".
Abbreviations are accepted. "unit of assignment" can be replaced by
"unitid" or "cluster" if the \code{StudySpecification} was created with
that element.}

\item{y}{Optionally, another string similar to \code{x}. A 1-dimensional
table is produced if \code{y} is left at its default, \code{NULL}.}

\item{sort}{Ignored if \code{y} is not \code{NULL}. If \code{FALSE}
(default), one-way table is sorted according to "names" of levels. If set
to \code{TRUE}, one-way table is sorted according to values.}

\item{decreasing}{If \code{sort} is \code{TRUE}, choose whether to sort
descending (\code{TRUE}, default) or ascending (\code{FALSE}).}

\item{use_var_names}{If \code{TRUE}, name dimensions of table returned by
variable names. If \code{FALSE} (default), name by their function (e.g.
"treatment" or "blocks"). Passing the \code{dnn} argument in \code{...}
(an argument of \code{\link[=table]{table()}}) overrides whatever is requested here.}

\item{...}{additional arguments \code{\link[=table]{table()}}}
}
\value{
A table of the requested variables.
}
\description{
Produces a table (1-dimensional, or 2-dimensional if \code{y}
is specified) of the elements of the \code{StudySpecification}.
}
\examples{
data(simdata)
spec <- obs_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid),
                  data = simdata)
specification_table(spec, "treatment")
specification_table(spec, "treatment", "block", sort = TRUE, use_var_names = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{rct_spec}
\alias{rct_spec}
\alias{rd_spec}
\alias{obs_spec}
\alias{rct_specification}
\alias{rd_specification}
\alias{obs_specification}
\alias{obsstudy_spec}
\alias{obsstudy_specification}
\title{Generates a \code{StudySpecification} object with the given
specifications.}
\usage{
rct_spec(formula, data, subset = NULL, na.fail = TRUE)

rd_spec(formula, data, subset = NULL, na.fail = TRUE)

obs_spec(formula, data, subset = NULL, na.fail = TRUE)

rct_specification(formula, data, subset = NULL, na.fail = TRUE)

rd_specification(formula, data, subset = NULL, na.fail = TRUE)

obs_specification(formula, data, subset = NULL, na.fail = TRUE)

obsstudy_spec(formula, data, subset = NULL, na.fail = TRUE)

obsstudy_specification(formula, data, subset = NULL, na.fail = TRUE)
}
\arguments{
\item{formula}{a \code{formula} defining the \code{StudySpecification}
components. See \code{Details} for specification.}

\item{data}{the data set from which to build the StudySpecification. Note
that this data need not be the same as used to estimate the treatment
effect; rather the \code{data} passed should contain information about the
units of treatment assignment (as opposed to the units of analysis).}

\item{subset}{optional, subset the data before creating the
\code{StudySpecification} object}

\item{na.fail}{If \code{TRUE} (default), any missing data found in the
variables specified in \code{formula} (excluding treatment) will trigger
an error. If \code{FALSE}, non-complete cases will be dropped before the
creation of the \code{StudySpecification}}
}
\value{
a \code{StudySpecification} object of the requested type for use in
further analysis.
}
\description{
Generate a randomized control treatment StudySpecification
(\code{\link[=rct_spec]{rct_spec()}}), or an observational StudySpecification (\code{\link[=obs_spec]{obs_spec()}}), or
a regression discontinuity StudySpecification (\code{\link[=rd_spec]{rd_spec()}}).
}
\details{
The formula should include exactly one \code{\link[=unit_of_assignment]{unit_of_assignment()}} to
identify the units of assignment (one or more variables). (\code{uoa},
\code{cluster}, or \code{unitid} are synonyms for
\code{unit_of_assignment}; the choice of which has no impact on the
analysis. See below for a limited exception in which the
\code{unit_of_assignment} specification may be omitted.) If defining an
\code{rd_spec}, the formula must also include a \code{\link[=forcing]{forcing()}} entry. The
formula may optionally include a \code{\link[=block]{block()}} as well. Each of these can take
in multiple variables, e.g. to pass both a household ID and individual ID
as unit of assignment, use \code{uoa(hhid, iid)} and not \code{uoa(hhid) +
  uoa(iid)}.

The treatment variable passed into the left-hand side of \code{formula}
can either be \code{logical}, \code{numeric}, or \code{character}. If it
is anything else, it attempts conversion to one of those types (for
example, \code{factor} and \code{ordered} are converted to \code{numeric}
if the levels are \code{numeric}, otherwise to \code{character}). If the
treatment is not \code{logical} or \code{numeric} with only values 0 and
1, in order to generate weights with \code{\link[=ate]{ate()}} or \code{\link[=ett]{ett()}}, the
\code{dichotomy} argument must be used in those functions to identify the
treatment and control groups. See \code{\link[=ett]{ett()}} for more details on specifying a
\code{dichotomy}.

There are a few aliases for each version.

If the formula excludes a \code{unit_of_assignment()}, data merges are
performed on row order. Such formulas can also be passed as the
specification argument to lmitt(), and that is their primary intended use
case. It is recommended that each formula argument passed to
*_specification() include a \code{unit_of_assignment()}, \code{uoa()} or
\code{cluster()} term identifying the key variable(s) with which
\code{StudySpecification} data is to be merged with analysis data.
Exceptions to this rule will be met with a warning. To disable the
warning, run \code{options("propertee_warn_on_no_unit_of_assignment" =
  FALSE)}.

The units of assignment, blocks, and forcing variables must be
\code{numeric} or \code{character}. If they are otherwise, an attempt is
made to cast them into \code{character}.
}
\examples{
data(simdata)
spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2) + block(bid),
                  data = simdata)

data(schooldata)
spec <- obs_spec(treatment ~ unit_of_assignment(schoolid) + block(state),
                  data = schooldata)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{..uoa..}
\alias{..uoa..}
\title{Return ..uoa.. column}
\usage{
..uoa..(spec)
}
\arguments{
\item{spec}{A studyspecification}
}
\value{
The ..uoa.. column
}
\description{
Return ..uoa.. column
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{var_table}
\alias{var_table}
\alias{var_names}
\title{Extract Variable Names from \code{StudySpecification}}
\usage{
var_table(specification, compress = TRUE, report_all = FALSE)

var_names(specification, type, implicitBlocks = FALSE)
}
\arguments{
\item{specification}{a \code{StudySpecification} object}

\item{compress}{should multiple variables be compressed into a
comma-separated string? Default \code{TRUE}. If \code{FALSE}, multiple
columns can be created instead.}

\item{report_all}{should we report all possible structures even if they
don't exist in the \code{StudySpecification}? Default \code{FALSE}.}

\item{type}{one of "t", "u", "b", "f"; for "treatment",
"unit_of_assignment", "block", and "forcing" respectively}

\item{implicitBlocks}{If the \code{StudySpecification} is created without
blocks, setting this to \code{TRUE} will return "\code{.blocks_internal}"
as the variable name corresponding to the blocks.}
}
\value{
\code{var_table} returns the requested table. \code{var_names}
returns a vector of variable names.
}
\description{
Methods to extract the variable names to the elements of the
structure of the \code{StudySpecification} (e.g. treatment, unit of
analysis, etc)
}
\details{
When \code{compress} is \code{TRUE}, the result will always have
two columns. When \code{FALSE}, the result will have number of columns
equal to the largest number of variables in a particular role, plus one.
E.g., a call such as \code{rct_spec(z ~ unitid(a, b, c, d) ...} will have
4+1=5 columns in the output matrix with \code{compress = FALSE}.

When \code{report_all} is \code{TRUE}, the matrix is guaranteed to have 3
rows (when the \code{specification} is an RCT or Obs) or 4 rows (when the
\code{specification} is a RD), with empty variable entries as appropriate.
When \code{FALSE}, the matrix will have minimum 2 rows (treatment and unit
of assignment/unitid/cluster), with additional rows for blocks and forcing
if included in the \code{StudySpecification}.
}
\examples{
spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
var_table(spec)
var_table(spec, compress = FALSE)
var_names(spec, "t")
var_names(spec, "u")
var_names(spec, "b")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{show,teeMod-method}
\alias{show,teeMod-method}
\title{Show a \code{teeMod}}
\usage{
\S4method{show}{teeMod}(object)
}
\arguments{
\item{object}{\code{teeMod} object, usually a result of a call to
\code{\link[=lmitt]{lmitt()}}.}
}
\value{
\code{object}, invisibly.
}
\description{
Display information about a \code{teeMod} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{bread.teeMod}
\alias{bread.teeMod}
\title{Extract bread matrix from a \code{teeMod} model fit}
\usage{
\method{bread}{teeMod}(x, values_from = x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{values_from}{optional, a fitted model object. Defaults to \code{x}.}

\item{...}{arguments passed to methods}
}
\value{
A variance-covariance matrix with row and column entries for the estimated
coefficients in \code{x}, the marginal mean outcome in the control condition,
the marginal mean \code{offset} in the control condition (if an \code{offset}
is provided), and if a moderator variable is specified in the formula for \code{x},
the mean interaction in the control condition of the outcome and \code{offset}
with the moderator variable
}
\description{
An S3method for \code{sandwich::bread} that extracts the bread of the
direct adjustment model sandwich covariance matrix.
}
\details{
This function is a thin wrapper around
\code{.get_tilde_a22_inverse()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validWeights.R
\name{.validWeights}
\alias{.validWeights}
\alias{.isValidWeightTarget}
\alias{.isValidWeightAlias}
\alias{.listValidWeightTargets}
\alias{.listValidWeightAliases}
\title{Valid Weights}
\usage{
.validWeights

.isValidWeightTarget(target)

.isValidWeightAlias(alias)

.listValidWeightTargets()

.listValidWeightAliases()
}
\arguments{
\item{target}{String}

\item{alias}{String}
}
\value{
Logical for \code{.isValid*}, and a string for \code{.listValid*}.
}
\description{
These track and access the weights we support as well as any aliases.
}
\details{
"target" refers to weight calculations we support.

"alias" refers to all possible names. Every "target" has at least one
"alias" (itself) and may have more.

\code{.isValidWeightTarget()} and \code{.isValidWeightAlias()} identify
whether a given input (from a user) is a value weighting name.

\code{.listValidWeightTargets()} and \code{.listValidWeightAliases()} are
for returning nicely formatted strings for messages to users.

IMPORTANT: Adding new aliases MUST correspond to new functions defined in
weights_exported.R, with possible adjustments to calculations in
weights_internal.R.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_by.R
\name{.update_by}
\alias{.update_by}
\title{(Internal) Use \code{by} to update \code{StudySpecification} with new
variable names}
\usage{
.update_by(specification, data, by)
}
\arguments{
\item{specification}{A \code{StudySpecification}}

\item{data}{\code{Data set}}

\item{by}{named vector or list connecting names of unit of
assignment/unitid/cluster variables in \code{specification} to unit of
assignment/unitid/cluster variables in \code{data}. Names represent
variables in the StudySpecification; values represent variables in the
data.}
}
\value{
A \code{StudySpecification} with updated variable names
}
\description{
Helper function used to update the variable names of a
\code{StudySpecification} when user passes a \code{by=} argument to align
variable names between data sets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{identify_small_blocks}
\alias{identify_small_blocks}
\title{Identify fine strata}
\usage{
identify_small_blocks(spec)
}
\arguments{
\item{spec}{A \code{StudySpecification} object.}
}
\value{
Logical vector with length given by the number of blocks in
\code{StudySpecification}
}
\description{
Identify blocks in a \code{StudySpecification} with exactly one
treated or one control unit of assignment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.base_S3class_estfun}
\alias{.base_S3class_estfun}
\title{(Internal) Extract empirical estimating equations from a
\code{teeMod} model using the S3 method associated with its
\code{.S3Class} slot}
\usage{
.base_S3class_estfun(x)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
S3 method
}
\description{
(Internal) Extract empirical estimating equations from a
\code{teeMod} model using the S3 method associated with its
\code{.S3Class} slot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.estfun_DB_blockabsorb}
\alias{.estfun_DB_blockabsorb}
\title{Design-based estimating equations contributions}
\usage{
.estfun_DB_blockabsorb(x, by = NULL, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} object}

\item{...}{arguments passed to methods}
}
\value{
An \eqn{n\times k} matrix
}
\description{
Design-based estimating equations contributions
}
\details{
calculate contributions to empirical estimating equations from a
\code{teeMod} model with absorbed intercepts from the design-based
perspective
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.compute_loo_resids}
\alias{.compute_loo_resids}
\title{Compute residuals for a \code{teeMod} object with leave-one-out estimates
of the \code{offset}}
\usage{
.compute_loo_resids(x, cluster, ...)
}
\arguments{
\item{x}{a \code{teeMod} object}

\item{cluster}{vector of column names that identify clusters}
}
\description{
Compute residuals for a \code{teeMod} object with leave-one-out estimates
of the \code{offset}
}
\details{
The residual for any observation also used for fitting the \code{fitted_covariance_model}
stored in the \code{offset} of \code{x} is replaced by an estimated residual
that uses a cluster leave-one-out estimate of the \code{fitted_covariance_model}
for generating a value of the \code{offset}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dichotomy.R
\name{.validate_dichotomy}
\alias{.validate_dichotomy}
\title{(Internal) Validate a dichotomy against other dichotomies found in the
call stack}
\usage{
.validate_dichotomy(possible_dichotomy)
}
\arguments{
\item{possible_dichotomy}{a list or a formula, the former coming from
\code{.find_dichotomies()}}
}
\value{
formula for a dichotomy or NULL
}
\description{
(Internal) Validate a dichotomy against other dichotomies found in the
call stack
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_DB_covadj_bread}
\alias{.get_DB_covadj_bread}
\title{(Internal) Bread matrix of design-based variance}
\usage{
.get_DB_covadj_bread(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
a list of bread matrices
}
\description{
(Internal) Bread matrix of design-based variance
}
\details{
Calculate bread matrix for design-based variance estimate for
\code{teeMod} models with covariance adjustment and without absorbed effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationStructure.R
\name{get_structure}
\alias{get_structure}
\alias{show,StudySpecificationStructure-method}
\title{\code{StudySpecification} Structure Information}
\usage{
get_structure(specification)

\S4method{show}{StudySpecificationStructure}(object)
}
\arguments{
\item{specification}{a \code{StudySpecification} object}

\item{object}{a \code{StudySpecificationStructure} object, typically the
output of \code{get_structure}}
}
\value{
A \code{StudySpecificationStructure} object containing the structure
of the \code{specification} as a \code{data.frame}.
}
\description{
Obtaining a \code{data.frame} which encodes the specification
information.
}
\examples{
data(simdata)
spec <- rct_spec(z ~ uoa(uoa1, uoa2) + block(bid), data = simdata)
get_structure(spec)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights_exported.R
\name{ett}
\alias{ett}
\alias{att}
\alias{ate}
\alias{etc}
\alias{atc}
\alias{ato}
\alias{olw}
\alias{owt}
\alias{pwt}
\title{Generate Direct Adjusted Weights for Treatment Effect Estimation}
\usage{
ett(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

att(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

ate(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

etc(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

atc(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

ato(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

olw(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

owt(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)

pwt(specification = NULL, dichotomy = NULL, by = NULL, data = NULL)
}
\arguments{
\item{specification}{optional; a \code{StudySpecification} object created by
one of \code{rct_spec()}, \code{rd_spec()}, or \code{obs_spec()}.}

\item{dichotomy}{optional; a formula defining the dichotomy of the treatment
variable if it isn't already \code{0}/\code{1}. See details for more
information. If \code{ett()} or \code{ate()} is called within a
\code{lmitt()} call that specifies a \code{dichotomy} argument, that
\code{dichotomy} will be used if the argument here has not been specified.}

\item{by}{optional; named vector or list connecting names of unit of
assignment/ variables in \code{specification} to unit of
assignment/unitid/cluster variables in \code{data}. Names represent
variables in the StudySpecification; values represent variables in the
data. Only needed if variable names differ.}

\item{data}{optional; the data for the analysis to be performed on. May be
excluded if these functions are included as the \code{weights} argument of
a model.}
}
\value{
a \code{WeightedStudySpecification} object, which is a vector of
numeric weights
}
\description{
These should primarily be used inside models. See Details.
}
\details{
These functions should primarily be used in the \code{weight}
argument of \code{\link[=lmitt]{lmitt()}} or\code{\link[=lm]{lm()}}. All arguments are optional if used within
those functions. If used on their own, \code{specification} and
\code{data} must be provided.
\itemize{
\item \code{ate} - Average treatment effect. Aliases: \code{ate()}.
\item \code{ett} - Effect of treatment on the treated. Aliases: \code{ett()},
\code{att()}.
\item \code{etc} - Effect of treatment on controls. Aliases: \code{etc()},
\code{atc()}.
\item \code{ato} - Overlap-weighted average effect. Aliases: \code{ato()},
\code{olw}, \code{owt}, \code{pwt}.
}

In a \code{StudySpecification} with \code{block}s, the weights are
generated as a function of the ratio of the number of treated units in a
block versus the total number of units in a block.

In any blocks where that ratio is 0 or 1 (that is, all units in the block
have the same treatment status), the weights will be 0. In effect this
removes from the target population any block in which there is no basis
for estimating either means under treatment or means under control.

If block is missing for a given observation, a weight of 0 is applied.

A \code{dichotomy} is specified by a \code{formula} consisting of a
conditional statement on both the left-hand side (identifying treatment
levels associated with "treatment") and the right hand side (identifying
treatment levels associated with "control"). For example, if your
treatment variable was called \code{dose} and doses above 250 are
considered treatment, you might write:

\code{ate(..., dichotomy = dose > 250 ~ dose <= 250}

The period (\code{.}) can be used to assign all other units of assignment.
For example, we could have written the same treatment regime as either

\code{etc(..., dichotomy = dose > 250 ~ .}

or

\code{olw(..., dichotomy = . ~ dose <= 250}

The \code{dichotomy} formula supports Relational Operators (see
\link{Comparison}), Logical Operators (see \link{Logic}), and \code{\%in\%} (see
\code{\link[=match]{match()}}).

The conditionals need not assign all values of treatment to control or
treatment, for example, \code{dose > 300 ~ dose < 200} does not assign
\code{200 <= dose <= 300} to either treatment or control. This would be
equivalent to manually generating a binary variable with \code{NA}
whenever \code{dose} is between 200 and 300. Standard errors will reflect
the sizes of the comparison groups specified by the \code{dichotomy}.

Tim Lycurgus contributed code for the computation of weights. The
\sQuote{overlap weight} concept is due to Li, Morgan and Zaslavsky
(2018), although the current implementation differs from that
discussed in their paper in that it avoids estimated propensity scores.
}
\examples{
data(simdata)
spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
summary(lmitt(y ~ 1, data = simdata, specification = spec, weights = ate()), vcov.type = "CR0")
}
\references{
Li, Fan, Kari Lock Morgan, and Alan M. Zaslavsky.
"Balancing covariates via propensity score weighting." Journal of the American Statistical Association 113, no. 521 (2018): 390-400.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{STARplus}
\alias{STARplus}
\title{STAR participants plus nonexperimental controls}
\format{
A \code{data.frame} with 13,382 rows and 56 columns.
\itemize{
\item {stdnid} Student ID
\item {gender} Student gender
\item {race} Student race
\item {birthmonth} Student month of birth
\item {birthday} Student day of birth
\item {birthyear} Student year of birth
\item {read_yr1} SAT reading scaled score from grade at which
student entered the study
\item {math_yr1} SAT math scaled score from grade at which
student entered the study
\item {gktreadss} Kindergarten reading scaled score (RCT participants only)
\item {gktmathss} Kindergarten math scaled score (RCT participants only)
\item {gktlistss} Kindergarten listening scaled score (RCT participants only)
\item {gkwordskillss} Kindergarten word study skills scaled score (RCT participants only)
\item {g1schid} Grade 1 School ID
\item {g1tchid} Grade 1 Teacher ID
\item {g1classsize} Class size of Grade 1
\item {g1treadss} Grade 1 SAT reading scaled score
\item {g1tmathss} Grade 1 SAT math scaled score
\item {g1tlistss} Grade 1 total listening scale score in SAT
\item {g1wordskillss} Grade 1 word study skills scale score in SAT
\item {g1readbsraw} Grade 1 reading raw score in Basic Skills First (BSF) tests
\item {g1mathbsraw} Grade 1 math raw score in BSF
\item {g1readbsobjpct} Grade 1 reading percent objectives mastered in BSF tests
\item {g1mathbsobjpct} Grade 1 math percent objectives mastered in BSF tests
\item {g2schid} Grade 2 School ID
\item {g2tchid} Grade 2 Teacher ID
\item {g2classsize} Class size of Grade 2
\item {g2treadss} Grade 2 total reading scale score in SAT
\item {g2tmathss} Grade 2 total math scale score in SAT
\item {g2tlistss} Grade 2 total listening scale score in SAT
\item {g2wordskillss} Grade 2 word study skills scale score in SAT
\item {g2readbsraw} Grade 2 reading raw score in BSF tests
\item {g2mathbsraw} Grade 2 math raw score in BSF test
\item {g2readbsobjpct} Grade 2 reading percent objectives mastered in BSF tests
\item {g3schid} Grade 3 School ID
\item {g3tchid} Grade 3 Teacher ID
\item {g3classsize} Class size of Grade 3
\item {g3treadss} Grade 3 total reading scale score in SAT
\item {g3tmathss} Grade 3 total math scale score in SAT
\item {g3langss} Grade 3 total language scale score in SAT
\item {g3tlistss} Grade 3 total listening scale score in SAT
\item {g3socialsciss} Grade 3 social science scale score in SAT
\item {g3spellss} Grade 3 spelling scale score in SAT
\item {g3vocabss} Grade 3 vocabulary scale score in SAT
\item {g3mathcomputss} Grade 3 math computation scale score in SAT
\item {g3mathnumconcss} Grade 3 concept of numbers scale score in SAT
\item {g3mathapplss} Grade 3 math applications scale score in SAT
\item {g3wordskillss} Grade 3 word study skills scale score in SAT
\item {g3readbsraw} Grade 3 reading raw score in BSF tests
\item {g3mathbsraw} Grade 3 math raw score in BSF tests
\item {g3readbsobjpct} Grade 3 reading percent objectives mastered in BSF tests
\item {g3mathbsobjpct} Grade 3 math percent objectives mastered in BSF tests
\item {dob} Date of birth (with NAs imputed RCT participant median)
\item {dobNA} Dat of birth not recorded
\item {grade_at_entry} Grade at which each student first entered the study
\item {school_at_entry} School ID corresponding to the student's grade
at entry into the study
\item {cond_at_entry} Classroom type corresponding to the student's grade
at entry into the study
}
}
\source{
\doi{doi:10.7910/DVN/SIWH9F}
}
\usage{
STARplus
}
\description{
Data from Tennessee’s Project STAR study.
This data frame describes student participants in the Project
STAR (Student-Teacher Achievement Ratio) field experiment
conducted in Tennessee, USA beginning in the mid-1980s, as well
as an external control group consisting of the contemporaneous
cohort of students attending a matched sample of Tennessee
schools that did not participate in the STAR experiment.
Variables are as described in Project \code{STAR} data
documentation (see references), with five exceptions.
Three \code{*_at_entry} variables were constructed
as follows: \code{grade_at_entry} indicates the grade of student's
first participation, while \code{school_at_entry} and
\code{cond_at_entry} reflect the school ID and classroom type corresponding
to the student's grade at entry to the study. Additionally, \code{read_yr1}
and \code{math_yr1} capture a student's scaled scores on the Scholastic
Assessment Test (SAT) administered to them during their \code{grade_of_entry},
i.e. their earliest available post-treatment SAT measurements.
}
\details{
Note: This dataset bears a Creative Commons Zero license (v1.0).
}
\references{
C.M. Achilles; Helen Pate Bain; Fred Bellott; Jayne Boyd-Zaharias;
Jeremy Finn; John Folger; John Johnston; Elizabeth Word, 2008,
"Tennessee's Student Teacher Achievement Ratio (STAR) project",
Harvard Dataverse, V1, https://doi.org/10.7910/DVN/SIWH9F UNF:3:Ji2Q+9HCCZAbw3csOdMNdA
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GV_data}
\alias{GV_data}
\title{Cluster-randomized experiment data on voter turnout in cable system markets}
\format{
A \code{data.frame} with 248 rows and 7 columns.
\itemize{
\item {age} Age of participant
\item {vote_04} Outcome variable indicating whether participant voted
\item {tv_company} Cable system serving participant's residential area
\item {treatment} Binary variable denoting treatment assignment
\item {pairs} A numeric indicator for the strata or matched pair group to which a cable system belongs (1-3)
\item {population_size} Total population size of residential area served by cable system
\item {sample_size} Number of individuals sampled from the cable system cluster
}
}
\source{
\url{https://dataverse.yale.edu/dataset.xhtml?persistentId=doi:10.60600/YU/OYIPLJ}
}
\usage{
GV_data
}
\description{
This dataset is a toy example derived from a cluster-randomized
field experiment that evaluates the effect of \dQuote{Rock the Vote} TV advertisements
on voter turnout rate. The original study included 23,869 first-time voters
across 85 cable television markets in 12 states. These markets were grouped
into matched sets based on their past voter turnout rates and then randomly
assigned to either a treatment or control condition. This toy dataset is
constructed by randomly sampling 10\% of individuals from selected cable
television markets in the original dataset.
}
\details{
The original dataset was drawn from a randomized controlled trial in
which 85 cable system areas were first grouped into 40 matched sets based on
historical voter turnout. Within each matched set, one cable system area was
randomly assigned to the treatment condition, while the others served
as controls.

This toy dataset includes a subset of the original replication data,
specifically individuals from matched sets 1–3, which encompass 7 of the 85
cable system areas. Within these selected clusters, a 10\% random sample of
individuals was taken.

The fuller Green-Vavreck dataset that this derives from bears a Creative
Commons BY-NC-ND license (v3.0) and is housed in Yale University's Institution
for Social and Policy Studies (ID: D005).
}
\references{
Green, Donald P. & Lynn Vavreck (2008) "Analysis of Cluster-Randomized
Experiments: A Comparison of Alternative Estimation Approaches." Political Analysis 16(2):138-152.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glmrobMethods.R
\name{estfun.glmrob}
\alias{estfun.glmrob}
\alias{bread.glmrob}
\title{Extract empirical estimating equations from a \code{glmbrob} model
fit}
\usage{
\method{estfun}{glmrob}(x, ...)

\method{bread}{glmrob}(x, ...)
}
\arguments{
\item{x}{a fitted \code{lmrob} object}

\item{...}{arguments passed to methods}
}
\value{
matrix, estimating functions evaluated at data points and fitted
parameters

matrix, inverse Hessian of loss as evaluated at fitted parameters
}
\description{
Extract empirical estimating equations from a \code{glmbrob} model
fit

Extract bread matrix from an \code{lmrob()} fit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_data_from_model.R
\name{.get_data_from_model}
\alias{.get_data_from_model}
\title{(Internal) Locate data in call stack}
\usage{
.get_data_from_model(which_fn, form = NULL, by = NULL)
}
\arguments{
\item{which_fn}{Identify calling function, "weights" or "assigned", helps
separate logic for the two functions.}

\item{form}{Formula on which to apply \code{model.frame()}. See details}

\item{by}{translation of unit of assignment/unitid/cluster ID names, passed
down from weights.}
}
\value{
\code{data.frame}
}
\description{
Whenever a function in a model
(\code{ate()}/\code{ett()}/\code{cov_adj()}/\code{assigned()}) is called
without an explicit \code{data=} argument, this will attempt to extract the
data from the model itself.
}
\details{
The \code{form} specifies what columns of the data are needed. For current
use cases (\code{ate()}/\code{ett()} and \code{assigned()}), this will be
only the unit of assignment variables, so e.g. \code{form = ~ uoavar}, to
enable merging of UOA level variables to the model data. However, this can
easily be expanded if other variables are needed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\docType{class}
\name{PreSandwichLayer-class}
\alias{PreSandwichLayer-class}
\title{(Internal) model predictions with some model artifacts, as S4 object}
\description{
(Internal) model predictions with some model artifacts, as S4 object
}
\section{Slots}{

\describe{
\item{\code{.Data}}{numeric vector of predictions}

\item{\code{fitted_covariance_model}}{model standing behind the predictions}

\item{\code{prediction_gradient}}{matrix, predictions gradient w/r/t model params}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{show,StudySpecification-method}
\alias{show,StudySpecification-method}
\title{Show a \code{StudySpecification}}
\usage{
\S4method{show}{StudySpecification}(object)
}
\arguments{
\item{object}{\code{StudySpecification} object, usually a result of a call
to \code{\link[=rct_spec]{rct_spec()}}, \code{\link[=obs_spec]{obs_spec()}}, or \code{\link[=rd_spec]{rd_spec()}}.}
}
\value{
\code{object}, invisibly.
}
\description{
Display information about a \code{StudySpecification} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.rcorrect}
\alias{.rcorrect}
\title{(Internal) Bias correct residuals contributing to standard errors of
a \code{teeMod}}
\usage{
.rcorrect(resids, x, model, type, ...)
}
\arguments{
\item{resids}{numeric vector of residuals to correct}

\item{x}{teeMod object}

\item{model}{string indicating which model the residuals are from.
\code{"itt"} indicates correction to the residuals of \code{x}, and
\code{"cov_adj"} indicates correction to the residuals of the covariance
adjustment model. This informs whether corrections should use information
from \code{x} or the \code{fitted_covariance_model} slot of the
\code{SandwichLayer} object in the \code{offset} for corrections.}

\item{type}{string indicating the desired bias correction. Can be one of
\code{"(HC/CR/MB)0"}, \code{"(HC/CR/MB)1"}, or \code{"(HC/CR/MB)2"}}

\item{...}{additional arguments passed from up the call stack; in
particular, the \code{cluster_cols} argument, which informs whether to
cluster and provide CR2 corrections instead of HC2 corrections, as well as
the correction for the number of clusters in the CR1 correction. This may
also include a \code{by} argument.}
}
\description{
(Internal) Bias correct residuals contributing to standard errors of
a \code{teeMod}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.convert_to_data.frame}
\alias{.convert_to_data.frame}
\title{(Internal) Ensures replacement column for \code{StudySpecification}
is a \code{data.frame}.}
\usage{
.convert_to_data.frame(value, specification, type)
}
\arguments{
\item{value}{A \code{vector} or \code{data.frame} containing a replacement.}

\item{specification}{A \code{StudySpecification}}

\item{type}{One of "t", "f", "u" or "b"}
}
\value{
\code{data.frame} containing named column(s)
}
\description{
Helper function for \code{StudySpecification} replacers to ensure
replacement is a properly named \code{data.frame}
}
\details{
When given a replacement set of values (e.g \code{vector} or \code{matrix}),
this ensures that the replacement is a named \code{data.frame}.

Input \code{vector}: Since it cannot be named, a vector can only be used to
replace an existing component. If the existing component has more than 1
column, uses the name of the first column.

Input \code{matrix} or \code{data.frame}: If unnamed and replacing existing
component, must have no more columns than original component. (If less
columns, uses the name of the first few columns.) If named, can replace any
number of columns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.add_mat_sqdif}
\alias{.add_mat_sqdif}
\title{(Internal) Helper function for design-based meat matrix
calculation}
\usage{
.add_mat_sqdif(X, zobs, bid, b, upper = TRUE)
}
\description{
(Internal) Helper function for design-based meat matrix
calculation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/c_weightedStudySpecification.R
\name{c,WeightedStudySpecification-method}
\alias{c,WeightedStudySpecification-method}
\title{Concatenate weights}
\usage{
\S4method{c}{WeightedStudySpecification}(x, ..., warn_dichotomy_not_equal = FALSE)
}
\arguments{
\item{x, }{.. a \code{WeightedStudySpecification} object, typically created
from \code{\link[=ate]{ate()}} or \code{\link[=ett]{ett()}}}

\item{...}{any number of additional \code{WeightedStudySpecification}
objects with equivalent \code{StudySpecification} to \code{x} and
eachother}

\item{warn_dichotomy_not_equal}{if \code{FALSE} (default),
\code{WeightedStudySpecification}s are considered equivalent even if their
\code{dichotomy} differs. If \code{TRUE}, a warning is produced.}
}
\value{
A numeric \code{vector} with the weights concatenated in the input
order.
}
\description{
Given several variations of weights generated from a single
\code{StudySpecification}, combine into a single weight.
}
\details{
Concatenating \code{WeightedStudySpecification} objects with \code{\link[=c]{c()}}
requires both individual \code{WeightedStudySpecification} objects to come
from the same \code{StudySpecification} and have the same target (e.g all
created with \code{\link[=ate]{ate()}} or all created with \code{\link[=ett]{ett()}}, no mixing-and-matching).
All arguments to \code{\link[=c]{c()}} must be \code{WeightedStudySpecification}.

\code{WeightedStudySpecification} objects may be concatenated together
even without having the same \code{@dichotomy} slot. This procedure only
prompts a warning for differing dichotomies if the argument
\code{warn_dichotomy_not_equal} is set to \code{TRUE}.
}
\examples{
data(simdata)
spec <- rct_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
w1 <- ate(spec, data = simdata[1:30,])
w2 <- ate(spec, data = simdata[31:40,])
w3 <- ate(spec, data = simdata[41:50,])
c_w <- c(w1, w2, w3)
c(length(w1), length(w2), length(w3), length(c_w))

spec <- rct_spec(dose ~ unit_of_assignment(uoa1, uoa2), data = simdata)
w1 <- ate(spec, data = simdata[1:10, ], dichotomy = dose >= 300 ~ .)
w2 <- ate(spec, data = simdata[11:30, ], dichotomy = dose >= 200 ~ .)
w3 <- ate(spec, data = simdata[31:50, ], dichotomy = dose >= 100 ~ .)
c_w <- c(w1, w2, w3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.expand_txt}
\alias{.expand_txt}
\title{(Internal) Expand treatment variable from a \code{StudySpecification}
to a dataframe with unit of assignment information}
\usage{
.expand_txt(txt, data, spec)
}
\arguments{
\item{txt}{A dataframe with one column corresponding to the treatment. Can
be dichotomized or as it's stored in \code{spec}}

\item{data}{A dataframe with unit of assignment information}

\item{spec}{A \code{StudySpecification}, used to align unit of assignment
information with \code{txt}}
}
\description{
(Internal) Expand treatment variable from a \code{StudySpecification}
to a dataframe with unit of assignment information
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.StudySpecification.R
\name{summary.StudySpecification}
\alias{summary.StudySpecification}
\alias{print.summary.StudySpecification}
\title{Summarizing \code{StudySpecification} objects}
\usage{
\method{summary}{StudySpecification}(object, ..., treatment_binary = TRUE)

\method{print}{summary.StudySpecification}(x, ..., max_unit_print = 3)
}
\arguments{
\item{object}{\code{StudySpecification} object, usually a result of a call
to \code{\link[=rct_spec]{rct_spec()}}, \code{\link[=obs_spec]{obs_spec()}}, or \code{\link[=rd_spec]{rd_spec()}}.}

\item{...}{Ignored}

\item{treatment_binary}{Should the treatment be dichotomized if
\code{object} contains a \code{dichotomy}? Ignored if \code{object} does
not contain a \code{dichotomy}.}

\item{x}{\code{summary.StudySpecification} object, usually as a result of a
call to \code{\link[=summary.StudySpecification]{summary.StudySpecification()}}}

\item{max_unit_print}{Maximum number of treatment levels to print in
treatment table}
}
\value{
The \code{StudySpecification} or
\code{summary.StudySpecification}object, invisibly
}
\description{
\code{\link[=summary]{summary()}} method for class \code{StudySpecification}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_data_from_model.R
\name{.fallback_data_search}
\alias{.fallback_data_search}
\title{(Internal) Fallback brute force method to locate \code{data} in the
call stack.}
\usage{
.fallback_data_search()
}
\value{
If found, the data.
}
\description{
We try to be intelligent about finding the appropriate data. If this fails,
we may have need for a brute force method that just loops through frames and
looks for a \code{data} object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{.rename_model_frame_columns}
\alias{.rename_model_frame_columns}
\title{(Internal) Rename columns to strip function calls}
\usage{
.rename_model_frame_columns(modframe)
}
\arguments{
\item{modframe}{A \code{data.frame}.}
}
\value{
The \code{data.frame} with function calls removed
}
\description{
After calling \code{model.frame()} on the formula input to
\code{.new_StudySpecification()}, the names of the columns will include
function names, e.g. "block(blockvar)". This function strips all these.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_spec.R
\name{.get_spec}
\alias{.get_spec}
\title{(Internal) Locate a \code{StudySpecification} in the call stack}
\usage{
.get_spec(NULL_on_error = FALSE)
}
\arguments{
\item{NULL_on_error}{if \code{TRUE}, returns \code{NULL} if a
\code{StudySpecification} object is not found rather than throwing an
error.}
}
\value{
A \code{StudySpecification}, or \code{NULL} if \code{NULL_on_error}
is \code{TRUE} and the \code{StudySpecification} can't be found.
}
\description{
\code{assigned()}/\code{ate()}/\code{ett()}/\code{cov_adj()} all need the
\code{StudySpecification} to operate. If any are called in the model without
a \code{specification=} argument, this function sees if it can find the
\code{StudySpecification} in another of these functions.
}
\details{
Note that it will never look inside \code{assigned()} (gets complicated in
formulas), only in weights or \code{cov_adj()}. E.g.

\code{lm(y ~ assigned(), weights = ate(spec), offest = cov_adj(mod1))}

\code{lm(y ~ assigned(), weights = ate(), offest = cov_adj(mod1, specification = spec))}

will both work, but

\code{lm(y ~ assigned(spec), weights = ate(), offest = cov_adj(mod1))}

will fail.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.add_vec}
\alias{.add_vec}
\title{(Internal) Helper function for design-based meat matrix calculation}
\usage{
.add_vec(a, upper = TRUE)
}
\description{
(Internal) Helper function for design-based meat matrix calculation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{as.SandwichLayer}
\alias{as.SandwichLayer}
\title{Convert a \code{PreSandwichLayer} to a \code{SandwichLayer} with a
\code{StudySpecification} object}
\usage{
as.SandwichLayer(x, specification, by = NULL, Q_data = NULL)
}
\arguments{
\item{x}{a \code{PreSandwichLayer} object}

\item{specification}{a \code{StudySpecification} object}

\item{by}{optional; a string or named vector of unique identifier columns in
the data used to create \code{specification} and the data used to fit the
covariance adjustment model. Default is NULL, in which case unit of
assignment columns are used for identification (even if they do not
uniquely identify units of observation). If a named vector is provided,
names should represent variables in the data used to create
\code{specification}, while values should represent variables in the
covariance adjustment data.}

\item{Q_data}{dataframe of direct adjustment sample, which is needed to
generate the \code{keys} slot of the \code{SandwichLayer} object. Defaults
to NULL, in which case if \code{by} is NULL, the data used to create
\code{specification} is used, and if \code{by} is not NULL, appropriate
data further up the call stack (passed as arguments to \code{cov_adj()} or
\code{lmitt.formula()}, for example) is used.}
}
\value{
a \code{SandwichLayer} object
}
\description{
\code{as.SandwichLayer()} uses the \code{StudySpecification}
object passed to the \code{specification} argument to populate the slots
in a \code{SandwichLayer} object that a \code{PreSandwichLayer} does not
have sufficient information for.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{.sanitize_C_ids}
\alias{.sanitize_C_ids}
\title{(Internal) Return ID's used to order observations in the covariance
adjustment sample}
\usage{
.sanitize_C_ids(x, id_col = NULL, sorted = FALSE, ...)
}
\arguments{
\item{x}{a \code{SandwichLayer} object}

\item{id_col}{character vector or list; optional. Specifies column names that
appear in botn the covariance adjustment and direct adjustmet samples.
Defaults to NULL, in which case unit of assignment columns in the
\code{SandwichLayer}'s \code{StudySpecification} slot will be used to
generate ID's.}

\item{...}{arguments passed to methods}
}
\value{
A vector of length equal to the number of units of observation used
to fit the covariance adjustment model
}
\description{
(Internal) Return ID's used to order observations in the covariance
adjustment sample
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\docType{class}
\name{SandwichLayer-class}
\alias{SandwichLayer-class}
\title{(Internal) model predictions with more model artifacts, as S4 object}
\description{
Contains PreSandwichLayer class.  Only additional slots listed here.
}
\section{Slots}{

\describe{
\item{\code{keys}}{a data.frame}

\item{\code{StudySpecification}}{a StudySpecification}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_dof}
\alias{.get_dof}
\title{(Internal) Compute the degrees of freedom of a contrast of a sandwich
variance estimate associated with a \code{teeMod}}
\usage{
.get_dof(
  x,
  ell,
  vcov.type,
  dof.type = c("stata", "IK"),
  cluster_ids = NULL,
  cluster = NULL,
  ...
)
}
\arguments{
\item{x}{\code{teeMod} object.}

\item{ell}{numeric.}

\item{vcov.type}{character.}

\item{dof.type}{character, either \code{IK} or \code{stata}.}

\item{cluster_ids}{optional, vector of ID's for clustering degrees of freedom
estimate. If not provided, default is the ID's associated with
\code{cluster}.}

\item{cluster}{optional, character identifiying the clustering variable
if \code{cluster_ids} is not provided. If not provided, defaults to the
unit of assignment columns specified in the \code{StudySpecification}.}

\item{...}{Additional arguments passed from calls higher up the stack. These
arguments are not used within this function.}
}
\description{
(Internal) Compute the degrees of freedom of a contrast of a sandwich
variance estimate associated with a \code{teeMod}
}
\details{
With \code{dof_type="IK"}, one obtains a degrees of freedom estimate
as given by the procedure propose in Imbens and Kolesár (2016). For more
details see the documentation of \code{.compute_IK_dof()}. With
\code{dof_type="stata"}, one obtains degrees of freedom equal to the number
of clusters less one, the default provided by STATA software (see Cameron
and Miller, 2015 or Bell and McCaffrey, 2002).

\code{ell} should be an integer specifying a location in the vector
of estimated coefficients (ignoring the coefficients suffixed by
\code{:(Intercept)}) or a vector of the same length as the vector of
estimated coefficients (ignoring the coefficients suffixed by
\code{:(Intercept)}).

\code{vcov.type} takes the same arguments as
the \code{type} argument in \code{vcov_tee()}.

\code{cluster_ids} should be ordered in alignment with the dataframe passed
to \code{lmitt()}. It should not exclude NA's because the function will
exclude them where necessary.
}
\references{
Guido W. Imbens and Michael Kolesár. Robust Standard Errors in
Small Samples: Some Practical Advice".
\emph{The Review of Economics and Statistics}, 98(4):701-712, October 2016.

A. Colin Cameron and Douglas L. Miller. A Practitioner's Guide to
Cluster-Robust Inference. \emph{The Journal of Human Resources}, 50(2):317-372,
2015.

Robert M. Bell and Daniel F. McCaffrey. Bias Reduction in
Standard Errors for Linear Regression with Multi-Stage Samples.
\emph{Survey Methodology}, 28(2):169-181, December 2002.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_a22_inverse}
\alias{.get_a22_inverse}
\alias{.get_a11_inverse}
\alias{.get_a21}
\alias{.get_tilde_a22_inverse}
\alias{.get_tilde_a21}
\title{(Internal) Estimate components of the sandwich covariance matrix
returned by \code{vcov_tee()}}
\usage{
.get_a22_inverse(x, values_from = x, ...)

.get_a11_inverse(x)

.get_a21(x, ...)

.get_tilde_a22_inverse(x, values_from = x, ...)

.get_tilde_a21(x)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{values_from}{optional, a fitted model object. Defaults to \code{x}.}

\item{...}{arguments passed to \code{bread} method}
}
\value{
\code{.get_a22_inverse()}/\code{.get_tilde_a22_inverse()}: A
\eqn{2\times 2} matrix corresponding to an intercept and the treatment
variable in the direct adjustment model

\code{.get_a11_inverse()}: A \eqn{p\times p} matrix where \eqn{p} is
the dimension of the covariance adjustment model, including an intercept

\code{.get_a21()}/\code{.get_tilde_a21()}: A \eqn{2\times p} matrix
where the number of rows are given by the intercept and the treatment
variable in the direct adjustment model, and the number of columns are
given by the dimension of the covariance adjustment model
}
\description{
(Internal) Estimate components of the sandwich covariance matrix
returned by \code{vcov_tee()}
}
\details{
\code{.get_a22_inverse()}/\code{.get_tilde_a22_inverse()}:
\eqn{A_{22}^{-1}} is the "bread" of the sandwich covariance matrix returned
by \code{vcov_tee()} whether one has fit a prior covariance adjustment
model or not.

\code{.get_a11_inverse()}: \eqn{A_{11}^{-1}} is the "bread" of the
sandwich covariance matrix for the covariance adjustment model. This matrix
contributes to the meat matrix of the direct adjustment sandwich covariance
matrix.

\code{.get_a21()}/\code{.get_tilde_a21()}: \eqn{A_{21}} is the
gradient of the estimating equations for the direct adjustment model taken
with respect to the covariance adjustment model parameters. This matrix is
the crossproduct of the prediction gradient for the units of observation in
\eqn{\mathcal{Q}} and the model matrix of the direct adjustment model.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.cov_mat_est}
\alias{.cov_mat_est}
\title{(Internal) Helper function for design-based meat matrix
calculation}
\usage{
.cov_mat_est(XXz, bidz)
}
\value{
estimated upper and lower bounds of covariance matrix of estimating
function vectors under either treatment or control
}
\description{
(Internal) Helper function for design-based meat matrix
calculation
}
\details{
Diagonal elements are estimated by sample variances Off-diagonal
elements are estimated using the Young's elementary inequality
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cov_adj.R
\name{default}
\alias{default}
\title{Instruct \code{cov_adj()} to find a default reference value for columns in
a covariance adjustment model}
\usage{
default()
}
\value{
empty list that inherits from class \code{set_to_reference_default}
}
\description{
Instruct \code{cov_adj()} to find a default reference value for columns in
a covariance adjustment model
}
\details{
When set as the entry of a list passed to the \code{set_to_reference}
argument of \code{cov_adj()}, \code{cov_adj()} uses the following
replacement values depending on the column type:
\itemize{
\item factor/character: minimum level
\item numeric: minimum value
\item logical: FALSE
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.apply_dichotomy}
\alias{.apply_dichotomy}
\title{(Internal) Applies dichotomy to treatment}
\usage{
.apply_dichotomy(txt, dichotomy)
}
\arguments{
\item{txt}{A named \code{data.frame} containing a column of the treatment,
such as that produed by \code{treatment(myspecification)}, and any variables
specified in \code{dichotomy}.}

\item{dichotomy}{A formula specifying how to dichotomize the non-binary
treatment column in \code{txt} (or a call that evaluates to a formula).
See the Details section of the \code{ett()} or \code{att()} help pages for
information on specifying this formula}
}
\value{
A \code{vector} of binary treatments
}
\description{
Given a dichotomy formula and a \code{data.frame} with a
treatment variable and any variables in the formula, returns a
\code{vector} containing only \code{0}, \code{1}, or \code{NA}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{schooldata}
\alias{schooldata}
\alias{studentdata}
\title{Student data}
\format{
Two \code{data.frame}s, one with school-level data (\code{schooldata})
including treatment assignment and a second with student-level data
(\code{studentdata}).
\code{schoolata}:
\itemize{
\item {schoolid} Unique school ID variable.
\item {treatment} Was this school in the intervention group?
\item {state} State which the school is in.
\item {pct}_disadvantage Percent of student body flagged as "disadvantaged".
}
\code{studentdata}:
\itemize{
\item {id} Unique student ID.
\item {schoolid} Unique school ID variable.
\item {grade} Student's grade, 3-5.
\item {gpa} Student GPA in prior year.
\item {math} Standarized math score (out of 100).
}

An object of class \code{data.frame} with 8713 rows and 5 columns.
}
\usage{
schooldata

studentdata
}
\description{
An example of data sets stored at two levels.
}
\details{
In this hypothetical data, schools were randomly assignment to
treatment status, but the unit of analysis is students. Thus the two data
sets, one encoding school information (including treatment status) and one
encoding student information (which does not include treatment status).
}
\examples{
soec <- obs_spec(treatment ~ uoa(schoolid), data = schooldata)

# Treatment effect
mod1 <- lmitt(math ~ 1, specification = soec, data = studentdata)

# Treatment effect by grade
mod2 <- lmitt(math ~ as.factor(grade), specification = soec, data = studentdata)
}
\keyword{dataset}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.get_col_from_new_data}
\alias{.get_col_from_new_data}
\title{(Internal) Extract specified \code{type} from new data set}
\usage{
.get_col_from_new_data(
  specification,
  newdata,
  type,
  by = NULL,
  implicitBlock = FALSE,
  ...
)
}
\arguments{
\item{specification}{A \code{StudySpecification}}

\item{newdata}{A \code{data.frame}, which may or may not be the one which
was used to create \code{specification}. It must have the units of
assignment variable(s) (though \code{by=} argument can be used if the name
differ), and will appropriately merge with the \code{specification} the
blocks, treatment or forcings.}

\item{type}{One of "t", "f", or "b".}

\item{by}{optional; named vector or list connecting names of unit of
assignment/unitid/cluster variables in \code{specification} to unit of
assignment/unitid/cluster variables in \code{data}. Names represent
variables in the StudySpecification; values represent variables in the
data. Only needed if variable names differ.}

\item{implicitBlock}{If the \code{StudySpecification} does not include a
block, \code{TRUE} will return a constant 1 for the blocks if \code{type}
requests it.}

\item{...}{Additional arguments to \code{merge()}.}
}
\value{
The column(s) belonging to the requested \code{type} in
}
\description{
(Internal) Extract specified \code{type} from new data set
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{confint.teeMod}
\alias{confint.teeMod}
\title{Confidence intervals with standard errors provided by
\code{vcov.teeMod()}}
\usage{
\method{confint}{teeMod}(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{a fitted \code{teeMod} model}

\item{parm}{a specification of which parameters are to be given confidence
intervals, either a vector of numbers or a vector of names. If missing,
all parameters are considered.}

\item{level}{the confidence level required.}

\item{...}{additional arguments to pass to \code{vcov.teeMod()}}
}
\value{
A matrix (or vector) with columns giving lower and upper confidence
limits for each parameter. These will be labelled as (1-level)/2 and 1 -
(1-level)/2 in \% (by default 2.5\% and 97.5\%)
}
\description{
An S3method for \code{stats::confint} that uses standard errors
computed using \code{vcov.teeMod()}. Additional arguments passed to this
function, such as \code{cluster} and \code{type}, specify the arguments of
the \code{vcov.teeMod()} call.
}
\details{
Rather than call \code{stats::confint.lm()},
\code{confint.teeMod()} calls \code{.confint_lm()}, a function internal to
the \code{propertee} package that ensures additional arguments in the
\code{...} of the \code{confint.teeMod()} call are passed to the internal
\code{vcov()} call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{specification_data_concordance}
\alias{specification_data_concordance}
\title{Check for variable agreement within units of assignment}
\usage{
specification_data_concordance(
  specification,
  data,
  by = NULL,
  warn_on_nonexistence = TRUE
)
}
\arguments{
\item{specification}{a \code{StudySpecification} object}

\item{data}{a new data set, presumably not the same used to create
\code{specification}.}

\item{by}{optional; named vector or list connecting names of variables in
\code{specification} to variables in \code{data}. Names represent
variables in \code{specification}; values represent variables in
\code{data}. Only needed if variable names differ.}

\item{warn_on_nonexistence}{default \code{TRUE}. If a variable does not
exist in \code{data}, should this be flagged? If \code{FALSE}, silently
move on if a variable doesn't exist in \code{data}.}
}
\value{
invisibly \code{TRUE} if no warnings are produced, \code{FALSE} if
any warnings are produced.
}
\description{
Useful for debugging purposes to ensure that there is
concordance between variables in the \code{StudySpecification} and data.
}
\details{
Consider the following scenario: A \code{StudySpecification} is
generated from some dataset, "data1", which includes a block variable
"b1". Within each unique unit of assignment/unitid/cluster of "data1", it
must be the case that "b1" is constant. (Otherwise the creation of the
\code{StudySpecification} will fail.)

Next, a model is fit which includes weights generated from the
\code{StudySpecification}, but on dataset "data2". In "data2", the block
variable "b1" also exists, but due to some issue with data cleaning, does
not agree with "b1" in "data1".

This could cause errors, either directly (via actual error messages) or
simply produce nonsense results. \code{\link[=specification_data_concordance]{specification_data_concordance()}} is
specificationed to help debug these scenarios by providing information on
whether variables in both the data used in the creation of
\code{specification} ("data1" in the above example) and some new dataset,
\code{data}, ("data2" in the above example) have any inconsistencies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.compute_IK_dof}
\alias{.compute_IK_dof}
\title{(Internal) Compute the degrees of freedom of a contrast of a
sandwich variance estimate as proposed in Imbens and Kolesár (2016)}
\usage{
.compute_IK_dof(
  tm,
  ell,
  vcov.type,
  cluster_ids = NULL,
  cluster = NULL,
  tol = 1e-09
)
}
\arguments{
\item{tm}{\code{teeMod} object.}

\item{ell}{numeric vector.}

\item{vcov.type}{character.}

\item{cluster_ids}{optional, vector of ID's for clustering degrees of freedom
estimate. If not provided, default is the ID's associated with
\code{cluster}.}

\item{cluster}{optional, character identifiying the clustering variable
if \code{cluster_ids} is not provided. If not provided, defaults to the
unit of assignment columns specified in the \code{StudySpecification}.}

\item{tol}{optional, numeric. Should not be changed.}
}
\description{
(Internal) Compute the degrees of freedom of a contrast of a
sandwich variance estimate as proposed in Imbens and Kolesár (2016)
}
\details{
\code{ell} should be a vector of length equal to the number of
estimated coefficients in the \code{teeMod} object. This excludes
coefficients printed in \code{show.teeMod} with \code{:(Intercept)}
suffixes. The degrees of freedom for a single standard error will specify
for \code{ell} a vector of all zeros except one element, which will have a
1 in the location corresponding to the coefficient of interest.

\code{cluster_ids} should be ordered in alignment with the dataframe passed
to \code{lmitt()}. It should not exclude NA's because the function will
exclude them where necessary.

\code{vcov.type} takes the same arguments
as the \code{type} argument in \code{vcov_tee()}.
}
\references{
Guido W. Imbens and Michael Kolesár. Robust Standard Errors in
Small Samples: Some Practical Advice".
\emph{The Review of Economics and Statistics}, 98(4):701-712, October 2016.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightedStudySpecification.R
\name{weights,WeightedStudySpecification-method}
\alias{weights,WeightedStudySpecification-method}
\title{Extract Weights from \code{WeightedStudySpecification}}
\usage{
\S4method{weights}{WeightedStudySpecification}(object, ...)
}
\arguments{
\item{object}{a \code{WeightedStudySpecification} object}

\item{...}{Ignored}
}
\value{
A numeric \code{vector} of the weights
}
\description{
A \code{WeightedStudySpecification} object contains a numeric
vector with a few additional slots, this extracts only the numeric vector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_phi_tilde}
\alias{.get_phi_tilde}
\title{(Internal) Calculate grave\{phi\}}
\usage{
.get_phi_tilde(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\description{
(Internal) Calculate grave\{phi\}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights_internal.R
\name{.weights_calc}
\alias{.weights_calc}
\title{(Internal) Worker function for weight calculation}
\usage{
.weights_calc(specification, target, weightAlias, dichotomy, by, data)
}
\arguments{
\item{specification}{a \code{StudySpecification} object created by one of
\code{rct_spec()}, \code{rd_spec()}, or \code{obs_spec()}.}

\item{target}{One of "ate" or "ett"; \code{ate()} and \code{ett()} chooses
these automatically.}

\item{weightAlias}{An alias for the weight target, currently one of "ate",
"ett", "att". Chosen by \code{ate()} and \code{ett()} automatically.}

\item{dichotomy}{optional; a formula defining the dichotomy of the treatment
variable if it isn't already \code{0}/\code{1}. See details of help for
\code{ate()} or \code{ett()} e.g. for details.}

\item{by}{optional; named vector or list connecting names of unit of
assignment/ variables in \code{specification} to unit of
assignment/cluster variables in \code{data}. Names represent variables in
the StudySpecification; values represent variables in the data. Only
needed if variable names differ.}

\item{data}{optionally the data for the analysis to be performed on. May be
excluded if these functions are included as the \code{weights} argument of
a model.}
}
\value{
a \code{WeightedStudySpecification} object
}
\description{
Called from \code{ate()} or \code{ett()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_preserve_order.R
\name{.merge_preserve_order}
\alias{.merge_preserve_order}
\title{(Internal) Merge \code{data.frame}s ensuring order of first
\code{data.frame} is maintained}
\usage{
.merge_preserve_order(x, ...)
}
\arguments{
\item{x}{\code{data.frame} whose ordering is to be maintained}

\item{...}{Additional arguments to \code{merge()}, particularly a second
\code{data.frame} and a \code{by=} argument.}
}
\value{
Merged \code{data.frame} with the same ordering as \code{x}.
}
\description{
(Internal) Merge \code{data.frame}s ensuring order of first
\code{data.frame} is maintained
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{cluster_iss}
\alias{cluster_iss}
\title{(Internal) Compute \eqn{(I_{i} - H_{ii})^{-1/2}} as part of CR2
variance estimates}
\usage{
cluster_iss(tm, cluster_unit, cluster_ids = NULL, ...)
}
\arguments{
\item{tm}{\code{teeMod} object.}

\item{cluster_unit}{cluster to subset observations to. Must be found in
\code{cluster_ids}.}

\item{cluster_ids}{optional, ID's for clustering standard errors. If not
provided, default is the ID's associated with \code{cluster} if provided,
otherwise the unit of assignment ID's.}

\item{...}{Additional arguments passed from calls higher in the stack. One
that may be used is \code{cluster}, which identifies the clustering variable
if \code{cluster_ids} is not provided. If \code{cluster} is not provided,
the default is the unit of assignment variable specified in the
\code{StudySpecification}.}
}
\description{
(Internal) Compute \eqn{(I_{i} - H_{ii})^{-1/2}} as part of CR2
variance estimates
}
\details{
The notation \eqn{I_{i}} and \eqn{H_{ii}} comes from Bell and
McCaffrey (2002). The matrix \eqn{I_{i}} is an identity matrix with number
of rows equal to the number of observations in cluster \eqn{i}, and
\eqn{H_{ii}} subsets the hat matrix associated with the regression fit
stored in \code{tm} to the rows associated with observations in cluster
\eqn{i}.

When possible, the function uses the method in Wasserman (2026)
to cheaply compute the inverse symmetric square root.

\code{cluster_ids} should be ordered in alignment with the dataframe passed
to \code{lmitt()}. It should not exclude NA's because the function will
exclude them.
}
\references{
Robert M. Bell and Daniel F. McCaffrey. Bias Reduction in
Standard Errors for Linear Regression with Multi-Stage Samples.
\emph{Survey Methodology}, 28(2):169-181, December 2002.

Joshua Wasserman. Methods for Causal Inference in Settings with Clustered
Data Subject to Missingness and Measurement Error. Unpublished thesis,
June 2026.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.specification_accessors_newdata_validate}
\alias{.specification_accessors_newdata_validate}
\title{(Internal) Checks newdata/by argument for specification accessors}
\usage{
.specification_accessors_newdata_validate(newdata, by)
}
\arguments{
\item{newdata}{newdata argument from e.g. \code{treatment()},
\code{blocks()}, etc}

\item{by}{from e.g. \code{treatment()}, \code{blocks()}, etc. See
\code{.check_by()}}
}
\value{
Invisibly \code{TRUE}. Warns or errors as appropriate.
}
\description{
(Internal) Checks newdata/by argument for specification accessors
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_by.R
\name{.check_by}
\alias{.check_by}
\title{(Internal) A few checks to ensure \code{by=} is valid}
\usage{
.check_by(by)
}
\arguments{
\item{by}{named vector or list connecting names of unit of
assignment/unitid/cluster variables in \code{specification} to unit of
assignment/unitid/cluster variables in \code{data}. Names represent
variables in the StudySpecification; values represent variables in the
data.}
}
\value{
\code{NULL} if no errors are found
}
\description{
Thie ensures that the \code{by=} argument is of the proper type, is named,
and consists of only unique entries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cov_adj.R
\name{cov_adj}
\alias{cov_adj}
\title{Prepare prognostic regression offset for \code{lmitt()} or \code{lm()}}
\usage{
cov_adj(
  model,
  newdata = NULL,
  specification = NULL,
  by = NULL,
  set_to_reference = NULL
)
}
\arguments{
\item{model}{any model that inherits from a \code{glm}, \code{lm}, or
\code{robustbase::lmrob} object.}

\item{newdata}{a dataframe; optional. See Details for how \code{newdata}
is found in the call stack if none is provided.}

\item{specification}{a \code{StudySpecification} object; optional. If not
provided, it can be retrieved from the call stack if passed to suitable
calls.}

\item{by}{named vector; optional.}

\item{set_to_reference}{named list; optional.}
}
\value{
A \code{SandwichLayer} if \code{specification} is not NULL or a
\code{StudySpecification} object is found in the call stack, otherwise a
\code{PreSandwichLayer} object
}
\description{
\code{cov_adj()} takes a fitted model object and predicts
outcomes under the counterfactual of being in the control condition.
Using these predictions as offsets in a linear model with just the
treatment variable as a regressor estimates the treatment effect with
covariance adjustments as provided by the previously fitted model.
In the special case that this covariance model was fit using only
(quasi-)experimental control group observations, the method is known
as Peters-Belson regression. \code{cov_adj()} assists the process by
reporting its predictions on the scale of the response (in contrast to
\code{predict.glm()}) and by retaining certain artifacts from
covariance model fitting.  These artifacts are used by \code{\link[=vcov_tee]{vcov_tee()}},
and by \code{\link[stats:vcov]{stats::vcov()}} as applied to a \code{teeMod} object, so
that the covariance model's contribution to sampling variability
is reflected in treatment coefficient standard errors.
\code{cov_adj()} can be used with \code{lm()}, but is better supported
in combination with propertee's \code{lm}-wrapper, \code{\link[=lmitt]{lmitt()}}.
}
\details{
If \code{specification} is provided or can be found in the call
stack, then prior to generating adjustments, \code{cov_adj()} will
identify the treatment variable and replace all values with a reference
level. If the treatment has logical type, this reference level is
\code{FALSE}, and if it has numeric type, this is the smallest
non-negative value (which means 0 for 0/1 binary). (Factor treatments are
not currently supported.)\cr\cr More
broadly, the \code{set_to_reference} argument can be used to set columns
to values specified by the user prior to generating predictions. The
argument takes a named list where the names specify the columns whose
values should be replaced, and the entries specify the replacement values.
If an entry is given as \code{default()}, internal routines will determine
the reference level (the first level of a factor/factorized character
variable or the minimum of a numeric/logical variable). \cr\cr If
\code{newdata} is provided, the values of the output vector represent
adjustments for the outcomes in \code{newdata}. If \code{cov_adj()} is
called in the \code{offset} argument of \code{lmitt()}, predictions will
correspond to rows in the \code{data} argument of \code{lmitt()}.
Otherwise, adjustments are made based on the data to which \code{model} is
fit. \cr\cr The \code{by} argument specifies columns for merging datasets.
The names of the argument specify columns in the dataframe for building
\code{specification}, while the entries specify columns in the dataframe
for fitting \code{model}.
}
\examples{
data("STARplus")

##' A prognostic model fitted to experimental + non-experimental controls
y0hat_read <- lm(read_yr1 ~ gender*dob +dobNA + race,
                 data = STARplus,
                 subset = cond_at_entry!="small")

STARspec <- rct_spec(cond_at_entry ~ unit_of_assignment(stdntid) +
                         block(grade_at_entry, school_at_entry),
                     subset=!is.na(grade_at_entry),# excludes non-experimentals
                     data = STARplus)
ett_wts    <- ett(STARspec, data = STARplus,
                  dichotomy= cond_at_entry =="small" ~.)

ett_read <- lm(read_yr1 ~ assigned(dichotomy= cond_at_entry =="small" ~.),
               offset = cov_adj(y0hat_read),
               data = STARplus,
               weights = ett_wts)
coef(ett_read)
ett_read |> as.lmitt() # brings in control-group means of outcome, predictions

ate_read <- lmitt(read_yr1 ~ 1, STARspec, STARplus,
                  dichotomy= cond_at_entry =="small" ~.,
                  offset = cov_adj(y0hat_read),
                  weights = "ate")
show(ate_read)
vcov(ate_read, type = "HC0", cov_adj_rcorrect = "HC0") |> unname()

ate_read_loc <-
    lmitt(read_yr1 ~ race, STARspec, STARplus,
          dichotomy= cond_at_entry =="small" ~.,
          offset = cov_adj(y0hat_read, newdata = STARplus),
          weights = "ate")
show(ate_read_loc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{treatment}
\alias{treatment}
\alias{treatment,StudySpecification-method}
\alias{treatment<-}
\alias{treatment<-,StudySpecification-method}
\alias{units_of_assignment}
\alias{units_of_assignment,StudySpecification-method}
\alias{units_of_assignment<-}
\alias{units_of_assignment<-,StudySpecification-method}
\alias{clusters}
\alias{clusters,StudySpecification-method}
\alias{clusters<-}
\alias{clusters<-,StudySpecification-method}
\alias{unitids}
\alias{unitids,StudySpecification-method}
\alias{unitids<-}
\alias{unitids<-,StudySpecification-method}
\alias{blocks}
\alias{blocks,StudySpecification-method}
\alias{blocks<-}
\alias{blocks<-,StudySpecification-method}
\alias{has_blocks}
\alias{forcings}
\alias{forcings,StudySpecification-method}
\alias{forcings<-}
\alias{forcings<-,StudySpecification-method}
\title{Accessors and Replacers for \code{StudySpecification} objects}
\usage{
treatment(x, newdata = NULL, dichotomy = NULL, by = NULL, ...)

\S4method{treatment}{StudySpecification}(x, newdata = NULL, dichotomy = NULL, by = NULL, ...)

treatment(x) <- value

\S4method{treatment}{StudySpecification}(x) <- value

units_of_assignment(x, newdata = NULL, by = NULL)

\S4method{units_of_assignment}{StudySpecification}(x, newdata = NULL, by = NULL)

units_of_assignment(x) <- value

\S4method{units_of_assignment}{StudySpecification}(x) <- value

clusters(x, newdata = NULL, by = NULL)

\S4method{clusters}{StudySpecification}(x, newdata = NULL, by = NULL)

clusters(x) <- value

\S4method{clusters}{StudySpecification}(x) <- value

unitids(x)

\S4method{unitids}{StudySpecification}(x)

unitids(x) <- value

\S4method{unitids}{StudySpecification}(x) <- value

blocks(x, newdata = NULL, by = NULL, ...)

\S4method{blocks}{StudySpecification}(x, newdata = NULL, by = NULL, ..., implicit = FALSE)

blocks(x) <- value

\S4method{blocks}{StudySpecification}(x) <- value

has_blocks(x)

forcings(x, newdata = NULL, by = NULL)

\S4method{forcings}{StudySpecification}(x, newdata = NULL, by = NULL)

forcings(x) <- value

\S4method{forcings}{StudySpecification}(x) <- value
}
\arguments{
\item{x}{a \code{StudySpecification} object}

\item{newdata}{optional; an additional \code{data.frame}. If passed, and the
unit of assignment variable is found in \code{newdata}, then the requested
variable type for each unit of \code{newdata} is returned. See \code{by}
argument if the name of the unit of assignment differs.}

\item{dichotomy}{optional; a formula specifying how to dichotomize a
non-binary treatment variable. See the Details section of the \code{ett()}
or \code{att()} help pages for information on specifying this formula}

\item{by}{optional; named vector or list connecting names of unit of
assignment/unitid/cluster variables in \code{x} to unit of
assignment/unitid/cluster variables in \code{data}. Names represent
variables in \code{x}; values represent variables in \code{newdata}. Only
needed if variable names differ.}

\item{...}{ignored.}

\item{value}{replacement. Either a \code{vector}/\code{matrix} of
appropriate dimension, or a named \code{data.frame} if renaming variable
as well. See \code{Details}.}

\item{implicit}{Should a block-less \code{StudySpecification} return a
constant 1 when extracting \code{blocks}?}
}
\value{
\code{data.frame} containing requested variable, or an updated
\code{StudySpecification}. \code{\link[=treatment]{treatment()}} works slightly differently, see
\code{Details}.
}
\description{
Allows access to the elements which define a
\code{StudySpecification}, enabling their extraction or replacement.
}
\details{
For \code{\link[=treatment]{treatment()}}, when argument \code{binary} is \code{FALSE}, the
treatment variable passed into the \code{StudySpecification} is returned
as a one-column \code{data.frame} regardless of whether it is binary or
\code{x} has a \code{dichotomy}

If a \code{dichotomy} is passed, a binary one-column \code{data.frame}
will be returned. If not and \code{binary} is \code{TRUE}, unless the
\code{StudySpecification} has a binary treatment, \code{\link[=treatment]{treatment()}} will
error. If \code{binary} is \code{"ifany"}, it will return the original
treatment in this case.

The one-column \code{data.frame} returned by \code{\link[=treatment]{treatment()}} is named as
entered in the \code{StudySpecification} creation, but if a
\code{dichotomy} is passed, the column name is \code{"__z"} to try and
avoid any name conflicts.

For the \code{value} when using replacers, the replacement must have the
same number of rows as the \code{StudySpecification} (the same number of
units of assignment). The number of columns can differ (e.g. if the
\code{StudySpecification} were defined with two variable uniquely
identifying blocks, you can replace that with a single variable uniquely
identifying blocks, as long as it respects other restrictions.)

If the replacement value is a \code{data.frame}, the name of the columns
is used as the new variable names. If the replacement is a \code{matrix}
or \code{vector}, the original names are retained. If reducing the number
of variables (e.g., moving from two variables uniquely identifying to a
single variable), the appropriate number of variable names are retained.
If increasing the number of variables, a \code{data.frame} with names must
be provided.
}
\examples{
data(simdata)
spec <- obs_spec(z ~ unit_of_assignment(uoa1, uoa2), data = simdata)
blocks(spec) # empty
blocks(spec) <- data.frame(blks = c(1, 1, 2, 2, 3, 3, 4, 4, 5, 5))
blocks(spec)
blocks(spec) <- c(5, 5, 4, 4, 3, 3, 2, 2, 1, 1)
blocks(spec) # notice that variable is not renamed
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.aggregate_to_cluster}
\alias{.aggregate_to_cluster}
\title{(Internal) Aggregate weights and outcomes to cluster level}
\usage{
.aggregate_to_cluster(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
a list of
\itemize{
\item a data frame of cluster weights, outcomes, treatments, and block ids;
\item treatment id column name;
\item block id column name
}
}
\description{
(Internal) Aggregate weights and outcomes to cluster level
}
\details{
aggregate individual weights and outcomes to cluster weighted sums
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{vcov.teeMod}
\alias{vcov.teeMod}
\title{Compute variance-covariance matrix for fitted \code{teeMod} model}
\usage{
\method{vcov}{teeMod}(object, ...)
}
\arguments{
\item{object}{a fitted \code{teeMod} model}

\item{...}{additional arguments to \code{vcov_tee()}.}
}
\value{
A variance-covariance matrix with row and column entries for the estimated
coefficients in \code{x}, the marginal mean outcome in the control condition,
the marginal mean \code{offset} in the control condition (if an \code{offset}
is provided), and if a moderator variable is specified in the formula for \code{x},
the mean interaction in the control condition of the outcome and \code{offset}
with the moderator variable
}
\description{
An S3method for \code{stats::vcov} that computes standard
errors for \code{teeMod} models using \code{vcov_tee()}.
}
\details{
\code{vcov.teeMod()} wraps around \code{vcov_tee()}, so additional
arguments passed to \code{...} will be passed to the \code{vcov_tee()}
call. See documentation for \code{vcov_tee()} for information about
necessary arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{vcov_tee}
\alias{vcov_tee}
\alias{.vcov_DB0}
\alias{.vcov_DB}
\title{Variance/Covariance for \code{teeMod} objects}
\usage{
vcov_tee(x, type = NULL, cluster = NULL, ...)

.vcov_DB0(x, ...)

.vcov_DB(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{type}{a string indicating the desired bias correction for the residuals
of \code{x}. Default makes no bias correction. See Details for supported types}

\item{cluster}{a vector indicating the columns that define clusters. The default
is the unit of assignment columns in the \code{StudySpecification} stored
in \code{x}. These columns should appear in the dataframe used for fitting
\code{x} as well as the dataframe passed to the covariance model fit in the case of
prior covariance adjustment. See Details}

\item{...}{arguments to be passed to the internal variance estimation
function, such as \code{cov_adj_rcorrect} and \code{loco_residuals}. If
\code{x} has a \code{SandwichLayer} object in its offset, The former
specifies the bias correction to the residuals of the covariance model, and
the latter indicates whether the offset should be replaced with predictions
from leave-one-cluster-out fits of the covariance adjustment model.
See Details}
}
\value{
A variance-covariance matrix with row and column entries for the estimated
coefficients in \code{x}, the marginal mean outcome in the control condition,
the marginal mean \code{offset} in the control condition (if an \code{offset}
is provided), and if a moderator variable is specified in the formula for \code{x},
the mean interaction in the control condition of the outcome and \code{offset}
with the moderator variable
}
\description{
Compute robust sandwich variance estimates with optional
covariance adjustment
}
\details{
Variance estimates will be clustered on the basis of the columns
provided to \code{cluster} (or obtained by the default behavior). As a result,
providing \code{"HCx"} or \code{"CRx"} to \code{type} will produce the same
variance estimate given that \code{cluster} remains the same.

With prior covariance adjustment, unless the \code{data} argument of the covariance
model fit is the same as the \code{data} argument for fitting \code{x} and the
\code{StudySpecification} of \code{x} has been created with a formula of the
form \code{trt_col ~ 1}, the column(s) provided to \code{cluster} must appear in the
dataframes in both \code{data} arguments, even if the clustering structure does
not exist, per se, in the covariance adjustment sample. For instance, in a
finely stratified randomized trial, one might desire standard errors clustered
at the block level, but the covariance adjustment model may include auxiliary
units that did not participate in the trial. In this case, in the \code{data}
argument of the fitted covariance model, the column(s) passed to \code{cluster}
should have the block ID's for rows overlapping with the \code{data} argument
used for fitting \code{x}, and NA's for any auxiliary units. \code{vcov_tee()}
will treat each row with an NA as its own cluster.

For ITT effect estimates without covariance adjustment, \code{type}
corresponds to the variance estimate desired. Supported options include:
\itemize{
\item \code{"MB0"}, \code{"HC0"}, and \code{"CR0"} for model-based HC/CR0 standard errors
\item \code{"MB1"}, \code{"HC1"}, and \code{"CR1"} for model-based HC/CR1 standard errors
(for \code{"MB1"} and \code{"HC1"}, this is \eqn{n/(n - 2)}, and for \code{"CR1"},
this is \eqn{g\cdot(n-1)/((g-1)\cdot(n-2))}, where \eqn{g} is the number of
clusters in the sample used for fitting \code{x})
\item \code{"MB2"}, \code{"HC2"}, and \code{"CR2"} for model-based HC/CR2 standard errors
\item \code{"DB0"} for design-based HC0 variance estimates
}

The \code{type} argument does not correspond to existing variance estimators in
the literature in the case of prior covariance adjustment. It specifies the
bias correction to the residuals of \code{x}, but the residuals of the covariance
model are corrected separately based on the \code{cov_adj_rcorrect} argument.
The \code{cov_adj_rcorrect} argument takes the same options as \code{type}
except \code{"DB0"}. When the covariance model includes rows in the treatment
condition for fitting, the residuals of \code{x} are further corrected by
having the values of \code{offset} replaced by predictions that use coefficient
estimates that leave out rows in the same cluster (as defined by the
\code{cluster} argument).

Design-based variance estimation is implemented for
\code{teeMod} models satisfying requirements including:
\itemize{
\item The model uses \code{rct_spec} as \code{StudySpecification}
\item The model only estimates a main treatment effect
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightedStudySpecification.R
\name{+,WeightedStudySpecification,numeric-method}
\alias{+,WeightedStudySpecification,numeric-method}
\alias{+,numeric,WeightedStudySpecification-method}
\alias{-,WeightedStudySpecification,numeric-method}
\alias{-,numeric,WeightedStudySpecification-method}
\alias{*,WeightedStudySpecification,numeric-method}
\alias{*,numeric,WeightedStudySpecification-method}
\alias{/,WeightedStudySpecification,numeric-method}
\alias{/,numeric,WeightedStudySpecification-method}
\title{\code{WeightedStudySpecification} Operations}
\usage{
\S4method{+}{WeightedStudySpecification,numeric}(e1, e2)

\S4method{+}{numeric,WeightedStudySpecification}(e1, e2)

\S4method{-}{WeightedStudySpecification,numeric}(e1, e2)

\S4method{-}{numeric,WeightedStudySpecification}(e1, e2)

\S4method{*}{WeightedStudySpecification,numeric}(e1, e2)

\S4method{*}{numeric,WeightedStudySpecification}(e1, e2)

\S4method{/}{WeightedStudySpecification,numeric}(e1, e2)

\S4method{/}{numeric,WeightedStudySpecification}(e1, e2)
}
\arguments{
\item{e1, e2}{\code{WeightedStudySpecification} or \code{numeric} objects}
}
\value{
a \code{WeightedStudySpecification} object
}
\description{
Algebraic operators on \code{WeightedStudySpecification}
objects and numeric vectors. \code{WeightedStudySpecification}s do not
support addition or subtraction.
}
\details{
These are primarily used to either combine weights via
multiplication, or to invert weights. Addition and subtraction are not
supported and will produce errors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.update_call_formula}
\alias{.update_call_formula}
\title{(Internal) Updates \code{spec@call}'s formula with the currently defined
variable names.}
\usage{
.update_call_formula(specification)
}
\arguments{
\item{specification}{A \code{StudySpecification}}
}
\value{
An updated formula
}
\description{
Helper function to update the \code{call} with the appropriate variable
names after they've been modified. Called within \code{StudySpecification}
replacers.
}
\details{
It's return should be stuck into the specification via
\code{spec@call$formula <- .update_call_formula(spec)}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{show,PreSandwichLayer-method}
\alias{show,PreSandwichLayer-method}
\title{Show a \code{PreSandwichLayer} or \code{SandwichLayer}}
\usage{
\S4method{show}{PreSandwichLayer}(object)
}
\arguments{
\item{object}{\code{PreSandwichLayer} or \code{SandwichLayer} object}
}
\value{
an invisible copy of \code{object}
}
\description{
Display information about a \code{PreSandwichLayer} or
\code{SandwichLayer} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.merge_block_id_cols}
\alias{.merge_block_id_cols}
\title{(Internal) Merge multiple block IDs}
\usage{
.merge_block_id_cols(df, ids)
}
\arguments{
\item{df}{a data frame}

\item{ids}{a vector of block IDs, column names of df}
}
\value{
a data frame with a column that contains unique block number IDs
}
\description{
(Internal) Merge multiple block IDs
}
\details{
merge multiple block ID columns by the value combinations and store
the new block ID in the column \code{ids[1]}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{estfun.teeMod}
\alias{estfun.teeMod}
\title{Extract empirical estimating equations from a \code{teeMod} model fit}
\usage{
\method{estfun}{teeMod}(x, values_from = x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{values_from}{optional, a fitted model object. Defaults to \code{x}.}

\item{...}{arguments passed to methods, most importantly those that define
the bias corrections for the residuals of \code{x} and, if applicable, a
\code{fitted_covariance_model} stored in its offset}
}
\value{
An \eqn{n\times k} matrix of empirical estimating equations for \code{x}.
\code{k} includes the model intercept, main effects of treatment and
moderator variables, any moderator effects, and marginal and conditional
means of the outcome (and \code{offset}, if provided) in the control condition.
See Details for definition of \eqn{n}.
}
\description{
An S3method for \code{sandwich::estfun} for producing a matrix
of contributions to the direct adjustment estimating equations.
}
\details{
If a prior covariance adjustment model has been passed to the
\code{offset} argument of the \code{teeMod} model using \code{cov_adj()},
\code{estfun.teeMod()} incorporates contributions to the estimating
equations of the covariance adjustment model.\cr\cr The covariance
adjustment sample may not fully overlap with the direct adjustment sample,
in which case \code{estfun.teeMod()} returns a matrix with the same number
of rows as the number of unique units of observation used to fit the two
models. Uniqueness is determined by matching units of assignment used to
fit the covariance adjustment model to units of assignment in the
\code{teeMod} model's \code{StudySpecification} slot; units of observation
within units of assignment that do not match are additional units that add
to the row count.\cr\cr The working residuals (and, if applicable, working
weights) from \code{values_from} will be used in the output matrix in
place of those from \code{x}. In other words, the contributions to the
the empirical estimating equations remain the same but are evaluated at
parameter estimates from \code{values_from} rather than \code{x}.\cr\cr The \code{by} argument in \code{cov_adj()} can
provide a column or a pair of columns (a named vector where the name
specifies a column in the direct adjustment sample and the value a column
in the covariance adjustment sample) that uniquely specifies units of
observation in each sample. This information can be used to align each
unit of observation's contributions to the two sets of estimating
equations. If no \code{by} argument is provided and units of observation
cannot be uniquely specified, contributions are aligned up to the unit of
assignment level. If standard errors are clustered no finer than that,
they will provide the same result as if each unit of observation's
contributions were aligned exactly.\cr\cr This method incorporates bias
corrections made to the residuals of \code{x} and, if applicable, the
covariance model stored in its \code{offset}. When its crossproduct is taken
(perhaps after suitable summing across rows within clusters), it provides
a heteroskedasticity- (or cluster-) robust estimate of the meat matrix of
the variance-covariance of the parameter estimates in \code{x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayer.R
\name{.show_layer}
\alias{.show_layer}
\title{(Internal) \code{show} helper for
\code{PreSandwichLayer}/\code{SandwichLayer}}
\usage{
.show_layer(object)
}
\arguments{
\item{object}{\code{PreSandwichLayer} or \code{SandwichLayer}}
}
\value{
\code{object}, invisibly
}
\description{
(Internal) \code{show} helper for
\code{PreSandwichLayer}/\code{SandwichLayer}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationAccessors.R
\name{.update_structure}
\alias{.update_structure}
\title{(Internal) Replaces \code{type} columns in \code{specification} with
\code{new}}
\usage{
.update_structure(specification, new, type)
}
\arguments{
\item{specification}{A \code{StudySpecification}}

\item{new}{A named \code{data.frame} with the replacement, should be the
output of \code{.convert_to_data.frame()}.}

\item{type}{One of "t", "f", "u" or "b\".}
}
\value{
The updated \code{StudySpecification}
}
\description{
Assumes \code{.convert_to_data.frame()} has already been called on
\code{new}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.align_and_extend_estfuns}
\alias{.align_and_extend_estfuns}
\title{(Internal) Align the dimensions and rows of direct adjustment and
covariance adjustment model estimating equations matrices}
\usage{
.align_and_extend_estfuns(
  x,
  values_from = x,
  ctrl_means_ef_mat = NULL,
  by = NULL,
  ...
)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{values_from}{optional, a fitted model object. Defaults to \code{x}.}

\item{ctrl_means_ef_mat}{optional, a matrix of estimating equations corresponding
to the estimates of the marginal (and possibly conditional) means of the outcome
and \code{offset} in the control condition. These are aligned and extended
in the same way as the matrix of estimating equations for \code{x} and
\code{cbind}ed to them}

\item{by}{optional, a character vector indicating columns that uniquely identify
rows in the dataframe used for fitting \code{x} and the dataframe passed to the
\code{data} argument of the covariance adjustment model fit. The default is
\code{NULL}, in which case the unit of assignment columns specified in the
\code{StudySpecification} slot of \code{x} are used.}

\item{...}{mostly arguments passed to methods, but the special case is the argument
\code{loco_residuals}, which indicates the offsets in the residuals of \code{x}
should be replaced by versions that use leave-one-cluster-out estimates of
the covariance model}
}
\value{
A list of two matrices, one being the aligned contributions to the
estimating equations for the direct adjustment model, and the other being
the aligned contributions to the covariance adjustment model.
}
\description{
(Internal) Align the dimensions and rows of direct adjustment and
covariance adjustment model estimating equations matrices
}
\details{
\code{.align_and_extend_estfuns()} first extracts the matrices of
contributions to the empirical estimating equations for the direct
adjustment and covariance adjustment models; then, it pads the matrices
with zeros to account for units of observation that appear in one
model-fitting sample but not the other; finally it orders the matrices so
units of observation (or if unit of observation-level ordering is
impossible, units of assignment) are aligned.\cr\cr As in
\code{estfun.teeMod()}, the working residuals (and, if applicable,
weights) from \code{values_from} will be used in the output matrix in
place of those from \code{x}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecification.R
\name{.new_StudySpecification}
\alias{.new_StudySpecification}
\title{(Internal) Create a new \code{StudySpecification} object.}
\usage{
.new_StudySpecification(
  form,
  data,
  type,
  subset = NULL,
  call = NULL,
  na.fail = TRUE,
  called_from_lmitt = FALSE
)
}
\arguments{
\item{form}{Formula to create StudySpecification, see help for
\code{rcr_spec()}, \code{rd_spec()} or \code{obs_spec()} for details.}

\item{data}{The data set}

\item{type}{One of "RCT", "RD", or "Obs"}

\item{subset}{Any subset information}

\item{call}{The call generating the \code{StudySpecification}.}

\item{na.fail}{Should it error on NA's (\code{TRUE}) or remove them
(\code{FALSE})?}

\item{called_from_lmitt}{Logical; was this called inside \code{lmitt()}, or
was it called from \code{*_spec()} (default).}
}
\value{
A new StudySpecification object
}
\description{
Helper function to create a new \code{StudySpecification}. Called internally
from \code{rct_spec()}, \code{rd_spec()} or \code{obs_spec()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{identical_StudySpecifications}
\alias{identical_StudySpecifications}
\title{Test equality of two \code{StudySpecification} objects}
\usage{
identical_StudySpecifications(x, y)
}
\arguments{
\item{x}{A \code{StudySpecification} object.}

\item{y}{A \code{StudySpecification} object.}
}
\value{
Logical, are \code{x} and \code{y} identical?
}
\description{
Check whether two \code{StudySpecification} objects are
identical.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{michigan_school_pairs}
\alias{michigan_school_pairs}
\title{Intervention data from a pair-matched study of schools in Michigan}
\format{
A \code{data.frame} with 14 rows and 13 columns.
\itemize{
\item{schoolid} school id
\item{blk} block
\item{z} treatment variable
\item{MALE_G11_PERC} percentage of G11 male students
\item{FEMALE_G11_PERC} percentage of G11 female students
\item{AM_G11_PERC} percentage of G11 American Indian/Alaska Native students
\item{ASIAN_G11_PERC} percentage of G11 Asian students
\item{HISP_G11_PERC} percentage of G11 Hispanic students
\item{BLACK_G11_PERC} percentage of G11 Black students
\item{WHITE_G11_PERC} percentage of G11 White students
\item{PACIFIC_G11_PERC} percentage of G11 Hawaiian Native/Pacific Islander students
\item{TR_G11_PERC} percentage of G11 Two or More Races students
\item{G11} Number of G11 students
}
}
\usage{
michigan_school_pairs
}
\description{
Michigan high schools, with a plausible cluster RCT
}
\details{
Grade 11 demographics for all Michigan high schools in 2013,
with mock block and treatment assignments for 14 high schools within
a large county in the metro Detroit area.
These schools were selected for this demonstration based on their
similarity to the 14 high schools from an adjacent Michigan
county that participated in the Pane et al (2013) study.  As a result,
they serve as an example of what one might expect to find as the
state-specific school-level subsample in a multi-state
paired cluster randomized trial featuring random assignment at
the school level.

The mock experimental schools were selected by optimal matching of
experimental schools to adjacent county schools, with substitute
schools grouped into the same pairs or triples (\sQuote{fine
strata}) as were their experimental counterparts.  The original
pairs and triples had been selected to reduce variation in baseline
variables predictive of outcomes, and the blocking structure the
substitute sample inherits may be expected to do this as well.  The
treatment/control distinction is also inherited from the
experimental sample, but there is of course no treatment effect
within the mock experiment.

The selection of mock experimental schools was based on both demographic
and student achievement variables, but the present data frame
includes only the demographic variables (as sourced from the
Common Core of Data [CCD; U.S. Department of Education]).  School
average outcomes in student test scores are available separately, from
Michigan's Center for Education Performance Information.  See the
vignette \sQuote{Real-data demonstration with a finely stratified
cluster RCT and a broader administrative database}, available on
the package website.
}
\examples{
data(michigan_school_pairs)
mi_spec <- rct_spec(z ~ uoa(schoolid)+block(blk),
data=michigan_school_pairs)
mi_spec
table(is.na(michigan_school_pairs$blk))
specification_table(mi_spec, "block", "treatment")
}
\references{
Pane, John F., et al.
"Effectiveness of cognitive tutor algebra I at scale."
\emph{Educational Evaluation and Policy Analysis} 36.2 (2014):
127-144.

U.S. Department of Education. Public Elementary/Secondary School
Universe Survey Data, v.2a. Institute of Education Sciences,
National Center for Education Statistics.
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{.update_form_to_unit_of_assignment}
\alias{.update_form_to_unit_of_assignment}
\title{(Internal) Rename cluster/unitid/uoa in a formula to
unit_of_assignment for internal consistency}
\usage{
.update_form_to_unit_of_assignment(form)
}
\arguments{
\item{form}{A formula passed to \code{.new_StudySpecification()}}
}
\value{
The formula with "cluster"/"unitid"/"uoa" replace with
"unit_of_assignment"
}
\description{
Internally, we always refer to uoa/cluster/unitid as "unit_of_assignment"
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.prepare_spec_matrix}
\alias{.prepare_spec_matrix}
\title{(Internal) Helper function for design-based meat matrix
calculation}
\usage{
.prepare_spec_matrix(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
a \eqn{m \times (p+2)} matrix of cluster sums of design-based
estimating equations scaled by \eqn{\sqrt{m_{b0}m_{b1}}/m_{b}}. Here
\eqn{m} is the number of clusters, \eqn{p} is the number of covariates used
in the prior covariance adjustment (excluding intercept)
}
\description{
(Internal) Helper function for design-based meat matrix
calculation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_DB_wo_covadj_se}
\alias{.get_DB_wo_covadj_se}
\title{(Internal) Design-based variance for models without
covariance adjustment}
\usage{
.get_DB_wo_covadj_se(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
design-based variance estimate of the main treatment effect
estimate
}
\description{
(Internal) Design-based variance for models without
covariance adjustment
}
\details{
Calculate bread matrix for design-based variance estimate for
\code{teeMod} models without covariance adjustment and without absorbed
effects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_data_frame.R
\name{.as_data_frame}
\alias{.as_data_frame}
\title{Convert object to \code{data.frame} or produce meaningful error}
\usage{
.as_data_frame(x)
}
\arguments{
\item{x}{An object}
}
\value{
\code{x} as a \code{data.frame}
}
\description{
Convert object to \code{data.frame} or produce meaningful error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SandwichLayerVariance.R
\name{.get_appinv_atp}
\alias{.get_appinv_atp}
\title{(Internal) Product of \eqn{A_{pp}^{-1} A_{\tau p}^T}}
\usage{
.get_appinv_atp(x, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}
}
\value{
An \eqn{s\times k} matrix \eqn{A_{pp}^{-1} A_{\tau p}^T}
}
\description{
(Internal) Product of \eqn{A_{pp}^{-1} A_{\tau p}^T}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.lmitt.R
\name{as.lmitt}
\alias{as.lmitt}
\alias{as.teeMod}
\title{Convert \code{lm} object into \code{teeMod}}
\usage{
as.lmitt(x, specification = NULL)

as.teeMod(x, specification = NULL)
}
\arguments{
\item{x}{\code{lm} object with weights containing a
\code{WeightedStudySpecification}, or an offset from \code{\link[=cov_adj]{cov_adj()}}.}

\item{specification}{Optional, explicitly specify the
\code{StudySpecification} to be used. If the \code{StudySpecification} is
specified elsewhere in \code{x} (e.g. passed as an argument to any of
\code{ate()}, \code{ett()}, \code{cov_adj()} or \code{assigned()}) it will
be found automatically and does not need to be passed here as well. If
different \code{StudySpecification} objects are passed (either through the
\code{lm} in weights or covariance adjustment, or through this argument),
an error will be produced.}
}
\value{
\code{teeMod} object
}
\description{
Converts the output of \code{\link[=lm]{lm()}} into a \code{teeMod}
object, for standard errors that account for block and cluster
information carried with the \code{lm}'s weights, and/or an
offset incorporating predictions of the outcome from a
separate model.
}
\details{
The formula with which \code{x} was created must include
a treatment identifier (e.g. \code{\link[=assigned]{assigned()}}).  If a model-based
offset is incorporated, the model's predictions would have to
have been extracted using \code{\link[=cov_adj]{cov_adj()}} as opposed to
\code{predict{}} in order for \code{teeMod} standard error
calculations to reflect propagation of error from these
predictions. This mechanism only supports treatment main effects:
to estimate interactions of treatment assignment with a moderator
variable, use \code{\link[=lmitt]{lmitt()}} instead of \code{lm()} followed by
\code{as.lmitt()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{.check_spec_formula}
\alias{.check_spec_formula}
\title{(Internal) Perform checks on formula for creation of
StudySpecification.}
\usage{
.check_spec_formula(form, allow_forcing = FALSE)
}
\arguments{
\item{form}{A formula passed to \code{.new_StudySpecification()}}

\item{allow_forcing}{Binary whether \code{forcing()} is allowed (\code{TRUE}
for RDD, \code{FALSE} for RCT and Obs).}
}
\value{
\code{TRUE} if all checks pass, otherwise errors.
}
\description{
Checks performed:
\itemize{
\item Ensure presence of no more than one of \code{unit_of_assignment()},
\code{cluster()} or \code{unitid()}.
\item Disallow multiple \code{block()} or multiple \code{forcing()} terms.
\item Disallow \code{forcing()} unless in RDD.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightedStudySpecification.R
\name{show,WeightedStudySpecification-method}
\alias{show,WeightedStudySpecification-method}
\title{Show a \code{WeightedStudySpecification}}
\usage{
\S4method{show}{WeightedStudySpecification}(object)
}
\arguments{
\item{object}{a \code{WeightedStudySpecification} object}
}
\value{
an invisible copy of \code{object}
}
\description{
Prints out the weights from a \code{WeightedStudySpecification}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudySpecificationUtilities.R
\name{unit_of_assignment}
\alias{unit_of_assignment}
\alias{unitid}
\alias{cluster}
\alias{uoa}
\alias{block}
\alias{forcing}
\title{Special terms in \code{StudySpecification} creation formula}
\usage{
unit_of_assignment(...)

unitid(...)

cluster(...)

uoa(...)

block(...)

forcing(...)
}
\arguments{
\item{...}{any number of variables of the same length.}
}
\value{
the variables with appropriate labels. No use outside of their
inclusion in the \code{formula} argument to \code{obs_spec},
\code{rct_spec}, or \code{rd_spec}
}
\description{
These are special functions used only in the definition of
\code{StudySpecification} objects. They identify the units of assignment,
blocks and forcing variables. They should never be used outside of the
\code{formula} argument to \code{obs_spec}, \code{rct_spec}, or
\code{rd_spec}.
}
\details{
These functions have no use outside of the formula in creating a
\code{StudySpecification}.

\code{unit_of_assignment}, \code{uoa}, \code{cluster} and \code{unitid}
are synonyms; you must include one and only one in each
\code{StudySpecification}. The choice of which to use will have no impact
on any analysis, only on some output and the name of the stored element in
the \code{StudySpecification}. Accessors/ replacers
(\code{units_of_assignment}, \code{unitids}, \code{clusters}) respect the
choice made at the point of creation of the \code{StudySpecification}, and
only the appropriate function will work.

See \code{rct_spec}, \code{obs_spec}, or \code{rd_spec} for examples of
their usage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.model.frame_tee.R
\name{.expand.model.frame_teeMod}
\alias{.expand.model.frame_teeMod}
\title{Add new variables to a model frame from a \code{teeMod} object}
\usage{
.expand.model.frame_teeMod(
  model,
  extras,
  envir = environment(formula(model)),
  na.expand = FALSE
)
}
\arguments{
\item{model}{A \code{teeMod} object}

\item{extras}{one-sided formula or vector of character strings describing
new variables to be added}

\item{envir}{an environment to evaluate things in}

\item{na.expand}{logical; see \code{stats::expand.model.frame} for details}
}
\value{
A \code{data.frame}
}
\description{
A variation of expand.model.frame which works for \code{teeMod}
objects
}
\details{
When building a \code{teeMod} object inside \code{lmitt()}, we do a lot of
manipulation of the variables involved in the model such that by the time
the \code{teeMod} is produced, neither the outcome nor predictors actually
fit in the model exist in the \code{data} passed into the call.

(E.g. to be specific, if a user calls \code{myda <- lmitt(y ~ 1, data =
mydata)}, then \code{model.frame(myda)} would contain column names not found
in \code{mydata}.)

This is a clone of \code{stats::expand.model.frame()} which has one addition
\itemize{
\item after extracting the \code{model$call$data} from \code{model}, it adds
columns from \code{model.frame(model)} to the object. This ensures that the
additional variables created during \code{lmitt()} can be found.
}

Trivial modifications from \code{stats::expand.model.frame()} include
ensuring \code{model} is a \code{teeMod} object, and using the \code{::}
syntax as appropriate.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.make_uoa_ids}
\alias{.make_uoa_ids}
\title{Make ID's to pass to the \code{cluster} argument of \code{vcov_tee()}}
\usage{
.make_uoa_ids(x, vcov_type, cluster = NULL, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} object}

\item{vcov_type}{a string indicating model-based or design-based
covariance estimation. Currently, "MB", "CR", and "HC" are the only strings
registered as indicating model-based estimation.}

\item{cluster}{character vector or list; optional. Specifies column names
that appear in both the covariance adjustment and direct adjustment model
dataframes. Defaults to NULL, in which case unit of assignment columns
indicated in the StudySpecification will be used for clustering. If there
are multiple clustering columns, they are concatenated together for each
row and separated by "_".}

\item{...}{arguments passed to methods}
}
\value{
A vector with length equal to the number of unique units of
observation used to fit the two models. See Details of
\code{estfun.teeMod()} for the method for determining uniqueness.
}
\description{
\code{.make_uoa_ids()} returns a factor vector of cluster ID's
that align with the order of the units of observations' contributions in
\code{estfun.teeMod()}. This is to ensure that when \code{vcov_tee()} calls
\code{sandwich::meatCL()}, the \code{cluster} argument aggregates the
correct contributions to estimating equations within clusters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights_internal.R
\name{.join_spec_weights}
\alias{.join_spec_weights}
\title{(Internal) Expand unit of assignment level weights to the level of
the data}
\usage{
.join_spec_weights(
  weights,
  specification,
  target,
  weightAlias,
  data,
  dichotomy
)
}
\arguments{
\item{weights}{a vector of weights sorted according to the
\code{StudySpecification}}

\item{specification}{a \code{StudySpecification}}

\item{target}{One of "ate" or "ett"}

\item{weightAlias}{Any currently supported alias}

\item{data}{New data}

\item{dichotomy}{formula used to specify a dichotomy of a non-binary
treatment variable. The output \code{WeightedStudySpecification} object
will store this as its \code{dichotomy} slot, unless it is NULL, in which
case it will be translated to an empty \code{formula}.}
}
\value{
a \code{WeightedStudySpecification}
}
\description{
Helper function called during creation of the weights via \code{ate()} or
\code{ett()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/teeMod.R
\name{.sanitize_Q_ids}
\alias{.sanitize_Q_ids}
\title{(Internal) Return ID's used to order observations in the direct
adjustment sample}
\usage{
.sanitize_Q_ids(x, id_col = NULL, ...)
}
\arguments{
\item{x}{a fitted \code{teeMod} model}

\item{id_col}{character vector; optional. Specifies column(s) whose ID's will
be returned. The column must exist in the data that created the
\code{StudySpecification} object. Default is NULL, in which case unit of
assignment columns indicated in the specification will be used to generate
ID's.}

\item{...}{arguments passed to methods}
}
\value{
A vector with length equal to the number of units of observation in
the direct adjustment sample
}
\description{
(Internal) Return ID's used to order observations in the direct
adjustment sample
}
\keyword{internal}
